/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.interfaces.simulation;

import com.net2plan.utils.Pair;
import com.net2plan.utils.StringUtils;
import java.awt.geom.Point2D;
import java.util.List;
import java.util.Map;

/**
 * <p>Provides a set of actions to be notified by event handlers. These actions are:</p>
 *
 * <ul>
 * <li>Layer/node/link/demand/route/protection segment/SRG added/modified/removed</li>
 * <li>Node/link up/down</li>
 * </ul>
 *
 * <p>Although the {@code SimAction} class is common for all actions, its meaning (i.e.
 * action type) depends on the static method used to get an instance. Basically, the action 
 * type is obtained through the {@link #getActionType() getActionType()} method, and then 
 * associated data can be retrieved from the {@code get()} methods corresponding to that action, 
 * for example {@code getDemandAdded_xxx()} for a {@code DEMAND_ADDED} action. Users are encouraged 
 * to take a look on the description of the static methods to obtain more information.</p>
 *
 * <p><b>Important</b>: Since Net2Plan 0.3.0, actions are assumed to be already 
 * applied by the algorithms to the network design. This class is used to notify actions 
 * between simulation modules (i.e. event generator and event processor).</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.1
 */
public class SimAction
{
	/**
	 * Type of action.
	 *
	 * @since 0.3.0
	 */
	public enum ActionType
	{
		/**
		 * A demand was added. Provided information includes layer and demand identifiers.
		 *
		 * @since 0.3.0
		 */
		DEMAND_ADDED,

		/**
		 * A demand was modified. Provided information includes layer and demand identifiers.
		 *
		 * @since 0.3.0
		 */
		DEMAND_MODIFIED,

		/**
		 * A demand was removed. Provided information includes layer and demand identifiers, 
		 * ingress and egress nodes, offered traffic, and attributes.
		 * 
		 * @since 0.3.0
		 */
		DEMAND_REMOVED,
		
		/**
		 * A forwarding rule was added. Provided information includes layer, demand and outgoing link identifiers.
		 *
		 * @since 0.3.0
		 */
		FORWARDING_RULE_ADDED,

		/**
		 * A forwarding rule was modified. Provided information includes layer, demand and outgoing link identifiers.
		 *
		 * @since 0.3.0
		 */
		FORWARDING_RULE_MODIFIED,

		/**
		 * A forwarding rule was removed. Provided information includes layer, demand and outgoing link identifiers, 
		 * splitting ratio, and attributes.
		 *
		 * @since 0.3.0
		 */
		FORWARDING_RULE_REMOVED,
		
		/**
		 * A layer was added. Provided information includes layer identifier.
		 * 
		 * @since 0.3.0
		 */
		LAYER_ADDED,

		/**
		 * A layer was modified. Provided information includes layer identifier.
		 * 
		 * @since 0.3.0
		 */
		LAYER_MODIFIED,

		/**
		 * A layer was removed. Provided information includes layer identifier, description, name, link capacity units name, demand traffic units name, and layer attributes.
		 * 
		 * @since 0.3.0
		 */
		LAYER_REMOVED,
		
		/**
		 * A node was added. Provided information includes node identifier.
		 * 
		 * @since 0.3.0
		 */
		NODE_ADDED,

		/**
		 * A node became down. Provided information includes node identifier.
		 * 
		 * @since 0.3.0
		 */
		NODE_DOWN,

		/**
		 * A node was modified. Provided information includes node identifier.
		 *
		 * @since 0.3.0
		 */
		NODE_MODIFIED,

		/**
		 * A node was removed. Provided information includes node identifier, coordinates, name and attributes.
		 *
		 * @since 0.3.0
		 */
		NODE_REMOVED,

		/**
		 * A node became up. Provided information includes node identifier.
		 * 
		 * @since 0.3.0
		 */
		NODE_UP,
		
		/**
		 * A link was added. Provided information includes layer and link identifiers.
		 * 
		 * @since 0.3.0
		 */
		LINK_ADDED,

		/**
		 * A link became down. Provided information includes layer and link identifiers.
		 * 
		 * @since 0.3.0
		 */
		LINK_DOWN,

		/**
		 * A link was modified. Provided information includes layer and link identifiers.
		 * 
		 * @since 0.3.0
		 */
		LINK_MODIFIED,

		/**
		 * A link was removed. Provided information includes layer and link identifiers, 
		 * origin and destination nodes, capacity, length in km, propagation speed in km 
		 * per second, and attributes.
		 * 
		 * @since 0.3.0
		 */
		LINK_REMOVED,

		/**
		 * A link became up. Provided information includes layer and link identifiers.
		 * 
		 * @since 0.3.0
		 */
		LINK_UP,
		
		/**
		 * A protection segment was added. Provided information includes layer and protection segment identifiers.
		 *
		 * @since 0.3.0
		 */
		PROTECTION_SEGMENT_ADDED,

		/**
		 * A protection segment was modified. Provided information includes layer and protection segment identifiers.
		 *
		 * @since 0.3.0
		 */
		PROTECTION_SEGMENT_MODIFIED,

		/**
		 * A protection segment was removed. Provided information includes layer and protection segment identifiers, 
		 * origin and destination nodes, sequence of links, reserved capacity, and attributes.
		 * 
		 * @since 0.3.0
		 */
		PROTECTION_SEGMENT_REMOVED,

		/**
		 * A route was added. Provided information includes layer and route identifiers.
		 *
		 * @since 0.3.0
		 */
		ROUTE_ADDED,

		/**
		 * A route was modified. Provided information includes layer and route identifiers.
		 *
		 * @since 0.3.0
		 */
		ROUTE_MODIFIED,

		/**
		 * A route was removed. Provided information includes layer and route identifiers, 
		 * demand identifier, ingress and egress nodes, carried traffic, occupied capacity, 
		 * current sequence of links, backup segment list, and attributes.
		 * 
		 * @since 0.3.0
		 */
		ROUTE_REMOVED,
		
		/**
		 * An SRG was added. Provided information includes SRG identifier.
		 *
		 * @since 0.3.0
		 */
		SRG_ADDED,

		/**
		 * An SRG was modified. Provided information includes SRG identifier.
		 *
		 * @since 0.3.0
		 */
		SRG_MODIFIED,

		/**
		 * An SRG was removed. Provided information includes SRG identifier, 
		 * MTTR and MTTF, and attributes.
		 * 
		 * @since 0.3.0
		 */
		SRG_REMOVED
	};
	
	private final ActionType type;
	
	private long layerAdded_layerId;
	private long layerModified_layerId;
	private long layerRemoved_layerId;
	private String layerRemoved_demandTrafficUnitsName;
	private String layerRemoved_linkCapacityUnitsName;
	private String layerRemoved_description;
	private String layerRemoved_name;
	private Map<String, String> layerRemoved_attributes;

	private long nodeAdded_nodeId;
	private long nodeDown_nodeId;
	private long nodeModified_nodeId;
	private long nodeRemoved_nodeId;
	private Point2D nodeRemoved_xyPosition;
	private String nodeRemoved_name;
	private Map<String, String> nodeRemoved_attributes;
	private long nodeUp_nodeId;
	
	private long linkAdded_layerId;
	private long linkAdded_linkId;
	private long linkDown_layerId;
	private long linkDown_linkId;
	private long linkModified_layerId;
	private long linkModified_linkId;
	private long linkRemoved_layerId;
	private long linkRemoved_linkId;
	private long linkRemoved_originNodeId;
	private long linkRemoved_destinationNodeId;
	private double linkRemoved_capacity;
	private double linkRemoved_lengthInKm;
	private double linkRemoved_propagationSpeedInKmPerSecond;
	private Map<String, String> linkRemoved_attributes;
	private long linkUp_layerId;
	private long linkUp_linkId;

	private long demandAdded_layerId;
	private long demandAdded_demandId;
	private long demandModified_layerId;
	private long demandModified_demandId;
	private long demandRemoved_layerId;
	private long demandRemoved_demandId;
	private long demandRemoved_ingressNodeId;
	private long demandRemoved_egressNodeId;
	private double demandRemoved_offeredTraffic;
	private Map<String, String> demandRemoved_attributes;

	private long routeAdded_layerId;
	private long routeAdded_routeId;
	private long routeModified_layerId;
	private long routeModified_routeId;
	private long routeRemoved_layerId;
	private long routeRemoved_routeId;
	private long routeRemoved_demandId;
	private long routeRemoved_ingressNodeId;
	private long routeRemoved_egressNodeId;
	private double routeRemoved_carriedTraffic;
	private double routeRemoved_occupiedCapacity;
	private List<Long> routeRemoved_seqLinks;
	private List<Long> routeRemoved_backupSegmentList;
	private Map<String, String> routeRemoved_attributes;

	private long forwardingRuleAdded_layerId;
	private Pair<Long, Long> forwardingRuleAdded_demandLinkPair;
	private long forwardingRuleModified_layerId;
	private Pair<Long, Long> forwardingRuleModified_demandLinkPair;
	private long forwardingRuleRemoved_layerId;
	private Pair<Long, Long> forwardingRuleRemoved_demandLinkPair;
	private double forwardingRuleRemoved_splittingRatio;
	private Map<String, String> forwardingRuleRemoved_attributes;

	private long segmentAdded_layerId;
	private long segmentAdded_segmentId;
	private long segmentModified_layerId;
	private long segmentModified_segmentId;
	private long segmentRemoved_layerId;
	private long segmentRemoved_segmentId;
	private long segmentRemoved_originNodeId;
	private long segmentRemoved_destinationNodeId;
	private double segmentRemoved_reservedCapacity;
	private List<Long> segmentRemoved_seqLinks;
	private Map<String, String> segmentRemoved_attributes;
	
	private long srgAdded_srgId;
	private long srgModified_srgId;
	private long srgRemoved_srgId;
	private double srgRemoved_mttfInHours;
	private double srgRemoved_mttrInHours;
	private Map<String, String> srgRemoved_attributes;
	
	private SimAction(ActionType type)
	{
		this.type = type;
		reset();
	}

	/**
	 * Returns the action type.
	 *
	 * @return Action type
	 * @since 0.2.1
	 */
	public SimAction.ActionType getActionType()
	{
		return type;
	}
	
	/**
	 * Returns the demand identifier associated to a {@link ActionType#DEMAND_ADDED DEMAND_ADDED} action.
	 * 
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long getDemandAdded_demandId() { return demandAdded_demandId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#DEMAND_ADDED DEMAND_ADDED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getDemandAdded_layerId() { return demandAdded_layerId; }
	
	/**
	 * Returns the demand identifier associated to a {@link ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action.
	 * 
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long getDemandModified_demandId() { return demandModified_demandId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getDemandModified_layerId() { return demandModified_layerId; }
	
	/**
	 * Returns the demand attributes associated to a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
	 * 
	 * @return Demand attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getDemandRemoved_attributes() { return demandRemoved_attributes; }
	
	/**
	 * Returns the demand identifier associated to a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
	 * 
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long getDemandRemoved_demandId() { return demandRemoved_demandId; }
	
	/**
	 * Returns the egress node identifier associated to a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
	 * 
	 * @return Egress node identifier
	 * @since 0.3.0
	 */
	public long getDemandRemoved_egressNodeId() { return demandRemoved_egressNodeId; }
	
	/**
	 * Returns the ingress node identifier associated to a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
	 * 
	 * @return Ingress node identifier
	 * @since 0.3.0
	 */
	public long getDemandRemoved_ingressNodeId() { return demandRemoved_ingressNodeId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getDemandRemoved_layerId() { return demandRemoved_layerId; }
	
	/**
	 * Returns the offered traffic associated to a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
	 * 
	 * @return Offered traffic
	 * @since 0.3.0
	 */
	public double getDemandRemoved_offeredTraffic() { return demandRemoved_offeredTraffic; }
	
	/**
	 * Returns the demand and outgoing link identifiers associated to a {@link ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} action.
	 *
	 * @return Demand-outgoing link identifier pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getForwardingRuleAdded_demandLinkPair() { return forwardingRuleAdded_demandLinkPair; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} action.

	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getForwardingRuleAdded_layerId() { return forwardingRuleAdded_layerId; }
	
	/**
	 * Returns the demand and outgoing link identifiers associated to a {@link ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action.
	 *
	 * @return Demand-outgoing link identifier pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getForwardingRuleModified_demandLinkPair() { return forwardingRuleModified_demandLinkPair; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getForwardingRuleModified_layerId() { return forwardingRuleModified_layerId; }
	
	/**
	 * Returns the attributes associated to a {@link ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action.
	 *
	 * @return Forwarding rule attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getForwardingRuleRemoved_attributes() { return forwardingRuleRemoved_attributes; }

	/**
	 * Returns the demand and outgoing link identifiers associated to a {@link ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action.
	 *
	 * @return Demand-outgoing link identifier pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getForwardingRuleRemoved_demandLinkPair() { return forwardingRuleRemoved_demandLinkPair; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getForwardingRuleRemoved_layerId() { return forwardingRuleRemoved_layerId; }
	
	/**
	 * Returns the splitting ratio associated to a {@link ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action.
	 *
	 * @return Splitting ratio
	 * @since 0.3.0
	 */
	public double getForwardingRuleRemoved_splittingRatio() { return forwardingRuleRemoved_splittingRatio; }

	/**
	 * Returns the layer identifier associated to a {@link ActionType#LAYER_ADDED LAYER_ADDED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLayerAdded_layerId() { return layerAdded_layerId; }

	/**
	 * Returns the layer identifier associated to a {@link ActionType#LAYER_MODIFIED LAYER_MODIFIED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLayerModified_layerId() { return layerModified_layerId; }
	
	/**
	 * Returns the layer attributes associated to a {@link ActionType#LAYER_REMOVED LAYER_REMOVED} action.
	 * 
	 * @return Layer attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getLayerRemoved_attributes() { return layerRemoved_attributes; }
	
	/**
	 * Returns the demand traffic units name associated to a {@link ActionType#LAYER_REMOVED LAYER_REMOVED} action.
	 *
	 * @return Demand traffic units name
	 * @since 0.3.0
	 */
	public String getLayerRemoved_demandTrafficUnitsName() { return layerRemoved_demandTrafficUnitsName; }
	
	/**
	 * Returns the layer description associated to a {@link ActionType#LAYER_REMOVED LAYER_REMOVED} action.
	 * 
	 * @return Layer description
	 * @since 0.3.0
	 */
	public String getLayerRemoved_description() { return layerRemoved_description; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#LAYER_REMOVED LAYER_REMOVED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLayerRemoved_layerId() { return layerRemoved_layerId; }
	
	/**
	 * Returns the link capacity units name associated to a {@link ActionType#LAYER_REMOVED LAYER_REMOVED} action.
	 *
	 * @return Link capacity units name
	 * @since 0.3.0
	 */
	public String getLayerRemoved_linkCapacityUnitsName() { return layerRemoved_linkCapacityUnitsName; }
	
	/**
	 * Returns the layer name associated to a {@link ActionType#LAYER_REMOVED LAYER_REMOVED} action.
	 *
	 * @return Layer name
	 * @since 0.3.0
	 */
	public String getLayerRemoved_name() { return layerRemoved_name; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#LINK_ADDED LINK_ADDED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLinkAdded_layerId() { return linkAdded_layerId; }
	
	/**
	 * Returns the link identifier associated to a {@link ActionType#LINK_ADDED LINK_ADDED} action.
	 * 
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long getLinkAdded_linkId() { return linkAdded_linkId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#LINK_DOWN LINK_DOWN} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLinkDown_layerId() { return linkDown_layerId; }

	/**
	 * Returns the link identifier associated to a {@link ActionType#LINK_DOWN LINK_DOWN} action.
	 * 
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long getLinkDown_linkId() { return linkDown_linkId; }

	/**
	 * Returns the layer identifier associated to a {@link ActionType#LINK_MODIFIED LINK_MODIFIED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLinkModified_layerId() { return linkModified_layerId; }
	
	/**
	 * Returns the link identifier associated to a {@link ActionType#LINK_MODIFIED LINK_MODIFIED} action.
	 * 
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long getLinkModified_linkId() { return linkModified_linkId; }
	
	/**
	 * Returns the attributes associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Link attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getLinkRemoved_attributes() { return linkRemoved_attributes; }

	/**
	 * Returns the capacity associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Link capacity
	 * @since 0.3.0
	 */
	public double getLinkRemoved_capacity() { return linkRemoved_capacity; }
	
	/**
	 * Returns the destination node identifier associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Destination node identifier
	 * @since 0.3.0
	 */
	public long getLinkRemoved_destinationNodeId() { return linkRemoved_destinationNodeId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLinkRemoved_layerId() { return linkRemoved_layerId; }
	
	/**
	 * Returns the length in km associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Link length in km
	 * @since 0.3.0
	 */
	public double getLinkRemoved_lengthInKm() { return linkRemoved_lengthInKm; }
	
	/**
	 * Returns the link identifier associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long getLinkRemoved_linkId() { return linkRemoved_linkId; }
	
	/**
	 * Returns the origin node identifier associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Origin node identifier
	 * @since 0.3.0
	 */
	public long getLinkRemoved_originNodeId() { return linkRemoved_originNodeId; }
	
	/**
	 * Returns the propagation speed in km per second associated to a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
	 * 
	 * @return Link propagation speed in km per second
	 * @since 0.3.0
	 */
	public double getLinkRemoved_propagationSpeedInKmPerSecond() { return linkRemoved_propagationSpeedInKmPerSecond; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#LINK_UP LINK_UP} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLinkUp_layerId() { return linkUp_layerId; }

	/**
	 * Returns the link identifier associated to a {@link ActionType#LINK_UP LINK_UP} action.
	 * 
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long getLinkUp_linkId() { return linkUp_linkId; }

	/**
	 * Returns the node identifier associated to a {@link ActionType#NODE_ADDED NODE_ADDED} action.
	 * 
	 * @return Node identifier
	 * @since 0.3.0
	 */
	public long getNodeAdded_nodeId() { return nodeAdded_nodeId; }
	
	/**
	 * Returns the node identifier associated to a {@link ActionType#NODE_DOWN NODE_DOWN} action.
	 * 
	 * @return Node identifier
	 * @since 0.3.0
	 */
	public long getNodeDown_nodeId() { return nodeDown_nodeId; }

	/**
	 * Returns the node identifier associated to a {@link ActionType#NODE_MODIFIED NODE_MODIFIED} action.
	 * 
	 * @return Node identifier
	 * @since 0.3.0
	 */
	public long getNodeModified_nodeId() { return nodeModified_nodeId; }

	/**
	 * Returns the attributes associated to a {@link ActionType#NODE_REMOVED NODE_REMOVED} action.
	 * 
	 * @return Node attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getNodeRemoved_attributes() { return nodeRemoved_attributes; }
	
	/**
	 * Returns the name associated to a {@link ActionType#NODE_REMOVED NODE_REMOVED} action.
	 * 
	 * @return Node name
	 * @since 0.3.0
	 */
	public String getNodeRemoved_name() { return nodeRemoved_name; }
	
	/**
	 * Returns the node identifier associated to a {@link ActionType#NODE_REMOVED NODE_REMOVED} action.
	 * 
	 * @return Node identifier
	 * @since 0.3.0
	 */
	public long getNodeRemoved_nodeId() { return nodeRemoved_nodeId; }

	/**
	 * Returns the coordinates associated to a {@link ActionType#NODE_REMOVED NODE_REMOVED} action.
	 * 
	 * @return Node position
	 * @since 0.3.0
	 */
	public Point2D getNodeRemoved_xyPosition() { return nodeRemoved_xyPosition; }

	/**
	 * Returns the node identifier associated to a {@link ActionType#NODE_UP NODE_UP} action.
	 * 
	 * @return Node identifier
	 * @since 0.3.0
	 */
	public long getNodeUp_nodeId() { return nodeUp_nodeId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#PROTECTION_SEGMENT_ADDED PROTECTION_SEGMENT_ADDED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentAdded_layerId() { return segmentAdded_layerId; }
	
	/**
	 * Returns the protection segment identifier associated to a {@link ActionType#PROTECTION_SEGMENT_ADDED PROTECTION_SEGMENT_ADDED} action.
	 * 
	 * @return Protection segment identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentAdded_segmentId() { return segmentAdded_segmentId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentModified_layerId() { return segmentModified_layerId; }
	
	/**
	 * Returns the protection segment identifier associated to a {@link ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action.
	 * 
	 * @return Protection segment identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentModified_segmentId() { return segmentModified_segmentId; }
	
	/**
	 * Returns the attributes associated to a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
	 * 
	 * @return Protection segment attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getProtectionSegmentRemoved_attributes() { return segmentRemoved_attributes; }
	
	/**
	 * Returns the destination node identifier associated to a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
	 * 
	 * @return Destination node identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentRemoved_destinationNodeId() { return segmentRemoved_destinationNodeId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentRemoved_layerId() { return segmentRemoved_layerId; }
	
	/**
	 * Returns the origin node identifier associated to a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
	 * 
	 * @return Origin node identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentRemoved_originNodeId() { return segmentRemoved_originNodeId; }
	
	/**
	 * Returns the reserved capacity associated to a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
	 * 
	 * @return Reserved capacity
	 * @since 0.3.0
	 */
	public double getProtectionSegmentRemoved_reservedCapacity() { return segmentRemoved_reservedCapacity; }
	
	/**
	 * Returns the protection segment identifier associated to a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
	 * 
	 * @return Protection segment identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentRemoved_segmentId() { return segmentRemoved_segmentId; }
	
	/**
	 * Returns the sequence of links associated to a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
	 * 
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> getProtectionSegmentRemoved_sequenceOfLinks() { return segmentRemoved_seqLinks; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#ROUTE_ADDED ROUTE_ADDED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getRouteAdded_layerId() { return routeAdded_layerId; }
	
	/**
	 * Returns the route identifier associated to a {@link ActionType#ROUTE_ADDED ROUTE_ADDED} action.
	 * 
	 * @return Route identifier
	 * @since 0.3.0
	 */
	public long getRouteAdded_routeId() { return routeAdded_routeId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getRouteModified_layerId() { return routeModified_layerId; }
	
	/**
	 * Returns the route identifier associated to a {@link ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action.
	 * 
	 * @return Route identifier
	 * @since 0.3.0
	 */
	public long getRouteModified_routeId() { return routeModified_routeId; }
	
	/**
	 * Returns the attributes associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Route attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getRouteRemoved_attributes() { return routeRemoved_attributes; }
	
	/**
	 * Returns the backup segment list associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Route backup segment list
	 * @since 0.3.0
	 */
	public List<Long> getRouteRemoved_backupSegmentList() { return routeRemoved_backupSegmentList; }
	
	/**
	 * Returns the carried traffic associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Route carried traffic
	 * @since 0.3.0
	 */
	public double getRouteRemoved_carriedTraffic() { return routeRemoved_carriedTraffic; }
	
	/**
	 * Returns the demand identifier associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long getRouteRemoved_demandId() { return routeRemoved_demandId; }
	
	/**
	 * Returns the egress node identifier associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Egress node identifier
	 * @since 0.3.0
	 */
	public long getRouteRemoved_egressNodeId() { return routeRemoved_egressNodeId; }
	
	/**
	 * Returns the ingress node identifier associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Ingress node identifier
	 * @since 0.3.0
	 */
	public long getRouteRemoved_ingressNodeId() { return routeRemoved_ingressNodeId; }
	
	/**
	 * Returns the layer identifier associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getRouteRemoved_layerId() { return routeRemoved_layerId; }
	
	/**
	 * Returns the occupied capacity associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Route occupied capacity
	 * @since 0.3.0
	 */
	public double getRouteRemoved_occupiedCapacity() { return routeRemoved_occupiedCapacity; }

	/**
	 * Returns the route identifier associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Route identifier
	 * @since 0.3.0
	 */
	public long getRouteRemoved_routeId() { return routeRemoved_routeId; }
	
	/**
	 * Returns the sequence of links associated to a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
	 * 
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> getRouteRemoved_sequenceOfLinks() { return routeRemoved_seqLinks; }
	
	/**
	 * Returns the SRG identifier associated to a {@link ActionType#SRG_ADDED SRG_ADDED} action.
	 * 
	 * @return SRG identifier
	 * @since 0.3.0
	 */
	public long getSRGAdded_srgId() { return srgAdded_srgId; }
	
	/**
	 * Returns the SRG identifier associated to a {@link ActionType#SRG_MODIFIED SRG_MODIFIED} action.
	 * 
	 * @return SRG identifier
	 * @since 0.3.0
	 */
	public long getSRGModified_srgId() { return srgModified_srgId; }

	/**
	 * Returns the attributes associated to a {@link ActionType#SRG_REMOVED SRG_REMOVED} action.
	 * 
	 * @return SRG attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getSRGRemoved_attributes() { return srgRemoved_attributes; }
	
	/**
	 * Returns the mean time to fail associated to a {@link ActionType#SRG_REMOVED SRG_REMOVED} action.
	 * 
	 * @return Mean time to fail
	 * @since 0.3.0
	 */
	public double getSRGRemoved_meanTimeToFailInHours() { return srgRemoved_mttfInHours; }
	
	/**
	 * Returns the mean time to repair associated to a {@link ActionType#SRG_REMOVED SRG_REMOVED} action.
	 * 
	 * @return Mean time to repair
	 * @since 0.3.0
	 */
	public double getSRGRemoved_meanTimeToRepairInHours() { return srgRemoved_mttrInHours; }
	
	/**
	 * Returns the SRG identifier associated to a {@link ActionType#SRG_REMOVED SRG_REMOVED} action.
	 * 
	 * @return SRG identifier
	 * @since 0.3.0
	 */
	public long getSRGRemoved_srgId() { return srgRemoved_srgId; }

	private void reset()
	{
		layerAdded_layerId = -1;
		layerModified_layerId = -1;
		layerRemoved_layerId = -1;
		layerRemoved_demandTrafficUnitsName = null;
		layerRemoved_linkCapacityUnitsName = null;
		layerRemoved_description = null;
		layerRemoved_name = null;
		layerRemoved_attributes = null;

		nodeAdded_nodeId = -1;
		nodeDown_nodeId = -1;
		nodeModified_nodeId = -1;
		nodeRemoved_nodeId = -1;
		nodeRemoved_xyPosition = null;
		nodeRemoved_name = null;
		nodeRemoved_attributes = null;
		nodeUp_nodeId = -1;

		linkAdded_layerId = -1;
		linkAdded_linkId = -1;
		linkDown_layerId = -1;
		linkDown_linkId = -1;
		linkModified_layerId = -1;
		linkModified_linkId = -1;
		linkRemoved_layerId = -1;
		linkRemoved_linkId = -1;
		linkRemoved_originNodeId = -1;
		linkRemoved_destinationNodeId = -1;
		linkRemoved_capacity = -1;
		linkRemoved_lengthInKm = -1;
		linkRemoved_propagationSpeedInKmPerSecond = -1;
		linkRemoved_attributes = null;
		linkUp_layerId = -1;
		linkUp_linkId = -1;

		demandAdded_layerId = -1;
		demandAdded_demandId = -1;
		demandModified_layerId = -1;
		demandModified_demandId = -1;
		demandRemoved_layerId = -1;
		demandRemoved_demandId = -1;
		demandRemoved_ingressNodeId = -1;
		demandRemoved_egressNodeId = -1;
		demandRemoved_offeredTraffic = -1;
		demandRemoved_attributes = null;

		routeAdded_layerId = -1;
		routeAdded_routeId = -1;
		routeModified_layerId = -1;
		routeModified_routeId = -1;
		routeRemoved_layerId = -1;
		routeRemoved_routeId = -1;
		routeRemoved_demandId = -1;
		routeRemoved_ingressNodeId = -1;
		routeRemoved_egressNodeId = -1;
		routeRemoved_carriedTraffic = -1;
		routeRemoved_occupiedCapacity = -1;
		routeRemoved_seqLinks = null;
		routeRemoved_backupSegmentList = null;
		routeRemoved_attributes = null;

		forwardingRuleAdded_layerId = -1;
		forwardingRuleAdded_demandLinkPair = Pair.unmodifiableOf(-1L, -1L);
		forwardingRuleModified_layerId = -1;
		forwardingRuleModified_demandLinkPair = Pair.unmodifiableOf(-1L, -1L);
		forwardingRuleRemoved_layerId = -1;
		forwardingRuleRemoved_demandLinkPair = Pair.unmodifiableOf(-1L, -1L);
		forwardingRuleRemoved_splittingRatio = -1;
		forwardingRuleRemoved_attributes = null;
		
		segmentAdded_layerId = -1;
		segmentAdded_segmentId = -1;
		segmentModified_layerId = -1;
		segmentModified_segmentId = -1;
		segmentRemoved_layerId = -1;
		segmentRemoved_segmentId = -1;
		segmentRemoved_originNodeId = -1;
		segmentRemoved_destinationNodeId = -1;
		segmentRemoved_reservedCapacity = -1;
		segmentRemoved_seqLinks = null;
		segmentRemoved_attributes = null;

		srgAdded_srgId = -1;
		srgModified_srgId = -1;
		srgRemoved_srgId = -1;
		srgRemoved_mttfInHours = -1;
		srgRemoved_mttrInHours = -1;
		srgRemoved_attributes = null;
	}
	
	/**
	 * Returns a {@code String} representation of the object.
	 *
	 * @return {@code String} representation of the object
	 * @since 0.2.1
	 */
	@Override
	public String toString()
	{
		switch (getActionType())
		{
			case DEMAND_ADDED:
				return String.format("Demand %d added (layer = %d)", demandAdded_demandId, demandAdded_layerId);

			case DEMAND_MODIFIED:
				return String.format("Demand %d modified (layer = %d)", demandModified_demandId, demandModified_layerId);

			case DEMAND_REMOVED:
				return String.format("Demand %d removed (layer = %d, ingress node = %d, egress node = %d, offered traffic = %f, attributes = %s)", demandRemoved_demandId, demandRemoved_layerId, demandRemoved_ingressNodeId, demandRemoved_egressNodeId, demandRemoved_offeredTraffic, demandRemoved_attributes == null || demandRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(demandRemoved_attributes));
				
			case FORWARDING_RULE_ADDED:
				return String.format("Forwarding rule [d%d, e%d] added (layer = %d)", forwardingRuleAdded_demandLinkPair.getFirst(), forwardingRuleAdded_demandLinkPair.getSecond(), forwardingRuleAdded_layerId);

			case FORWARDING_RULE_MODIFIED:
				return String.format("Forwarding rule [d%d, e%d] modified (layer = %d)", forwardingRuleModified_demandLinkPair.getFirst(), forwardingRuleModified_demandLinkPair.getSecond(), forwardingRuleModified_layerId);

			case FORWARDING_RULE_REMOVED:
				return String.format("Forwarding rule [d%d, e%d] removed (layer = %d, splitting ratio = %f, attributes = %s)", forwardingRuleRemoved_demandLinkPair.getFirst(), forwardingRuleRemoved_demandLinkPair.getSecond(), forwardingRuleRemoved_layerId, forwardingRuleRemoved_splittingRatio, forwardingRuleRemoved_attributes == null || forwardingRuleRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(forwardingRuleRemoved_attributes));
				
			case LAYER_ADDED:
				return String.format("Layer %d added", layerAdded_layerId);
				
			case LAYER_MODIFIED:
				return String.format("Layer %d modified", layerModified_layerId);
				
			case LAYER_REMOVED:
				return String.format("Layer %d removed (description = %s, name = %s, link capacity units name = %s, demand traffic units = %s, attributes = %s)", layerRemoved_layerId, layerRemoved_description, layerRemoved_name, layerRemoved_linkCapacityUnitsName, layerRemoved_demandTrafficUnitsName, layerRemoved_attributes == null || layerRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(layerRemoved_attributes));
				
			case LINK_ADDED:
				return String.format("Link %d added (layer = %d)", linkAdded_linkId, linkAdded_layerId);

			case LINK_DOWN:
				return String.format("Link %d down (layer = %d)", linkDown_linkId, linkDown_layerId);
				
			case LINK_MODIFIED:
				return String.format("Link %d modified (layer = %d)", linkModified_linkId, linkModified_layerId);

			case LINK_REMOVED:
				return String.format("Link %d removed (layer = %d, origin node = %d, destination node = %d, capacity = %f, lengthInKm = %f, propagationSpeedInKmPerSecond = %f, attributes = %s)", linkRemoved_linkId, linkRemoved_layerId, linkRemoved_originNodeId, linkRemoved_destinationNodeId, linkRemoved_capacity, linkRemoved_lengthInKm, linkRemoved_propagationSpeedInKmPerSecond, linkRemoved_attributes == null || linkRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(linkRemoved_attributes));
				
			case LINK_UP:
				return String.format("Link %d up (layer = %d)", linkUp_linkId, linkUp_layerId);
				
			case NODE_ADDED:
				return String.format("Node %d added", nodeAdded_nodeId);
				
			case NODE_DOWN:
				return String.format("Node %d down", nodeDown_nodeId);
				
			case NODE_MODIFIED:
				return String.format("Node %d modified", nodeModified_nodeId);
				
			case NODE_REMOVED:
				return String.format("Node %d removed (position = %s, name = %s, attributes = %s)", nodeRemoved_nodeId, nodeRemoved_xyPosition, nodeRemoved_name, nodeRemoved_attributes == null || nodeRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(nodeRemoved_attributes));
				
			case NODE_UP:
				return String.format("Node %d up", nodeUp_nodeId);
				
			case PROTECTION_SEGMENT_ADDED:
				return String.format("Protection segment %d added (layer = %d)", segmentAdded_segmentId, segmentAdded_layerId);

			case PROTECTION_SEGMENT_MODIFIED:
				return String.format("Protection segment %d modified (layer = %d)", segmentModified_segmentId, segmentModified_layerId);

			case PROTECTION_SEGMENT_REMOVED:
				return String.format("Protection segment %d removed (layer = %d, origin node = %d, destination node = %d, reserved capacity = %f, seq. links = %s, attributes = %s)", segmentRemoved_segmentId, segmentRemoved_layerId, segmentRemoved_originNodeId, segmentRemoved_destinationNodeId, segmentRemoved_reservedCapacity, segmentRemoved_seqLinks, segmentRemoved_attributes == null || segmentRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(segmentRemoved_attributes));

			case ROUTE_ADDED:
				return String.format("Route %d added (layer = %d)", routeAdded_routeId, routeAdded_layerId);

			case ROUTE_MODIFIED:
				return String.format("Route %d modified (layer = %d)", routeModified_routeId, routeModified_layerId);
				
			case ROUTE_REMOVED:
				return String.format("Route %d removed (layer = %d, demand = %d, ingress node = %d, egress node = %d, carried traffic = %f, occupied capacity = %f, seq. links = %s, backup segment list = %s, attributes = %s)", routeRemoved_routeId, routeRemoved_layerId, routeRemoved_demandId, routeRemoved_ingressNodeId, routeRemoved_egressNodeId, routeRemoved_carriedTraffic, routeRemoved_occupiedCapacity, routeRemoved_seqLinks, routeRemoved_backupSegmentList, routeRemoved_attributes == null || routeRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(routeRemoved_attributes));

			case SRG_ADDED:
				return String.format("SRG %d added", srgAdded_srgId);
				
			case SRG_MODIFIED:
				return String.format("SRG %d modified", srgModified_srgId);
				
			case SRG_REMOVED:
				return String.format("SRG %d removed (mttf = %f, mttr = %f, attributes = %s)", srgRemoved_srgId, srgRemoved_mttfInHours, srgRemoved_mttrInHours, srgRemoved_attributes == null || srgRemoved_attributes.isEmpty() ? "none" : StringUtils.mapToString(srgRemoved_attributes));
				
			default:
				throw new RuntimeException("Bad");
		}
	}
	
	/**
	 * Factory methods to create actions. In general it should not be used by users, 
	 * {@link com.net2plan.interfaces.networkDesign.NetPlan NetPlan} class is able 
	 * to collect state changes using the {@link com.net2plan.interfaces.networkDesign.NetPlan#activateActionMonitoring() activateActionMonitoring} 
	 * method.
	 * 
	 * @since 0.3.0
	 */
	public static class Factory
	{
		/**
		 * Creates a {@link ActionType#DEMAND_ADDED DEMAND_ADDED} action.
		 * 
		 * @param demandId Demand identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction demandAdded(long demandId)
		{
			return demandAdded(-1, demandId);
		}

		/**
		 * Creates a {@link ActionType#DEMAND_ADDED DEMAND_ADDED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param demandId Demand identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction demandAdded(long layerId, long demandId)
		{
			SimAction simAction = new SimAction(ActionType.DEMAND_ADDED);
			simAction.demandAdded_layerId = layerId;
			simAction.demandAdded_demandId = demandId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action.
		 * 
		 * @param demandId Demand identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction demandModified(long demandId)
		{
			return demandModified(-1, demandId);
		}

		/**
		 * Creates a {@link ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param demandId Demand identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction demandModified(long layerId, long demandId)
		{
			SimAction simAction = new SimAction(ActionType.DEMAND_MODIFIED);
			simAction.demandModified_layerId = layerId;
			simAction.demandModified_demandId = demandId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
		 * 
		 * @param demandId Demand identifier
		 * @param ingressNodeId Ingress node identifier
		 * @param egressNodeId Egress node identifier
		 * @param offeredTraffic Offered traffic
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction demandRemoved(long demandId, long ingressNodeId, long egressNodeId, double offeredTraffic, Map<String, String> attributes)
		{
			return demandRemoved(-1, demandId, ingressNodeId, egressNodeId, offeredTraffic, attributes);
		}

		/**
		 * Creates a {@link ActionType#DEMAND_REMOVED DEMAND_REMOVED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param demandId Demand identifier
		 * @param ingressNodeId Ingress node identifier
		 * @param egressNodeId Egress node identifier
		 * @param offeredTraffic Offered traffic
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction demandRemoved(long layerId, long demandId, long ingressNodeId, long egressNodeId, double offeredTraffic, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.DEMAND_REMOVED);
			simAction.demandRemoved_layerId = layerId;
			simAction.demandRemoved_demandId = demandId;
			simAction.demandRemoved_ingressNodeId = ingressNodeId;
			simAction.demandRemoved_egressNodeId = egressNodeId;
			simAction.demandRemoved_offeredTraffic = offeredTraffic;
			simAction.demandRemoved_attributes = attributes;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} action.
		 * 
		 * @param demandLinkPair Demand-outgoing link pair
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction forwardingRuleAdded(Pair<Long, Long> demandLinkPair)
		{
			return Factory.forwardingRuleAdded(-1, demandLinkPair);
		}

		/**
		 * Creates a {@link ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param demandLinkPair Demand-outgoing link pair
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction forwardingRuleAdded(long layerId, Pair<Long, Long> demandLinkPair)
		{
			SimAction simAction = new SimAction(ActionType.FORWARDING_RULE_ADDED);
			simAction.forwardingRuleAdded_layerId = layerId;
			simAction.forwardingRuleAdded_demandLinkPair = demandLinkPair;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action.
		 * 
		 * @param demandLinkPair Demand-outgoing link pair
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction forwardingRuleModified(Pair<Long, Long> demandLinkPair)
		{
			return forwardingRuleModified(-1, demandLinkPair);
		}

		/**
		 * Creates a {@link ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param demandLinkPair Demand-outgoing link pair
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction forwardingRuleModified(long layerId, Pair<Long, Long> demandLinkPair)
		{
			SimAction simAction = new SimAction(ActionType.FORWARDING_RULE_MODIFIED);
			simAction.forwardingRuleModified_layerId = layerId;
			simAction.forwardingRuleModified_demandLinkPair = demandLinkPair;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action.
		 * 
		 * @param demandLinkPair Demand-outgoing link pair
		 * @param splittingRatio Splitting ratio
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction forwardingRuleRemoved(Pair<Long, Long> demandLinkPair, double splittingRatio, Map<String, String> attributes)
		{
			return Factory.forwardingRuleRemoved(-1, demandLinkPair, splittingRatio, attributes);
		}

		/**
		 * Creates a {@link ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param demandLinkPair Demand-outgoing link pair
		 * @param splittingRatio Splitting ratio
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction forwardingRuleRemoved(long layerId, Pair<Long, Long> demandLinkPair, double splittingRatio, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.FORWARDING_RULE_REMOVED);
			simAction.forwardingRuleRemoved_layerId = layerId;
			simAction.forwardingRuleRemoved_demandLinkPair = demandLinkPair;
			simAction.forwardingRuleRemoved_splittingRatio = splittingRatio;
			simAction.forwardingRuleRemoved_attributes = attributes;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LAYER_ADDED LAYER_ADDED} action.
		 * 
		 * @param layerId Layer identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction layerAdded(long layerId)
		{
			SimAction simAction = new SimAction(ActionType.LAYER_ADDED);
			simAction.layerAdded_layerId = layerId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LAYER_MODIFIED LAYER_MODIFIED} action.
		 * 
		 * @param layerId Layer identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction layerModified(long layerId)
		{
			SimAction simAction = new SimAction(ActionType.LAYER_MODIFIED);
			simAction.layerModified_layerId = layerId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LAYER_REMOVED LAYER_REMOVED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param name Name
		 * @param description Description
		 * @param linkCapacityUnitsName Link capacity units name
		 * @param demandTrafficUnitsName Demand traffic units name
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction layerRemoved(long layerId, String name, String description, String linkCapacityUnitsName, String demandTrafficUnitsName, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.LAYER_REMOVED);
			simAction.layerRemoved_layerId = layerId;
			simAction.layerRemoved_demandTrafficUnitsName = demandTrafficUnitsName;
			simAction.layerRemoved_linkCapacityUnitsName = linkCapacityUnitsName;
			simAction.layerRemoved_description = description;
			simAction.layerRemoved_name = name;
			simAction.layerRemoved_attributes = attributes;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LINK_ADDED LINK_ADDED} action.
		 * 
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkAdded(long linkId)
		{
			return linkAdded(-1, linkId);
		}

		/**
		 * Creates a {@link ActionType#LINK_ADDED LINK_ADDED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkAdded(long layerId, long linkId)
		{
			SimAction simAction = new SimAction(ActionType.LINK_ADDED);
			simAction.linkAdded_layerId = layerId;
			simAction.linkAdded_linkId = linkId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LINK_DOWN LINK_DOWN} action.
		 * 
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkDown(long linkId)
		{
			return linkDown(-1, linkId);
		}

		/**
		 * Creates a {@link ActionType#LINK_DOWN LINK_DOWN} action.
		 * 
		 * @param layerId Layer identifier
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkDown(long layerId, long linkId)
		{
			SimAction simAction = new SimAction(ActionType.LINK_DOWN);
			simAction.linkDown_layerId = layerId;
			simAction.linkDown_linkId = linkId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LINK_MODIFIED LINK_MODIFIED} action.
		 * 
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkModified(long linkId)
		{
			return linkModified(-1, linkId);
		}

		/**
		 * Creates a {@link ActionType#LINK_MODIFIED LINK_MODIFIED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkModified(long layerId, long linkId)
		{
			SimAction simAction = new SimAction(ActionType.LINK_MODIFIED);
			simAction.linkModified_layerId = layerId;
			simAction.linkModified_linkId = linkId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
		 * 
		 * @param linkId Link identifier
		 * @param originNodeId Origin node identifier
		 * @param destinationNodeId Destination node identifier
		 * @param capacity Capacity
		 * @param lengthInKm Length in km
		 * @param propagationSpeedInKmPerSecond Propagation speed in km per second
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkRemoved(long linkId, long originNodeId, long destinationNodeId, double capacity, double lengthInKm, double propagationSpeedInKmPerSecond, Map<String, String> attributes)
		{
			return linkRemoved(-1, linkId, originNodeId, destinationNodeId, capacity, lengthInKm, propagationSpeedInKmPerSecond, attributes);
		}

		/**
		 * Creates a {@link ActionType#LINK_REMOVED LINK_REMOVED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param linkId Link identifier
		 * @param originNodeId Origin node identifier
		 * @param destinationNodeId Destination node identifier
		 * @param capacity Capacity
		 * @param lengthInKm Length in km
		 * @param propagationSpeedInKmPerSecond Propagation speed in km per second
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkRemoved(long layerId, long linkId, long originNodeId, long destinationNodeId, double capacity, double lengthInKm, double propagationSpeedInKmPerSecond, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.LINK_REMOVED);
			simAction.linkRemoved_layerId = layerId;
			simAction.linkRemoved_linkId = linkId;
			simAction.linkRemoved_originNodeId = originNodeId;
			simAction.linkRemoved_destinationNodeId = destinationNodeId;
			simAction.linkRemoved_capacity = capacity;
			simAction.linkRemoved_lengthInKm = lengthInKm;
			simAction.linkRemoved_propagationSpeedInKmPerSecond = propagationSpeedInKmPerSecond;
			simAction.linkRemoved_attributes = attributes;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#LINK_UP LINK_UP} action.
		 * 
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkUp(long linkId)
		{
			return linkUp(-1, linkId);
		}

		/**
		 * Creates a {@link ActionType#LINK_UP LINK_UP} action.
		 * 
		 * @param layerId Layer identifier
		 * @param linkId Link identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction linkUp(long layerId, long linkId)
		{
			SimAction simAction = new SimAction(ActionType.LINK_UP);
			simAction.linkUp_layerId = layerId;
			simAction.linkUp_linkId = linkId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#NODE_ADDED NODE_ADDED} action.
		 * 
		 * @param nodeId Node identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction nodeAdded(long nodeId)
		{
			SimAction simAction = new SimAction(ActionType.NODE_ADDED);
			simAction.nodeAdded_nodeId = nodeId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#NODE_DOWN NODE_DOWN} action.
		 * 
		 * @param nodeId Node identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction nodeDown(long nodeId)
		{
			SimAction simAction = new SimAction(ActionType.NODE_DOWN);
			simAction.nodeDown_nodeId = nodeId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#NODE_MODIFIED NODE_MODIFIED} action.
		 * 
		 * @param nodeId Node identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction nodeModified(long nodeId)
		{
			SimAction simAction = new SimAction(ActionType.NODE_MODIFIED);
			simAction.nodeModified_nodeId = nodeId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#NODE_REMOVED NODE_REMOVED} action.
		 * 
		 * @param nodeId Node identifier
		 * @param xyPosition Coordinates
		 * @param name Name
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction nodeRemoved(long nodeId, Point2D xyPosition, String name, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.NODE_REMOVED);
			simAction.nodeRemoved_nodeId = nodeId;
			simAction.nodeRemoved_name = name;
			simAction.nodeRemoved_xyPosition = xyPosition;
			simAction.nodeRemoved_attributes = attributes;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#NODE_UP NODE_UP} action.
		 * 
		 * @param nodeId Node identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction nodeUp(long nodeId)
		{
			SimAction simAction = new SimAction(ActionType.NODE_UP);
			simAction.nodeUp_nodeId = nodeId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#PROTECTION_SEGMENT_ADDED PROTECTION_SEGMENT_ADDED} action.
		 * 
		 * @param segmentId Protection segment identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction protectionSegmentAdded(long segmentId)
		{
			return protectionSegmentAdded(-1, segmentId);
		}

		/**
		 * Creates a {@link ActionType#PROTECTION_SEGMENT_ADDED PROTECTION_SEGMENT_ADDED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param segmentId Protection segment identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction protectionSegmentAdded(long layerId, long segmentId)
		{
			SimAction simAction = new SimAction(ActionType.PROTECTION_SEGMENT_ADDED);
			simAction.segmentAdded_layerId = layerId;
			simAction.segmentAdded_segmentId = segmentId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action.
		 * 
		 * @param segmentId Protection segment identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction protectionSegmentModified(long segmentId)
		{
			return protectionSegmentModified(-1, segmentId);
		}

		/**
		 * Creates a {@link ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action.
		 * 
		 * @param layerId Layer identifier
		 * @param segmentId Protection segment identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction protectionSegmentModified(long layerId, long segmentId)
		{
			SimAction simAction = new SimAction(ActionType.PROTECTION_SEGMENT_MODIFIED);
			simAction.segmentModified_layerId = layerId;
			simAction.segmentModified_segmentId = segmentId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
		 *
		 * @param segmentId Protection segment identifier
		 * @param reservedCapacity Reserved capacity
		 * @param seqLinks Sequence of links
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction protectionSegmentRemoved(long segmentId, double reservedCapacity, List<Long> seqLinks, Map<String, String> attributes)
		{
			return protectionSegmentRemoved(-1, segmentId, reservedCapacity, seqLinks, attributes);
		}

		/**
		 * Creates a {@link ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action.
		 *
		 * @param layerId Layer identifier
		 * @param segmentId Protection segment identifier
		 * @param reservedCapacity Reserved capacity
		 * @param seqLinks Sequence of links
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction protectionSegmentRemoved(long layerId, long segmentId, double reservedCapacity, List<Long> seqLinks, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.PROTECTION_SEGMENT_REMOVED);
			simAction.segmentRemoved_layerId = layerId;
			simAction.segmentRemoved_segmentId = segmentId;
			simAction.segmentRemoved_reservedCapacity = reservedCapacity;
			simAction.segmentRemoved_seqLinks = seqLinks;
			if (attributes != null) simAction.segmentRemoved_attributes = attributes;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#ROUTE_ADDED ROUTE_ADDED} action.
		 *
		 * @param routeId Route identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction routeAdded(long routeId)
		{
			return routeAdded(-1, routeId);
		}

		/**
		 * Creates a {@link ActionType#ROUTE_ADDED ROUTE_ADDED} action.
		 *
		 * @param layerId Layer identifier
		 * @param routeId Route identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction routeAdded(long layerId, long routeId)
		{
			SimAction simAction = new SimAction(ActionType.ROUTE_ADDED);
			simAction.routeAdded_layerId = layerId;
			simAction.routeAdded_routeId = routeId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action.
		 *
		 * @param routeId Route identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction routeModified(long routeId)
		{
			return routeModified(-1, routeId);
		}

		/**
		 * Creates a {@link ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action.
		 *
		 * @param layerId Layer identifier
		 * @param routeId Route identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction routeModified(long layerId, long routeId)
		{
			SimAction simAction = new SimAction(ActionType.ROUTE_MODIFIED);
			simAction.routeModified_layerId = layerId;
			simAction.routeModified_routeId = routeId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
		 *
		 * @param routeId Route identifier
		 * @param demandId Demand identifier
		 * @param ingressNodeId Ingress node identifier
		 * @param egressNodeId Egress node identifier
		 * @param carriedTraffic Carried traffic from the demand
		 * @param occupiedCapacity Occupied capacity in the links
		 * @param seqLinks Sequence of traversed links (and segments)
		 * @param backupSegmentList Backup segment list
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction routeRemoved(long routeId, long demandId, long ingressNodeId, long egressNodeId, double carriedTraffic, double occupiedCapacity, List<Long> seqLinks, List<Long> backupSegmentList, Map<String, String> attributes)
		{
			return routeRemoved(-1, routeId, demandId, ingressNodeId, egressNodeId, carriedTraffic, occupiedCapacity, seqLinks, backupSegmentList, attributes);
		}

		/**
		 * Creates a {@link ActionType#ROUTE_REMOVED ROUTE_REMOVED} action.
		 *
		 * @param layerId Layer identifier
		 * @param routeId Route identifier
		 * @param demandId Demand identifier
		 * @param ingressNodeId Ingress node identifier
		 * @param egressNodeId Egress node identifier
		 * @param carriedTraffic Carried traffic from the demand
		 * @param occupiedCapacity Occupied capacity in the links
		 * @param seqLinks Sequence of traversed links (and segments)
		 * @param backupSegmentList Backup segment list
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction routeRemoved(long layerId, long routeId, long demandId, long ingressNodeId, long egressNodeId, double carriedTraffic, double occupiedCapacity, List<Long> seqLinks, List<Long> backupSegmentList, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.ROUTE_REMOVED);
			simAction.routeRemoved_layerId = layerId;
			simAction.routeRemoved_demandId = demandId;
			simAction.routeRemoved_routeId = routeId;
			simAction.routeRemoved_ingressNodeId = ingressNodeId;
			simAction.routeRemoved_egressNodeId = egressNodeId;
			simAction.routeRemoved_carriedTraffic = carriedTraffic;
			simAction.routeRemoved_occupiedCapacity = occupiedCapacity;
			simAction.routeRemoved_seqLinks = seqLinks;
			simAction.routeRemoved_backupSegmentList = backupSegmentList;
			simAction.routeRemoved_attributes = attributes;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#SRG_ADDED SRG_ADDED} action.
		 *
		 * @param srgId SRG identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction srgAdded(long srgId)
		{
			SimAction simAction = new SimAction(ActionType.SRG_ADDED);
			simAction.srgAdded_srgId = srgId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#SRG_MODIFIED SRG_MODIFIED} action.
		 *
		 * @param srgId SRG identifier
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction srgModified(long srgId)
		{
			SimAction simAction = new SimAction(ActionType.SRG_MODIFIED);
			simAction.srgModified_srgId = srgId;

			return simAction;
		}

		/**
		 * Creates a {@link ActionType#SRG_REMOVED SRG_REMOVED} action.
		 *
		 * @param srgId SRG identifier
		 * @param mttfInHours Mean time to fail (in hours)
		 * @param mttrInHours Mean time to repair (in hours)
		 * @param attributes Attributes
		 * @return {@code SimAction}
		 * @since 0.3.0
		 */
		public static SimAction srgRemoved(long srgId, double mttfInHours, double mttrInHours, Map<String, String> attributes)
		{
			SimAction simAction = new SimAction(ActionType.SRG_REMOVED);
			simAction.srgRemoved_srgId = srgId;
			simAction.srgRemoved_mttfInHours = mttfInHours;
			simAction.srgRemoved_mttrInHours = mttrInHours;
			simAction.srgRemoved_attributes = attributes;

			return simAction;
		}
	}
}