/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import cern.colt.matrix.tdouble.DoubleFactory2D;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.XMLUtils;
import com.net2plan.libraries.GraphUtils.GraphPath;
import com.net2plan.libraries.GraphUtils.JUNGUtils;
import com.net2plan.libraries.GraphUtils.YenAlgorithm;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.LongUtils;
import com.net2plan.utils.Pair;
import edu.uci.ics.jung.graph.Graph;
import java.io.*;
import java.util.*;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.XMLEvent;
import org.codehaus.stax2.XMLInputFactory2;
import org.codehaus.stax2.XMLOutputFactory2;
import org.codehaus.stax2.XMLStreamReader2;
import org.codehaus.stax2.XMLStreamWriter2;

/**
 * <p>A candidate path list is an object containing a set of paths computed for
 * each demand in the network. This object is commonly used for solving
 * flow-path formulations. Each path is characterized by a demand and a sequence
 * of traversed links (route). If more than one demand exists between two nodes,
 * the same route appears in different paths, each for one demand. There are
 * several forms of initializing a candidate path list, based on the k-shortest
 * path idea. In general, for every demand k paths are computed with the
 * shortest weight according to some weights assigned to the links.</p>
 *
 * <p>The computation of paths can be configured via {@code "parameter=value"} options in the constructor. There are several options to
 * configure, which can be combined:</p>
 *
 * <ul>
 * <li>{@code computePaths}: Indicates whether or not paths should be computed (default: true)</li>
 * <li>{@code K}: Number of desired loopless shortest paths (default: 3). If <i>K'</i><{@code K} different paths are found between the demand node pairs, then only <i>K'</i> paths are included in the candidate path list</li>
 * <li>{@code maxLengthInKm}: Maximum path length measured in kilometers allowed (default: Double.MAX_VALUE)</li>
 * <li>{@code maxNumHops}: Maximum number of hops allowed (default: Integer.MAX_VALUE)</li>
 * <li>{@code maxPropDelayInMs}: Maximum propagation delay in miliseconds allowed in a path (default: Double.MAX_VALUE)</li>
 * <li>{@code maxWeight}: Maximum path weight allowed (default: Double.MAX_VALUE)</li>
 * <li>{@code maxWeightFactorRespectToShortestPath}: Maximum path weight factor with respect to the shortest path weight (default: Double.MAX_VALUE)</li>
 * <li>{@code maxWeightRespectToShortestPath}: Maximum path weight with respect to the shortest path weight (default: Double.MAX_VALUE). While the previous one is a multiplicative factor, this one is an additive factor</li>
 * <li>{@code useRoutesWithinNetPlan}: Indicates whether or not existing routes are added to the candidate path list (default: false)</li>
 * <li>{@code weights}: Link weight vector (default: vector of 1s, which corresponds to a shortest path algorithm using number of hops as metric)</li>
 * </ul>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 */
public class CandidatePathList
{
	private final static String TEMPLATE_NOT_ACTIVE_PATH = "Path %d is not in the candidate list";
	private final static String TEMPLATE_NO_MORE_PATHS_ALLOWED = "Maximum path identifier reached";

	/* Cached data from NetPlan object */
	private Set<Long> linkIds, demandIds;
	private Map<Long, Pair<Long, Long>> linkMap, demandMap;
	private Map<Long, Integer> linkId2LinearIndex, demandId2LinearIndex;
	private Map<Pair<Long, Long>, Long> firstDemandThisNodePair;

	/* Parameters */
	private int K;
	private double maxLengthInKm, maxPropDelayInMs, maxWeight, maxWeightFactorRespectToShortestPath, maxWeightRespectToShortestPath;
	private int maxNumHops;
	private Map<Long, Double> linkWeightMap;
	
	/* Path list data */
	private long nextPathId;
	private final Set<Long> pathIds;
	private final Map<Long, Long> pathDemandIds;
	private final Map<Long, List<Long>> pathSequenceOfLinks;
	private final Map<Long, Set<Long>> demand2PathMap;
	private final Map<Long, Set<Long>> link2PathMap;

	private CandidatePathList()
	{
		nextPathId = 0;
		pathIds = new LinkedHashSet<Long>();
		pathDemandIds = new LinkedHashMap<Long, Long>();
		pathSequenceOfLinks = new LinkedHashMap<Long, List<Long>>();
		demand2PathMap = new LinkedHashMap<Long, Set<Long>>();
		link2PathMap = new LinkedHashMap<Long, Set<Long>>();
		firstDemandThisNodePair = new LinkedHashMap<Pair<Long, Long>, Long>();
	}
	
	/**
	 * Initializes the candidate path list, previously stored in a system file.
	 *
	 * @param f File containing a previously computed {@code CandidatePathList}
	 * @since 0.3.0
	 */
	public CandidatePathList(File f)
	{
		this();
		
		linkMap = new LinkedHashMap<Long, Pair<Long, Long>>();
		demandMap = new LinkedHashMap<Long, Pair<Long, Long>>();
		
		try
		{
			XMLInputFactory2 xmlInputFactory = (XMLInputFactory2) XMLInputFactory.newInstance();
			XMLStreamReader2 xmlStreamReader = (XMLStreamReader2) xmlInputFactory.createXMLStreamReader(f);

			while(xmlStreamReader.hasNext())
			{
				int eventType = xmlStreamReader.next();
				switch (eventType)
				{
					case XMLEvent.START_ELEMENT:
						String elementName = xmlStreamReader.getName().toString();
						if (!elementName.equals("candidatePathList")) throw new RuntimeException("Bad");

						while(xmlStreamReader.hasNext())
						{
							xmlStreamReader.next();
							switch(xmlStreamReader.getEventType())
							{
								case XMLEvent.START_ELEMENT:
									String startElementName = xmlStreamReader.getName().toString();
									switch(startElementName)
									{
										case "demand":
											long demandId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
											long ingressNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "ingressNodeId"));
											long egressNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "egressNodeId"));
											Pair<Long, Long> aux = Pair.of(ingressNodeId, egressNodeId);
											demandMap.put(demandId, aux);
											
											if (!firstDemandThisNodePair.containsKey(aux))
												firstDemandThisNodePair.put(aux, demandId);
											
											break;

										case "link":
											long linkId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
											long originNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "originNodeId"));
											long destinationNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "destinationNodeId"));
											linkMap.put(linkId, Pair.of(originNodeId, destinationNodeId));
											break;

										case "path":
											long demandId_thisPath = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "demandId"));
											List<Long> seqLinks = LongUtils.toList(xmlStreamReader.getAttributeAsLongArray(xmlStreamReader.getAttributeIndex(null, "seqLinks")));
											addPath(demandId_thisPath, seqLinks);
											break;

										default:
											throw new RuntimeException("Bad");
									}

									break;

								case XMLEvent.END_ELEMENT:
									String endElementName = xmlStreamReader.getName().toString();
									if (endElementName.equals("candidatePathList"))
									{
										linkIds = new LinkedHashSet<Long>(linkMap.keySet());
										demandIds = new LinkedHashSet<Long>(demandMap.keySet());
										linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
										demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds);
										
										return;
									}
									break;
							}
						}

						throw new RuntimeException("Bad");

					default:
						break;
				}
			}		
		}
		catch(FactoryConfigurationError | XMLStreamException | RuntimeException e)
		{
			throw new RuntimeException(e);
		}
	}
	
	/**
	 * <p>Initializes the candidate path list, computing all the paths for each
	 * demand. A candidate path list is an object containing a set of paths
	 * computed for each demand in the network. This object is commonly used for
	 * solving flow-path formulations. Each path is characterized by a demand
	 * and a sequence of traversed links (route). If more than one demand exists
	 * between two nodes, one path per demand is created, even though they follow 
	 * the same route.</p>
	 * 
	 * <p>There are several forms of initializing a candidate path
	 * list, based on the k-shortest path idea. In general, for every demand k
	 * paths are computed with the shortest weight according to some weights
	 * assigned to the links.</p>
	 *
	 * <p>Here, all link weights are initialized to 1, meaning that shortest 
	 * path will be computed according to the number of hops.</p>
	 *
	 * @param netPlan The network plan, containing at least the nodes, links and demands in the network. Routes already defined will be ommitted
	 * @param paramValuePairs Parameters to be passed to the class to tune its operation. An even number of {@code String} is to be passed. For each {@code String} pair, first {@code String} must be the name of the parameter, second a {@code String} with its value. If no name-value pairs are set, default values are used
	 * @since 0.2.0
	 */
	public CandidatePathList(NetPlan netPlan, String... paramValuePairs)
	{
		this(netPlan, netPlan.getLayerDefaultId(), DoubleUtils.ones(netPlan.getLinkIds()), paramValuePairs);
	}

	/**
	 * <p>Initializes the candidate path list, computing all the paths for each
	 * demand. A candidate path list is an object containing a set of paths
	 * computed for each demand in the network. This object is commonly used for
	 * solving flow-path formulations. Each path is characterized by a demand
	 * and a sequence of traversed links (route). If more than one demand exists
	 * between two nodes, one path per demand is created, even though they follow 
	 * the same route.</p>
	 * 
	 * <p>There are several forms of initializing a candidate path
	 * list, based on the k-shortest path idea. In general, for every demand k
	 * paths are computed with the shortest weight according to some weights
	 * assigned to the links.</p>
	 *
	 * <p>Here, all link weights are initialized to 1, meaning that shortest 
	 * path will be computed according to the number of hops.</p>
	 *
	 * @param netPlan The network plan, containing at least the nodes, links and demands in the network. Routes already defined will be ommitted
	 * @param layerId Layer identifier
	 * @param paramValuePairs Parameters to be passed to the class to tune its operation. An even number of {@code String} is to be passed. For each {@code String} pair, first {@code String} must be the name of the parameter, second a {@code String} with its value. If no name-value pairs are set, default values are used
	 * @since 0.3.0
	 */
	public CandidatePathList(NetPlan netPlan, long layerId, String... paramValuePairs)
	{
		this(netPlan, layerId, DoubleUtils.ones(netPlan.getLinkIds(layerId)), paramValuePairs);
	}
	
	/**
	 * Initializes the candidate path list, computing all the paths for each
	 * demand. A candidate path list is an object containing a set of paths
	 * computed for each demand in the network. This object is commonly used for
	 * solving flow-path formulations. Each path is characterized by a demand
	 * and a sequence of traversed links (route). If more than one demand exists
	 * between two nodes, one path per demand is created, even though they follow the same route.
	 * There are several forms of initializing a candidate path
	 * list, based on the k-shortest path idea. In general, for every demand k
	 * paths are computed with the shortest weight according to some weights
	 * assigned to the links.
	 *
	 * @param netPlan The network plan, containing at least the nodes, links and demands in the network. Routes already defined will be ommitted
	 * @param linkWeightMap Link weight for the shortest path algorithm (null means 'all-one' map)
	 * @param paramValuePairs Parameters to be passed to the class to tune its operation. An even number of {@code String} is to be passed. For each {@code String} pair, first {@code String} must be the name of the parameter, second a {@code String} with its value. If no name-value pairs are set, default values are used
	 * @since 0.2.0
	 */
	public CandidatePathList(NetPlan netPlan, Map<Long, Double> linkWeightMap, String... paramValuePairs)
	{
		this(netPlan, netPlan.getLayerDefaultId(), linkWeightMap, paramValuePairs);
	}

	/**
	 * Initializes the candidate path list, computing all the paths for each
	 * demand. A candidate path list is an object containing a set of paths
	 * computed for each demand in the network. This object is commonly used for
	 * solving flow-path formulations. Each path is characterized by a demand
	 * and a sequence of traversed links (route). If more than one demand exists
	 * between two nodes, one path per demand is created, even though they follow the same route.
	 * There are several forms of initializing a candidate path
	 * list, based on the k-shortest path idea. In general, for every demand k
	 * paths are computed with the shortest weight according to some weights
	 * assigned to the links.
	 *
	 * @param netPlan The network plan, containing at least the nodes, links and demands in the network. Routes already defined will be ommitted
	 * @param layerId Layer identifier
	 * @param linkWeightMap Link weight for the shortest path algorithm (null means 'all-one' map)
	 * @param paramValuePairs Parameters to be passed to the class to tune its operation. An even number of {@code String} is to be passed. For each {@code String} pair, first {@code String} must be the name of the parameter, second a {@code String} with its value. If no name-value pairs are set, default values are used
	 * @since 0.3.0
	 */
	public CandidatePathList(final NetPlan netPlan, long layerId, Map<Long, Double> linkWeightMap, String... paramValuePairs)
	{
		this();

		linkIds = netPlan.getLinkIds(layerId);
		demandIds = netPlan.getDemandIds(layerId);
		linkMap = netPlan.getLinkMap(layerId);
		demandMap = netPlan.getDemandMap(layerId);
		linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds);

		for(long linkId : linkIds) link2PathMap.put(linkId, new TreeSet<Long>());
		for(long demandId : demandIds) demand2PathMap.put(demandId, new TreeSet<Long>());

		if (linkWeightMap != null)
		{
			int E = linkIds.size();
			if (linkWeightMap.size() != E) throw new Net2PlanException("Number of weights (" + linkWeightMap.size() + ") does not match the number of links (" + E + ")");
			this.linkWeightMap = Collections.unmodifiableMap(linkWeightMap);
		}

		boolean computePaths = true;
		K = 3;
		maxLengthInKm = Double.MAX_VALUE;
		maxNumHops = Integer.MAX_VALUE;
		maxPropDelayInMs = Double.MAX_VALUE;
		maxWeight = Double.MAX_VALUE;
		maxWeightFactorRespectToShortestPath = Double.MAX_VALUE;
		maxWeightRespectToShortestPath = Double.MAX_VALUE;
		boolean useRoutesWithinNetPlan = false;
		
		int numParameters = (int) (paramValuePairs.length / 2);
		if (numParameters * 2 != paramValuePairs.length) throw new Net2PlanException("A parameter has not assigned its value");

		for (int contParam = 0; contParam < numParameters; contParam++)
		{
			String parameter = paramValuePairs[contParam * 2];
			String value = paramValuePairs[contParam * 2 + 1];

			if (parameter.equalsIgnoreCase("computePaths"))
			{
				computePaths = Boolean.parseBoolean(value);
			}
			else if (parameter.equalsIgnoreCase("useRoutesWithinNetPlan"))
			{
				useRoutesWithinNetPlan = Boolean.parseBoolean(value);
			}
			else if (parameter.equalsIgnoreCase("K"))
			{
				K = Integer.parseInt(value);
				if (K <= 0) throw new Net2PlanException("'K' parameter must be greater than zero");
			}
			else if (parameter.equalsIgnoreCase("maxLengthInKm"))
			{
				maxLengthInKm = Double.parseDouble(value);
				if (maxLengthInKm <= 0) throw new Net2PlanException("'maxLengthInKm' parameter must be greater than zero");
			}
			else if (parameter.equalsIgnoreCase("maxPropDelayInMs"))
			{
				maxPropDelayInMs = Double.parseDouble(value);
				if (maxPropDelayInMs <= 0) throw new Net2PlanException("'maxPropDelayInMs' parameter must be greater than zero");
			}
			else if (parameter.equalsIgnoreCase("maxNumHops"))
			{
				maxNumHops = Integer.parseInt(value);
				if (maxNumHops <= 0) throw new Net2PlanException("'maxNumHops' parameter must be greater than zero");
			}
			else if (parameter.equalsIgnoreCase("maxWeight"))
			{
				maxWeight = Double.parseDouble(value);
				if (maxWeight <= 0) throw new Net2PlanException("'maxWeight' parameter must be greater than zero");
			}
			else if (parameter.equalsIgnoreCase("maxWeightFactorRespectToShortestPath"))
			{
				maxWeightFactorRespectToShortestPath = Double.parseDouble(value);
				if (maxWeightFactorRespectToShortestPath <= 0) throw new Net2PlanException("'maxWeightFactorRespectToShortestPath' parameter must be greater than zero");
			}
			else if (parameter.equalsIgnoreCase("maxWeightRespectToShortestPath"))
			{
				maxWeightRespectToShortestPath = Double.parseDouble(value);
				if (maxWeightRespectToShortestPath < 0) throw new Net2PlanException("'maxWeightRespectToShortestPath' parameter must be greater or equal than zero");
			}
			else
			{
				throw new RuntimeException("Unknown parameter " + parameter);
			}
		}

		if (useRoutesWithinNetPlan)
		{
			Set<Long> routeIds = netPlan.getRouteIds(layerId);
			for(long routeId : routeIds) addPath(netPlan.getRouteDemand(layerId, routeId), netPlan.getRouteSequenceOfLinks(layerId, routeId));
		}

		if (!computePaths) return;

		Set<Long> nodeIds = netPlan.getNodeIds();
		Graph<Long, Long> g = JUNGUtils.getGraphFromLinkMap(nodeIds, linkMap);

		YenAlgorithm<Long, Long> paths = new YenAlgorithm<Long, Long>(g, JUNGUtils.getEdgeWeightTransformer(linkWeightMap))
		{
			@Override
			public boolean acceptPath(GraphPath<Long> candidate)
			{
				CandidatePathList cpl = CandidatePathList.this;
				if (cpl.maxNumHops != Integer.MAX_VALUE && candidate.getPathLength() > cpl.maxNumHops) return false;
				
				if (cpl.maxLengthInKm == Double.MAX_VALUE && cpl.maxPropDelayInMs == Double.MAX_VALUE) return true;
				
				double pathLengthInKm = 0;
				double pathPropDelayInMs = 0;
				for (long linkId : candidate.getPath())
				{
					pathLengthInKm += netPlan.getLinkLengthInKm(linkId);
					pathPropDelayInMs += netPlan.getLinkPropagationDelayInSeconds(linkId) * 1000;
				}

				return pathLengthInKm <= cpl.maxLengthInKm && pathPropDelayInMs <= cpl.maxPropDelayInMs;
			}

			@Override
			public boolean compareCandidateToShortestPath(GraphPath<Long> candidate, GraphPath<Long> shortestPath)
			{
				CandidatePathList cpl = CandidatePathList.this;
				if (cpl.maxWeight != Double.MAX_VALUE && candidate.getPathWeight() > cpl.maxWeight)
				{
					return false;
				}
				
				if (cpl.maxWeightFactorRespectToShortestPath != Double.MAX_VALUE && candidate.getPathWeight() > shortestPath.getPathWeight() * cpl.maxWeightFactorRespectToShortestPath)
				{
					return false;
				}
				
				return !(cpl.maxWeightRespectToShortestPath != Double.MAX_VALUE && candidate.getPathWeight() > shortestPath.getPathWeight() + cpl.maxWeightRespectToShortestPath);
			}
		};
		
		for(long demandId : demandIds)
		{
			long ingressNodeId = netPlan.getDemandIngressNode(layerId, demandId);
			long egressNodeId = netPlan.getDemandEgressNode(layerId, demandId);

			Pair<Long, Long> aux = Pair.of(ingressNodeId, egressNodeId);
			if (firstDemandThisNodePair.containsKey(aux))
			{
				long firstDemandId = firstDemandThisNodePair.get(aux);
				Set<Long> pathIds_thisDemand = demand2PathMap.get(firstDemandId);
				for(long pathId : pathIds_thisDemand)
					addPath(demandId, getPathSequenceOfLinks(pathId));
			}
			else
			{
				List<List<Long>> paths_thisNodePair = paths.getPaths(ingressNodeId, egressNodeId, K);

				for (List<Long> seqLinks : paths_thisNodePair)
					addPath(demandId, seqLinks);
				
				firstDemandThisNodePair.put(aux, demandId);
			}
		}
	}
	
	/**
	 * Adds a new path to the path list.
	 *
	 * @param demandId Demand identifier
	 * @param seqLinks Sequence of traversed links
	 * @return Path identifier
	 * @since 0.3.0
	 */
	public long addPath(long demandId, List<Long> seqLinks)
	{
		GraphUtils.checkRouteContinuity(linkMap, seqLinks, Constants.CheckRoutingCycleType.NO_CHECK);

		long initialNodeId = linkMap.get(seqLinks.get(0)).getFirst();
		long lastNodeId = linkMap.get(seqLinks.get(seqLinks.size() - 1)).getSecond();
		long ingressNodeId = demandMap.get(demandId).getFirst();
		long egressNodeId = demandMap.get(demandId).getSecond();
		if (ingressNodeId != initialNodeId) throw new Net2PlanException("Initial node in the path is not equal to the ingress node of the demand");
		if (egressNodeId != lastNodeId) throw new Net2PlanException("Last node in the path is not equal to the egress node of the demand");
		
		long pathId = getPathNextId();
		pathSequenceOfLinks.put(pathId, Collections.unmodifiableList(new LinkedList<Long>(seqLinks)));
		pathDemandIds.put(pathId, demandId);
		
		demand2PathMap.get(demandId).add(pathId);
		for(long linkId : seqLinks)
			link2PathMap.get(linkId).add(pathId);
		
		pathIds.add(pathId);
		
		nextPathId++;
		return pathId;
	}
	
	private void checkPathIsActive(long pathId)
	{
		if (!isPathActive(pathId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_PATH, pathId));
	}

	/**
	 * Computes the amount of traffic carried for each demand, if each path
	 * carries the traffic given in x_p vector
	 *
	 * @param x_p The traffic carried by each path in the list
	 * @return The traffic carried by each demand
	 * @since 0.2.0
	 */
	public Map<Long, Double> computeDemandCarriedTraffic(double[] x_p)
	{
		Map<Long, Double> r_d = DoubleUtils.zeros(demandIds);
		
		Map<Long, Integer> pathId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(pathIds);
		for(long pathId : pathIds)
		{
			int p = pathId2LinearIndex.get(pathId);
			long demandId = getPathDemand(pathId);
			
			r_d.put(demandId, r_d.get(demandId) + x_p[p]);
		}
		
		return r_d;
	}
	
	/**
	 * Returns the shortest path cost for every demand.
	 *
	 * @param linkWeightMap Link weights
	 * @return Shortest path cost for every demand
	 * @since 0.2.0
	 */
	public Map<Long, Double> computeDemandShortestPathWeight(Map<Long, Double> linkWeightMap)
	{
		Map<Long, Double> pathWeightMap = getPathCostMap(linkWeightMap);
		
		Map<Long, Double> demandShortestPathWeightMap = DoubleUtils.constantMap(demandIds, Double.MAX_VALUE);
		for(long pathId : pathIds)
		{
			double weight_thisPath = pathWeightMap.get(pathId);

			long demandId = getPathDemand(pathId);
			double weight_thisDemand = demandShortestPathWeightMap.get(demandId);
			if (weight_thisPath < weight_thisDemand) demandShortestPathWeightMap.put(demandId, weight_thisPath);
		}
		
		return pathWeightMap;
	}
	
	/**
	 * Computes the amount of traffic in each link, if each path carries the
	 * traffic given in x_p vector.
	 *
	 * @param y_p Capacity units occupied by each path
	 * @return The traffic carried in each link
	 * @since 0.2.0
	 */
	public Map<Long, Double> computeLinkCarriedTrafficMap(Map<Long, Double> y_p)
	{
		Map<Long, Double> y_e = DoubleUtils.zeros(linkIds);
		
		for(long pathId : pathIds)
		{
			double y_p_thisPath = y_p.get(pathId);
			List<Long> seqLinks = getPathSequenceOfLinks(pathId);
			for(long linkId : seqLinks)
				y_e.put(linkId, y_e.get(linkId) + y_p_thisPath);
		}
		
		return y_e;
	}
	
	/**
	 * Computes the amount of traffic in each link, if each path carries the
	 * traffic given in x_p vector.
	 *
	 * @param y_p Capacity units occupied by each path
	 * @return The traffic carried in each link
	 * @since 0.2.0
	 */
	public Map<Long, Double> computeLinkCarriedTrafficMap(double[] y_p)
	{
		Map<Long, Double> y_e = DoubleUtils.zeros(linkIds);
		
		Map<Long, Integer> pathId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(pathIds);
		for(long pathId : pathIds)
		{
			int p = pathId2LinearIndex.get(pathId);
	
			List<Long> seqLinks = getPathSequenceOfLinks(pathId);
			for(long linkId : seqLinks)
				y_e.put(linkId, y_e.get(linkId) + y_p[p]);
		}
		
		return y_e;
	}

	/**
	 * Computes the amount of traffic in each link, if each path carries the
	 * traffic given in y_p map.
	 *
	 * @param y_p Capacity units occupied by each path
	 * @return The traffic carried in each link
	 * @since 0.2.0
	 */
	public double[] computeLinkCarriedTrafficVector(Map<Long, Double> y_p)
	{
		return DoubleUtils.toArray(computeLinkCarriedTrafficMap(y_p).values());
	}
	
	/**
	 * Computes the amount of traffic in each link, if each path carries the
	 * traffic given in x_p vector.
	 *
	 * @param y_p Capacity units occupied by each path
	 * @return The traffic carried in each link
	 * @since 0.2.0
	 */
	public double[] computeLinkCarriedTrafficVector(double[] y_p)
	{
		return DoubleUtils.toArray(computeLinkCarriedTrafficMap(y_p).values());
	}

	/**
	 * Computes for each path the maximum amount of traffic that is expected to 
	 * carry in case of non-bifurcated routing, that is, the offered traffic of 
	 * its corresponding demand.
	 *
	 * @param h_d Offered traffic per demand
	 * @return Maximum traffic that is expected to carry each path in case of non-bifurcated routing
	 * @since 0.3.0
	 */
	public Map<Long, Double> computePathMaximumCarriedTrafficMap(Map<Long, Double> h_d)
	{
		Map<Long, Double> h_p = new LinkedHashMap<Long, Double>();
		for(long pathId : pathIds) h_p.put(pathId, h_d.get(getPathDemand(pathId)));
		
		return h_p;
	}

	/**
	 * Computes for each path the maximum amount of traffic that is expected to 
	 * carry in case of non-bifurcated routing, that is, the offered traffic of 
	 * its corresponding demand.
	 *
	 * @param h_d Offered traffic per demand
	 * @return Maximum traffic that is expected to carry each path in case of non-bifurcated routing
	 * @since 0.3.1
	 */
	public Map<Long, Double> computePathMaximumCarriedTrafficMap(double[] h_d)
	{
		Map<Long, Double> h_p = new LinkedHashMap<Long, Double>();
		for(long demandId : demandIds)
		{
			int d = demandId2LinearIndex.get(demandId);
			Set<Long> pathIds_thisDemand = getDemandPaths(demandId);
			for(long pathId : pathIds_thisDemand) h_p.put(pathId, h_d[d]);
		}
		
		return h_p;
	}

	/**
	 * Computes for each path the maximum amount of traffic that is expected to 
	 * carry in case of non-bifurcated routing, that is, the offered traffic of 
	 * its corresponding demand.
	 *
	 * @param h_d Offered traffic per demand
	 * @return Maximum traffic that is expected to carry each path in case of non-bifurcated routing
	 * @since 0.3.0
	 */
	public double[] computePathMaximumCarriedTrafficVector(Map<Long, Double> h_d)
	{
		return DoubleUtils.toArray(computePathMaximumCarriedTrafficMap(h_d).values());
	}
	
	/**
	 * Computes for each path the maximum amount of traffic that is expected to 
	 * carry in case of non-bifurcated routing, that is, the offered traffic of 
	 * its corresponding demand.
	 *
	 * @param h_d Offered traffic per demand
	 * @return Maximum traffic that is expected to carry each path in case of non-bifurcated routing
	 * @since 0.3.1
	 */
	public double[] computePathMaximumCarriedTrafficVector(double[] h_d)
	{
		return DoubleUtils.toArray(computePathMaximumCarriedTrafficMap(h_d).values());
	}
	
	/**
	 * Returns the demand-link incidence matrix (a <i>D</i>x<i>E</i> matrix in 
	 * which an element <i>&delta;<sub>de</sub></i> is equal to the number of 
	 * times which traffic routes carrying traffic from demand <i>d</i> traverse 
	 * link <i>e</i>).
	 *
	 * @return The demand-link incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemand2LinkAssignmentMatrix()
	{
		DoubleMatrix2D delta_dp = getDemand2PathAssignmentMatrix();
		DoubleMatrix2D delta_ep = getLink2PathAssignmentMatrix();
		
		return delta_dp.zMult(delta_ep.viewDice(), null);
	}

	/**
	 * Returns the demand-path incidence matrix (a <i>D</i>x<i>P</i> matrix in 
	 * which an element <i>&delta;<sub>dp</sub></i> is equal to 1 if traffic 
	 * route <i>p</i> is able to carry traffic from demand <i>d</i>).
	 *
	 * @return The demand-path incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemand2PathAssignmentMatrix()
	{
		int D = demandIds.size();
		int P = getNumberOfPaths();

		DoubleMatrix2D delta_dp = DoubleFactory2D.sparse.make(D, P);
		long[] d_p = getPathDemandVector();
		for (int p = 0; p < P; p++)
		{
			int d = demandId2LinearIndex.get(d_p[p]);
			delta_dp.setQuick(d, p, 1);
		}

		return delta_dp;
	}

	/**
	 * Returns the set of identifiers of the paths that are associated to the
	 * demand.
	 *
	 * @param demandId Demand identifier
	 * @return A set with the path identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getDemandPaths(long demandId)
	{
		return demand2PathMap.get(demandId);
	}

	/**
	 * Returns the array of identifiers of the paths that are associated to the
	 * demand.
	 *
	 * @param demandId Demand identifier
	 * @return An array with the path identifiers
	 * @since 0.3.0
	 */
	public long[] getDemandPathsVector(long demandId)
	{
		return LongUtils.toArray(getDemandPaths(demandId));
	}

	/**
	 * Returns the link-path incidence matrix (an <i>E</i>x<i>P</i> matrix in 
	 * which an element <i>&delta;<sub>ep</sub></i> is equal to the number of 
	 * times which traffic route <i>p</i> traverses link <i>e</i>).
	 *
	 * @return The link-path incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getLink2PathAssignmentMatrix()
	{
		int E = linkIds.size();
		int P = getNumberOfPaths();

		DoubleMatrix2D delta_ep = DoubleFactory2D.sparse.make(E, P);
		for (int p = 0; p < P; p++)
		{
			List<Long> seqLinks = getPathSequenceOfLinks(p);
			for(long linkId : seqLinks)
			{
				int e = linkId2LinearIndex.get(linkId);
				delta_ep.setQuick(e, p, delta_ep.getQuick(e, p) + 1);
			}
		}

		return delta_ep;
	}

	/**
	 * Returns the set of identifiers of the paths that traverse a link.
	 *
	 * @param linkId Link identifier
	 * @return A set with path identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLinkAssociatedPaths(long linkId)
	{
		return link2PathMap.get(linkId);
	}

	/**
	 * Returns the array of identifiers of the paths that traverse a link.
	 *
	 * @param linkId Link identifier
	 * @return An array with path identifiers
	 * @since 0.3.0
	 */
	public long[] getLinkAssociatedPathsVector(long linkId)
	{
		return LongUtils.toArray(getLinkAssociatedPaths(linkId));
	}
	
	/**
	 * Returns the set of identifiers of the paths that are associated to the
	 * first demand of a node pair.
	 *
	 * @param ingressNodeId Ingress node identifier
	 * @param egressNodeId Ingress node identifier
	 * @return A set with path identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairPaths(long ingressNodeId, long egressNodeId)
	{
		Pair<Long, Long> aux = Pair.of(ingressNodeId, egressNodeId);
		Long demandId = firstDemandThisNodePair.get(aux);
		if (demandId == null) return new LinkedHashSet<Long>();
		
		return getDemandPaths(demandId);
	}

	/**
	 * Returns the array of identifiers of the paths that are associated to the
	 * first demand of a node pair.
	 *
	 * @param ingressNodeId Ingress node identifier
	 * @param egressNodeId Ingress node identifier
	 * @return An array with the path identifiers
	 * @since 0.3.0
	 */
	public long[] getNodePairPathsVector(long ingressNodeId, long egressNodeId)
	{
		return LongUtils.toArray(getNodePairPaths(ingressNodeId, egressNodeId));
	}
	
	/**
	 * Returns the number of paths in the list.
	 *
	 * @return The number of paths
	 * @since 0.2.0
	 */
	public int getNumberOfPaths()
	{
		return pathDemandIds.size();
	}
	
	/**
	 * Returns the sequence of links for each path in ascending order of 
	 * path identifier.
	 *
	 * @return List of sequence of links
	 * @since 0.3.0
	 */
	public List<List<Long>> getPathAllSequencesOfLinksList()
	{
		return new LinkedList<List<Long>>(pathSequenceOfLinks.values());
	}

	/**
	 * Returns the sequence of links for each path.
	 *
	 * @return Sequence of links per path
	 * @since 0.3.0
	 */
	public Map<Long, List<Long>> getPathAllSequencesOfLinksMap()
	{
		return Collections.unmodifiableMap(pathSequenceOfLinks);
	}

	/**
	 * Returns the cost of a path, given a set of link weights.
	 *
	 * @param pathId Path identifier
	 * @return Path cost
	 * @since 0.3.0
	 */
	public double getPathCost(long pathId)
	{
		return getPathCost(pathId, linkWeightMap);
	}

	/**
	 * Returns the cost of a path, given a set of link weights.
	 *
	 * @param pathId Path identifier
	 * @param linkWeightMap Link weights
	 * @return Path cost
	 * @since 0.3.0
	 */
	public double getPathCost(long pathId, Map<Long, Double> linkWeightMap)
	{
		List<Long> seqLinks = getPathSequenceOfLinks(pathId);
		return GraphUtils.convertPath2PathCost(seqLinks, linkWeightMap);
	}

	/**
	 * Computes the weight of each path in the list, for the given weight of the
	 * links. The weight of a path is the sum of the weights of the traversing
	 * links
	 *
	 * @return An array with the cost of each path
	 * @since 0.2.0
	 */
	public Map<Long, Double> getPathCostMap()
	{
		return getPathCostMap(linkWeightMap);
	}

	/**
	 * Computes the weight of each path in the list, for the given weight of the
	 * links. The weight of a path is the sum of the weights of the traversing
	 * links
	 *
	 * @param linkWeightMap The set of weight per link
	 * @return An array with the cost of each path
	 * @since 0.2.0
	 */
	public Map<Long, Double> getPathCostMap(Map<Long, Double> linkWeightMap)
	{
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		for(long pathId : getPathIds())
			out.put(pathId, getPathCost(pathId, linkWeightMap));
		
		return out;
	}
	
	/**
	 * Computes the weight of each path in the list, for the given weight of the
	 * links. The weight of a path is the sum of the weights of the traversing
	 * links
	 *
	 * @return An array with the cost of each path
	 * @since 0.2.0
	 */
	public double[] getPathCostVector()
	{
		return DoubleUtils.toArray(getPathCostMap().values());
	}

	/**
	 * Computes the weight of each path in the list, for the given weight of the
	 * links. The weight of a path is the sum of the weights of the traversing
	 * links
	 *
	 * @param linkWeightMap The set of weight per link
	 * @return An array with the cost of each path
	 * @since 0.2.0
	 */
	public double[] getPathCostVector(Map<Long, Double> linkWeightMap)
	{
		return DoubleUtils.toArray(getPathCostMap(linkWeightMap).values());
	}
	
	/**
	 * Obtains the demand associated to the path.
	 *
	 * @param pathId Path identifier
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long getPathDemand(long pathId)
	{
		checkPathIsActive(pathId);
		return pathDemandIds.get(pathId);
	}
	
	/**
	 * Obtains the demand associated to each path in ascending order of path 
	 * identifiers.
	 *
	 * @return A list with the demand identifier for each path
	 * @since 0.3.0
	 */
	public List<Long> getPathDemandList()
	{
		return new LinkedList<Long>(pathDemandIds.values());
	}

	/**
	 * Obtains the demand associated to each path.
	 *
	 * @return A map with the demand identifier for each path
	 * @since 0.3.0
	 */
	public Map<Long, Long> getPathDemandMap()
	{
		return Collections.unmodifiableMap(pathDemandIds);
	}

	/**
	 * Obtains a vector with the demand associated to each path.
	 *
	 * @return A vector with the demand identifiers
	 * @since 0.3.0
	 */
	public long[] getPathDemandVector()
	{
		return LongUtils.toArray(getPathDemandMap().values());
	}
	
	/**
	 * Returns the identifiers of the active paths.
	 *
	 * @return An ordered set with the path identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getPathIds()
	{
		return Collections.unmodifiableSet(pathIds);
	}
	
	/**
	 * Returns the identifiers of the active paths.
	 *
	 * @return An ordered vector with the path identifiers
	 * @since 0.3.0
	 */
	public long[] getPathIdsVector()
	{
		return LongUtils.toArray(getPathIds());
	}
	
	/**
	 * Returns the identifier of the next added path.
	 *
	 * @return Next path identifier
	 * @since 0.3.0
	 */
	public long getPathNextId()
	{
		long pathId = nextPathId;
		if (pathId < 0) throw new Net2PlanException(TEMPLATE_NO_MORE_PATHS_ALLOWED);
		
		return pathId;
	}

	/**
	 * Returns the sequence of links associated to a path.
	 *
	 * @param pathId Path identifier
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> getPathSequenceOfLinks(long pathId)
	{
		checkPathIsActive(pathId);
		return Collections.unmodifiableList(pathSequenceOfLinks.get(pathId));
	}
	
	/**
	 * Obtains the sequence of nodes of the given path.
	 *
	 * @param pathId Path identifier
	 * @return Sequence of nodes
	 * @since 0.3.0
	 */
	public List<Long> getPathSequenceOfNodes(long pathId)
	{
		checkPathIsActive(pathId);
		return GraphUtils.convertSequenceOfLinksToSequenceOfNodes(linkMap, getPathSequenceOfLinks(pathId));
	}
	
	/**
	 * Indicates whether a path is active.
	 *
	 * @param pathId Path identifier
	 * @return {@code true} in case the path is active. Otherwise, {@code false}
	 * @since 0.3.0
	 */
	public boolean isPathActive(long pathId)
	{
		return pathIds.contains(pathId);
	}

	/**
	 * Checks whether a link belongs to a path.
	 *
	 * @param linkId Link identifier
	 * @param pathId Path identifier
	 * @return {@code true} if link {@code e} belongs to path {@code p}. Otherwise, {@code false}.
	 * @since 0.3.0
	 */
	public boolean linkBelongsToPath(long linkId, long pathId)
	{
		checkPathIsActive(pathId);
		return CollectionUtils.contains(pathSequenceOfLinks.get(pathId), linkId);
	}
	
	/**
	 * Removes a path from the candidate path list.
	 *
	 * @param pathId Path identifier
	 * @since 0.3.0
	 */
	public void removePath(long pathId)
	{
		checkPathIsActive(pathId);
		pathDemandIds.remove(pathId);
		
		long demandId = getPathDemand(pathId);
		demand2PathMap.get(demandId).remove(pathId);

		List<Long> seqLinks = getPathSequenceOfLinks(pathId);
		for(long linkId : seqLinks)
			link2PathMap.get(linkId).remove(pathId);
	}

	/**
	 * Removes a set of paths from the candidate path list.
	 *
	 * @param pathIds Path identifiers
	 * @since 0.3.0
	 */
	public void removePaths(Set<Long> pathIds)
	{
		for(long pathId : pathIds) removePath(pathId);
	}
	
	/**
	 * Saves the current candidate path list to a given file for further usage.
	 *
	 * @param f Output file
	 * @since 0.3.0
	 */
	public void saveToFile(File f)
	{
		try (OutputStream os = new FileOutputStream(f))
		{
			XMLOutputFactory2 output = (XMLOutputFactory2) XMLOutputFactory.newFactory();
			XMLStreamWriter2 writer = (XMLStreamWriter2) output.createXMLStreamWriter(os);
			
			writer.writeStartDocument("UTF-8", "1.0");
			
			XMLUtils.indent(writer, 0);
			writer.writeStartElement("candidatePathList");
			
			for (long linkId : linkIds)
			{
				Pair<Long, Long> endNodes = linkMap.get(linkId);
				double weight = linkWeightMap.get(linkId);
				
				XMLUtils.indent(writer, 1);
				writer.writeEmptyElement("link");
				writer.writeAttribute("id", Long.toString(linkId));
				writer.writeAttribute("originNodeId", Long.toString(endNodes.getFirst()));
				writer.writeAttribute("destinationNodeId", Long.toString(endNodes.getSecond()));
				writer.writeAttribute("weight", Double.toString(weight));
			}

			for (long demandId : demandIds)
			{
				Pair<Long, Long> endNodes = linkMap.get(demandId);
				
				XMLUtils.indent(writer, 1);
				writer.writeEmptyElement("demand");
				writer.writeAttribute("id", Long.toString(demandId));
				writer.writeAttribute("ingressNodeId", Long.toString(endNodes.getFirst()));
				writer.writeAttribute("egressNodeId", Long.toString(endNodes.getSecond()));
			}

			for (long pathId : pathIds)
			{
				long demandId = getPathDemand(pathId);
				List<Long> seqLinks = getPathSequenceOfLinks(pathId);
				
				XMLUtils.indent(writer, 1);
				writer.writeEmptyElement("path");
				writer.writeAttribute("demandId", Long.toString(demandId));
				writer.writeAttribute("seqLinks", CollectionUtils.join(seqLinks, " "));
			}
			
			XMLUtils.indent(writer, 0);
			writer.writeEndElement();
			writer.writeEndDocument();
			writer.flush();
			writer.close();
		}
		catch(Throwable e)
		{
			throw new RuntimeException(e);
		}
	}
	
	/**
	 * A formatted {@code String} describing the paths in the list, their 
	 * associated demands and sequence of links.
	 *
	 * @return String
	 * @since 0.2.0
	 */
	@Override
	public String toString()
	{
		StringBuilder out = new StringBuilder();
		
		for(long pathId : pathIds)
		{
			long demandId = getPathDemand(pathId);
			List<Long> seqLinks = getPathSequenceOfLinks(pathId);
			List<Long> seqNodes = getPathSequenceOfNodes(pathId);
			String auxSeqLinks = CollectionUtils.join(seqLinks, " => ");
			String auxSeqNodes = CollectionUtils.join(seqNodes, " => ");
			out.append(String.format("path %d: demand = %d, sequence of links = %s, sequence of nodes = %s%n", pathId, demandId, auxSeqLinks, auxSeqNodes));
		}

		return out.toString();
	}
}
