/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import cern.colt.matrix.tdouble.DoubleFactory2D;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import cern.colt.matrix.tdouble.algo.DenseDoubleAlgebra;
import cern.colt.matrix.tdouble.algo.decomposition.DenseDoubleEigenvalueDecomposition;
import cern.jet.math.tdouble.DoubleFunctions;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.libraries.GraphUtils.JGraphTUtils;
import com.net2plan.libraries.GraphUtils.JUNGUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.IntUtils;
import com.net2plan.utils.Pair;
import edu.uci.ics.jung.algorithms.scoring.BetweennessCentrality;
import edu.uci.ics.jung.graph.Graph;
import java.util.*;
import java.util.Map.Entry;
import org.apache.commons.collections15.Transformer;
import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.EdmondsKarpMaximumFlow;
import org.jgrapht.alg.StrongConnectivityInspector;

/**
 * <p>Class to deal with graph-theory metrics computation.</p>
 *
 * <p><b>Important</b>: Internal computations (like shortest-paths) are cached in order to improve efficiency.</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 */
public class GraphTheoryMetrics
{
	private final Set<Long> nodeIds, linkIds;
	private final Map<Long, Integer> nodeId2LinearIndexMap;
	private final Map<Long, Integer> linkId2LinearIndexMap;
	private final Map<Long, Pair<Long, Long>> linkMap;
	private Map<Long, Double> costMap;
	private final int N;
	private final int E;
	private DoubleMatrix2D adjacencyMatrix;
	private double[] adjacencyMatrixEigenvalues;
	private double averageSPLength, diameter, heterogeneity;
	private DirectedGraph<Long, Long> graph_jgrapht;
	private Graph<Long, Long> graph_jung;
	private DoubleMatrix2D incidenceMatrix;
	private DoubleMatrix2D laplacianMatrix;
	private double[] laplacianMatrixEigenvalues;
	private Transformer<Long, Double> nev;
	
	private Map<Long, Double> linkBetweenessCentrality;
	private Map<Long, Double> nodeBetweenessCentrality;
	private Map<Long, Integer> outNodeDegree;

	/**
	 * Default constructor.
	 *
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @since 0.3.0
	 */
	public GraphTheoryMetrics(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Double> linkCostMap)
	{
		this.nodeIds = new TreeSet<Long>(nodeIds);
		linkIds = new TreeSet<Long>(linkMap.keySet());
		
		N = nodeIds.size();
		E = linkIds.size();
		if (N > 0 && this.nodeIds.iterator().next() < 0) throw new Net2PlanException("Node identifiers must be greater or equal than zero");
		if (E > 0 && linkIds.iterator().next() < 0) throw new Net2PlanException("Link identifiers must be greater or equal than zero");
		
		nodeId2LinearIndexMap = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		linkId2LinearIndexMap = CollectionUtils.convertId2LinearIndexMap(linkIds);

		this.linkMap = new LinkedHashMap<Long, Pair<Long, Long>>(linkMap);

		adjacencyMatrix = null;
		adjacencyMatrixEigenvalues = null;
		graph_jgrapht = null;
		graph_jung = null;
		incidenceMatrix = null;
		laplacianMatrix = null;
		laplacianMatrixEigenvalues = null;
		linkBetweenessCentrality = null;
		nev = null;
		nodeBetweenessCentrality = null;
		outNodeDegree = null;

		configureLinkCostMap(linkCostMap);
	}

	private void computeBetweenessCentrality()
	{
		Graph<Long, Long> aux_graph = getGraph_JUNG();
		BetweennessCentrality bc = new BetweennessCentrality(aux_graph, getCostTransformer());			

		nodeBetweenessCentrality = new LinkedHashMap<Long, Double>();
		for (long nodeId : nodeId2LinearIndexMap.keySet()) nodeBetweenessCentrality.put(nodeId, bc.getVertexScore(nodeId));

		linkBetweenessCentrality = new LinkedHashMap<Long, Double>();
		for (long linkId : linkId2LinearIndexMap.keySet()) linkBetweenessCentrality.put(linkId, bc.getEdgeScore(linkId));
	}
	
	private void computeSPDistanceMetrics()
	{
		diameter = 0;
		averageSPLength = 0;
		heterogeneity = 0;
		
		Graph<Long, Long> aux_graph = getGraph_JUNG();
		Transformer<Long, Double> aux_nev = getCostTransformer();

		/* Compute network diameter using naïve Floyd-Warshall algorithm */
		double[][] costMatrix = new double[N][N];
		for(int n = 0; n < N; n++)
		{
			Arrays.fill(costMatrix[n], Double.MAX_VALUE);
			costMatrix[n][n] = 0;
		}
		
		for(long edge : aux_graph.getEdges())
		{
			int a_e = nodeId2LinearIndexMap.get(aux_graph.getSource(edge));
			int b_e = nodeId2LinearIndexMap.get(aux_graph.getDest(edge));
			double newCost = aux_nev.transform(edge);
			if (newCost < costMatrix[a_e][b_e]) costMatrix[a_e][b_e] = newCost;
		}
		
		for(int k = 0; k < N; k++)
		{
			for(int i = 0; i < N; i++)
			{
				if (i == k) continue;
				
				for(int j = 0; j < N; j++)
				{
					if (j == k || j == i) continue;
					
					double newValue = costMatrix[i][k] + costMatrix[k][j];
					if (newValue < costMatrix[i][j]) costMatrix[i][j] = newValue;
				}
			}
		}
		
		int numPaths = 0;
		double sum = 0;
		double M = 0.0;
		double S = 0.0;

		for(int i = 0; i < N; i++)
		{
			for(int j = i + 1; j < N; j++)
			{
				double dist_ij = costMatrix[i][j];
				if (dist_ij < Double.MAX_VALUE)
				{
					sum += dist_ij;
					numPaths++;

					double tmpM = M;
					M += (dist_ij - tmpM) / numPaths;
					S += (dist_ij - tmpM) * (dist_ij - M);

					if (dist_ij > diameter) diameter = dist_ij;
				}
				
				double dist_ji = costMatrix[j][i];
				if (dist_ji < Double.MAX_VALUE)
				{
					sum += dist_ji;
					numPaths++;

					double tmpM = M;
					M += (dist_ji - tmpM) / numPaths;
					S += (dist_ji - tmpM) * (dist_ji - M);
					
					if (dist_ji > diameter) diameter = dist_ji;
				}
			}
		}
		
		if (numPaths == 0) return;
		
		averageSPLength = numPaths == 0 ? 0 : sum / numPaths;
		heterogeneity = averageSPLength == 0 ? 0 : Math.sqrt(S / numPaths) / averageSPLength;
	}

	/**
	 * Re-configures link cost setting. Related information, such as shortest paths, is cleared.
	 * 
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @since 0.3.0
	 */
	public void configureLinkCostMap(Map<Long, Double> linkCostMap)
	{
		if (linkCostMap == null) this.costMap = null;
		else this.costMap = new LinkedHashMap<Long, Double>(linkCostMap);

		averageSPLength = -1;
		diameter = -1;
		heterogeneity = -1;
		nev = null;
	}

	/**
	 * Returns the adjacency matrix of the network. The adjacency matrix is a
	 * <i>NxN</i> matrix (where <i>N</i> is the number of nodes in the network),
	 * where each position (<i>i</i>,<i>j</i>) represents the number of directed
	 * links from <i>i</i> to <i>j</i>.
	 *
	 * @return Adjacency matrix
	 * @since 0.3.1
	 */
	private DoubleMatrix2D getAdjacencyMatrix()
	{
		if (adjacencyMatrix == null)
		{
			adjacencyMatrix = DoubleFactory2D.sparse.make(N, N);

			for (long linkId : linkMap.keySet())
			{
				long originNodeId = linkMap.get(linkId).getFirst();
				long destinationNodeId = linkMap.get(linkId).getSecond();
				int a_e = nodeId2LinearIndexMap.get(originNodeId);
				int b_e = nodeId2LinearIndexMap.get(destinationNodeId);
				adjacencyMatrix.setQuick(a_e, b_e, adjacencyMatrix.get(a_e, b_e) + 1);
			}
		}

		return adjacencyMatrix;
	}

	/**
	 * Returns the eigenvalues of the adjacency matrix.
	 *
	 * @return Eigenvalues of the adjacency matrix
	 * @since 0.2.0
	 */
	private double[] getAdjacencyMatrixEigenvalues()
	{
		if (adjacencyMatrixEigenvalues == null)
		{
			DoubleMatrix2D A_nn = getAdjacencyMatrix().copy();
			A_nn.assign(A_nn.viewDice(), DoubleFunctions.max);

			DenseDoubleAlgebra alg = new DenseDoubleAlgebra();
			DenseDoubleEigenvalueDecomposition eig = alg.eig(A_nn);

			adjacencyMatrixEigenvalues = eig.getRealEigenvalues().toArray();
			DoubleUtils.sort(adjacencyMatrixEigenvalues, Constants.OrderingType.ASCENDING);
		}

		return adjacencyMatrixEigenvalues;
	}

	/**
	 * <p>Returns the algebraic connectivity of the network. The algebraic connectivity
	 * is equal to the second smallest eigenvalue of the laplacian matrix.</p>
	 *
	 * <p>For symmetric (or undirected) networks, if the algebraic connectivity
	 * is different from zero, it is ensured that the network is connected, that is,
	 * it is possible to find a path between each node pair.
	 *
	 * @return Algebraic connectivity
	 * @since 0.2.0
	 */
	public double getAlgebraicConnectivity()
	{
		double[] eig = getLaplacianMatrixEigenvalues();
		return eig[1];
	}

	/**
	 * Returns the assortativity of the network.
	 *
	 * @return Assortativity
	 * @since 0.2.0
	 */
	public double getAssortativity()
	{
		if (E == 0) return 0;

		Map<Long, Integer> aux_outNodeDegree = getOutNodeDegree();

		double a = 0;
		double b = 0;
		double y = 0;

		for (long linkId : linkIds)
		{
			Pair<Long, Long> aux = linkMap.get(linkId);
			long originNodeId = aux.getFirst();
			long destinationNodeId = aux.getSecond();

			int j_e = aux_outNodeDegree.get(originNodeId);
			int k_e = aux_outNodeDegree.get(destinationNodeId);

			y += j_e + k_e;
			a += j_e * k_e;
			b += j_e * j_e + k_e * k_e;
		}

		y /= 2.0D * E;
		y *= y;
		a /= E;
		b /= 2.0D * E;

		return (a - y) / (b - y);
	}

	/**
	 * Returns the average neighbor connectivity.
	 *
	 * @return Average neighbor connectivity
	 * @since 0.2.0
	 */
	public double getAverageNeighborConnectivity()
	{
		if (E == 0) return 0;

		Map<Long, Integer> aux_outNodeDegree = getOutNodeDegree();

		int maxNodeDegree = IntUtils.maxValue(aux_outNodeDegree);

		double[] knn = new double[maxNodeDegree + 1];
		DoubleMatrix2D m = DoubleFactory2D.sparse.make(maxNodeDegree + 1, maxNodeDegree + 1);

		for (long linkId : linkIds)
		{
			Pair<Long, Long> aux = linkMap.get(linkId);
			long originNodeId = aux.getFirst();
			long destinationNodeId = aux.getSecond();

			int degree_k1 = aux_outNodeDegree.get(originNodeId);
			int degree_k2 = aux_outNodeDegree.get(destinationNodeId);

			m.set(degree_k1, degree_k2, m.get(degree_k1, degree_k2) + 1);
		}

		for (int k_1 = 1; k_1 <= maxNodeDegree; k_1++)
		{
			knn[k_1] = k_1 * m.viewRow(k_1).zSum();
		}

		return DoubleUtils.averageNonZeros(knn) / (E - 1);
	}

	/**
	 * Returns the average number of outgoing links per node.
	 *
	 * @return Average number of outgoing links per node
	 * @since 0.2.0
	 */
	public double getAverageOutNodeDegree()
	{
		return IntUtils.average(getOutNodeDegree());
	}

	/**
	 * Returns the average shortest path distance among all node-pair shortest paths.
	 *
	 * @return Average shortest path distance
	 * @since 0.2.0
	 */
	public double getAverageShortestPathDistance()
	{
		if (averageSPLength == -1) computeSPDistanceMetrics();
		return averageSPLength;
	}

	/**
	 * Returns the average two-term reliability (A2TR) of the network. A2TR is computed
	 * as the ratio between the number of node-pair for which a path can be found
	 * and the same number when the network is connected (<i>Nx(N-1)</i>, where
	 * <i>N</i> is the number of nodes in the network). The value is in range [0, 1].
	 *
	 * @return Average two-term reliability
	 * @since 0.2.0
	 */
	public double getAverageTwoTermReliability()
	{
		if (E == 0) return 0;

		DirectedGraph<Long, Long> graph = getGraph_JGraphT();
		StrongConnectivityInspector<Long, Long> ci = new StrongConnectivityInspector<Long, Long>(graph);
		List<Set<Long>> connectedComponents = ci.stronglyConnectedSets();

		double sum = 0;
		Iterator<Set<Long>> it = connectedComponents.iterator();
		while (it.hasNext())
		{
			int componentSize = it.next().size();
			sum += componentSize * (componentSize - 1);
		}

		return sum / (N * (N - 1));
	}
	
	/**
	 * Returns the clustering coefficient of the network.
	 *
	 * @return Clustering coefficient
	 * @since 0.2.0
	 */
	public double getClusteringCoefficient()
	{
		if (E == 0) return 0;

		Map<Long, Integer> aux_outNodeDegree = getOutNodeDegree();

		Map<Long, Double> clusteringCoefficient = new LinkedHashMap<Long, Double>();
		for (long nodeId : nodeId2LinearIndexMap.keySet())
		{
			switch (aux_outNodeDegree.get(nodeId))
			{
				case 0:
					break;

				case 1:
					clusteringCoefficient.put(nodeId, 1.0);
					break;

				default:
					Collection<Long> neighbors = getNeighbors(nodeId);
					int aux = 0;
					for (long i : neighbors)
					{
						Collection<Long> aux_neighbors = getNeighbors(i);
						for (long j : neighbors)
						{
							if (i == j) continue;

							if (CollectionUtils.contains(aux_neighbors, j)) aux++;
						}
					}

					clusteringCoefficient.put(nodeId, (double) aux / neighbors.size());
					break;
			}
		}

		return DoubleUtils.average(clusteringCoefficient);
	}

	private Transformer<Long, Double> getCostTransformer()
	{
		if (nev == null) nev = JUNGUtils.getEdgeWeightTransformer(costMap);

		return nev;
	}

	/**
	 * Returns the density of the network. The density represents the ratio
	 * between the number of links in the network and the number of links needed
	 * to build a full-mesh network (<i>Nx(N-1)</i>, where <i>N</i> is the number of
	 * nodes in the network).
	 *
	 * @return Density
	 * @since 0.2.0
	 */
	public double getDensity()
	{
		if (N == 0) return 0;

		return (double) E / (N * (N - 1));
	}

	/**
	 * Returns the diameter of the network. The diameter is the longest path distance
	 * among all node-pair shortest paths.
	 *
	 * @return Network diameter
	 * @since 0.2.0
	 */
	public double getDiameter()
	{
		if (diameter == -1) computeSPDistanceMetrics();
		return diameter;
	}

	private DirectedGraph<Long, Long> getGraph_JGraphT()
	{
		if (graph_jgrapht == null) graph_jgrapht = (DirectedGraph<Long, Long>) JGraphTUtils.getGraphFromLinkMap(nodeIds, linkMap);
		return graph_jgrapht;
	}

	private Graph<Long, Long> getGraph_JUNG()
	{
		if (graph_jung == null) graph_jung = JUNGUtils.getGraphFromLinkMap(nodeIds, linkMap);

		return graph_jung;
	}

	/**
	 * Returns the heterogeneity of the network. The heterogeneity is equal to the
	 * standard deviation of all node-pair shortest paths divided by the average
	 * shortest path distance.
	 *
	 * @return Heterogeneity
	 * @since 0.2.0
	 */
	public double getHeterogeneity()
	{
		if (heterogeneity == -1) computeSPDistanceMetrics();
		return heterogeneity;
	}

	/**
	 * Returns the incidence matrix of the network. The incidence matrix is a
	 * <i>NxE</i> matrix (where <i>N</i> and <i>E</i> are the number of nodes
	 * and links in the network, respectively), where each position (<i>i</i>,
	 * <i>j</i>) is equal to: '1', if node <i>i</i> is the origin node of the link
	 * <i>j</i>; '-1', if node <i>i</i> is the destination node of the link
	 * <i>j</i>; and '0', otherwise.
	 *
	 * @return Incidence matrix
	 * @since 0.3.1
	 */
	private DoubleMatrix2D getIncidenceMatrix()
	{
		if (incidenceMatrix == null)
		{
			incidenceMatrix = DoubleFactory2D.sparse.make(N, E);
			for (Entry<Long, Integer> entry : linkId2LinearIndexMap.entrySet())
			{
				long linkId = entry.getKey();
				int e = entry.getValue();

				long originNodeId = linkMap.get(linkId).getFirst();
				long destinationNodeId = linkMap.get(linkId).getSecond();

				int a_e = nodeId2LinearIndexMap.get(originNodeId);
				int b_e = nodeId2LinearIndexMap.get(destinationNodeId);
				incidenceMatrix.setQuick(a_e, e, 1);
				incidenceMatrix.setQuick(b_e, e, -1);
			}
		}

		return incidenceMatrix;
	}

	/**
	 * Returns the laplacian matrix of the network. The laplacian matrix is equal
	 * to the product of the incidence matrix by its transpose matrix.
	 *
	 * @return Laplacian matrix
	 * @since 0.3.1
	 */
	private DoubleMatrix2D getLaplacianMatrix()
	{
		if (laplacianMatrix == null)
		{
			DoubleMatrix2D A_ne = getIncidenceMatrix().copy();
			laplacianMatrix = A_ne.zMult(A_ne.viewDice(), null);
		}

		return laplacianMatrix;
	}

	/**
	 * Returns the eigenvalues of the laplacian matrix of the network.
	 *
	 * @return Eigenvalues of the laplacian matrix
	 * @since 0.2.0
	 */
	private double[] getLaplacianMatrixEigenvalues()
	{
		if (laplacianMatrixEigenvalues == null)
		{
			DenseDoubleAlgebra alg = new DenseDoubleAlgebra();
			DenseDoubleEigenvalueDecomposition eig = alg.eig(getLaplacianMatrix());

			laplacianMatrixEigenvalues = eig.getRealEigenvalues().toArray();
			DoubleUtils.sort(laplacianMatrixEigenvalues, Constants.OrderingType.ASCENDING);
		}

		return laplacianMatrixEigenvalues;
	}

	/**
	 * <p>Returns the betweeness centrality of each link. The betweeness
	 * centrality of a link is equal to the number of node-pair shortest paths which traverses
	 * the link.</p>
	 *
	 * <p>Internally it makes use of the Brandes' algorithm.</p>
	 *
	 * @return Betweeness centrality of each link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkBetweenessCentrality()
	{
		if (linkBetweenessCentrality == null) computeBetweenessCentrality();

		return linkBetweenessCentrality;
	}

	/**
	 * <p>Returns the link connectivity. The link connectivity is equal to the smallest
	 * number of link-disjoint paths between each node pair.</p>
	 *
	 * <p>Internally it makes use of the Edmonds-Karp algorithm to compute the maximum
	 * flow between each node pair, assuming a link capacity equal to one for every link.</p>
	 *
	 * @return Link connectivity
	 * @since 0.2.0
	 */
	public int getLinkConnectivity()
	{
		if (E == 0) return 0;

		DirectedGraph<Long, Long> graph = getGraph_JGraphT();
		EdmondsKarpMaximumFlow<Long, Long> ek = new EdmondsKarpMaximumFlow(graph);
		int k = Integer.MAX_VALUE;

		for (long originNodeId : nodeIds)
		{
			for (long destinationNodeId : nodeIds)
			{
				if (originNodeId == destinationNodeId) continue;

				ek.calculateMaximumFlow(originNodeId, destinationNodeId);
				k = Math.min(k, ek.getMaximumFlowValue().intValue());

				if (k == 0) break;
			}
		}

		return k == Integer.MAX_VALUE ? 0 : k;
	}
	
	/**
	 * Returns the set of nodes reachable from a given node.
	 *
	 * @param nodeId Node identifier
	 * @return Set of reachable nodes
	 * @since 0.3.0
	 */
	public Collection<Long> getNeighbors(long nodeId)
	{
		return Collections.unmodifiableCollection(getGraph_JUNG().getSuccessors(nodeId));
	}

	/**
	 * <p>Returns the betweeness centrality of each node. The betweeness
	 * centrality of a node is equal to the number of node-pair shortest paths which traverses
	 * the node.</p>
	 *
	 * <p>Internally it makes use of the Brandes' algorithm.</p>
	 *
	 * @return Betweeness centrality of each node
	 * @since 0.3.0
	 */
	public Map<Long, Double> getNodeBetweenessCentrality()
	{
		if (nodeBetweenessCentrality == null) computeBetweenessCentrality();

		return nodeBetweenessCentrality;
	}
	
	/**
	 * Returns the node connectivity. The node connectivity is equal to the smallest
	 * number of node-disjoint paths between each node pair.
	 *
	 * <p>Internally it makes use of the (modified) Edmonds-Karp algorithm to compute the maximum
	 * flow between each node pair, assuming a link capacity equal to one for every link.</p>
	 *
	 * @return Node connectivity
	 * @since 0.2.0
	 */
	public int getNodeConnectivity()
	{
		if (E == 0) return 0;
		
		DirectedGraph<Long, Long> graph = getGraph_JGraphT();
		int k = Integer.MAX_VALUE;

		for (long originNodeId : nodeIds)
		{
			for (long destinationNodeId : nodeIds)
			{
				if (originNodeId == destinationNodeId) continue;
				
				DirectedGraph<Long, Long> auxGraph = (DirectedGraph<Long, Long>) JGraphTUtils.buildAuxiliaryNodeDisjointGraph(graph, originNodeId, destinationNodeId);
				EdmondsKarpMaximumFlow<Long, Long> ek = new EdmondsKarpMaximumFlow(auxGraph);
				ek.calculateMaximumFlow(originNodeId, destinationNodeId);
				k = Math.min(k, ek.getMaximumFlowValue().intValue());

				if (k == 0) break;
			}
		}

		return k == Integer.MAX_VALUE ? 0 : k;
	}

	/**
	 * Returns the number of outgoing links for each node.
	 *
	 * @return Number of outgoing links per node
	 * @since 0.2.0
	 */
	public Map<Long, Integer> getOutNodeDegree()
	{
		if (outNodeDegree == null)
		{
			Graph<Long, Long> aux_graph = getGraph_JUNG();
			outNodeDegree = new LinkedHashMap<Long, Integer>();

			for (long nodeId : nodeIds) outNodeDegree.put(nodeId, aux_graph.outDegree(nodeId));
		}

		return Collections.unmodifiableMap(outNodeDegree);
	}

	/**
	 * Returns the spectral radius of the network. The spectral radius is equal
	 * to the largest eigenvalue of the adjacency matrix.
	 *
	 * @return Spectral radius
	 * @since 0.2.0
	 */
	public double getSpectralRadius()
	{
		if (E == 0) return 0;

		double[] eig = getAdjacencyMatrixEigenvalues();
		return eig[eig.length - 1];
	}

	/**
	 * Returns the symmetry ratio. The symmetry ratio is equal to the number
	 * of distinct eigenvalues of the adjacency matrix divided by the network
	 * density plus one.
	 *
	 * @return Symmetry ratio
	 * @since 0.2.0
	 */
	public double getSymmetryRatio()
	{
		if (E == 0) return 0;

		double[] eig = getAdjacencyMatrixEigenvalues();
		return (double) DoubleUtils.unique(eig).length / (getDensity() + 1);
	}
}
