/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import cern.colt.list.tdouble.DoubleArrayList;
import cern.colt.list.tint.IntArrayList;
import cern.colt.matrix.tdouble.DoubleFactory2D;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import cern.colt.matrix.tdouble.algo.DenseDoubleAlgebra;
import com.net2plan.interfaces.networkDesign.Configuration;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.utils.ImageUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.Constants.CheckRoutingCycleType;
import com.net2plan.utils.Constants.RoutingCycleType;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.LongUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Quadruple;
import edu.uci.ics.jung.algorithms.filters.EdgePredicateFilter;
import edu.uci.ics.jung.algorithms.layout.Layout;
import edu.uci.ics.jung.algorithms.layout.StaticLayout;
import edu.uci.ics.jung.algorithms.shortestpath.DijkstraShortestPath;
import edu.uci.ics.jung.graph.DirectedOrderedSparseMultigraph;
import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.Context;
import edu.uci.ics.jung.graph.util.EdgeType;
import static edu.uci.ics.jung.graph.util.EdgeType.DIRECTED;
import edu.uci.ics.jung.visualization.Layer;
import edu.uci.ics.jung.visualization.VisualizationImageServer;
import edu.uci.ics.jung.visualization.control.LayoutScalingControl;
import edu.uci.ics.jung.visualization.decorators.EdgeShape;
import edu.uci.ics.jung.visualization.decorators.ToStringLabeller;
import edu.uci.ics.jung.visualization.renderers.DefaultVertexLabelRenderer;
import edu.uci.ics.jung.visualization.renderers.Renderer;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.util.*;
import java.util.Map.Entry;
import javax.swing.JComponent;
import org.apache.commons.collections15.ListUtils;
import org.apache.commons.collections15.Predicate;
import org.apache.commons.collections15.Transformer;
import org.apache.commons.collections15.functors.ConstantTransformer;
import org.apache.commons.collections15.functors.MapTransformer;
import org.jgrapht.DirectedGraph;
import org.jgrapht.UndirectedGraph;
import org.jgrapht.alg.ConnectivityInspector;
import org.jgrapht.alg.StrongConnectivityInspector;
import org.jgrapht.graph.AsWeightedGraph;
import org.jgrapht.graph.DirectedSubgraph;
import org.jgrapht.graph.DirectedWeightedMultigraph;
import org.jgrapht.graph.Subgraph;
import org.jgrapht.graph.UndirectedSubgraph;

/**
 * <p>Auxiliary static methods to work with graphs.</p>
 *
 * <p>These methods make intensive use of several Java libraries (i.e. <a href='#jom'>JOM</a>, <a href='#jgrapht'>JGraphT</a> or <a href='#jung'>JUNG</a>) hiding low-level details to users.</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 * @see <a name='jom'></a><a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
 * @see <a name='jgrapht'></a><a href='http://jgrapht.org/'>JGraphT website</a>
 * @see <a name='jung'></a><a href='http://jung.sourceforge.net/'>Java Universal Network/Graph Framework (JUNG) website</a>
 */
public class GraphUtils
{
	private GraphUtils() { }
	
	/**
	 * <p>Checks for validity of a given path (continuity and, optionally, no loops).</p>
	 *
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqLinks Sequence of traversed links
	 * @param checkCycles Indicates whether (and how) or not to check if there are cycles
	 * @since 0.3.0
	 */
	public static void checkRouteContinuity(Map<Long, Pair<Long, Long>> linkMap, List<Long> seqLinks, CheckRoutingCycleType checkCycles)
	{
		if (seqLinks.isEmpty()) throw new Net2PlanException("No path");

		if (checkCycles == CheckRoutingCycleType.NO_REPEAT_LINK && LongUtils.unique(seqLinks).size() != seqLinks.size())
			throw new Net2PlanException("There is a loop, seq. links = " + CollectionUtils.join(seqLinks, " => "));
		
		List<Long> seqNodes = convertSequenceOfLinksToSequenceOfNodes(linkMap, seqLinks);
		if (checkCycles == CheckRoutingCycleType.NO_REPEAT_NODE && LongUtils.unique(seqNodes).size() != seqNodes.size())
			throw new Net2PlanException("There is a loop, seq. links = " + CollectionUtils.join(seqLinks, " => ") + ", seq. nodes = " + CollectionUtils.join(seqNodes, " => "));
	}
	
	/**
	 * This function checks the validity of a destination-based routing (i.e IP routing). The destination-based routing is represented by the
	 * <i>f<sub>te</sub></i> variables. In particular, the function checks (i) if the routing has closed routing cycles
	 * (so that, the 100% of the traffic that enter these cycles, never reaches the destination), (ii) if the routing has open cycles (so that the traffic
	 * entering the cycle has in every loop, a positive probability of leaving the cycle). Note that if a routing has cycles of type (i) or (ii), it cannot be
	 * represented by a finite amount of routes in {@code Net2Plan}.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.1
	 */
	private static RoutingCycleType checkRouting_fte(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Pair<Long, Long>, Double> f_te)
	{
		/* For each target node t, detect if the routing is cycleless and well defined */
		boolean hasOpenCycles = false;
		for (long egressNodeId : nodeIds)
		{
			try { if (computeM_fte(egressNodeId, nodeIds, linkMap, f_te).getSecond() == RoutingCycleType.OPEN_CYCLES) hasOpenCycles = true; }
			catch(ClosedCycleRoutingException e) { return RoutingCycleType.CLOSED_CYCLES; }
		}

		return hasOpenCycles ? RoutingCycleType.OPEN_CYCLES : RoutingCycleType.LOOPLESS;
	}
	
	/**
	 * This function checks the validity of a destination-based routing (i.e IP routing). The destination-based routing is represented by the
	 * <i>f<sub>te</sub></i> variables. In particular, the function checks (i) if the routing has closed routing cycles
	 * (so that, the 100% of the traffic that enter these cycles, never reaches the destination), (ii) if the routing has open cycles (so that the traffic
	 * entering the cycle has in every loop, a positive probability of leaving the cycle). Note that if a routing has cycles of type (i) or (ii), it cannot be
	 * represented by a finite amount of routes in {@code Net2Plan}.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.1
	 */
	private static RoutingCycleType checkRouting_fte(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, DoubleMatrix2D f_te)
	{
		/* For each target node t, detect if the routing is cycleless and well defined */
		boolean hasOpenCycles = false;
		for (long egressNodeId : nodeIds)
		{
			try { if (computeM_fte(egressNodeId, nodeIds, linkMap, f_te).getSecond() == RoutingCycleType.OPEN_CYCLES) hasOpenCycles = true; }
			catch(ClosedCycleRoutingException e) { return RoutingCycleType.CLOSED_CYCLES; }
		}

		return hasOpenCycles ? RoutingCycleType.OPEN_CYCLES : RoutingCycleType.LOOPLESS;
	}
	
	/**
	 * Computes the Euclidean distance between two points.
	 *
	 * @param point1 Point 1
	 * @param point2 Point 2
	 * @return Euclidean distance between two points
	 * @since 0.3.0
	 */
	public static double computeEuclideanDistance(Point2D point1, Point2D point2)
	{
		return point1.distance(point2);
	}
	
	/**
	 * Computes the Haversine distance between two points, that is, the shortest 
	 * distance over the Earth's surface.
	 *
	 * <p><b>Important</b>: It is assumed an Earth's radius equal to {@link com.net2plan.utils.Constants#EARTH_RADIUS_IN_KM Constants.EARTH_RADIUS_IN_KM }</p>
	 *
	 * <p><b>Important</b>: Coordinates are assumed to be given in degress, conversions 
	 * to radians are made internally.</p>
	 *
	 * @param point1 Point 1 (x-coord is equal to longitude, y-coord is equal to latitude)
	 * @param point2 Point 2 (x-coord is equal to longitude, y-coord is equal to latitude)
	 * @return Haversine distance between two points
	 * @since 0.3.0
	 * @see <a href="http://www.movable-type.co.uk/scripts/latlong.html">Calculate distance, bearing and more between Latitude/Longitude points</a>
	 */
	public static double computeHaversineDistanceInKm(Point2D point1, Point2D point2)
	{
		double lon1 = point1.getX();
		double lat1 = point1.getY();
		double lon2 = point2.getX();
		double lat2 = point2.getY();
		
		double dLat = Math.toRadians(lat2-lat1);
		double dLon = Math.toRadians(lon2-lon1);
		double sindLat = Math.sin(dLat / 2);
		double sindLon = Math.sin(dLon / 2);
		double a = Math.pow(sindLat, 2) + Math.pow(sindLon, 2) * Math.cos(Math.toRadians(lat1)) * Math.cos(Math.toRadians(lat2));
		double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a));
		return Constants.EARTH_RADIUS_IN_KM * c;
	}
	
	/**
	 * Computes the fundamental matrix of the absorbing Markov chain represented by 
	 * a given demand-link routing for a given demand.
	 * 
	 * @param demandId Demand identifier
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
     * @param demandIds Set of demand identifiers. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_de For each demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandIds}, <i>d = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_de[d][e]} sets the fraction of the traffic from demand <i>d</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>b(d)</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return <i>N</i>x<i>N</i> fundamental matrix and {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.1
	 */
	private static Pair<DoubleMatrix2D, RoutingCycleType> computeM_fde(long demandId, Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Set<Long> demandIds, DoubleMatrix2D f_de)
	{
		int N = nodeIds.size();
		
		Set<Long> linkIds = linkMap.keySet();
		long[] linkIdsVector = LongUtils.toArray(linkIds);
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds);
		DenseDoubleAlgebra algebra = new DenseDoubleAlgebra();

		int d = demandId2LinearIndex.get(demandId);

		/* Compute the I-Q matrix */
		DoubleMatrix2D IminusQ = DoubleFactory2D.dense.identity(N);
		IntArrayList e_thisDemand = new IntArrayList();
		DoubleArrayList f_de_thisDemand = new DoubleArrayList();
		f_de.viewRow(d).getNonZeros(e_thisDemand, f_de_thisDemand);
		
		int elements = e_thisDemand.size();
		for(int i = 0; i < elements; i++)
		{
			int e = e_thisDemand.get(i);
			long linkId = linkIdsVector[e];

			Pair<Long, Long> aux = linkMap.get(linkId);
			long originNodeId = aux.getFirst();
			long destinationNodeId = aux.getSecond();

			int a_e = nodeId2LinearIndex.get(originNodeId);
			int b_e = nodeId2LinearIndex.get(destinationNodeId);
			IminusQ.set(a_e, b_e, IminusQ.get(a_e, b_e) - f_de_thisDemand.getQuick(i));
		}
		
		/* Compute the chain fundamental matrix M (IllegalArgumentException is catched if det(IminusQ) == 0 */
		DoubleMatrix2D M;
		try { M = algebra.inverse(IminusQ); }
		catch(IllegalArgumentException e) { throw new ClosedCycleRoutingException("Closed routing cycle for demand " + demandId); }

		for (int contN = 0; contN < N; contN++)
			if (Math.abs(M.get(contN, contN) - 1) > 1e-5)
				return Pair.of(M, RoutingCycleType.OPEN_CYCLES);
		
		return Pair.of(M, RoutingCycleType.LOOPLESS);
	}
	
	/**
	 * Computes the fundamental matrix of the absorbing Markov chain represented by 
	 * a given demand-link routing for a given demand.
	 * 
	 * @param demandId Demand identifier
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_de For each demand-outgoing link pair sets the fraction of the traffic from demand <i>d</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>b(d)</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return <i>N</i>x<i>N</i> fundamental matrix and {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.0
	 */
	private static Pair<DoubleMatrix2D, RoutingCycleType> computeM_fde(long demandId, Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Pair<Long, Long>, Double> f_de)
	{
		int N = nodeIds.size();

		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		DenseDoubleAlgebra algebra = new DenseDoubleAlgebra();

		/* Compute the I-Q matrix */
		DoubleMatrix2D IminusQ = DoubleFactory2D.dense.identity(N);
		for (Entry<Pair<Long, Long>, Double> entry : f_de.entrySet())
		{
			long demandId_thisForwardingRule = entry.getKey().getFirst();
			if (demandId_thisForwardingRule != demandId) continue;
			
			long linkId = entry.getKey().getSecond();
			Pair<Long, Long> aux = linkMap.get(linkId);
			long originNodeId = aux.getFirst();
			long destinationNodeId = aux.getSecond();

			int a_e = nodeId2LinearIndex.get(originNodeId);
			int b_e = nodeId2LinearIndex.get(destinationNodeId);
			IminusQ.set(a_e, b_e, IminusQ.get(a_e, b_e) - entry.getValue());
		}

		/* Compute the chain fundamental matrix M (IllegalArgumentException is catched if det(IminusQ) == 0 */
		DoubleMatrix2D M;
		try { M = algebra.inverse(IminusQ); }
		catch(IllegalArgumentException e) { throw new ClosedCycleRoutingException("Closed routing cycle for demand " + demandId); }
		
		for (int contN = 0; contN < N; contN++)
			if (Math.abs(M.get(contN, contN) - 1) > 1e-5)
				return Pair.of(M, RoutingCycleType.OPEN_CYCLES);
		
		return Pair.of(M, RoutingCycleType.LOOPLESS);
	}

	/**
	 * Computes the fundamental matrix of the absorbing Markov chain represented by 
	 * a given destination-based routing for a given egress node.
	 * 
	 * @param egressNodeId Egress node identifier
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i> and each link <i>e</i>, {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return <i>N</i>x<i>N</i> fundamental matrix and {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.0
	 */
	private static Pair<DoubleMatrix2D, RoutingCycleType> computeM_fte(long egressNodeId, Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Pair<Long, Long>, Double> f_te)
	{
		int N = nodeIds.size();
		
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		DenseDoubleAlgebra algebra = new DenseDoubleAlgebra();

		/* Compute the I-Q matrix */
		DoubleMatrix2D IminusQ = DoubleFactory2D.dense.identity(N);
		for (Entry<Pair<Long, Long>, Double> entry : f_te.entrySet())
		{
			long egressNodeId_thisForwardingRule = entry.getKey().getFirst();
			if (egressNodeId_thisForwardingRule != egressNodeId) continue;
			
			long linkId = entry.getKey().getSecond();
			Pair<Long, Long> aux = linkMap.get(linkId);
			long originNodeId = aux.getFirst();
			long destinationNodeId = aux.getSecond();

			int a_e = nodeId2LinearIndex.get(originNodeId);
			int b_e = nodeId2LinearIndex.get(destinationNodeId);
			IminusQ.set(a_e, b_e, IminusQ.get(a_e, b_e) - entry.getValue());
		}

		/* Compute the chain fundamental matrix M (IllegalArgumentException is catched if det(IminusQ) == 0) */
		DoubleMatrix2D M;
		try { M = algebra.inverse(IminusQ); }
		catch(IllegalArgumentException e) { throw new ClosedCycleRoutingException("Closed routing cycle for destination " + egressNodeId); }

		for (int contN = 0; contN < N; contN++)
			if (Math.abs(M.get(contN, contN) - 1) > 1e-5)
				return Pair.of(M, RoutingCycleType.OPEN_CYCLES);
		
		return Pair.of(M, RoutingCycleType.LOOPLESS);
	}
	
	/**
	 * Computes the fundamental matrix of the absorbing Markov chain represented by 
	 * a given destination-based routing for a given egress node.
	 * 
	 * @param egressNodeId Egress node identifier
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return <i>N</i>x<i>N</i> fundamental matrix and {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.1
	 */
	private static Pair<DoubleMatrix2D, RoutingCycleType> computeM_fte(long egressNodeId, Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, DoubleMatrix2D f_te)
	{
		int N = f_te.rows();
		
		Set<Long> linkIds = linkMap.keySet();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		long[] linkIdsVector = LongUtils.toArray(linkIds);
		DenseDoubleAlgebra algebra = new DenseDoubleAlgebra();

		int t = nodeId2LinearIndex.get(egressNodeId);

		/* Compute the I-Q matrix */
		DoubleMatrix2D IminusQ = DoubleFactory2D.dense.identity(N);
		IntArrayList e_thisDemand = new IntArrayList();
		DoubleArrayList f_te_thisDemand = new DoubleArrayList();
		f_te.viewRow(t).getNonZeros(e_thisDemand, f_te_thisDemand);
		
		int elements = e_thisDemand.size();
		for(int i = 0; i < elements; i++)
		{
			int e = e_thisDemand.get(i);
			long linkId = linkIdsVector[e];

			Pair<Long, Long> aux = linkMap.get(linkId);
			long originNodeId = aux.getFirst();
			long destinationNodeId = aux.getSecond();
			if (originNodeId == egressNodeId || destinationNodeId == egressNodeId) continue;

			int a_e = nodeId2LinearIndex.get(originNodeId);
			int b_e = nodeId2LinearIndex.get(destinationNodeId);
			IminusQ.set(a_e, b_e, IminusQ.get(a_e, b_e) - f_te_thisDemand.getQuick(i));
		}

		/* Compute the chain fundamental matrix M (IllegalArgumentException is catched if det(IminusQ) == 0 */
		DoubleMatrix2D M;
		try { M = algebra.inverse(IminusQ); }
		catch(IllegalArgumentException e) { throw new ClosedCycleRoutingException("Closed routing cycle for destination " + egressNodeId); }

		for (int contN = 0; contN < N; contN++)
			if (Math.abs(M.get(contN, contN) - 1) > 1e-5)
				return Pair.of(M, RoutingCycleType.OPEN_CYCLES);
		
		return Pair.of(M, RoutingCycleType.LOOPLESS);
	}
	
	/**
	 * Applies a correction to a given airline distance (i.e. Euclidean 
	 * distance, Haversine distance...) considering cable (i.e. optical fiber) 
	 * deployment issues to predict a more realistic cable length. For undersea 
	 * cables, the airline distance is considered as valid.
	 *
	 * @param airlineDistanceInKm Airline distance in km (i.e. Euclidean distance, Haversine distance...)
	 * @return Cable length in km
	 * @since 0.3.0
	 * @see <a href="http://www.etsi.org/deliver/etsi_en/300400_300499/300416/01.02.01_60/en_300416v010201p.pdf">ETSI, "Network Aspects (NA): Availability performance of path elements of international digital paths," European Standard EN 300 416 V1.2.1, August 1998.</a>
	 */
	public static double computeRoadDistanceInKm(double airlineDistanceInKm)
	{
		if (airlineDistanceInKm < 0) throw new Net2PlanException("Bad - Airline distance must be greater or equal than zero");
		else if (airlineDistanceInKm < 1000) return 1.5 * airlineDistanceInKm;
		else if (airlineDistanceInKm < 1200) return 1500;
		else return 1.25 * airlineDistanceInKm;
	}

	/**
	 * Returns the total cost for a given path, wjocj os equal to the sum of the cost of 
	 * each traversed link.
	 * 
	 * @param seqLinks Sequence of traversed links
	 * @param linkCostMap Cost per link (null means 'all one'), where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return Cost of the path
	 * @since 0.3.0
	 */
	public static double convertPath2PathCost(List<Long> seqLinks, Map<Long, Double> linkCostMap)
	{
		double pathCost = 0;
		for(long linkId : seqLinks) pathCost += linkCostMap == null ? 1 : linkCostMap.get(linkId);

		return pathCost;
	}

	/**
	 * Returns the total cost for a given a list of paths.
	 * 
	 * @param pathList List of paths, where each path is represented by its sequence of traversed links
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return Cost per path
	 * @since 0.3.0
	 */
	public static List<Double> convertPathList2PathCost(List<List<Long>> pathList, Map<Long, Double> linkCostMap)
	{
		List<Double> pathCost = new LinkedList<Double>();

		ListIterator<List<Long>> it = pathList.listIterator();
		while (it.hasNext())
		{
			List<Long> path = it.next();
			pathCost.add(convertPath2PathCost(path, linkCostMap));
		}

		return pathCost;
	}
	
	/**
	 * Converts a given sequence of links to the corresponding sequence of nodes.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqLinks Sequence of traversed links
	 * @return Sequence of nodes
	 * @since 0.3.0
	 */
	public static List<Long> convertSequenceOfLinksToSequenceOfNodes(Map<Long, Pair<Long, Long>> linkMap, List<Long> seqLinks)
	{
		if (seqLinks.isEmpty()) throw new Net2PlanException("No path");
		
		List<Long> sequenceOfNodes = new LinkedList<Long>();
		boolean isFirstLink = true;
		long lastVisitedNodeId = -1;
		for(long linkId : seqLinks)
		{
			Pair<Long, Long> aux = linkMap.get(linkId);
			if (aux == null) throw new Net2PlanException("Invalid sequence of links");
			
			long originNodeId = aux.getFirst();
			long destinationNodeId = aux.getSecond();
			
			if (isFirstLink)
			{
				sequenceOfNodes.add(originNodeId);
				isFirstLink = false;
			}
			else
			{
				if (lastVisitedNodeId != originNodeId)
				{
					throw new RuntimeException("Route is not feasible (destination node of each link must be the origin node of the following link)");
				}
			}
			
			sequenceOfNodes.add(destinationNodeId);
			lastVisitedNodeId = destinationNodeId;
		}
		
		return sequenceOfNodes;
	}

	/**
	 * Given a forwarding rule mapping (fractions of traffic entering a node from 
	 * demand 'd', leaving that node through link 'e'), and an offered 
	 * traffic to the network, it generates the resulting demand-link routing in 
	 * the form x_de (amount of traffic from demand d, transmitted through link e).

	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param h_d The amount of traffic offered for each demand
	 * @param f_de For each demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandIds}, <i>d = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_de[d][e]} sets the fraction of the traffic from demand <i>d</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>b(d)</i>, the sum of the fractions <i>f<sub>de</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e), total carried traffic per demand, carried traffic per link, and routing cycle type (loopless, with open cycles...) per demand
	 * @since 0.3.1
	 */
	public static Quadruple<DoubleMatrix2D, double[], double[], RoutingCycleType[]> convert_fde2xde(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Double> h_d, DoubleMatrix2D f_de)
	{
		int E = linkMap.size();
		int D = demandMap.size();
		
		Set<Long> linkIds = linkMap.keySet();
		Set<Long> demandIds = demandMap.keySet();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds);
		
		DoubleMatrix2D x_de = DoubleFactory2D.sparse.make(D, E);
		double[] r_d = new double[D];
		double[] y_e = new double[E];
		RoutingCycleType[] routingCycleType = new RoutingCycleType[D];
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));

		/* For each demand 'd', detect if the routing is cycleless and well defined */
		for (long demandId : demandIds)
		{
			final int d = demandId2LinearIndex.get(demandId);

			Pair<DoubleMatrix2D, RoutingCycleType> out = computeM_fde(demandId, nodeIds, linkMap, demandIds, f_de);
			DoubleMatrix2D M = out.getFirst();
			routingCycleType[d] = out.getSecond();
			
			final Pair<Long, Long> ingressEgressNodePair = demandMap.get(demandId);
			final long ingressNodeId = ingressEgressNodePair.getFirst();
			final long egressNodeId = ingressEgressNodePair.getSecond();
			final double h_d_thisDemand = h_d.get(demandId);

			final int a_d = nodeId2LinearIndex.get(ingressNodeId);
			final int b_d = nodeId2LinearIndex.get(egressNodeId);
			
			for (long linkId : linkIds)
			{
				long originNodeId = linkMap.get(linkId).getFirst();
				
				int e = linkId2LinearIndex.get(linkId);
				int a_e = nodeId2LinearIndex.get(originNodeId);
				
				x_de.setQuick(d, e, h_d_thisDemand * M.getQuick(a_d, a_e) * f_de.getQuick(d, e));
				if (x_de.getQuick(d, e) < PRECISION_FACTOR) x_de.setQuick(d, e, 0);
				
				y_e[e] += x_de.getQuick(d, e);
			}
			
			r_d[d] = h_d_thisDemand * M.getQuick(a_d, b_d);
			if (r_d[d] < PRECISION_FACTOR) r_d[d] = 0;
		}

		return Quadruple.of(x_de, r_d, y_e, routingCycleType);
	}
	
	/**
	 * Given a forwarding rule mapping (fractions of traffic entering a node from 
	 * demand 'd', leaving that node through link 'e'), and an offered 
	 * traffic to the network, it generates the resulting demand-link routing in 
	 * the form x_de (amount of traffic from demand d, transmitted through link e).

	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param h_d The amount of traffic offered for each demand
	 * @param f_de For each demand-outgoing link pair sets the fraction of the traffic from demand <i>d</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>b(d)</i>, the sum of the fractions <i>f<sub>de</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e), total carried traffic per demand, carried traffic per link, and routing cycle type (loopless, with open cycles...) per demand
	 * @since 0.3.0
	 */
	public static Quadruple<Map<Pair<Long, Long>, Double>, Map<Long, Double>, Map<Long, Double>, List<RoutingCycleType>> convert_fde2xde(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Double> h_d, Map<Pair<Long, Long>, Double> f_de)
	{
		Set<Long> linkIds = linkMap.keySet();
		Set<Long> demandIds = demandMap.keySet();
	
		Map<Pair<Long, Long>, Double> x_de = new LinkedHashMap<Pair<Long, Long>, Double>();
		Map<Long, Double> r_d = new LinkedHashMap<Long, Double>();
		Map<Long, Double> y_e = DoubleUtils.zeros(linkIds);
		List<RoutingCycleType> routingCycleType = new LinkedList<RoutingCycleType>();
		
		/* For each demand 'd', detect if the routing is cycleless and well defined */
		for (long demandId : demandIds)
		{
			Quadruple<Map<Pair<Long, Long>, Double>, Double, Map<Long, Double>, RoutingCycleType> out_thisDemand = convert_fde2xde(demandId, nodeIds, linkMap, demandMap, h_d, f_de);
			x_de.putAll(out_thisDemand.getFirst());
			r_d.put(demandId, out_thisDemand.getSecond());
			routingCycleType.add(out_thisDemand.getFourth());
			Map<Long, Double> y_e_thisDemand = out_thisDemand.getThird();
			for(Entry<Long, Double> entry : y_e_thisDemand.entrySet())
			{
				long linkId = entry.getKey();
				double y_e_thisDemand_thisLink = entry.getValue();
				y_e.put(linkId, y_e.get(linkId) + y_e_thisDemand_thisLink);
			}
		}
		
		return Quadruple.of(x_de, r_d, y_e, routingCycleType);
	}
	
	/**
	 * Given a forwarding rule mapping (fractions of traffic entering a node from 
	 * demand 'd', leaving that node through link 'e'), and an offered 
	 * traffic to the network, it generates the resulting demand-link routing in 
	 * the form x_de (amount of traffic from demand d, transmitted through link e), 
	 * only for a given demand.
	 * 
	 * @param demandId Demand identiifer
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param h_d The amount of traffic offered for each demand
	 * @param f_de For each demand-outgoing link pair sets the fraction of the traffic from demand <i>d</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>b(d)</i>, the sum of the fractions <i>f<sub>de</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e), total carried traffic, carried traffic per link (including only those traversed by the demand), and routing cycle type (loopless, with open cycles...)
	 * @since 0.3.0
	 */
	public static Quadruple<Map<Pair<Long, Long>, Double>, Double, Map<Long, Double>, RoutingCycleType> convert_fde2xde(long demandId, Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Double> h_d, Map<Pair<Long, Long>, Double> f_de)
	{
		Set<Long> linkIds = linkMap.keySet();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		
		Map<Pair<Long, Long>, Double> x_de = new LinkedHashMap<Pair<Long, Long>, Double>();
		Map<Long, Double> y_e = new LinkedHashMap<Long, Double>();
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));

		Pair<DoubleMatrix2D, RoutingCycleType> out = computeM_fde(demandId, nodeIds, linkMap, f_de);
		DoubleMatrix2D M = out.getFirst();
		RoutingCycleType routingCycleType = out.getSecond();
		
		final Pair<Long, Long> ingressEgressNodePair = demandMap.get(demandId);
		final long ingressNodeId = ingressEgressNodePair.getFirst();
		final long egressNodeId = ingressEgressNodePair.getSecond();
		final double h_d_thisDemand = h_d.get(demandId);

		final int a_d = nodeId2LinearIndex.get(ingressNodeId);
		final int b_d = nodeId2LinearIndex.get(egressNodeId);
		for (long linkId : linkIds)
		{
			Pair<Long, Long> demandPairLink = Pair.of(demandId, linkId);
			Double f_de_thisLink = f_de.get(demandPairLink);
			if (f_de_thisLink == null) continue;

			long originNodeId = linkMap.get(linkId).getFirst();
			int a_e = nodeId2LinearIndex.get(originNodeId);
			double x_de_thisLink = h_d_thisDemand * M.getQuick(a_d, a_e) * f_de.get(demandPairLink);
			if (x_de_thisLink < PRECISION_FACTOR) x_de_thisLink = 0;
			y_e.put(linkId, x_de_thisLink);

			if (x_de_thisLink == 0) continue;
			x_de.put(demandPairLink, x_de_thisLink);
		}

		double r_d = h_d_thisDemand * M.getQuick(a_d, b_d);
		if (r_d < PRECISION_FACTOR) r_d = 0;

		return Quadruple.of(x_de, r_d, y_e, routingCycleType);
	}
	
	/**
	 * Given a destination-based routing in the form f_te (fractions of traffic in 
	 * a node, that is forwarded through each of its output links), it generates 
	 * the resulting demand-link routing in the form f_de (fractions of traffic in 
	 * a node from demand d, transmitted through link e).
	 * 
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i>, and each link <i>e</i>, {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return Demand-link routing in the form f_de (fractions of traffic in a node from demand d, transmitted through link e)
	 * @since 0.3.0
	 */
	public static Map<Pair<Long, Long>, Double> convert_fte2fde(Map<Long, Pair<Long, Long>> demandMap, Map<Pair<Long, Long>, Double> f_te)
	{
		Map<Pair<Long, Long>, Double> f_de = new LinkedHashMap<Pair<Long, Long>, Double>();
		
		Map<Long, Set<Long>> demandIdsPerEgressNodeId = new LinkedHashMap<Long, Set<Long>>();
		for(Entry<Long, Pair<Long, Long>> entry : demandMap.entrySet())
		{
			long egressNodeId = entry.getValue().getSecond();
			Set<Long> demandIds_thisEgressNode = demandIdsPerEgressNodeId.get(egressNodeId);
			if (demandIds_thisEgressNode == null)
			{
				demandIds_thisEgressNode = new LinkedHashSet<Long>();
				demandIdsPerEgressNodeId.put(egressNodeId, demandIds_thisEgressNode);
			}
			
			demandIds_thisEgressNode.add(entry.getKey());
		}
		
		for(Entry<Pair<Long, Long>, Double> entry : f_te.entrySet())
		{
			long egressNodeId = entry.getKey().getFirst();
			long linkId = entry.getKey().getSecond();
			double f_te_thisEgressNode_thisLink = entry.getValue();
			
			for(long demandId : demandIdsPerEgressNodeId.get(egressNodeId))
				f_de.put(Pair.of(demandId, linkId), f_te_thisEgressNode_thisLink);
		}
		
		return f_de;
	}
	
	/**
	 * Given a destination-based routing in the form f_te (fractions of traffic in 
	 * a node, that is forwarded through each of its output links), it generates 
	 * the resulting demand-link routing in the form f_de (fractions of traffic in 
	 * a node from demand d, transmitted through link e).
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return Demand-link routing in the form f_de (fractions of traffic in a node from demand d, transmitted through link e)
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D convert_fte2fde(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> demandMap, DoubleMatrix2D f_te)
	{
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandMap.keySet());
		
		int E = f_te.columns();
		int D = demandMap.size();
		DoubleMatrix2D f_de = DoubleFactory2D.sparse.make(D, E);
		for(Entry<Long, Pair<Long, Long>> entry : demandMap.entrySet())
		{
			long egressNodeId = entry.getValue().getSecond();
			int t = nodeId2LinearIndex.get(egressNodeId);
			
			long demandId = entry.getKey();
			int d = demandId2LinearIndex.get(demandId);
			f_de.viewRow(d).assign(f_te.viewRow(t));
		}
		
		return f_de;
	}
	
	/**
	 * Given a destination-based routing in the form f_te (fractions of traffic in 
	 * a node, that is forwarded through each of its output links), and an offered 
	 * traffic to the network, it generates the resulting demand-link routing in 
	 * the form x_de (amount of traffic from demand d, transmitted through link e).
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param h_d The amount of traffic offered for each demand
	 * @param f_te For each destination node <i>t</i> and each link <i>e</i>, {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @return Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e)
	 * @since 0.3.0
	 */
	public static Quadruple<Map<Pair<Long, Long>, Double>, Map<Long, Double>, Map<Long, Double>, List<RoutingCycleType>> convert_fte2xde(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Double> h_d, Map<Pair<Long, Long>, Double> f_te)
	{
		int N = nodeIds.size();
		int E = linkMap.size();
		int D = demandMap.size();
		
		Map<Pair<Long, Long>, Double> x_de = new LinkedHashMap<Pair<Long, Long>, Double>();
		Map<Long, Double> r_d = DoubleUtils.constantMap(demandMap.keySet(), 0);
		Map<Long, Double> y_e = DoubleUtils.constantMap(linkMap.keySet(), 0);
		Map<Long, RoutingCycleType> routingCycleType = new TreeMap<Long, RoutingCycleType>();
		if (N == 0 || E == 0 || D == 0) return Quadruple.of(x_de, r_d, y_e, (List<RoutingCycleType>) new LinkedList<RoutingCycleType>(routingCycleType.values()));

		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Set<Long>> demandIdsPerEgressNodeId = new LinkedHashMap<Long, Set<Long>>();
		for(Entry<Long, Pair<Long, Long>> entry : demandMap.entrySet())
		{
			long egressNodeId = entry.getValue().getSecond();
			Set<Long> demandIds_thisEgressNode = demandIdsPerEgressNodeId.get(egressNodeId);
			if (demandIds_thisEgressNode == null)
			{
				demandIds_thisEgressNode = new LinkedHashSet<Long>();
				demandIdsPerEgressNodeId.put(egressNodeId, demandIds_thisEgressNode);
			}
			
			demandIds_thisEgressNode.add(entry.getKey());
		}
		
		Set<Long> linkIds = linkMap.keySet();
		for(Entry<Long, Set<Long>> entry : demandIdsPerEgressNodeId.entrySet())
		{
			long egressNodeId = entry.getKey();
			int t = nodeId2LinearIndex.get(egressNodeId);
			
			Pair<DoubleMatrix2D, RoutingCycleType> out = computeM_fte(egressNodeId, nodeIds, linkMap, f_te);
			DoubleMatrix2D M = out.getFirst();
			RoutingCycleType routingCycleType_thisEgressNode = out.getSecond();

			Set<Long> demandIds_thisEgressNode = entry.getValue();
			for(long demandId : demandIds_thisEgressNode)
			{
				long ingressNodeId = demandMap.get(demandId).getFirst();
				int a_d = nodeId2LinearIndex.get(ingressNodeId);
				double h_d_thisDemand = h_d.get(demandId);
				for (long linkId : linkIds)
				{
					Pair<Long, Long> egressNodeLinkPair = Pair.of(egressNodeId, linkId);
					Double f_te_thisLink = f_te.get(egressNodeLinkPair);
					if (f_te_thisLink == null) continue;

					long originNodeId = linkMap.get(linkId).getFirst();
					int a_e = nodeId2LinearIndex.get(originNodeId);
					double x_de_thisLink = h_d_thisDemand * M.getQuick(a_d, a_e) * f_te.get(egressNodeLinkPair);
					if (x_de_thisLink < PRECISION_FACTOR) x_de_thisLink = 0;

					if (x_de_thisLink == 0) continue;
					y_e.put(linkId, y_e.get(linkId) + x_de_thisLink);

					Pair<Long, Long> demandLinkPair = Pair.of(demandId, linkId);
					x_de.put(demandLinkPair, x_de_thisLink);
				}

				double r_d_thisDemand = h_d_thisDemand * M.getQuick(a_d, t);
				if (r_d_thisDemand < PRECISION_FACTOR) r_d_thisDemand = 0;
				
				r_d.put(demandId, r_d_thisDemand);
				routingCycleType.put(demandId, routingCycleType_thisEgressNode);
			}
		}
		
		return Quadruple.of(x_de, r_d, y_e, (List<RoutingCycleType>) new LinkedList<RoutingCycleType>(routingCycleType.values()));
	}
	
	/**
	 * Given a destination-based routing in the form f_te (fractions of traffic in a node, that is forwarded through each of its output links), and
	 * an offered traffic to the network, it generates the resulting set of paths that are produced.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param h_d The amount of traffic offered for each demand
	 * @param f_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @param d_p (Output parameter) Demand corresponding to each path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param x_p (Output parameter) Carried traffic per path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param pathList (Output parameter) List of paths, where each path is represented by its sequence of traversed links. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @since 0.3.1
	 */
	public static void convert_fte2xp(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Double> h_d, Map<Pair<Long, Long>, Double> f_te, List<Long> d_p, List<Double> x_p, List<List<Long>> pathList)
	{
		RoutingCycleType routingCycleType = checkRouting_fte(nodeIds, linkMap, f_te);
		if (routingCycleType == RoutingCycleType.OPEN_CYCLES) throw new Net2PlanException("There are some open cycles");
		if (routingCycleType == RoutingCycleType.CLOSED_CYCLES) throw new Net2PlanException("There are some closed cycles");

		Set<Long> demandIds = demandMap.keySet();

		/* For each demand */
		long nextPathId = 0;
		Map<Long, LinkedList<Long>> aux_seqLinks_p = new LinkedHashMap<Long, LinkedList<Long>>();
		Map<Long, Double> aux_x_p = new LinkedHashMap<Long, Double>();
		for (long demandId : demandIds)
		{
			long ingressNodeId = demandMap.get(demandId).getFirst();
			long egressNodeId = demandMap.get(demandId).getSecond();
			double offeredTraffic_thisDemand = h_d.get(demandId);
			LinkedList<Long> activePaths = new LinkedList<Long>();
			Set<Long> outLinks_ingressNode = getNodeOutgoingLinks(linkMap, ingressNodeId);

			for (long linkId : outLinks_ingressNode)
			{
				double f_te_thisEgressNode_thisLink = f_te.get(Pair.of(egressNodeId, linkId));
				if (f_te_thisEgressNode_thisLink > 0)
				{
					long pathId = nextPathId;
					LinkedList<Long> seqLinks = new LinkedList<Long>(); seqLinks.add(linkId);
					aux_seqLinks_p.put(pathId, seqLinks);
					aux_x_p.put(pathId, offeredTraffic_thisDemand * f_te_thisEgressNode_thisLink);
					d_p.add(demandId);
					if (linkMap.get(linkId).getSecond() != egressNodeId) activePaths.add(pathId);

					nextPathId++;
				}
			}

			while (activePaths.size() > 0)
			{
				final long pathId = activePaths.getFirst();
				final LinkedList<Long> seqLinks = aux_seqLinks_p.get(pathId);
				final double current_xp = aux_x_p.get(pathId);
				final long currentLastNode = linkMap.get(seqLinks.getLast()).getSecond();
				Set<Long> outLinks = getNodeOutgoingLinks(linkMap, currentLastNode);
				boolean alreadyOneOutputLinkWithTraffic = false;
				for (long linkId : outLinks)
				{
					double f_te_thisEgressNode_thisLink = f_te.get(Pair.of(egressNodeId, linkId));
					if (f_te_thisEgressNode_thisLink > 0)
					{
						if (!alreadyOneOutputLinkWithTraffic)
						{
							alreadyOneOutputLinkWithTraffic = true;
							seqLinks.add(linkId);
							aux_x_p.put(pathId, current_xp * f_te_thisEgressNode_thisLink);
							if (linkMap.get(linkId).getSecond() == egressNodeId)
							{
								activePaths.removeFirst();
							}
						}
						else
						{
							long newPathId = nextPathId;
							LinkedList<Long> extendedSeqLinks = new LinkedList<Long>(seqLinks);
							extendedSeqLinks.removeLast();
							extendedSeqLinks.add(linkId);
							aux_seqLinks_p.put(newPathId, extendedSeqLinks);
							aux_x_p.put(newPathId, current_xp * f_te_thisEgressNode_thisLink);
							d_p.add(demandId);
							if (linkMap.get(linkId).getSecond() != egressNodeId) activePaths.add(newPathId);

							nextPathId++;
						}
					}
				}

				if (alreadyOneOutputLinkWithTraffic == false)
				{
					throw new RuntimeException("Unexpected error");
				}
			}
		}

		x_p.addAll(aux_x_p.values());
		pathList.addAll(aux_seqLinks_p.values());
	}

	/**
	 * Given a destination-based routing in the form f_te (fractions of traffic in a node, that is forwarded through each of its output links), and
	 * an offered traffic to the network, it generates the resulting set of paths that are produced.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param h_d The amount of traffic offered for each demand
	 * @param f_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @param d_p (Output parameter) Demand corresponding to each path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param x_p (Output parameter) Carried traffic per path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param pathList (Output parameter) List of paths, where each path is represented by its sequence of traversed links. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @since 0.3.1
	 */
	public static void convert_fte2xp(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Double> h_d, DoubleMatrix2D f_te, List<Long> d_p, List<Double> x_p, List<List<Long>> pathList)
	{
		RoutingCycleType routingCycleType = checkRouting_fte(nodeIds, linkMap, f_te);
		if (routingCycleType == RoutingCycleType.OPEN_CYCLES) throw new Net2PlanException("There are some open cycles");
		if (routingCycleType == RoutingCycleType.CLOSED_CYCLES) throw new Net2PlanException("There are some closed cycles");

		Set<Long> linkIds = linkMap.keySet();
		Set<Long> demandIds = demandMap.keySet();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);

		/* For each demand */
		long nextPathId = 0;
		Map<Long, LinkedList<Long>> aux_seqLinks_p = new LinkedHashMap<Long, LinkedList<Long>>();
		Map<Long, Double> aux_x_p = new LinkedHashMap<Long, Double>();
		for (long demandId : demandIds)
		{
			long ingressNodeId = demandMap.get(demandId).getFirst();
			long egressNodeId = demandMap.get(demandId).getSecond();
			double offeredTraffic_thisDemand = h_d.get(demandId);
			LinkedList<Long> activePaths = new LinkedList<Long>();
			Set<Long> outLinks_ingressNode = getNodeOutgoingLinks(linkMap, ingressNodeId);

			int b_d = nodeId2LinearIndex.get(egressNodeId);

			for (long linkId : outLinks_ingressNode)
			{
				int e = linkId2LinearIndex.get(linkId);

				if (f_te.getQuick(b_d, e) > 0)
				{
					long pathId = nextPathId;
					LinkedList<Long> seqLinks = new LinkedList<Long>(); seqLinks.add(linkId);
					aux_seqLinks_p.put(pathId, seqLinks);
					aux_x_p.put(pathId, offeredTraffic_thisDemand * f_te.getQuick(b_d, e));
					d_p.add(demandId);
					if (linkMap.get(linkId).getSecond() != egressNodeId) activePaths.add(pathId);

					nextPathId++;
				}
			}

			while (activePaths.size() > 0)
			{
				final long pathId = activePaths.getFirst();
				final LinkedList<Long> seqLinks = aux_seqLinks_p.get(pathId);
				final double current_xp = aux_x_p.get(pathId);
				final long currentLastNode = linkMap.get(seqLinks.getLast()).getSecond();
				Set<Long> outLinks = getNodeOutgoingLinks(linkMap, currentLastNode);
				boolean alreadyOneOutputLinkWithTraffic = false;
				for (long linkId : outLinks)
				{
					int e = linkId2LinearIndex.get(linkId);
					
					if (f_te.getQuick(b_d, e) > 0)
					{
						if (!alreadyOneOutputLinkWithTraffic)
						{
							alreadyOneOutputLinkWithTraffic = true;
							seqLinks.add(linkId);
							aux_x_p.put(pathId, current_xp * f_te.getQuick(b_d, e));
							if (linkMap.get(linkId).getSecond() == egressNodeId)
							{
								activePaths.removeFirst();
							}
						}
						else
						{
							long newPathId = nextPathId;
							LinkedList<Long> extendedSeqLinks = new LinkedList<Long>(seqLinks);
							extendedSeqLinks.removeLast();
							extendedSeqLinks.add(linkId);
							aux_seqLinks_p.put(newPathId, extendedSeqLinks);
							aux_x_p.put(newPathId, current_xp * f_te.getQuick(b_d, e));
							d_p.add(demandId);
							if (linkMap.get(linkId).getSecond() != egressNodeId) activePaths.add(newPathId);

							nextPathId++;
						}
					}
				}

				if (alreadyOneOutputLinkWithTraffic == false)
				{
					throw new RuntimeException("Unexpected error");
				}
			}
		}

		x_p.addAll(aux_x_p.values());
		pathList.addAll(aux_seqLinks_p.values());
	}

	/**
	 * Given a destination-based routing in the form f_te (fractions of traffic in a node, that is forwarded through each of its output links), and
	 * an offered traffic to the network, it generates the resulting destination-based
	 * routing in the form x_te (amount of traffic targeted to node t, transmitted through link e).
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @param trafficMatrix Traffic matrix
	 * @return Destination-based routing in the form <i>f<sub>te</sub></i> (fractions of traffic in a node, that is forwarded through each of its output links)
	 * @since 0.3.1
	 */
	public static Quadruple<DoubleMatrix2D, double[], double[], RoutingCycleType[]> convert_fte2xte(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, DoubleMatrix2D f_te, DoubleMatrix2D trafficMatrix)
	{
		int N = f_te.rows();
		int E = linkMap.size();

		DoubleMatrix2D x_te = DoubleFactory2D.sparse.make(N, E);
		double[] r_t = new double[N];
		double[] y_e = new double[E];
		RoutingCycleType[] routingCycleType = new RoutingCycleType[N];
		if (N == 0 || E == 0) return Quadruple.of(x_te, r_t, y_e, routingCycleType);
		
		Set<Long> linkIds = linkMap.keySet();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		
		for (int t = 0; t < N; t++)
		{
			Pair<DoubleMatrix2D, RoutingCycleType> out = computeM_fte(t, nodeIds, linkMap, f_te);
			DoubleMatrix2D M = out.getFirst();
			RoutingCycleType routingCycleType_thisEgressNode = out.getSecond();
			for (long linkId : linkIds)
			{
				long originNodeId = linkMap.get(linkId).getFirst();
				
				int e = linkId2LinearIndex.get(linkId);
				int a_e = nodeId2LinearIndex.get(originNodeId);
				
				for (long nodeId : nodeIds)
				{
					int n = nodeId2LinearIndex.get(nodeId);
					x_te.setQuick(t, e, trafficMatrix.getQuick(n, t) * M.getQuick(n, a_e) * f_te.getQuick(t, e));
					y_e[e] += x_te.getQuick(t, e);
				}
			}

			for (long nodeId : nodeIds)
			{
				int n = nodeId2LinearIndex.get(nodeId);
				r_t[t] += trafficMatrix.getQuick(n, t) * M.getQuick(n, t);
			}
			
			routingCycleType[t] = routingCycleType_thisEgressNode;
		}

		return Quadruple.of(x_te, r_t, y_e, routingCycleType);
	}
	
	/**
	 * Given a destination-based routing in the form f_te (fractions of traffic in a node, that is forwarded through each of its output links), and
	 * an offered traffic to the network, it generates the resulting destination-based
	 * routing in the form x_te (amount of traffic targeted to node t, transmitted through link e).
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param f_te For each destination node <i>t</i> and each link <i>e</i>, {@code f_te[t][e]} sets the fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>t</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1 (unchecked)
	 * @param trafficMatrix Traffic matrix
	 * @return Destination-based routing in the form <i>x<sub>te</sub></i> (amount of traffic targeted to node t, transmitted through link e)
	 * @since 0.3.1
	 */
	public static Quadruple<Map<Pair<Long, Long>, Double>, Map<Long, Double>, Map<Long, Double>, List<RoutingCycleType>> convert_fte2xte(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Pair<Long, Long>, Double> f_te, DoubleMatrix2D trafficMatrix)
	{
		int N = nodeIds.size();
		int E = linkMap.size();
		
		Map<Pair<Long, Long>, Double> x_te = new LinkedHashMap<Pair<Long, Long>, Double>();
		Map<Long, Double> r_t = DoubleUtils.constantMap(nodeIds, 0);
		Map<Long, Double> y_e = DoubleUtils.constantMap(linkMap.keySet(), 0);
		List<RoutingCycleType> routingCycleType = new LinkedList<RoutingCycleType>();
		if (N == 0 || E == 0) return Quadruple.of(x_te, r_t, y_e, routingCycleType);
		
		Set<Long> linkIds = linkMap.keySet();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		y_e = DoubleUtils.constantMap(linkIds, 0);
		
		for(long egressNodeId : nodeIds)
		{
			int t = nodeId2LinearIndex.get(egressNodeId);
				
			Pair<DoubleMatrix2D, RoutingCycleType> out = computeM_fte(t, nodeIds, linkMap, f_te);
			DoubleMatrix2D M = out.getFirst();
			RoutingCycleType routingCycleType_thisEgressNode = out.getSecond();
			
			for (long linkId : linkIds)
			{
				Pair<Long, Long> egressNodeLinkPair = Pair.of(egressNodeId, linkId);
				Double f_te_thisLink = f_te.get(egressNodeLinkPair);
				if (f_te_thisLink == null) continue;
				
				long originNodeId = linkMap.get(linkId).getFirst();
				
				int a_e = nodeId2LinearIndex.get(originNodeId);
				double x_te_thisEgressNode_thisLink = 0;
				
				for (long nodeId : nodeIds)
				{
					int n = nodeId2LinearIndex.get(nodeId);
					x_te_thisEgressNode_thisLink += trafficMatrix.getQuick(n, t) * M.getQuick(n, a_e) * f_te_thisLink;
				}
				
				if (x_te_thisEgressNode_thisLink < PRECISION_FACTOR) x_te_thisEgressNode_thisLink = 0;
				if (x_te_thisEgressNode_thisLink == 0) continue;
				
				x_te.put(egressNodeLinkPair, x_te_thisEgressNode_thisLink);
				y_e.put(linkId, y_e.get(linkId) + x_te_thisEgressNode_thisLink);
			}

			double r_t_thisEgressNode = 0;
			for (long nodeId : nodeIds)
			{
				int n = nodeId2LinearIndex.get(nodeId);
				r_t_thisEgressNode += trafficMatrix.getQuick(n, t) * M.getQuick(n, t);
			}
			
			r_t.put(egressNodeId, r_t_thisEgressNode);
			routingCycleType.add(routingCycleType_thisEgressNode);
		}
		
		return Quadruple.of(x_te, r_t, y_e, routingCycleType);
	}
	
	/**
	 * Given a demand-link routing in the form x_de (amount of traffic from demand 
	 * d, transmitted through link e), returns the equivalent forwarding rule 
	 * mapping (fractions of traffic entering a node from demand 'd', leaving that 
	 * node through link 'e').
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
     * @param demandIds Set of demand identifiers. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param x_de Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e)
	 * @return Forwarding rule mapping (fractions of traffic entering a node from demand 'd', leaving that node through link 'e')
	 * @since 0.3.0
	 */
	public static Map<Pair<Long, Long>, Double> convert_xde2fde(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Set<Long> demandIds, Map<Pair<Long, Long>, Double> x_de)
	{
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(nodeIds, linkMap);
		Map<Pair<Long, Long>, Double> f_de = new LinkedHashMap<Pair<Long, Long>, Double>();
		
		/* For each node 'n', where we are filling the routing table */
		for (long nodeId : nodeIds)
		{
			/* Compute the outgoing links from node n */
			Collection<Long> outLinks = graph.getOutEdges(nodeId);
			if (outLinks == null) outLinks = new LinkedHashSet<Long>();
			
			/* For each demand 'd', fill the f_de fractions */
			for (long demandId : demandIds)
			{
				double outTraffic = 0;
				for(long linkId : outLinks)
				{
					Pair<Long, Long> demandLinkPair = Pair.of(demandId, linkId);
					if (x_de.containsKey(demandLinkPair)) outTraffic += x_de.get(demandLinkPair);
				}
				
				if (outTraffic > 0) /* there is traffic leaving the node */
				{
					for (long linkId : outLinks)
					{
						Pair<Long, Long> demandLinkPair = Pair.of(demandId, linkId);
						if (x_de.containsKey(demandLinkPair)) f_de.put(demandLinkPair, x_de.get(demandLinkPair) / outTraffic);
					}
				}
			}
		}
		
		return f_de;
	}
	
	/**
	 * Given the amount of traffic for each demand <i>d</i> and traversing link <i>e</i>, it computes
	 * the equivalent path-based routing.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param x_de Matrix containing the amount of traffic from demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandMap.keySet()}, <i>d = 1</i> refers to the second one, and so on) which traverses link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on)
	 * @param d_p (Output parameter) Demand corresponding to each path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param x_p (Output parameter) Carried traffic per path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param pathList (Output parameter) List of paths, where each path is represented by its sequence of traversed links. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @return Number of new paths
	 * @since 0.3.1
	 */
	public static int convert_xde2xp(Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, DoubleMatrix2D x_de, List<Long> d_p, List<Double> x_p, List<List<Long>> pathList)
	{
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));

		Set<Long> linkIds = linkMap.keySet();
		Set<Long> demandIds = demandMap.keySet();
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds);

		int numPaths = 0;
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(linkMap);
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(null);
		for (long demandId : demandIds)
		{
			int d = demandId2LinearIndex.get(demandId);
			long ingressNodeId = demandMap.get(demandId).getFirst();
			long egressNodeId = demandMap.get(demandId).getSecond();

			double[] x_e = x_de.viewRow(d).toArray();
			for (int e = 0; e < x_e.length; e++) if (x_e[e] < 0) x_e[e] = 0;
			
			Collection<Long> incomingLinksToIngressNode = graph.getInEdges(ingressNodeId);
			Collection<Long> outgoingLinksFromIngressNode = graph.getOutEdges(ingressNodeId);
			if (incomingLinksToIngressNode == null) incomingLinksToIngressNode = new LinkedHashSet<Long>();
			if (outgoingLinksFromIngressNode == null) outgoingLinksFromIngressNode = new LinkedHashSet<Long>();

			double divAtIngressNode = 0;
			for(long linkId : outgoingLinksFromIngressNode)
			{
				int e = linkId2LinearIndex.get(linkId);
				divAtIngressNode += x_e[e];
			}
			
			for(long linkId : incomingLinksToIngressNode)
			{
				int e = linkId2LinearIndex.get(linkId);
				divAtIngressNode -= x_e[e];
			}
			
			while (divAtIngressNode > PRECISION_FACTOR)
			{
				List<Long> candidateLinks = new LinkedList<Long>();
				for (long linkId : linkIds)
				{
					int e = linkId2LinearIndex.get(linkId);
					if (x_e[e] > 0) candidateLinks.add(linkId);
				}

				if (candidateLinks.isEmpty()) break;
				
				Graph<Long, Long> auxGraph = JUNGUtils.filterGraph(graph, null, null, candidateLinks, null);
				List<Long> seqLinks = JUNGUtils.getShortestPath(auxGraph, nev, ingressNodeId, egressNodeId);
				if (seqLinks.isEmpty()) break;

				double trafficInPath = Double.MAX_VALUE;
				for(long linkId : seqLinks)
				{
					int e = linkId2LinearIndex.get(linkId);
					trafficInPath = Math.min(trafficInPath, x_e[e]);
				}
				
				trafficInPath = Math.min(trafficInPath, divAtIngressNode);
				divAtIngressNode -= trafficInPath;

				x_p.add(trafficInPath);
				d_p.add(demandId);
				pathList.add(seqLinks);
				for (long linkId : seqLinks)
				{
					int e = linkId2LinearIndex.get(linkId);
					x_e[e] -= trafficInPath;
				}
				
				numPaths++;

				if (divAtIngressNode <= PRECISION_FACTOR) break;
			}
		}

		return numPaths;
	}
	
	/**
	 * Given the amount of traffic for each demand <i>d</i> and traversing link <i>e</i>, it computes
	 * the equivalent path-based routing.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param x_de Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e)
	 * @param d_p (Output parameter) Demand corresponding to each path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param x_p (Output parameter) Carried traffic per path. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @param pathList (Output parameter) List of paths, where each path is represented by its sequence of traversed links. User should pass a initialized {@code List} object (i.e. {@code LinkedList} or {@code ArrayList})
	 * @return Number of new paths
	 * @since 0.3.0
	 */
	public static int convert_xde2xp(Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Pair<Long, Long>, Double> x_de, List<Long> d_p, List<Double> x_p, List<List<Long>> pathList)
	{
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));

		Set<Long> linkIds = linkMap.keySet();
		Set<Long> demandIds = demandMap.keySet();

		int numPaths = 0;
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(linkMap);
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(null);
		for (long demandId : demandIds)
		{
			long ingressNodeId = demandMap.get(demandId).getFirst();
			long egressNodeId = demandMap.get(demandId).getSecond();

			Map<Long, Double> x_e = new LinkedHashMap<Long, Double>();
			for(long linkId : linkIds)
			{
				Pair<Long, Long> demandLinkPair = Pair.of(demandId, linkId);
				if (x_de.containsKey(demandLinkPair))
					x_e.put(linkId, Math.max(0, x_de.get(demandLinkPair)));
			}
			
			Collection<Long> incomingLinksToIngressNode = graph.getInEdges(ingressNodeId);
			Collection<Long> outgoingLinksFromIngressNode = graph.getOutEdges(ingressNodeId);
			if (incomingLinksToIngressNode == null) incomingLinksToIngressNode = new LinkedHashSet<Long>();
			if (outgoingLinksFromIngressNode == null) outgoingLinksFromIngressNode = new LinkedHashSet<Long>();

			double divAtIngressNode = 0;
			for(long linkId : outgoingLinksFromIngressNode)
				if (x_e.containsKey(linkId))
					divAtIngressNode += x_e.get(linkId);
			
			for(long linkId : incomingLinksToIngressNode)
				if (x_e.containsKey(linkId))
					divAtIngressNode -= x_e.get(linkId);
			
			while (divAtIngressNode > PRECISION_FACTOR)
			{
				List<Long> candidateLinks = new LinkedList<Long>(x_e.keySet());
				if (candidateLinks.isEmpty()) break;
				
				Graph<Long, Long> auxGraph = JUNGUtils.filterGraph(graph, null, null, candidateLinks, null);
				List<Long> seqLinks = JUNGUtils.getShortestPath(auxGraph, nev, ingressNodeId, egressNodeId);
				if (seqLinks.isEmpty()) break;

				double trafficInPath = Double.MAX_VALUE;
				for(long linkId : seqLinks) trafficInPath = Math.min(trafficInPath, x_e.get(linkId));
				
				trafficInPath = Math.min(trafficInPath, divAtIngressNode);
				divAtIngressNode -= trafficInPath;

				x_p.add(trafficInPath);
				d_p.add(demandId);
				pathList.add(seqLinks);
				for (long linkId : seqLinks)
				{
					double newTrafficValue_thisLink = x_e.get(linkId) - trafficInPath;
					if (newTrafficValue_thisLink < PRECISION_FACTOR) x_e.remove(linkId);
					else x_e.put(linkId, newTrafficValue_thisLink);
				}
				
				numPaths++;

				if (divAtIngressNode <= PRECISION_FACTOR) break;
			}
		}

		return numPaths;
	}
	
	/**
	 * Returns the carried traffic per link.
	 * 
	 * @param linkIds Set of link identifiers. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param x_de Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e)
	 * @return Carried traffic per link
	 * @since 0.3.0
	 */
	public static Map<Long, Double> convert_xde2ye(Set<Long> linkIds, Map<Pair<Long, Long>, Double> x_de)
	{
		Map<Long, Double> y_e = DoubleUtils.zeros(linkIds);
		for(Entry<Pair<Long, Long>, Double> entry : x_de.entrySet())
		{
			Pair<Long, Long> forwardingRule = entry.getKey();
			double x_de_thisForwardingRule = entry.getValue();
			
			long linkId = forwardingRule.getSecond();
			y_e.put(linkId, y_e.get(linkId) + x_de_thisForwardingRule);
		}
		
		return y_e;
	}

	/**
	 * Returns the carried traffic per link.
	 * 
	 * @param x_de Matrix containing the amount of traffic from demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandMap.keySet()}, <i>d = 1</i> refers to the second one, and so on) which traverses link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on)
	 * @return Carried traffic per link
	 * @since 0.3.1
	 */
	public static double[] convert_xde2ye(DoubleMatrix2D x_de)
	{
		int D = x_de.rows();
		if (D == 0) return new double[0];

		int E = x_de.columns();
		double[] y_e = new double[E];
		for(int e = 0; e < E; e++)
			for(int d = 0; d < D; d++)
				y_e[e] += x_de.getQuick(d, e);
		
		return y_e;
	}

	/**
	 * Given a set of traffic routes and their carried traffic returns a destination-based
	 * routing in the form <i>f<sub>te</sub></i> (fractions of traffic in a node, that is forwarded
	 * through each of its output links).
	 *
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param d_p Demand corresponding to each path
	 * @param x_p Carried traffic per path
	 * @param pathList List of paths, where each path is represented by its sequence of traversed links
	 * @return Destination-based routing in the form <i>f<sub>te</sub></i> (fractions of traffic in a node, that is forwarded through each of its output links)
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D convert_xp2fte(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Long> d_p, Map<Long, Double> x_p, Map<Long, List<Long>> pathList)
	{
		DoubleMatrix2D x_te = convert_xp2xte(nodeIds, linkMap, demandMap, d_p, x_p, pathList);
		return convert_xte2fte(nodeIds, linkMap, x_te);
	}

	/**
	 * Given a set of traffic routes and their carried traffic returns a destination-based
	 * routing in the form x_te (amount of traffic targeted to node t, transmitted through link e).
	 *
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param d_p Demand corresponding to each path
	 * @param x_p Carried traffic per path
	 * @param pathList List of paths, where each path is represented by its sequence of traversed links
	 * @return Destination-based routing in the form x_te (amount of traffic targeted to node t, transmitted through link e)
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D convert_xp2xte(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Long> d_p, Map<Long, Double> x_p, Map<Long, List<Long>> pathList)
	{
		int N = nodeIds.size();
		int E = linkMap.size();
		Set<Long> linkIds = linkMap.keySet();
		Set<Long> demandIds = demandMap.keySet();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds);
		
		DoubleMatrix2D x_de = convert_xp2xde(linkMap, demandIds, d_p, x_p, pathList);
		DoubleMatrix2D x_te = DoubleFactory2D.sparse.make(N, E);
		for (long demandId : demandIds)
		{
			int d = demandId2LinearIndex.get(demandId);
			long egressNodeId = demandMap.get(demandId).getSecond();
			int t = nodeId2LinearIndex.get(egressNodeId);
			for (long linkId : linkIds)
			{
				int e = linkId2LinearIndex.get(linkId);
				x_te.setQuick(t, e, x_te.getQuick(t, e) + x_de.getQuick(d, e));
			}
		}

		return x_te;
	}
	
	/**
	 * Returns the carried traffic per link.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param x_p Carried traffic per path
	 * @param pathList List of paths, where each path is represented by its sequence of traversed links
	 * @return Carried traffic per link
	 * @since 0.3.0
	 */
	public static double[] convert_xp2ye(Map<Long, Pair<Long, Long>> linkMap, List<Double> x_p, List<List<Long>> pathList)
	{
		if (x_p.size() != pathList.size()) throw new Net2PlanException("'x_p' must have the same length as the number of paths");

		Set<Long> linkIds = linkMap.keySet();
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		int E = linkMap.size();
		double[] y_e = new double[E];

		Iterator<Double> x_p_it = x_p.iterator();
		Iterator<List<Long>> seqLinks_it = pathList.iterator();
		while (seqLinks_it.hasNext())
		{
			double x_p_thisPath = x_p_it.next();
			List<Long> seqLinks_thisPath = seqLinks_it.next();
			for (long linkId : seqLinks_thisPath)
			{
				int linkIndex = linkId2LinearIndex.get(linkId);
				y_e[linkIndex] += x_p_thisPath;
			}
		}

		return y_e;
	}
	
	/**
	 * Returns the carried traffic per link.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param x_p Carried traffic per path
	 * @param pathList List of paths, where each path is represented by its sequence of traversed links
	 * @return Carried traffic per link
	 * @since 0.2.0
	 */
	public static double[] convert_xp2ye(Map<Long, Pair<Long, Long>> linkMap, double[] x_p, List<List<Long>> pathList)
	{
		if (x_p.length != pathList.size()) throw new Net2PlanException("'x_p' must have the same length as the number of paths");

		Set<Long> linkIds = linkMap.keySet();
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		int E = linkMap.size();
		double[] y_e = new double[E];

		ListIterator<List<Long>> seqLinks_it = pathList.listIterator();
		while (seqLinks_it.hasNext())
		{
			int p = seqLinks_it.nextIndex();
			double x_p_thisPath = x_p[p];
			List<Long> seqLinks_thisPath = seqLinks_it.next();
			for (long linkId : seqLinks_thisPath)
			{
				int e = linkId2LinearIndex.get(linkId);
				y_e[e] += x_p_thisPath;
			}
		}

		return y_e;
	}

	/**
	 * Given a destination-based routing in the form of an array x_te (amount of traffic targeted to node t, transmitted through link e), it
	 * returns the associated destination-based routing in the form of fractions f_te (fraction of the traffic targeted to node t that arrives
	 * (or is generated in) node a(e) (the initial node of link e), that is forwarded through link e). If a node n does not forward any traffic to a destination t,
	 * it is not possible to determine the f_te fractions for the output links of the node. Then, the function arbitrarily assumes that the 100% of the traffic
	 * to node t, would be forwarded through the shortest path (in number of hops) from n to t. Finally note that for every destination t, f_te = 0
	 * for all the links e outgoing of node t.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param x_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} represents the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>
	 * @return The f_te array created
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D convert_xte2fte(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, DoubleMatrix2D x_te)
	{
		int N = x_te.rows();
		int E = linkMap.size();

		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkMap.keySet());
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(nodeIds, linkMap);
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(null);
		DoubleMatrix2D f_te = DoubleFactory2D.sparse.make(N, E);
		
		/* For each node n, where we are filling the routing table */
		for (long nodeId : nodeIds)
		{
			/* Compute the outgoing links from node n */
			Collection<Long> outLinks = graph.getOutEdges(nodeId);
			if (outLinks == null) outLinks = new LinkedHashSet<Long>();
			
			/* For each destination t, fill the f_te fractions */
			for (long egressNodeId : nodeIds)
			{
				if (nodeId == egressNodeId) continue; /* f_te = 0 for all the outgoing links of node t */
				
				int t = nodeId2LinearIndex.get(egressNodeId);
				
				double outTraffic = 0;
				for(long linkId : outLinks)
				{
					int e = linkId2LinearIndex.get(linkId);
					outTraffic += x_te.getQuick(t, e);
				}
				
				if (outTraffic != 0) /* there is traffic leaving the node */
				{
					for (long linkId : outLinks)
					{
						int e = linkId2LinearIndex.get(linkId);
						f_te.setQuick(t, e, x_te.getQuick(t, e) / outTraffic);
					}
				}
				else /* no traffic leaving the node: all the routing in the shortest path from n to t */
				{
					List<Long> seqLinks = JUNGUtils.getShortestPath(graph, nev, nodeId, egressNodeId);
					if (seqLinks.isEmpty()) continue;
					
					long firstLinkId = seqLinks.get(0);
					int e = linkId2LinearIndex.get(firstLinkId);
					f_te.setQuick(t, e, 1);
				}
			}
		}
		return f_te;
	}

	/**
	 * Given a path-based routing, returns the amount of traffic for each demand d
	 * traversing each link e.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
     * @param demandIds Set of demand identifiers. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param d_p Demand corresponding to each path
	 * @param x_p Carried traffic per path
	 * @param pathMap Map of paths, where the key is the unique path identifier and the value is the sequence of traversed links
	 * @return Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e)
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D convert_xp2xde(Map<Long, Pair<Long, Long>> linkMap, Set<Long> demandIds, Map<Long, Long> d_p, Map<Long, Double> x_p, Map<Long, List<Long>> pathMap)
	{
		int E = linkMap.size();
		int D = demandIds.size();
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkMap.keySet());
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds);
		
		DoubleMatrix2D x_de = DoubleFactory2D.sparse.make(D, E, 0);
		
		Set<Long> routeIds = x_p.keySet();
		for(long routeId : routeIds)
		{
			long demandId = d_p.get(routeId);
			int d = demandId2LinearIndex.get(demandId);
			
			double x_p_thisRoute = x_p.get(routeId);
			List<Long> seqLinks_thisRoute = pathMap.get(routeId);
			for(long linkId : seqLinks_thisRoute)
			{
				int e = linkId2LinearIndex.get(linkId);
				x_de.setQuick(d, e, x_de.getQuick(d, e) + x_p_thisRoute);
			}
		}

		return x_de;
	}

	/**
	 * Given a path-based routing, returns the amount of traffic for each demand d
	 * traversing each link e.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param d_p Demand corresponding to each path
	 * @param x_p Carried traffic per path
	 * @param pathMap Map of paths, where the key is the unique path identifier and the value is the sequence of traversed links
	 * @return Demand-link routing in the form x_de (amount of traffic from demand d, transmitted through link e)
	 * @since 0.3.0
	 */
	public static Map<Pair<Long, Long>, Double> convert_xp2xde(Map<Long, Pair<Long, Long>> linkMap, Map<Long, Long> d_p, Map<Long, Double> x_p, Map<Long, List<Long>> pathMap)
	{
		Map<Pair<Long, Long>, Double> x_de = new LinkedHashMap<Pair<Long, Long>, Double>();
		
		Set<Long> routeIds = x_p.keySet();
		for(long routeId : routeIds)
		{
			long demandId = d_p.get(routeId);
			
			double x_p_thisRoute = x_p.get(routeId);
			List<Long> seqLinks_thisRoute = pathMap.get(routeId);
			for(long linkId : seqLinks_thisRoute)
			{
				Pair<Long, Long> demandLinkPair = Pair.of(demandId, linkId);
				double previous_x_de = x_de.containsKey(demandLinkPair) ? x_de.get(demandLinkPair) : 0;
				x_de.put(demandLinkPair, previous_x_de + x_p_thisRoute);
			}
		}

		return x_de;
	}
	
	/**
	 * Returns the carried traffic per link.
	 * 
	 * @param x_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} represents the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>
	 * @return Carried traffic per link
	 * @since 0.3.1
	 */
	public static double[] convert_xte2ye(DoubleMatrix2D x_te)
	{
		return convert_xde2ye(x_te);
	}

	/**
	 * Returns all the loopless shortest paths between two nodes. All these paths
	 * have the same total cost.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param originNodeId Origin node
	 * @param destinationNodeId Destination node
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return All loopless shortest paths
	 * @since 0.3.0
	 */
	public static List<List<Long>> getAllLooplessShortestPaths(Map<Long, Pair<Long, Long>> linkMap, long originNodeId, long destinationNodeId, Map<Long, Double> linkCostMap)
	{
		Graph<Long, Long> g = JUNGUtils.getGraphFromLinkMap(linkMap);
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkCostMap);
		g = JUNGUtils.filterGraph(g, nev);

		YenAlgorithm<Long, Long> paths = new YenAlgorithm<Long, Long>(g, nev)
		{
			@Override
			public boolean compareCandidateToShortestPath(GraphPath<Long> candidate, GraphPath<Long> shortestPath)
			{
				return Math.abs(candidate.getPathWeight() - shortestPath.getPathWeight()) < 1E-10;
			}
		};

		List<List<Long>> pathList = paths.getPaths(originNodeId, destinationNodeId, Integer.MAX_VALUE);
		return pathList;
	}

	/**
	 * Returns the shortest path that fulfills a given minimum capacity requirement
	 * along its traversed links. In case no path can be found, an empty list will
	 * be returned.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param originNodeId Origin node
	 * @param destinationNodeId Destination node
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @param linkSpareCapacityMap Current available capacity per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @param capacityGoal Minimum capacity required
	 * @return Shortest path fulfilling a minimum capacity requirement
	 * @since 0.3.0
	 */
	public static List<Long> getCapacitatedShortestPath(Map<Long, Pair<Long, Long>> linkMap, long originNodeId, long destinationNodeId, final Map<Long, Double> linkCostMap, final Map<Long, Double> linkSpareCapacityMap, final double capacityGoal)
	{
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(linkMap);
		if (!graph.containsVertex(originNodeId)) return new LinkedList<Long>();
		if (!graph.containsVertex(destinationNodeId)) return new LinkedList<Long>();

		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkCostMap);
		Transformer<Long, Double> capacityTransformer = JUNGUtils.getEdgeWeightTransformer(linkSpareCapacityMap);
		return JUNGUtils.getCapacitatedShortestPath(graph, nev, originNodeId, destinationNodeId, capacityTransformer, capacityGoal);
	}

	/**
	 * Returns the K-loopless shortest paths between two nodes. If only <i>n<K</i>
	 * could be found, only these ones will be returned.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param originNodeId Origin node
	 * @param destinationNodeId Destination node
	 * @param K Number of different paths
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return K-shortest paths
	 * @since 0.3.0
	 */
	public static List<List<Long>> getKLooplessShortestPaths(Map<Long, Pair<Long, Long>> linkMap, long originNodeId, long destinationNodeId, int K, Map<Long, Double> linkCostMap)
	{
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(linkMap);
		final Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkCostMap);
		return JUNGUtils.getKLooplessShortestPaths(graph, nev, originNodeId, destinationNodeId, K);
	}

	/**
	 * Returns the shortest pair of link-disjoint paths, where each item represents a path.
	 * The number of returned items will be equal to the number of paths found: when empty, 
	 * no path was found; when {@code size()} = 1, only one path was found; and when 
	 * {@code size()} = 2, the link-disjoint paths were found. Internally it uses 
	 * the Suurballe-Tarjan algorithm.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param originNodeId Origin node
	 * @param destinationNodeId Destination node
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return Shortest pair of link-disjoint paths
	 * @since 0.3.0
	 */
	public static List<List<Long>> getTwoLinkDisjointPaths(Map<Long, Pair<Long, Long>> linkMap, long originNodeId, long destinationNodeId, Map<Long, Double> linkCostMap)
	{
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(linkMap);
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkCostMap);
		List<List<Long>> linkDisjointSPs = JUNGUtils.getTwoLinkDisjointPaths(graph, nev, originNodeId, destinationNodeId);
		return linkDisjointSPs;
	}

	/**
	 * Returns the shortest pair of node-disjoint paths, where each item represents a path.
	 * The number of returned items will be equal to the number of paths found: when empty, 
	 * no path was found; when {@code size()} = 1, only one path was found; and when 
	 * {@code size()} = 2, the node-disjoint paths were found. Internally it uses 
	 * the Suurballe-Tarjan algorithm.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param originNodeId Origin node
	 * @param destinationNodeId Destination node
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return Shortest pair of node-disjoint paths
	 * @since 0.3.0
	 */
	public static List<List<Long>> getTwoNodeDisjointPaths(Map<Long, Pair<Long, Long>> linkMap, long originNodeId, long destinationNodeId, Map<Long, Double> linkCostMap)
	{
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(linkMap);
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkCostMap);
		List<List<Long>> nodeDisjointSPs = JUNGUtils.getTwoNodeDisjointPaths(graph, nev, originNodeId, destinationNodeId);
		return nodeDisjointSPs;
	}

	/**
	 * <p>Given a map of links, it computes the adjacency matrix. This is a matrix 
	 * with as many rows and columns as nodes. Position (<i>i</i>,<i>j</i>) accounts 
	 * for the number of links/demands/paths from node <i>i</i> (<i>i = 0</i> refers 
	 * to the first node in {@code nodeIds}, <i>i = 1</i> refers to the second one, 
	 * and so on) to node <i>j</i> (<i>j = 0</i> refers to the first node in 
	 * {@code nodeIds}, <i>j = 1</i> refers to the second one, and so on).</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could 
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to 
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getAdjacencyMatrix(nodeIds, linkMap).toArray();}</p>
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @return Adjacency matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom/'>Java Optimization Modeler (JOM) website</a>
	 */
	public static DoubleMatrix2D getAdjacencyMatrix(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap)
	{
		int N = nodeIds.size();
		DoubleMatrix2D A_nn = DoubleFactory2D.sparse.make(N, N);
		
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		for (Pair<Long, Long> nodePair_thisLink : linkMap.values())
		{
			long originNodeId = nodePair_thisLink.getFirst();
			long destinationNodeId = nodePair_thisLink.getSecond();
			int a_e = nodeId2LinearIndex.get(originNodeId);
			int b_e = nodeId2LinearIndex.get(destinationNodeId);
			A_nn.setQuick(a_e, b_e, A_nn.get(a_e, b_e) + 1);
		}

		return A_nn;
	}
	
	/**
	 * <p>Given a map of links representing a bidirectional topology, with the same number 
	 * of links on each direction for each node pair, it bundles into opposite link 
	 * pairs and computes the bidirectional matrix. This is a matrix with as many 
	 * rows as opposite link pairs (or bidirectional links) and columns as links. 
	 * For each bidirectional link <i>i</i>, position (<i>i</i>,<i>j</i>) is equal 
	 * to 1 for link <i>j</i> (<i>e = 0</i> refers to the first link in 
	 * {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on) 
	 * in one direction, -1 for the link <i>j</i> in the opposite direction, and 
	 * zero otherwise. By convention, the selection order of links for each bidirectional link 
	 * is given by the following scheme: first, for each node <i>i</i> (in ascending order), 
	 * it iterates over all node <i>j</i> (in ascending order); then, for each node 
	 * pair (<i>i</i>, <i>j</i>) it retrieves the identifiers of links from <i>i</i> to <i>j</i>,
	 * and viceversa. Finally, it adds a new row to the matrix for each bidirectional link 
	 * consisting of a link from <i>i</i> to <i>j</i> and another one in the 
	 * opposite direction (also, in ascending order).</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so 
	 * that could be directly used along with the <a href='#jom'>JOM</a> library 
	 * in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getBidirectionalMatrix(nodeIds, linkMap).toArray();}</p>
	 *
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @return Bidirectional matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom/'>Java Optimization Modeler (JOM) website</a>
	 */
	public static DoubleMatrix2D getBidirectionalMatrix(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap)
	{
		Set<Long> linkIds = linkMap.keySet();
		
		try { CollectionUtils.convertId2LinearIndexMap(nodeIds); } /* Do not delete! It is used to check that nodeIds is in ascending order */
		catch(Throwable e) { throw(e); }
		
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		int E = linkIds.size();
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(nodeIds, linkMap);
		
		DoubleMatrix2D M_ij = DoubleFactory2D.sparse.make(E / 2, E);
		int i = 0;
		for(long nodeId_1 : nodeIds)
		{
			for(long nodeId_2 : nodeIds)
			{
				if (nodeId_1 >= nodeId_2) continue;
				
				Collection<Long> downstreamLinks = graph.findEdgeSet(nodeId_1, nodeId_2);
				Collection<Long> upstreamLinks = graph.findEdgeSet(nodeId_2, nodeId_1);
				if (downstreamLinks.size() != upstreamLinks.size()) throw new Net2PlanException("Link map must be bidirectional (same number of links on each direction)");
				
				Iterator<Long> downstreamLinks_it = downstreamLinks.iterator();
				Iterator<Long> upstreamLinks_it = upstreamLinks.iterator();
				while (downstreamLinks_it.hasNext())
				{
					int e_1 = linkId2LinearIndex.get(downstreamLinks_it.next());
					int e_2 = linkId2LinearIndex.get(upstreamLinks_it.next());
					
					M_ij.setQuick(i, e_1, 1);
					M_ij.setQuick(i, e_2, -1);
					i++;
				}
			}
		}
		
		return M_ij;
	}
	
	/**
	 * <p>Given a map of links, it computes the node-link incidence matrix. This 
	 * is a matrix with as many rows as nodes, and as many columns as links. 
	 * Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers
	 * to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers
	 * to the first node in {@code nodeIds}, <i>n = 1</i> refers to the second
	 * one, and so on), -1 if it ends in node n, and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getIncidenceMatrix(nodeIds, linkMap).toArray();}</p>
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @return Incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public static DoubleMatrix2D getIncidenceMatrix(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap)
	{
		int N = nodeIds.size();
		int E = linkMap.size();
		
		DoubleMatrix2D A_ne = DoubleFactory2D.sparse.make(N, E);		
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Set<Long> anyIds = linkMap.keySet();
		Map<Long, Integer> anyId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(anyIds);
		for (long anyId : anyIds)
		{
			Pair<Long, Long> nodePair_thisLink = linkMap.get(anyId);
			long originNodeId = nodePair_thisLink.getFirst();
			long destinationNodeId = nodePair_thisLink.getSecond();
			
			int a_e = nodeId2LinearIndex.get(originNodeId);
			int b_e = nodeId2LinearIndex.get(destinationNodeId);
			int e = anyId2LinearIndex.get(anyId);
			A_ne.set(a_e, e, 1);
			A_ne.set(b_e, e, -1);
		}

		return A_ne;
	}

	/**
	 * <p>Given a map of links, it computes the node-link incoming incidence 
	 * matrix. This is a matrix with as many rows as nodes, and as many columns
	 * as links. Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i>
	 * (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()},
	 * <i>e = 1</i> refers to the second one, and so on) is terminated in node
	 * <i>n</i> (<i>n = 0</i> refers to the first node in {@code nodeIds},
	 * <i>n = 1</i> refers to the second one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getIncomingIncidenceMatrix(nodeIds, linkMap).toArray();}</p>
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @return Node-link incoming incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public static DoubleMatrix2D getIncomingIncidenceMatrix(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap)
	{
		int N = nodeIds.size();
		int E = linkMap.size();
		
		DoubleMatrix2D A_ne = DoubleFactory2D.sparse.make(N, E);
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Set<Long> anyIds = linkMap.keySet();
		Map<Long, Integer> anyId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(anyIds);
		for (long anyId : anyIds)
		{
			Pair<Long, Long> nodePair_thisLink = linkMap.get(anyId);
			long destinationNodeId = nodePair_thisLink.getSecond();
			
			int b_e = nodeId2LinearIndex.get(destinationNodeId);
			int e = anyId2LinearIndex.get(anyId);
			A_ne.set(b_e, e, 1);
		}

		return A_ne;
	}

	/**
	 * <p>Given a map of links, it computes the node-link outgoing incidence matrix. 
	 * This is a matrix with as many rows as nodes, and as many columns as links. 
	 * Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers 
	 * to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the 
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers 
	 * to the first node in {@code nodeIds}, <i>n = 1</i> refers to the second 
	 * one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getOutgoingIncidenceMatrix(nodeIds, linkMap).toArray();}</p>
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @return Node-link outgoing incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public static DoubleMatrix2D getOutgoingIncidenceMatrix(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap)
	{
		int N = nodeIds.size();
		int E = linkMap.size();
		
		DoubleMatrix2D A_ne = DoubleFactory2D.sparse.make(N, E);
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Set<Long> anyIds = linkMap.keySet();
		Map<Long, Integer> anyId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(anyIds);
		for (long anyId : anyIds)
		{
			Pair<Long, Long> nodePair_thisLink = linkMap.get(anyId);
			long originNodeId = nodePair_thisLink.getFirst();
			
			int a_e = nodeId2LinearIndex.get(originNodeId);
			int e = anyId2LinearIndex.get(anyId);
			A_ne.set(a_e, e, 1);
		}

		return A_ne;
	}
	
	/**
	 * Given a link map and a node, it returns the set of links that enter to that node.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param nodeId Node identifier
	 * @return Incoming links of the node
	 * @since 0.3.0
	 */
	public static Set<Long> getNodeIncomingLinks(Map<Long, Pair<Long, Long>> linkMap, long nodeId)
	{
		Set<Long> out = new TreeSet<Long>();
		for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
			if (entry.getValue().getSecond() == nodeId)
				out.add(entry.getKey());
		
		return out;
	}

	/**
	 * Given a link map and a node, it returns the set of links that leave that node.
	 * 
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param nodeId Node identifier
	 * @return Outgoing links of the node
	 * @since 0.3.0
	 */
	public static Set<Long> getNodeOutgoingLinks(Map<Long, Pair<Long, Long>> linkMap, long nodeId)
	{
		Set<Long> out = new TreeSet<Long>();
		for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
			if (entry.getValue().getFirst() == nodeId)
				out.add(entry.getKey());
		
		return out;
	}
	
	/**
	 * Obtains the sequence of links representing the (unidirectional) shortest path between two nodes.
	 *
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param originNodeId Origin node identifier
	 * @param destinationNodeId Destination node identifier
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return Sequence of links in the shortest path (empty, if destination not reachable from origin)
	 * @since 0.3.0
	 */
	public static List<Long> getShortestPath(Map<Long, Pair<Long, Long>> linkMap, long originNodeId, long destinationNodeId, Map<Long, Double> linkCostMap)
	{
		Graph<Long, Long> graph = JUNGUtils.getGraphFromLinkMap(linkMap);
		if (!graph.containsVertex(originNodeId)) return new LinkedList<Long>();
		if (!graph.containsVertex(destinationNodeId)) return new LinkedList<Long>();

		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkCostMap);
		return JUNGUtils.getShortestPath(graph, nev, originNodeId, destinationNodeId);
	}

	/**
	 * Checks whether the physical topology has the same number of links between each node pair in both directions (assuming multi-digraphs).
	 *
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @return {@code true} if the physical topology is bidirectional, and false otherwise
	 * @since 0.3.0
	 */
	public static boolean isBidirectional(Map<Long, Pair<Long, Long>> linkMap)
	{
		org.jgrapht.Graph<Long, Long> graph = JGraphTUtils.getGraphFromLinkMap(linkMap);
		return JGraphTUtils.isWeightedBidirectional(graph);
	}
	
	/**
	 * Check whether the physical topology is connected, that is, if it is possible 
	 * to connect every node to each other, but only in a subset of nodes (subgraph).
	 *
	 * @param nodeIds Set of node identifiers
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @return {@code true} if the subgraph is connected, and false otherwise
	 * @since 0.3.0
	 */
	public static boolean isConnected(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap)
	{
		org.jgrapht.Graph<Long, Long> graph = JGraphTUtils.getGraphFromLinkMap(linkMap);
		return JGraphTUtils.isConnected(graph, nodeIds);
	}

	/**
	 * Check whether the physical topology is simple, that is, if it has at most one unidirectional link from a node to each other.
	 *
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @return {@code true} if the physical topology is simple, and false otherwise
	 * @since 0.3.0
	 */
	public static boolean isSimple(Map<Long, Pair<Long, Long>> linkMap)
	{
		org.jgrapht.Graph<Long, Long> graph = JGraphTUtils.getGraphFromLinkMap(linkMap);
		return JGraphTUtils.isSimple(graph);
	}

	/**
	 * Checks whether the physical topology has the same number of links between each node pair in both directions (assuming multi-digraphs) and same weights per direction.
	 *
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return {@code true} if the physical topology is bidirectional and symmetric, and false otherwise
	 * @since 0.3.0
	 */
	public static boolean isWeightedBidirectional(Map<Long, Pair<Long, Long>> linkMap, Map<Long, Double> linkCostMap)
	{
		org.jgrapht.Graph<Long, Long> auxGraph = JGraphTUtils.getGraphFromLinkMap(linkMap);
		org.jgrapht.Graph<Long, Long> graph = JGraphTUtils.getAsWeightedGraph(auxGraph, linkCostMap);
		return JGraphTUtils.isWeightedBidirectional(graph);
	}
	
	/**
	 * Given a link map that may contain multiple links between some node pairs,
	 * returns a new link map where only appears, for each node pair, the link having 
	 * the lowest weight (links whose weight is equal to {@code Double.MAX_VALUE} 
	 * are excluded). Ties are broken arbitrarely.
	 *
	 * <p><b>Important</b>: Returned link map is not backed in the input one, so changes
	 * will not be reflected on it.
	 *
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively
	 * @param linkCostMap Cost per link, where the key is the link identifier and the value is the cost of traversing the link. No special iteration-order (i.e. ascending) is required
	 * @return Copy of the input link map with at most a link between each node pair
	 * @since 0.3.0
	 */
	public static Map<Long, Pair<Long, Long>> simplifyLinkMap(Map<Long, Pair<Long, Long>> linkMap, Map<Long, Double> linkCostMap)
	{
		Map<Pair<Long, Long>, Long> originDestinationNodePairLinkIdMapping = new LinkedHashMap<Pair<Long, Long>, Long>();
		for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
		{
			long linkId = entry.getKey();
			Pair<Long, Long> originDestinationNodePair = entry.getValue();

			double cost_thisLink = linkCostMap == null ? 1 : linkCostMap.get(linkId);
			if (cost_thisLink == Double.MAX_VALUE) continue;
			
			if (originDestinationNodePairLinkIdMapping.containsKey(originDestinationNodePair) && linkCostMap != null)
			{
				long bestLinkId = originDestinationNodePairLinkIdMapping.get(originDestinationNodePair);
				double bestCostId = linkCostMap.get(bestLinkId);
				if (cost_thisLink < bestCostId) originDestinationNodePairLinkIdMapping.put(originDestinationNodePair, linkId);
			}
			else
			{
				originDestinationNodePairLinkIdMapping.put(originDestinationNodePair, linkId);
			}
		}

		Set<Long> linksToMaintain = new TreeSet<Long>(originDestinationNodePairLinkIdMapping.values());
		return CollectionUtils.selectEntries(linkMap, linksToMaintain);
	}
	
	/**
	 * <p>Class to represent a path in a Graph. Note that a path is defined in terms of edges (rather than vertices) so that multiple edges between the same pair of vertices can be discriminated.</p>
	 *
	 * <p>It implements the {@code {@link java.lang.Comparable Comparable}} interface to impose order between different paths. First, try to order using the path weight, and if equals, using the number of hops.</p>
	 *
	 * @param <E> Edge type
	 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
	 * @since 0.2.0
	 */
	public static class GraphPath<E> implements Comparable<GraphPath>
	{
		private final List<E> path;
		private final double pathWeight;

		/**
		 * Default constructor.
		 *
		 * @param path Sequence of links
		 * @param pathWeight Path weight
		 * @since 0.2.0
		 */
		public GraphPath(List<E> path, double pathWeight)
		{
			this.path = path;
			this.pathWeight = pathWeight;
		}

		/**
		 * Returns the edges making up the path.
		 *
		 * @return An unmodifiable list with the sequence of edges followed by the path
		 * @since 0.2.0
		 */
		public List<E> getPath()
		{
			return Collections.unmodifiableList(path);
		}

		/**
		* Returns the path length measured in number of hops or edges. It is equivalent to {@code {@link #getPath getPath()}.size()}.
		*
		* @return The path length
		* @since 0.2.0
		*/
		public int getPathLength()
		{
			return getPath().size();
		}

		/**
		* Returns the weight assigned to the path.
		*
		* @return The weight assigned to the path
		* @since 0.2.0
		*/
		public double getPathWeight()
		{
			return pathWeight;
		}

		/**
		* Indicates whether some other object is "equal to" this one.
		 *
		 * @param o Reference object with which to compare
		 * @return {@code true} if this object is the same as the {@code o} argument; {@code false} otherwise
		 * @since 0.2.0
		 */
		@Override
		public boolean equals(Object o)
		{
			if (o == null) return false;
			if (o == this) return true;
			if (!(o instanceof GraphPath)) return false;

			GraphPath p = (GraphPath) o;
			return getPath().equals(p.getPath()) && Math.abs(getPathWeight() - p.getPathWeight()) < 1e-10;
		}

		/**
		* Returns a hash code value for the object. This method is supported for the benefit of hash tables such as those provided by {@code HashMap}.
		 *
		 * @return Hash code value for this object
		 * @since 0.2.0
		 */
		@Override
		public int hashCode()
		{
			int hash = 7;
			hash = 37 * hash + (this.path != null ? this.path.hashCode() : 0);
			hash = 37 * hash + (int) (Double.doubleToLongBits(this.pathWeight) ^ (Double.doubleToLongBits(this.pathWeight) >>> 32));
			return hash;
		}

		/**
		* Compares this object with the specified object for order.
		 *
		 * @param o The object to be compared
		 * @return A negative integer, zero, or a positive integer as this object is less than, equal to, or greater than the specified object
		 * @since 0.2.0
		 */
		@Override
		public int compareTo(GraphPath o)
		{
			if (pathWeight < o.pathWeight) return -1;
			if (pathWeight > o.pathWeight) return 1;

			return path.size() == o.path.size() ? 0 : (path.size() > o.path.size() ? -1 : 1);
		}
	}

	/**
	 * <p>Auxiliary class to work with the graph library <a href='GraphUtils.html#jgrapht'>JGraphT</a>.</p>
	 *
	 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
	 * @since 0.2.0
	 */
	public static class JGraphTUtils
	{
		/**
		 * Builds an auxiliary graph for the application of edge-disjoint 
		 * shortest-path pair algorithms to node-disjoint problems.
		 * 
		 * @param graph Graph representing the network
		 * @param originNode Origin node
		 * @param destinationNode Destination node
		 * @return Auxiliary graph
		 * @since 0.3.0
		 */
		public static org.jgrapht.Graph<Long, Long> buildAuxiliaryNodeDisjointGraph(final org.jgrapht.Graph<Long, Long> graph, long originNode, long destinationNode)
		{
			org.jgrapht.Graph<Long, Long> auxGraph = new DirectedWeightedMultigraph<Long, Long>(Long.class);
			Map<Long, Long> originalNodeId2AuxIdMapping = new LinkedHashMap<Long, Long>();

			for(long vertex : graph.vertexSet())
			{
				auxGraph.addVertex(vertex);
				
				if (vertex == originNode || vertex == destinationNode)
				{
					originalNodeId2AuxIdMapping.put(vertex, vertex);
				}
				else
				{
					long auxVertex = -(10000 + vertex);
					originalNodeId2AuxIdMapping.put(vertex, auxVertex);
					long auxEdge = auxVertex;

					auxGraph.addVertex(auxVertex);
					auxGraph.addEdge(vertex, auxVertex, auxEdge);
				}
			}

			for(long edge : graph.edgeSet())
			{
				if (edge < 0) throw new Net2PlanException("Bad - Edge indexes must be greater or equal than zero");
				
				long originNode_thisLink = originalNodeId2AuxIdMapping.get(graph.getEdgeSource(edge));
				long destinationNode_thisLink = graph.getEdgeTarget(edge);
				auxGraph.addEdge(originNode_thisLink, destinationNode_thisLink, edge);
			}

			return auxGraph;
		}

		/**
		 * <p>It generates a weighted view of the backing graph specified in the constructor. This graph allows modules to apply algorithms designed for weighted graphs to an unweighted graph by providing an explicit edge weight mapping.</p>
		 *
		 * <p>Query operations on this graph "read through" to the backing graph. Vertex addition/removal and edge addition/removal are supported.</p>
		 * 
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param graph The backing graph over which a weighted view is to be created
		 * @param edgeWeightMap A mapping of edges to weights (null means all to one)
		 * @return Returns a weighted view of the backing graph specified in the constructor
		 * @since 0.3.0
		 */
		public static <V, E> org.jgrapht.WeightedGraph<V, E> getAsWeightedGraph(org.jgrapht.Graph<V, E> graph, Map<E, Double> edgeWeightMap)
		{
			if (edgeWeightMap == null)
			{
				edgeWeightMap = new LinkedHashMap<E, Double>();
				for(E edge : graph.edgeSet()) edgeWeightMap.put(edge, 1.0);
			}
			
			return new AsWeightedGraph<V, E>(graph, edgeWeightMap);
		}

		/**
		 * <p>Obtains a {@code JGraphT} graph from a given link map.</p>
		 * 
		 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. Origin/destination nodes will be added as needed
		 * @return {@code JGraphT} graph
		 * @since 0.3.0
		 */
		public static org.jgrapht.Graph<Long, Long> getGraphFromLinkMap(Map<Long, Pair<Long, Long>> linkMap)
		{
			org.jgrapht.Graph<Long, Long> graph = new DirectedWeightedMultigraph<Long, Long>(Long.class);
			
			if (linkMap != null)
			{
				for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
				{
					long linkId = entry.getKey();
					long originNodeId = entry.getValue().getFirst();
					long destinationNodeId = entry.getValue().getSecond();

					if (!graph.containsVertex(originNodeId)) graph.addVertex(originNodeId);
					if (!graph.containsVertex(destinationNodeId)) graph.addVertex(destinationNodeId);

					graph.addEdge(originNodeId, destinationNodeId, linkId);
				}
			}

			return graph;
		}

		/**
		 * <p>Obtains a {@code JGraphT} graph from a given link map.</p>
		 * 
		 * @param nodeIds Set of node identifiers
		 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. Only links whose end-nodes are in {@code nodeIds} will be added to the graph
		 * @return {@code JGraphT} graph
		 * @since 0.3.0
		 */
		public static org.jgrapht.Graph<Long, Long> getGraphFromLinkMap(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap)
		{
			org.jgrapht.Graph<Long, Long> graph = new DirectedWeightedMultigraph<Long, Long>(Long.class);

			if (nodeIds != null)
				for(long nodeId : nodeIds) graph.addVertex(nodeId);
			
			if (linkMap != null)
			{
				for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
				{
					long linkId = entry.getKey();
					long originNodeId = entry.getValue().getFirst();
					long destinationNodeId = entry.getValue().getSecond();

					if (!graph.containsVertex(originNodeId)) continue;
					if (!graph.containsVertex(destinationNodeId)) continue;

					graph.addEdge(originNodeId, destinationNodeId, linkId);
				}
			}

			return graph;
		}
		
		/**
		 * Check whether the topology has the same number of links between each node pair in both directions (assuming multi-digraphs).
		 *
		 * @param graph The graph to analyze
		 * @return {@code true} if the graph is bidirectional, and false otherwise
		 * @since 0.2.0
		 */
		public static boolean isBidirectional(org.jgrapht.Graph graph)
		{
			Object[] vertices = graph.vertexSet().toArray();
			for(int v1 = 0; v1 < vertices.length; v1++)
				for(int v2 = v1 + 1; v2 < vertices.length; v2++)
					if (graph.getAllEdges(vertices[v1], vertices[v2]).size() != graph.getAllEdges(vertices[v2], vertices[v1]).size())
						return false;

			return true;
		}

		/**
		 * Check whether the graph is connected, that is, if it is possible to connect every node to each other.
		 *
		 * @param graph The graph to analyze
		 * @return {@code true} if the graph is connected, and false otherwise
		 * @since 0.2.0
		 */
		public static boolean isConnected(org.jgrapht.Graph graph)
		{
			if (graph instanceof DirectedGraph)
			{
				StrongConnectivityInspector ci = new StrongConnectivityInspector((DirectedGraph) graph);
				return ci.isStronglyConnected();
			}
			else if (graph instanceof UndirectedGraph)
			{
				ConnectivityInspector ci = new ConnectivityInspector((UndirectedGraph) graph);
				return ci.isGraphConnected();
			}

			throw new RuntimeException("Bad");
		}

		/**
		 * Check whether the graph is connected, that is, if it is possible to connect every node to each other, but only in a subset of vertices (subgraph).
		 *
		 * @param graph The graph to analyze
		 * @param vertices Subset of vertices
		 * @return {@code true} if the subgraph is connected, and false otherwise
		 * @since 0.2.0
		 */
		public static boolean isConnected(org.jgrapht.Graph graph, Set vertices)
		{
			Subgraph subgraph;
			if (graph instanceof DirectedGraph) subgraph = new DirectedSubgraph((DirectedGraph) graph, vertices, null);
			else if (graph instanceof UndirectedGraph) subgraph = new UndirectedSubgraph((UndirectedGraph) graph, vertices, null);
			else throw new RuntimeException("Bad");

			return isConnected(subgraph);
		}

		/**
		 * Check whether the graph is simple, that is, if it has at most one link between each node pair (one per direction under directed graphs, one under undirected graphs).
		 *
		 * @param graph The graph to analyze
		 * @return {@code true} if the graph is simple, and false otherwise
		 * @since 0.2.0
		 */
		public static boolean isSimple(org.jgrapht.Graph graph)
		{
			Object[] vertices = graph.vertexSet().toArray();
			for(int v1 = 0; v1 < vertices.length; v1++)
			{
				for(int v2 = v1 + 1; v2 < vertices.length; v2++)
				{
					if (graph.getAllEdges(vertices[v1], vertices[v2]).size() > 1) return false;
					if (graph.getAllEdges(vertices[v2], vertices[v1]).size() > 1) return false;
				}
			}

			return true;
		}

		/**
		 * Checks whether the graph has the same number of links between each node pair in both directions (assuming multi-digraphs) and same individual weights per direction.
		 *
		 * @param graph The graph to analyze
		 * @return {@code true} if the graph is weighted-bidirectional, and false otherwise. By convention returns {@code false} if network is empty 
		 * @since 0.2.0
		 */
		public static boolean isWeightedBidirectional(org.jgrapht.Graph graph)
		{
			Object[] vertices = graph.vertexSet().toArray();
			if (vertices.length == 0)
			{
				return false;
			}

			for (int vertexId_1 = 0; vertexId_1 < vertices.length; vertexId_1++)
			{
				for (int vertexId_2 = vertexId_1 + 1; vertexId_2 < vertices.length; vertexId_2++)
				{
					Set links_12 = graph.getAllEdges(vertices[vertexId_1], vertices[vertexId_2]);
					Set links_21 = graph.getAllEdges(vertices[vertexId_2], vertices[vertexId_1]);

					if (links_12.size() != links_21.size())
					{
						return false;
					}

					Iterator it_12 = links_12.iterator();
					while (it_12.hasNext())
					{
						Object aux_12 = it_12.next();

						Iterator it_21 = links_21.iterator();
						while (it_21.hasNext())
						{
							Object aux_21 = it_21.next();

							if (Math.abs(graph.getEdgeWeight(aux_12) - graph.getEdgeWeight(aux_21)) < 1E-10)
							{
								it_12.remove();
								it_21.remove();
								break;
							}
						}
					}

					if (!links_12.isEmpty() || !links_12.isEmpty())
					{
						return false;
					}
				}
			}

			return true;
		}
	}
	
	/**
	 * <p>Auxiliary class to work with the graph library <a href='GraphUtils.html#jung'>JUNG</a>.</p>
	 *
	 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
	 * @since 0.2.0
	 */
	public static class JUNGUtils
	{
		/**
		 * Builds an auxiliary graph for the application of edge-disjoint 
		 * shortest-path pair algorithms to node-disjoint problems.
		 * 
		 * @param graph Graph representing the network
		 * @param originNode Origin node
		 * @param destinationNode Destination node
		 * @return Auxiliary graph
		 * @since 0.3.0
		 */
		public static Graph<Long, Long> buildAuxiliaryNodeDisjointGraph(final Graph<Long, Long> graph, long originNode, long destinationNode)
		{
			Graph<Long, Long> auxGraph = new DirectedSparseMultigraph<Long, Long>();
			Map<Long, Long> originalNodeId2AuxIdMapping = new LinkedHashMap<Long, Long>();

			for(long vertex : graph.getVertices())
			{
				auxGraph.addVertex(vertex);
				
				if (vertex == originNode || vertex == destinationNode)
				{
					originalNodeId2AuxIdMapping.put(vertex, vertex);
				}
				else
				{
					long auxVertex = -(10000 + vertex);
					originalNodeId2AuxIdMapping.put(vertex, auxVertex);
					long auxEdge = auxVertex;

					auxGraph.addVertex(auxVertex);
					auxGraph.addEdge(auxEdge, vertex, auxVertex);
				}
			}

			for(long edge : graph.getEdges())
			{
				if (edge < 0) throw new Net2PlanException("Bad - Edge indexes must be greater or equal than zero");
				
				switch(graph.getEdgeType(edge))
				{
					case DIRECTED:
						long originNode_thisLink = originalNodeId2AuxIdMapping.get(graph.getSource(edge));
						long destinationNode_thisLink = graph.getDest(edge);
						auxGraph.addEdge(edge, originNode_thisLink, destinationNode_thisLink);
						break;

					default:
						throw new Net2PlanException("This method is only for directed graphs");
				}
			}

			return auxGraph;
		}

		/**
		 * Builds an auxiliary graph for the application of edge-disjoint 
		 * shortest-path pair algorithms to node-disjoint problems.
		 * 
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @param originNode Origin node
		 * @param destinationNode Destination node
		 * @return Auxiliary graph, and its corresponding object returning edge weights
		 * @since 0.3.0
		 */
		public static Pair<Graph<Long, Long>, Transformer<Long, Double>> buildAuxiliaryNodeDisjointGraph(final Graph<Long, Long> graph, final Transformer<Long, Double> nev, long originNode, long destinationNode)
		{
			final Graph<Long, Long> auxGraph = buildAuxiliaryNodeDisjointGraph(graph, originNode, destinationNode);
			
			Transformer<Long, Double> auxNev = new Transformer<Long, Double>()
			{
				@Override
				public Double transform(Long edge)
				{
					if (graph.containsEdge(edge))
					{
						if (edge < 0) throw new RuntimeException("Bad");
						return nev.transform(edge);
					}
					else if (auxGraph.containsEdge(edge))
					{
						return 1.0;
					}
					
					throw new RuntimeException("Bad");
				}
			};
			
			return Pair.of(auxGraph, auxNev);
		}
		
		/**
		 * Creates an image of the network topology. The returned image will be 
		 * trimmed to remove white borders.
		 * 
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param graph Graph representing the network
		 * @param vertexPositionTransformer Object responsible for returning position for vertices
		 * @param width Expected maximum width (it may be lower due to white borders removal)
		 * @param height Expected maximum height (it may be lower due to white borders removal)
		 * @param fontSize Font size for vertex label (in points)
		 * @param nodeSize Node size
		 * @param linkThickness Link thickness
		 * @return Image
		 * @since 0.3.1
		 */
		public static <V, E> BufferedImage createImage(Graph<V, E> graph, final Transformer<V, Point2D> vertexPositionTransformer, int width, int height, int fontSize, final float nodeSize, final float linkThickness)
		{
			Transformer<V, Point2D> actualVertexPositionTransformer = new Transformer<V, Point2D>()
			{
				@Override
				public Point2D transform(V input)
				{
					Point2D p = new Point2D.Double();
					p.setLocation(vertexPositionTransformer.transform(input));
					p.setLocation(p.getX(), -p.getY());
					
					return p;
				}
			};
			
			Layout<V, E> layout = new StaticLayout<V, E>(graph, actualVertexPositionTransformer, new Dimension(width, height));
			VisualizationImageServer<V, E> vis = new VisualizationImageServer<V, E>(layout, layout.getSize());
			vis.setBackground(Color.WHITE);
			vis.getRenderContext().setEdgeArrowPredicate(new NoArrowPredicate<V, E>());
			vis.getRenderContext().setEdgeShapeTransformer(new EdgeShape.Line<V, E>());
			vis.getRenderContext().setEdgeStrokeTransformer(new ConstantTransformer(new BasicStroke(linkThickness)));
			vis.getRenderContext().setVertexDrawPaintTransformer(new ConstantTransformer(Color.BLACK));
			vis.getRenderContext().setVertexFillPaintTransformer(new ConstantTransformer(Color.BLACK));
			vis.getRenderContext().setVertexFontTransformer(new ConstantTransformer(new Font("Helvetica", Font.BOLD, fontSize)));
			vis.getRenderContext().setVertexLabelRenderer(new MyDefaultVertexLabelRenderer<V>(Color.WHITE));
			vis.getRenderContext().setVertexLabelTransformer(new ToStringLabeller<V>());
			vis.getRenderContext().setVertexShapeTransformer(new Transformer<V, Shape>()
			{
				@Override
				public Shape transform(V vertex)
				{
					return new Ellipse2D.Double(-nodeSize / 2, -nodeSize / 2, nodeSize, nodeSize);
				}
			});
			
			vis.getRenderer().getVertexLabelRenderer().setPosition(Renderer.VertexLabel.Position.CNTR);

			double aux_xmax = Double.NEGATIVE_INFINITY;
			double aux_xmin = Double.POSITIVE_INFINITY;
			double aux_ymax = Double.NEGATIVE_INFINITY;
			double aux_ymin = Double.POSITIVE_INFINITY;

			for(V vertex : graph.getVertices())
			{
				Point2D aux = layout.transform(vertex);
				if (aux_xmax < aux.getX()) aux_xmax = aux.getX();
				if (aux_xmin > aux.getX()) aux_xmin = aux.getX();
				if (aux_ymax < aux.getY()) aux_ymax = aux.getY();
				if (aux_ymin > aux.getY()) aux_ymin = aux.getY();
			}

			final double PRECISION_FACTOR = 0.00001;
			Rectangle viewInLayoutUnits = vis.getRenderContext().getMultiLayerTransformer().inverseTransform(vis.getBounds()).getBounds();
			float ratio_h = Math.abs(aux_xmax - aux_xmin) < PRECISION_FACTOR ? 1 : (float) (viewInLayoutUnits.getWidth() / (aux_xmax - aux_xmin));
			float ratio_v = Math.abs(aux_ymax - aux_ymin) < PRECISION_FACTOR ? 1 : (float) (viewInLayoutUnits.getHeight() / (aux_ymax - aux_ymin));
			float ratio = (float) (0.8 * Math.min(ratio_h, ratio_v));
			LayoutScalingControl scalingControl = new LayoutScalingControl();
			scalingControl.scale(vis, ratio, vis.getCenter());

			Point2D q = new Point2D.Double((aux_xmin + aux_xmax) / 2, (aux_ymin + aux_ymax) / 2);
			Point2D lvc = vis.getCenter();
			double dx = (lvc.getX() - q.getX());
			double dy = (lvc.getY() - q.getY());
			vis.getRenderContext().getMultiLayerTransformer().getTransformer(Layer.LAYOUT).translate(dx, dy);

			BufferedImage bi = (BufferedImage) vis.getImage(new Point2D.Double(layout.getSize().getWidth() / 2, layout.getSize().getHeight() / 2), new Dimension(layout.getSize()));
			return ImageUtils.trim(bi);
		}
	
		/**
		 * <p>Filters a graph maintaining/removing some vertices/edges. In the same invocation, 
		 * user must choose between 'accept' or 'block' some elements (vertices and/or edges).</p>
		 *
		 * <p><b>Important</b>: Returned graph is not backed in the input one, so changes
		 * will not be reflected on it.
		 * 
		 * @param <V> Class type for vertices
		 * @param <E> Class type for edges
		 * @param graph Graph representing the network
		 * @param acceptedVertices Collection of accepted vertices (null means empty)
		 * @param blockedVertices Collection of blocked vertices (null means empty)
		 * @param acceptedEdges Collection of accepted edges (null means empty)
		 * @param blockedEdges Collection of blocked edges (null means empty)
		 * @return Filtered graph
		 * @since 0.3.0
		 */
		public static <V, E> Graph<V, E> filterGraph(Graph<V, E> graph, final Collection<V> acceptedVertices, final Collection<V> blockedVertices, final Collection<E> acceptedEdges, final Collection<E> blockedEdges)
		{
			if (acceptedVertices != null && blockedVertices != null) throw new Net2PlanException("Over-specified filter: only accepted or blocked vertices can be specified at the same time");
			if (acceptedEdges != null && blockedEdges != null) throw new Net2PlanException("Over-specified filter: only accepted or blocked edges can be specified at the same time");
			
			Graph<V, E> filteredGraph;
			try { filteredGraph = graph.getClass().newInstance(); }
			catch (InstantiationException | IllegalAccessException e) { throw new RuntimeException("Unable to create a copy of existing graph: ", e); }

			if (acceptedVertices != null)			
			{
				for (V vertex : graph.getVertices())
					if (acceptedVertices.contains(vertex))
						filteredGraph.addVertex(vertex);
			}
			else if (blockedVertices != null && !blockedVertices.isEmpty())
			{
				for (V vertex : graph.getVertices())
					if (!blockedVertices.contains(vertex))
						filteredGraph.addVertex(vertex);
			}
			else
			{
				for (V vertex : graph.getVertices())
					filteredGraph.addVertex(vertex);
			}
			
			if (acceptedEdges != null)
			{
				for (E edge : graph.getEdges())
				{
					if (!acceptedEdges.contains(edge)) continue;
					Collection<V> incident = graph.getIncidentVertices(edge);
					
					boolean allIncidents = true;
					for(V vertex : incident)
					{
						if (!filteredGraph.containsVertex(vertex))
						{
							allIncidents = false;
							break;
						}
					}
					
					if (allIncidents) filteredGraph.addEdge(edge, incident);
				}
			}
			else if (blockedEdges != null && !blockedEdges.isEmpty())
			{
				for (E edge : graph.getEdges())
				{
					if (blockedEdges.contains(edge)) continue;
					
					Collection<V> incident = graph.getIncidentVertices(edge);
					
					boolean allIncidents = true;
					for(V vertex : incident)
					{
						if (!filteredGraph.containsVertex(vertex))
						{
							allIncidents = false;
							break;
						}
					}
					
					if (allIncidents) filteredGraph.addEdge(edge, incident);
				}
			}
			else
			{
				for (E edge : graph.getEdges())
				{
					Collection<V> incident = graph.getIncidentVertices(edge);
					
					boolean allIncidents = true;
					for(V vertex : incident)
					{
						if (!filteredGraph.containsVertex(vertex))
						{
							allIncidents = false;
							break;
						}
					}
					
					if (allIncidents) filteredGraph.addEdge(edge, incident);
				}
			}
			
			return filteredGraph;
		}

		/**
		 * <p>Filters a graph removing those edges whose weight is equal to {@code Double.MAX_VALUE}.</p>
		 *
		 * <p><b>Important</b>: Returned graph is not backed in the input one, so changes
		 * will not be reflected on it.
		 * 
		 * @param <V> Class type for vertices
		 * @param <E> Class type for edges
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @return Filtered graph
		 * @since 0.3.0
		 */
		public static <V, E> Graph<V, E> filterGraph(Graph<V, E> graph, final Transformer<E, Double> nev)
		{
			return filterGraph(graph, nev, null, 0);
		}
		
		/**
		 * <p>Filters a graph removing those edges whose weight is equal to {@code Double.MAX_VALUE}, 
		 * or whose capacity is below a certain threshold.</p>
		 *
		 * <p><b>Important</b>: Returned graph is not backed in the input one, so changes
		 * will not be reflected on it.
		 * 
		 * @param <V> Class type for vertices
		 * @param <E> Class type for edges
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @param edgeSpareCapacityTransformer Object responsible for returning capacity for edges (if null, it will not applied)
		 * @param requiredCapacity Capacity threshold. Edges whose capacity is below that value it will be removed
		 * @return Filtered graph
		 * @since 0.3.0
		 */
		public static <V, E> Graph<V, E> filterGraph(Graph<V, E> graph, final Transformer<E, Double> nev, final Transformer<E, Double> edgeSpareCapacityTransformer, final double requiredCapacity)
		{
			EdgePredicateFilter<V, E> linkFilter = new EdgePredicateFilter<V, E>(new Predicate<E>()
			{
				@Override
				public boolean evaluate(E edge)
				{
					if (nev != null && nev.transform(edge) == Double.MAX_VALUE) return false;
					else if (edgeSpareCapacityTransformer != null && edgeSpareCapacityTransformer.transform(edge) < requiredCapacity) return false;
					
					return true;
				}
			});

			return linkFilter.transform(graph);
		}
		
		/**
		 * Returns the shortest path that fulfills a given minimum capacity requirement
		 * along its traversed edges. In case no path can be found, an empty list will
		 * be returned.
		 * 
		 * @param <V> Class type for vertices
		 * @param <E> Class type for edges
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @param originNodeId Origin node
		 * @param destinationNodeId Destination node
		 * @param edgeSpareCapacityTransformer Object responsible for returning capacity for edges (if null, it will not applied)
		 * @param requiredCapacity Capacity threshold. Edges whose capacity is below that value it will be removed
		 * @return Shortest path fulfilling a minimum capacity requirement
		 * @since 0.3.0
		 */
		public static <V, E> List<E> getCapacitatedShortestPath(Graph<V, E> graph, Transformer<E, Double> nev, V originNodeId, V destinationNodeId, Transformer<E, Double> edgeSpareCapacityTransformer, double requiredCapacity)
		{
			if (!graph.containsVertex(originNodeId)) return new LinkedList<E>();
			if (!graph.containsVertex(destinationNodeId)) return new LinkedList<E>();
			
			if (nev == null) nev = getEdgeWeightTransformer(null);
			
			Graph<V, E> filteredGraph = filterGraph(graph, nev, edgeSpareCapacityTransformer, requiredCapacity);
			DijkstraShortestPath<V, E> dsp = new DijkstraShortestPath<V, E>(filteredGraph, nev);
			List<E> path = dsp.getPath(originNodeId, destinationNodeId);
			return path;
		}
		
		/**
		 * Obtains a transformer for returning link weight from link identifier.
		 *
		 * @param <E> Edge type
		 * @param edgeWeightMap A mapping of edges to weights (null means 'all-one' map)
		 * @return A transformer for returning weights for edges
		 * @since 0.3.0
		 */
		public static <E> Transformer<E, Double> getEdgeWeightTransformer(final Map<E, Double> edgeWeightMap)
		{
			if (edgeWeightMap == null)
			{
				return new Transformer<E, Double>()
				{
					@Override
					public Double transform(E edge)
					{
						return 1.0;
					}
				};
			}
			else
			{
				return new Transformer<E, Double>()
				{
					@Override
					public Double transform(E edge)
					{
						Double value = edgeWeightMap.get(edge);
						if (value == null) throw new Net2PlanException("Bad - No weight for link " + edge);
						return value;
					}
				};
			}
		}

		/**
		 * Obtains a {@code JUNG} graph from a given link map.
		 * 
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. Origin/destination nodes will be added as needed
		 * @return {@code JUNG} graph
		 * @since 0.3.1
		 */
		public static <V, E> Graph<V, E> getGraphFromLinkMap(Map<E, Pair<V, V>> linkMap)
		{
			Graph<V, E> graph = new DirectedOrderedSparseMultigraph<V, E>();
			
			if (linkMap != null)
			{
				for(Entry<E, Pair<V, V>> entry : linkMap.entrySet())
				{
					E linkId = entry.getKey();
					V originNodeId = entry.getValue().getFirst();
					V destinationNodeId = entry.getValue().getSecond();

					if (!graph.containsVertex(originNodeId)) graph.addVertex(originNodeId);
					if (!graph.containsVertex(destinationNodeId)) graph.addVertex(destinationNodeId);

					graph.addEdge(linkId, originNodeId, destinationNodeId);
				}
			}

			return graph;
		}

		/**
		 * <p>Obtains a {@code JUNG} graph from a given link map.</p>
		 * 
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param nodeIds Set of node identifiers
		 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. Only links whose end-nodes are in {@code nodeIds} will be added to the graph
		 * @return {@code JUNG} graph
		 * @since 0.3.1
		 */
		public static <V, E> Graph<V, E> getGraphFromLinkMap(Set<V> nodeIds, Map<E, Pair<V, V>> linkMap)
		{
			Graph<V, E> graph = new DirectedOrderedSparseMultigraph<V, E>();
			
			if (nodeIds != null)
				for(V nodeId : nodeIds)
					graph.addVertex(nodeId);
			
			if (linkMap != null)
			{
				for(Entry<E, Pair<V, V>> entry : linkMap.entrySet())
				{
					E linkId = entry.getKey();
					V originNodeId = entry.getValue().getFirst();
					V destinationNodeId = entry.getValue().getSecond();

					if (!graph.containsVertex(originNodeId)) continue;
					if (!graph.containsVertex(destinationNodeId)) continue;

					graph.addEdge(linkId, originNodeId, destinationNodeId);
				}
			}

			return graph;
		}
		
		/**
		 * Returns the K-loopless shortest paths between two nodes. If only <i>n<K</i>
		 * could be found, only these ones will be returned.
		 * 
		 * @param <V> Class type for vertices
		 * @param <E> Class type for edges
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @param originNodeId Origin node
		 * @param destinationNodeId Destination node
		 * @param K Number of different paths
		 * @return K-shortest paths
		 * @since 0.3.0
		 */
		public static <V, E> List<List<E>> getKLooplessShortestPaths(Graph<V, E> graph, Transformer<E, Double> nev, V originNodeId, V destinationNodeId, int K)
		{
			if (!graph.containsVertex(originNodeId)) return new LinkedList<List<E>>();
			if (!graph.containsVertex(destinationNodeId)) return new LinkedList<List<E>>();
			
			if (nev == null) nev = getEdgeWeightTransformer(null);
			
			Graph<V, E> filteredGraph = filterGraph(graph, nev);
			YenAlgorithm<V, E> paths = new YenAlgorithm<V, E>(filteredGraph, nev);
			List<List<E>> pathList = paths.getPaths(originNodeId, destinationNodeId, K);
			return pathList;
		}
		
		/**
		 * Returns the weight of a path given the sequence of edges.
		 *
		 * @param <E> Class type for edges
		 * @param path Sequence of edges
		 * @param nev Object responsible for returning weights for edges
		 * @return Path weight
		 * @since 0.2.0
		 */
		public static <E> double getPathWeight(List<E> path, Transformer<E, Double> nev)
		{
			double pathWeight = 0;
			if (nev == null) nev = getEdgeWeightTransformer(null);
			for (E edge : path) pathWeight += nev.transform(edge);

			return pathWeight;
		}
		
		/**
		 * Returns the shortest path between two nodes using Dijkstra's algorithm.
		 * 
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @param originNodeId Origin node
		 * @param destinationNodeId Destination node
		 * @return Shortest path
		 * @since 0.3.0
		 */
		public static <V, E> List<E> getShortestPath(Graph<V, E> graph, Transformer<E, Double> nev, V originNodeId, V destinationNodeId)
		{
			return getCapacitatedShortestPath(graph, nev, originNodeId, destinationNodeId, null, 0);
		}

		/**
		 * Returns the shortest pair of link-disjoint paths, where each item represents a path.
		 * The number of returned items will be equal to the number of paths found: when empty, 
		 * no path was found; when {@code size()} = 1, only one path was found; and when 
		 * {@code size()} = 2, the link-disjoint paths were found. Internally it uses the Suurballe-Tarjan algorithm.
		 * 
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @param originNodeId Origin node
		 * @param destinationNodeId Destination node
		 * @return Shortest pair of link-disjoint paths
		 * @since 0.3.0
		 */
		public static <V, E> List<List<E>> getTwoLinkDisjointPaths(Graph<V, E> graph, Transformer<E, Double> nev, V originNodeId, V destinationNodeId)
		{
			SuurballeTarjanAlgorithm<V, E> suurballeTarjanAlgorithm = new SuurballeTarjanAlgorithm<V, E>(graph, nev);
			return suurballeTarjanAlgorithm.getDisjointPaths(originNodeId, destinationNodeId);
		}
		
		/**
		 * Returns the shortest pair of node-disjoint paths, where each item represents a path.
		 * The number of returned items will be equal to the number of paths found: when empty, 
		 * no path was found; when {@code size()} = 1, only one path was found; and when 
		 * {@code size()} = 2, the node-disjoint paths were found. Internally it uses the Suurballe-Tarjan algorithm.
		 * 
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @param originNode Origin node
		 * @param destinationNode Origin node
		 * @return Shortest pair of node-disjoint paths
		 * @since 0.3.0
		 */
		public static List<List<Long>> getTwoNodeDisjointPaths(final Graph<Long, Long> graph, final Transformer<Long, Double> nev, long originNode, long destinationNode)
		{
			List<List<Long>> nodeDisjointSPs = new LinkedList<List<Long>>();
			if (graph.getVertexCount() < 2 || !graph.containsVertex(originNode) || !graph.containsVertex(destinationNode))
				return nodeDisjointSPs;

			Pair<Graph<Long, Long>, Transformer<Long, Double>> aux = buildAuxiliaryNodeDisjointGraph(graph, nev, originNode, destinationNode);
			Graph<Long, Long> auxGraph = aux.getFirst();
			Transformer<Long, Double> auxNev = aux.getSecond();
			
			nodeDisjointSPs = getTwoLinkDisjointPaths(auxGraph, auxNev, originNode, destinationNode);
			for(List<Long> auxSP : nodeDisjointSPs)
			{
				Iterator<Long> it = auxSP.iterator();
				while(it.hasNext())
				{
					Long edge = it.next();
					if (!graph.containsEdge(edge)) it.remove();
				}
			}

			return nodeDisjointSPs;
		}

		/**
		 * Check whether the topology has the same number of links between each node pair in both directions (assuming multi-digraphs).
		 *
		 * @param graph The graph to analyze
		 * @return {@code true} if the graph is bidirectional, and false otherwise
		 * @since 0.2.0
		 */
		public static boolean isBidirectional(Graph graph)
		{
			Object[] vertices = graph.getVertices().toArray();
			for(int v1 = 0; v1 < vertices.length; v1++)
			{
				for(int v2 = v1 + 1; v2 < vertices.length; v2++)
				{
					if (graph.findEdgeSet(vertices[v1], vertices[v2]).size() != graph.findEdgeSet(vertices[v2], vertices[v1]).size())
						return false;
				}
			}
			
			return true;
		}

		/**
		 * Check whether the graph is simple, that is, if it has at most one link between each node pair (one per direction under directed graphs, one under undirected graphs).
		 *
		 * @param graph The graph to analyze
		 * @return {@code true} if the graph is simple, and false otherwise
		 * @since 0.2.0
		 */
		public static boolean isSimple(Graph graph)
		{
			Object[] vertices = graph.getVertices().toArray();
			for(int v1 = 0; v1 < vertices.length; v1++)
			{
				for(int v2 = v1 + 1; v2 < vertices.length; v2++)
				{
					if (graph.findEdgeSet(vertices[v1], vertices[v2]).size() > 1) return false;
					if (graph.findEdgeSet(vertices[v2], vertices[v1]).size() > 1) return false;
				}
			}
			
			return true;
		}

		/**
		 * Checks whether the graph has the same number of links between each node pair in both directions (assuming multi-digraphs) and same individual weights per direction.
		 *
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param graph The graph to analyze
		 * @param nev Object responsible for returning weights for edges
		 * @return {@code true} if the graph is weighted-bidirectional, and false otherwise. By convention returns {@code false} if network is empty 
		 * @since 0.2.0
		 */
		public static <V, E> boolean isWeightedBidirectional(Graph<V, E> graph, Transformer<E, Double> nev)
		{
			Object[] vertices = graph.getVertices().toArray();
			if (vertices.length == 0) return false;
			
			for(int v1 = 0; v1 < vertices.length; v1++)
			{
				for(int v2 = v1 + 1; v2 < vertices.length; v2++)
				{
					Collection<E> links_12 = graph.findEdgeSet((V) vertices[v1], (V) vertices[v2]);
					Collection<E> links_21 = graph.findEdgeSet((V) vertices[v2], (V) vertices[v1]);
					
					if (links_12.size() != links_21.size()) return false;
					
					Iterator<E> it_12 = links_12.iterator();
					while (it_12.hasNext())
					{
						E aux_12 = it_12.next();

						Iterator<E> it_21 = links_21.iterator();
						while (it_21.hasNext())
						{
							E aux_21 = it_21.next();

							if (Math.abs(nev.transform(aux_12) - nev.transform(aux_21)) < 1e-10)
							{
								it_12.remove();
								it_21.remove();
								break;
							}
						}
					}

					if (!links_12.isEmpty() || !links_12.isEmpty()) return false;
				}
			}
			
			return true;
		}
		
		/**
		 * Given an input graph that may contain multiple edges between some vertex pairs,
		 * returns a new graph where only appears, for each vertex pair, the edge having 
		 * the lowest weight (edges whose weight is equal to {@code Double.MAX_VALUE} 
		 * are excluded). Ties are broken arbitrarely.
		 *
		 * <p><b>Important</b>: Returned graph is not backed in the input one, so changes
		 * will not be reflected on it.
		 *
		 * @param <V> Vertex type
		 * @param <E> Edge type
		 * @param graph Graph representing the network
		 * @param nev Object responsible for returning weights for edges
		 * @return Copy of the input graph with at most an edge between each vertex pair
		 * @since 0.3.0
		 */
		public static <V, E> Graph<V, E> simplifyGraph(Graph<V, E> graph, Transformer<E, Double> nev)
		{
			Collection<V> vertices = graph.getVertices();
			Set<E> edgesToMaintain = new LinkedHashSet<E>();
			for (V originVertex : vertices)
			{
				for (V destinationVertex : vertices)
				{
					if (originVertex == destinationVertex) continue;

					Collection<E> edges = graph.findEdgeSet(originVertex, destinationVertex);
					if (edges.isEmpty()) continue;

					E bestEdge = null;
					double bestWeight = Double.MAX_VALUE;

					for(E edge : edges)
					{
						double weight_thisEdge = nev.transform(edge);
						if (weight_thisEdge < bestWeight)
						{
							bestWeight = weight_thisEdge;
							bestEdge = edge;
						}
					}

					if (bestEdge != null)
						edgesToMaintain.add(bestEdge);
				}
			}

			return JUNGUtils.filterGraph(graph, null, null, edgesToMaintain, null);
		}
		
		private static class MyDefaultVertexLabelRenderer<V> extends DefaultVertexLabelRenderer
		{
			private final Color color;

			public MyDefaultVertexLabelRenderer(Color color)
			{
				super(color);
				this.color = color;
			}

			@Override
			public <V> Component getVertexLabelRendererComponent(JComponent vv, Object value, Font font, boolean isSelected, V vertex)
			{
				Component c = super.getVertexLabelRendererComponent(vv, value, font, isSelected, vertex);
				c.setForeground(color);

				return c;
			}
		}

		private static class NoArrowPredicate<V,E> implements Predicate<Context<Graph<V,E>,E>>
		{
			@Override
			public boolean evaluate(Context<Graph<V,E>,E> context)
			{
				return false;
			}
		}
	}
	
	/**
	 * <p>Class to calculate the shortest link-disjoint path pair between two nodes 
	 * using Suurballe-Tarjan's algorithm.</p>
	 *
	 * <p>Reference: {@code J.W. Suurballe, R.E. Tarjan, "A Quick Method for Finding Shortest Pairs of Disjoint Paths," <i>Networks</i>, vol. 14, no. 2, pp. 325-335, 1984}</p>
	 * 
	 * @param <V> Vertex type
	 * @param <E> Edge type
	 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo Zaragoza
	 * @since 0.3.0
	 */
	public static class SuurballeTarjanAlgorithm<V, E>
	{
		private final Graph<V, E> graph;
		private final Transformer<E, Double> nev;
		private final DijkstraShortestPath<V, E> dijkstra;
		private final boolean cached;

		/**
		 * Default constructor. Previous results from the shortest-path algorithm 
		 * are cached.
		 *
		 * @param graph Graph on which shortest paths are searched
		 * @param nev The class responsible for returning weights for edges
		 * @since 0.3.0
		 */
		public SuurballeTarjanAlgorithm(Graph<V, E> graph, Transformer<E, Double> nev)
		{
			this(graph, nev, true);
		}

		/**
		 * This constructor allows to configure if the shortest-path algorithm should 
		 * cached previous computations.
		 *
		 * @param graph Graph on which shortest paths are searched
		 * @param nev The class responsible for returning weights for edges
		 * @param cached Indicates whether previous computations from the shortest-path algorithm should be cached
		 * @since 0.3.0
		 */
		public SuurballeTarjanAlgorithm(Graph<V, E> graph, Transformer<E, Double> nev, boolean cached)
		{
			this.graph = graph;
			this.nev = nev;
			this.cached = cached;

			dijkstra = new DijkstraShortestPath<V, E>(graph, nev, cached);
		}

		/**
		 * <p>Returns the shortest link-disjoint path pair (in increasing order of weight).</p>
		 * <p><b>Important</b>: If only one path can be found, only such a path will be returned.</p>
		 *
		 * @param startVertex Start vertex of the calculated paths
		 * @param endVertex Target vertex of the calculated paths
		 * @return List of paths in increasing order of weight
		 * @since 0.3.0
		 */
		public List<List<E>> getDisjointPaths(V startVertex, V endVertex)
		{
			List<List<E>> linkDisjointSPs = new LinkedList<List<E>>();

			if (!graph.containsVertex(startVertex) || !graph.containsVertex(endVertex) || startVertex.equals(endVertex))
				return linkDisjointSPs;

			/* If target is not reachable, return */
			if (dijkstra.getDistance(startVertex, endVertex) == null) return linkDisjointSPs;

			List<E> sp = dijkstra.getPath(startVertex, endVertex);
			
			/* Determine length of shortest path from "source" to any other node */
			Map<V, Number> lengthMap = dijkstra.getDistanceMap(startVertex);

			/* Length transformation */
			Transformer<E, Double> lengthTrans = lengthTransformation(graph, MapTransformer.getInstance(lengthMap));

			/* Get shortest path in g with reversed shortest path... */
			Graph<V, E> revG = reverseEdges(graph, sp);
			DijkstraShortestPath<V, E> revDijkstra = new DijkstraShortestPath<V, E>(revG, lengthTrans, cached);

			Number revDistance = revDijkstra.getDistance(startVertex, endVertex);
			if (revDistance == null || revDistance.doubleValue() == Double.MAX_VALUE)
			{
				/* no alternate path, return */
				linkDisjointSPs.add(sp);
				return linkDisjointSPs;
			}

			List<E> revSp = revDijkstra.getPath(startVertex, endVertex);
			
			validatePath(graph, startVertex, endVertex, sp);
			validatePath(revG, startVertex, endVertex, revSp);

			List<E> spCopy = new LinkedList<E>(sp);
			List<List<E>> paths = findDisjointPaths(sp, revSp);

			if (paths == null)
			{
				/* no disjoint solution found, just return shortest path */
				linkDisjointSPs.add(spCopy);
				return linkDisjointSPs;
			}

			/* Check path validity */
			for (List<E> path : paths) validatePath(graph, startVertex, endVertex, path);

			return paths;
		}

		private static <V, E> void validatePath(Graph<V, E> graph, V source, V target, List<E> path)
		{
			if (!graph.isSource(source, path.get(0))) throw new RuntimeException("Bad - Source node is not the first node in the path");

			Iterator<E> it = path.iterator();
			E originVertex = it.next();

			while (it.hasNext())
			{
				E destinationVertex = it.next();
				if (!graph.isSource(graph.getDest(originVertex), destinationVertex)) throw new RuntimeException("Bad - Path is not contiguous");

				originVertex = destinationVertex;
			}

			if (!graph.isDest(target, path.get(path.size() - 1))) throw new RuntimeException("Bad - ");
		}

		/**
		 * Combines two disjoint paths from two SuurballeTarjan input paths.
		 * 
		 * @param path1 Dijkstra shortest path
		 * @param path2 Dijkstra shortest path in partly reverted graph
		 * @return the two disjoint paths
		 * @since 0.3.0
		 */
		private List<List<E>> findDisjointPaths(List<E> path1, List<E> path2)
		{
			final V source = graph.getSource(path1.get(0));
			final V target = graph.getDest(path1.get(path1.size() - 1));

			/* First, remove common links */
			Iterator<E> path1_it = path1.iterator();
			while(path1_it.hasNext())
			{
				E e1 = path1_it.next();

				Iterator<E> path2_it = path2.iterator();
				while(path2_it.hasNext())
				{
					E e2 = path2_it.next();

					if (e1.equals(e2))
					{
						if (graph.isSource(source, e1) || graph.isSource(source, e2) || graph.isDest(target, e1) || graph.isDest(target, e2))
							return null;

						path1_it.remove();
						path2_it.remove();
						break;
					}
				}
			}

			/* no disjoint solution found */
			if (path1.isEmpty() || path2.isEmpty()) return null; 
			
			/* Now recombine the two paths */
			List<E> union = ListUtils.union(path1, path2); /* concatenate */

			List<E> p1 = recombinePaths(path1, target, union);
			if (p1 == null) return null;

			List<E> p2 = recombinePaths(path2, target, union);
			if (p2 == null) return null;

//			if (!union.isEmpty()) throw new RuntimeException("Bad"); /* ToDo: It is an error? */

			List<List<E>> solution = new LinkedList<List<E>>();

			double path1_cost = 0;
			for(E edge : p1) path1_cost += nev.transform(edge);

			double path2_cost = 0;
			for(E edge : p2) path2_cost += nev.transform(edge);

			if (path1_cost <= path2_cost)
			{
				solution.add(p1);
				solution.add(p2);
			}
			else
			{
				solution.add(p2);
				solution.add(p1);
			}

			return solution;
		}

		private List<E> recombinePaths(List<E> path, V target, List<E> union)
		{
			LinkedList<E> p = new LinkedList<E>(); /* provides getLast */
			p.add(path.get(0));
			union.remove(path.get(0));

			V curDest;
			while (!(curDest = graph.getDest(p.getLast())).equals(target))
			{
				boolean progress = false;
				for (E e : union)
				{
					if (graph.isSource(curDest, e))
					{
						p.add(e);
						progress = true;
						union.remove(e);
						break;
					}
				}

				if (!progress) return null;

				if (union.isEmpty())
				{
					if (!graph.isDest(target, p.getLast())) throw new RuntimeException("Bad");
					else break;
				}
			}
			return p;
		}

		/**
		 * This method reverse the path "path" in the graph "graph" and returns it.
		 * 
		 * @param graph the input graph which will not be changed.
		 * @param path the path to reverse
		 * @return a new graph with the reversed path
		 * @since 0.3.0
		 */
		private static <V, E> Graph<V, E> reverseEdges(Graph<V, E> graph, List<E> path)
		{
			if (graph == null || path == null) throw new IllegalArgumentException();
			Graph<V, E> clone = new DirectedOrderedSparseMultigraph<V, E>();

			for (V v : graph.getVertices()) clone.addVertex(v);
			for (E e : graph.getEdges()) clone.addEdge(e, graph.getEndpoints(e));

			for (E link : path)
			{
				V src = clone.getSource(link);
				V dst = clone.getDest(link);
				clone.removeEdge(link);
				clone.addEdge(link, dst, src, EdgeType.DIRECTED);
			}

			return clone;
		}

		/**
		 * This method does the following length transformation:
		 * 
		 * <pre>
		 *  c'(v,w) = c(v,w) - d (s,w) + d (s,v)
		 * </pre>
		 * 
		 * @param graph1 the graph
		 * @param slTrans The shortest length transformer
		 * @return the transformed graph
		 * @since 0.3.0
		 */
		private Transformer<E, Double> lengthTransformation(Graph<V, E> graph1, Transformer<V, Number> slTrans)
		{
			Map<E, Double> map = new LinkedHashMap<E, Double>();

			for (E link : graph1.getEdges())
			{
				double newWeight;
				
				if (slTrans.transform(graph1.getSource(link)) == null)
				{
					newWeight = Double.MAX_VALUE;
				}
				else
				{
					newWeight = nev.transform(link) - slTrans.transform(graph1.getDest(link)).doubleValue() + slTrans.transform(graph1.getSource(link)).doubleValue();
					if (Math.abs(newWeight) < 1e-6) newWeight = 0; /* Numerical errors */
				}
				
				map.put(link, newWeight);
			}

			return MapTransformer.getInstance(map);
		}
	}

	/**
	 * <p>Class to calculate the (loopless) <i>k</i>-shortest paths between a node 
	 * pair using Yen's algorithm.</p>
	 * 
	 * <p>Reference: {@code J.Y. Yen, "Finding the K Shortest Loopless Paths in a Network," <i>Management Science</i>, vol. 17, no. 11, pp. 712-716, Jul. 1971}</p>
	 *
	 * @param <V> Vertex type
	 * @param <E> Edge type
	 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo Zaragoza
	 * @since 0.3.0
	 */
	public static class YenAlgorithm<V, E>
	{
		private final Graph<V, E> graph;
		private final Transformer<E, Double> nev;
		private final DijkstraShortestPath<V, E> dijkstra;

		/**
		 * Default constructor.
		 *
		 * @param graph Graph on which shortest paths are searched
		 * @param nev The class responsible for returning weights for edges
		 * @since 0.3.0
		 */
		public YenAlgorithm(Graph<V, E> graph, Transformer<E, Double> nev)
		{
			if (nev == null)
			{
				this.graph = graph;
				this.nev = JUNGUtils.getEdgeWeightTransformer(null);
			}
			else
			{
				this.nev = nev;
				this.graph = JUNGUtils.filterGraph(graph, nev);
			}

			dijkstra = new DijkstraShortestPath<V, E>(graph, nev);
		}

		/**
		 * <p>Returns the (loopless) <i>k</i>-shortest simple paths in increasing order of weight.</p>
		 * <p><b>Important</b>: If only <i>n</i> < <i>k</i> paths can be found, only such <i>n</i> paths will be returned.</p>
		 *
		 * @param startVertex Start vertex of the calculated paths
		 * @param endVertex Target vertex of the calculated paths
		 * @param k Number of paths to be computed
		 * @return List of paths in increasing order of weight
		 * @since 0.2.0
		 */
		public List<List<E>> getPaths(V startVertex, V endVertex, int k)
		{
			LinkedList<List<E>> paths = new LinkedList<List<E>>();

			if (!graph.containsVertex(startVertex) || !graph.containsVertex(endVertex) || startVertex.equals(endVertex)) return paths;
			if (dijkstra.getDistance(startVertex, endVertex) == null) return paths;

			PriorityQueue<GraphPath> priorityQueue = new PriorityQueue<GraphPath>();
			List<E> aux = dijkstra.getPath(startVertex, endVertex);
			double cost = GraphUtils.JUNGUtils.getPathWeight(aux, nev);
			GraphPath<E> shortestPath = new GraphPath<E>(aux, cost);

			if (!acceptPath(shortestPath)) return paths;
			paths.add(aux);

			DijkstraShortestPath<V, E> blockedDijkstra;

			while (paths.size() < k)
			{
				List<E> curShortestPath = paths.getLast();

				int currentPathLength = curShortestPath.size();

				/* Split path into Head and NextEdge */
				for (int deviationId = 0; deviationId < currentPathLength; deviationId++)
				{
					List<E> head = curShortestPath.subList(0, deviationId);
					V deviationVertex = head.isEmpty() ? startVertex : graph.getDest(head.get(deviationId - 1));

					/* 1. Block edges */
					Graph<V, E> blocked = blockFilter(head, deviationVertex, paths);

					/* 2. Get shortest path in graph with blocked edges */
					blockedDijkstra = new DijkstraShortestPath<V, E>(blocked, nev);

					Number dist = blockedDijkstra.getDistance(deviationVertex, endVertex);
					if (dist == null) continue;

					List<E> tail = blockedDijkstra.getPath(deviationVertex, endVertex);

					/* 3. Combine head and tail into new path */
					List<E> candidatePath = new ArrayList<E>(deviationId + tail.size());
					candidatePath.addAll(head);
					candidatePath.addAll(tail);

					GraphPath<E> candidate = new GraphPath<E>(candidatePath, JUNGUtils.getPathWeight(candidatePath, nev));

					/* Check if we already found this solution */
					if (priorityQueue.contains(candidate)) continue;

					if (!acceptPath(candidate) || !compareCandidateToShortestPath(candidate, shortestPath)) continue;

					priorityQueue.add(candidate);
				}

				if (priorityQueue.isEmpty()) break; /* No more candidate paths */
				paths.add(priorityQueue.poll().getPath());
			}

			return paths;
		}

		/**
		 * Blocks all incident edges of the vertices in head as well as the edge connecting head to the next node by creating a new filtered graph.
		 *
		 * @param head The current head, from source to deviation node
		 * @param deviation The edge to the next node
		 * @param foundPaths The solutions already found and to check against
		 * @return The filtered graph without the blocked edges
		 * @since 0.3.0
		 */
		private Graph<V, E> blockFilter(List<E> head, V deviation, List<List<E>> foundPaths)
		{
			final Set<E> blocked = new LinkedHashSet<E>();

			/* Block incident edges to make all vertices in head unreachable. */
			for (E e : head)
			{
				for (E e2 : graph.getIncidentEdges(graph.getSource(e)))
				{
					blocked.add(e2);
				}
			}

			/* Block all outgoing edges that have been used at deviation vertex */
			for (List<E> path : foundPaths)
			{
				if (path.size() > head.size() && ListUtils.isEqualList(path.subList(0, head.size()), head))
				{
					for (E e : path)
					{
						if (graph.isSource(deviation, e))
						{
							blocked.add(e);
							break; /* Continue with next path */
						}
					}
				}
			}
			
			return JUNGUtils.filterGraph(graph, null, null, null, blocked);
		}

		/**
		 * Indicates whether a path is valid. It can be override to implement checks 
		 * like maximum path length. In contrast, {@link #compareCandidateToShortestPath(com.net2plan.libraries.GraphUtils.GraphPath, com.net2plan.libraries.GraphUtils.GraphPath) compareCandidateToShortestPath()} method.
		 * method can be used to compare current candidate path with the shortest path.
		 * 
		 * @param candidate Candidate path
		 * @return {@code true} must be accepted. Otherwise, {@code false}
		 * @since 0.3.0
		 */
		public boolean acceptPath(GraphPath<E> candidate)
		{
			return true;
		}

		/**
		 * Indicates whether a path is valid when compared to the shortest path. It 
		 * can be override to implement checks like maximum distance from the shortest path.
		 * 
		 * @param candidate Candidate path
		 * @param shortestPath
		 * @return {@code true} must be accepted. Otherwise, {@code false}
		 * @since 0.3.0
		 */
		public boolean compareCandidateToShortestPath(GraphPath<E> candidate, GraphPath<E> shortestPath)
		{
			return true;
		}
	}
	
	/**
	 * Exception thrown when hop-by-hop routing includes closed cycles.
	 *
	 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
	 * @since 0.3.0
	 */
	public static class ClosedCycleRoutingException extends Net2PlanException
	{
		/**
		 * Default constructor.
		 *
		 * @since 0.3.0
		 */
		public ClosedCycleRoutingException()
		{
			super();
		}
		
		/**
		 * Constructs a new {@code Net2PlanException} exception with the specified detail message.
		 *
		 * @param message Message to be retrieved by the {@link #getMessage()} method.
		 * @since 0.2.0
		 */
		public ClosedCycleRoutingException(String message)
		{
			super(message);
		}
	}
}