/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import cern.colt.matrix.tdouble.DoubleFactory2D;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils.JUNGUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Quadruple;
import java.util.Map.Entry;
import java.util.*;
import org.apache.commons.collections15.Transformer;

/**
 * Class for destination-based routing (IP-like).
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 */
public class IPUtils
{
	/**
	 * Link attribute name for IP weight setting.
	 * 
	 * @since 0.3.0
	 */
	public final static String IP_WEIGHT_ATTRIBUTE_NAME = "linkWeight";
	
	private IPUtils() { }
	
	/**
	 * <p>Returns the cost of a routing scheme according to the carried traffic per 
	 * link and link capacities, assuming the IGP-WO cost model.</p>
	 * 
	 * <p>Let <i>E</i> be a set of links, where <i>u<sub>e</sub></i> is the capacity 
	 * for link <i>e &isin; E</i>, and <i>y<sub>e</sub></i> is the carried traffic 
	 * by link <i>e &isin; E</i>. Then, the total cost of the associated routing, 
	 * denoted as <i>C</i>, is the sum for every link <i>e &isin; E</i> of its 
	 * associated cost <i>c<sub>e</sub></i>, which is computed as follows:</p>
	 * 
	 * <ul>
	 * <li><i>c<sub>e</sub></i> =        <i>y<sub>e</sub></i>                                     , if <i>&rho;<sub>e</sub> &isin; [0, 1/3)</i></li>
	 * <li><i>c<sub>e</sub></i> =    3 * <i>y<sub>e</sub></i> -     (2 / 3) * <i>u<sub>e</sub></i>, if <i>&rho;<sub>e</sub> &isin; [1/3, 2/3)</i></li>
	 * <li><i>c<sub>e</sub></i> =   10 * <i>y<sub>e</sub></i> -    (16 / 3) * <i>u<sub>e</sub></i>, if <i>&rho;<sub>e</sub> &isin; [2/3, 9/10)</i></li>
	 * <li><i>c<sub>e</sub></i> =   70 * <i>y<sub>e</sub></i> -   (178 / 3) * <i>u<sub>e</sub></i>, if <i>&rho;<sub>e</sub> &isin; [9/10, 1)</i></li>
	 * <li><i>c<sub>e</sub></i> =  500 * <i>y<sub>e</sub></i> -  (1468 / 3) * <i>u<sub>e</sub></i>, if <i>&rho;<sub>e</sub> &isin; [1, 11/10)</i></li>
	 * <li><i>c<sub>e</sub></i> = 5000 * <i>y<sub>e</sub></i> - (16318 / 3) * <i>u<sub>e</sub></i>, otherwise</li>
	 * </ul>
	 * 
	 * <p>where <i>&rho;<sub>e</sub></i> is the ratio <i>y<sub>e</sub></i> / <i>u<sub>e</sub></i>.</p>
	 * 
	 * <p>Reference: {@code H. Ümit, "Techniques and Tools for Intra-domain Traffic Engineering," Ph.D. Thesis, Université Catholique de Louvain (Belgium), December 2009}</p>
	 * 
	 * @param y_e Carried traffic per link
	 * @param u_e Capacity per link
	 * @return IGP cost
	 * @since 0.3.0
	 */
	public static double calculateIGPCost(Map<Long, Double> y_e, Map<Long, Double> u_e)
	{
		double cost = 0;
		
		Set<Long> linkIds = y_e.keySet();
		for(long linkId : linkIds)
		{
			if (!u_e.containsKey(linkId)) throw new Net2PlanException("Bad");
			double y_e_thisLink = y_e.get(linkId);
			double u_e_thisLink = u_e.get(linkId);

			double phi = 0;

			double aux = y_e_thisLink;
			if (aux >= phi) phi = aux;

			aux = 3.0 * y_e_thisLink - (2.0 / 3.0) * u_e_thisLink;
			if (aux >= phi) phi = aux;

			aux = 10.0 * y_e_thisLink - (16.0 / 3.0) * u_e_thisLink;
			if (aux >= phi) phi = aux;

			aux = 70.0 * y_e_thisLink - (178.0 / 3.0) * u_e_thisLink;
			if (aux >= phi) phi = aux;

			aux = 500.0 * y_e_thisLink - (1468.0 / 3.0) * u_e_thisLink;
			if (aux >= phi) phi = aux;

			aux = 5000.0 * y_e_thisLink - (16318.0 / 3.0) * u_e_thisLink;
			if (aux >= phi) phi = aux;

			cost += phi;
		}

		return cost;
	}
	
	/**
	 * <p>Returns the cost of a routing scheme according to the carried traffic per 
	 * link and the link capacity vectors, assuming the IGP-WO cost model.</p>
	 * 
	 * <p>For more information, see {@link #calculateIGPCost(java.util.Map, java.util.Map)}</p>
	 * 
	 * @param y_e Carried traffic per link vector
	 * @param u_e Link capacity vector
	 * @return IGP cost
	 * @since 0.2.3
	 * @see #calculateIGPCost(java.util.Map, java.util.Map)
	 */
	public static double calculateIGPCost(double[] y_e, double[] u_e)
	{
		int E = u_e.length;
		double cost = 0;

		for (int linkId = 0; linkId < E; linkId++)
		{
			double phi = 0;

			double aux = y_e[linkId];
			if (aux >= phi) phi = aux;

			aux = 3.0 * y_e[linkId] - (2.0 / 3.0) * u_e[linkId];
			if (aux >= phi) phi = aux;

			aux = 10.0 * y_e[linkId] - (16.0 / 3.0) * u_e[linkId];
			if (aux >= phi) phi = aux;

			aux = 70.0 * y_e[linkId] - (178.0 / 3.0) * u_e[linkId];
			if (aux >= phi) phi = aux;

			aux = 500.0 * y_e[linkId] - (1468.0 / 3.0) * u_e[linkId];
			if (aux >= phi) phi = aux;

			aux = 5000.0 * y_e[linkId] - (16318.0 / 3.0) * u_e[linkId];
			if (aux >= phi) phi = aux;

			cost += phi;
		}

		return cost;
	}
	
	private static void checkIPWeight(double linkWeight)
	{
		if (linkWeight < 1) throw new Net2PlanException("Link weights must be greater or equal than 1");
	}
	
	/**
	 * Computes the resulting carried traffic according to a link weight setting 
	 * and OSPF/ECMP routing.
	 * 
	 * @param netPlan Network design
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means 'takes weight from link attribute'). No special iteration-order (i.e. ascending) is required
	 * @return A {@link com.net2plan.utils.Quadruple Quadruple} composed of {@code f_de} (splitting rules), {@code x_de} (carried traffic for demand 'd' through link 'e'), {@code r_d} (end-to-end carried traffic for demand 'd'), and {@code y_e} (carried traffic by link 'e')
	 * @since 0.3.0
	 */
	public static Quadruple<Map<Pair<Long, Long>, Double>, Map<Pair<Long, Long>, Double>, Map<Long, Double>, Map<Long, Double>> computeCarriedTrafficFromIGPWeights(NetPlan netPlan, Map<Long, Double> linkWeightMap)
	{
		long layerId = netPlan.getLayerDefaultId();
		return computeCarriedTrafficFromIGPWeights(netPlan, layerId, linkWeightMap);
	}
	
	/**
	 * Computes the resulting carried traffic according to a link weight setting 
	 * and OSPF/ECMP routing.
	 * 
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means 'takes weight from link attribute'). No special iteration-order (i.e. ascending) is required
	 * @return A {@link com.net2plan.utils.Quadruple Quadruple} composed of {@code f_de} (splitting rules), {@code x_de} (carried traffic for demand 'd' through link 'e'), {@code r_d} (end-to-end carried traffic for demand 'd'), and {@code y_e} (carried traffic by link 'e')
	 * @since 0.3.0
	 */
	public static Quadruple<Map<Pair<Long, Long>, Double>, Map<Pair<Long, Long>, Double>, Map<Long, Double>, Map<Long, Double>> computeCarriedTrafficFromIGPWeights(NetPlan netPlan, long layerId, Map<Long, Double> linkWeightMap)
	{
		if (linkWeightMap == null) linkWeightMap = IPUtils.getLinkWeightMap(netPlan, layerId);
		
		final Set<Long> nodeIds = netPlan.getNodeIds();
		final Map<Long, Pair<Long, Long>> linkTable = netPlan.getLinkMap(layerId);
		final Map<Long, Pair<Long, Long>> demandTable = netPlan.getDemandMap(layerId);
		final Map<Long, Double> h_d = netPlan.getDemandOfferedTrafficMap(layerId);

		final Map<Pair<Long, Long>, Double> f_te = IPUtils.computeECMPRoutingTableMap(nodeIds, linkTable, linkWeightMap);
		final Quadruple<Map<Pair<Long, Long>, Double>, Map<Long, Double>, Map<Long, Double>, List<Constants.RoutingCycleType>> q_noFailure = GraphUtils.convert_fte2xde(nodeIds, linkTable, demandTable, h_d, f_te);
		final Map<Pair<Long, Long>, Double> x_de = q_noFailure.getFirst();
		final Map<Long, Double> r_d = q_noFailure.getSecond();
		final Map<Long, Double> y_e = q_noFailure.getThird();
		
		final Map<Pair<Long, Long>, Double> f_de = new LinkedHashMap<Pair<Long, Long>, Double>();
		for(Entry<Pair<Long, Long>, Double> entry : f_te.entrySet())
		{
			long egressNodeId = entry.getKey().getFirst();
			long linkId = entry.getKey().getSecond();
			double f_te_thisEgressNode_thisLink = entry.getValue();
			
			Set<Long> demandIds_thisEgressNode = netPlan.getNodeIncomingDemands(layerId, egressNodeId);
			for(long demandId : demandIds_thisEgressNode) f_de.put(Pair.of(demandId, linkId), f_te_thisEgressNode_thisLink);
		}
		
		return Quadruple.of(f_de, x_de, r_d, y_e);
	}
	
	/**
	 * Computes the forwarding rules according to an OSPF/ECMP scheme. Each 
	 * forwarding rule is determined by a demand-outgoing link pair, and its 
	 * corresponding splitting ratio. Splitting ratio is defined as the fraction 
	 * of traffic sent through the given outgoing link, with respect to the total 
	 * traffic entering to the node for the given demand.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param demandMap Map of demands, where the key is the unique demand identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the ingress node and the egress node of the demand, respectively. It is mandatory that can be iterated in ascending order of demand identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means each link will have a weight equal to '1'). No special iteration-order (i.e. ascending) is required
	 * @return Forwarding rule mapping, where key is the demand-outgoing link pair and the value is the splitting ratio
	 * @since 0.3.0
	 */
	public static Map<Pair<Long, Long>, Double> computeECMPForwardingRules(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Pair<Long, Long>> demandMap, Map<Long, Double> linkWeightMap)
	{
		Map<Pair<Long, Long>, Double> splittingRatioMap = new LinkedHashMap<Pair<Long, Long>, Double>();
		
		int N = nodeIds.size();
		Map<Long, Integer> nodeId2LinearIndexMap = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		double[][] costMatrix = new double[N][N];
		for(int n = 0; n < N; n++)
		{
			Arrays.fill(costMatrix[n], Double.MAX_VALUE);
			costMatrix[n][n] = 0;
		}
		
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkWeightMap);
		
		Map<Pair<Long, Long>, Set<Long>> linkIdsPerNodePair = new LinkedHashMap<Pair<Long, Long>, Set<Long>>();
		for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
		{
			long linkId = entry.getKey();
			Pair<Long, Long> nodePair_thisLink = entry.getValue();
			int a_e = nodeId2LinearIndexMap.get(nodePair_thisLink.getFirst());
			int b_e = nodeId2LinearIndexMap.get(nodePair_thisLink.getSecond());
			costMatrix[a_e][b_e] = Math.min(costMatrix[a_e][b_e], nev.transform(linkId));
			
			Set<Long> linkIds_thisNodePair = linkIdsPerNodePair.get(nodePair_thisLink);
			if (linkIds_thisNodePair == null)
			{
				linkIds_thisNodePair = new LinkedHashSet<Long>();
				linkIdsPerNodePair.put(nodePair_thisLink, linkIds_thisNodePair);
			}
			
			linkIds_thisNodePair.add(linkId);
		}

		for(int k = 0; k < N; k++)
		{
			for(int i = 0; i < N; i++)
			{
				if (i == k) continue;
				
				for(int j = 0; j < N; j++)
				{
					if (j == k || j == i) continue;
					
					double newValue = costMatrix[i][k] + costMatrix[k][j];
					if (newValue < costMatrix[i][j]) costMatrix[i][j] = newValue;
				}
			}
		}
		
		for(Entry<Long, Pair<Long, Long>> entry : demandMap.entrySet())
		{
			long demandId = entry.getKey();
			int b_d = nodeId2LinearIndexMap.get(entry.getValue().getSecond());
			for (long nodeId : nodeIds)
			{
				int n = nodeId2LinearIndexMap.get(nodeId);
				if (n == b_d) continue;

				double distNodeToEgress = costMatrix[n][b_d];
				if (distNodeToEgress == Double.MAX_VALUE) continue;

				Set<Long> A_t = new LinkedHashSet<Long>();
				for (long intermediateNodeId : nodeIds)
				{
					if (nodeId == intermediateNodeId) continue;

					int m = nodeId2LinearIndexMap.get(intermediateNodeId);

					double distIntermediateToEgress = costMatrix[m][b_d];
					if (distIntermediateToEgress == Double.MAX_VALUE) continue;

					Collection<Long> linksFromNodeToIntermediate = linkIdsPerNodePair.get(Pair.of(nodeId, intermediateNodeId));
					if (linksFromNodeToIntermediate == null) continue;

					for (long linkId : linksFromNodeToIntermediate)
					{
						double weight_thisLink = linkWeightMap.get(linkId);
						checkIPWeight(weight_thisLink);

						if (Math.abs(weight_thisLink - (distNodeToEgress - distIntermediateToEgress)) < 1E-10)
							A_t.add(linkId);
					}
				}

				int outdegree = A_t.size();

				if (outdegree > 0)
					for (long linkId : A_t)
						splittingRatioMap.put(Pair.of(demandId, linkId), 1.0 / outdegree);
			}
		}

		return splittingRatioMap;
	}
	
	/**
	 * Computes the routing table matrix according to an OSPF/ECMP scheme. For 
	 * each destination node <i>t</i>, and each link <i>e</i>, {@code f_te[t][e]} 
	 * sets the fraction of the traffic targeted to node <i>t</i> that arrives 
	 * (or is generated in) node <i>a(e)</i> (the initial node of link <i>e</i>), 
	 * that is forwarded through link <i>e</i>. It must hold that for every 
	 * node <i>n</i> different of <i>t</i>, the sum of the fractions 
	 * <i>f<sub>te</sub></i> along its outgoing links must be 1. For every 
	 * destination <i>t</i>, <i>f<sub>te</sub></i> = 0 for all the links <i>e</i> 
	 * that are outgoing links of <i>t</i>.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means each link will have a weight equal to '1'). No special iteration-order (i.e. ascending) is required
	 * @return Destination-based routing in the form <i>f<sub>te</sub></i> (fractions of traffic in a node, that is forwarded through each of its output links to node {@code t})
	 * @since 0.3.0
	 */
	public static Map<Pair<Long, Long>, Double> computeECMPRoutingTableMap(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Double> linkWeightMap)
	{
		int N = nodeIds.size();
		Map<Pair<Long, Long>, Double> f_te = new LinkedHashMap<Pair<Long, Long>, Double>();
		
		Map<Long, Integer> nodeId2LinearIndexMap = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		double[][] costMatrix = new double[N][N];
		for(int n = 0; n < N; n++)
		{
			Arrays.fill(costMatrix[n], Double.MAX_VALUE);
			costMatrix[n][n] = 0;
		}
		
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkWeightMap);
		
		Map<Pair<Long, Long>, Set<Long>> linkIdsPerNodePair = new LinkedHashMap<Pair<Long, Long>, Set<Long>>();
		for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
		{
			long linkId = entry.getKey();
			Pair<Long, Long> nodePair_thisLink = entry.getValue();
			int a_e = nodeId2LinearIndexMap.get(nodePair_thisLink.getFirst());
			int b_e = nodeId2LinearIndexMap.get(nodePair_thisLink.getSecond());
			costMatrix[a_e][b_e] = Math.min(costMatrix[a_e][b_e], nev.transform(linkId));
			
			Set<Long> linkIds_thisNodePair = linkIdsPerNodePair.get(nodePair_thisLink);
			if (linkIds_thisNodePair == null)
			{
				linkIds_thisNodePair = new LinkedHashSet<Long>();
				linkIdsPerNodePair.put(nodePair_thisLink, linkIds_thisNodePair);
			}
			
			linkIds_thisNodePair.add(linkId);
		}

		for(int k = 0; k < N; k++)
		{
			for(int i = 0; i < N; i++)
			{
				if (i == k) continue;
				
				for(int j = 0; j < N; j++)
				{
					if (j == k || j == i) continue;
					
					double newValue = costMatrix[i][k] + costMatrix[k][j];
					if (newValue < costMatrix[i][j]) costMatrix[i][j] = newValue;
				}
			}
		}
		
		for (long egressNodeId : nodeIds)
		{
			int t = nodeId2LinearIndexMap.get(egressNodeId);
			for (long nodeId : nodeIds)
			{
				int n = nodeId2LinearIndexMap.get(nodeId);
				if (n == t) continue;

				double distNodeToEgress = costMatrix[n][t];
				if (distNodeToEgress == Double.MAX_VALUE) continue;

				Set<Long> A_t = new LinkedHashSet<Long>();
				for (long intermediateNodeId : nodeIds)
				{
					if (nodeId == intermediateNodeId) continue;

					int m = nodeId2LinearIndexMap.get(intermediateNodeId);

					double distIntermediateToEgress = costMatrix[m][t];
					if (distIntermediateToEgress == Double.MAX_VALUE) continue;

					Collection<Long> linksFromNodeToIntermediate = linkIdsPerNodePair.get(Pair.of(nodeId, intermediateNodeId));
					if (linksFromNodeToIntermediate == null) continue;

					for (long linkId : linksFromNodeToIntermediate)
					{
						double weight_thisLink = linkWeightMap.get(linkId);
						checkIPWeight(weight_thisLink);

						if (Math.abs(weight_thisLink - (distNodeToEgress - distIntermediateToEgress)) < 1E-10)
							A_t.add(linkId);
					}
				}

				int outdegree = A_t.size();

				if (outdegree > 0)
					for (long linkId : A_t)
						f_te.put(Pair.of(egressNodeId, linkId), 1.0 / outdegree);
			}
		}

		return f_te;
	}

	/**
	 * Computes the routing table matrix according to an OSPF/ECMP scheme. For 
	 * each destination node <i>t</i>, and each link <i>e</i>, {@code f_te[t][e]} 
	 * sets the fraction of the traffic targeted to node <i>t</i> that arrives 
	 * (or is generated in) node <i>a(e)</i> (the initial node of link <i>e</i>), 
	 * that is forwarded through link <i>e</i>. It must hold that for every 
	 * node <i>n</i> different of <i>t</i>, the sum of the fractions 
	 * <i>f<sub>te</sub></i> along its outgoing links must be 1. For every 
	 * destination <i>t</i>, <i>f<sub>te</sub></i> = 0 for all the links <i>e</i> 
	 * that are outgoing links of <i>t</i>.
	 * 
	 * @param nodeIds Set of node identifiers. It is mandatory that can be iterated in ascending order of node identifier (i.e. using {@code TreeSet} or those {@code Set} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkMap Map of links, where the key is the unique link identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively. It is mandatory that can be iterated in ascending order of link identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means each link will have a weight equal to '1'). No special iteration-order (i.e. ascending) is required
	 * @return Destination-based routing in the form <i>f<sub>te</sub></i> (fractions of traffic in a node, that is forwarded through each of its output links to node {@code t})
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D computeECMPRoutingTableMatrix(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, Map<Long, Double> linkWeightMap)
	{
		int N = nodeIds.size();
		int E = linkMap.size();
		DoubleMatrix2D f_te = DoubleFactory2D.sparse.make(N, E);
		
		Map<Long, Integer> nodeId2LinearIndexMap = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		double[][] costMatrix = new double[N][N];
		for(int n = 0; n < N; n++)
		{
			Arrays.fill(costMatrix[n], Double.MAX_VALUE);
			costMatrix[n][n] = 0;
		}
		
		Transformer<Long, Double> nev = JUNGUtils.getEdgeWeightTransformer(linkWeightMap);
		
		Map<Pair<Long, Long>, Set<Long>> linkIdsPerNodePair = new LinkedHashMap<Pair<Long, Long>, Set<Long>>();
		for(Entry<Long, Pair<Long, Long>> entry : linkMap.entrySet())
		{
			long linkId = entry.getKey();
			Pair<Long, Long> nodePair_thisLink = entry.getValue();
			int a_e = nodeId2LinearIndexMap.get(nodePair_thisLink.getFirst());
			int b_e = nodeId2LinearIndexMap.get(nodePair_thisLink.getSecond());
			costMatrix[a_e][b_e] = Math.min(costMatrix[a_e][b_e], nev.transform(linkId));
			
			Set<Long> linkIds_thisNodePair = linkIdsPerNodePair.get(nodePair_thisLink);
			if (linkIds_thisNodePair == null)
			{
				linkIds_thisNodePair = new LinkedHashSet<Long>();
				linkIdsPerNodePair.put(nodePair_thisLink, linkIds_thisNodePair);
			}
			
			linkIds_thisNodePair.add(linkId);
		}

		for(int k = 0; k < N; k++)
		{
			for(int i = 0; i < N; i++)
			{
				if (i == k) continue;
				
				for(int j = 0; j < N; j++)
				{
					if (j == k || j == i) continue;
					
					double newValue = costMatrix[i][k] + costMatrix[k][j];
					if (newValue < costMatrix[i][j]) costMatrix[i][j] = newValue;
				}
			}
		}
		
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkMap.keySet());
		for (long egressNodeId : nodeIds)
		{
			int t = nodeId2LinearIndexMap.get(egressNodeId);
			for (long nodeId : nodeIds)
			{
				int n = nodeId2LinearIndexMap.get(nodeId);
				if (n == t) continue;

				double distNodeToEgress = costMatrix[n][t];
				if (distNodeToEgress == Double.MAX_VALUE) continue;

				Set<Long> A_t = new LinkedHashSet<Long>();
				for (long intermediateNodeId : nodeIds)
				{
					if (nodeId == intermediateNodeId) continue;

					int m = nodeId2LinearIndexMap.get(intermediateNodeId);

					double distIntermediateToEgress = costMatrix[m][t];
					if (distIntermediateToEgress == Double.MAX_VALUE) continue;

					Collection<Long> linksFromNodeToIntermediate = linkIdsPerNodePair.get(Pair.of(nodeId, intermediateNodeId));
					if (linksFromNodeToIntermediate == null) continue;

					for (long linkId : linksFromNodeToIntermediate)
					{
						double weight_thisLink = linkWeightMap.get(linkId);
						checkIPWeight(weight_thisLink);

						if (Math.abs(weight_thisLink - (distNodeToEgress - distIntermediateToEgress)) < 1E-10)
							A_t.add(linkId);
					}
				}

				int outdegree = A_t.size();

				if (outdegree > 0)
				{
					for (long linkId : A_t)
					{
						int e = linkId2LinearIndex.get(linkId);
						f_te.setQuick(t, e, 1.0 / outdegree);
					}
				}
			}
		}

		return f_te;
	}

	/**
	 * Returns the weight associated to a given link.
	 * 
	 * @param netPlan Network design
	 * @param linkId Link identifier
	 * @return Link weight (or {@code Double.MAX_VALUE} if link is down)
	 * @since 0.3.0
	 */
	public static double getLinkWeight(NetPlan netPlan, long linkId)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getLinkWeight(netPlan, layerId, linkId);
	}

	/**
	 * Returns the weight associated to a given link.
	 * 
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link weight (or {@code Double.MAX_VALUE} if link is down)
	 * @since 0.3.0
	 */
	public static double getLinkWeight(NetPlan netPlan, long layerId, long linkId)
	{
		if (netPlan.isLinkDown(layerId, linkId)) return Double.MAX_VALUE;
		
		double linkWeight = 1;
		String str_linkWeight = netPlan.getLinkAttribute(layerId, linkId, IP_WEIGHT_ATTRIBUTE_NAME);
		try	{ linkWeight = Double.parseDouble(str_linkWeight); }
		catch(Throwable e) { }
		
		checkIPWeight(linkWeight);
		
		return linkWeight;
	}
	
	/**
	 * Obtains the set of link weights (link attribute 'linkWeight', default: 1) from a given a network design.
	 *
	 * @param netPlan Network design
	 * @return Set of link weights (must be greater than zero, default: 1)
	 * @since 0.3.0
	 */
	public static Map<Long, Double> getLinkWeightMap(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getLinkWeightMap(netPlan, layerId);
	}

	/**
	 * Obtains the set of link weights (link attribute 'linkWeight', default: 1) from a given a network design.
	 *
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @return Set of link weights (must be greater than zero, default: 1)
	 * @since 0.3.0
	 */
	public static Map<Long, Double> getLinkWeightMap(NetPlan netPlan, long layerId)
	{
		Map<Long, Double> linkWeights = new LinkedHashMap<Long, Double>();
		Set<Long> linkIds = netPlan.getLinkIds(layerId);
		for(long linkId : linkIds) linkWeights.put(linkId, getLinkWeight(netPlan, layerId, linkId));
	
		return linkWeights;
	}
	
	/**
	 * Obtains the set of link weights (link attribute 'linkWeight', default: 1) from a given a network design.
	 *
	 * @param netPlan Network design
	 * @return Set of link weights (must be greater than zero, default: 1)
	 * @since 0.3.0
	 */
	public static double[] getLinkWeightVector(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getLinkWeightVector(netPlan, layerId);
	}

	/**
	 * Obtains the set of link weights (link attribute 'linkWeight', default: 1) from a given a network design.
	 *
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @return Set of link weights (must be greater than zero, default: 1)
	 * @since 0.3.0
	 */
	public static double[] getLinkWeightVector(NetPlan netPlan, long layerId)
	{
		return DoubleUtils.toArray(getLinkWeightMap(netPlan, layerId).values());
	}
	
	/**
	 * Obtains a destination-based routing from a given network design.
	 *
	 * @param netPlan Network design
	 * @return A destination-based routing in the form of fractions <i>f<sub>te</sub></i> (fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a</i>(<i>e</i>) (the initial node of link <i>e</i>), that is forwarded through link <i>e</i>)
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D getRoutingTableMatrix(NetPlan netPlan)
	{
		Set<Long> nodeIds = netPlan.getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap();
		Map<Long, Pair<Long, Long>> demandMap = netPlan.getDemandMap();
		Map<Long, Long> d_p = netPlan.getRouteDemandMap();
		Map<Long, Double> x_p = netPlan.getRouteCarriedTrafficMap();
		Map<Long, List<Long>> seqLinks = netPlan.getRouteAllSequenceOfLinks();

		return GraphUtils.convert_xp2fte(nodeIds, linkMap, demandMap, d_p, x_p, seqLinks);
	}

	/**
	 * Outputs a given set of routing tables to a {@code String}. For debugging purposes.
	 *
	 * @param nodeIds Set of node identifiers
	 * @param linkMap Map of links, where the key is the link identifier and the value is {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the link, respectively.
	 * @param f_te Destination-based routing in the form of fractions <i>f<sub>te</sub></i> (fraction of the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a</i>(<i>e</i>) (the initial node of link <i>e</i>), that is forwarded through link <i>e</i>)
	 * @return A {@code String} from the given routing tables
	 * @since 0.3.0
	 */
	public static String routingTableMatrixToString(Set<Long> nodeIds, Map<Long, Pair<Long, Long>> linkMap, double[][] f_te)
	{
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkMap.keySet());
		
		StringBuilder out = new StringBuilder();
		int N = nodeIds.size();
		int i = 0;
		for (long nodeId : nodeIds)
		{
			out.append(String.format("Routing table for n%d%n", nodeId));

			Set<Long> outgoingLinks_thisNode = GraphUtils.getNodeOutgoingLinks(linkMap, nodeId);

			boolean thereAreRoutes = false;

			if (!outgoingLinks_thisNode.isEmpty())
			{
				for (long egressNodeId : nodeIds)
				{
					if (nodeId == egressNodeId) continue;

					int t = nodeId2LinearIndex.get(egressNodeId);
					thereAreRoutes = false;
					
					StringBuilder out_thisEgressNode = new StringBuilder();
					out_thisEgressNode.append(String.format("dst: n%d, gw:", egressNodeId));
					for (long linkId : outgoingLinks_thisNode)
					{
						int e = linkId2LinearIndex.get(linkId);
						
						double trafficValue = f_te[t][e];
						if (trafficValue < 0) continue;
						
						out_thisEgressNode.append(String.format(" [e%d (n%d): %f]", linkId, linkMap.get(linkId).getSecond(), trafficValue));
						thereAreRoutes = true;
					}

					out_thisEgressNode.append(String.format("%n"));
					if (thereAreRoutes) out.append(out_thisEgressNode);
				}
			}

			if (!thereAreRoutes) out.append("Empty");
			
			i++;
			if (i < N)
			{
				out.append(String.format("%n"));
			}
		}

		return out.toString();
	}
	
	/**
	 * Sets the OSPF/ECMP forwarding rules in the given design, according to the 
	 * given IGP weight setting. Any previous routing information (either source 
	 * routing or hop-by-hop routing) will be removed. This method calls 
	 * sequentually to {@code computeECMPForwardingRules} and {@code setForwardingRules} methods.
	 * 
	 * @param netPlan Network design
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means 'takes weight from link attribute'). No special iteration-order (i.e. ascending) is required
	 * @since 0.3.0
	 */
	public static void setECMPForwardingRulesFromLinkWeights(NetPlan netPlan, Map<Long, Double> linkWeightMap)
	{
		long layerId = netPlan.getLayerDefaultId();
		setECMPForwardingRulesFromLinkWeights(netPlan, layerId, linkWeightMap);
	}
	
	/**
	 * Sets the OSPF/ECMP forwarding rules in the given design, according to the 
	 * given IGP weight setting. Any previous routing information (either source 
	 * routing or hop-by-hop routing) will be removed. This method calls 
	 * sequentually to {@code computeECMPForwardingRules} and {@code setForwardingRules} methods.
	 * 
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means 'takes weight from link attribute'). No special iteration-order (i.e. ascending) is required
	 * @since 0.3.0
	 */
	public static void setECMPForwardingRulesFromLinkWeights(NetPlan netPlan, long layerId, Map<Long, Double> linkWeightMap)
	{
		Map<Pair<Long, Long>, Double> splittingRatioMap = computeCarriedTrafficFromIGPWeights(netPlan, layerId, linkWeightMap).getFirst();
		netPlan.setForwardingRules(layerId, splittingRatioMap);
	}
	
	/**
	 * Sets the explicit traffic routing according to an OSPF/ECMP scheme. Any 
	 * previous routing information (either source routing or hop-by-hop routing) 
	 * will be removed.
	 * 
	 * @param netPlan Network design
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means 'takes weight from link attribute'). No special iteration-order (i.e. ascending) is required
	 * @since 0.3.1
	 */
	public static void setECMPRoutesFromLinkWeights(NetPlan netPlan, Map<Long, Double> linkWeightMap)
	{
		long layerId = netPlan.getLayerDefaultId();
		setECMPRoutesFromLinkWeights(netPlan, layerId, linkWeightMap);
	}

	/**
	 * Sets the explicit traffic routing according to an OSPF/ECMP scheme. Any 
	 * previous routing information (either source routing or hop-by-hop routing) 
	 * will be removed.
	 * 
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkWeightMap Cost per link, where the key is the link identifier and the value is the link weight (null means 'takes weight from link attribute'). No special iteration-order (i.e. ascending) is required
	 * @since 0.3.1
	 */
	public static void setECMPRoutesFromLinkWeights(NetPlan netPlan, long layerId, Map<Long, Double> linkWeightMap)
	{
		if (linkWeightMap == null) linkWeightMap = IPUtils.getLinkWeightMap(netPlan, layerId);
		
		Set<Long> nodeIds = netPlan.getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap(layerId);
		Map<Long, Pair<Long, Long>> demandMap = netPlan.getDemandMap(layerId);
		Map<Long, Double> h_d = netPlan.getDemandOfferedTrafficMap(layerId);
		
		/* Compute 'f_te' */
		Map<Pair<Long, Long>, Double> f_te = IPUtils.computeECMPForwardingRules(nodeIds, linkMap, demandMap, linkWeightMap);
		
		/* Convert 'f_te' to 'x_p' */
		List<Long> d_p = new LinkedList<Long>();
		List<Double> x_p = new LinkedList<Double>();
		List<List<Long>> seqLinks_p = new LinkedList<List<Long>>();
		GraphUtils.convert_fte2xp(nodeIds, linkMap, demandMap, h_d, f_te, d_p, x_p, seqLinks_p);
		
		/* Update routing according to 'x_p' */
		netPlan.removeAllRoutingInformation(layerId);
		netPlan.setRoutingType(layerId, Constants.RoutingType.SOURCE_ROUTING);
		Iterator<Long> demands_it = d_p.iterator();
		Iterator<Double> x_p_it = x_p.iterator();
		Iterator<List<Long>> seqLinks_it = seqLinks_p.iterator();
		while(x_p_it.hasNext())
		{
			long demandId_thisPath = demands_it.next();
			List<Long> seqLinks_thisPath = seqLinks_it.next();
			double x_p_thisPath = x_p_it.next();
			netPlan.addRoute(layerId, demandId_thisPath, x_p_thisPath, seqLinks_thisPath, null);
		}
	}

	/**
	 * Sets the weight associated to the link.
	 * 
	 * @param netPlan Network design
	 * @param linkId Link identifier
	 * @param linkWeight IGP weight associated to the link (must be greater or equal than one)
	 * @since 0.3.0
	 */
	public static void setLinkWeight(NetPlan netPlan, long linkId, double linkWeight)
	{
		long layerId = netPlan.getLayerDefaultId();
		setLinkWeight(netPlan, layerId, linkId, linkWeight);
	}

	/**
	 * Sets the weight associated to the link.
	 * 
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param linkWeight IGP weight associated to the link (must be greater or equal than one)
	 * @since 0.3.0
	 */
	public static void setLinkWeight(NetPlan netPlan, long layerId, long linkId, double linkWeight)
	{
		checkIPWeight(linkWeight);
		netPlan.setLinkAttribute(layerId, linkId, IP_WEIGHT_ATTRIBUTE_NAME, Double.toString(linkWeight));
	}
	
	/**
	 * Sets the weight associated to every link.
	 *
	 * @param netPlan Network design
	 * @param linkWeight IGP weight associated to each link (must be greater or equal than one)
	 * @since 0.2.0
	 */
	public static void setLinkWeights(NetPlan netPlan, double linkWeight)
	{
		long layerId = netPlan.getLayerDefaultId();
		setLinkWeights(netPlan, layerId, linkWeight);
	}

	/**
	 * Sets the weight associated to every link.
	 *
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkWeight IGP weight associated to each link (must be greater or equal than one)
	 * @since 0.3.0
	 */
	public static void setLinkWeights(NetPlan netPlan, long layerId, double linkWeight)
	{
		Set<Long> linkIds = netPlan.getLinkIds(layerId);
		for(long linkId : linkIds)
			setLinkWeight(netPlan, layerId, linkId, linkWeight);
	}

	/**
	 * Sets the weight associated to each link.
	 *
	 * @param netPlan Network design
	 * @param linkWeightMap IGP weight per link (must be greater or equal than one)
	 * @since 0.3.0
	 */
	public static void setLinkWeights(NetPlan netPlan, Map<Long, Double> linkWeightMap)
	{
		long layerId = netPlan.getLayerDefaultId();
		setLinkWeights(netPlan, layerId, linkWeightMap);
	}

	/**
	 * Sets the weight associated to each link.
	 *
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkWeightMap IGP weight per link (must be greater or equal than one)
	 * @since 0.3.0
	 */
	public static void setLinkWeights(NetPlan netPlan, long layerId, Map<Long, Double> linkWeightMap)
	{
		for(Entry<Long, Double> entry : linkWeightMap.entrySet())
		{
			long linkId = entry.getKey();
			double linkWeight = entry.getValue();
			setLinkWeight(netPlan, layerId, linkId, linkWeight);
		}		
	}
	
	/**
	 * Sets the weight associated to each link.
	 *
	 * @param netPlan Network design
	 * @param linkWeightVector IGP weight per link (must be greater or equal than one)
	 * @since 0.2.0
	 */
	public static void setLinkWeights(NetPlan netPlan, double[] linkWeightVector)
	{
		long layerId = netPlan.getLayerDefaultId();
		setLinkWeights(netPlan, layerId, linkWeightVector);
	}

	/**
	 * Sets the weight associated to each link.
	 *
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @param linkWeightVector IGP weight per link (must be greater or equal than one)
	 * @since 0.3.0
	 */
	public static void setLinkWeights(NetPlan netPlan, long layerId, double[] linkWeightVector)
	{
		Set<Long> linkIds = netPlan.getLinkIds(layerId);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		for(long linkId : linkIds)
		{
			int e = linkId2LinearIndex.get(linkId);
			setLinkWeight(netPlan, layerId, linkId, linkWeightVector[e]);
		}
	}
}