/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.IntUtils;
import com.net2plan.utils.LongUtils;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * <p>Class providing static methods to compute several performance
 * metrics using classical formulae (i.e. Erlang-B or Kaufman-Roberts recursion
 * for call blocking probability...).</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 */
public class NetworkPerformanceMetrics
{
	private NetworkPerformanceMetrics() { }
	
	/**
	 * <p>Checks the current network state and, maybe, returns some warnings about 
	 * link oversubscription, demand blocking, and so on.</p>
	 * 
	 * <p><b>Important</b>: {@code NetPlan} objects are always valid and consistent from a technology-agnostic point of view</p>
	 *
	 * @param netPlan Input network design
	 * @param net2planParameters A key-value map with {@code Net2Plan}-wide configuration options
	 * @return List of warnings (empty if none)
	 * @since 0.3.0
	 */
	public static List<String> checkNetworkState(NetPlan netPlan, Map<String, String> net2planParameters)
	{
		double PRECISIONFACTOR = Double.parseDouble(net2planParameters.get("precisionFactor"));

		List<String> warnings = new LinkedList<String>();

		if (!netPlan.hasNodes()) warnings.add("Node set is not defined");

		boolean isMultiLayer = netPlan.isMultiLayer();
		Set<Long> layerIds_thisNetPlan = netPlan.getLayerIds();
		for (long layerId : layerIds_thisNetPlan)
		{
			String layerName = netPlan.getLayerName(layerId);
			String thisLayerMsg = "";
			if (isMultiLayer)
			{
				thisLayerMsg = String.format(" at layer %d", layerId);
				if (!layerName.isEmpty()) thisLayerMsg += " (" + layerName + ")";
			}
			
			if (!netPlan.hasLinks(layerId))
			{
				warnings.add("Link set is not defined" + thisLayerMsg);
			}
			else
			{
				double max_u_e = netPlan.getLinkMaximumCapacity(layerId);
				if (max_u_e < PRECISIONFACTOR) warnings.add("All the links have zero capacity" + thisLayerMsg);

				Set<Long> linkIds_thisLayer = netPlan.getLinkIds(layerId);
				Map<Long, Double> rho_e = netPlan.getLinkUtilizationMap(layerId);
				boolean someLinksOvercongested = false;
				boolean someLinksCongested = false;
				int numCoupledLinks = 0;
				for (long linkId : linkIds_thisLayer)
				{
					if (netPlan.isLinkCoupled(layerId, linkId)) numCoupledLinks++;
					
					double rho_e_thisLink = rho_e.get(linkId);
					if (DoubleUtils.isEqualWithinRelativeTolerance(rho_e_thisLink, 1, PRECISIONFACTOR))
					{
						someLinksCongested = true;
					}
					else if (rho_e_thisLink > 1)
					{
						someLinksOvercongested = true;
					}
				}

				if (someLinksOvercongested) warnings.add("Some links are overcongested (carried traffic > link capacity)" + thisLayerMsg);
				else if (someLinksCongested) warnings.add("Some links have 100% utilization" + thisLayerMsg);
				
				if (numCoupledLinks > 0 && numCoupledLinks < linkIds_thisLayer.size()) warnings.add("Not all links are coupled" + thisLayerMsg);
			}
			
			Set<Long> demandIds_thisLayer = netPlan.getDemandIds(layerId);
			if (!netPlan.hasDemands(layerId))
			{
				warnings.add("Traffic demand set is not defined" + thisLayerMsg);
			}
			else
			{
				int numCoupledDemands = 0;
				for (long demandId : demandIds_thisLayer)
					if (netPlan.isDemandCoupled(layerId, demandId))
						numCoupledDemands++;

				if (!netPlan.getDemandsBlocked(layerId).isEmpty()) warnings.add("Traffic losses: Not all the traffic is being carried" + thisLayerMsg);

				if (numCoupledDemands > 0 && numCoupledDemands < demandIds_thisLayer.size()) warnings.add("Not all demands are coupled" + thisLayerMsg);
			}

			if (netPlan.getRoutingType(layerId) == Constants.RoutingType.SOURCE_ROUTING)
			{
				if (!netPlan.hasRoutes(layerId))
				{
					warnings.add("Routing is not defined" + thisLayerMsg);
				}
				else
				{
					Set<Long> routeIds_thisLayer = netPlan.getRouteIds(layerId);
					boolean routingCycles = false;
					for (long routeId : routeIds_thisLayer)
					{
						List<Long> sequenceOfNodes = netPlan.getRouteSequenceOfNodes(layerId, routeId);
						Set<Long> uniqueNodeIds = LongUtils.unique(sequenceOfNodes);

						if (uniqueNodeIds.size() != sequenceOfNodes.size())
						{
							routingCycles = true;
							break;
						}
					}

					if (routingCycles) warnings.add("Some routes have cycles" + thisLayerMsg);
				}
			}
			else
			{
				if (!netPlan.hasForwardingRules(layerId))
				{
					warnings.add("Forwarding rules not defined" + thisLayerMsg);
				}
				else
				{
					boolean hasClosedCycles = false;
					boolean hasOpenCycles = false;
					for (long demandId : demandIds_thisLayer)
					{
						Constants.RoutingCycleType routingCycleType_thisDemand = netPlan.getDemandRoutingCycleType(layerId, demandId);
						if (routingCycleType_thisDemand == Constants.RoutingCycleType.CLOSED_CYCLES)
						{
							hasClosedCycles = true;
							break;
						}
						else if (routingCycleType_thisDemand == Constants.RoutingCycleType.OPEN_CYCLES)
						{
							hasOpenCycles = true;
						}
					}
					
					if (hasClosedCycles) warnings.add("Forwarding rules not well-defined (traffic enters into a closed cycle) " + thisLayerMsg);
					else if (hasOpenCycles) warnings.add("Forwarding rules non-strictly well-defined (some node is visited more than once)" + thisLayerMsg);
				}
			}
		}
		
		return warnings;
	}

	/**
	 * <p>Returns the probability of call blocking in a {@code M/M/n/n} queue 
	 * system.</p>
	 * 
	 * <p>Reference implementation: {@code [1] S. Qiao, L. Qiao, "A Robust and Efficient Algorithm for Evaluating Erlang B Formula", Technical Report CAS98-03, McMaster University (Canada), October 1998}</p>
	 *
	 * @param numberOfServers Number of servers (i.e. link capacity in integer units). It must be greater or equal than zero
	 * @param load Traffic load (i.e. carried traffic by the link). It must be greater or equal than zero
	 * @return Call blocking probability
	 * @since 0.2.0
	 */
	public static double erlangBLossProbability(int numberOfServers, double load)
	{
		if (numberOfServers < 0) throw new Net2PlanException("Number of server must be greater or equal than 0");
		if (load < 0) throw new Net2PlanException("System load must be greater or equal than 0");
		
		/* Initialize output variable */
		double gradeOfService = 0;
		if (load > 1.0E-10)
		{
			double s = 0;
			for (int i = 1; i <= numberOfServers; i++) s = (1 + s) * (i / load);

			gradeOfService = 1 / (1 + s);
		}
		return gradeOfService;
	}

	/**
	 * <p>Returns the number of servers (i.e. link capacity) to achieve a given grade 
	 * of service (i.e. call blocking probability) under a given load in a 
	 * {@code M/M/n/n} queue system.</p>
	 *
	 * <p>Reference implementation: {@code [1] S. Qiao, L. Qiao, "A Robust and Efficient Algorithm for Evaluating Erlang B Formula", Technical Report CAS98-03, McMaster University (Canada), October 1998}</p>
	 *
	 * @param gradeOfService Grade of service (i.e. call blocking probability). It must be greater or equal than zero
	 * @param load Traffic load (i.e. carried traffic by the link). It must be greater or equal than zero
	 * @return Number of servers
	 * @since 0.2.0
	 */
	public static int inverseErlangB(double gradeOfService, double load)
	{
		if (gradeOfService < 0) throw new Net2PlanException("Grade of service must be greater or equal than 0");
		if (load < 0) throw new Net2PlanException("System load must be greater or equal than 0");
		
		/* Initialize output variable */
		int numberOfServers = 0;
		if (load > 1.0E-10)
		{
			int l = 0;
			int r = (int) Math.ceil(load);
			double fR = erlangBLossProbability(r, load);
			while (fR > gradeOfService)
			{
				l = r;
				r += 32;
				fR = erlangBLossProbability(r, load);
			}
			
			while (r - l > 1)
			{
				int m = (int) Math.ceil((double) (l + r) / 2);
				double fMid = erlangBLossProbability(m, load);
				if (fMid > gradeOfService) l = m;
				else r = m;
			}
			
			numberOfServers = r;
		}
		
		return numberOfServers;
	}

	/**
	 * Computes the Kaufman-Roberts recursion for a multi-rate loss model system.
	 *
	 * @param u_e Link capacity (in integer units). It must be greater or equal than zero
	 * @param h_p Traffic volume vector. Each element is referred to a connection of type {@code p} and must be greater than zero
	 * @param s_p Capacity units occupied in the link by each accepted connection of type {@code p}. Each element must be greater or equal than one
	 * @return Vector of connection blocking probability. Each element is referred to a connection of type {@code p}
	 * @since 0.2.0
	 */
	public static double[] kaufmanRobertsRecursion(int u_e, double[] h_p, int[] s_p)
	{
		if (u_e <= 0) throw new Net2PlanException("Link capacity must be greater or equal than zero");
		
		if (h_p.length != s_p.length) throw new Net2PlanException("Length of offered traffic vector and connection size vector don't match");
		
		double[] A_k = h_p;
		if (DoubleUtils.minValue(A_k) < 0) throw new Net2PlanException("Offered traffic must be greater or equal than zero");

		int[] b_k = s_p;
		int[] b_k_maxMinValues = IntUtils.maxMinValues(b_k);
		if (b_k_maxMinValues[1] < 1) throw new Net2PlanException("Connection size must be greater or equal than one");

		int K = h_p.length;
		int offset = b_k_maxMinValues[0];
		double[] g = new double[1 + u_e + offset];
		Arrays.fill(g, 0);
		g[offset - 1] = 1;
		for (int c = 1; c <= u_e; c++)
		{
			for (int k = 1; k <= K; k++)
				g[offset + c - 1] += A_k[k - 1] * g[offset + c - b_k[k - 1] - 1];
			
			g[offset + c - 1] = g[offset + c - 1] / c;
		}
		double sumG = DoubleUtils.sum(g);
		
		double[] pb_k = new double[K];
		Arrays.fill(pb_k, 0);
		for (int k = 1; k <= K; k++)
			for (int c = u_e - b_k[k - 1] + 1; c <= u_e; c++)
				pb_k[k - 1] += (1 / sumG) * g[offset + c - 1];
		
		return pb_k;
	}
}
