/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import cern.colt.matrix.tdouble.DoubleFactory2D;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.LongUtils;
import com.net2plan.utils.Pair;
import java.util.*;
import java.util.Map.Entry;

/**
 * Provides a set of static methods which can be useful when dealing with network resilience.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.3
 */
public class SRGUtils
{
	private SRGUtils() { }
	
	/**
	 * Type of shared-risk model.
	 *
	 * @since 0.2.3
	 */
	public enum SharedRiskModel
	{
		/**
		 * Defines a SRG per each node. Links are not associated to any SRG.
		 */
		PER_NODE,
		
		/**
		 * Defines a SRG per each unidirectional link. Nodes are not associated to any SRG.
		 */
		PER_LINK,
		
		/**
		 * Defines a SRG (one per direction) containing all links between a node pair. Nodes are not associated to any SRG.
		 */
		PER_DIRECTIONAL_LINK_BUNDLE,
		
		/**
		 * Defines a SRG containing all links between a node pair. Nodes are not associated to any SRG.
		 */
		PER_BIDIRECTIONAL_LINK_BUNDLE
	};
	
	/**
	 * <p>Given a network design with a set of shared-risk groups (SRGs) and a set 
	 * of failure states, each of them containing the associated set of simultaneous 
	 * SRG failures, the failure state-link unavailability matrix. This is a matrix 
	 * with as many rows as failure states, and as many columns as links. Position 
	 * (<i>f</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers to the 
	 * first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second 
	 * one, and so on) is affected by failure state <i>f</i> (<i>f = 0</i> refers 
	 * failure state, <i>f = 1</i> refers to the second one, and so on), considering 
	 * either the link itself or some of the end nodes.</p>
	 * 
	 * <p>This method is an extended version of {@link #computeSRG2LinkUnavailabilityMatrix(com.net2plan.interfaces.networkDesign.NetPlan) computeSRG2LinkUnavailabilityMatrix}, 
	 * since it allows to consider not only single SRG failures, but also simultaneous 
	 * SRG failures.</p>
	 * 
	 * @param netPlan Network design containing a set of shared-risk groups
	 * @param F_s List of failure states, each of them containing the associated set of simultaneous SRG failures
	 * @return Failure state-link unavailability matrix
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D computeFailureState2LinkUnavailabilityMatrix(NetPlan netPlan, List<Set<Long>> F_s)
	{
		long layerId = netPlan.getLayerDefaultId();
		return computeFailureState2LinkUnavailabilityMatrix(netPlan, layerId, F_s);
	}
	
	/**
	 * <p>Given a network design with a set of shared-risk groups (SRGs) and a set 
	 * of failure states, each of them containing the associated set of simultaneous 
	 * SRG failures, the failure state-link unavailability matrix. This is a matrix 
	 * with as many rows as failure states, and as many columns as links. Position 
	 * (<i>f</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers to the 
	 * first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second 
	 * one, and so on) is affected by failure state <i>f</i> (<i>f = 0</i> refers 
	 * failure state, <i>f = 1</i> refers to the second one, and so on), considering 
	 * either the link itself or some of the end nodes.</p>
	 * 
	 * <p>This method is an extended version of {@link #computeSRG2LinkUnavailabilityMatrix(com.net2plan.interfaces.networkDesign.NetPlan, long) computeSRG2LinkUnavailabilityMatrix}, 
	 * since it allows to consider not only single SRG failures, but also simultaneous 
	 * SRG failures.</p>
	 * 
	 * @param netPlan Network design containing a set of shared-risk groups
	 * @param layerId Layer identifier
	 * @param F_s List of failure states, each of them containing the associated set of simultaneous SRG failures
	 * @return Failure state-link unavailability matrix
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D computeFailureState2LinkUnavailabilityMatrix(NetPlan netPlan, long layerId, List<Set<Long>> F_s)
	{
		int F = F_s.size();
		if (F == 0) throw new Net2PlanException("No failure state");
		
		int E = netPlan.getNumberOfLinks(layerId);
		if (E == 0) throw new Net2PlanException("No links");
		
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(netPlan.getLinkIds(layerId));
		DoubleMatrix2D A_fe = DoubleFactory2D.sparse.make(F, E);
		
		int f = 0;
		for(Set<Long> srgIds_thisFailureState : F_s)
		{
			Set<Long> linkIds_thisFailureState = new LinkedHashSet<Long>();
			for(long srgId : srgIds_thisFailureState)
			{
				Set<Long> nodeIds_thisSRG = netPlan.getSRGNodes(srgId);
				for(long nodeId : nodeIds_thisSRG)
					linkIds_thisFailureState.addAll(netPlan.getNodeAssociatedLinks(layerId, nodeId));
				
				Set<Long> linkIds_thisSRG = netPlan.getSRGLinks(layerId, srgId);
				linkIds_thisFailureState.addAll(linkIds_thisSRG);
			}
			
			for(long linkId : linkIds_thisFailureState)
			{
				int e = linkId2LinearIndex.get(linkId);
				A_fe.setQuick(f, e, 1);
			}
			
			f++;
		}
		
		return A_fe;
	}
	
	/**
	 * <p>Given a network design with a set of shared-risk groups (SRGs), it computes 
	 * the SRG-link unavailability matrix. This is a matrix with as many rows as SRGs, 
	 * and as many columns as links. Position (<i>s</i>, <i>e</i>) has a 1 if link 
	 * <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, 
	 * <i>e = 1</i> refers to the second one, and so on) is affected by the failure 
	 * of SRG <i>s</i> (<i>s = 0</i> refers to the first defined SRG, <i>s = 1</i> 
	 * refers to the second one, and so on), considering either the link itself or 
	 * some of the end nodes.</p>
	 * 
	 * @param netPlan Network design containing a set of shared-risk groups
	 * @return SRG-link unavailability matrix
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D computeSRG2LinkUnavailabilityMatrix(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return computeSRG2LinkUnavailabilityMatrix(netPlan, layerId);
	}
	
	/**
	 * <p>Given a network design with a set of shared-risk groups (SRGs), it computes 
	 * the SRG-link unavailability matrix. This is a matrix with as many rows as SRGs, 
	 * and as many columns as links. Position (<i>s</i>, <i>e</i>) has a 1 if link 
	 * <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, 
	 * <i>e = 1</i> refers to the second one, and so on) is affected by the failure 
	 * of SRG <i>s</i> (<i>s = 0</i> refers to the first defined SRG, <i>s = 1</i> 
	 * refers to the second one, and so on), considering either the link itself or 
	 * some of the end nodes.</p>
	 * 
	 * @param netPlan Network design containing a set of shared-risk groups
	 * @param layerId Layer identifier
	 * @return SRG-link unavailability matrix
	 * @since 0.3.1
	 */
	public static DoubleMatrix2D computeSRG2LinkUnavailabilityMatrix(NetPlan netPlan, long layerId)
	{
		List<Set<Long>> F_s = new LinkedList<Set<Long>>();
		Set<Long> srgIds = netPlan.getSRGIds();
		for(long srgId : srgIds) F_s.add(LongUtils.setOf(srgId));
		
		return computeFailureState2LinkUnavailabilityMatrix(netPlan, layerId, F_s);
	}
	
	/**
	 * Computes the probability to find the network on each failure state.
	 * 
	 * @param F_s Set of SRGs failing per each failure state
	 * @param A_f Availability value per SRG
	 * @return Probability to find the network in each failure state
	 * @since 0.3.0
	 */
	public static List<Double> computeStateProbabilities(List<Set<Long>> F_s, Map<Long, Double> A_f)
	{
		Set<Long> srgIds = A_f.keySet();
		for(Entry<Long, Double> entry : A_f.entrySet())
		{
			long srgId = entry.getKey();
			double availability = entry.getValue();

			if (availability < 0 || availability > 1) throw new RuntimeException("Availability for SRG " + srgId + " must be in range [0, 1] (current " + availability + ")");
		}

		List<Double> pi_s = new LinkedList<Double>();
		ListIterator<Set<Long>> it = F_s.listIterator();
		while (it.hasNext())
		{
			Set<Long> srgIds_thisState = it.next();

			double pi_s_thisState = 1;
			for (long srgId : srgIds) pi_s_thisState *= srgIds_thisState.contains(srgId) ? (1 - A_f.get(srgId)) : A_f.get(srgId);
			
			pi_s.add(pi_s_thisState);
		}

		double aux = DoubleUtils.sum(pi_s);
		if (aux > 1) throw new RuntimeException("Bad - Summation of state probabilities is greater than one (repeated states?) " + aux);

		return pi_s;
	}
	
	/**
	 * Configures the SRGs into the network design at the first layer. Existing SRGs will be removed.
	 *
	 * @param netPlan A network design containing a physical topology
	 * @param defaultMTTF Default value for mean time to fail (in hours). Zero or negative value means {@code Double.MAX_VALUE}
	 * @param defaultMTTR Default value for mean time to repair (in hours). Zero or negative value are not allowed
	 * @param sharedRiskModel Model defining SRGs
	 * @param removeExistingSRGs Indicates whether or not existing SRGs should be removed
	 * @since 0.2.3
	 */
	public static void configureSRGs(NetPlan netPlan, double defaultMTTF, double defaultMTTR, SharedRiskModel sharedRiskModel, boolean removeExistingSRGs)
	{
		long layerId = netPlan.getLayerDefaultId();
		configureSRGs(netPlan, layerId, defaultMTTF, defaultMTTR, sharedRiskModel, removeExistingSRGs);
	}

	/**
	 * Configures the SRGs into the network design at the given layer. Existing SRGs will be removed.
	 *
	 * @param netPlan A network design containing a physical topology
	 * @param layerId Layer identifier
	 * @param defaultMTTF Default value for mean time to fail (in hours). Zero or negative value means {@code Double.MAX_VALUE}
	 * @param defaultMTTR Default value for mean time to repair (in hours). Zero or negative value are not allowed
	 * @param sharedRiskModel Model defining SRGs
	 * @param removeExistingSRGs Indicates whether or not existing SRGs should be removed
	 * @since 0.3.0
	 */
	public static void configureSRGs(NetPlan netPlan, long layerId, double defaultMTTF, double defaultMTTR, SharedRiskModel sharedRiskModel, boolean removeExistingSRGs)
	{
		long[] nodeIdsVector = netPlan.getNodeIdsVector();
		Set<Long> linkIds = netPlan.getLinkIds(layerId);

		if (defaultMTTF <= 0) defaultMTTF = Double.MAX_VALUE;
		if (defaultMTTR <= 0) throw new Net2PlanException("'defaultMTTR' must be greater than zero");

		if (removeExistingSRGs) netPlan.removeAllSRGs();
		
		switch (sharedRiskModel)
		{
			case PER_BIDIRECTIONAL_LINK_BUNDLE:

				for (int n1 = 0; n1 < nodeIdsVector.length; n1++)
				{
					for (int n2 = n1 + 1; n2 < nodeIdsVector.length; n2++)
					{
						Set<Long> nodePairLinks = netPlan.getNodePairBidirectionalLinks(layerId, nodeIdsVector[n1], nodeIdsVector[n2]);
						if (nodePairLinks.isEmpty()) continue;

						long srgId = netPlan.addSRG(defaultMTTF, defaultMTTR, null);
						for(long linkId : nodePairLinks)
							netPlan.addLinkToSRG(layerId, linkId, srgId);
					}
				}

				break;

			case PER_DIRECTIONAL_LINK_BUNDLE:

				for (long originNodeId : nodeIdsVector)
				{
					for (long destinationNodeId : nodeIdsVector)
					{
						if (originNodeId == destinationNodeId) continue;
						
						Set<Long> nodePairLinks = netPlan.getNodePairLinks(layerId, originNodeId, destinationNodeId);
						if (nodePairLinks.isEmpty())
						{
							continue;
						}

						long srgId = netPlan.addSRG(defaultMTTF, defaultMTTR, null);
						for(long linkId : nodePairLinks)
							netPlan.addLinkToSRG(layerId, linkId, srgId);
					}
				}

				break;

			case PER_LINK:

				for (long linkId : linkIds)
				{
					long srgId = netPlan.addSRG(defaultMTTF, defaultMTTR, null);
					netPlan.addLinkToSRG(layerId, linkId, srgId);
				}

				break;

			case PER_NODE:

				for (long nodeId : nodeIdsVector)
				{
					long srgId = netPlan.addSRG(defaultMTTF, defaultMTTR, null);
					netPlan.addNodeToSRG(nodeId, srgId);
				}

				break;

			default:
				throw new RuntimeException("Bad - Invalid failure model type");
		}
	}
	
	/**
	 * Returns the set of SRGs going down on each failure state.
	 * 
	 * @param srgIds SRG identifiers
	 * @param considerNoFailureState Flag to indicate whether or not no failure state is included
	 * @param considerDoubleFailureStates Flag to indicate whether or not double failure states are included
	 * @return Set of SRGs going down on each failure state
	 * @since 0.2.3
	 */
	public static List<Set<Long>> enumerateFailureStates(Set<Long> srgIds, boolean considerNoFailureState, boolean considerDoubleFailureStates)
	{
		List<Set<Long>> F_s = new LinkedList<Set<Long>>();

		/* No failure */
		if (considerNoFailureState) F_s.add(new TreeSet<Long>());

		/* Single failure */
		for (long srgId : srgIds)
		{
			Set<Long> aux = new TreeSet<Long>();
			aux.add(srgId);
			F_s.add(aux);
		}

		/* Double failures */
		if (considerDoubleFailureStates)
		{
			for (long srgId_1 : srgIds)
			{
				for (long srgId_2 : srgIds)
				{
					if (srgId_2 <= srgId_1) continue;
					
					Set<Long> aux = new TreeSet<Long>();
					aux.add(srgId_1);
					aux.add(srgId_2);
					F_s.add(aux);
				}
			}
		}

		return F_s;
	}

	/**
	 * Indicates whether SRG definition follows one of the predefined models (per 
	 * node, per link...), or 'Mixed' otherwise (or 'None' if no SRGs are defined).
	 * 
	 * @param netPlan Current network design
	 * @return Description of the current SRG model in the default layer of the {@code netPlan} object
	 * @since 0.2.3
	 */
	public static String getSRGModel(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getSRGModel(netPlan, layerId);
	}

	/**
	 * Indicates whether SRG definition follows one of the predefined models (per 
	 * node, per link...), or 'Mixed' otherwise (or 'None' if no SRGs are defined).
	 * 
	 * @param netPlan Current network design
	 * @param layerId Layer identifier
	 * @return Description of the current SRG model
	 * @since 0.3.0
	 */
	public static String getSRGModel(NetPlan netPlan, long layerId)
	{
		netPlan.checkLayerIsActive(layerId);
		
		boolean isAnOneSRGPerNodeModel = true;

		int N = netPlan.getNumberOfNodes();
		int E = netPlan.getNumberOfLinks(layerId);
		int numSRGs = netPlan.getNumberOfSRGs();
		Set<Long> srgIds_thisNetPlan = netPlan.getSRGIds();
		Set<Long> nodesInSRGs = new LinkedHashSet<Long>();
		for (long srgId : srgIds_thisNetPlan)
		{
			Set<Long> nodeIds_thisSRG = netPlan.getSRGNodes(srgId);
			Set<Long> linkIds_thisSRG = netPlan.getSRGLinks(layerId, srgId);

			if (!linkIds_thisSRG.isEmpty())
			{
				isAnOneSRGPerNodeModel = false;
				break;
			}

			switch (nodeIds_thisSRG.size())
			{
				case 0:
					continue;

				case 1:
					long nodeId = nodeIds_thisSRG.iterator().next();
					if (nodesInSRGs.contains(nodeId)) isAnOneSRGPerNodeModel = false;
					else nodesInSRGs.add(nodeId);

					break;

				default:
					isAnOneSRGPerNodeModel = false;
					break;
			}

			if (!isAnOneSRGPerNodeModel) break;
		}

		if (N == 0 || nodesInSRGs.size() != N) isAnOneSRGPerNodeModel = false;
		nodesInSRGs.clear();

		boolean isAnOneSRGPerLinkModel = true;

		Set<Long> linksInSRG = new LinkedHashSet<Long>();
		for (long srgId : srgIds_thisNetPlan)
		{
			Set<Long> nodeIds_thisSRG = netPlan.getSRGNodes(srgId);
			Set<Long> linkIds_thisSRG = netPlan.getSRGLinks(layerId, srgId);

			if (!nodeIds_thisSRG.isEmpty())
			{
				isAnOneSRGPerLinkModel = false;
				break;
			}

			switch (linkIds_thisSRG.size())
			{
				case 0:
					continue;

				case 1:
					long linkId = linkIds_thisSRG.iterator().next();
					if (linksInSRG.contains(linkId)) isAnOneSRGPerLinkModel = false;
					else linksInSRG.add(linkId);

					break;

				default:
					isAnOneSRGPerLinkModel = false;
					break;
			}

			if (!isAnOneSRGPerLinkModel) break;
		}

		if (E == 0 || linksInSRG.size() != E) isAnOneSRGPerLinkModel = false;
		linksInSRG.clear();

		boolean isAnOneSRGPerLinkBundleModel = true;

		for (long srgId : srgIds_thisNetPlan)
		{
			Set<Long> nodeIds_thisSRG = netPlan.getSRGNodes(srgId);
			Set<Long> linkIds_thisSRG = netPlan.getSRGLinks(layerId, srgId);

			if (!nodeIds_thisSRG.isEmpty())
			{
				isAnOneSRGPerLinkBundleModel = false;
				break;
			}

			if (linkIds_thisSRG.isEmpty()) continue;
			
			long firstLinkId = linkIds_thisSRG.iterator().next();
			long originNodeId = netPlan.getLinkOriginNode(layerId, firstLinkId);
			long destinationNodeId = netPlan.getLinkDestinationNode(layerId, firstLinkId);
			Set<Long> linkIds_thisNodePair = netPlan.getNodePairLinks(layerId, originNodeId, destinationNodeId);
			
			for (long linkId : linkIds_thisNodePair)
			{
				if (linksInSRG.contains(linkId) || !linkIds_thisNodePair.contains(linkId)) isAnOneSRGPerLinkBundleModel = false;
				if (!isAnOneSRGPerLinkBundleModel) break;

				linksInSRG.add(linkId);
			}

			if (!isAnOneSRGPerLinkBundleModel) break;
		}

		if (E == 0 || linksInSRG.size() != E) isAnOneSRGPerLinkBundleModel = false;
		linksInSRG.clear();

		boolean isAnOneSRGPerBidiLinkBundleModel = true;

		for (long srgId : srgIds_thisNetPlan)
		{
			Set<Long> nodeIds_thisSRG = netPlan.getSRGNodes(srgId);
			Set<Long> linkIds_thisSRG = netPlan.getSRGLinks(layerId, srgId);

			if (!nodeIds_thisSRG.isEmpty())
			{
				isAnOneSRGPerBidiLinkBundleModel = false;
				break;
			}

			if (linkIds_thisSRG.isEmpty()) continue;
			
			long firstLinkId = linkIds_thisSRG.iterator().next();
			long originNodeId = netPlan.getLinkOriginNode(layerId, firstLinkId);
			long destinationNodeId = netPlan.getLinkDestinationNode(layerId, firstLinkId);
			Set<Long> linkIds_thisNodePair = netPlan.getNodePairBidirectionalLinks(layerId, originNodeId, destinationNodeId);
			
			for (long linkId : linkIds_thisNodePair)
			{
				if (linksInSRG.contains(linkId) || !linkIds_thisNodePair.contains(linkId)) isAnOneSRGPerBidiLinkBundleModel = false;
				if (!isAnOneSRGPerBidiLinkBundleModel) break;

				linksInSRG.add(linkId);
			}

			if (!isAnOneSRGPerBidiLinkBundleModel) break;
		}

		if (E == 0 || linksInSRG.size() != E) isAnOneSRGPerBidiLinkBundleModel = false;
		linksInSRG.clear();
		
		String srgModel;
		if (isAnOneSRGPerNodeModel && (!isAnOneSRGPerLinkModel && !isAnOneSRGPerLinkBundleModel && !isAnOneSRGPerBidiLinkBundleModel))
			srgModel = "One SRG per node";
		else if (isAnOneSRGPerBidiLinkBundleModel && !isAnOneSRGPerNodeModel)
			srgModel = "One SRG per bidirectional link bundle";
		else if (isAnOneSRGPerLinkBundleModel && !isAnOneSRGPerNodeModel && !isAnOneSRGPerBidiLinkBundleModel)
			srgModel = "One SRG per unidirectional link bundle";
		else if (isAnOneSRGPerLinkModel && !isAnOneSRGPerNodeModel && !isAnOneSRGPerLinkBundleModel && !isAnOneSRGPerBidiLinkBundleModel)
			srgModel = "One SRG per unidirectional link";
		else
			srgModel = numSRGs > 0 ? "Mixed" : "None";

		return srgModel;
	}
	
	/**
	 * Returns the percentage of SRG disjointness of traffic routes and
	 * protection segments.
	 * 
	 * @param netPlan Current network design
	 * @return Two SRG disjointness values (either considering or not end nodes) for the default layer of the {@code netPlan} object
	 * @since 0.2.3
	 */
	public static Pair<Double, Double> getSRGDisjointnessPercentage(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getSRGDisjointnessPercentage(netPlan, layerId);
	}

	/**
	 * <p>Returns the percentage of SRG disjointness of traffic routes and
	 * protection segments.</p>
	 * 
	 * <p><b>Important</b>: In case there are no SRGs, then it will return zero.</p>
	 * 
	 * @param netPlan Current network design
	 * @param layerId Layer identifier
	 * @return Two SRG disjointness values (either considering or not end nodes)
	 * @since 0.2.3
	 */
	public static Pair<Double, Double> getSRGDisjointnessPercentage(NetPlan netPlan, long layerId)
	{
		netPlan.checkLayerIsActive(layerId);
		if (!netPlan.hasSRGs()) return Pair.of(0.0, 0.0);
		
		int accum_routeSRGDisjoint_withEndNodes = 0;
		int accum_routeSRGDisjoint_withoutEndNodes = 0;
		
		int R = netPlan.getNumberOfRoutes(layerId);
		Set<Long> routeIds = netPlan.getRouteIds(layerId);
		for (long routeId : routeIds)
		{
			List<Long> segmentIds = netPlan.getRouteBackupSegmentList(layerId, routeId);
			if (segmentIds.isEmpty()) continue;

			boolean srgDisjoint_withEndNodes = true;
			boolean srgDisjoint_withoutEndNodes = true;

			List<Long> seqLinks = netPlan.convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, netPlan.getRouteSequenceOfLinks(layerId, routeId));
			List<Long> seqNodes = netPlan.getRouteSequenceOfNodes(layerId, routeId);

			Set<Long> routeSRGIds = netPlan.getRouteSRGs(layerId, routeId);
			Set<Long> segmentSRGIds = new LinkedHashSet<Long>();
			for(long segmentId : segmentIds) segmentSRGIds.addAll(netPlan.getProtectionSegmentSRGs(layerId, segmentId));
			
			Set<Long> commonSRGs_withEndNodes = CollectionUtils.intersect(routeSRGIds, segmentSRGIds);
			if (!commonSRGs_withEndNodes.isEmpty()) srgDisjoint_withEndNodes = false;

			Set<Long> routeSRGIds_withoutEndNodes = new LinkedHashSet<Long>();
			Iterator<Long> nodeIt = seqNodes.iterator(); nodeIt.next();
			while(nodeIt.hasNext())
				if (nodeIt.hasNext())
					routeSRGIds_withoutEndNodes.addAll(netPlan.getNodeSRGs(nodeIt.next()));
			
			for (long linkId : seqLinks)
				routeSRGIds_withoutEndNodes.addAll(netPlan.getLinkSRGs(layerId, linkId));
				
			Set<Long> commonSRGs_withoutEndNodes = CollectionUtils.intersect(routeSRGIds_withoutEndNodes, segmentSRGIds);
			if (!commonSRGs_withoutEndNodes.isEmpty()) srgDisjoint_withoutEndNodes = false;

			if (srgDisjoint_withEndNodes) accum_routeSRGDisjoint_withEndNodes++;
			if (srgDisjoint_withoutEndNodes) accum_routeSRGDisjoint_withoutEndNodes++;
		}

		double percentageRouteSRGDisjointness_withoutEndNodes = R > 0 ? 100 * accum_routeSRGDisjoint_withoutEndNodes / (double) R : 0;
		double percentageRouteSRGDisjointness_withEndNodes = R > 0 ? 100 * accum_routeSRGDisjoint_withEndNodes / (double) R : 0;

		return Pair.of(percentageRouteSRGDisjointness_withEndNodes, percentageRouteSRGDisjointness_withoutEndNodes);
	}
}
