/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import com.net2plan.interfaces.networkDesign.Configuration;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

/**
 * Class implementing different traffic metrics.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.3.0
 */
public class TrafficComputationEngine
{
	private TrafficComputationEngine() { }
	
	/**
	 * Returns the bottleneck (lowest) capacity among the set of traversed links and protection 
	 * segments. This method takes into account multiple passes through the same link 
	 * or protection segment.
	 * 
	 * @param netPlan Current network design
	 * @param path Sequence of links and protection segments (not checked for continuity)
	 * @return Bottleneck capacity
	 * @since 0.3.0
	 */
	public static double getPathBottleneckCapacity(NetPlan netPlan, List<Long> path)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getPathBottleneckCapacity(netPlan, layerId, path);
	}

	/**
	 * Returns the bottleneck (lowest) capacity among the set of traversed links and protection 
	 * segments. This method takes into account multiple passes through the same link 
	 * or protection segment.
	 * 
	 * @param netPlan Current network design
	 * @param layerId Layer identifier
	 * @param path Sequence of links and protection segments (not checked for continuity)
	 * @return Bottleneck capacity
	 * @since 0.3.0
	 */
	public static double getPathBottleneckCapacity(NetPlan netPlan, long layerId, List<Long> path)
	{
		if (path.isEmpty()) throw new Net2PlanException("Path is empty");
		
		double bottleneckCapacity = Double.MAX_VALUE;
		Map<Long, Integer> linkPasses = CollectionUtils.count(path);
		for(long linkId : new LinkedHashSet<Long>(linkPasses.keySet()))
		{
			if (linkId >= 0)
				bottleneckCapacity = Math.min(bottleneckCapacity, netPlan.getLinkSpareCapacity(layerId, linkId) / linkPasses.get(linkId));
			else
				bottleneckCapacity = Math.min(bottleneckCapacity, netPlan.getProtectionSegmentSpareCapacity(layerId, -1 - linkId) / linkPasses.get(linkId));
		}
		
		return bottleneckCapacity;
	}
	
	/**
	 * Obtains the average route length among the current routes according to
	 * certain link cost metric. Route lengths are weighted by their carried traffic.
	 * 
	 * @param netPlan Current network design
	 * @param linkCostMap Cost metric per link
	 * @return Average route length
	 * @since 0.3.0
	 */
	public static double getRouteAverageLength(NetPlan netPlan, Map<Long, Double> linkCostMap)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getRouteAverageLength(netPlan, layerId, linkCostMap);
	}

	/**
	 * Obtains the average route length among the current routes according to
	 * certain link cost metric. Route lengths are weighted by their carried traffic.
	 * 
	 * @param netPlan Current network design
	 * @param layerId Layer identifier
	 * @param linkCostMap Cost metric per link
	 * @return Average route length
	 * @since 0.2.2
	 */
	public static double getRouteAverageLength(NetPlan netPlan, long layerId, Map<Long, Double> linkCostMap)
	{
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		double R_d = netPlan.getDemandTotalCarriedTraffic(layerId);
		if (R_d < PRECISION_FACTOR) return 0;

		double routeAverageLength = 0;
		Set<Long> routeIds = netPlan.getRouteIds(layerId);
		for (long routeId : routeIds)
		{
			double x_p = netPlan.getRouteCarriedTraffic(layerId, routeId);
			List<Long> seqLinks = netPlan.convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, netPlan.getRouteSequenceOfLinks(layerId, routeId));
			routeAverageLength += x_p * GraphUtils.convertPath2PathCost(seqLinks, linkCostMap);
		}

		routeAverageLength /= R_d;

		return routeAverageLength;
	}
	
	/**
	 * <p>Returns the statistics for protection degree carried traffic. Returned
	 * values are the following:</p>
	 * 
	 * <ul>
	 * <li>Unprotected: Percentage of carried traffic which is not backed up by protection segments</li>
	 * <li>Complete and dedicated protection: Percentage of carried traffic which
	 * has one or more dedicated protection segments covering the whole route with enough bandwidth</li>
	 * <li>Partial and/or shared protection: For cases not covered by the above definitions</li>
	 * </ul>
	 * 
	 * @param netPlan Current network design
	 * @return Traffic protection degree
	 * @since 0.2.2
	 */
	public static Triple<Double, Double, Double> getTrafficProtectionDegree(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getTrafficProtectionDegree(netPlan, layerId);
	}
	
	/**
	 * <p>Returns the statistics for protection degree carried traffic. Returned
	 * values are the following:</p>
	 * 
	 * <ul>
	 * <li>Unprotected: Percentage of carried traffic which is not backed up by protection segments</li>
	 * <li>Complete and dedicated protection: Percentage of carried traffic which
	 * has one or more dedicated protection segments covering the whole route with enough bandwidth</li>
	 * <li>Partial and/or shared protection: For cases not covered by the above definitions</li>
	 * </ul>
	 * 
	 * @param netPlan Current network design
	 * @param layerId Layer identifier
	 * @return Traffic protection degree
	 * @since 0.3.0
	 */
	public static Triple<Double, Double, Double> getTrafficProtectionDegree(NetPlan netPlan, long layerId)
	{
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		
		double R_d = netPlan.getDemandTotalCarriedTraffic(layerId);
		double percentageUnprotected = 0;
		double percentageDedicated = 0;
		double percentageShared = 0;

		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap(layerId);
		Set<Long> dedicatedSegmentIds = netPlan.getProtectionSegmentsDedicated(layerId);
		Map<Long, Double> u_s = netPlan.getProtectionSegmentReservedCapacityMap(layerId);
		Set<Long> routeIds = netPlan.getRouteIds(layerId);
		for (long routeId : routeIds)
		{
			double x_p = netPlan.getRouteCarriedTraffic(layerId, routeId);
			List<Long> segmentIds_thisRoute = netPlan.getRouteBackupSegmentList(layerId, routeId);
			if (segmentIds_thisRoute.isEmpty())
			{
				percentageUnprotected += x_p;
				continue;
			}

			List<Long> seqLinks_thisRoute_noSegments = new LinkedList<Long>(netPlan.getRouteSequenceOfLinks(layerId, routeId));
			Iterator<Long> linkIt = seqLinks_thisRoute_noSegments.iterator();
			while(linkIt.hasNext())
			{
				long linkId = linkIt.next();
				if (linkId < 0) linkIt.remove();
			}
			
			List<Long> seqNodes_thisRoute_noSegments = new LinkedList<Long>();
			if (!seqLinks_thisRoute_noSegments.isEmpty()) seqNodes_thisRoute_noSegments.addAll(GraphUtils.convertSequenceOfLinksToSequenceOfNodes(linkMap, seqLinks_thisRoute_noSegments));

			List<Double> u_s_thisRoute = CollectionUtils.select(u_s, segmentIds_thisRoute);
			double[] maxMinValue = DoubleUtils.maxMinValues(u_s_thisRoute);
			double u_s_reservedForProtection_thisRoute_max = maxMinValue[0];
			double u_s_reservedForProtection_thisRoute_min = maxMinValue[1];

			if (u_s_reservedForProtection_thisRoute_min < PRECISION_FACTOR)
			{
				percentageUnprotected += x_p;
			}
			else
			{
				boolean isCompleteAndDedicated = false;

				if (u_s_reservedForProtection_thisRoute_max > x_p - PRECISION_FACTOR)
				{
					Set<Long> totallyProtectedLinks = new LinkedHashSet<Long>();
					for (long segmentId : segmentIds_thisRoute)
					{
						if (!dedicatedSegmentIds.contains(segmentId)) continue;
						
						if (u_s.get(segmentId) > x_p - PRECISION_FACTOR)
						{
							if (seqLinks_thisRoute_noSegments.isEmpty()) continue;
							
							List<Long> seqLinks_thisSegment = netPlan.getProtectionSegmentSequenceOfLinks(layerId, segmentId);
							if (CollectionUtils.containsAny(seqLinks_thisRoute_noSegments, seqLinks_thisSegment)) continue;
							
							List<Long> seqNodes_thisSegment = netPlan.getProtectionSegmentSequenceOfNodes(layerId, segmentId);
							long originNodeId = seqNodes_thisSegment.get(0);
							long destinationNodeId = seqNodes_thisSegment.get(seqNodes_thisSegment.size() - 1);

							int head_pos = seqNodes_thisRoute_noSegments.indexOf(originNodeId);
							int tail_pos = seqNodes_thisRoute_noSegments.lastIndexOf(destinationNodeId);
							if (head_pos == -1 || tail_pos == -1 || head_pos > tail_pos) throw new RuntimeException("Bad");
							
							int firstHopId = -1;
							int lastHopId = -1;
							
							ListIterator<Long> it = seqLinks_thisRoute_noSegments.listIterator();
							while(it.hasNext())
							{
								int hopId = it.nextIndex();
								
								if (netPlan.getLinkOriginNode(layerId, it.next()) == originNodeId)
								{
									firstHopId = hopId;
									break;
								}							
							}
							
							it = seqLinks_thisRoute_noSegments.listIterator(seqLinks_thisRoute_noSegments.size());
							while(it.hasPrevious())
							{
								int hopId = it.previousIndex();
								
								if (netPlan.getLinkDestinationNode(layerId, it.previous()) == destinationNodeId)
								{
									lastHopId = hopId;
									break;
								}							
							}
							
							if (firstHopId > lastHopId) throw new RuntimeException("Bad");
							
							for (int hopId = firstHopId; hopId <= lastHopId; hopId++)
							{
								totallyProtectedLinks.add(seqLinks_thisRoute_noSegments.get(hopId));
							}
						}
					}

					seqLinks_thisRoute_noSegments.removeAll(totallyProtectedLinks);
					if (seqLinks_thisRoute_noSegments.isEmpty()) isCompleteAndDedicated = true;
				}

				if (isCompleteAndDedicated) percentageDedicated += x_p;
				else percentageShared += x_p;
			}
		}

		percentageUnprotected = R_d == 0 ? 0 : 100 * percentageUnprotected / R_d;
		percentageDedicated = R_d == 0 ? 0 : 100 * percentageDedicated / R_d;
		percentageShared = R_d == 0 ? 0 : 100 * percentageShared / R_d;

		return Triple.of(percentageUnprotected, percentageDedicated, percentageShared);
	}
}