/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.libraries;

import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.IntUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.StringUtils;
import java.util.*;
import org.apache.commons.collections15.Transformer;

/**
 * Class to deal with optical topologies including wavelength assignment and regenerator placement.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.1
 */
public class WDMUtils
{
	/**
	 * Route/protection segment attribute name for sequence of regenerators.
	 * 
	 * @since 0.3.0
	 */
	public final static String SEQUENCE_OF_REGENERATORS_ATTRIBUTE_NAME = "seqRegenerators";
	
	/**
	 * Route/protection segment attribute name for sequence of wavelengths.
	 * 
	 * @since 0.3.0
	 */
	public final static String SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME = "seqWavelengths";
	
	private WDMUtils() { }
	
	/**
	 * <p>Wavelength assignment algorithm based on a first-fit fashion. Wavelengths
	 * are indexed from 0 to <i>W<sub>f</sub></i>-1, where <i>W<sub>f</sub></i>
	 * is the number of wavelengths supported by fiber <i>f</i>. Then, the wavelength
	 * assigned to each lightpath (along the whole physical route) is the minimum
	 * index among the common free-wavelength set for all traversed fibers.</p>
	 *
	 * <p>In case a lightpath cannot be allocated, the corresponding sequence of
	 * wavelengths ({@code seqWavelengths} parameter) will be an empty array.</p>
	 * 
	 * <p><b>Important</b>: {@code wavelengthFiberOccupancy} is not updated, so 
	 * subsequent usage of {@code allocateResources} method is encouraged.</p>
	 *
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param w_f Number of wavelengths per fiber
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @return Sequence of wavelengths traversed by each lightpath
	 * @since 0.3.0
	 */
	public static int[] WA_firstFit(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, Map<Long, Integer> w_f, List<Set<Long>> wavelengthFiberOccupancy)
	{
		GraphUtils.checkRouteContinuity(fiberMap, seqFibers, Constants.CheckRoutingCycleType.NO_REPEAT_LINK);

		int W = IntUtils.maxValue(CollectionUtils.select(w_f, seqFibers));
		for(int wavelengthId = 0; wavelengthId < W; wavelengthId++)
		{
			Set<Long> occupiedFibers_thisWavelength = wavelengthFiberOccupancy.get(wavelengthId);
			if (CollectionUtils.containsAny(occupiedFibers_thisWavelength, seqFibers)) continue;
			
			int[] seqWavelengths = new int[seqFibers.size()];
			Arrays.fill(seqWavelengths, wavelengthId);
			return seqWavelengths;
		}

		return new int[0];
	}

	/**
	 * <p>Wavelength assignment algorithm based on a first-fit fashion assuming
	 * full wavelength conversion and regeneration. Each node selects the first
	 * free wavelength for its output fiber, and next nodes in the lightpath try
	 * to maintain it. If not possible, or regeneration is needed, then include
	 * a regenerator (can act also as a full wavelength converter) and search
	 * for the first free wavelength, and so on.</p>
	 *
	 * <p>In case a lightpath cannot be allocated, the corresponding sequence of
	 * wavelengths ({@code seqWavelengths} parameter) will be an empty array.</p>
	 *
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param w_f Number of wavelengths per fiber
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @param l_f Physical length in km per fiber
	 * @param nodeRegeneratorOccupancy Number of regenerators installed per node
	 * @param maxRegeneratorDistanceInKm Maximum regeneration distance
	 * @return Sequence of wavelengths traversed by each lightpath, and a 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static Pair<int[], int[]> WA_RPP_firstFit(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, Map<Long, Integer> w_f, List<Set<Long>> wavelengthFiberOccupancy, Map<Long, Double> l_f, Map<Long, Integer> nodeRegeneratorOccupancy, double maxRegeneratorDistanceInKm)
	{
		int W = IntUtils.maxValue(w_f);

		List<Integer> seqWavelengths = new LinkedList<Integer>();
		List<Integer> seqRegenerators = new LinkedList<Integer>();

		double control_accumDistance = 0;
		Set<Integer> control_occupied_w = new LinkedHashSet<Integer>();
		int control_firstFitValidWavelengthForSubpath = -1;
		List<Long> control_currentSubpathSeqLinks = new LinkedList<Long>();

		boolean lpAllocated = true;

		Map<Long, Set<Integer>> avoidLoopWavelengthClash = new LinkedHashMap<Long, Set<Integer>>();

		for (long fiberId : seqFibers)
		{
			int numWavelengths_thisFiber = w_f.get(fiberId);
			double fiberLengthInKm = l_f.get(fiberId);

			if (fiberLengthInKm > maxRegeneratorDistanceInKm)
			{
				throw new WDMException(String.format("Fiber %d is longer (%f km) than the maximum distance without regenerators (%f km)", fiberId, fiberLengthInKm, maxRegeneratorDistanceInKm));
			}

			/* update the info as if this link was included in the subpath */
			final double plusLink_accumDistance = control_accumDistance + fiberLengthInKm;
			int plusLink_firstFitValidWavelengthForSubpath = -1;
			Set<Integer> plusLink_occupied_w = new LinkedHashSet<Integer>(control_occupied_w);
			for (int wavelengthId = numWavelengths_thisFiber; wavelengthId < W; wavelengthId++)
			{
				plusLink_occupied_w.add(wavelengthId);
			}

			if (avoidLoopWavelengthClash.containsKey(fiberId))
			{
				plusLink_occupied_w.addAll(avoidLoopWavelengthClash.get(fiberId));
			}

			for (int wavelengthId = numWavelengths_thisFiber - 1; wavelengthId >= 0; wavelengthId--)
			{
				if (!plusLink_occupied_w.contains(wavelengthId) && !wavelengthFiberOccupancy.get(wavelengthId).contains(fiberId))
				{
					plusLink_firstFitValidWavelengthForSubpath = wavelengthId;
				}
				else
				{
					plusLink_occupied_w.add(wavelengthId);
				}
			}

			if (!control_currentSubpathSeqLinks.contains(fiberId) && plusLink_accumDistance <= maxRegeneratorDistanceInKm && plusLink_firstFitValidWavelengthForSubpath != -1)
			{
				/* we do not have to put a regenerator in the origin node of e: the subpath is valid up to now */
				control_accumDistance = plusLink_accumDistance;
				control_occupied_w = plusLink_occupied_w;
				control_firstFitValidWavelengthForSubpath = plusLink_firstFitValidWavelengthForSubpath;
				control_currentSubpathSeqLinks.add(fiberId);
				seqRegenerators.add(0);
				continue;
			}

			/* Here if we have to put a regenerator in initial node of this link, add a subpath */
			if (control_firstFitValidWavelengthForSubpath == -1)
			{
				lpAllocated = false;
				break;
			}

			seqRegenerators.add(1);
			int numFibersSubPath = control_currentSubpathSeqLinks.size();
			for (int cont = 0; cont < numFibersSubPath; cont++)
			{
				seqWavelengths.add(control_firstFitValidWavelengthForSubpath);

				long aux_fiberId = control_currentSubpathSeqLinks.get(cont);
				if (!avoidLoopWavelengthClash.containsKey(aux_fiberId))
					avoidLoopWavelengthClash.put(aux_fiberId, new LinkedHashSet<Integer>());

				avoidLoopWavelengthClash.get(aux_fiberId).add(control_firstFitValidWavelengthForSubpath);
			}

			/* new span includes just this link */
			control_accumDistance = fiberLengthInKm;
			control_currentSubpathSeqLinks = new LinkedList<Long>();
			control_currentSubpathSeqLinks.add(fiberId);
			control_occupied_w = new LinkedHashSet<Integer>();
			if (avoidLoopWavelengthClash.containsKey(fiberId))
				control_occupied_w.addAll(avoidLoopWavelengthClash.get(fiberId));

			control_firstFitValidWavelengthForSubpath = -1;
			for (int wavelengthId = 0; wavelengthId < numWavelengths_thisFiber; wavelengthId++)
			{
				if (wavelengthFiberOccupancy.get(wavelengthId).contains(fiberId))
					control_occupied_w.add(wavelengthId);
				else if (control_firstFitValidWavelengthForSubpath == -1)
					control_firstFitValidWavelengthForSubpath = wavelengthId;
			}

			if (control_firstFitValidWavelengthForSubpath == -1)
			{
				lpAllocated = false;
				break;
			}
		}

		/* Add the last subpath */
		/* Here if we have to put a regenerator in initial node of this link, add a subpath */
		if (control_firstFitValidWavelengthForSubpath == -1) lpAllocated = false;

		if (!lpAllocated) return Pair.of(new int[0], new int[0]);
		
		int numFibersSubPath = control_currentSubpathSeqLinks.size();
		for (int cont = 0; cont < numFibersSubPath; cont++)
			seqWavelengths.add(control_firstFitValidWavelengthForSubpath);

		return Pair.of(IntUtils.toArray(seqWavelengths), IntUtils.toArray(seqRegenerators));
	}
	
	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a new lightpath is occupying 
	 * a wavelength in each fiber.
	 * 
	 * @param seqFibers Sequence of traversed fibers
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @since 0.3.0
	 */
	public static void allocateResources(List<Long> seqFibers, int[] seqWavelengths, List<Set<Long>> wavelengthFiberOccupancy)
	{
		allocateResources(null, seqFibers, seqWavelengths, wavelengthFiberOccupancy, null, null);
	}

	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a new lightpath is occupying 
	 * a wavelength in each fiber.
	 * 
	 * @param seqFibers Sequence of traversed fibers
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @since 0.3.0
	 */
	public static void allocateResources(List<Long> seqFibers, int wavelengthId, List<Set<Long>> wavelengthFiberOccupancy)
	{
		allocateResources(null, seqFibers, wavelengthId, wavelengthFiberOccupancy, null, null);
	}

	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a new lightpath is occupying 
	 * a wavelength in each fiber.
	 * 
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @param nodeRegeneratorOccupancy Number of regenerators installed per node
	 * @since 0.3.0
	 */
	public static void allocateResources(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, int[] seqWavelengths, List<Set<Long>> wavelengthFiberOccupancy, int[] seqRegenerators, Map<Long, Integer> nodeRegeneratorOccupancy)
	{
		if (seqWavelengths.length == 0) return;
		
		ListIterator<Long> fiberIt = seqFibers.listIterator();
		while(fiberIt.hasNext())
		{
			int hopId = fiberIt.nextIndex();
			long fiberId = fiberIt.next();
			int wavelengthId = seqWavelengths[hopId];
			wavelengthFiberOccupancy.get(wavelengthId).add(fiberId);

			if (seqRegenerators != null && seqRegenerators[hopId] == 1)
			{
				long nodeId = fiberMap.get(fiberId).getFirst();
				nodeRegeneratorOccupancy.put(nodeId, nodeRegeneratorOccupancy.get(nodeId) + 1);
			}
		}
	}

	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a new lightpath is occupying 
	 * a wavelength in each fiber.
	 * 
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @param nodeRegeneratorOccupancy Number of regenerators installed per node
	 * @since 0.3.0
	 */
	public static void allocateResources(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, int wavelengthId, List<Set<Long>> wavelengthFiberOccupancy, int[] seqRegenerators, Map<Long, Integer> nodeRegeneratorOccupancy)
	{
		ListIterator<Long> fiberIt = seqFibers.listIterator();
		while(fiberIt.hasNext())
		{
			int hopId = fiberIt.nextIndex();
			long fiberId = fiberIt.next();
			boolean wasFree = wavelengthFiberOccupancy.get(wavelengthId).add(fiberId);
			if (!wasFree) throw new WDMException("Wavelength " + wavelengthId + " was already occupied in fiber " + fiberId);

			if (seqRegenerators != null && seqRegenerators[hopId] == 1)
			{
				long nodeId = fiberMap.get(fiberId).getFirst();
				nodeRegeneratorOccupancy.put(nodeId, nodeRegeneratorOccupancy.get(nodeId) + 1);
			}
		}
	}

	/**
	 * <p>Performs extra checks to those applicable to every network design, especially
	 * focused on WDM networks.</p>
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @since 0.2.3
	 */
	public static void checkConsistency(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		checkConsistency(netPlan, layerId);
	}

	/**
	 * <p>Performs extra checks to those applicable to every network design, especially
	 * focused on WDM networks.</p>
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public static void checkConsistency(NetPlan netPlan, long layerId)
	{
		getFiberNumWavelengthsMap(netPlan, layerId);

		Set<Long> lpRouteIds = netPlan.getRouteIds(layerId);
		for (long lpRouteId : lpRouteIds)
		{
			double y_p_thisRoute = netPlan.getRouteOccupiedCapacity(layerId, lpRouteId);
			if (y_p_thisRoute != 1) throw new WDMException("Lightpath route " + lpRouteId + " must occupy only one wavelength per fiber");

			int numHops_thisLp = netPlan.getRouteNumberOfHops(layerId, lpRouteId);
			int[] seqWavelengths_thisLp = getLightpathSeqWavelengths(netPlan, layerId, lpRouteId);
			int[] seqRegenerators_thisLp = getLightpathSeqRegenerators(netPlan, layerId, lpRouteId);

			if (netPlan.isRouteDown(layerId, lpRouteId))
			{
				if (seqWavelengths_thisLp.length != 0) throw new WDMException("Failing lightpath route " + lpRouteId + " must not occupy wavelength resources");
				if (seqRegenerators_thisLp.length != 0) throw new WDMException("Failing lightpath route " + lpRouteId + " must not occupy regenerator resources");
			}
			else
			{
				if (seqWavelengths_thisLp.length != numHops_thisLp) throw new WDMException("Length of sequence of wavelengths for lightpath " + lpRouteId + " does not match the number of traversed fibers");
				if (seqRegenerators_thisLp.length > 0 && seqRegenerators_thisLp.length != numHops_thisLp) throw new WDMException("Length of sequence of regenerators/wavelength converters for lightpath " + lpRouteId + " does not match the number of traversed fibers");
			}
		}

		Set<Long> protectionLpIds = netPlan.getProtectionSegmentIds(layerId);
		for (long protectionLpId : protectionLpIds)
		{
			double u_e_thisSegment = netPlan.getProtectionSegmentReservedCapacity(layerId, protectionLpId);
			if (u_e_thisSegment != 1) throw new WDMException("Protection lightpath " + protectionLpId + " must occupy only one wavelength per fiber");

			int numHops_thisProtectionLp = netPlan.getProtectionSegmentNumberOfHops(layerId, protectionLpId);
			int[] seqWavelengths_thisProtectionLp = getProtectionLightpathSeqWavelengths(netPlan, layerId, protectionLpId);
			int[] seqRegenerators_thisProtectionLp = getProtectionLightpathSeqRegenerators(netPlan, layerId, protectionLpId);
			
			if (netPlan.isProtectionSegmentDown(layerId, protectionLpId))
			{
				if (seqWavelengths_thisProtectionLp.length != 0) throw new WDMException("Failing protection lightpath " + protectionLpId + " must not occupy wavelength resources");
				if (seqRegenerators_thisProtectionLp.length != 0) throw new WDMException("Failing protection lightpath " + protectionLpId + " must not occupy regenerator resources");
			}
			else
			{
				if (seqWavelengths_thisProtectionLp.length != numHops_thisProtectionLp) throw new WDMException("Length of sequence of wavelengths for protection lightpath " + protectionLpId + " does not match the number of traversed fibers");
				if (seqRegenerators_thisProtectionLp.length > 0 && seqRegenerators_thisProtectionLp.length != numHops_thisProtectionLp) throw new WDMException("Length of sequence of regenerators/wavelength converters for protection lightpath " + protectionLpId + " does not match the number of traversed fibers");
			}
		}

		getNodeRegeneratorOccupancyMap(netPlan, layerId);
		getWavelengthFiberOccupancy(netPlan, layerId);
	}
	
	/**
	 * Returns the list of nodes within the lightpath route containing a regenerator,
	 * only following a distance criterium, assuming no wavelength conversion is required.
	 *
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param l_f Physical length in km per fiber
	 * @param maxRegeneratorDistanceInKm Maximum regeneration distance
	 * @return A vector with as many elements as traversed links in the route/segment. Each element is a 1 if an optical regenerator is used at the origin node of the corresponding link, and a 0 if not. First element is always 0.
	 * @since 0.3.0
	 */
	public static int[] computeRegeneratorPositions(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, Map<Long, Double> l_f, double maxRegeneratorDistanceInKm)
	{
		int numHops = seqFibers.size();

		double accumDistance = 0;
		int[] seqRegenerators = new int[numHops];

		ListIterator<Long> it = seqFibers.listIterator();
		while (it.hasNext())
		{
			int hopId = it.nextIndex();
			long fiberId = it.next();
			double fiberLengthInKm = l_f.get(fiberId);

			if (fiberLengthInKm > maxRegeneratorDistanceInKm)
			{
				throw new WDMException(String.format("Fiber %d is longer (%f km) than the maximum distance without regenerators (%f km)", fiberId, fiberLengthInKm, maxRegeneratorDistanceInKm));
			}

			accumDistance += fiberLengthInKm;

			if (accumDistance > maxRegeneratorDistanceInKm)
			{
				seqRegenerators[hopId] = 1;
				accumDistance = fiberLengthInKm;
			}
			else
			{
				seqRegenerators[hopId] = 0;
			}
		}

		return seqRegenerators;
	}

	/**
	 * Returns the list of nodes within the lightpath route containing a regenerator,
	 * only following a distance criterium, assuming wavelength conversion is required.
	 *
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @param l_f Physical length in km per fiber
	 * @param maxRegeneratorDistanceInKm Maximum regeneration distance
	 * @return A vector with as many elements as traversed links in the route/segment. Each element is a 1 if an optical regenerator is used at the origin node of the corresponding link, and a 0 if not. First element is always 0.
	 * @since 0.3.0
	 */
	public static int[] computeRegeneratorPositions(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, int[] seqWavelengths, Map<Long, Double> l_f, double maxRegeneratorDistanceInKm)
	{
		int numHops = seqFibers.size();

		double accumDistance = 0;
		int[] seqRegenerators = new int[numHops];

		ListIterator<Long> it = seqFibers.listIterator();
		while (it.hasNext())
		{
			int hopId = it.nextIndex();
			long fiberId = it.next();
			double fiberLengthInKm = l_f.get(fiberId);

			if (fiberLengthInKm > maxRegeneratorDistanceInKm)
			{
				throw new WDMException(String.format("Fiber %d is longer (%f km) than the maximum distance without regenerators (%f km)", fiberId, fiberLengthInKm, maxRegeneratorDistanceInKm));
			}

			accumDistance += fiberLengthInKm;

			if (accumDistance > maxRegeneratorDistanceInKm || (hopId > 0 && seqWavelengths[hopId - 1] != seqWavelengths[hopId]))
			{
				seqRegenerators[hopId] = 1;
				accumDistance = fiberLengthInKm;
			}
			else
			{
				seqRegenerators[hopId] = 0;
			}
		}

		return seqRegenerators;
	}

	/**
	 * Returns the number of wavelengths for the given fiber. It is equivalent to 
	 * the method {@link com.net2plan.interfaces.networkDesign.NetPlan#getLinkCapacity(long) getLinkCapacity()} 
	 * from the {@link com.net2plan.interfaces.networkDesign.NetPlan NetPlan} object, 
	 * but converting capacity value from {@code double} to {@code int}.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param fiberId Fiber identifier
	 * @return Number of wavelengths
	 * @since 0.3.0
	 */
	public static int getFiberNumWavelengths(NetPlan netPlan, long fiberId)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getFiberNumWavelengths(netPlan, layerId, fiberId);
	}
	
	/**
	 * Returns the number of wavelengths for the given fiber. It is equivalent to 
	 * the method {@link com.net2plan.interfaces.networkDesign.NetPlan#getLinkCapacity(long) getLinkCapacity()} 
	 * from the {@link com.net2plan.interfaces.networkDesign.NetPlan NetPlan} object, 
	 * but converting capacity value from {@code double} to {@code int}.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param fiberId Fiber identifier
	 * @return Number of wavelengths
	 * @since 0.3.0
	 */
	public static int getFiberNumWavelengths(NetPlan netPlan, long layerId, long fiberId)
	{
		double u_e = netPlan.getLinkCapacity(layerId, fiberId);
		int w_f_thisFiber = (int) u_e;
		if (Math.abs(u_e - w_f_thisFiber) > 1e-6) throw new WDMException("Link capacity must be a non-negative integer representing the number of wavelengths of the fiber");

		return w_f_thisFiber;
	}
	
	/**
	 * Returns the total number of wavelengths in each fiber. It is equivalent to 
	 * the method {@link com.net2plan.interfaces.networkDesign.NetPlan#getLinkCapacityMap() getLinkCapacityMap()} 
	 * from the {@link com.net2plan.interfaces.networkDesign.NetPlan NetPlan} object, 
	 * but converting capacity values from {@code Double} to {@code Integer}.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @return Number of wavelengths per fiber
	 * @since 0.3.0
	 */
	public static Map<Long, Integer> getFiberNumWavelengthsMap(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getFiberNumWavelengthsMap(netPlan, layerId);
	}

	/**
	 * Returns the total number of wavelengths in each fiber. It is equivalent to 
	 * the method {@link com.net2plan.interfaces.networkDesign.NetPlan#getLinkCapacityMap(long) getLinkCapacityMap(layerId)} 
	 * from the {@link com.net2plan.interfaces.networkDesign.NetPlan NetPlan} object, 
	 * but converting capacity values from {@code Double} to {@code Integer}.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @return Number of wavelengths per fiber
	 * @since 0.3.0
	 */
	public static Map<Long, Integer> getFiberNumWavelengthsMap(NetPlan netPlan, long layerId)
	{
		Set<Long> fiberIds = netPlan.getLinkIds(layerId);
		Map<Long, Integer> w_f = new LinkedHashMap<Long, Integer>();
		for (long fiberId : fiberIds) w_f.put(fiberId, getFiberNumWavelengths(netPlan, layerId, fiberId));

		return w_f;
	}
	
	/**
	 * Returns the total number of wavelengths in each fiber, where item 0 corresponds 
	 * to first fiber in the network, item 1 to the second one, and so on.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @return Number of wavelengths per fiber
	 * @since 0.3.0
	 */
	public static int[] getFiberNumWavelengthsVector(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getFiberNumWavelengthsVector(netPlan, layerId);
	}
	
	/**
	 * Returns the total number of wavelengths in each fiber, where item 0 corresponds 
	 * to first fiber in the network, item 1 to the second one, and so on.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @return Number of wavelengths per fiber
	 * @since 0.3.0
	 */
	public static int[] getFiberNumWavelengthsVector(NetPlan netPlan, long layerId)
	{
		return IntUtils.toArray(getFiberNumWavelengthsMap(netPlan, layerId).values());
	}
	
	/**
	 * Returns the set of used wavelengths per fiber.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a physical topology
	 * @return Set of used wavelengths per fiber
	 * @since 0.3.0
	 */
	public static Map<Long, Set<Integer>> getFiberWavelengthOccupancyMap(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getFiberWavelengthOccupancyMap(netPlan, layerId);
	}

	/**
	 * Returns the set of used wavelengths per fiber.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a physical topology
	 * @param layerId Layer identifier
	 * @return Set of used wavelengths per fiber
	 * @since 0.3.0
	 */
	public static Map<Long, Set<Integer>> getFiberWavelengthOccupancyMap(NetPlan netPlan, long layerId)
	{
		Map<Long, Integer> w_f = getFiberNumWavelengthsMap(netPlan, layerId);
		Set<Long> fiberIds = netPlan.getLinkIds(layerId);
		Map<Long, Set<Integer>> fiberWavelengthOccupancyMap = new LinkedHashMap<Long, Set<Integer>>();
		for (long fiberId : fiberIds)
			fiberWavelengthOccupancyMap.put(fiberId, new LinkedHashSet<Integer>());
		
		Set<Long> lpIds = netPlan.getRouteIds(layerId);
		Set<Long> protectionLpIds = netPlan.getProtectionSegmentIds(layerId);
		for (long lpId : lpIds)
		{
			if (netPlan.isRouteDown(layerId, lpId)) continue;
			
			List<Long> seqFibers = netPlan.getRouteSequenceOfLinks(layerId, lpId);
			int[] seqWavelengths = getLightpathSeqWavelengths(netPlan, layerId, lpId);

			for (int hopIndex = 0; hopIndex < seqFibers.size(); hopIndex++)
			{
				long fiberId = seqFibers.get(hopIndex);
				int wavelengthId = seqWavelengths[hopIndex];

				int numWavelengths_thisFiber = w_f.get(fiberId);
				if (numWavelengths_thisFiber <= wavelengthId)
					throw new WDMException(String.format("Fiber %d only has %d wavelengths (lightpath %d, wavelength %d)", fiberId, numWavelengths_thisFiber, lpId, wavelengthId));

				if (fiberWavelengthOccupancyMap.get(fiberId).contains(wavelengthId))
					throw new WDMException(String.format("Two lightpaths/segments cannot share a wavelength (fiber %d, wavelength %d)", fiberId, wavelengthId));

				fiberWavelengthOccupancyMap.get(fiberId).add(wavelengthId);
			}
		}

		for (long protectionLpId : protectionLpIds)
		{
			if (netPlan.isProtectionSegmentDown(layerId, protectionLpId)) continue;
			
			List<Long> seqFibers = netPlan.getProtectionSegmentSequenceOfLinks(layerId, protectionLpId);
			int[] seqWavelengths = getProtectionLightpathSeqWavelengths(netPlan, layerId, protectionLpId);

			for (int hopIndex = 0; hopIndex < seqFibers.size(); hopIndex++)
			{
				long fiberId = seqFibers.get(hopIndex);
				int wavelengthId = seqWavelengths[hopIndex];

				int numWavelengths_thisFiber = w_f.get(fiberId);
				if (numWavelengths_thisFiber <= wavelengthId)
					throw new WDMException(String.format("Fiber %d only has %d wavelengths (segment %d, wavelength %d)", fiberId, numWavelengths_thisFiber, protectionLpId, wavelengthId));

				if (fiberWavelengthOccupancyMap.get(fiberId).contains(wavelengthId))
					throw new WDMException(String.format("Two lightpaths/segments cannot share a wavelength (fiber %d, wavelength %d)", fiberId, wavelengthId));

				fiberWavelengthOccupancyMap.get(fiberId).add(wavelengthId);
			}
		}

		return fiberWavelengthOccupancyMap;
	}

	/**
	 * Returns the sequence of regenerators/wavelength converters for the given lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param lightpathId Lightpath identifier
	 * @return A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static int[] getLightpathSeqRegenerators(NetPlan netPlan, long lightpathId)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getLightpathSeqRegenerators(netPlan, layerId, lightpathId);
	}

	/**
	 * Returns the sequence of regenerators/wavelength converters for the given lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param lightpathId Lightpath identifier
	 * @return A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static int[] getLightpathSeqRegenerators(NetPlan netPlan, long layerId, long lightpathId)
	{
		return parseSeqRegenerators(netPlan.getRouteAttributeMap(layerId, lightpathId));
	}
	
	/**
	 * Returns the sequence of wavelengths for the given lightpath.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param lightpathId Lightpath identifier
	 * @return Sequence of wavelengths
	 * @since 0.3.0
	 */
	public static int[] getLightpathSeqWavelengths(NetPlan netPlan, long lightpathId)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getLightpathSeqWavelengths(netPlan, layerId, lightpathId);
	}

	/**
	 * Returns the sequence of wavelengths for the given lightpath.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param lightpathId Lightpath identifier
	 * @return Sequence of wavelengths
	 * @since 0.3.0
	 */
	public static int[] getLightpathSeqWavelengths(NetPlan netPlan, long layerId, long lightpathId)
	{
		return parseSeqWavelengths(netPlan.getRouteAttributeMap(layerId, lightpathId));
	}
	
	/**
	 * Returns the number of regenerators installed per node.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a physical topology
	 * @return Number of regenerators installed per node
	 * @since 0.3.0
	 */
	public static Map<Long, Integer> getNodeRegeneratorOccupancyMap(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getNodeRegeneratorOccupancyMap(netPlan, layerId);
	}

	/**
	 * Returns the number of regenerators installed per node.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a physical topology
	 * @param layerId Layer identifier
	 * @return Number of regenerators installed per node
	 * @since 0.3.0
	 */
	public static Map<Long, Integer> getNodeRegeneratorOccupancyMap(NetPlan netPlan, long layerId)
	{
		Set<Long> nodeIds = netPlan.getNodeIds();
		Map<Long, Integer> regeneratorOccupancy = IntUtils.zeros(nodeIds);

		Set<Long> lightpathIds = netPlan.getRouteIds(layerId);
		Set<Long> segmentIds = netPlan.getProtectionSegmentIds(layerId);
		for (long lightpathId : lightpathIds)
		{
			List<Long> seqFibers = netPlan.getRouteSequenceOfLinks(layerId, lightpathId);
			int[] seqRegenerators = getLightpathSeqRegenerators(netPlan, layerId, lightpathId);
			if (seqRegenerators.length == 0)
			{
				continue;
			}

			for (int hopIndex = 0; hopIndex < seqFibers.size(); hopIndex++)
			{
				if (seqRegenerators[hopIndex] == 0)
				{
					continue;
				}

				long fiberId = seqFibers.get(hopIndex);
				long nodeId = netPlan.getLinkOriginNode(layerId, fiberId);
				regeneratorOccupancy.put(nodeId, regeneratorOccupancy.get(nodeId) + 1);
			}
		}

		for (long segmentId : segmentIds)
		{
			List<Long> seqFibers = netPlan.getProtectionSegmentSequenceOfLinks(layerId, segmentId);
			int[] seqRegenerators = getProtectionLightpathSeqRegenerators(netPlan, layerId, segmentId);
			if (seqRegenerators.length == 0) continue;

			for (int hopIndex = 0; hopIndex < seqFibers.size(); hopIndex++)
			{
				if (seqRegenerators[hopIndex] == 0) continue;

				long fiberId = seqFibers.get(hopIndex);
				long nodeId = netPlan.getLinkOriginNode(layerId, fiberId);
				regeneratorOccupancy.put(nodeId, regeneratorOccupancy.get(nodeId) + 1);
			}
		}

		return regeneratorOccupancy;
	}
	
	/**
	 * Returns the sequence of regenerators/wavelength converters for the given lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param protectionLightpathId Protection lightpath identifier
	 * @return A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static int[] getProtectionLightpathSeqRegenerators(NetPlan netPlan, long protectionLightpathId)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getProtectionLightpathSeqRegenerators(netPlan, layerId, protectionLightpathId);
	}

	/**
	 * Returns the sequence of regenerators/wavelength converters for the given lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param protectionLightpathId Protection lightpath identifier
	 * @return A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static int[] getProtectionLightpathSeqRegenerators(NetPlan netPlan, long layerId, long protectionLightpathId)
	{
		return parseSeqRegenerators(netPlan.getProtectionSegmentAttributeMap(layerId, protectionLightpathId));
	}
	
	/**
	 * Returns the sequence of wavelengths for a given protection lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param protectionLightpathId Protection lightpath identifier
	 * @return Sequence of wavelengths
	 * @since 0.3.0
	 */
	public static int[] getProtectionLightpathSeqWavelengths(NetPlan netPlan, long protectionLightpathId)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getProtectionLightpathSeqWavelengths(netPlan, layerId, protectionLightpathId);
	}

	/**
	 * Returns the sequence of wavelengths for a given protection lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param protectionLightpathId Protection lightpath identifier
	 * @return Sequence of wavelengths
	 * @since 0.3.0
	 */
	public static int[] getProtectionLightpathSeqWavelengths(NetPlan netPlan, long layerId, long protectionLightpathId)
	{
		return parseSeqWavelengths(netPlan.getProtectionSegmentAttributeMap(layerId, protectionLightpathId));
	}
	
	/**
	 * <p>Returns the set of fibers occupied in each wavelength.</p>
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a physical topology
	 * @return Set of fibers occupied in each wavelength
	 * @since 0.3.0
	 */
	public static List<Set<Long>> getWavelengthFiberOccupancy(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		return getWavelengthFiberOccupancy(netPlan, layerId);
	}

	/**
	 * <p>Returns the set of fibers occupied in each wavelength.</p>
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a physical topology
	 * @param layerId Layer identifier
	 * @return Set of fibers occupied in each wavelength
	 * @since 0.3.0
	 */
	public static List<Set<Long>> getWavelengthFiberOccupancy(NetPlan netPlan, long layerId)
	{
		List<Set<Long>> wavelengthFiberOccupancy = new ArrayList<Set<Long>>();
		Map<Long, Integer> w_f = getFiberNumWavelengthsMap(netPlan, layerId);

		int W = IntUtils.maxValue(w_f);
		for(int wavelengthId = 0; wavelengthId < W; wavelengthId++)
			wavelengthFiberOccupancy.add(new TreeSet<Long>());

		Set<Long> lpRouteIds = netPlan.getRouteIds(layerId);
		for (long lpRouteId : lpRouteIds)
		{
			if (netPlan.isRouteDown(layerId, lpRouteId)) continue;

			List<Long> seqFibers = netPlan.getRouteSequenceOfLinks(layerId, lpRouteId);
			int[] seqWavelengths = getLightpathSeqWavelengths(netPlan, layerId, lpRouteId);

			for (int hopIndex = 0; hopIndex < seqFibers.size(); hopIndex++)
			{
				long fiberId = seqFibers.get(hopIndex);
				int wavelengthId = seqWavelengths[hopIndex];

				int numWavelengths_thisFiber = w_f.get(fiberId);
				if (numWavelengths_thisFiber <= wavelengthId)
					throw new WDMException(String.format("Fiber %d only has %d wavelengths (lightpath %d, wavelength %d)", fiberId, numWavelengths_thisFiber, lpRouteId, wavelengthId));

				if (wavelengthFiberOccupancy.get(wavelengthId).contains(fiberId))
					throw new WDMException(String.format("Two lightpaths/segments cannot share a wavelength (fiber %d, wavelength %d)", fiberId, wavelengthId));

				wavelengthFiberOccupancy.get(wavelengthId).add(fiberId);
			}
		}

		Set<Long> segmentIds = netPlan.getProtectionSegmentIds(layerId);
		for (long segmentId : segmentIds)
		{
			if (netPlan.isProtectionSegmentDown(layerId, segmentId)) continue;

			List<Long> seqFibers = netPlan.getProtectionSegmentSequenceOfLinks(layerId, segmentId);
			int[] seqWavelengths = getProtectionLightpathSeqWavelengths(netPlan, layerId, segmentId);

			for (int hopIndex = 0; hopIndex < seqFibers.size(); hopIndex++)
			{
				long fiberId = seqFibers.get(hopIndex);
				int wavelengthId = seqWavelengths[hopIndex];

				int numWavelengths_thisFiber = w_f.get(fiberId);
				if (numWavelengths_thisFiber <= wavelengthId)
					throw new WDMException(String.format("Fiber %d only has %d wavelengths (segment %d, wavelength %d)", fiberId, numWavelengths_thisFiber, segmentId, wavelengthId));

				if (wavelengthFiberOccupancy.get(wavelengthId).contains(fiberId))
					throw new WDMException(String.format("Two lightpaths/segments cannot share a wavelength (fiber %d, wavelength %d)", fiberId, wavelengthId));

				wavelengthFiberOccupancy.get(wavelengthId).add(fiberId);
			}
		}

		return wavelengthFiberOccupancy;
	}
	
	/**
	 * Returns the sequence of regenerators/wavelength converters from an attribute 
	 * map, where the associated key corresponds to {@link #SEQUENCE_OF_REGENERATORS_ATTRIBUTE_NAME}.
	 * 
	 * @param attributeMap Attribute map, where key is the attribute name and the value is the parameter value
	 * @return Sequence of regenerators/wavelength converters
	 * @since 0.3.0
	 */
	public static int[] parseSeqRegenerators(Map<String, String> attributeMap)
	{
		String attributeValue = attributeMap.get(SEQUENCE_OF_REGENERATORS_ATTRIBUTE_NAME);
		if (attributeValue == null) return new int[0];

		return StringUtils.toIntArray(StringUtils.split(attributeValue, " "));
	}
	
	/**
	 * Returns the sequence of wavelengths from an attribute map, where the associated 
	 * key corresponds to {@link #SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME}.
	 * 
	 * @param attributeMap Attribute map, where key is the attribute name and the value is the parameter value
	 * @return Sequence of wavelengths
	 * @since 0.3.0
	 */
	public static int[] parseSeqWavelengths(Map<String, String> attributeMap)
	{
		String attributeValue = attributeMap.get(SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME);
		if (attributeValue == null) throw new Net2PlanException("No '" + SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME + "' attribute defined");

		return StringUtils.toIntArray(StringUtils.split(attributeMap.get(SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME), " "));
	}
	
	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a lightpath is releasing
	 * used wavelengths.
	 * 
	 * @param seqFibers Sequence of traversed fibers
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @since 0.3.0
	 */
	public static void releaseResources(List<Long> seqFibers, int[] seqWavelengths, List<Set<Long>> wavelengthFiberOccupancy)
	{
		releaseResources(null, seqFibers, seqWavelengths, wavelengthFiberOccupancy, null, null);
	}

	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a lightpath is releasing
	 * used wavelengths.
	 * 
	 * @param seqFibers Sequence of traversed fibers
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @since 0.3.0
	 */
	public static void releaseResources(List<Long> seqFibers, int wavelengthId, List<Set<Long>> wavelengthFiberOccupancy)
	{
		releaseResources(null, seqFibers, wavelengthId, wavelengthFiberOccupancy, null, null);
	}

	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a lightpath is releasing
	 * used wavelengths.
	 * 
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @param nodeRegeneratorOccupancy Number of regenerators installed per node
	 * @since 0.3.0
	 */
	public static void releaseResources(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, int[] seqWavelengths, List<Set<Long>> wavelengthFiberOccupancy, int[] seqRegenerators, Map<Long, Integer> nodeRegeneratorOccupancy)
	{
		if (seqWavelengths.length == 0) return;
		
		ListIterator<Long> fiberIt = seqFibers.listIterator();
		while(fiberIt.hasNext())
		{
			int hopId = fiberIt.nextIndex();
			long fiberId = fiberIt.next();
			int wavelengthId = seqWavelengths[hopId];
			boolean wasOccupied = wavelengthFiberOccupancy.get(wavelengthId).remove(fiberId);
			if (!wasOccupied) throw new WDMException("Wavelength " + wavelengthId + " was unused in fiber " + fiberId);

			if (seqRegenerators != null && seqRegenerators[hopId] == 1)
			{
				long nodeId = fiberMap.get(fiberId).getFirst();
				nodeRegeneratorOccupancy.put(nodeId, nodeRegeneratorOccupancy.get(nodeId) - 1);
			}
		}
	}

	/**
	 * Updates {@code wavelengthFiberOccupancy} to consider that a lightpath is releasing
	 * used wavelengths.
	 * 
	 * @param fiberMap Map of fibers, where the key is the unique fiber identifier and the value is a {@link com.net2plan.utils.Pair Pair} representing the origin node and the destination node of the fiber, respectively. It is mandatory that can be iterated in ascending order of fiber identifier (i.e. using {@code TreeMap} or those {@code Map} objects returned from {@link com.net2plan.interfaces.networkDesign.NetPlan Net2Plan} object
	 * @param seqFibers Sequence of traversed fibers
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @param wavelengthFiberOccupancy Set of occupied fibers in each wavelength
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @param nodeRegeneratorOccupancy Number of regenerators installed per node
	 * @since 0.3.0
	 */
	public static void releaseResources(Map<Long, Pair<Long, Long>> fiberMap, List<Long> seqFibers, int wavelengthId, List<Set<Long>> wavelengthFiberOccupancy, int[] seqRegenerators, Map<Long, Integer> nodeRegeneratorOccupancy)
	{
		ListIterator<Long> fiberIt = seqFibers.listIterator();
		while(fiberIt.hasNext())
		{
			int hopId = fiberIt.nextIndex();
			long fiberId = fiberIt.next();
			wavelengthFiberOccupancy.get(wavelengthId).remove(fiberId);

			if (seqRegenerators != null && seqRegenerators[hopId] == 1)
			{
				long nodeId = fiberMap.get(fiberId).getFirst();
				nodeRegeneratorOccupancy.put(nodeId, nodeRegeneratorOccupancy.get(nodeId) - 1);
			}
		}
	}

	/**
	 * Sets the number of wavelengths available on the given fiber.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param fiberId Fiber identifier
	 * @param numWavelengths Number of wavelengths for the given fiber
	 * @since 0.3.0
	 */
	public static void setFiberNumWavelengths(NetPlan netPlan, long fiberId, int numWavelengths)
	{
		long layerId = netPlan.getLayerDefaultId();
		setFiberNumWavelengths(netPlan, layerId, fiberId, numWavelengths);
	}
	
	/**
	 * Sets the number of wavelengths available on the given fiber.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param fiberId Fiber identifier
	 * @param numWavelengths Number of wavelengths for the given fiber
	 * @since 0.3.0
	 */
	public static void setFiberNumWavelengths(NetPlan netPlan, long layerId, long fiberId, int numWavelengths)
	{
		if (numWavelengths < 0) throw new WDMException("'numWavelengths' must be a non-negative integer");
		netPlan.setLinkCapacity(layerId, fiberId, numWavelengths);
	}
	
	/**
	 * Sets the number of wavelengths available in each fiber to the same value.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param numWavelengths Number of wavelengths for all fibers
	 * @since 0.2.3
	 */
	public static void setFibersNumWavelengths(NetPlan netPlan, int numWavelengths)
	{
		long layerId = netPlan.getLayerDefaultId();
		setFibersNumWavelengths(netPlan, layerId, numWavelengths);
	}

	/**
	 * Sets the number of wavelengths available in each fiber to the same value.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param numWavelengths Number of wavelengths for all fibers
	 * @since 0.3.0
	 */
	public static void setFibersNumWavelengths(NetPlan netPlan, long layerId, int numWavelengths)
	{
		Set<Long> fiberIds = netPlan.getLinkIds(layerId);
		for (long fiberId : fiberIds) setFiberNumWavelengths(netPlan, layerId, fiberId, numWavelengths);
	}
	
	/**
	 * Sets the number of wavelengths available in each fiber.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param w_f Number of wavelengths per fiber
	 * @since 0.2.3
	 */
	public static void setFibersNumWavelengths(NetPlan netPlan, int[] w_f)
	{
		long layerId = netPlan.getLayerDefaultId();
		setFibersNumWavelengths(netPlan, layerId, w_f);
	}

	/**
	 * Sets the number of wavelengths available in each fiber.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param w_f Number of wavelengths per fiber
	 * @since 0.3.0
	 */
	public static void setFibersNumWavelengths(NetPlan netPlan, long layerId, int[] w_f)
	{
		Set<Long> fiberIds = netPlan.getLinkIds(layerId);
		Map<Long, Integer> fiberId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(fiberIds);
		for (long fiberId : fiberIds)
		{
			int f = fiberId2LinearIndex.get(fiberId);
			setFiberNumWavelengths(netPlan, layerId, fiberId, w_f[f]);
		}
	}
	
	/**
	 * Sets the number of wavelengths available in each fiber.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param w_f Number of wavelengths per fiber
	 * @since 0.3.0
	 */
	public static void setFibersNumWavelengths(NetPlan netPlan, Map<Long, Integer> w_f)
	{
		long layerId = netPlan.getLayerDefaultId();
		setFibersNumWavelengths(netPlan, layerId, w_f);
	}

	/**
	 * Sets the number of wavelengths available in each fiber.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param w_f Number of wavelengths per fiber
	 * @since 0.3.0
	 */
	public static void setFibersNumWavelengths(NetPlan netPlan, long layerId, Map<Long, Integer> w_f)
	{
		Set<Long> fiberIds = netPlan.getLinkIds(layerId);
		for (long fiberId : fiberIds)
			setFiberNumWavelengths(netPlan, layerId, fiberId, w_f.get(fiberId));
	}
	
	/**
	 * Sets the sequence of regenerators/wavelength converters for a given lightpath.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param lpId Lightpath identifier
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static void setLightpathSeqRegenerators(NetPlan netPlan, long lpId, int[] seqRegenerators)
	{
		long layerId = netPlan.getLayerDefaultId();
		setLightpathSeqRegenerators(netPlan, layerId, lpId, seqRegenerators);
	}

	/**
	 * Sets the sequence of regenerators/wavelength converters for a given lightpath.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param lpId Lightpath identifier
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static void setLightpathSeqRegenerators(NetPlan netPlan, long layerId, long lpId, int[] seqRegenerators)
	{
		netPlan.setRouteAttribute(layerId, lpId, SEQUENCE_OF_REGENERATORS_ATTRIBUTE_NAME, IntUtils.join(seqRegenerators, " "));
	}
	
	/**
	 * Sets the sequence of wavelengths for the given lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param lpId Lightpath identifier
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @since 0.3.0
	 */
	public static void setLightpathSeqWavelengths(NetPlan netPlan, long lpId, int[] seqWavelengths)
	{
		long layerId = netPlan.getLayerDefaultId();
		setLightpathSeqWavelengths(netPlan, layerId, lpId, seqWavelengths);
	}

	/**
	 * Sets the sequence of wavelengths for the given lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param lpId Lightpath identifier
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @since 0.3.0
	 */
	public static void setLightpathSeqWavelengths(NetPlan netPlan, long layerId, long lpId, int[] seqWavelengths)
	{
		netPlan.setRouteAttribute(layerId, lpId, SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME, IntUtils.join(seqWavelengths, " "));
	}
	
	/**
	 * Sets the current wavelength for the given lightpath, assuming no wavelength conversion.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param lpId Lightpath identifier
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @since 0.3.0
	 */
	public static void setLightpathSeqWavelengths(NetPlan netPlan, long lpId, int wavelengthId)
	{
		long layerId = netPlan.getLayerDefaultId();
		setLightpathSeqWavelengths(netPlan, layerId, lpId, wavelengthId);
	}

	/**
	 * Sets the current wavelength for the given lightpath, assuming no wavelength conversion.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param lpId Lightpath identifier
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @since 0.3.0
	 */
	public static void setLightpathSeqWavelengths(NetPlan netPlan, long layerId, long lpId, int wavelengthId)
	{
		netPlan.setRouteAttribute(layerId, lpId, SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME, IntUtils.join(IntUtils.constantArray(netPlan.getRouteNumberOfHops(layerId, lpId), wavelengthId), " "));
	}
	
	/**
	 * Sets the sequence of regenerators/wavelength converters for a given protection lightpath.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param protectionLightpathId Protection lightpath identifier
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static void setProtectionLightpathSeqRegenerators(NetPlan netPlan, long protectionLightpathId, int[] seqRegenerators)
	{
		long layerId = netPlan.getLayerDefaultId();
		setProtectionLightpathSeqRegenerators(netPlan, layerId, protectionLightpathId, seqRegenerators);
	}

	/**
	 * Sets the sequence of regenerators/wavelength converters for a given protection lightpath.
	 *
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param protectionLightpathId Protection lightpath identifier
	 * @param seqRegenerators A 0-1 array indicating whether (1) or not (0) a regenerator/wavelength converter is required at the origin node of the corresponding fiber
	 * @since 0.3.0
	 */
	public static void setProtectionLightpathSeqRegenerators(NetPlan netPlan, long layerId, long protectionLightpathId, int[] seqRegenerators)
	{
		netPlan.setProtectionSegmentAttribute(layerId, protectionLightpathId, SEQUENCE_OF_REGENERATORS_ATTRIBUTE_NAME, IntUtils.join(seqRegenerators, " "));
	}
	
	/**
	 * Sets the sequence of wavelengths for a protection lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param protectionLightpathId Protection lightpath identifier
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @since 0.3.0
	 */
	public static void setProtectionLightpathSeqWavelengths(NetPlan netPlan, long protectionLightpathId, int[] seqWavelengths)
	{
		long layerId = netPlan.getLayerDefaultId();
		setProtectionLightpathSeqWavelengths(netPlan, layerId, protectionLightpathId, seqWavelengths);
	}

	/**
	 * Sets the sequence of wavelengths for a protection lightpath.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param protectionLightpathId Protection lightpath identifier
	 * @param seqWavelengths Sequence of wavelengths (as many as the number of links in the lightpath)
	 * @since 0.3.0
	 */
	public static void setProtectionLightpathSeqWavelengths(NetPlan netPlan, long layerId, long protectionLightpathId, int[] seqWavelengths)
	{
		netPlan.setProtectionSegmentAttribute(layerId, protectionLightpathId, SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME, IntUtils.join(seqWavelengths, " "));
	}
	
	/**
	 * Sets the sequence of wavelengths for a protection lightpath, assuming no wavelength conversion.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param protectionLightpathId Protection lightpath identifier
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @since 0.3.0
	 */
	public static void setProtectionLightpathSeqWavelengths(NetPlan netPlan, long protectionLightpathId, int wavelengthId)
	{
		long layerId = netPlan.getLayerDefaultId();
		setProtectionLightpathSeqWavelengths(netPlan, layerId, protectionLightpathId, wavelengthId);
	}

	/**
	 * Sets the sequence of wavelengths for a protection lightpath, assuming no wavelength conversion.
	 * 
	 * @param netPlan A {@link com.net2plan.interfaces.networkDesign.NetPlan} representing a WDM physical topology
	 * @param layerId Layer identifier
	 * @param protectionLightpathId Protection lightpath identifier
	 * @param wavelengthId Wavelength identifier (the same for all traversed fibers)
	 * @since 0.3.0
	 */
	public static void setProtectionLightpathSeqWavelengths(NetPlan netPlan, long layerId, long protectionLightpathId, int wavelengthId)
	{
		netPlan.setProtectionSegmentAttribute(layerId, protectionLightpathId, SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME, IntUtils.join(IntUtils.constantArray(netPlan.getProtectionSegmentNumberOfHops(layerId, protectionLightpathId), wavelengthId), " "));
	}
	
	private static class WDMException extends Net2PlanException
	{
		public WDMException(String message)
		{
			super("WDM: " + message);
		}
	}

	/**
	 * Fiber-to-weight transformer for shortest path algorithms which returns 
	 * {@code Double.MAX_VALUE} for already-occupied fibers, and the associated 
	 * weight otherwise.
	 * 
	 * @since 0.3.0
	 */
	public static class WavelengthOccupancyTransformer implements Transformer<Long, Double>
	{
		private final Transformer<Long, Double> nev;
		private final Set<Long> alreadyOccupiedFibers;

		/**
		 * Default constructor.
		 * 
		 * @param nev Original fiber-to-weight transformer
		 * @param alreadyOccupiedFibers Set of already-occupied fibers
		 * @since 0.3.0
		 */
		public WavelengthOccupancyTransformer(Transformer<Long, Double> nev, Set<Long> alreadyOccupiedFibers)
		{
			if (nev == null) this.nev = GraphUtils.JUNGUtils.getEdgeWeightTransformer(null);
			else this.nev = nev;
			
			this.alreadyOccupiedFibers = alreadyOccupiedFibers;
		}
		
		/**
		 * Returns {@code Double.MAX_VALUE} for already occupied fibers, or the 
		 * fiber weight otherwise.
		 * 
		 * @param fiberId Fiber identifier
		 * @return {@code Double.MAX_VALUE} for already occupied fibers, or the fiber weight otherwise
		 * @since 0.3.0
		 */
		@Override
		public Double transform(Long fiberId)
		{
			return alreadyOccupiedFibers.contains(fiberId) ? Double.MAX_VALUE : nev.transform(fiberId);
		}
	}
}