/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.interfaces.networkDesign;

import cern.colt.matrix.tdouble.DoubleFactory2D;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;
import com.google.common.collect.Table.Cell;
import com.net2plan.interfaces.simulation.SimAction;
import com.net2plan.interfaces.simulation.SimAction.Factory;
import com.net2plan.internal.AttributeMap;
import com.net2plan.internal.ErrorHandling;
import com.net2plan.internal.UnmodifiablePoint2D;
import com.net2plan.internal.Version;
import com.net2plan.internal.XMLUtils;
import com.net2plan.libraries.CandidatePathList;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.Constants.CheckRoutingCycleType;
import com.net2plan.utils.Constants.RoutingCycleType;
import com.net2plan.utils.Constants.RoutingType;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.IntUtils;
import com.net2plan.utils.LongUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Quadruple;
import com.net2plan.utils.StringUtils;
import java.awt.geom.Point2D;
import java.io.*;
import java.util.*;
import java.util.Map.Entry;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.XMLEvent;
import org.apache.commons.lang3.mutable.MutableLong;
import org.codehaus.stax2.XMLInputFactory2;
import org.codehaus.stax2.XMLOutputFactory2;
import org.codehaus.stax2.XMLStreamReader2;
import org.codehaus.stax2.XMLStreamWriter2;
import org.jgrapht.experimental.dag.DirectedAcyclicGraph;

/**
 * <p>Class defining a complete multi-layer network structure. Layers may 
 * contain links, demands, routes and protection segments or forwarding rules, 
 * while nodes and SRGs are defined with a network-wide scope, that is, they 
 * appear in all layers. To simplify handling of single-layer designs, methods that may refer to 
 * layer-specific elements are replicated to include a 'layer identifier' input parameter. 
 * In case of using the method with no 'layer identifier' parameter, the action 
 * will be applied to the default layer (by default it is first defined layer), 
 * which can be modified using the {@link #setLayerDefaultId(long) setLayerDefaultId()} 
 * method.</p>
 * 
 * <p>Internal representation of the {@code NetPlan} object is based on maps, 
 * aiming to maximize the performance in read/write operations. For this reason, 
 * identifiers for network elements (i.e., nodes, links...) are represented as {@code long}, 
 * and they are also incremental, that is, e.g. after removal identifiers are not renumbered.</p>
 *
 * <p>In order to improve the speed of algorithms, some methods return directly 
 * the information from the internal representation of the {@code Net2Plan} object, 
 * so that data is backed there. Also, some metrics such as link/demand carried traffic, 
 * incoming/outgoing links/demands to/from a node... are cached, and internally 
 * updated by the kernel (transparently to users). Therefore, users should copy 
 * the output (in general a {@code Set} or a {@code Map}, so users are encouraged 
 * to make copies using {@code LinkedHashSet} and {@code LinkedHashMap}, respectively, 
 * to preserve ordering) in case they want to modify it. Each method returning 
 * data directly from the internals is advertised in its description.</p>
 *
 * <p>An unmodifiable version of NetPlan object can be obtained through the 
 * {@link NetPlan#unmodifiableView() unmodifiableView} method. Instances work 
 * transparently as {@code NetPlan} object unless you try to change it. Calling 
 * any method that can potentially change the network (e.g. add/set/remove methods) 
 * throws an {@code UnsupportedOperationException}.</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 */
public class NetPlan
{
	private final static String TEMPLATE_CROSS_LAYER_UNITS_NOT_MATCHING = "Link capacity units at layer %d (%s) do not match demand traffic units at layer %d (%s)";
	private final static String TEMPLATE_NO_MORE_LAYERS_ALLOWED = "Maximum layer identifier reached";
	private final static String TEMPLATE_NO_MORE_NODES_ALLOWED = "Maximum node identifier reached";
	private final static String TEMPLATE_NO_MORE_LINKS_ALLOWED = "Maximum link identifier for layer %d reached";
	private final static String TEMPLATE_NO_MORE_DEMANDS_ALLOWED = "Maximum demand identifier for layer %d reached";
	private final static String TEMPLATE_NO_MORE_ROUTES_ALLOWED = "Maximum route identifier for layer %d reached";
	private final static String TEMPLATE_NO_MORE_SEGMENTS_ALLOWED = "Maximum protection segment identifier for layer %d reached";
	private final static String TEMPLATE_NO_MORE_SRGS_ALLOWED = "Maximum SRG identifier reached";
	private final static String TEMPLATE_NOT_ACTIVE_LAYER = "Layer %d is not in the network";
	private final static String TEMPLATE_NOT_ACTIVE_NODE = "Node %d is not in the network";
	private final static String TEMPLATE_NOT_ACTIVE_LINK = "Link %d is not at layer %d";
	private final static String TEMPLATE_NOT_ACTIVE_DEMAND = "Demand %d is not at layer %d";
	private final static String TEMPLATE_NOT_ACTIVE_ROUTE = "Route %d is not at layer %d";
	private final static String TEMPLATE_NOT_ACTIVE_FORWARDING_RULE = "Forwarding rule (%d, %d) is not at layer %d";
	private final static String TEMPLATE_NOT_ACTIVE_SEGMENT = "Protection segment %d is not at layer %d";
	private final static String TEMPLATE_NOT_ACTIVE_SRG = "SRG %d is not in the network";
	private final static String UNMODIFIABLE_EXCEPTION_STRING = "Unmodifiable NetState object - can't be changed";
	
	private RoutingType DEFAULT_ROUTING_TYPE = RoutingType.SOURCE_ROUTING;
	private boolean hasReadLayerZero;
	private boolean isModifiable;
	private final LinkedList<SimAction> actionsPerformed;
	private boolean monitoringActivated;
	private boolean addingItem, disableCacheUpdating;
	private Map<String, String> networkAttributes;
	private String networkDescription;
	private String networkName;
	
	private long layerDefaultId;
	
	private LinkedHashMap<Long, AttributeMap> layerAttributeMap;
	private LinkedHashSet<Long> layerIds;
	private LinkedHashMap<Long, String> layerDescriptionMap;
	private LinkedHashMap<Long, String> layerNameMap;
	private LinkedHashMap<Long, String> layerDemandTrafficUnitsNameMap;
	private LinkedHashMap<Long, String> layerLinkCapacityUnitsNameMap;
	private MutableLong nextLayerId;
	private MutableLong nextNodeId;
	private MutableLong nextSRGId;
	private LinkedHashMap<Long, Long> nextLinkId;
	private LinkedHashMap<Long, Long> nextDemandId;
	private LinkedHashMap<Long, Long> nextRouteId;
	private LinkedHashMap<Long, Long> nextSegmentId;
	
	private LinkedHashMap<Long, AttributeMap> nodeAttributes;
	private LinkedHashSet<Long> nodeIds;
	private LinkedHashMap<Long, String> nodeNames;
	private LinkedHashMap<Long, Point2D> nodeXYPositionMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>> cache_nodeIncomingLinks;
	private LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>> cache_nodeOutgoingLinks;
	private LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>> cache_nodeIncomingDemands;
	private LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>> cache_nodeOutgoingDemands;
	private LinkedHashMap<Long, Set<Long>> cache_nodeSRGIds;
	
	private LinkedHashMap<Long, LinkedHashSet<Long>> linkIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>> linkSRGIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>> linkRouteIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>> linkSegmentIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, Pair<Long, Long>>> linkMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> linkCapacityMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> cache_linkCarriedTrafficMap;
	private LinkedHashMap<Long, TreeSet<Long>> cache_linksOverSubscribed;
	private LinkedHashMap<Long, TreeSet<Long>> cache_linksWithZeroCapacity;
	private LinkedHashMap<Long, TreeSet<Long>> cache_linksWithUnlimitedCapacity;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> linkReservedCapacityMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> linkLengthInKmMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> linkPropagationSpeedInKmPerSecondMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>> linkAttributesMap;

	private LinkedHashMap<Long, Set<Long>> linksDown;
	private LinkedHashMap<Long, TreeSet<Long>> routesDown;
	private LinkedHashMap<Long, TreeSet<Long>> segmentsDown;
	private LinkedHashMap<Long, Set<Pair<Long, Long>>> forwardingRuleDemandLinkPairDownMap;
	private TreeSet<Long> nodesDown;
	
	private LinkedHashMap<Long, LinkedHashSet<Long>> demandIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, Pair<Long, Long>>> demandMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> demandOfferedTrafficMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>> demandAttributesMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>> demandRouteIdsMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> cache_demandCarriedTrafficMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> cache_demandBlockedTrafficMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, RoutingCycleType>> cache_demandRoutingCycleType;
	private LinkedHashMap<Long, TreeSet<Long>> cache_demandBlockedIds;
	
	private LinkedHashMap<Long, LinkedHashSet<Long>> routeIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, Long>> routeDemandIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>> routeSequenceOfLinks;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> routeCarriedTrafficMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> routeOccupiedCapacityMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>> routeBackupSegmentList;
	private LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>> routeAttributes;
	private LinkedHashMap<Long, Set<Long>> routeSameOccupiedCapacityAsCarriedTraffic;
	private LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>> routeOriginalSequenceOfLinks;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> routeOriginalCarriedTrafficMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> routeOriginalOccupiedCapacityMap;
	
	private LinkedHashMap<Long, LinkedHashSet<Long>> segmentIds;
	private LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>> segmentSequenceOfLinks;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> segmentReservedCapacity;
	private LinkedHashMap<Long, LinkedHashMap<Long, Double>> cache_segmentCarriedTrafficMap;
	private LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>> segmentRoutes;
	private LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>> segmentAttributes;
	
	private LinkedHashSet<Long> srgIds;
	private LinkedHashMap<Long, Set<Long>> srgNodes;
	private LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>> srgLinks;
	private LinkedHashMap<Long, Double> srgMeanTimeToFailInHours;
	private LinkedHashMap<Long, Double> srgMeanTimeToRepairInHours;
	private LinkedHashMap<Long, AttributeMap> srgAttributes;
	
	private DirectedAcyclicGraph<Long, DemandLinkMapping> interLayerCoupling;
	private Table<Long, Long, Pair<Long, Long>> upperLayerToLowerLayerMapping;
	private Table<Long, Long, Pair<Long, Long>> lowerLayerToUpperLayerMapping;

	private LinkedHashMap<Long, RoutingType> layerRoutingType;
	private LinkedHashMap<Long, LinkedHashSet<Pair<Long, Long>>> forwardingRuleDemandLinkPairMap;
	private LinkedHashMap<Long, Map<Pair<Long, Long>, Double>> forwardingRuleSplittingRatioMap;
	private LinkedHashMap<Long, Map<Pair<Long, Long>, Double>> forwardingRuleCarriedTrafficMap;
	private LinkedHashMap<Long, Map<Pair<Long, Long>, Map<String, String>>> forwardingRuleAttributeMap;
	private LinkedHashMap<Long, Map<Long, LinkedHashSet<Pair<Long, Long>>>> cache_demandForwardingRuleMap;
	private LinkedHashMap<Long, Map<Long, LinkedHashSet<Pair<Long, Long>>>> cache_linkForwardingRuleMap;
	private LinkedHashMap<Long, Map<Long, LinkedHashSet<Pair<Long, Long>>>> cache_nodeForwardingRuleMap;
	
	/**
	 * Default constructor.
	 *
	 * @since 0.2.0
	 */
	public NetPlan()
	{
		isModifiable = true;
		
		networkAttributes = new AttributeMap();
		networkDescription = "";
		networkName = "";
		
		actionsPerformed = new LinkedList<SimAction>();
		monitoringActivated = false;
		disableCacheUpdating = false;
		
		layerDefaultId = 0;
		nextLayerId = new MutableLong(0);
		
		layerAttributeMap = new LinkedHashMap<Long, AttributeMap>();
		layerDescriptionMap = new LinkedHashMap<Long, String>();
		layerIds = new LinkedHashSet<Long>();
		layerNameMap = new LinkedHashMap<Long, String>();
		layerDemandTrafficUnitsNameMap = new LinkedHashMap<Long, String>();
		layerLinkCapacityUnitsNameMap = new LinkedHashMap<Long, String>();
		layerRoutingType = new LinkedHashMap<Long, RoutingType>();
		forwardingRuleDemandLinkPairMap = new LinkedHashMap<Long, LinkedHashSet<Pair<Long, Long>>>();
		forwardingRuleSplittingRatioMap = new LinkedHashMap<Long, Map<Pair<Long, Long>, Double>>();
		forwardingRuleCarriedTrafficMap = new LinkedHashMap<Long, Map<Pair<Long, Long>, Double>>();
		forwardingRuleAttributeMap = new LinkedHashMap<Long, Map<Pair<Long, Long>, Map<String, String>>>();
		cache_demandForwardingRuleMap = new LinkedHashMap<Long, Map<Long, LinkedHashSet<Pair<Long, Long>>>>();
		cache_linkForwardingRuleMap = new LinkedHashMap<Long, Map<Long, LinkedHashSet<Pair<Long, Long>>>>();
		cache_nodeForwardingRuleMap = new LinkedHashMap<Long, Map<Long, LinkedHashSet<Pair<Long, Long>>>>();

		nextLinkId = new LinkedHashMap<Long, Long>();
		nextDemandId = new LinkedHashMap<Long, Long>();
		nextRouteId = new LinkedHashMap<Long, Long>();
		nextSegmentId = new LinkedHashMap<Long, Long>();

		nextNodeId = new MutableLong(0);
		nodeAttributes = new LinkedHashMap<Long, AttributeMap>();
		nodeIds = new LinkedHashSet<Long>();
		nodeNames = new LinkedHashMap<Long, String>();
		nodeXYPositionMap = new LinkedHashMap<Long, Point2D>();
		cache_nodeIncomingLinks = new LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>>();
		cache_nodeOutgoingLinks = new LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>>();
		cache_nodeIncomingDemands = new LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>>();
		cache_nodeOutgoingDemands = new LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>>();
		cache_nodeSRGIds = new LinkedHashMap<Long, Set<Long>>();

		nextSRGId = new MutableLong(0);
		linkIds = new LinkedHashMap<Long, LinkedHashSet<Long>>();
		linkSRGIds = new LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>>();
		linkRouteIds = new LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>>();
		linkSegmentIds = new LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>>();
		linkMap = new LinkedHashMap<Long, LinkedHashMap<Long, Pair<Long, Long>>>();
		linkCapacityMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		cache_linkCarriedTrafficMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		cache_linksOverSubscribed = new LinkedHashMap<Long, TreeSet<Long>>();
		cache_linksWithZeroCapacity = new LinkedHashMap<Long, TreeSet<Long>>();
		cache_linksWithUnlimitedCapacity = new LinkedHashMap<Long, TreeSet<Long>>();
		linkReservedCapacityMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		linkLengthInKmMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		linkPropagationSpeedInKmPerSecondMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		linkAttributesMap = new LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>>();

		linksDown = new LinkedHashMap<Long, Set<Long>>();
		routesDown = new LinkedHashMap<Long, TreeSet<Long>>();
		segmentsDown = new LinkedHashMap<Long, TreeSet<Long>>();
		nodesDown = new TreeSet<Long>();
		forwardingRuleDemandLinkPairDownMap = new LinkedHashMap<Long, Set<Pair<Long, Long>>>();
		
		demandIds = new LinkedHashMap<Long, LinkedHashSet<Long>>();
		demandMap = new LinkedHashMap<Long, LinkedHashMap<Long, Pair<Long, Long>>>();
		demandOfferedTrafficMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		cache_demandCarriedTrafficMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		cache_demandBlockedTrafficMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		cache_demandBlockedIds = new LinkedHashMap<Long, TreeSet<Long>>();
		demandAttributesMap = new LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>>();
		demandRouteIdsMap = new LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>>();
		cache_demandRoutingCycleType = new LinkedHashMap<Long, LinkedHashMap<Long, RoutingCycleType>>();

		routeIds = new LinkedHashMap<Long, LinkedHashSet<Long>>();
		routeDemandIds = new LinkedHashMap<Long, LinkedHashMap<Long, Long>>();
		routeSequenceOfLinks = new LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>>();
		routeCarriedTrafficMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		routeOccupiedCapacityMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		routeBackupSegmentList = new LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>>();
		routeAttributes = new LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>>();
		routeSameOccupiedCapacityAsCarriedTraffic = new LinkedHashMap<Long, Set<Long>>();
		routeOriginalSequenceOfLinks = new LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>>();
		routeOriginalCarriedTrafficMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		routeOriginalOccupiedCapacityMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();

		segmentIds = new LinkedHashMap<Long, LinkedHashSet<Long>>();
		segmentSequenceOfLinks = new LinkedHashMap<Long, LinkedHashMap<Long, List<Long>>>();
		segmentReservedCapacity = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		cache_segmentCarriedTrafficMap = new LinkedHashMap<Long, LinkedHashMap<Long, Double>>();
		segmentRoutes = new LinkedHashMap<Long, LinkedHashMap<Long, TreeSet<Long>>>();
		segmentAttributes = new LinkedHashMap<Long, LinkedHashMap<Long, AttributeMap>>();

		srgIds = new LinkedHashSet<Long>();
		srgNodes = new LinkedHashMap<Long, Set<Long>>();
		srgLinks = new LinkedHashMap<Long, LinkedHashMap<Long, Set<Long>>>();
		srgMeanTimeToFailInHours = new LinkedHashMap<Long, Double>();
		srgMeanTimeToRepairInHours = new LinkedHashMap<Long, Double>();
		srgAttributes = new LinkedHashMap<Long, AttributeMap>();

		upperLayerToLowerLayerMapping = HashBasedTable.create();
		lowerLayerToUpperLayerMapping = HashBasedTable.create();
		interLayerCoupling = new DirectedAcyclicGraph<Long, DemandLinkMapping>(DemandLinkMapping.class);

		layerAttributeMap.clear();
		layerIds.clear();
		layerNameMap.clear();
		layerDemandTrafficUnitsNameMap.clear();
		layerLinkCapacityUnitsNameMap.clear();

		nextLinkId.clear();
		nextDemandId.clear();
		nextRouteId.clear();
		nextSegmentId.clear();
		
		layerDefaultId = addLayer(null, null, null, null, null);
	}
	
	/**
	 * Generates a new network design from a given {@code .n2p} file.
	 * 
	 * @param file {@code .n2p} file
	 * @since 0.2.0
	 */
	public NetPlan(File file)
	{
		this();
		assignFrom(loadFromFile(file));
	}

	/**
	 * Generates a new network design from an input stream.
	 *
	 * @param inputStream Input stream
	 * @since 0.3.1
	 */
	public NetPlan(InputStream inputStream)
	{
		this();
		
		try
		{
			XMLInputFactory2 xmlInputFactory = (XMLInputFactory2) XMLInputFactory2.newInstance();
			XMLStreamReader2 xmlStreamReader = (XMLStreamReader2) xmlInputFactory.createXMLStreamReader(inputStream);
		
			while(xmlStreamReader.hasNext())
			{
				int eventType = xmlStreamReader.next();
				switch (eventType)
				{
					case XMLEvent.START_ELEMENT:
						String elementName = xmlStreamReader.getName().toString();
						if (!elementName.equals("network")) throw new RuntimeException("Root element must be 'network'");
						
						INetPlanFormat netPlanFormat;
						int index = xmlStreamReader.getAttributeIndex(null, "version");
						if (index == -1)
						{
							netPlanFormat = new NetPlanFormat_v1();
						}
						else
						{
							int version = xmlStreamReader.getAttributeAsInt(index);
							switch(version)
							{
								case 2:
									netPlanFormat = new NetPlanFormat_v2();
									break;

								case 3:
									netPlanFormat = new NetPlanFormat_v3();
									break;

								default:
									throw new Net2PlanException("Wrong version number");
							}
						}

						netPlanFormat.create(this, xmlStreamReader);
						
						return;

					default:
						break;
				}
			}		
		}
		catch (Net2PlanException e)
		{
			if (ErrorHandling.isDebugEnabled()) ErrorHandling.printStackTrace(e);
			throw(e);
		}
		catch (FactoryConfigurationError | Exception e)
		{
			if (ErrorHandling.isDebugEnabled()) ErrorHandling.printStackTrace(e);
			throw new RuntimeException(e);
		}
		
		throw new Net2PlanException("Not a valid .n2p file");
	}

	/**
	 * Activate monitoring of actions. This will save all the actions (i.e. add node, 
	 * modify link capacity, and so on) in a list, which could be retrieved by 
	 * {@link #getActionsPerformed() getActionsPerformed()} method.
	 * 
	 * @since 0.3.0
	 */
	public void activateActionMonitoring()
	{
		checkIsModifiable();
		
		if (isActionMonitoringActivated()) return;
		monitoringActivated = true;
	}

	/**
	 * <p>Adds a new traffic demand.</p>
	 * 
	 * <p><b>Important</b>: Self-demands are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_ADDED DEMAND_ADDED} action will be issued.</p>
	 *
	 * @param ingressNodeId Identifier of the ingress node
	 * @param egressNodeId Identifier of the egress node
	 * @param offeredTraffic Offered traffic by this demand. It must be greater or equal than zero
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Demand identifier
	 * @since 0.2.0
	 */
	public long addDemand(long ingressNodeId, long egressNodeId, double offeredTraffic, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addDemand(layerId, ingressNodeId, egressNodeId, offeredTraffic, attributes);
	}

	/**
	 * <p>Adds a new traffic demand.</p>
	 * 
	 * <p><b>Important</b>: Self-demands are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_ADDED DEMAND_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param ingressNodeId Identifier of the ingress node
	 * @param egressNodeId Identifier of the egress node
	 * @param offeredTraffic Offered traffic by this demand. It must be greater or equal than zero
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long addDemand(long layerId, long ingressNodeId, long egressNodeId, double offeredTraffic, Map<String, String> attributes)
	{
		checkIsModifiable();

		checkLayerIsActive(layerId);
		checkNodeIsActive(ingressNodeId);
		checkNodeIsActive(egressNodeId);

		if (ingressNodeId == egressNodeId) throw new Net2PlanException("Self-demands are not allowed");

		long demandId = getDemandNextId(layerId);
		
		try
		{
			addingItem = true;
			demandIds.get(layerId).add(demandId);
			demandAttributesMap.get(layerId).put(demandId, new AttributeMap());
			demandMap.get(layerId).put(demandId, Pair.unmodifiableOf(ingressNodeId, egressNodeId));
			demandRouteIdsMap.get(layerId).put(demandId, new TreeSet<Long>());
			cache_nodeIncomingDemands.get(egressNodeId).get(layerId).add(demandId);
			cache_nodeOutgoingDemands.get(ingressNodeId).get(layerId).add(demandId);
			cache_demandCarriedTrafficMap.get(layerId).put(demandId, 0.0);
			cache_demandBlockedTrafficMap.get(layerId).put(demandId, offeredTraffic);
			cache_demandBlockedIds.get(layerId).add(demandId);
			cache_demandForwardingRuleMap.get(layerId).put(demandId, new LinkedHashSet<Pair<Long, Long>>());
			cache_demandRoutingCycleType.get(layerId).put(demandId, RoutingCycleType.LOOPLESS);
			setDemandOfferedTraffic(layerId, demandId, offeredTraffic);
			setDemandAttributeMap(layerId, demandId, attributes);
			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			demandIds.get(layerId).remove(demandId);
			demandMap.get(layerId).remove(demandId);
			cache_nodeIncomingDemands.get(egressNodeId).get(layerId).remove(demandId);
			cache_nodeOutgoingDemands.get(ingressNodeId).get(layerId).remove(demandId);
			demandOfferedTrafficMap.get(layerId).remove(demandId);
			cache_demandCarriedTrafficMap.get(layerId).remove(demandId);
			cache_demandBlockedTrafficMap.get(layerId).remove(demandId);
			cache_demandBlockedIds.get(layerId).remove(demandId);
			cache_demandForwardingRuleMap.get(layerId).remove(demandId);
			demandAttributesMap.get(layerId).remove(demandId);
			cache_demandRoutingCycleType.get(layerId).remove(demandId);
			demandRouteIdsMap.get(layerId).remove(demandId);
			throw(e);
		}
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.demandAdded(layerId, demandId));
		nextDemandId.put(layerId, demandId + 1);
		
		return demandId;
	}

	/**
	 * <p>Adds a new bidirectional traffic demand, that is, a demand on each direction.</p>
	 * 
	 * <p><b>Important</b>: Self-demands are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, two {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_ADDED DEMAND_ADDED} actions will be issued.</p>
	 *
	 * @param ingressNodeId Identifier of the ingress node
	 * @param egressNodeId Identifier of the egress node
	 * @param offeredTraffic Offered traffic by this demand. It must be greater or equal than zero
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addDemandBidirectional(long ingressNodeId, long egressNodeId, double offeredTraffic, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addDemandBidirectional(layerId, ingressNodeId, egressNodeId, offeredTraffic, attributes);
	}

	/**
	 * <p>Adds a new bidirectional traffic demand, that is, a demand on each direction.</p>
	 * 
	 * <p><b>Important</b>: Self-demands are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, two {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_ADDED DEMAND_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param ingressNodeId Identifier of the ingress node
	 * @param egressNodeId Identifier of the egress node
	 * @param offeredTraffic Offered traffic by this demand. It must be greater or equal than zero
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addDemandBidirectional(long layerId, long ingressNodeId, long egressNodeId, double offeredTraffic, Map<String, String> attributes)
	{
		long demandId_1 = addDemand(layerId, ingressNodeId, egressNodeId, offeredTraffic, attributes);
		long demandId_2 = addDemand(layerId, egressNodeId, ingressNodeId, offeredTraffic, attributes);
		
		return Pair.of(demandId_1, demandId_2);
	}

	/**
	 * <p>Adds a new forwarding rule (or override an existing one).</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} action will be issued.</p>
	 *
	 * @param demandId Demand identifier
	 * @param linkId Link identifier
	 * @param splittingRatio Splitting ratio (fraction of traffic from demand 'd' entering to the origin node of link 'e', going through link 'e')
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addForwardingRule(long demandId, long linkId, double splittingRatio, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addForwardingRule(layerId, demandId, linkId, splittingRatio, attributes);
	}
	
	/**
	 * <p>Adds a new forwarding rule (or override an existing one).</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param linkId Link identifier
	 * @param splittingRatio Splitting ratio (fraction of traffic from demand 'd' entering to the origin node of link 'e', going through link 'e')
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addForwardingRule(long layerId, long demandId, long linkId, double splittingRatio, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		checkDemandIsActive(layerId, demandId);
		checkLinkIsActive(layerId, linkId);
		
		Pair<Long, Long> demandLinkPair = Pair.unmodifiableOf(demandId, linkId);
		if (splittingRatio < 0) throw new Net2PlanException("Splitting ratio must be greater or equal than zero");
		
		long nodeId = getLinkOriginNode(layerId, linkId);
		try
		{
			addingItem = true;
			
			forwardingRuleAttributeMap.get(layerId).put(demandLinkPair, new AttributeMap());
			forwardingRuleDemandLinkPairMap.get(layerId).add(demandLinkPair);
			forwardingRuleCarriedTrafficMap.get(layerId).put(demandLinkPair, 0.0);
			forwardingRuleSplittingRatioMap.get(layerId).put(demandLinkPair, 0.0);
			cache_demandForwardingRuleMap.get(layerId).get(demandId).add(demandLinkPair);
			cache_linkForwardingRuleMap.get(layerId).get(linkId).add(demandLinkPair);
			cache_nodeForwardingRuleMap.get(nodeId).get(layerId).add(demandLinkPair);
			setForwardingRuleSplittingRatio(layerId, demandLinkPair, splittingRatio);
			setForwardingRuleAttributeMap(layerId, demandLinkPair, attributes);
			if (isNodeDown(nodeId) || isLinkDown(layerId, linkId)) forwardingRuleDemandLinkPairDownMap.get(layerId).add(demandLinkPair);

			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			
			forwardingRuleDemandLinkPairMap.get(layerId).remove(demandLinkPair);
			cache_demandForwardingRuleMap.get(layerId).get(demandId).remove(demandLinkPair);
			cache_linkForwardingRuleMap.get(layerId).get(linkId).remove(demandLinkPair);
			cache_nodeForwardingRuleMap.get(nodeId).get(layerId).remove(demandLinkPair);
			forwardingRuleCarriedTrafficMap.get(layerId).remove(demandLinkPair);
			forwardingRuleSplittingRatioMap.get(layerId).remove(demandLinkPair);
			forwardingRuleAttributeMap.get(layerId).remove(demandLinkPair);
			forwardingRuleDemandLinkPairDownMap.get(layerId).remove(demandLinkPair);
			
			throw(e);
		}
		
		return demandLinkPair;
	}
	
	/**
	 * <p>Adds a set of forwarding rules (or override existing ones).</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} actions will be issued.</p>
	 * 
	 * @param splittingRatioMap Splitting ratio map (fraction of traffic from demand 'd' entering to the origin node of link 'e', going through link 'e')
	 * @param includeUnusedRules Indicates whether or not rules whose splitting ratio is below a precision factor will be added
	 * @since 0.3.0
	 */
	public void addForwardingRules(Map<Pair<Long, Long>, Double> splittingRatioMap, boolean includeUnusedRules)
	{
		long layerId = getLayerDefaultId();
		addForwardingRules(layerId, splittingRatioMap, includeUnusedRules);
	}
	
	/**
	 * <p>Adds a set of forwarding rules (or override existing ones).</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param splittingRatioMap Splitting ratio map (fraction of traffic from demand 'd' entering to the origin node of link 'e', going through link 'e')
	 * @param includeUnusedRules Indicates whether or not rules whose splitting ratio is below a precision factor will be added
	 * @since 0.3.0
	 */
	public void addForwardingRules(long layerId, Map<Pair<Long, Long>, Double> splittingRatioMap, boolean includeUnusedRules)
	{
		checkIsModifiable();
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);

		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		
		try
		{
			disableCacheUpdating = true;
			
			Set<Long> demandIdsToUpdate = new LinkedHashSet<Long>();
			Set<Long> linkIdsToUpdate = new LinkedHashSet<Long>();
			for(Entry<Pair<Long, Long>, Double> entry : splittingRatioMap.entrySet())
			{
				double splittingRatio_thisDemandLinkPair = entry.getValue();
				if (splittingRatio_thisDemandLinkPair < PRECISION_FACTOR && !includeUnusedRules) continue;

				Pair<Long, Long> demandLinkPair = entry.getKey();
				long demandId = demandLinkPair.getFirst();
				long linkId = demandLinkPair.getSecond();
				if (!demandIdsToUpdate.contains(demandId))
				{
					Set<Pair<Long, Long>> forwardingRules_thisDemand = getDemandForwardingRules(layerId, demandId);
					for(Pair<Long, Long> auxDemandLinkPair : forwardingRules_thisDemand)
						linkIdsToUpdate.add(auxDemandLinkPair.getSecond());

					demandIdsToUpdate.add(demandId);
				}

				linkIdsToUpdate.add(linkId);
				addForwardingRule(layerId, demandLinkPair.getFirst(), demandLinkPair.getSecond(), splittingRatio_thisDemandLinkPair, null);
			}
			
			for(long demandId : demandIdsToUpdate) updateCacheDemandForwarding(layerId, demandId);
			disableCacheUpdating = false;
			for(long linkId : linkIdsToUpdate) updateCacheLinkForwarding(layerId, linkId);
		}
		catch(Throwable e)
		{
			disableCacheUpdating = false;
			throw(e);
		}
	}
	
	/**
	 * <p>Adds a new layer.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_ADDED LAYER_ADDED} action will be issued.</p>
	 *
	 * @param name Layer name ({@code null} means empty)
	 * @param description Layer description ({@code null} means empty)
	 * @param linkCapacityUnitsName Textual description of link capacity units ({@code null} means empty)
	 * @param demandTrafficUnitsName Textual description of demand traffic units ({@code null} means empty)
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long addLayer(String name, String description, String linkCapacityUnitsName, String demandTrafficUnitsName, Map<String, String> attributes)
	{
		checkIsModifiable();

		long layerId = getLayerNextId();
		try
		{
			addingItem = true;
			layerIds.add(layerId);
			layerAttributeMap.put(layerId, new AttributeMap());

			nextLinkId.put(layerId, 0L);
			nextDemandId.put(layerId, 0L);
			nextRouteId.put(layerId, 0L);
			nextSegmentId.put(layerId, 0L);

			linkIds.put(layerId, new LinkedHashSet<Long>());
			linkMap.put(layerId, new LinkedHashMap<Long, Pair<Long, Long>>());
			linkSRGIds.put(layerId, new LinkedHashMap<Long, Set<Long>>());
			linkRouteIds.put(layerId, new LinkedHashMap<Long, Set<Long>>());
			linkSegmentIds.put(layerId, new LinkedHashMap<Long, Set<Long>>());
			linkCapacityMap.put(layerId, new LinkedHashMap<Long, Double>());
			cache_linkCarriedTrafficMap.put(layerId, new LinkedHashMap<Long, Double>());
			cache_linksOverSubscribed.put(layerId, new TreeSet<Long>());
			cache_linksWithUnlimitedCapacity.put(layerId, new TreeSet<Long>());
			cache_linksWithZeroCapacity.put(layerId, new TreeSet<Long>());
			linkReservedCapacityMap.put(layerId, new LinkedHashMap<Long, Double>());
			linkLengthInKmMap.put(layerId, new LinkedHashMap<Long, Double>());
			linkPropagationSpeedInKmPerSecondMap.put(layerId, new LinkedHashMap<Long, Double>());
			linkAttributesMap.put(layerId, new LinkedHashMap<Long, AttributeMap>());
			linksDown.put(layerId, new TreeSet<Long>());
			routesDown.put(layerId, new TreeSet<Long>());
			segmentsDown.put(layerId, new TreeSet<Long>());
			forwardingRuleDemandLinkPairDownMap.put(layerId, new LinkedHashSet<Pair<Long, Long>>());

			demandIds.put(layerId, new LinkedHashSet<Long>());
			demandMap.put(layerId, new LinkedHashMap<Long, Pair<Long, Long>>());
			demandOfferedTrafficMap.put(layerId, new LinkedHashMap<Long, Double>());
			cache_demandCarriedTrafficMap.put(layerId, new LinkedHashMap<Long, Double>());
			cache_demandBlockedTrafficMap.put(layerId, new LinkedHashMap<Long, Double>());
			cache_demandBlockedIds.put(layerId, new TreeSet<Long>());
			demandAttributesMap.put(layerId, new LinkedHashMap<Long, AttributeMap>());
			demandRouteIdsMap.put(layerId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_demandRoutingCycleType.put(layerId, new LinkedHashMap<Long, RoutingCycleType>());

			routeIds.put(layerId, new LinkedHashSet<Long>());
			routeDemandIds.put(layerId, new LinkedHashMap<Long, Long>());
			routeSequenceOfLinks.put(layerId, new LinkedHashMap<Long, List<Long>>());
			routeCarriedTrafficMap.put(layerId, new LinkedHashMap<Long, Double>());
			routeOccupiedCapacityMap.put(layerId, new LinkedHashMap<Long, Double>());
			routeBackupSegmentList.put(layerId, new LinkedHashMap<Long, List<Long>>());
			routeAttributes.put(layerId, new LinkedHashMap<Long, AttributeMap>());
			routeSameOccupiedCapacityAsCarriedTraffic.put(layerId, new LinkedHashSet<Long>());
			routeOriginalSequenceOfLinks.put(layerId, new LinkedHashMap<Long, List<Long>>());
			routeOriginalCarriedTrafficMap.put(layerId, new LinkedHashMap<Long, Double>());
			routeOriginalOccupiedCapacityMap.put(layerId, new LinkedHashMap<Long, Double>());

			segmentIds.put(layerId, new LinkedHashSet<Long>());
			segmentSequenceOfLinks.put(layerId, new LinkedHashMap<Long, List<Long>>());
			segmentReservedCapacity.put(layerId, new LinkedHashMap<Long, Double>());
			cache_segmentCarriedTrafficMap.put(layerId, new LinkedHashMap<Long, Double>());
			segmentRoutes.put(layerId, new LinkedHashMap<Long, TreeSet<Long>>());
			segmentAttributes.put(layerId, new LinkedHashMap<Long, AttributeMap>());

			Set<Long> nodeIds_thisNetPlan = getNodeIds();
			for(long nodeId : nodeIds_thisNetPlan)
			{
				cache_nodeIncomingLinks.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeOutgoingLinks.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeIncomingDemands.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeOutgoingDemands.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeForwardingRuleMap.get(nodeId).put(layerId, new LinkedHashSet<Pair<Long, Long>>());
			}

			Set<Long> srgIds_thisNetPlan = getSRGIds();
			for(long srgId : srgIds_thisNetPlan)
				srgLinks.get(srgId).put(layerId, new TreeSet<Long>());
		
			layerDemandTrafficUnitsNameMap.put(layerId, "");
			layerDescriptionMap.put(layerId, "");
			layerNameMap.put(layerId, "");
			layerLinkCapacityUnitsNameMap.put(layerId, "");

			setLayerAttributeMap(layerId, attributes);
			setDemandTrafficUnitsName(layerId, demandTrafficUnitsName);
			setLayerDescription(layerId, description);
			setLayerName(layerId, name);
			setLinkCapacityUnitsName(layerId, linkCapacityUnitsName);
			layerRoutingType.put(layerId, DEFAULT_ROUTING_TYPE);
			
			forwardingRuleDemandLinkPairMap.put(layerId, new LinkedHashSet<Pair<Long, Long>>());
			forwardingRuleSplittingRatioMap.put(layerId, new LinkedHashMap<Pair<Long, Long>, Double>());
			forwardingRuleCarriedTrafficMap.put(layerId, new LinkedHashMap<Pair<Long, Long>, Double>());
			cache_demandForwardingRuleMap.put(layerId, new LinkedHashMap<Long, LinkedHashSet<Pair<Long, Long>>>());
			cache_linkForwardingRuleMap.put(layerId, new LinkedHashMap<Long, LinkedHashSet<Pair<Long, Long>>>());
			forwardingRuleAttributeMap.put(layerId, new LinkedHashMap<Pair<Long, Long>, Map<String, String>>());
			interLayerCoupling.addVertex(layerId);
			
			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			layerIds.remove(layerId);
			layerAttributeMap.remove(layerId);
			layerDemandTrafficUnitsNameMap.remove(layerId);
			layerDescriptionMap.remove(layerId);
			layerNameMap.remove(layerId);
			layerLinkCapacityUnitsNameMap.remove(layerId);
			layerRoutingType.remove(layerId);
			forwardingRuleDemandLinkPairMap.remove(layerId);
			forwardingRuleSplittingRatioMap.remove(layerId);
			forwardingRuleCarriedTrafficMap.remove(layerId);
			cache_demandForwardingRuleMap.remove(layerId);
			cache_linkForwardingRuleMap.remove(layerId);
			forwardingRuleAttributeMap.remove(layerId);

			nextLinkId.remove(layerId);
			nextDemandId.remove(layerId);
			nextRouteId.remove(layerId);
			nextSegmentId.remove(layerId);

			linkIds.remove(layerId);
			linkMap.remove(layerId);
			linkSRGIds.remove(layerId);
			linkRouteIds.remove(layerId);
			linkSegmentIds.remove(layerId);
			linkCapacityMap.remove(layerId);
			cache_linkCarriedTrafficMap.remove(layerId);
			cache_linksOverSubscribed.remove(layerId);
			cache_linksWithUnlimitedCapacity.remove(layerId);
			cache_linksWithZeroCapacity.remove(layerId);
			linkReservedCapacityMap.remove(layerId);
			linkLengthInKmMap.remove(layerId);
			linkPropagationSpeedInKmPerSecondMap.remove(layerId);
			linkAttributesMap.remove(layerId);
			linksDown.remove(layerId);
			routesDown.remove(layerId);
			segmentsDown.remove(layerId);
			forwardingRuleDemandLinkPairDownMap.remove(layerId);

			demandIds.remove(layerId);
			demandMap.remove(layerId);
			demandOfferedTrafficMap.remove(layerId);
			cache_demandCarriedTrafficMap.remove(layerId);
			cache_demandBlockedTrafficMap.remove(layerId);
			cache_demandBlockedIds.remove(layerId);
			demandAttributesMap.remove(layerId);
			demandRouteIdsMap.remove(layerId);
			cache_demandRoutingCycleType.remove(layerId);

			routeIds.remove(layerId);
			routeDemandIds.remove(layerId);
			routeSequenceOfLinks.remove(layerId);
			routeCarriedTrafficMap.remove(layerId);
			routeOccupiedCapacityMap.remove(layerId);
			routeSameOccupiedCapacityAsCarriedTraffic.remove(layerId);
			routeBackupSegmentList.remove(layerId);
			routeAttributes.remove(layerId);
			routeOriginalSequenceOfLinks.remove(layerId);
			routeOriginalCarriedTrafficMap.remove(layerId);
			routeOriginalOccupiedCapacityMap.remove(layerId);

			segmentIds.remove(layerId);
			segmentSequenceOfLinks.remove(layerId);
			segmentReservedCapacity.remove(layerId);
			cache_segmentCarriedTrafficMap.remove(layerId);
			segmentRoutes.remove(layerId);
			segmentAttributes.remove(layerId);
			
			Set<Long> nodeIds_thisNetPlan = getNodeIds();
			for(long nodeId : nodeIds_thisNetPlan)
			{
				cache_nodeIncomingLinks.get(nodeId).remove(layerId);
				cache_nodeOutgoingLinks.get(nodeId).remove(layerId);
				cache_nodeIncomingDemands.get(nodeId).remove(layerId);
				cache_nodeOutgoingDemands.get(nodeId).remove(layerId);
				cache_nodeForwardingRuleMap.get(nodeId).remove(layerId);
			}
			
			Set<Long> srgIds_thisNetPlan = getSRGIds();
			for(long srgId : srgIds_thisNetPlan)
				srgLinks.get(srgId).remove(layerId);
			
			interLayerCoupling.removeVertex(layerId);
			throw(e);
		}
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.layerAdded(layerId));

		nextLayerId.increment();
		return layerId;
	}
	
	/**
	 * <p>Adds a new link. Propagation speed is assumed to be 200000 km/s.</p>
	 *
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} action will be issued.</p>
	 *
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long addLink(long originNodeId, long destinationNodeId, double capacity, double lengthInKm, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addLink(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, attributes);
	}

	/**
	 * <p>Adds a new link. Propagation speed is assumed to be 200000 km/s.</p>
	 *
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long addLink(long layerId, long originNodeId, long destinationNodeId, double capacity, double lengthInKm, Map<String, String> attributes)
	{
		return addLink(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, 200000, attributes);
	}

	/**
	 * <p>Adds a new link.</p>
	 *
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} action will be issued.</p>
	 *
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param propagationSpeedInKmPerSecond Link propagation speed. It must be greater than zero ({@code Double.MAX_VALUE} means no propagation delay})
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long addLink(long originNodeId, long destinationNodeId, double capacity, double lengthInKm, double propagationSpeedInKmPerSecond, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addLink(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, propagationSpeedInKmPerSecond, attributes);
	}

	/**
	 * <p>Adds a new link.</p>
	 *
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param propagationSpeedInKmPerSecond Link propagation speed. It must be greater than zero ({@code Double.MAX_VALUE} means no propagation delay})
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long addLink(long layerId, long originNodeId, long destinationNodeId, double capacity, double lengthInKm, double propagationSpeedInKmPerSecond, Map<String, String> attributes)
	{
		checkIsModifiable();
		
		checkLayerIsActive(layerId);
		checkNodeIsActive(originNodeId);
		checkNodeIsActive(destinationNodeId);

		if (originNodeId == destinationNodeId) throw new Net2PlanException("Self-links are not allowed");
		long linkId = getLinkNextId(layerId);
		
		try
		{
			addingItem = true;
			linkIds.get(layerId).add(linkId);
			linkAttributesMap.get(layerId).put(linkId, new AttributeMap());
			linkMap.get(layerId).put(linkId, Pair.unmodifiableOf(originNodeId, destinationNodeId));
			linkSRGIds.get(layerId).put(linkId, new TreeSet<Long>());
			linkRouteIds.get(layerId).put(linkId, new TreeSet<Long>());
			linkSegmentIds.get(layerId).put(linkId, new TreeSet<Long>());
			cache_nodeIncomingLinks.get(destinationNodeId).get(layerId).add(linkId);
			cache_nodeOutgoingLinks.get(originNodeId).get(layerId).add(linkId);
			cache_linkCarriedTrafficMap.get(layerId).put(linkId, 0.0);
			linkReservedCapacityMap.get(layerId).put(linkId, 0.0);
			cache_linkForwardingRuleMap.get(layerId).put(linkId, new LinkedHashSet<Pair<Long, Long>>());
			setLinkCapacity(layerId, linkId, capacity);
			setLinkLengthInKm(layerId, linkId, lengthInKm);
			setLinkPropagationSpeedInKmPerSecond(layerId, linkId, propagationSpeedInKmPerSecond);
			setLinkAttributeMap(layerId, linkId, attributes);
			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			linkIds.get(layerId).remove(linkId);
			linkMap.get(layerId).remove(linkId);
			linkSRGIds.get(layerId).remove(linkId);
			linkRouteIds.get(layerId).remove(linkId);
			linkSegmentIds.get(layerId).remove(linkId);
			cache_nodeIncomingLinks.get(destinationNodeId).get(layerId).remove(linkId);
			cache_nodeOutgoingLinks.get(originNodeId).get(layerId).remove(linkId);
			linkCapacityMap.get(layerId).remove(linkId);
			cache_linkCarriedTrafficMap.get(layerId).remove(linkId);
			cache_linksOverSubscribed.get(layerId).remove(linkId);
			cache_linksWithUnlimitedCapacity.get(layerId).remove(linkId);
			cache_linksWithZeroCapacity.get(layerId).remove(linkId);
			cache_linkForwardingRuleMap.get(layerId).remove(linkId);
			linkReservedCapacityMap.get(layerId).remove(linkId);
			linkLengthInKmMap.get(layerId).remove(linkId);
			linkPropagationSpeedInKmPerSecondMap.get(layerId).remove(linkId);
			linkAttributesMap.get(layerId).remove(linkId);
			throw(e);
		}
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkAdded(layerId, linkId));
		nextLinkId.put(layerId, linkId + 1);
		
		return linkId;
	}
	
	/**
	 * <p>Adds a new bidirectional link, that is, a link on each direction. Propagation speed is assumed to be 200000 km/s.</p>
	 * 
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, two {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} actions will be issued.</p>
	 *
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addLinkBidirectional(long originNodeId, long destinationNodeId, double capacity, double lengthInKm, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addLinkBidirectional(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, attributes);
	}

	/**
	 * <p>Adds a new bidirectional link, that is, a link on each direction. Propagation speed is assumed to be 200000 km/s.</p>
	 * 
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, two {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addLinkBidirectional(long layerId, long originNodeId, long destinationNodeId, double capacity, double lengthInKm, Map<String, String> attributes)
	{
		return addLinkBidirectional(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, 200000, attributes);
	}

	/**
	 * <p>Adds a new bidirectional link, that is, a link on each direction.</p>
	 * 
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, two {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} actions will be issued.</p>
	 *
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param propagationSpeedInKmPerSecond Link propagation speed. It must be greater than zero ({@code Double.MAX_VALUE} means no propagation delay})
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addLinkBidirectional(long originNodeId, long destinationNodeId, double capacity, double lengthInKm, double propagationSpeedInKmPerSecond, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addLinkBidirectional(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, propagationSpeedInKmPerSecond, attributes);
	}

	/**
	 * <p>Adds a new bidirectional link, that is, a link on each direction.</p>
	 * 
	 * <p><b>Important</b>: Self-links are not allowed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, two {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_ADDED LINK_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param originNodeId Identifier of the origin node
	 * @param destinationNodeId Identifier of the destination node
	 * @param capacity Link capacity. It must be greater or equal than zero
	 * @param lengthInKm Link length. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairEuclideanDistance()} (for Euclidean distance) or {@link #getNodePairHaversineDistanceInKm(long, long) getNodePairHaversineDistanceInKm()} (for airline distance) methods
	 * @param propagationSpeedInKmPerSecond Link propagation speed. It must be greater than zero ({@code Double.MAX_VALUE} means no propagation delay})
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> addLinkBidirectional(long layerId, long originNodeId, long destinationNodeId, double capacity, double lengthInKm, double propagationSpeedInKmPerSecond, Map<String, String> attributes)
	{
		long linkId_1 = addLink(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, propagationSpeedInKmPerSecond, attributes);
		long linkId_2 = addLink(layerId, destinationNodeId, originNodeId, capacity, lengthInKm, propagationSpeedInKmPerSecond, attributes);
		
		return Pair.of(linkId_1, linkId_2);
	}
	
	/**
	 * <p>Adds a link to a given SRG.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param linkId Link identifier
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void addLinkToSRG(long linkId, long srgId)
	{
		long layerId = getLayerDefaultId();
		addLinkToSRG(layerId, linkId, srgId);
	}

	/**
	 * <p>Adds a link to a given SRG.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void addLinkToSRG(long layerId, long linkId, long srgId)
	{
		checkIsModifiable();
		
		checkLinkIsActive(layerId, linkId);
		checkSRGIsActive(srgId);
		
		srgLinks.get(srgId).get(layerId).add(linkId);
		linkSRGIds.get(layerId).get(linkId).add(srgId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.srgModified(srgId));
	}
	
	/**
	 * Adds a new node to the network. Nodes are associated to all layers.
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_ADDED NODE_ADDED} action will be issued.</p>
	 *
	 * @param xCoord Node position in x-axis
	 * @param yCoord Node position in y-axis
	 * @param name Node name ({@code null} will be converted to "Node " + node identifier)
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Node identifier
	 * @since 0.3.0
	 */
	public long addNode(double xCoord, double yCoord, String name, Map<String, String> attributes)
	{
		checkIsModifiable();

		long nodeId = getNodeNextId();
		
		try
		{
			addingItem = true;
			nodeIds.add(nodeId);
			nodeAttributes.put(nodeId, new AttributeMap());
			cache_nodeSRGIds.put(nodeId, new TreeSet<Long>());
			cache_nodeIncomingLinks.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeOutgoingLinks.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeIncomingDemands.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeOutgoingDemands.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeForwardingRuleMap.put(nodeId, new LinkedHashMap<Long, LinkedHashSet<Pair<Long, Long>>>());
			
			Set<Long> layerIds_thisNetPlan = getLayerIds();
			for(long layerId : layerIds_thisNetPlan)
			{
				cache_nodeIncomingLinks.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeOutgoingLinks.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeIncomingDemands.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeOutgoingDemands.get(nodeId).put(layerId, new TreeSet<Long>());
				cache_nodeForwardingRuleMap.get(nodeId).put(layerId, new LinkedHashSet<Pair<Long, Long>>());
			}
			
			setNodeXYPosition(nodeId, xCoord, yCoord);
			setNodeName(nodeId, name);
			setNodeAttributeMap(nodeId, attributes);
			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			nodeIds.remove(nodeId);
			nodeXYPositionMap.remove(nodeId);
			nodeNames.remove(nodeId);
			cache_nodeSRGIds.remove(nodeId);
			nodeAttributes.remove(nodeId);
			cache_nodeIncomingLinks.remove(nodeId);
			cache_nodeOutgoingLinks.remove(nodeId);
			cache_nodeIncomingDemands.remove(nodeId);
			cache_nodeOutgoingDemands.remove(nodeId);
			cache_nodeForwardingRuleMap.remove(nodeId);
			throw(e);
		}

		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeAdded(nodeId));
		nextNodeId.increment();

		return nodeId;
	}

	/**
	 * <p>Adds a node to a given SRG.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void addNodeToSRG(long nodeId, long srgId)
	{
		checkIsModifiable();
		
		checkNodeIsActive(nodeId);
		checkSRGIsActive(srgId);
		
		srgNodes.get(srgId).add(nodeId);
		cache_nodeSRGIds.get(nodeId).add(srgId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.srgModified(srgId));
	}

	/**
	 * <p>Adds a new protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_ADDED PROTECTION_SEGMENT_ADDED} action will be issued.</p>
	 *
	 * @param sequenceOfLinks Sequence of links
	 * @param reservedCapacity Reserved capacity in each link from {@code seqLinks}. It must be greater or equal than zero
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Protection segment identifier
	 * @since 0.3.0
	 */
	public long addProtectionSegment(List<Long> sequenceOfLinks, double reservedCapacity, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addProtectionSegment(layerId, sequenceOfLinks, reservedCapacity, attributes);
	}

	/**
	 * <p>Adds a new protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_ADDED PROTECTION_SEGMENT_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param sequenceOfLinks Sequence of links
	 * @param reservedCapacity Reserved capacity in each link from {@code seqLinks}. It must be greater or equal than zero
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Protection segment identifier
	 * @since 0.3.0
	 */
	public long addProtectionSegment(long layerId, List<Long> sequenceOfLinks, double reservedCapacity, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		GraphUtils.checkRouteContinuity(getLinkMap(layerId), sequenceOfLinks, Constants.CheckRoutingCycleType.NO_CHECK);
		if (reservedCapacity <= 0) throw new Net2PlanException("Reserved capacity must be greater than zero");
		
		long segmentId = getProtectionSegmentNextId(layerId);
		try
		{
			addingItem = true;
			segmentIds.get(layerId).add(segmentId);
			segmentAttributes.get(layerId).put(segmentId, new AttributeMap());
			segmentSequenceOfLinks.get(layerId).put(segmentId, Collections.unmodifiableList(new ArrayList<Long>(sequenceOfLinks)));
			
			boolean segmentDown = false;
			for(long linkId : sequenceOfLinks)
			{
				if (!segmentDown && isLinkDown(layerId, linkId)) segmentDown = true;
				linkSegmentIds.get(layerId).get(linkId).add(segmentId);
			}
			
			if (!segmentDown)
			{
				List<Long> seqNodes = getProtectionSegmentSequenceOfNodes(layerId, segmentId);
				for(long nodeId : seqNodes)
				{
					if (isNodeDown(nodeId))
					{
						segmentDown = true;
						break;
					}
				}
			}
			
			if (segmentDown) segmentsDown.get(layerId).add(segmentId);
			cache_segmentCarriedTrafficMap.get(layerId).put(segmentId, 0.0);
			segmentRoutes.get(layerId).put(segmentId, new TreeSet<Long>());
			setProtectionSegmentAttributeMap(layerId, segmentId, attributes);
			setProtectionSegmentReservedCapacity(layerId, segmentId, reservedCapacity);
			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			segmentIds.get(layerId).remove(segmentId);
			segmentSequenceOfLinks.get(layerId).remove(segmentId);
			segmentReservedCapacity.get(layerId).remove(segmentId);
			cache_segmentCarriedTrafficMap.get(layerId).remove(segmentId);
			segmentRoutes.get(layerId).remove(segmentId);
			segmentAttributes.get(layerId).remove(segmentId);
			segmentsDown.get(layerId).remove(segmentId);
			for(long linkId : sequenceOfLinks) linkSegmentIds.get(layerId).get(linkId).remove(segmentId);
			throw(e);
		}
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.protectionSegmentAdded(layerId, segmentId));
		nextSegmentId.put(layerId, segmentId + 1);
		
		return segmentId;
	}
	
	/**
	 * <p>Adds a protection segment in the default layer to the list of backup protection segments of a route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 *
	 * @param segmentId Protection segment identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void addProtectionSegmentToRouteBackupSegmentList(long segmentId, long routeId)
	{
		long layerId = getLayerDefaultId();
		addProtectionSegmentToRouteBackupSegmentList(layerId, segmentId, routeId);
	}

	/**
	 * <p>Adds a protection segment in the given layer to the list of backup protection segments of a route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void addProtectionSegmentToRouteBackupSegmentList(long layerId, long segmentId, long routeId)
	{
		checkIsModifiable();
		
		checkLayerIsActive(layerId);
		checkProtectionSegmentIsActive(layerId, segmentId);
		checkRouteIsActive(layerId, routeId);
		
		routeBackupSegmentList.get(layerId).get(routeId).add(segmentId);
		segmentRoutes.get(layerId).get(segmentId).add(routeId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
	}
	
	/**
	 * <p>Adds a new traffic route. Capacity occupied in links is assumed to be the demand carried traffic.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} action will be issued.</p>
	 *
	 * @param demandId Demand identifier
	 * @param carriedTraffic Carried traffic. It must be greater or equal than zero
	 * @param sequenceOfLinks Sequence of links
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Route identifier
	 * @since 0.3.0
	 */
	public long addRoute(long demandId, double carriedTraffic, List<Long> sequenceOfLinks, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addRoute(layerId, demandId, carriedTraffic, sequenceOfLinks, attributes);
	}

	/**
	 * <p>Adds a new traffic route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} action will be issued.</p>
	 *
	 * @param demandId Demand identifier
	 * @param carriedTraffic Carried traffic. It must be greater or equal than zero
	 * @param occupiedLinkCapacity Occupied link capacity. If -1, it will be assumed to be equal to the carried traffic. Otherwise, it must be greater or equal than zero.
	 * @param sequenceOfLinks Sequence of links
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Route identifier
	 * @since 0.3.0
	 */
	public long addRoute(long demandId, double carriedTraffic, double occupiedLinkCapacity, List<Long> sequenceOfLinks, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		return addRoute(layerId, demandId, carriedTraffic, occupiedLinkCapacity, sequenceOfLinks, attributes);
	}

	/**
	 * <p>Adds a new traffic route. Capacity occupied in links is assumed to be the demand carried traffic.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param carriedTraffic Carried traffic. It must be greater or equal than zero
	 * @param sequenceOfLinks Sequence of links
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Route identifier
	 * @since 0.3.0
	 */
	public long addRoute(long layerId, long demandId, double carriedTraffic, List<Long> sequenceOfLinks, Map<String, String> attributes)
	{
		return addRoute(layerId, demandId, carriedTraffic, -1, sequenceOfLinks, attributes);
	}

	/**
	 * <p>Adds a new traffic route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param carriedTraffic Carried traffic. It must be greater or equal than zero
	 * @param occupiedLinkCapacity Occupied link capacity. If -1, it will be assumed to be equal to the carried traffic. Otherwise, it must be greater or equal than zero.
	 * @param sequenceOfLinks Sequence of links
	 * @param attributes Map for user-defined attributes ({@code null} means 'no attribute'). Each key represents the attribute name, whereas value represents the attribute value
	 * @return Route identifier
	 * @since 0.3.0
	 */
	public long addRoute(long layerId, long demandId, double carriedTraffic, double occupiedLinkCapacity, List<Long> sequenceOfLinks, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		checkPathValidityForDemand(layerId, sequenceOfLinks, demandId);
		
		if (carriedTraffic < 0) throw new Net2PlanException("Carried traffic must be greater or equal than zero");
		if (occupiedLinkCapacity < 0 && occupiedLinkCapacity != -1) throw new Net2PlanException("Occupied capacity must be greater or equal than zero, or -1");
		
		long routeId = getRouteNextId(layerId);
		try
		{
			addingItem = true;
			routeIds.get(layerId).add(routeId);
			routeAttributes.get(layerId).put(routeId, new AttributeMap());
			routeDemandIds.get(layerId).put(routeId, demandId);
			routeBackupSegmentList.get(layerId).put(routeId, new LinkedList<Long>());
			demandRouteIdsMap.get(layerId).get(demandId).add(routeId);
			routeSameOccupiedCapacityAsCarriedTraffic.get(layerId).add(routeId);
			routeOccupiedCapacityMap.get(layerId).put(routeId, 0.0);
			setRouteAttributeMap(layerId, routeId, attributes);
			setRouteSequenceOfLinks(layerId, routeId, sequenceOfLinks);
			setRouteCarriedTraffic(layerId, routeId, carriedTraffic);
			setRouteOccupiedCapacity(layerId, routeId, occupiedLinkCapacity);
			routeOriginalSequenceOfLinks.get(layerId).put(routeId, Collections.unmodifiableList(new LinkedList<Long>(getRouteSequenceOfLinks(layerId, routeId))));
			routeOriginalCarriedTrafficMap.get(layerId).put(routeId, getRouteCarriedTraffic(layerId, routeId));
			routeOriginalOccupiedCapacityMap.get(layerId).put(routeId, getRouteOccupiedCapacity(layerId, routeId));
			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			routeIds.get(layerId).remove(routeId);
			routeAttributes.get(layerId).remove(routeId);
			routeCarriedTrafficMap.get(layerId).remove(routeId);
			routeOccupiedCapacityMap.get(layerId).remove(routeId);
			routeSequenceOfLinks.get(layerId).remove(routeId);
			routeOriginalCarriedTrafficMap.get(layerId).remove(routeId);
			routeOriginalOccupiedCapacityMap.get(layerId).remove(routeId);
			routeOriginalSequenceOfLinks.get(layerId).remove(routeId);
			routeBackupSegmentList.get(layerId).remove(routeId);
			routeDemandIds.get(layerId).remove(routeId);
			demandRouteIdsMap.get(layerId).get(demandId).remove(routeId);
			routeSameOccupiedCapacityAsCarriedTraffic.get(layerId).remove(routeId);
			routesDown.get(layerId).remove(routeId);
			for(long linkId : sequenceOfLinks) linkRouteIds.get(layerId).get(linkId).remove(routeId);

			throw(e);
		}
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.routeAdded(layerId, routeId));
		nextRouteId.put(layerId, routeId + 1);

		return routeId;
	}
	
	/**
	 * <p>Adds traffic routes from a candidate path list. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param cpl Candidate path list
	 * @since 0.3.0
	 */
	public void addRoutes(CandidatePathList cpl)
	{
		long layerId = getLayerDefaultId();
		addRoutes(layerId, cpl);
	}
	
	/**
	 * <p>Adds traffic routes from a candidate path list. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param cpl Candidate path list
	 * @since 0.3.0
	 */
	public void addRoutes(long layerId, CandidatePathList cpl)
	{
		addRoutes(layerId, cpl, new double[cpl.getNumberOfPaths()], null, true);
	}
	
	/**
	 * <p>Adds traffic routes from a candidate path list. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param cpl Candidate path list
	 * @param x_p Carried traffic per route. It must be greater or equal than zero
	 * @param includeUnusedRoutes Indicate whether routes carrying no traffic will be included
	 * @since 0.2.0
	 */
	public void addRoutes(CandidatePathList cpl, double[] x_p, boolean includeUnusedRoutes)
	{
		long layerId = getLayerDefaultId();
		addRoutes(layerId, cpl, x_p, includeUnusedRoutes);
	}

	/**
	 * <p>Adds traffic routes from a candidate path list. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param cpl Candidate path list
	 * @param x_p Carried traffic per route. It must be greater or equal than zero
	 * @param y_p Occupied link capacity per route. If null, it will be assumed to be equal to the carried traffic. Otherwise, it must be greater or equal than zero.
	 * @param includeUnusedRoutes Indicate whether routes carrying no traffic will be included
	 * @since 0.3.0
	 */
	public void addRoutes(CandidatePathList cpl, double[] x_p, double[] y_p, boolean includeUnusedRoutes)
	{
		long layerId = getLayerDefaultId();
		addRoutes(layerId, cpl, x_p, y_p, includeUnusedRoutes);
	}
	
	/**
	 * <p>Adds traffic routes from a candidate path list. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param cpl Candidate path list
	 * @param x_p Carried traffic per route. It must be greater or equal than zero
	 * @param includeUnusedRoutes Indicate whether routes carrying no traffic will be included
	 * @since 0.3.0
	 */
	public void addRoutes(long layerId, CandidatePathList cpl, double[] x_p, boolean includeUnusedRoutes)
	{
		addRoutes(layerId, cpl, x_p, null, includeUnusedRoutes);
	}
	
	/**
	 * <p>Adds traffic routes from a candidate path list. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param cpl Candidate path list
	 * @param x_p Carried traffic per route. It must be greater or equal than zero
	 * @param y_p Occupied link capacity per route. If null, it will be assumed to be equal to the carried traffic. Otherwise, it must be greater or equal than zero.
	 * @param includeUnusedRoutes Indicate whether routes carrying no traffic will be included
	 * @since 0.3.0
	 */
	public void addRoutes(long layerId, CandidatePathList cpl, double[] x_p, double[] y_p, boolean includeUnusedRoutes)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));

		long[] pathIdsVector = cpl.getPathIdsVector();
		int P = cpl.getNumberOfPaths();
		for (int p = 0; p < P; p++)
		{
			if (x_p[p] < PRECISION_FACTOR && !includeUnusedRoutes) continue;

			long pathId = pathIdsVector[p];
			long demandId = cpl.getPathDemand(pathId);
			List<Long> seqLinks = cpl.getPathSequenceOfLinks(pathId);
			double x_p_thisPath = x_p[p];
			double y_p_thisPath = y_p == null ? -1 : y_p[p];
			addRoute(layerId, demandId, x_p_thisPath, y_p_thisPath, seqLinks, null);
		}
	}
	
	/**
	 * <p>Adds traffic routes from demand-link routing. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param x_de Matrix containing the amount of traffic from demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandMap.keySet()}, <i>d = 1</i> refers to the second one, and so on) which traverses link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on)
	 * @since 0.3.1
	 */
	public void addRoutesFromDemandLinkCarriedTraffic(DoubleMatrix2D x_de)
	{
		long layerId = getLayerDefaultId();
		addRoutesFromDemandLinkCarriedTraffic(layerId, x_de);
	}
	
	/**
	 * <p>Adds traffic routes from demand-link routing. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param x_de Matrix containing the amount of traffic from demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandMap.keySet()}, <i>d = 1</i> refers to the second one, and so on) which traverses link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on)
	 * @since 0.3.1
	 */
	public void addRoutesFromDemandLinkCarriedTraffic(long layerId, DoubleMatrix2D x_de)
	{
		/* Convert the x_de variables into a set of routes for each demand */
		List<Long> demands = new LinkedList<Long>();
		List<Double> x_p = new LinkedList<Double>();
		List<List<Long>> seqLinks = new LinkedList<List<Long>>();
		GraphUtils.convert_xde2xp(getLinkMap(layerId), getDemandMap(layerId), x_de, demands, x_p, seqLinks);

		/* Update netPlan object adding the calculated routes */
		Iterator<Long> demands_it = demands.iterator();
		Iterator<List<Long>> seqLinks_it = seqLinks.iterator();
		Iterator<Double> x_p_it = x_p.iterator();
		while(x_p_it.hasNext())
		{
			long demandId_thisPath = demands_it.next();
			List<Long> seqLinks_thisPath = seqLinks_it.next();
			double x_p_thisPath = x_p_it.next();
			addRoute(demandId_thisPath, x_p_thisPath, seqLinks_thisPath, null);
		}
	}
	
	/**
	 * <p>Adds traffic routes from destination-link routing. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param x_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} represents the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>
	 * @since 0.3.1
	 */
	public void addRoutesFromDestinationLinkCarriedTraffic(DoubleMatrix2D x_te)
	{
		long layerId = getLayerDefaultId();
		addRoutesFromDestinationLinkCarriedTraffic(layerId, x_te);
	}
	
	/**
	 * <p>Adds traffic routes from destination-link routing. Existing routes will not be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_ADDED ROUTE_ADDED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param x_te For each destination node <i>t</i> (<i>t = 0</i> refers to the first node in {@code nodeIds}, <i>t = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the second one, and so on), {@code f_te[t][e]} represents the traffic targeted to node <i>t</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>
	 * @since 0.3.1
	 */
	public void addRoutesFromDestinationLinkCarriedTraffic(long layerId, DoubleMatrix2D x_te)
	{
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = getLinkMap(layerId);
		Map<Long, Pair<Long, Long>> demandMap_thisLayer = getDemandMap(layerId);
		
		/* Convert the x_te variables into a set of routes for each demand */
		DoubleMatrix2D f_te = GraphUtils.convert_xte2fte(nodeIds_thisNetPlan, linkMap_thisLayer, x_te);
		List<Long> demands_p = new LinkedList<Long>();
		List<Double> x_p = new LinkedList<Double>();
		List<List<Long>> seqLinks_p = new LinkedList<List<Long>>();
		GraphUtils.convert_fte2xp(nodeIds_thisNetPlan, linkMap_thisLayer, demandMap_thisLayer, getDemandOfferedTrafficMap(layerId), f_te, demands_p, x_p, seqLinks_p);

		/* Update netPlan object adding the calculated routes */
		Iterator<Long> demands_it = demands_p.iterator();
		Iterator<Double> x_p_it = x_p.iterator();
		Iterator<List<Long>> seqLinks_it = seqLinks_p.iterator();
		while(x_p_it.hasNext())
		{
			long demandId_thisPath = demands_it.next();
			List<Long> seqLinks_thisPath = seqLinks_it.next();
			double x_p_thisPath = x_p_it.next();
			addRoute(demandId_thisPath, x_p_thisPath, seqLinks_thisPath, null);
		}
	}
	
	/**
	 * <p>Adds a new SRG.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, an {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_ADDED SRG_ADDED} action will be issued.</p>
	 *
	 * @param mttfInHours Mean Time To Fail (in hours). Zero or negative value means {@code Double.MAX_VALUE}
	 * @param mttrInHours Mean Time To Repair (in hours). Zero or negative value are not allowed
	 * @param attributes  Map for user-defined attributes. The key (String value) in the map will be the attribute name and the value (String value) will be whatever value is stored in that attribute. If null, it will be assumed to be an empty LinkedHashMap
	 * @return SRG identifier
	 * @since 0.3.0
	 */
	public long addSRG(double mttfInHours, double mttrInHours, Map<String, String> attributes)
	{
		checkIsModifiable();
		
		long srgId = getSRGNextId();
		try
		{
			addingItem = true;
			srgIds.add(srgId);
			srgNodes.put(srgId, new TreeSet<Long>());
			srgLinks.put(srgId, new LinkedHashMap<Long, Set<Long>>());
			for(long layerId : getLayerIds())
				srgLinks.get(srgId).put(layerId, new TreeSet<Long>());
			srgAttributes.put(srgId, new AttributeMap());
			setSRGMeanTimeToFailInHours(srgId, mttfInHours);
			setSRGMeanTimeToRepairInHours(srgId, mttrInHours);
			setSRGAttributeMap(srgId, attributes);
			addingItem = false;
		}
		catch(Throwable e)
		{
			addingItem = false;
			srgIds.remove(srgId);
			srgNodes.remove(srgId);
			srgLinks.remove(srgId);
			srgMeanTimeToFailInHours.remove(srgId);
			srgMeanTimeToRepairInHours.remove(srgId);
			srgAttributes.remove(srgId);
			throw(e);
		}
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.srgAdded(srgId));
		nextSRGId.increment();

		return srgId;
	}
	
	/**
	 * <p>Assigns the information from the input {@code NetPlan}.</p>
	 * 
	 * <p><b>Important</b>: A shadow copy is made, so changes in the input object will be reflected in this one. For deep copies use {@link #copyFrom(com.net2plan.interfaces.networkDesign.NetPlan) copyFrom()}</p>
	 * 
	 * @param netPlan Network plan to be copied
	 * @since 0.3.0
	 */
	public void assignFrom(NetPlan netPlan)
	{
		isModifiable = netPlan.isModifiable;
		
		networkAttributes = netPlan.networkAttributes;
		networkDescription = netPlan.networkDescription;
		networkName = netPlan.networkName;
		
		layerDefaultId = netPlan.layerDefaultId;
		
		layerAttributeMap = netPlan.layerAttributeMap;
		layerDescriptionMap = netPlan.layerDescriptionMap;
		layerIds = netPlan.layerIds;
		layerNameMap = netPlan.layerNameMap;
		layerDemandTrafficUnitsNameMap = netPlan.layerDemandTrafficUnitsNameMap;
		layerLinkCapacityUnitsNameMap = netPlan.layerLinkCapacityUnitsNameMap;
		layerRoutingType = netPlan.layerRoutingType;
		forwardingRuleSplittingRatioMap = netPlan.forwardingRuleSplittingRatioMap;
		forwardingRuleCarriedTrafficMap = netPlan.forwardingRuleCarriedTrafficMap;
		forwardingRuleDemandLinkPairMap = netPlan.forwardingRuleDemandLinkPairMap;
		forwardingRuleAttributeMap = netPlan.forwardingRuleAttributeMap;

		nextLayerId = netPlan.nextLayerId;
		nextNodeId = netPlan.nextNodeId;
		nextLinkId = netPlan.nextLinkId;
		nextDemandId = netPlan.nextDemandId;
		nextRouteId = netPlan.nextRouteId;
		nextSegmentId = netPlan.nextSegmentId;
		nextSRGId = netPlan.nextSRGId;

		nodeAttributes = netPlan.nodeAttributes;
		nodeIds = netPlan.nodeIds;
		nodeNames = netPlan.nodeNames;
		cache_nodeSRGIds = netPlan.cache_nodeSRGIds;
		nodeXYPositionMap = netPlan.nodeXYPositionMap;
		cache_nodeIncomingLinks = netPlan.cache_nodeIncomingLinks;
		cache_nodeOutgoingLinks = netPlan.cache_nodeOutgoingLinks;
		cache_nodeIncomingDemands = netPlan.cache_nodeIncomingDemands;
		cache_nodeOutgoingDemands = netPlan.cache_nodeOutgoingDemands;
		cache_nodeForwardingRuleMap = netPlan.cache_nodeForwardingRuleMap;

		linkIds = netPlan.linkIds;
		linkMap = netPlan.linkMap;
		linkSRGIds = netPlan.linkSRGIds;
		linkRouteIds = netPlan.linkRouteIds;
		linkSegmentIds = netPlan.linkSegmentIds;
		linkCapacityMap = netPlan.linkCapacityMap;
		cache_linkCarriedTrafficMap = netPlan.cache_linkCarriedTrafficMap;
		cache_linksOverSubscribed = netPlan.cache_linksOverSubscribed;
		cache_linksWithUnlimitedCapacity = netPlan.cache_linksWithUnlimitedCapacity;
		cache_linksWithZeroCapacity = netPlan.cache_linksWithZeroCapacity;
		linkReservedCapacityMap = netPlan.linkReservedCapacityMap;
		linkLengthInKmMap = netPlan.linkLengthInKmMap;
		linkPropagationSpeedInKmPerSecondMap = netPlan.linkPropagationSpeedInKmPerSecondMap;
		linkAttributesMap = netPlan.linkAttributesMap;
		cache_linkForwardingRuleMap = netPlan.cache_linkForwardingRuleMap;

		linksDown = netPlan.linksDown;
		routesDown = netPlan.routesDown;
		segmentsDown = netPlan.segmentsDown;
		forwardingRuleDemandLinkPairDownMap = netPlan.forwardingRuleDemandLinkPairDownMap;
		nodesDown = netPlan.nodesDown;

		demandIds = netPlan.demandIds;
		demandMap = netPlan.demandMap;
		demandOfferedTrafficMap = netPlan.demandOfferedTrafficMap;
		cache_demandCarriedTrafficMap = netPlan.cache_demandCarriedTrafficMap;
		cache_demandBlockedTrafficMap = netPlan.cache_demandBlockedTrafficMap;
		cache_demandBlockedIds = netPlan.cache_demandBlockedIds;
		demandAttributesMap = netPlan.demandAttributesMap;
		demandRouteIdsMap = netPlan.demandRouteIdsMap;
		cache_demandRoutingCycleType = netPlan.cache_demandRoutingCycleType;
		cache_demandForwardingRuleMap = netPlan.cache_demandForwardingRuleMap;

		routeIds = netPlan.routeIds;
		routeDemandIds = netPlan.routeDemandIds;
		routeSequenceOfLinks = netPlan.routeSequenceOfLinks;
		routeCarriedTrafficMap = netPlan.routeCarriedTrafficMap;
		routeOccupiedCapacityMap = netPlan.routeOccupiedCapacityMap;
		routeSameOccupiedCapacityAsCarriedTraffic = netPlan.routeSameOccupiedCapacityAsCarriedTraffic;
		routeBackupSegmentList = netPlan.routeBackupSegmentList;
		routeAttributes = netPlan.routeAttributes;
		routeOriginalSequenceOfLinks = netPlan.routeOriginalSequenceOfLinks;
		routeOriginalCarriedTrafficMap = netPlan.routeOriginalCarriedTrafficMap;
		routeOriginalOccupiedCapacityMap = netPlan.routeOriginalOccupiedCapacityMap;

		segmentIds = netPlan.segmentIds;
		segmentSequenceOfLinks = netPlan.segmentSequenceOfLinks;
		segmentReservedCapacity = netPlan.segmentReservedCapacity;
		cache_segmentCarriedTrafficMap = netPlan.cache_segmentCarriedTrafficMap;
		segmentRoutes = netPlan.segmentRoutes;
		segmentAttributes = netPlan.segmentAttributes;

		srgIds = netPlan.srgIds;
		srgNodes = netPlan.srgNodes;
		srgLinks = netPlan.srgLinks;
		srgMeanTimeToFailInHours = netPlan.srgMeanTimeToFailInHours;
		srgMeanTimeToRepairInHours = netPlan.srgMeanTimeToRepairInHours;
		srgAttributes = netPlan.srgAttributes;

		upperLayerToLowerLayerMapping = netPlan.upperLayerToLowerLayerMapping;
		lowerLayerToUpperLayerMapping = netPlan.lowerLayerToUpperLayerMapping;
		interLayerCoupling = netPlan.interLayerCoupling;
	}
	
	/**
	 * Checks whether a demand is active. When negative, an exception will be thrown.
	 * 
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void checkDemandIsActive(long demandId)
	{
		long layerId = getLayerDefaultId();
		checkDemandIsActive(layerId, demandId);
	}

	/**
	 * Checks whether a demand is active. When negative, an exception will be thrown.
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void checkDemandIsActive(long layerId, long demandId)
	{
		if (!isDemandActive(layerId, demandId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_DEMAND, demandId, layerId));
	}

	/**
	 * <p>Checks whether a forwarding rule is active. When negative, an exception will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public void checkForwardingRuleIsActive(Pair<Long, Long> demandLinkPair)
	{
		long layerId = getLayerDefaultId();
		checkForwardingRuleIsActive(layerId, demandLinkPair);
	}

	/**
	 * <p>Checks whether a forwarding rule is active. When negative, an exception will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public void checkForwardingRuleIsActive(long layerId, Pair<Long, Long> demandLinkPair)
	{
		if (!isForwardingRuleActive(layerId, demandLinkPair)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_FORWARDING_RULE, demandLinkPair.getFirst(), demandLinkPair.getSecond(), layerId));
	}

	/**
	 * Checks whether the {@code NetPlan} object is modifiable. When negative, an exception will be thrown.
	 * 
	 * @since 0.3.0
	 */
	private void checkIsModifiable()
	{
		if (!isModifiable) throw new UnsupportedOperationException(UNMODIFIABLE_EXCEPTION_STRING);
	}
	
	/**
	 * Checks whether a layer is active. When negative, an exception will be thrown.
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void checkLayerIsActive(long layerId)
	{
		if (!isLayerActive(layerId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_LAYER, layerId));
	}

	/**
	 * Checks whether a link is active. When negative, an exception will be thrown.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void checkLinkIsActive(long linkId)
	{
		long layerId = getLayerDefaultId();
		checkLinkIsActive(layerId, linkId);
	}

	/**
	 * Checks whether a link is active. When negative, an exception will be thrown.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void checkLinkIsActive(long layerId, long linkId)
	{
		if (!isLinkActive(layerId, linkId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_LINK, linkId, layerId));
	}

	/**
	 * Checks whether a node is active. When negative, an exception will be thrown.
	 * 
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void checkNodeIsActive(long nodeId)
	{
		if (!isNodeActive(nodeId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_NODE, nodeId));
	}

	/**
	 * Checks whether a sequence of links (and maybe protection segments) is valid for a given demand.
	 *
	 * @param path Sequence of links and segments
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void checkPathValidityForDemand(List<Long> path, long demandId)
	{
		long layerId = getLayerDefaultId();
		checkPathValidityForDemand(layerId, path, demandId);
	}
	
	/**
	 * Checks whether a sequence of links (and maybe protection segments) is valid for a given demand.
	 *
	 * @param layerId Layer identifier
	 * @param path Sequence of links and segments
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void checkPathValidityForDemand(long layerId, List<Long> path, long demandId)
	{
		LinkedList<Long> aux_sequenceOfLinks = (LinkedList<Long>) convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, path);
		GraphUtils.checkRouteContinuity(getLinkMap(layerId), aux_sequenceOfLinks, Constants.CheckRoutingCycleType.NO_CHECK);
		
		long ingressNodeId = getDemandIngressNode(layerId, demandId);
		long egressNodeId = getDemandEgressNode(layerId, demandId);
		
		long originNodeIdFirstLink = getLinkOriginNode(layerId, aux_sequenceOfLinks.getFirst());
		long destinationNodeIdLastLink = getLinkDestinationNode(layerId, aux_sequenceOfLinks.getLast());

		if (ingressNodeId != originNodeIdFirstLink) throw new Net2PlanException(String.format("Ingress node of the demand (%d) and origin node of the first link in the route (%d) doesn't match", ingressNodeId, originNodeIdFirstLink));
		if (egressNodeId != destinationNodeIdLastLink) throw new Net2PlanException(String.format("Egress node of the demand (%d) and destination node of the last link in the route (%d) doesn't match", egressNodeId, destinationNodeIdLastLink));
	}

	/**
	 * <p>Checks whether a protection segment is active. When negative, an exception will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @since 0.3.0
	 */
	public void checkProtectionSegmentIsActive(long segmentId)
	{
		long layerId = getLayerDefaultId();
		checkProtectionSegmentIsActive(layerId, segmentId);
	}

	/**
	 * <p>Checks whether a protection segment is active. When negative, an exception will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @since 0.3.0
	 */
	public void checkProtectionSegmentIsActive(long layerId, long segmentId)
	{
		if (!isProtectionSegmentActive(layerId, segmentId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_SEGMENT, segmentId, layerId));
	}

	/**
	 * <p>Checks whether a route is active. When negative, an exception will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void checkRouteIsActive(long routeId)
	{
		long layerId = getLayerDefaultId();
		checkRouteIsActive(layerId, routeId);
	}

	/**
	 * <p>Checks whether a route is active. When negative, an exception will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void checkRouteIsActive(long layerId, long routeId)
	{
		if (!isRouteActive(layerId, routeId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_ROUTE, routeId, layerId));
	}

	/**
	 * Checks whether routing type is the expected one. When negative, an exception will be thrown.
	 * 
	 * @param routingType Expected {@link com.net2plan.utils.Constants.RoutingType RoutingType}
	 * @since 0.3.0
	 */
	public void checkRoutingType(RoutingType routingType)
	{
		long layerId = getLayerDefaultId();
		checkRoutingType(layerId, routingType);
	}
	
	/**
	 * Checks whether routing type is the expected one. When negative, an exception will be thrown.
	 * 
	 * @param layerId Layer identifier
	 * @param routingType Expected {@link com.net2plan.utils.Constants.RoutingType RoutingType}
	 * @since 0.3.0
	 */
	public void checkRoutingType(long layerId, RoutingType routingType)
	{
		RoutingType routingType_thisLayer = getRoutingType(layerId);
		if (routingType_thisLayer != routingType)
		{
			if (isSingleLayer()) throw new Net2PlanException("Routing type must be " + routingType);
			else throw new Net2PlanException("Routing type for layer " + layerId + " must be " + routingType);
		}
	}
	
	/**
	 * Checks whether an SRG is active. When negative, an exception will be thrown.
	 * 
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void checkSRGIsActive(long srgId)
	{
		if (!isSRGActive(srgId)) throw new Net2PlanException(String.format(TEMPLATE_NOT_ACTIVE_SRG, srgId));
	}
	
	/**
	 * Checks the current state.
	 *
	 * @param net2planParameters A key-value map with {@code Net2Plan}-wide configuration options
	 * @param allowLinkOversubscription {@code true} if link capacity constraint may be violated. Otherwise, {@code false}
	 * @param allowExcessCarriedTraffic {@code true} if carried traffic may be greater than the offered one for some demand. Otherwise, {@code false}
	 * @since 0.2.3
	 */
	public void checkValidity(Map<String, String> net2planParameters, boolean allowLinkOversubscription, boolean allowExcessCarriedTraffic)
	{
		long layerId = getLayerDefaultId();
		checkValidity(layerId, net2planParameters, allowLinkOversubscription, allowExcessCarriedTraffic);
	}
	
	/**
	 * Checks the current state.
	 *
	 * @param layerId Layer identifier
	 * @param net2planParameters A key-value map with {@code Net2Plan}-wide configuration options
	 * @param allowLinkOversubscription {@code true} if link capacity constraint may be violated. Otherwise, {@code false}
	 * @param allowExcessCarriedTraffic {@code true} if carried traffic may be greater than the offered one for some demand. Otherwise, {@code false}
	 * @since 0.3.0
	 */
	public void checkValidity(long layerId, Map<String, String> net2planParameters, boolean allowLinkOversubscription, boolean allowExcessCarriedTraffic)
	{
		double PRECISIONFACTOR = Double.parseDouble(net2planParameters.get("precisionFactor"));

		if (!allowLinkOversubscription)
		{
			Set<Long> oversubscribedLinks = getLinksOversubscribed(layerId);
			for (long linkId : oversubscribedLinks)
			{
				double u_e = getLinkCapacity(layerId, linkId);
				double y_e = getLinkCarriedTraffic(layerId, linkId);
				double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
				throw new Net2PlanException(String.format("Link capacity constraint violated for link %d (carried = %f, reserved for protection = %f, capacity = %f) at layer %d", linkId, y_e, r_e, u_e, layerId));
			}
		}

		if (!allowExcessCarriedTraffic)
		{
			Set<Long> demandIds_thisLayer = getDemandIds(layerId);
			Map<Long, Double> h_d = getDemandOfferedTrafficMap(layerId);
			Map<Long, Double> r_d = getDemandCarriedTrafficMap(layerId);
			for (long demandId : demandIds_thisLayer)
			{
				double h_d_thisDemand = h_d.get(demandId);
				double r_d_thisDemand = r_d.get(demandId);
				if (h_d_thisDemand < r_d_thisDemand - PRECISIONFACTOR)
					throw new Net2PlanException(String.format("Carried traffic for demand %d overcomes the offered traffic (offered = %f, carried = %f) at layer %d", demandId, h_d_thisDemand, r_d_thisDemand, layerId));
			}
		}
	}
	
	/**
	 * Converts a sequence of links and protection segments to the actual sequence of links.
	 * 
	 * @param path Sequence of links and protection segments
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(List<Long> path)
	{
		long layerId = getLayerDefaultId();
		return convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, path);
	}
	
	/**
	 * Converts a sequence of links and protection segments to the actual sequence of links.
	 * 
	 * @param layerId Layer identifier
	 * @param path Sequence of links and protection segments
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(long layerId, List<Long> path)
	{
		LinkedList<Long> aux_sequenceOfLinks = new LinkedList<Long>();
		for(long linkId : path)
		{
			if (linkId >= 0)
			{
				aux_sequenceOfLinks.add(linkId);
			}
			else
			{
				long segmentId = -1 - linkId;
				aux_sequenceOfLinks.addAll(getProtectionSegmentSequenceOfLinks(layerId, segmentId));
			}
		}
		
		return aux_sequenceOfLinks;
	}
	
	/**
	 * Returns a deep copy of the current design.
	 * 
	 * @return Deep copy of the current design
	 * @since 0.2.0
	 */
	public NetPlan copy()
	{
		NetPlan netPlan = new NetPlan();
		netPlan.copyFrom(this);

		return netPlan;
	}
	
	/**
	 * Removes all the current information from the current {@code NetPlan} and copy the 
	 * information from the input {@code NetPlan}.
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_ADDED LAYER_ADDED} actions will be issued.</p>
	 *
	 * @param netPlan Network plan to be copied
	 * @since 0.2.3
	 */
	public void copyFrom(NetPlan netPlan)
	{
		if (netPlan == this) return;
		copyFrom(netPlan, netPlan.getLayerIds());
	}

	/**
	 * Removes all the current information from the {@code NetPlan} and copy the 
	 * information all the specified layers from the input {@code NetPlan}.
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_ADDED LAYER_ADDED} actions will be issued.</p>
	 *
	 * @param netPlan Network plan to be copied
	 * @param layerIds Set of layers to be copied
	 * @since 0.3.0
	 */
	public void copyFrom(NetPlan netPlan, Set<Long> layerIds)
	{
		if (netPlan == this) return;

		checkIsModifiable();
		reset();
		
		/* Copy network-wide information */
		networkAttributes.clear();
		networkAttributes.putAll(netPlan.networkAttributes);
		networkDescription = netPlan.networkDescription;
		networkName = netPlan.networkName;
		
		/* Copy node-wide information */
		nextNodeId = new MutableLong(netPlan.nextNodeId);
		nodesDown = new TreeSet<Long>(netPlan.nodesDown);
		Set<Long> nodeIds_thisNetPlan = netPlan.getNodeIds();
		for(long nodeId : nodeIds_thisNetPlan)
		{
			nodeAttributes.put(nodeId, new AttributeMap(netPlan.nodeAttributes.get(nodeId)));
			nodeIds.add(nodeId);
			nodeNames.put(nodeId, netPlan.nodeNames.get(nodeId));
			cache_nodeSRGIds.put(nodeId, new TreeSet<Long>(netPlan.cache_nodeSRGIds.get(nodeId)));
			nodeXYPositionMap.put(nodeId, new UnmodifiablePoint2D(netPlan.nodeXYPositionMap.get(nodeId).getX(), netPlan.nodeXYPositionMap.get(nodeId).getY()));
			cache_nodeIncomingLinks.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeOutgoingLinks.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeIncomingDemands.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeOutgoingDemands.put(nodeId, new LinkedHashMap<Long, TreeSet<Long>>());
			cache_nodeForwardingRuleMap.put(nodeId, new LinkedHashMap<Long, LinkedHashSet<Pair<Long, Long>>>());
		}
		
		/* Copy SRG-wide information */
		nextSRGId =  new MutableLong(netPlan.nextSRGId);
		Set<Long> srgIds_thisNetPlan = netPlan.getSRGIds();
		for(long srgId : srgIds_thisNetPlan)
		{
			srgAttributes.put(srgId, new AttributeMap(netPlan.srgAttributes.get(srgId)));
			srgIds.add(srgId);
			srgMeanTimeToFailInHours.put(srgId, netPlan.srgMeanTimeToFailInHours.get(srgId));
			srgMeanTimeToRepairInHours.put(srgId, netPlan.srgMeanTimeToRepairInHours.get(srgId));
			srgNodes.put(srgId, new TreeSet<Long>(netPlan.srgNodes.get(srgId)));
			srgLinks.put(srgId, new LinkedHashMap<Long, Set<Long>>());
		}

		/* Copy layers */
		long firstLayerId = new TreeSet<Long>(layerIds).iterator().next();
		long maxLayerId = 0;
		for(long layerId : layerIds)
		{
			nextLayerId.setValue(layerId);
			copyLayerFrom(netPlan, layerId);
			
			maxLayerId = Math.max(maxLayerId, layerId);
		}
		
		nextLayerId.setValue(maxLayerId + 1);
		
		/* Copy layer coupling information */
		for(Cell<Long, Long, Pair<Long, Long>> entry : netPlan.upperLayerToLowerLayerMapping.cellSet())
		{
			Pair<Long, Long> upperLayerLink = Pair.of(entry.getRowKey(), entry.getColumnKey());
			Pair<Long, Long> lowerLayerDemand = entry.getValue();
			long upperLayerId = upperLayerLink.getFirst();
			long linkId = upperLayerLink.getSecond();
			long lowerLayerId = lowerLayerDemand.getFirst();
			long demandId = lowerLayerDemand.getSecond();
			
			if (layerIds.contains(upperLayerId) && layerIds.contains(lowerLayerId))
			{
				upperLayerToLowerLayerMapping.put(upperLayerId, linkId, lowerLayerDemand);
				lowerLayerToUpperLayerMapping.put(lowerLayerId, demandId, upperLayerLink);
				
				DemandLinkMapping coupling_thisLayerPair = interLayerCoupling.getEdge(lowerLayerId, upperLayerId);
				if (coupling_thisLayerPair == null)
				{
					coupling_thisLayerPair = new DemandLinkMapping();
					interLayerCoupling.addEdge(lowerLayerId, upperLayerId, coupling_thisLayerPair);
				}
				
				coupling_thisLayerPair.put(demandId, upperLayerLink.getSecond());
			}
		}
		
		if (firstLayerId != 0) removeLayer(0);
		layerDefaultId = isLayerActive(netPlan.getLayerDefaultId()) ? netPlan.getLayerDefaultId() : firstLayerId;
	}
	
	/**
	 * <p>Copies the default layer as a new layer. Naturally, inter-layer couplings are omitted.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_ADDED LAYER_ADDED} action will be issued.</p>
	 *
	 * @return New layer identifier
	 * @since 0.3.0
	 */
	public long copyLayer()
	{
		long layerId = getLayerDefaultId();
		return copyLayer(layerId);
	}

	/**
	 * <p>Copies a given layer as a new layer. Naturally, inter-layer couplings are omitted.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_ADDED LAYER_ADDED} action will be issued.</p>
	 *
	 * @param layerId Layer to copy from
	 * @return New layer identifier
	 * @since 0.3.0
	 */
	public long copyLayer(long layerId)
	{
		return copyLayerFrom(this, layerId);
	}
	
	/**
	 * <p>Copies the default layer from a given network design as a new layer. Naturally, inter-layer couplings are omitted.</p>
	 * 
	 * <p><b>Important</b>: Set of nodes and SRGs must be the same in both designs</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_ADDED LAYER_ADDED} action will be issued.</p>
	 *
	 * @param netPlan Network design
	 * @return New layer identifier
	 * @since 0.2.3
	 */
	public long copyLayerFrom(NetPlan netPlan)
	{
		long layerId = getLayerDefaultId();
		return copyLayerFrom(netPlan, layerId);
	}
	
	/**
	 * <p>Copies a given layer from a given network as a new layer. Naturally, inter-layer couplings are omitted.</p>
	 * 
	 * <p><b>Important</b>: Set of nodes and SRGs must be the same in both designs</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_ADDED LAYER_ADDED} action will be issued.</p>
	 *
	 * @param netPlan Network design
	 * @param layerId Layer to copy from
	 * @return New layer identifier
	 * @since 0.3.0
	 */
	public long copyLayerFrom(NetPlan netPlan, long layerId)
	{
		checkIsModifiable();
		
		if (netPlan != this)
		{
			Set<Long> nodeIds_thisNetPlan = getNodeIds();
			Set<Long> nodeIds_templateNetPlan = netPlan.getNodeIds();
			if (!nodeIds_thisNetPlan.containsAll(nodeIds_templateNetPlan)) throw new Net2PlanException("Set of nodes differs from the current one");
			
			Set<Long> nodesDown_thisNetPlan = getNodesDown();
			Set<Long> nodesDown_templateNetPlan = netPlan.getNodesDown();
			if (!nodesDown_thisNetPlan.equals(nodesDown_templateNetPlan)) throw new Net2PlanException("Set of failing nodes differs from the current one");

			Set<Long> srgIds_thisNetPlan = getSRGIds();
			Set<Long> srgIds_templateNetPlan = netPlan.getSRGIds();
			if (!srgIds_thisNetPlan.equals(srgIds_templateNetPlan)) throw new Net2PlanException("Set of SRGs differs from the current one");
		}
		
		long newLayerId = addLayer(netPlan.getLayerName(layerId), netPlan.getLayerDescription(layerId), netPlan.getLinkCapacityUnitsName(layerId), netPlan.getDemandTrafficUnitsName(layerId), netPlan.getLayerAttributeMap(layerId));
		
		RoutingType routingType_thisLayer = netPlan.getRoutingType(layerId);
		layerRoutingType.put(newLayerId, routingType_thisLayer);
		nextLinkId.put(newLayerId, netPlan.nextLinkId.get(layerId));
		nextDemandId.put(newLayerId, netPlan.nextDemandId.get(layerId));
		nextRouteId.put(newLayerId, netPlan.nextRouteId.get(layerId));
		nextSegmentId.put(newLayerId, netPlan.nextSegmentId.get(layerId));
		
		linksDown.put(newLayerId, new TreeSet<Long>(netPlan.linksDown.get(layerId)));
		routesDown.put(newLayerId, new TreeSet<Long>(netPlan.routesDown.get(layerId)));
		segmentsDown.put(newLayerId, new TreeSet<Long>(netPlan.segmentsDown.get(layerId)));
		forwardingRuleDemandLinkPairDownMap.put(newLayerId, new LinkedHashSet<Pair<Long, Long>>(netPlan.forwardingRuleDemandLinkPairDownMap.get(layerId)));
			
		cache_linksOverSubscribed.put(newLayerId, new TreeSet<Long>(netPlan.cache_linksOverSubscribed.get(layerId)));
		cache_linksWithUnlimitedCapacity.put(newLayerId, new TreeSet<Long>(netPlan.cache_linksWithUnlimitedCapacity.get(layerId)));
		cache_linksWithZeroCapacity.put(newLayerId, new TreeSet<Long>(netPlan.cache_linksWithZeroCapacity.get(layerId)));
		Map<Long, Double> linkCapacityMap_thisLayer = netPlan.linkCapacityMap.get(layerId);
		Map<Long, Double> cache_linkCarriedTrafficMap_thisLayer = netPlan.cache_linkCarriedTrafficMap.get(layerId);
		Map<Long, Double> linkReservedCapacityMap_thisLayer = netPlan.linkReservedCapacityMap.get(layerId);
		Map<Long, Double> linkLengthInKmMap_thisLayer = netPlan.linkLengthInKmMap.get(layerId);
		Map<Long, Double> linkPropagationSpeedInKmPerSecondMap_thisLayer = netPlan.linkPropagationSpeedInKmPerSecondMap.get(layerId);
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = netPlan.linkMap.get(layerId);

		Map<Long, Double> linkCapacityMap_newLayer = linkCapacityMap.get(newLayerId);
		Map<Long, Double> cache_linkCarriedTrafficMap_newLayer = cache_linkCarriedTrafficMap.get(newLayerId);
		Map<Long, Double> linkReservedCapacityMap_newLayer = linkReservedCapacityMap.get(newLayerId);
		Map<Long, Double> linkLengthInKmMap_newLayer = linkLengthInKmMap.get(newLayerId);
		Map<Long, Double> linkPropagationSpeedInKmPerSecondMap_newLayer = linkPropagationSpeedInKmPerSecondMap.get(newLayerId);
		Map<Long, Pair<Long, Long>> linkMap_newLayer = linkMap.get(newLayerId);
		
		Set<Long> linkIds_thisLayer = netPlan.getLinkIds(layerId);
		Set<Long> linkIds_newLayer = linkIds.get(newLayerId);
		for(long linkId : linkIds_thisLayer)
		{
			linkCapacityMap_newLayer.put(linkId, linkCapacityMap_thisLayer.get(linkId));
			cache_linkCarriedTrafficMap_newLayer.put(linkId, cache_linkCarriedTrafficMap_thisLayer.get(linkId));
			linkReservedCapacityMap_newLayer.put(linkId, linkReservedCapacityMap_thisLayer.get(linkId));
			linkLengthInKmMap_newLayer.put(linkId, linkLengthInKmMap_thisLayer.get(linkId));
			linkPropagationSpeedInKmPerSecondMap_newLayer.put(linkId, linkPropagationSpeedInKmPerSecondMap_thisLayer.get(linkId));
			linkMap_newLayer.put(linkId, linkMap_thisLayer.get(linkId));

			linkAttributesMap.get(newLayerId).put(linkId, new AttributeMap(netPlan.linkAttributesMap.get(layerId).get(linkId)));
			linkIds_newLayer.add(linkId);
			linkSRGIds.get(newLayerId).put(linkId, new TreeSet<Long>(netPlan.linkSRGIds.get(layerId).get(linkId)));
			linkRouteIds.get(newLayerId).put(linkId, new TreeSet<Long>(netPlan.linkRouteIds.get(layerId).get(linkId)));
			linkSegmentIds.get(newLayerId).put(linkId, new TreeSet<Long>(netPlan.linkSegmentIds.get(layerId).get(linkId)));
			cache_linkForwardingRuleMap.get(newLayerId).put(linkId, new LinkedHashSet<Pair<Long, Long>>(netPlan.cache_linkForwardingRuleMap.get(layerId).get(linkId)));
		}

		cache_demandBlockedIds.put(newLayerId, new TreeSet<Long>(netPlan.cache_demandBlockedIds.get(layerId)));
		Set<Long> demandIds_thisLayer = netPlan.getDemandIds(layerId);
		for(long demandId : demandIds_thisLayer)
		{
			demandOfferedTrafficMap.get(newLayerId).put(demandId, netPlan.demandOfferedTrafficMap.get(layerId).get(demandId));
			cache_demandCarriedTrafficMap.get(newLayerId).put(demandId, netPlan.cache_demandCarriedTrafficMap.get(layerId).get(demandId));
			cache_demandBlockedTrafficMap.get(newLayerId).put(demandId, netPlan.cache_demandBlockedTrafficMap.get(layerId).get(demandId));
			demandMap.get(newLayerId).put(demandId, netPlan.demandMap.get(layerId).get(demandId));
			cache_demandRoutingCycleType.get(newLayerId).put(demandId, netPlan.cache_demandRoutingCycleType.get(layerId).get(demandId));

			demandAttributesMap.get(newLayerId).put(demandId, new AttributeMap(netPlan.demandAttributesMap.get(layerId).get(demandId)));
			demandIds.get(newLayerId).add(demandId);
			demandRouteIdsMap.get(newLayerId).put(demandId, new TreeSet<Long>(netPlan.demandRouteIdsMap.get(layerId).get(demandId)));
			cache_demandForwardingRuleMap.get(newLayerId).put(demandId, new LinkedHashSet<Pair<Long, Long>>(netPlan.cache_demandForwardingRuleMap.get(layerId).get(demandId)));
		}

		switch(routingType_thisLayer)
		{
			case SOURCE_ROUTING:
				routeSameOccupiedCapacityAsCarriedTraffic.put(newLayerId, new LinkedHashSet<Long>(netPlan.routeSameOccupiedCapacityAsCarriedTraffic.get(layerId)));
				Set<Long> routeIds_thisLayer = netPlan.getRouteIds(layerId);
				for(long routeId : routeIds_thisLayer)
				{
					routeCarriedTrafficMap.get(newLayerId).put(routeId, netPlan.routeCarriedTrafficMap.get(layerId).get(routeId));
					routeOccupiedCapacityMap.get(newLayerId).put(routeId, netPlan.routeOccupiedCapacityMap.get(layerId).get(routeId));
					routeOriginalCarriedTrafficMap.get(newLayerId).put(routeId, netPlan.routeOriginalCarriedTrafficMap.get(layerId).get(routeId));
					routeOriginalOccupiedCapacityMap.get(newLayerId).put(routeId, netPlan.routeOriginalOccupiedCapacityMap.get(layerId).get(routeId));

					routeAttributes.get(newLayerId).put(routeId, new AttributeMap(netPlan.routeAttributes.get(layerId).get(routeId)));
					routeBackupSegmentList.get(newLayerId).put(routeId, new LinkedList<Long>(netPlan.routeBackupSegmentList.get(layerId).get(routeId)));
					routeDemandIds.get(newLayerId).put(routeId, netPlan.routeDemandIds.get(layerId).get(routeId));
					routeIds.get(newLayerId).add(routeId);
					routeSequenceOfLinks.get(newLayerId).put(routeId, Collections.unmodifiableList(new LinkedList<Long>(netPlan.routeSequenceOfLinks.get(layerId).get(routeId))));
					routeOriginalSequenceOfLinks.get(newLayerId).put(routeId, Collections.unmodifiableList(new LinkedList<Long>(netPlan.routeOriginalSequenceOfLinks.get(layerId).get(routeId))));
				}

				Set<Long> segmentIds_thisLayer = netPlan.getProtectionSegmentIds(layerId);
				for(long segmentId : segmentIds_thisLayer)
				{
					segmentReservedCapacity.get(newLayerId).put(segmentId, netPlan.segmentReservedCapacity.get(layerId).get(segmentId));
					cache_segmentCarriedTrafficMap.get(newLayerId).put(segmentId, netPlan.cache_segmentCarriedTrafficMap.get(layerId).get(segmentId));
					
					segmentAttributes.get(newLayerId).put(segmentId, new AttributeMap(netPlan.segmentAttributes.get(layerId).get(segmentId)));
					segmentIds.get(newLayerId).add(segmentId);
					segmentSequenceOfLinks.get(newLayerId).put(segmentId, Collections.unmodifiableList(new LinkedList<Long>(netPlan.segmentSequenceOfLinks.get(layerId).get(segmentId))));
					segmentRoutes.get(newLayerId).put(segmentId, new TreeSet<Long>(netPlan.segmentRoutes.get(layerId).get(segmentId)));
				}
				
				break;
				
			case HOP_BY_HOP_ROUTING:
				Set<Pair<Long, Long>> forwardingRules_thisLayer = netPlan.getForwardingRules(layerId);
				forwardingRuleDemandLinkPairMap.put(newLayerId, new LinkedHashSet<Pair<Long, Long>>(forwardingRules_thisLayer));
				forwardingRuleSplittingRatioMap.put(newLayerId, new LinkedHashMap<Pair<Long, Long>, Double>(netPlan.forwardingRuleSplittingRatioMap.get(layerId)));
				forwardingRuleCarriedTrafficMap.put(newLayerId, new LinkedHashMap<Pair<Long, Long>, Double>(netPlan.forwardingRuleCarriedTrafficMap.get(layerId)));
				for(Pair<Long, Long> forwardingRule : forwardingRules_thisLayer)
					forwardingRuleAttributeMap.get(newLayerId).put(forwardingRule, new AttributeMap(netPlan.forwardingRuleAttributeMap.get(layerId).get(forwardingRule)));
				
				break;
				
			default:
				throw new RuntimeException("Bad");
		}

		Set<Long> nodeIds_thisNetPlan = netPlan.getNodeIds();
		for(long nodeId : nodeIds_thisNetPlan)
		{
			cache_nodeIncomingLinks.get(nodeId).put(newLayerId, new TreeSet<Long>(netPlan.cache_nodeIncomingLinks.get(nodeId).get(layerId)));
			cache_nodeOutgoingLinks.get(nodeId).put(newLayerId, new TreeSet<Long>(netPlan.cache_nodeOutgoingLinks.get(nodeId).get(layerId)));
			cache_nodeIncomingDemands.get(nodeId).put(newLayerId, new TreeSet<Long>(netPlan.cache_nodeIncomingDemands.get(nodeId).get(layerId)));
			cache_nodeOutgoingDemands.get(nodeId).put(newLayerId, new TreeSet<Long>(netPlan.cache_nodeOutgoingDemands.get(nodeId).get(layerId)));
			cache_nodeForwardingRuleMap.get(nodeId).put(newLayerId, new LinkedHashSet<Pair<Long, Long>>(netPlan.cache_nodeForwardingRuleMap.get(nodeId).get(layerId)));
		}

		Set<Long> srgIds_thisNetPlan = netPlan.getSRGIds();
		for(long srgId : srgIds_thisNetPlan) srgLinks.get(srgId).put(newLayerId, new TreeSet<Long>(netPlan.srgLinks.get(srgId).get(layerId)));
		
		return newLayerId;
	}
	
	/**
	 * Couples a demand in the default layer to a link in an upper layer.
	 * 
	 * @param demandId Demand identifier
	 * @param upperLayerId Upper layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void coupleDemandWithUpperLayerLink(long demandId, long upperLayerId, long linkId)
	{
		long lowerLayerId = getLayerDefaultId();
		coupleDemandWithUpperLayerLink(lowerLayerId, demandId, upperLayerId, linkId);
	}

	/**
	 * Couples a demand in the given layer to a link in an upper layer.
	 * 
	 * @param lowerLayerId Lower layer identifier
	 * @param demandId Demand identifier
	 * @param upperLayerId Upper layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void coupleDemandWithUpperLayerLink(long lowerLayerId, long demandId, long upperLayerId, long linkId)
	{
		coupleLinkWithLowerLayerDemand(upperLayerId, linkId, lowerLayerId, demandId);
	}

	/**
	 * Couples a link in the default layer to a demand in a lower layer.
	 *
	 * @param linkId Link identifier
	 * @param lowerLayerId Lower layer identifier
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void coupleLinkWithLowerLayerDemand(long linkId, long lowerLayerId, long demandId)
	{
		long upperLayerId = getLayerDefaultId();
		coupleLinkWithLowerLayerDemand(upperLayerId, linkId, lowerLayerId, demandId);
	}
	
	/**
	 * Couples a link in the given layer to a demand in a lower layer.
	 *
	 * @param upperLayerId Upper layer identifier
	 * @param linkId Link identifier
	 * @param lowerLayerId Lower layer identifier
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void coupleLinkWithLowerLayerDemand(long upperLayerId, long linkId, long lowerLayerId, long demandId)
	{
		checkIsModifiable();
		
		checkLinkIsActive(upperLayerId, linkId);
		checkDemandIsActive(lowerLayerId, demandId);
		
		if (upperLayerId == lowerLayerId) throw new Net2PlanException("Bad - Trying to couple links and demands in the same layer");
		
		DemandLinkMapping coupling_thisLayerPair = interLayerCoupling.getEdge(lowerLayerId, upperLayerId);
		if (coupling_thisLayerPair == null)
		{
			coupling_thisLayerPair = new DemandLinkMapping();
			boolean valid;
			try { valid = interLayerCoupling.addDagEdge(lowerLayerId, upperLayerId, coupling_thisLayerPair); }
			catch (DirectedAcyclicGraph.CycleFoundException ex) { valid = false; }
			if (!valid) throw new Net2PlanException("Coupling between link " + linkId + " at layer " + upperLayerId + " and demand " + demandId + " at layer " + lowerLayerId + " would induce a cycle between layers");
		}

		String upperLayerLinkCapacityUnitsName = getLinkCapacityUnitsName(upperLayerId);
		String lowerLayerDemandTrafficUnitsName = getDemandTrafficUnitsName(lowerLayerId);
		if (!upperLayerLinkCapacityUnitsName.equals(lowerLayerDemandTrafficUnitsName))
			throw new Net2PlanException(String.format(TEMPLATE_CROSS_LAYER_UNITS_NOT_MATCHING, upperLayerId, upperLayerLinkCapacityUnitsName, lowerLayerId, lowerLayerDemandTrafficUnitsName));
		
		if (isLinkCoupled(upperLayerId, linkId)) throw new Net2PlanException("Link " + linkId + " at layer " + upperLayerId + " is already associated to a lower layer demand");
		if (isDemandCoupled(lowerLayerId, demandId)) throw new Net2PlanException("Demand " + demandId + " at layer " + lowerLayerId + " is already associated to an upper layer demand");
		
		long originNodeId_upperLayer = getLinkOriginNode(upperLayerId, linkId);
		long ingressNodeId_lowerLayer = getDemandIngressNode(lowerLayerId, demandId);
		if (originNodeId_upperLayer != ingressNodeId_lowerLayer) throw new Net2PlanException(String.format("Origin node of link %d at layer %d must be equal to the ingress node of demand %d at layer %d", linkId, upperLayerId, demandId, lowerLayerId));
		
		long destinationNodeId_lowerLayer = getLinkDestinationNode(upperLayerId, linkId);
		long egressNodeId_lowerLayer = getDemandEgressNode(lowerLayerId, demandId);
		if (destinationNodeId_lowerLayer != egressNodeId_lowerLayer) throw new Net2PlanException(String.format("Destination node of link %d at layer %d must be equal to the egress node of demand %d at layer %d", linkId, upperLayerId, demandId, lowerLayerId));

		/* Link capacity at the upper layer is equal to the carried traffic at the lower layer */
		setLinkCapacity(upperLayerId, linkId, getDemandCarriedTraffic(lowerLayerId, demandId));
		
		Pair<Long, Long> upperLayerLink = Pair.unmodifiableOf(upperLayerId, linkId);
		Pair<Long, Long> lowerLayerDemand = Pair.of(lowerLayerId, demandId);
		upperLayerToLowerLayerMapping.put(upperLayerId, linkId, lowerLayerDemand);
		lowerLayerToUpperLayerMapping.put(lowerLayerId, demandId, upperLayerLink);
		coupling_thisLayerPair.put(demandId, linkId);
	}

	/**
	 * Creates a lower layer demand from a given link.
	 * 
	 * @param linkId Link identifier
	 * @param lowerLayerId Identifier of the layer where the demand will be created
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long createLowerLayerDemandFromLink(long linkId, long lowerLayerId)
	{
		long upperLayerId = getLayerDefaultId();
		return createLowerLayerDemandFromLink(upperLayerId, linkId, lowerLayerId);
	}
	
	/**
	 * Creates a lower layer demand from a given link.
	 * 
	 * @param upperLayerId Upper layer identifier
	 * @param linkId Link identifier
	 * @param lowerLayerId Identifier of the layer where the demand will be created
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long createLowerLayerDemandFromLink(long upperLayerId, long linkId, long lowerLayerId)
	{
		if (!isLayerCouplingValid(lowerLayerId, upperLayerId))
			throw new Net2PlanException("Coupling between links at layer " + upperLayerId + " and demands at layer " + lowerLayerId + " would induce a cycle between layers");
		
		long originNodeId = getLinkOriginNode(upperLayerId, linkId);
		long destinationNodeId = getLinkDestinationNode(upperLayerId, linkId);
		double u_e = getLinkCapacity(upperLayerId, linkId);
		Map<String, String> attributes = getLinkAttributeMap(upperLayerId, linkId);

		long demandId = addDemand(lowerLayerId, originNodeId, destinationNodeId, u_e, attributes);
		coupleLinkWithLowerLayerDemand(upperLayerId, linkId, lowerLayerId, demandId);
		
		return demandId;
	}
	
	/**
	 * <p>Creates lower layer demands from existing links.</p>
	 * 
	 * @param lowerLayerId Identifier of the layer where demands will be created
	 * @return Link to demand correspondence map
	 * @since 0.3.0
	 */
	public Map<Long, Long> createLowerLayerDemandsFromLinks(long lowerLayerId)
	{
		long upperLayerId = getLayerDefaultId();
		return createLowerLayerDemandsFromLinks(upperLayerId, lowerLayerId);
	}
	
	/**
	 * <p>Creates lower layer demands from existing links.</p>
	 * 
	 * @param upperLayerId Upper layer identifier
	 * @param lowerLayerId Identifier of the layer where demands will be created
	 * @return Link to demand correspondence map
	 * @since 0.3.0
	 */
	public Map<Long, Long> createLowerLayerDemandsFromLinks(long upperLayerId, long lowerLayerId)
	{
		Set<Long> linkIds_upperLayer = getLinkIds(upperLayerId);
		Map<Long, Long> out = new LinkedHashMap<Long, Long>();
		for(long linkId : linkIds_upperLayer)
		{
			if (isLinkCoupled(upperLayerId, linkId)) continue;
			
			long demandId = createLowerLayerDemandFromLink(upperLayerId, linkId, lowerLayerId);
			out.put(linkId, demandId);
		}
		
		return out;
	}

	/**
	 * Creates an upper layer link from a given demand.
	 * 
	 * @param demandId Demand identifier
	 * @param upperLayerId Identifier of the layer where the link will be created
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long createUpperLayerLinkFromDemand(long demandId, long upperLayerId)
	{
		long lowerLayerId = getLayerDefaultId();
		return createUpperLayerLinkFromDemand(lowerLayerId, demandId, upperLayerId);
	}
	
	/**
	 * Creates an upper layer link from a given demand.
	 * 
	 * @param lowerLayerId Lower layer identifier
	 * @param demandId Demand identifier
	 * @param upperLayerId Identifier of the layer where the link will be created
	 * @return Link identifier
	 * @since 0.3.0
	 */
	public long createUpperLayerLinkFromDemand(long lowerLayerId, long demandId, long upperLayerId)
	{
		if (!isLayerCouplingValid(lowerLayerId, upperLayerId))
			throw new Net2PlanException("Coupling between links at layer " + upperLayerId + " and demands at layer " + lowerLayerId + " would induce a cycle between layers");
		
		long ingressNodeId = getDemandIngressNode(lowerLayerId, demandId);
		long egressNodeId = getDemandEgressNode(lowerLayerId, demandId);
		double h_d = getDemandOfferedTraffic(lowerLayerId, demandId);
		Map<String, String> attributes = getDemandAttributeMap(lowerLayerId, demandId);

		long linkId = addLink(upperLayerId, ingressNodeId, egressNodeId, h_d, 0, attributes);
		coupleDemandWithUpperLayerLink(lowerLayerId, demandId, upperLayerId, linkId);
		
		return linkId;
	}
	
	/**
	 * <p>Creates upper layer links from existing demands.</p>
	 * 
	 * @param upperLayerId Identifier of the layer where demands will be created
	 * @return Demand to link correspondence map
	 */
	public Map<Long, Long> createUpperLayerLinksFromDemands(long upperLayerId)
	{
		long lowerLayerId = getLayerDefaultId();
		return createUpperLayerLinksFromDemands(lowerLayerId, upperLayerId);
	}

	/**
	 * <p>Creates upper layer links from existing demands.</p>
	 * 
	 * @param lowerLayerId Lower layer identifier
	 * @param upperLayerId Identifier of the layer where demands will be created
	 * @return Demand to link correspondence map
	 * @since 0.3.0
	 */
	public Map<Long, Long> createUpperLayerLinksFromDemands(long lowerLayerId, long upperLayerId)
	{
		Set<Long> demandIds_lowerLayer = getDemandIds(lowerLayerId);
		Map<Long, Long> out = new LinkedHashMap<Long, Long>();
		for(long demandId : demandIds_lowerLayer)
		{
			if (isDemandCoupled(lowerLayerId, demandId)) continue;
			
			long linkId = createUpperLayerLinkFromDemand(lowerLayerId, demandId, upperLayerId);
			out.put(demandId, linkId);
		}
		
		return out;
	}
	
	/**
	 * Deactivates action monitoring. This will remove the list of actions monitored.
	 * Users should not call this method, this is automatically done by the 
	 * {@link #getActionsPerformed() getActionsPerformed()} method.
	 * 
	 * @since 0.3.0
	 */
	public void deactivateActionMonitoring()
	{
		monitoringActivated = false;
		actionsPerformed.clear();
	}

	/**
	 * Decouples a demand from an upper layer link.
	 * 
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void decoupleDemand(long demandId)
	{
		long layerId = getLayerDefaultId();
		decoupleDemand(layerId, demandId);
	}
	
	/**
	 * Decouples a demand from an upper layer link.
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void decoupleDemand(long layerId, long demandId)
	{
		Pair<Long, Long> upperLayerLink = getDemandCoupledUpperLayerLink(layerId, demandId);
		decoupleLink(upperLayerLink.getFirst(), upperLayerLink.getSecond());
	}
	
	/**
	 * Decouples a link from a lower layer demand.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void decoupleLink(long linkId)
	{
		long layerId = getLayerDefaultId();
		decoupleLink(layerId, linkId);
	}

	/**
	 * Decouples a link from a lower layer demand.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void decoupleLink(long layerId, long linkId)
	{
		checkIsModifiable();
		
		Pair<Long, Long> lowerLayerDemand = getLinkCoupledLowerLayerDemand(layerId, linkId);
		lowerLayerToUpperLayerMapping.remove(lowerLayerDemand.getFirst(), lowerLayerDemand.getSecond());
		upperLayerToLowerLayerMapping.remove(layerId, linkId);
		DemandLinkMapping coupling_thisLayerPair = interLayerCoupling.getEdge(lowerLayerDemand.getFirst(), layerId);
		coupling_thisLayerPair.remove(lowerLayerDemand.getSecond());
		if (coupling_thisLayerPair.isEmpty()) interLayerCoupling.removeEdge(lowerLayerDemand.getFirst(), layerId);
	}

	/**
	 * Returns the list of monitored actions since {@link #activateActionMonitoring() activateActionMonitoring()} was called. 
	 * {@link #deactivateActionMonitoring() deactivateActionMonitoring()} is called internally.
	 * 
	 * @return List of monitored actions
	 * @since 0.3.0
	 */
	public List<SimAction> getActionsPerformed()
	{
		List<SimAction> actions = new LinkedList<SimAction>(actionsPerformed);
		deactivateActionMonitoring();
		
		return actions;
	}
	
	/**
	 * Returns the set of consecutive reachable nodes from a given one to the last node in the path.
	 * 
	 * @param nodeId Node identifier
	 * @param path Sequence of links (and maybe protection segments)
	 * @return Set of reachable nodes from the given one following the input path
	 * @since 0.3.0
	 */
	public Set<Long> getAvailableNodesFromNode(long nodeId, List<Long> path)
	{
		long layerId = getLayerDefaultId();
		return getAvailableNodesFromNode(layerId, nodeId, path);
	}

	/**
	 * Returns the set of consecutive reachable nodes from a given one to the last node in the path.
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @param path Sequence of links (and maybe protection segments)
	 * @return Set of reachable nodes from the given one following the input path
	 * @since 0.3.0
	 */
	public Set<Long> getAvailableNodesFromNode(long layerId, long nodeId, List<Long> path)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> availableNodesFromNode = new LinkedHashSet<Long>();
		ListIterator<Long> it = path.listIterator(path.size());
		long linkId;
		while(it.hasPrevious())
		{
			linkId = it.previous();
			long destinationNodeId = linkId >= 0 ? getLinkDestinationNode(layerId, linkId) : getProtectionSegmentDestinationNode(layerId, -1 - linkId);
			if (destinationNodeId == nodeId)
			{
				linkId = it.next(); /* (do not delete!) alternating calls to next and previous will return the same element repeatedly */
				while(it.hasNext())
				{
					linkId = it.next();
					long originNodeId = linkId >= 0 ? getLinkOriginNode(layerId, linkId) : getProtectionSegmentOriginNode(layerId, -1 - linkId);
					if (originNodeId != nodeId && !isNodeDown(nodeId)) availableNodesFromNode.add(originNodeId);
					
					if (!it.hasNext())
					{
						destinationNodeId = linkId >= 0 ? getLinkDestinationNode(layerId, linkId) : getProtectionSegmentDestinationNode(layerId, -1 - linkId);
						if (!isNodeDown(destinationNodeId) && destinationNodeId != nodeId) availableNodesFromNode.add(destinationNodeId);
					}
				}
				
				break;
			}
			
			if (!it.hasPrevious())
			{
				long originNodeId = linkId >= 0 ? getLinkOriginNode(layerId, linkId) : getProtectionSegmentOriginNode(layerId, -1 - linkId);
				if (originNodeId == nodeId)
				{
					do
					{
						destinationNodeId = linkId >= 0 ? getLinkDestinationNode(layerId, linkId) : getProtectionSegmentDestinationNode(layerId, -1 - linkId);
						if (destinationNodeId != nodeId && !isNodeDown(destinationNodeId)) availableNodesFromNode.add(destinationNodeId);

						if (it.hasNext()) linkId = it.next();
						else break;
					} while(true);
					
					break;
				}
				else
				{
					throw new Net2PlanException("Node not in path");
				}
			}
		}

		return availableNodesFromNode;
	}
	
	/**
	 * Returns the demand-link incidence matrix (a <i>D</i>x<i>E</i> matrix in 
	 * which an element <i>&delta;<sub>de</sub></i> is equal to the number of 
	 * times which traffic routes carrying traffic from demand <i>d</i> traverse 
	 * link <i>e</i>).
	 *
	 * @return The demand-link incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemand2LinkAssignmentMatrix()
	{
		long layerId = getLayerDefaultId();
		return getDemand2LinkAssignmentMatrix(layerId);
	}

	/**
	 * Returns the demand-link incidence matrix (a <i>D</i>x<i>E</i> matrix in 
	 * which an element <i>&delta;<sub>de</sub></i> is equal to the number of 
	 * times which traffic routes carrying traffic from demand <i>d</i> traverse 
	 * link <i>e</i>).
	 *
	 * @param layerId Layer identifier
	 * @return The demand-link incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemand2LinkAssignmentMatrix(long layerId)
	{
		DoubleMatrix2D delta_dr = getDemand2RouteAssignmentMatrix(layerId);
		DoubleMatrix2D delta_er = getLink2RouteAssignmentMatrix(layerId);
		
		return delta_dr.zMult(delta_er.viewDice(), null);
	}

	/**
	 * Returns the demand-route incidence matrix (a <i>D</i>x<i>R</i> matrix in 
	 * which an element <i>&delta;<sub>dr</sub></i> is equal to 1 if traffic 
	 * route <i>r</i> is able to carry traffic from demand <i>d</i>).
	 *
	 * @return The demand-route incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemand2RouteAssignmentMatrix()
	{
		long layerId = getLayerDefaultId();
		return getDemand2RouteAssignmentMatrix(layerId);
	}

	/**
	 * Returns the demand-route incidence matrix (a <i>D</i>x<i>R</i> matrix in 
	 * which an element <i>&delta;<sub>dr</sub></i> is equal to 1 if traffic 
	 * route <i>r</i> is able to carry traffic from demand <i>d</i>).
	 *
	 * @param layerId Layer identifier
	 * @return The demand-route incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemand2RouteAssignmentMatrix(long layerId)
	{
		int D = getNumberOfDemands(layerId);
		int R = getNumberOfRoutes(layerId);

		DoubleMatrix2D delta_dr = DoubleFactory2D.sparse.make(D, R);
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getDemandIds(layerId));
		Map<Long, Integer> routeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getRouteIds(layerId));
		for(Entry<Long, Long> entry : getRouteDemandMap(layerId).entrySet())
		{
			int d = demandId2LinearIndex.get(entry.getValue());
			int r = routeId2LinearIndex.get(entry.getKey());
			delta_dr.setQuick(d, r, 1);
		}
		
		return delta_dr;
	}

	/**
	 * Returns the value of a given attribute for a demand. If not defined,
	 * it is search for the default layer in the network.
	 *
	 * @param demandId Demand identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getDemandAttribute(long demandId, String key)
	{
		long layerId = getLayerDefaultId();
		return getDemandAttribute(layerId, demandId, key);
	}

	/**
	 * Returns the value of a given attribute for a demand. If not defined,
	 * it is search for the given layer.
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getDemandAttribute(long layerId, long demandId, String key)
	{
		checkDemandIsActive(layerId, demandId);
		
		String value = demandAttributesMap.get(layerId).get(demandId).get(key);
		return value == null ? getLayerAttribute(layerId, key) : value;
	}

	
	/**
	 * <p>Returns the attributes defined for a given demand. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param demandId Demand identifier
	 * @return Demand attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getDemandAttributeMap(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandAttributeMap(layerId, demandId);
	}
	
	/**
	 * <p>Returns the attributes defined for a given demand. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Demand attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getDemandAttributeMap(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return Collections.unmodifiableMap(demandAttributesMap.get(layerId).get(demandId));
	}
	
	/**
	 * <p>Returns the number of routes, in up state, carrying traffic for a given demand.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param demandId Demand identifier
	 * @return Number of routes carrying traffic for a given demand
	 * @since 0.3.0
	 */
	public int getDemandBifurcationDegree(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandBifurcationDegree(layerId, demandId);
	}

	/**
	 * <p>Returns the number of routes, in up state, carrying traffic for a given demand.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Number of routes carrying traffic for a given demand
	 * @since 0.3.0
	 */
	public int getDemandBifurcationDegree(long layerId, long demandId)
	{
		int bifurcationDegree = 0;
		Set<Long> routeIds_thisDemand = getDemandRoutes(layerId, demandId);
		for(long routeId : routeIds_thisDemand)
			if (!isRouteDown(layerId, routeId) && getRouteCarriedTraffic(layerId, routeId) > 0)
				bifurcationDegree++;

		return bifurcationDegree;
	}
	
	/**
	 * <p>Returns the bifurcation degree per demand.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Bifurcation degree per demand
	 * @since 0.3.0
	 */
	public Map<Long, Integer> getDemandBifurcationDegreeMap()
	{
		long layerId = getLayerDefaultId();
		return getDemandBifurcationDegreeMap(layerId);
	}

	/**
	 * <p>Returns the bifurcation degree per demand.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Bifurcation degree per demand
	 * @since 0.3.0
	 */
	public Map<Long, Integer> getDemandBifurcationDegreeMap(long layerId)
	{
		Map<Long, Integer> bifurcationDegree_d = new LinkedHashMap<Long, Integer>();
		Set<Long> demandIds_thisLayer = getDemandIds(layerId);
		for (long demandId : demandIds_thisLayer)
			bifurcationDegree_d.put(demandId, getDemandBifurcationDegree(layerId, demandId));
		
		return bifurcationDegree_d;
	}
	
	/**
	 * <p>Returns the bifurcation degree per demand, in increasing order of demand identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Bifurcation degree per demand
	 * @since 0.3.0
	 */
	public int[] getDemandBifurcationDegreeVector()
	{
		long layerId = getLayerDefaultId();
		return getDemandBifurcationDegreeVector(layerId);
	}

	/**
	 * <p>Returns the bifurcation degree per demand, in increasing order of demand identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Bifurcation degree per demand
	 * @since 0.3.0
	 */
	public int[] getDemandBifurcationDegreeVector(long layerId)
	{
		return IntUtils.toArray(getDemandBifurcationDegreeMap(layerId).values());
	}
	
	/**
	 * Returns the blocked traffic of a demand.
	 * 
	 * @param demandId Demand identifier
	 * @return Blocked traffic
	 * @since 0.3.0
	 */
	public double getDemandBlockedTraffic(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandBlockedTraffic(layerId, demandId);
	}

	/**
	 * Returns the blocked traffic of a demand.
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Blocked traffic
	 * @since 0.3.0
	 */
	public double getDemandBlockedTraffic(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return cache_demandBlockedTrafficMap.get(layerId).get(demandId);
	}
	
	/**
	 * <p>Returns the blocked traffic per demand. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Blocked traffic per demand
	 * @since 0.3.0
	 */
	public Map<Long, Double> getDemandBlockedTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getDemandBlockedTrafficMap(layerId);
	}

	/**
	 * <p>Returns the blocked traffic per demand. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Blocked traffic per demand
	 * @since 0.3.0
	 */
	public Map<Long, Double> getDemandBlockedTrafficMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(cache_demandBlockedTrafficMap.get(layerId));
	}
	
	/**
	 * Returns the blocked traffic per demand, in increasing order of demand identifier.
	 * 
	 * @return Blocked traffic per demand
	 * @since 0.3.0
	 */
	public double[] getDemandBlockedTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getDemandBlockedTrafficVector(layerId);
	}

	/**
	 * Returns the blocked traffic per demand, in increasing order of demand identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Blocked traffic per demand
	 * @since 0.3.0
	 */
	public double[] getDemandBlockedTrafficVector(long layerId)
	{
		return DoubleUtils.toArray(getDemandBlockedTrafficMap(layerId).values());
	}
	
	/**
	 * Returns the identifier of the first demand with a given value in a 
	 * certain attribute, or -1 if not found.
	 * 
	 * @param attribute Attribute (not null)
	 * @param value Value (not null)
	 * @return Identifier of the first demand with a given value in a certain attribute, or -1 if not found
	 * @since 0.3.0
	 */
	public long getDemandByAttribute(String attribute, String value)
	{
		long layerId = getLayerDefaultId();
		return getDemandByAttribute(layerId, attribute, value);
	}

	/**
	 * Returns the identifier of the first demand with a given value in a 
	 * certain attribute, or -1 if not found.
	 * 
	 * @param layerId Layer identifier
	 * @param attribute Attribute (not null)
	 * @param value Value (not null)
	 * @return Identifier of the first demand with a given value in a certain attribute, or -1 if not found
	 * @since 0.3.0
	 */
	public long getDemandByAttribute(long layerId, String attribute, String value)
	{
		if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

		Set<Long> demandIds_thisNetPlan = getDemandIds(layerId);
		for(long demandId : demandIds_thisNetPlan)
		{
			Map<String, String> attributeMap_thisDemand = getDemandAttributeMap(layerId, demandId);
			if (value.equals(attributeMap_thisDemand.get(attribute)))
				return demandId;
		}

		return -1;
	}

	/**
	 * Returns the carried traffic for a given demand.
	 *
	 * @param demandId Demand identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getDemandCarriedTraffic(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandCarriedTraffic(layerId, demandId);
	}

	/**
	 * Returns the carried traffic for a given demand.
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getDemandCarriedTraffic(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return cache_demandCarriedTrafficMap.get(layerId).get(demandId);
	}

	/**
	 * <p>Returns the carried traffic per demand. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Carried traffic per demand
	 * @since 0.3.0
	 */
	public Map<Long, Double> getDemandCarriedTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getDemandCarriedTrafficMap(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic per demand. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per demand
	 * @since 0.3.0
	 */
	public Map<Long, Double> getDemandCarriedTrafficMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(cache_demandCarriedTrafficMap.get(layerId));
	}
	
	/**
	 * Returns the carried traffic per demand, in increasing order of demand identifier.
	 * 
	 * @return Carried traffic per demand
	 * @since 0.2.0
	 */
	public double[] getDemandCarriedTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getDemandCarriedTrafficVector(layerId);
	}
	
	/**
	 * Returns the carried traffic per demand, in increasing order of demand identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per demand
	 * @since 0.3.0
	 */
	public double[] getDemandCarriedTrafficVector(long layerId)
	{
		checkLayerIsActive(layerId);
		return DoubleUtils.toArray(getDemandCarriedTrafficMap(layerId).values());
	}
	
	/**
	 * Returns the upper layer link coupled to the demand. If not coupled, then an exception will be thrown.
	 * 
	 * @param demandId Demand identifier
	 * @return Upper layer and link identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getDemandCoupledUpperLayerLink(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandCoupledUpperLayerLink(layerId, demandId);
	}

	/**
	 * Returns the upper layer link coupled to the demand. If not coupled, then an exception will be thrown.
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Upper layer and link identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getDemandCoupledUpperLayerLink(long layerId, long demandId)
	{
		if (isDemandCoupled(layerId, demandId)) return lowerLayerToUpperLayerMapping.get(layerId, demandId);
		else throw new Net2PlanException("Demand " + demandId + " at layer " + layerId + " is not associated to any upper layer link");
	}
	
	/**
	 * Returns the identifier of the egress node of the given demand.
	 *
	 * @param demandId Demand identifier
	 * @return Egress node identifier
	 * @since 0.3.0
	 */
	public long getDemandEgressNode(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandEgressNode(layerId, demandId);
	}

	/**
	 * Returns the identifier of the egress node of the given demand.
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Egress node identifier
	 * @since 0.3.0
	 */
	public long getDemandEgressNode(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return demandMap.get(layerId).get(demandId).getSecond();
	}

	/**
	 * <p>Returns the forwarding rules and splitting ratios associated to a given demand. 
	 * </p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandId Demand identifier
	 * @return Splitting ratio for each forwarding rule associated to the demand
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getDemandForwardingRuleSplittingRatioMap(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandForwardingRuleSplittingRatioMap(layerId, demandId);
	}
	
	/**
	 * <p>Returns the forwarding rules and splitting ratios associated to a given demand. 
	 * </p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Splitting ratio for each forwarding rule associated to the demand
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getDemandForwardingRuleSplittingRatioMap(long layerId, long demandId)
	{
		Map<Pair<Long, Long>, Double> splittingRatioMap = new LinkedHashMap<Pair<Long, Long>, Double>();
		for(Pair<Long, Long> forwardingRule : getDemandForwardingRules(layerId, demandId))
			splittingRatioMap.put(forwardingRule, getForwardingRuleSplittingRatio(layerId, forwardingRule));
		
		return splittingRatioMap;
	}
	
	/**
	 * <p>Returns the splitting ratio vector (fractions of traffic entering a node from 
	 * the given demand, leaving that node through link 'e'), in increasing order of link identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandId Demand identifier
	 * @return Splitting ratio matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemandForwardingRuleSplittingRatioMatrix(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandForwardingRuleSplittingRatioMatrix(layerId, demandId);
	}
	
	/**
	 * <p>Returns the splitting ratio vector (fractions of traffic entering a node from 
	 * the given demand, leaving that node through link 'e'), in increasing order of link identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Splitting ratio matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getDemandForwardingRuleSplittingRatioMatrix(long layerId, long demandId)
	{
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getLinkIds(layerId));
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getDemandIds(layerId));
		int E = getNumberOfLinks(layerId);
		int D = getNumberOfDemands(layerId);
		int d = demandId2LinearIndex.get(demandId);
		DoubleMatrix2D f_de = DoubleFactory2D.sparse.make(D, E);
		for(Entry<Pair<Long, Long>, Double> entry : getDemandForwardingRuleSplittingRatioMap(layerId, demandId).entrySet())
		{
			long linkId = entry.getKey().getSecond();
			int e = linkId2LinearIndex.get(linkId);
			f_de.setQuick(d, e, entry.getValue());
		}
		
		return f_de;
	}
	
	/**
	 * <p>Returns the set of forwarding rules associated to a demand. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandId Demand identifier
	 * @return Set of forwarding rules
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getDemandForwardingRules(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandForwardingRules(layerId, demandId);
	}
	
	/**
	 * <p>Returns the set of forwarding rules associated to a demand. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Set of forwarding rules
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getDemandForwardingRules(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		return Collections.unmodifiableSet(cache_demandForwardingRuleMap.get(layerId).get(demandId));
	}
	
	/**
	 * <p>Returns the set of identifiers of the demands. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getDemandIds()
	{
		long layerId = getLayerDefaultId();
		return getDemandIds(layerId);
	}

	/**
	 * <p>Returns the set of identifiers of the demands. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getDemandIds(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableSet(demandIds.get(layerId));
	}

	/**
	 * Returns the set of identifiers of the demands, in increasing order of identifier.
	 *
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public long[] getDemandIdsVector()
	{
		long layerId = getLayerDefaultId();
		return getDemandIdsVector(layerId);
	}

	/**
	 * Returns the set of identifiers of the demands, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public long[] getDemandIdsVector(long layerId)
	{
		return LongUtils.toArray(getDemandIds(layerId));
	}

	/**
	 * Returns the identifiers of the ingress and egress nodes.
	 *
	 * @param demandId Demand identifier
	 * @return Ingress and egress node identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getDemandIngressEgressNodePair(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandIngressEgressNodePair(layerId, demandId);
	}

	/**
	 * Returns the identifiers of the ingress and egress nodes.
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Ingress and egress node identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getDemandIngressEgressNodePair(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return demandMap.get(layerId).get(demandId);
	}
	
	/**
	 * Returns the identifier of the ingress node of the given demand.
	 *
	 * @param demandId Demand identifier
	 * @return Ingress node identifier
	 * @since 0.3.0
	 */
	public long getDemandIngressNode(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandIngressNode(layerId, demandId);
	}

	/**
	 * Returns the identifier of the ingress node of the given demand.
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Ingress node identifier
	 * @since 0.3.0
	 */
	public long getDemandIngressNode(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return demandMap.get(layerId).get(demandId).getFirst();
	}
	
	/**
	 * <p>Returns the ingress-egress node pair per demand. The output map is 
	 * backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Ingress-egress node pair per demand
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getDemandMap()
	{
		long layerId = getLayerDefaultId();
		return getDemandMap(layerId);
	}

	/**
	 * <p>Returns the ingress-egress node pair per demand. The output map is 
	 * backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Ingress-egress node pair per demand
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getDemandMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(demandMap.get(layerId));
	}
	
	/**
	 * Returns the next identifier for a new demand.
	 *
	 * @return Next demand identifier
	 * @since 0.3.0
	 */
	public long getDemandNextId()
	{
		long layerId = getLayerDefaultId();
		return getDemandNextId(layerId);
	}

	/**
	 * Returns the next identifier for a new demand.
	 *
	 * @param layerId Layer identifier
	 * @return Next demand identifier
	 * @since 0.3.0
	 */
	public long getDemandNextId(long layerId)
	{
		checkLayerIsActive(layerId);
		
		long demandId = nextDemandId.get(layerId);
		if (demandId < 0) throw new Net2PlanException(String.format(TEMPLATE_NO_MORE_DEMANDS_ALLOWED, layerId));
		
		return demandId;
	}
	
	/**
	 * Returns the offered traffic for a given demand.
	 *
	 * @param demandId Demand identifier
	 * @return Offered traffic
	 * @since 0.3.0
	 */
	public double getDemandOfferedTraffic(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandOfferedTraffic(layerId, demandId);
	}

	/**
	 * Returns the offered traffic for a given demand.
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Offered traffic
	 * @since 0.3.0
	 */
	public double getDemandOfferedTraffic(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return demandOfferedTrafficMap.get(layerId).get(demandId);
	}
	
	/**
	 * <p>Returns the offered traffic per demand. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Offered traffic per demand
	 * @since 0.3.0
	 */
	public Map<Long, Double> getDemandOfferedTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getDemandOfferedTrafficMap(layerId);
	}
	
	/**
	 * <p>Returns the offered traffic per demand. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Offered traffic per demand
	 * @since 0.3.0
	 */
	public Map<Long, Double> getDemandOfferedTrafficMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(demandOfferedTrafficMap.get(layerId));
	}
	
	/**
	 * Returns the offered traffic per demand, in increasing order of demand identifier.
	 * 
	 * @return Offered traffic per demand
	 * @since 0.2.0
	 */
	public double[] getDemandOfferedTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getDemandOfferedTrafficVector(layerId);
	}
	
	/**
	 * Returns the offered traffic per demand, in increasing order of demand identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Offered traffic per demand
	 * @since 0.3.0
	 */
	public double[] getDemandOfferedTrafficVector(long layerId)
	{
		checkLayerIsActive(layerId);
		return DoubleUtils.toArray(demandOfferedTrafficMap.get(layerId).values());
	}
	
	/**
	 * <p>Returns the routes which can carry traffic from a given demand. The output 
	 * set is backed in the {@code NetPlan} object, hence users should copy in 
	 * case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param demandId Demand identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getDemandRoutes(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandRoutes(layerId, demandId);
	}
	
	/**
	 * <p>Returns the routes which can carry traffic from a given demand. The output 
	 * set is backed in the {@code NetPlan} object, hence users should copy in 
	 * case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getDemandRoutes(long layerId, long demandId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		checkDemandIsActive(layerId, demandId);
		return Collections.unmodifiableSet(demandRouteIdsMap.get(layerId).get(demandId));
	}
	
	/**
	 * Returns the routing cycle type for a given demand.
	 * 
	 * @param demandId Demand identifier
	 * @return {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.0
	 */
	public RoutingCycleType getDemandRoutingCycleType(long demandId)
	{
		long layerId = getLayerDefaultId();
		return getDemandRoutingCycleType(layerId, demandId);
	}
	
	/**
	 * Returns the routing cycle type for a given demand.
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return {@link com.net2plan.utils.Constants.RoutingCycleType RoutingCycleType}
	 * @since 0.3.0
	 */
	public RoutingCycleType getDemandRoutingCycleType(long layerId, long demandId)
	{
		switch(getRoutingType(layerId))
		{
			case SOURCE_ROUTING:
				Set<Long> routeIds_thisLayer = getDemandRoutes(layerId, demandId);
				for(long routeId : routeIds_thisLayer)
				{
					RoutingCycleType routingCycleType_thisRoute = getRouteCycleType(layerId, routeId);
					if (routingCycleType_thisRoute != RoutingCycleType.LOOPLESS) return routingCycleType_thisRoute;
				}
				
				return RoutingCycleType.LOOPLESS;
				
			case HOP_BY_HOP_ROUTING:
				checkDemandIsActive(layerId, demandId);
				return cache_demandRoutingCycleType.get(layerId).get(demandId);
				
			default:
				throw new RuntimeException("Bad");
		}
	}
	
	/**
	 * Returns the total blocked traffic, summing up for all demands.
	 * 
	 * @return Total blocked traffic
	 * @since 0.3.0
	 */
	public double getDemandTotalBlockedTraffic()
	{
		long layerId = getLayerDefaultId();
		return getDemandTotalBlockedTraffic(layerId);
	}

	/**
	 * Returns the total blocked traffic, summing up for all demands.
	 * 
	 * @param layerId Layer identifier
	 * @return Total blocked traffic
	 * @since 0.3.0
	 */
	public double getDemandTotalBlockedTraffic(long layerId)
	{
		return DoubleUtils.sum(getDemandBlockedTrafficMap(layerId));
	}
	
	/**
	 * Returns the total carried traffic, summing up for all demands.
	 * 
	 * @return Total carried traffic
	 * @since 0.3.0
	 */
	public double getDemandTotalCarriedTraffic()
	{
		long layerId = getLayerDefaultId();
		return getDemandTotalCarriedTraffic(layerId);
	}

	/**
	 * Returns the total carried traffic, summing up for all demands.
	 * 
	 * @param layerId Layer identifier
	 * @return Total carried traffic
	 * @since 0.3.0
	 */
	public double getDemandTotalCarriedTraffic(long layerId)
	{
		return DoubleUtils.sum(getDemandCarriedTrafficMap(layerId));
	}
	
	/**
	 * Returns the total offered traffic, summing up for all demands.
	 * 
	 * @return Total offered traffic
	 * @since 0.3.0
	 */
	public double getDemandTotalOfferedTraffic()
	{
		long layerId = getLayerDefaultId();
		return getDemandTotalOfferedTraffic(layerId);
	}
	
	/**
	 * Returns the total offered traffic, summing up for all demands.
	 * 
	 * @param layerId Layer identifier
	 * @return Total offered traffic
	 * @since 0.3.0
	 */
	public double getDemandTotalOfferedTraffic(long layerId)
	{
		return DoubleUtils.sum(getDemandOfferedTrafficMap(layerId));
	}
	
	/**
	 * Returns the demand traffic units name.
	 * 
	 * @return Demand traffic units name
	 * @since 0.3.0
	 */
	public String getDemandTrafficUnitsName()
	{
		long layerId = getLayerDefaultId();
		return getDemandTrafficUnitsName(layerId);
	}

	/**
	 * Returns the demand traffic units name.
	 * 
	 * @param layerId Layer identifier
	 * @return Demand traffic units name
	 * @since 0.3.0
	 */
	public String getDemandTrafficUnitsName(long layerId)
	{
		checkLayerIsActive(layerId);
		return layerDemandTrafficUnitsNameMap.get(layerId);
	}
	
	/**
	 * Returns the value of the same attribute for each demand.
	 *
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getDemandsAttributeMap(String key)
	{
		long layerId = getLayerDefaultId();
		return getDemandsAttributeMap(layerId, key);
	}
	
	/**
	 * Returns the value of the same attribute for each demand.
	 *
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getDemandsAttributeMap(long layerId, String key)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, String> out = new LinkedHashMap<Long, String>();
		Set<Long> demandIds_thisLayer = getDemandIds(layerId);
		for(long demandId : demandIds_thisLayer)
			out.put(demandId, getDemandAttribute(layerId, demandId, key));
		
		return out;
	}
	
	/**
	 * <p>Returns the set of demands with some blocked traffic. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Blocked demands
	 * @since 0.3.0
	 */
	public Set<Long> getDemandsBlocked()
	{
		long layerId = getLayerDefaultId();
		return getDemandsBlocked(layerId);
	}
	
	/**
	 * <p>Returns the set of demands with some blocked traffic. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Blocked demands
	 * @since 0.3.0
	 */
	public Set<Long> getDemandsBlocked(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableSet(cache_demandBlockedIds.get(layerId));
	}

	/**
	 * Returns the demands with some blocked traffic, in increasing order of demand identifier.
	 * 
	 * @return Blocked demands
	 * @since 0.3.0
	 */
	public long[] getDemandsBlockedVector()
	{
		long layerId = getLayerDefaultId();
		return getDemandsBlockedVector(layerId);
	}
	
	/**
	 * Returns the demands with some blocked traffic, in increasing order of demand identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Blocked demands
	 * @since 0.3.0
	 */
	public long[] getDemandsBlockedVector(long layerId)
	{
		return LongUtils.toArray(getDemandsBlocked(layerId));
	}
	
    /**
     * Returns the identifiers of the demands with a given value in a certain attribute.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the demands with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getDemandsByAttribute(String attribute, String value)
    {
        long layerId = getLayerDefaultId();
        return getDemandsByAttribute(layerId, attribute, value);
    }

    /**
     * Returns the identifiers of the demands with a given value in a certain attribute.
     * 
     * @param layerId Layer identifier
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the demands with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getDemandsByAttribute(long layerId, String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> out = new LinkedHashSet<Long>();

        Set<Long> demandIds_thisNetPlan = getDemandIds(layerId);
        for(long demandId : demandIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisDemand = getDemandAttributeMap(layerId, demandId);
            if (value.equals(attributeMap_thisDemand.get(attribute)))
                out.add(demandId);
        }

        return out;
    }

	/**
	 * <p>Returns the upper layer link coupled to each demand (among the coupled ones). 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 * 
	 * @return Upper layer link coupled to each demand
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getDemandsCoupledLinkMap()
	{
		long layerId = getLayerDefaultId();
		return getDemandsCoupledLinkMap(layerId);
	}
	
	/**
	 * <p>Returns the upper layer link coupled to each demand (among the coupled ones). 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Upper layer link coupled to each demand
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getDemandsCoupledLinkMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(lowerLayerToUpperLayerMapping.row(layerId));
	}
	
	/**
	 * Returns the first node that is in up state after all failures in a path. In 
	 * case there is no failure, it will return the last node in the sequence. In 
	 * case the last node in the path is failed, it will return -1.
	 * 
	 * @param path Sequence of links and protection segments (not checked for continuity)
	 * @return First node in up state after finding a failed resource
	 * @since 0.3.0
	 */
	public long getFirstAvailableNodeAfterFailures(List<Long> path)
	{
		long layerId = getLayerDefaultId();
		return getFirstAvailableNodeAfterFailures(layerId, path);
	}
	
	/**
	 * Returns the first node that is in up state after all failures in a path. In 
	 * case there is no failure, it will return the last node in the sequence. In 
	 * case the last node in the path is failed, it will return -1.
	 * 
	 * @param layerId Layer identifier
	 * @param path Sequence of links and protection segments (not checked for continuity)
	 * @return First node in up state after finding a failed resource
	 * @since 0.3.0
	 */
	public long getFirstAvailableNodeAfterFailures(long layerId, List<Long> path)
	{
		checkLayerIsActive(layerId);
		
		long firstAvailableNodeAfterFailures = -1;
		ListIterator<Long> it = path.listIterator(path.size());
		while(it.hasPrevious())
		{
			long linkId = it.previous();
			long destinationNodeId = linkId >= 0 ? getLinkDestinationNode(layerId, linkId) : getProtectionSegmentDestinationNode(layerId, -1 - linkId);
			if (isNodeDown(destinationNodeId)) break;
			firstAvailableNodeAfterFailures = destinationNodeId;
			
			if (isLinkDown(layerId, linkId))
			{
				break;
			}
			else if (!it.hasPrevious())
			{
				long originNodeId = linkId >= 0 ? getLinkOriginNode(layerId, linkId) : getProtectionSegmentOriginNode(layerId, -1 - linkId);
				if (!isNodeDown(originNodeId)) firstAvailableNodeAfterFailures = originNodeId;
			}
		}
		
		return firstAvailableNodeAfterFailures;
	}
	
	/**
	 * Returns the first node that is in up state before any failed resource. In 
	 * case there is no failure, it will return the first node in the sequence. In 
	 * case the first node in the path is failed, it will return -1.
	 * 
	 * @param path Sequence of links and protection segments (not checked for continuity)
	 * @return First node in up state before finding a failed resource
	 * @since 0.3.0
	 */
	public long getFirstAvailableNodeBeforeFailures(List<Long> path)
	{
		long layerId = getLayerDefaultId();
		return getFirstAvailableNodeBeforeFailures(layerId, path);
	}
	
	/**
	 * Returns the first node that is in up state before any failed resource. In 
	 * case there is no failure, it will return the first node in the sequence. In 
	 * case the first node in the path is failed, it will return -1.
	 * 
	 * @param layerId Layer identifier
	 * @param path Sequence of links and protection segments (not checked for continuity)
	 * @return First node in up state before finding a failed resource
	 * @since 0.3.0
	 */
	public long getFirstAvailableNodeBeforeFailures(long layerId, List<Long> path)
	{
		checkLayerIsActive(layerId);
		
		long firstAvailableNodeBeforeFailures = -1;
		ListIterator<Long> it = path.listIterator();
		while(it.hasNext())
		{
			long linkId = it.next();
			long originNodeId = linkId >= 0 ? getLinkOriginNode(layerId, linkId) : getProtectionSegmentOriginNode(layerId, -1 - linkId);
			if (isNodeDown(originNodeId)) break;
			firstAvailableNodeBeforeFailures = originNodeId;
			
			if (isLinkDown(layerId, linkId))
			{
				break;
			}
			else if (!it.hasNext())
			{
				long destinationNodeId = linkId >= 0 ? getLinkDestinationNode(layerId, linkId) : getProtectionSegmentDestinationNode(layerId, -1 - linkId);
				if (!isNodeDown(destinationNodeId)) firstAvailableNodeBeforeFailures = destinationNodeId;
			}
		}
		
		return firstAvailableNodeBeforeFailures;
	}

	/**
	 * <p>Returns the value of a given attribute for a forwarding rule. If not defined,
	 * it is search for the default layer.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getForwardingRuleAttribute(Pair<Long, Long> demandLinkPair, String key)
	{
		long layerId = getLayerDefaultId();
		return getForwardingRuleAttribute(layerId, demandLinkPair, key);
	}

	/**
	 * <p>Returns the value of a given attribute for a forwarding rule. If not defined,
	 * it is search for the given layer.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getForwardingRuleAttribute(long layerId, Pair<Long, Long> demandLinkPair, String key)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		checkForwardingRuleIsActive(layerId, demandLinkPair);
		
		String value = forwardingRuleAttributeMap.get(layerId).get(demandLinkPair).get(key);
		return value == null ? getLayerAttribute(layerId, key) : value;
	}

	/**
	 * <p>Returns the attributes defined for a given forwarding rule. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return Forwarding rule attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getForwardingRuleAttributeMap(Pair<Long, Long> demandLinkPair)
	{
		long layerId = getLayerDefaultId();
		return getForwardingRuleAttributeMap(layerId, demandLinkPair);
	}
	
	/**
	 * <p>Returns the attributes defined for a given forwarding rule. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return Forwarding rule attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getForwardingRuleAttributeMap(long layerId, Pair<Long, Long> demandLinkPair)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		checkForwardingRuleIsActive(layerId, demandLinkPair);
		return Collections.unmodifiableMap(forwardingRuleAttributeMap.get(layerId).get(demandLinkPair));
	}
	
	/**
	 * <p>Returns the carried traffic using a forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getForwardingRuleCarriedTraffic(Pair<Long, Long> demandLinkPair)
	{
		long layerId = getLayerDefaultId();
		return getForwardingRuleCarriedTraffic(layerId, demandLinkPair);
	}
	
	/**
	 * <p>Returns the carried traffic using a forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getForwardingRuleCarriedTraffic(long layerId, Pair<Long, Long> demandLinkPair)
	{
		checkForwardingRuleIsActive(layerId, demandLinkPair);
		return forwardingRuleCarriedTrafficMap.get(layerId).get(demandLinkPair);
	}
	
	/**
	 * <p>Returns the carried traffic per forwarding rule. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @return Carried traffic per forwarding rule
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getForwardingRuleCarriedTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getForwardingRuleCarriedTrafficMap(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic per forwarding rule. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per forwarding rule
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getForwardingRuleCarriedTrafficMap(long layerId)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		return Collections.unmodifiableMap(forwardingRuleCarriedTrafficMap.get(layerId));
	}
	
	/**
	 * <p>Returns the splitting ratio associated to a forwarding rule (fractions of 
	 * traffic entering a node from demand 'd', leaving that node through link 'e').</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return Splitting ratio
	 * @since 0.3.0
	 */
	public double getForwardingRuleSplittingRatio(Pair<Long, Long> demandLinkPair)
	{
		long layerId = getLayerDefaultId();
		return getForwardingRuleSplittingRatio(layerId, demandLinkPair);
	}
	
	/**
	 * <p>Returns the splitting ratio associated to a forwarding rule (fractions of 
	 * traffic entering a node from demand 'd', leaving that node through link 'e').</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return Splitting ratio
	 * @since 0.3.0
	 */
	public double getForwardingRuleSplittingRatio(long layerId, Pair<Long, Long> demandLinkPair)
	{
		if (isForwardingRuleActive(layerId, demandLinkPair))
			return forwardingRuleSplittingRatioMap.get(layerId).get(demandLinkPair);
		else
			return 0.0;
	}
	
	/**
	 * <p>Returns the splitting ratio associated to each forwarding rule (fractions of 
	 * traffic entering a node from demand 'd', leaving that node through link 'e'). 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @return Splitting ratio map
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getForwardingRuleSplittingRatioMap()
	{
		long layerId = getLayerDefaultId();
		return getForwardingRuleSplittingRatioMap(layerId);
	}
	
	/**
	 * <p>Returns the splitting ratio associated to each forwarding rule (fractions of 
	 * traffic entering a node from demand 'd', leaving that node through link 'e'). 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Splitting ratio map
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getForwardingRuleSplittingRatioMap(long layerId)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		return Collections.unmodifiableMap(forwardingRuleSplittingRatioMap.get(layerId));
	}
	
	/**
	 * <p>Returns the splitting ratio matrix (fractions of traffic entering a node from 
	 * demand 'd', leaving that node through link 'e'). Rows and columns are in 
	 * increasing order of demand and link identifiers, respectively.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @return Splitting ratio matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getForwardingRuleSplittingRatioMatrix()
	{
		long layerId = getLayerDefaultId();
		return getForwardingRuleSplittingRatioMatrix(layerId);
	}
	
	/**
	 * <p>Returns the splitting ratio matrix (fractions of traffic entering a node from 
	 * demand 'd', leaving that node through link 'e'). Rows and columns are in 
	 * increasing order of demand and link identifiers, respectively.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Splitting ratio matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getForwardingRuleSplittingRatioMatrix(long layerId)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getLinkIds(layerId));
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getDemandIds(layerId));
		int E = getNumberOfLinks(layerId);
		int D = getNumberOfDemands(layerId);
		DoubleMatrix2D f_de = DoubleFactory2D.sparse.make(D, E);
		for(Entry<Pair<Long, Long>, Double> entry : forwardingRuleSplittingRatioMap.get(layerId).entrySet())
		{
			long demandId = entry.getKey().getFirst();
			long linkId = entry.getKey().getSecond();
			int d = demandId2LinearIndex.get(demandId);
			int e = linkId2LinearIndex.get(linkId);
			f_de.setQuick(d, e, entry.getValue());
		}
		
		return f_de;
	}
	
	/**
	 * <p>Returns the identifiers of active forwarding rules. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @return Forwarding rule identifiers
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getForwardingRules()
	{
		long layerId = getLayerDefaultId();
		return getForwardingRules(layerId);
	}
	
	/**
	 * <p>Returns the identifiers of active forwarding rules. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Forwarding rule identifiers
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getForwardingRules(long layerId)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		return Collections.unmodifiableSet(forwardingRuleDemandLinkPairMap.get(layerId));
	}
	
	/**
	 * <p>Returns the value of the same attribute for each forwarding rule.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, String> getForwardingRulesAttributeMap(String key)
	{
		long layerId = getLayerDefaultId();
		return getForwardingRulesAttributeMap(layerId, key);
	}
	
	/**
	 * <p>Returns the value of the same attribute for each forwarding rule.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, String> getForwardingRulesAttributeMap(long layerId, String key)
	{
		Map<Pair<Long, Long>, String> forwardingRulesAttributeMap = new LinkedHashMap<Pair<Long, Long>, String>();
		for(Pair<Long, Long> demandLinkPair : getForwardingRules(layerId))
			forwardingRulesAttributeMap.put(demandLinkPair, getForwardingRuleAttribute(layerId, demandLinkPair, key));
		
		return forwardingRulesAttributeMap;
	}
	
	/**
	 * <p>Returns the set of forwarding rules that are down. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @return Forwarding rule identifiers
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getForwardingRulesDown()
	{
		long layerId = getLayerDefaultId();
		return getForwardingRulesDown(layerId);
	}
	
	/**
	 * <p>Returns the set of forwarding rules that are down. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Forwarding rule identifiers
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getForwardingRulesDown(long layerId)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		return Collections.unmodifiableSet(forwardingRuleDemandLinkPairDownMap.get(layerId));
	}
	
	/**
	 * <p>Returns a copy of the current network design where only the given layer 
	 * is present.</p>
	 * 
	 * <p>Note that the returned network design is a deep copy of the current one, 
	 * so subsequent changes in any of them is not reflected into the other one.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Copy of the current network design where only the given layer is present
	 * @since 0.3.0
	 */
	public NetPlan getLayer(long layerId)
	{
		checkLayerIsActive(layerId);
		Set<Long> aux = LongUtils.setOf(layerId);
		NetPlan netPlan_thisLayer = new NetPlan();
		netPlan_thisLayer.copyFrom(this, aux);
		
		return netPlan_thisLayer;
	}
	
	/**
	 * Returns the value of a layer attribute.
	 *
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @return Attribute value
	 * @since 0.3.0
	 */
	public String getLayerAttribute(long layerId, String key)
	{
		checkLayerIsActive(layerId);
		String value = layerAttributeMap.get(layerId).get(key);
		return value == null ? getNetworkAttribute(key) : value;
	}

	/**
	 * Returns the layer attributes. The output map is backed in the {@code NetPlan}
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @return Attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getLayerAttributeMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(layerAttributeMap.get(layerId));
	}
	
	/**
	 * Returns the layer description.
	 *
	 * @param layerId Layer identifier
	 * @return Layer description
	 * @since 0.3.0
	 */
	public String getLayerDescription(long layerId)
	{
		checkLayerIsActive(layerId);
		return layerDescriptionMap.get(layerId);
	}

	/**
	 * Returns the identifier of the default layer.
	 *
	 * @return Layer identifier
	 * @since 0.3.0
	 */
	public long getLayerDefaultId()
	{
		return layerDefaultId;
	}
	
    /**
     * Returns the identifier of the first layer with a given value in a 
     * certain attribute, or -1 if not found.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first layer with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getLayerByAttribute(String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> layerIds_thisNetPlan = getLayerIds();
        for(long layerId : layerIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisLayer = getLayerAttributeMap(layerId);
            if (value.equals(attributeMap_thisLayer.get(attribute)))
                return layerId;
        }

        return -1;
    }

    /**
     * Returns the identifier of the first layer matching the given name, or -1 if 
     * not found.
     * 
     * @param name Layer name
     * @return Identifier of the first layer matching the given name, or -1 if not found
     * @since 0.3.0
     */
    public long getLayerByName(String name)
    {
		Set<Long> layerIds_thisNetPlan = getLayerIds();
		for(long layerId : layerIds_thisNetPlan)
			if (getLayerName(layerId).equals(name))
				return layerId;

		return -1;
    }

	/**
	 * <p>Returns the set of identifiers of the active layers. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Layer identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLayerIds()
	{
		return Collections.unmodifiableSet(layerIds);
	}

	/**
	 * Returns the set of identifiers of the active layers, in increasing order of identifier.
	 *
	 * @return Layer identifiers
	 * @since 0.3.0
	 */
	public long[] getLayerIdsVector()
	{
		return LongUtils.toArray(getLayerIds());
	}
	
	/**
	 * Returns the name of the given layer.
	 *
	 * @param layerId Layer identifier
	 * @return Layer name
	 * @since 0.3.0
	 */
	public String getLayerName(long layerId)
	{
		checkLayerIsActive(layerId);
		return layerNameMap.get(layerId);
	}
	
	/**
	 * <p>Returns the name for each layer. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Layer names
	 * @since 0.3.0
	 */
	public Map<Long, String> getLayerNameMap()
	{
		return Collections.unmodifiableMap(layerNameMap);
	}
	
	/**
	 * Returns the name for each layer, in increasing order of identifier.
	 *
	 * @return Layer names
	 * @since 0.3.0
	 */
	public String[] getLayerNameVector()
	{
		return StringUtils.toArray(layerNameMap.values());
	}
	
	/**
	 * Returns the next identifier for a new layer.
	 *
	 * @return Next layer identifier
	 * @since 0.3.0
	 */
	public long getLayerNextId()
	{
		long layerId = nextLayerId.longValue();
		if (layerId < 0) throw new Net2PlanException(TEMPLATE_NO_MORE_LAYERS_ALLOWED);
		
		return layerId;
	}
	
	/**
	 * Returns layer identifiers in bottom-up order, that is, starting from the 
	 * lower layers to the upper layers following coupling relationships. For layers 
	 * at the same hierarchical level, no order is guarantee. The output set is not 
	 * backed in the {@code NetPlan} object, hence changes will not affect it.
	 * 
	 * @return Layer identifiers in topological order
	 * @since 0.3.0
	 */
	public Set<Long> getLayerTopologicalSortIds()
	{
		Set<Long> layerIds_topologicalSort = new LinkedHashSet<Long>();
		Iterator<Long> it = interLayerCoupling.iterator();
		while(it.hasNext())	layerIds_topologicalSort.add(it.next());
		
		return layerIds_topologicalSort;
	}

	/**
	 * Returns layer identifiers in bottom-up order, that is, starting from the 
	 * lower layers to the upper layers following coupling relationships. For layers 
	 * at the same hierarchical level, no order is guarantee.
	 * 
	 * @return Layer identifiers in topological order
	 * @since 0.3.0
	 */
	public long[] getLayerTopologicalSortIdsVector()
	{
		return LongUtils.toArray(getLayerTopologicalSortIds());
	}
	
	/**
	 * Returns the value of the same attribute for each layer.
	 *
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getLayersAttributeMap(String key)
	{
		Map<Long, String> out = new LinkedHashMap<Long, String>();
		Set<Long> layerIds_thisNetPlan = getLayerIds();
		for(long layerId : layerIds_thisNetPlan)
		{
			out.put(layerId, getLayerAttribute(layerId, key));
		}
		
		return out;
	}

    /**
     * Returns the identifiers of the layers with a given value in a certain 
	 * attribute. The output set is not backed in the {@code NetPlan} object, 
	 * hence changes will not affect it.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the layers with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getLayersByAttribute(String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> out = new LinkedHashSet<Long>();

        Set<Long> layerIds_thisNetPlan = getLayerIds();
        for(long layerId : layerIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisLayer = getLayerAttributeMap(layerId);
            if (value.equals(attributeMap_thisLayer.get(attribute)))
                out.add(layerId);
        }

        return out;
    }

	/**
	 * Returns the link-route incidence matrix (an <i>E</i>x<i>R</i> matrix in 
	 * which an element <i>&delta;<sub>ep</sub></i> is equal to the number of 
	 * times which traffic route <i>r</i> traverses link <i>e</i>).
	 *
	 * @return The link-route incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getLink2RouteAssignmentMatrix()
	{
		long layerId = getLayerDefaultId();
		return getLink2RouteAssignmentMatrix(layerId);
	}

	/**
	 * Returns the link-route incidence matrix (an <i>E</i>x<i>R</i> matrix in 
	 * which an element <i>&delta;<sub>ep</sub></i> is equal to the number of 
	 * times which traffic route <i>r</i> traverses link <i>e</i>).
	 *
	 * @param layerId Layer identifier
	 * @return The link-route incidence matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getLink2RouteAssignmentMatrix(long layerId)
	{
		int E = getNumberOfLinks(layerId);
		int R = getNumberOfRoutes(layerId);

		DoubleMatrix2D delta_er = DoubleFactory2D.sparse.make(E, R);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getLinkIds(layerId));
		Map<Long, Integer> routeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(getRouteIds(layerId));
		for(Entry<Long, Integer> entry : routeId2LinearIndex.entrySet())
		{
			long routeId = entry.getKey();
			int r = entry.getValue();
			
			List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
			for(long linkId : seqLinks)
			{
				int e = linkId2LinearIndex.get(linkId);
				delta_er.setQuick(e, r, delta_er.getQuick(e, r) + 1);
			}
		}

		return delta_er;
	}

	/**
	 * <p>Returns the protection segments traversing a given link. The output set 
	 * is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param linkId Link identifier
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLinkAssociatedProtectionSegments(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkAssociatedProtectionSegments(layerId, linkId);
	}

	/**
	 * <p>Returns the protection segments traversing a given link. The output set 
	 * is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLinkAssociatedProtectionSegments(long layerId, long linkId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		checkLinkIsActive(layerId, linkId);
		return linkSegmentIds.get(layerId).get(linkId);
	}
	
	/**
	 * <p>Returns the routes traversing a given link. The output set 
	 * is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param linkId Link identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLinkAssociatedRoutes(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkAssociatedRoutes(layerId, linkId);
	}

	/**
	 * <p>Returns the routes traversing a given link. The output set 
	 * is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLinkAssociatedRoutes(long layerId, long linkId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		checkLinkIsActive(layerId, linkId);
		return linkRouteIds.get(layerId).get(linkId);
	}
	
	/**
	 * <p>Returns the value of a given attribute for a link. If not defined,
	 * it is search for the default layer in the network.</p>
	 *
	 * @param linkId Link identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getLinkAttribute(long linkId, String key)
	{
		long layerId = getLayerDefaultId();
		return getLinkAttribute(layerId, linkId, key);
	}

	/**
	 * <p>Returns the value of a given attribute for a link. If not defined,
	 * it is search for the given layer.</p>
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getLinkAttribute(long layerId, long linkId, String key)
	{
		checkLinkIsActive(layerId, linkId);
		String value = linkAttributesMap.get(layerId).get(linkId).get(key);
		return value == null ? getLayerAttribute(layerId, key) : value;
	}

	/**
	 * <p>Returns the attributes defined for a given link. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param linkId Link identifier
	 * @return Link attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getLinkAttributeMap(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkAttributeMap(layerId, linkId);
	}
	
	/**
	 * <p>Returns the attributes defined for a given link. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getLinkAttributeMap(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return Collections.unmodifiableMap(linkAttributesMap.get(layerId).get(linkId));
	}

    /**
     * Returns the identifier of the first link with a given value in a 
     * certain attribute, or -1 if not found.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first link with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getLinkByAttribute(String attribute, String value)
    {
        long layerId = getLayerDefaultId();
        return getLinkByAttribute(layerId, attribute, value);
    }

    /**
     * Returns the identifier of the first link with a given value in a 
     * certain attribute, or -1 if not found.
     * 
     * @param layerId Layer identifier
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first link with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getLinkByAttribute(long layerId, String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> linkIds_thisNetPlan = getLinkIds(layerId);
        for(long linkId : linkIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisLink = getLinkAttributeMap(layerId, linkId);
            if (value.equals(attributeMap_thisLink.get(attribute)))
                return linkId;
        }

        return -1;
    }

	/**
	 * Returns the capacity of the given link.
	 *
	 * @param linkId Link identifier
	 * @return Link capacity
	 * @since 0.3.0
	 */
	public double getLinkCapacity(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacity(layerId, linkId);
	}

	/**
	 * Returns the capacity of the given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link capacity
	 * @since 0.3.0
	 */
	public double getLinkCapacity(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkCapacityMap.get(layerId).get(linkId);
	}
	
	/**
	 * <p>Returns the link capacity per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Link capacities
	 * @since 0.2.0
	 */
	public Map<Long, Double> getLinkCapacityMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityMap(layerId);
	}

	/**
	 * <p>Returns the capacity per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Link capacities
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkCapacityMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(linkCapacityMap.get(layerId));
	}

	/**
	 * Returns the capacity for each link, in increasing order of identifier.
	 *
	 * @return Link capacities
	 * @since 0.2.0
	 */
	public double[] getLinkCapacityVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityVector(layerId);
	}

	/**
	 * Returns the capacity for each link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Link capacities
	 * @since 0.3.0
	 */
	public double[] getLinkCapacityVector(long layerId)
	{
		checkLayerIsActive(layerId);
		return DoubleUtils.toArray(linkCapacityMap.get(layerId).values());
	}

	/**
	 * Returns the capacity not reserved for protection of a given link.
	 *
	 * @param linkId Link identifier
	 * @return Link capacity not reserved for protection
	 * @since 0.3.0
	 */
	public double getLinkCapacityNotReservedForProtection(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityNotReservedForProtection(layerId, linkId);
	}

	/**
	 * Returns the capacity not reserved for protection of a given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link capacity not reserved for protection
	 * @since 0.3.0
	 */
	public double getLinkCapacityNotReservedForProtection(long layerId, long linkId)
	{
		double u_e = getLinkCapacity(layerId, linkId);
		double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
		
		return Math.max(u_e - r_e, 0);
	}
	
	/**
	 * <p>Returns the capacity not reserved for protection per link.</p>
	 * 
	 * @return Link capacities not reserved for protection
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkCapacityNotReservedForProtectionMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityNotReservedForProtectionMap(layerId);
	}
	
	/**
	 * <p>Returns the capacity not reserved for protection per link.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Link capacities not reserved for protection
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkCapacityNotReservedForProtectionMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> linkIds_thisLayer = getLinkIds();
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		for(long linkId : linkIds_thisLayer)
			out.put(linkId, getLinkCapacityNotReservedForProtection(layerId, linkId));
		
		return out;
	}
	
	/**
	 * Returns the capacity not reserved for protection per link, in increasing order of identifier.
	 *
	 * @return Link capacity vector reserved for protection
	 * @since 0.2.0
	 */
	public double[] getLinkCapacityNotReservedForProtectionVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityNotReservedForProtectionVector(layerId);
	}

	/**
	 * Returns the capacity not reserved for protection per link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Link capacity vector reserved for protection
	 * @since 0.3.0
	 */
	public double[] getLinkCapacityNotReservedForProtectionVector(long layerId)
	{
		return DoubleUtils.toArray(getLinkCapacityNotReservedForProtectionMap(layerId).values());
	}

	/**
	 * Returns the capacity reserved for protection of a link.
	 *
	 * @param linkId Link identifier
	 * @return Link capacity reserved for protection
	 * @since 0.3.0
	 */
	public double getLinkCapacityReservedForProtection(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityReservedForProtection(layerId, linkId);
	}

	/**
	 * Returns the capacity reserved for protection of a link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link capacity reserved for protection
	 * @since 0.3.0
	 */
	public double getLinkCapacityReservedForProtection(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkReservedCapacityMap.get(layerId).get(linkId);
	}
	
	/**
	 * <p>Returns the capacity reserved for protection per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Capacities reserved for protection per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkCapacityReservedForProtectionMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityReservedForProtectionMap(layerId);
	}
	
	/**
	 * <p>Returns the capacity reserved for protection per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Capacities reserved for protection per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkCapacityReservedForProtectionMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(linkReservedCapacityMap.get(layerId));
	}
	
	/**
	 * Returns the capacity reserved for protection per link, in increasing order of identifier.
	 *
	 * @return Link capacity vector reserved for protection
	 * @since 0.2.0
	 */
	public double[] getLinkCapacityReservedForProtectionVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityReservedForProtectionVector(layerId);
	}

	/**
	 * Returns the capacity reserved for protection per link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Link capacity vector reserved for protection
	 * @since 0.3.0
	 */
	public double[] getLinkCapacityReservedForProtectionVector(long layerId)
	{
		return DoubleUtils.toArray(getLinkCapacityReservedForProtectionMap(layerId).values());
	}

	/**
	 * Returns the link capacity units name.
	 * 
	 * @return Capacity units name
	 * @since 0.3.0
	 */
	public String getLinkCapacityUnitsName()
	{
		long layerId = getLayerDefaultId();
		return getLinkCapacityUnitsName(layerId);
	}
	
	/**
	 * Returns the link capacity units name.
	 * 
	 * @param layerId Layer identifier
	 * @return Capacity units name
	 * @since 0.3.0
	 */
	public String getLinkCapacityUnitsName(long layerId)
	{
		checkLayerIsActive(layerId);
		return layerLinkCapacityUnitsNameMap.get(layerId);
	}
	
	/**
	 * Returns the carried traffic by a given link.
	 *
	 * @param linkId Link identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getLinkCarriedTraffic(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkCarriedTraffic(layerId, linkId);
	}

	/**
	 * Returns the carried traffic by a given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getLinkCarriedTraffic(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return cache_linkCarriedTrafficMap.get(layerId).get(linkId);
	}
	
	/**
	 * <p>Returns the carried traffic per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Carried traffic per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkCarriedTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkCarriedTrafficMap(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkCarriedTrafficMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(cache_linkCarriedTrafficMap.get(layerId));
	}
	
	/**
	 * Returns the carried traffic per link, in increasing order of identifier.
	 *
	 * @return Carried traffic per link
	 * @since 0.2.0
	 */
	public double[] getLinkCarriedTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkCarriedTrafficVector(layerId);
	}

	/**
	 * Returns the carried traffic per link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Carried traffic per link
	 * @since 0.3.0
	 */
	public double[] getLinkCarriedTrafficVector(long layerId)
	{
		return DoubleUtils.toArray(getLinkCarriedTrafficMap(layerId).values());
	}
	
	/**
	 * Returns the lower layer demand coupled to the link. If not coupled, then an exception will be thrown.
	 * 
	 * @param linkId Link identifier
	 * @return Lower layer and demand identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getLinkCoupledLowerLayerDemand(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkCoupledLowerLayerDemand(layerId, linkId);
	}
	
	/**
	 * Returns the lower layer demand coupled to the link. If not coupled, then an exception will be thrown.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Lower layer and demand identifiers
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getLinkCoupledLowerLayerDemand(long layerId, long linkId)
	{
		if (isLinkCoupled(layerId, linkId)) return upperLayerToLowerLayerMapping.get(layerId, linkId);
		else throw new Net2PlanException("Link " + linkId + " at layer " + layerId + " is not associated to any lower layer demand");
	}
	
	/**
	 * Returns the identifier of the destination node of the given link.
	 *
	 * @param linkId Link identifier
	 * @return Destination node identifier
	 * @since 0.3.0
	 */
	public long getLinkDestinationNode(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkDestinationNode(layerId, linkId);
	}

	/**
	 * Returns the identifier of the destination node of the given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Destination node identifier
	 * @since 0.3.0
	 */
	public long getLinkDestinationNode(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkMap.get(layerId).get(linkId).getSecond();
	}
	
	/**
	 * <p>Returns the set of forwarding rules associated to a link. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param linkId Link identifier
	 * @return Set of forwarding rules
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getLinkForwardingRules(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkForwardingRules(layerId, linkId);
	}
	
	/**
	 * <p>Returns the set of forwarding rules associated to a link. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Set of forwarding rules
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getLinkForwardingRules(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		return Collections.unmodifiableSet(cache_linkForwardingRuleMap.get(layerId).get(linkId));
	}
	
	/**
	 * <p>Returns the set of identifiers of the links. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLinkIds()
	{
		long layerId = getLayerDefaultId();
		return getLinkIds(layerId);
	}
	
	/**
	 * <p>Returns the set of identifiers of the links. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getLinkIds(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableSet(linkIds.get(layerId));
	}
	
	/**
	 * Returns the set of identifiers of the links, in increasing order of identifier.
	 *
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public long[] getLinkIdsVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkIdsVector(layerId);
	}
	
	/**
	 * Returns the set of identifiers of the links, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public long[] getLinkIdsVector(long layerId)
	{
		return LongUtils.toArray(getLinkIds(layerId));
	}
	
	/**
	 * Returns the length in km of a link.
	 *
	 * @param linkId Link identifier
	 * @return Link length (measured in km)
	 * @since 0.3.0
	 */
	public double getLinkLengthInKm(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkLengthInKm(layerId, linkId);
	}

	/**
	 * Returns the length in km of a link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link length (measured in km)
	 * @since 0.3.0
	 */
	public double getLinkLengthInKm(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkLengthInKmMap.get(layerId).get(linkId);
	}
	
	/**
	 * <p>Returns the length in km per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Link lengths (measured in km)
	 * @since 0.2.0
	 */
	public Map<Long, Double> getLinkLengthInKmMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkLengthInKmMap(layerId);
	}

	/**
	 * <p>Returns the length in km per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Link lengths (measured in km)
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkLengthInKmMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(linkLengthInKmMap.get(layerId));
	}

	/**
	 * Returns the length in km for each link, in increasing order of identifier.
	 *
	 * @return Link lengths (measured in km)
	 * @since 0.2.0
	 */
	public double[] getLinkLengthInKmVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkLengthInKmVector(layerId);
	}

	/**
	 * Returns the length in km for each link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Link lengths (measured in km)
	 * @since 0.3.0
	 */
	public double[] getLinkLengthInKmVector(long layerId)
	{
		checkLayerIsActive(layerId);
		return DoubleUtils.toArray(linkLengthInKmMap.get(layerId).values());
	}

	/**
	 * <p>Returns the origin-destination node pair per link. The output map is 
	 * backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Origin-destination node pair per link
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getLinkMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkMap(layerId);
	}

	/**
	 * <p>Returns the origin-destination node pair per link. The output map is 
	 * backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Origin-destination node pair per link
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getLinkMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(linkMap.get(layerId));
	}
	
	/**
	 * Returns the maximum capacity among all links.
	 * 
	 * @return Maximum capacity value
	 * @since 0.3.0
	 */
	public double getLinkMaximumCapacity()
	{
		long layerId = getLayerDefaultId();
		return getLinkMaximumCapacity(layerId);
	}
	
	/**
	 * Returns the maximum capacity among all links.
	 * 
	 * @param layerId Layer identifier
	 * @return Maximum capacity value
	 * @since 0.3.0
	 */
	public double getLinkMaximumCapacity(long layerId)
	{
		return DoubleUtils.maxValue(getLinkCapacityMap(layerId));
	}
	
	/**
	 * Returns the maximum link utilization among all the links.
	 *
	 * @return Maximum link utilization
	 * @since 0.2.0
	 */
	public double getLinkMaximumUtilization()
	{
		long layerId = getLayerDefaultId();
		return getLinkMaximumUtilization(layerId);
	}

	/**
	 * Returns the maximum link utilization among all the links.
	 *
	 * @param layerId Layer identifier
	 * @return Maximum link utilization
	 * @since 0.3.0
	 */
	public double getLinkMaximumUtilization(long layerId)
	{
		return hasLinks(layerId) ? DoubleUtils.maxValue(getLinkUtilizationVector(layerId)) : 0;
	}

	/**
	 * Returns the maximum link utilization among all links (excluding reserved 
	 * capacity for protection).
	 *
	 * @return Maximum link utilization
	 * @since 0.2.0
	 */
	public double getLinkMaximumUtilizationWithoutConsiderReservedCapacityForProtection()
	{
		long layerId = getLayerDefaultId();
		return getLinkMaximumUtilizationWithoutConsiderReservedCapacityForProtection(layerId);
	}

	/**
	 * Returns the maximum link utilization among all links (excluding reserved 
	 * capacity for protection).
	 *
	 * @param layerId Layer identifier
	 * @return Maximum link utilization
	 * @since 0.3.0
	 */
	public double getLinkMaximumUtilizationWithoutConsiderReservedCapacityForProtection(long layerId)
	{
		return hasLinks(layerId) ? DoubleUtils.maxValue(getLinkUtilizationWithoutConsiderReservedCapacityForProtectionVector(layerId)) : 0;
	}
	
	/**
	 * Returns the next identifier for a new link.
	 *
	 * @return Next link identifier
	 * @since 0.3.0
	 */
	public long getLinkNextId()
	{
		long layerId = getLayerDefaultId();
		return getLinkNextId(layerId);
	}

	/**
	 * Returns the next identifier for a new link.
	 *
	 * @param layerId Layer identifier
	 * @return Next link identifier
	 * @since 0.3.0
	 */
	public long getLinkNextId(long layerId)
	{
		checkLayerIsActive(layerId);
		long linkId = nextLinkId.get(layerId);
		if (linkId < 0) throw new Net2PlanException(String.format(TEMPLATE_NO_MORE_LINKS_ALLOWED, layerId));
		
		return linkId;
	}

	/**
	 * Returns the occupied capacity for each link (excluding protection segments).
	 *
	 * @param linkId Link identifier
	 * @return Occupied capacity
	 * @since 0.2.0
	 */
	public double getLinkOccupiedCapacity(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkOccupiedCapacity(layerId, linkId);
	}
	
	/**
	 * Returns the occupied capacity for each link (excluding protection segments).
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Occupied capacity
	 * @since 0.3.0
	 */
	public double getLinkOccupiedCapacity(long layerId, long linkId)
	{
		if (isLinkDown(layerId, linkId)) return 0;
		
		double y_e = getLinkCarriedTraffic(layerId, linkId);
		double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
		
		return y_e + r_e;
	}

	/**
	 * <p>Returns the occupied capacity per link.</p>
	 *
	 * @return Occupied capacity per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkOccupiedCapacityMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkOccupiedCapacityMap(layerId);
	}
	
	/**
	 * <p>Returns the occupied capacity per link. </p>
	 *
	 * @param layerId Layer identifier
	 * @return Occupied capacity per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkOccupiedCapacityMap(long layerId)
	{
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		for(long linkId : linkIds_thisLayer) out.put(linkId, getLinkOccupiedCapacity(layerId, linkId));
		
		return out;
	}

	/**
	 * Returns the occupied capacity per link, in increasing order of identifier.
	 *
	 * @return Capacity available vector
	 * @since 0.2.0
	 */
	public double[] getLinkOccupiedCapacityVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkOccupiedCapacityVector(layerId);
	}
	
	/**
	 * Returns the occupied capacity per link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Occupied capacity per link
	 * @since 0.3.0
	 */
	public double[] getLinkOccupiedCapacityVector(long layerId)
	{
		checkLayerIsActive(layerId);
		return DoubleUtils.toArray(getLinkOccupiedCapacityMap(layerId).values());
	}
	
	/**
	 * Returns the origin-destination node pair for a given link.
	 *
	 * @param linkId Link identifier
	 * @return Origin-destination node pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getLinkOriginDestinationNodePair(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkOriginDestinationNodePair(layerId, linkId);
	}

	/**
	 * Returns the origin-destination node pair for a given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Origin-destination node pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getLinkOriginDestinationNodePair(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkMap.get(layerId).get(linkId);
	}
	
	/**
	 * Returns the identifier of the origin node of a given link.
	 *
	 * @param linkId Link identifier
	 * @return Origin node identifier
	 * @since 0.3.0
	 */
	public long getLinkOriginNode(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkOriginNode(layerId, linkId);
	}

	/**
	 * Returns the identifier of the origin node of a given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Origin node identifier
	 * @since 0.3.0
	 */
	public long getLinkOriginNode(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkMap.get(layerId).get(linkId).getFirst();
	}
	
	/**
	 * Returns the propagation delay in seconds of a given link.
	 *
	 * @param linkId Link identifier
	 * @return Propagation delay
	 * @since 0.3.0
	 */
	public double getLinkPropagationDelayInSeconds(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkPropagationDelayInSeconds(layerId, linkId);
	}

	/**
	 * Returns the propagation delay in seconds of a given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Propagation delay
	 * @since 0.3.0
	 */
	public double getLinkPropagationDelayInSeconds(long layerId, long linkId)
	{
		double l_e = getLinkLengthInKm(layerId, linkId);
		double v_prop = getLinkPropagationSpeedInKmPerSecond(layerId, linkId);
		return v_prop == 0 ? 0 : l_e / v_prop;
	}

	/**
	 * <p>Returns the propagation delay in seconds per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Propagation delays
	 * @since 0.2.0
	 */
	public Map<Long, Double> getLinkPropagationDelayInSecondsMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkPropagationDelayInSecondsMap(layerId);
	}

	/**
	 * <p>Returns the propagation delay in seconds per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Propagation delays
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkPropagationDelayInSecondsMap(long layerId)
	{
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		for(long linkId : linkIds_thisLayer) out.put(linkId, getLinkPropagationDelayInSeconds(layerId, linkId));
		
		return out;
	}
	
	/**
	 * Returns the propagation delay in seconds per link, in increasing order of identifier.
	 *
	 * @return Propagation delay vector
	 * @since 0.2.0
	 */
	public double[] getLinkPropagationDelayInSecondsVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkPropagationDelayInSecondsVector(layerId);
	}

	/**
	 * Returns the propagation delay in seconds per link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Propagation delay vector
	 * @since 0.3.0
	 */
	public double[] getLinkPropagationDelayInSecondsVector(long layerId)
	{
		return DoubleUtils.toArray(getLinkPropagationDelayInSecondsMap(layerId).values());
	}

	/**
	 * Returns the propagation speed in km/s of a given link.
	 * 
	 * @param linkId Link identifier
	 * @return Propagation speed in km/s
	 * @since 0.3.0
	 */
	public double getLinkPropagationSpeedInKmPerSecond(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkPropagationSpeedInKmPerSecond(layerId, linkId);
	}

	/**
	 * Returns the propagation speed in km/s of a given link.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Propagation speed in km/s
	 * @since 0.3.0
	 */
	public double getLinkPropagationSpeedInKmPerSecond(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkPropagationSpeedInKmPerSecondMap.get(layerId).get(linkId);
	}
	
	/**
	 * <p>Returns the propagation speed in km/s per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @return Propagation speeds
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkPropagationSpeedInKmPerSecondMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkPropagationSpeedInKmPerSecondMap(layerId);
	}
	
	/**
	 * <p>Returns the propagation speed in km/s per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Propagation speeds
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkPropagationSpeedInKmPerSecondMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(linkPropagationSpeedInKmPerSecondMap.get(layerId));
	}

	/**
	 * Returns the propagation speed in km/s per link, in increasing order of identifier.
	 * 
	 * @return Propagation speeds
	 * @since 0.3.0
	 */
	public double[] getLinkPropagationSpeedInKmPerSecondVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkPropagationSpeedInKmPerSecondVector(layerId);
	}
	
	/**
	 * Returns the propagation speed in km/s per link, in increasing order of identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Propagation speeds
	 * @since 0.3.0
	 */
	public double[] getLinkPropagationSpeedInKmPerSecondVector(long layerId)
	{
		return DoubleUtils.toArray(getLinkPropagationSpeedInKmPerSecondMap(layerId).values());
	}
	
	/**
	 * Returns all SRGs associated to a given link. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @param linkId Link identifier
	 * @return SRGs associated to the link
	 * @since 0.3.0
	 */
	public Set<Long> getLinkSRGs(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkSRGs(layerId, linkId);
	}

	/**
	 * Returns all SRGs associated to a given link. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return SRGs associated to the link
	 * @since 0.3.0
	 */
	public Set<Long> getLinkSRGs(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linkSRGIds.get(layerId).get(linkId);
	}
	
	/**
	 * Returns all SRGs associated to each link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @return SRGs associated to each link
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getLinkSRGsMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkSRGsMap(layerId);
	}

	/**
	 * Returns all SRGs associated to each link. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @return SRGs associated to each link
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getLinkSRGsMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		for (long linkId : linkIds_thisLayer)
		{
			out.put(linkId, getLinkSRGs(layerId, linkId));
		}

		return out;
	}
	
	/**
	 * Returns the available capacity in a link.
	 *
	 * @param linkId Link identifier
	 * @return Available capacity
	 * @since 0.3.0
	 */
	public double getLinkSpareCapacity(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkSpareCapacity(layerId, linkId);
	}
	
	/**
	 * Returns the available capacity in a link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Available capacity
	 * @since 0.3.0
	 */
	public double getLinkSpareCapacity(long layerId, long linkId)
	{
		if (isLinkDown(layerId, linkId)) return 0;
		
		double u_e = getLinkCapacity(layerId, linkId);
		double y_e = getLinkCarriedTraffic(layerId, linkId);
		double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
		
		return Math.max(u_e - y_e - r_e, 0);
	}

	/**
	 * Returns the available capacity per link.
	 * 
	 * @return Available capacities
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkSpareCapacityMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkSpareCapacityMap(layerId);
	}
	
	/**
	 * Returns the available capacity per link.
	 * 
	 * @param layerId Layer identifier
	 * @return Available capacities
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkSpareCapacityMap(long layerId)
	{
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		for(long linkId : linkIds_thisLayer) out.put(linkId, getLinkSpareCapacity(layerId, linkId));
		
		return out;
	}

	/**
	 * Returns the available capacity per link, in increasing order of identifier.
	 *
	 * @return Available capacity vector
	 * @since 0.2.0
	 */
	public double[] getLinkSpareCapacityVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkSpareCapacityVector(layerId);
	}
	
	/**
	 * Returns the available capacity per link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Available capacity vector
	 * @since 0.3.0
	 */
	public double[] getLinkSpareCapacityVector(long layerId)
	{
		checkLayerIsActive(layerId);
		return DoubleUtils.toArray(getLinkSpareCapacityMap(layerId).values());
	}
	
	/**
	 * Returns the total installed capacity.
	 * 
	 * @return Total installed capacity
	 * @since 0.3.0
	 */
	public double getLinkTotalCapacity()
	{
		long layerId = getLayerDefaultId();
		return getLinkTotalCapacity(layerId);
	}

	/**
	 * Returns the total installed capacity.
	 * 
	 * @param layerId Layer identifier
	 * @return Total installed capacity
	 * @since 0.3.0
	 */
	public double getLinkTotalCapacity(long layerId)
	{
		return DoubleUtils.sum(getLinkCapacityMap(layerId));
	}
	
	/**
	 * Returns the total carried traffic by links (excluding protection segments).
	 * 
	 * @return Total carried traffic by links
	 * @since 0.3.0
	 */
	public double getLinkTotalCarriedTraffic()
	{
		long layerId = getLayerDefaultId();
		return getLinkTotalCarriedTraffic(layerId);
	}

	/**
	 * Returns the total carried traffic by links (excluding protection segments).
	 * 
	 * @param layerId Layer identifier
	 * @return Total carried traffic by links
	 * @since 0.3.0
	 */
	public double getLinkTotalCarriedTraffic(long layerId)
	{
		return DoubleUtils.sum(getLinkCarriedTrafficMap(layerId));
	}
	
	/**
	 * Returns the utilization of a given link.
	 *
	 * @param linkId Link identifier
	 * @return Link utilization
	 * @since 0.3.0
	 */
	public double getLinkUtilization(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkUtilization(layerId, linkId);
	}

	/**
	 * Returns the utilization of a given link.
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link utilization
	 * @since 0.3.0
	 */
	public double getLinkUtilization(long layerId, long linkId)
	{
		double u_e = getLinkCapacity(layerId, linkId);
		double y_e = getLinkCarriedTraffic(layerId, linkId) + getLinkCapacityReservedForProtection(layerId, linkId);
		return y_e == 0 ? 0 : Math.max(y_e / u_e, 0);
	}
	
	/**
	 * Returns the utilization per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @return Utilization per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkUtilizationMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkUtilizationMap(layerId);
	}

	/**
	 * Returns the utilization per link. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @return Utilization per link
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkUtilizationMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		for(long linkId : linkIds_thisLayer)
			out.put(linkId, getLinkUtilization(layerId, linkId));
		
		return out;
	}

	/**
	 * Returns the utilization per link, in increasing order of identifier.
	 *
	 * @return Link utilization vector
	 * @since 0.2.0
	 */
	public double[] getLinkUtilizationVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkUtilizationVector(layerId);
	}

	/**
	 * Returns the utilization per link, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Link utilization vector
	 * @since 0.2.0
	 */
	public double[] getLinkUtilizationVector(long layerId)
	{
		return DoubleUtils.toArray(getLinkUtilizationMap(layerId).values());
	}
	
	/**
	 * Returns the utilization of a link (excluding protection segments).
	 *
	 * @param linkId Link identifier
	 * @return Link utilization without reserved capacity
	 * @since 0.3.0
	 */
	public double getLinkUtilizationWithoutConsiderReservedCapacityForProtection(long linkId)
	{
		long layerId = getLayerDefaultId();
		return getLinkUtilizationWithoutConsiderReservedCapacityForProtection(layerId, linkId);
	}

	/**
	 * Returns the utilization of a link (excluding protection segments).
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return Link utilization without reserved capacity
	 * @since 0.3.0
	 */
	public double getLinkUtilizationWithoutConsiderReservedCapacityForProtection(long layerId, long linkId)
	{
		double u_e = getLinkCapacity(layerId, linkId);
		double y_e = getLinkCarriedTraffic(layerId, linkId);
		return y_e == 0 ? 0 : y_e / u_e;
	}

	/**
	 * Returns the utilization per link (excluding protection segments).
	 *
	 * @return Link utilizations without reserved capacity
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkUtilizationWithoutConsiderReservedCapacityForProtectionMap()
	{
		long layerId = getLayerDefaultId();
		return getLinkUtilizationWithoutConsiderReservedCapacityForProtectionMap(layerId);
	}

	/**
	 * Returns the utilization per link (excluding protection segments).
	 *
	 * @param layerId Layer identifier
	 * @return Link utilizations without reserved capacity
	 * @since 0.3.0
	 */
	public Map<Long, Double> getLinkUtilizationWithoutConsiderReservedCapacityForProtectionMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		for(long linkId : linkIds_thisLayer)
			out.put(linkId, getLinkUtilizationWithoutConsiderReservedCapacityForProtection(layerId, linkId));
		
		return out;
	}
	
	/**
	 * Returns the utilization per link (excluding protection segments), in increasing order of identifier.
	 *
	 * @return Link utilizations without reserved capacity
	 * @since 0.2.0
	 */
	public double[] getLinkUtilizationWithoutConsiderReservedCapacityForProtectionVector()
	{
		long layerId = getLayerDefaultId();
		return getLinkUtilizationWithoutConsiderReservedCapacityForProtectionVector(layerId);
	}

	/**
	 * Returns the utilization per link (excluding protection segments), in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Link utilizations without reserved capacity
	 * @since 0.3.0
	 */
	public double[] getLinkUtilizationWithoutConsiderReservedCapacityForProtectionVector(long layerId)
	{
		return DoubleUtils.toArray(getLinkUtilizationWithoutConsiderReservedCapacityForProtectionMap(layerId).values());
	}
	
	/**
	 * Returns the value of an attribute for each demand at the first layer.
	 *
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.2.0
	 */
	public Map<Long, String> getLinksAttributeMap(String key)
	{
		long layerId = getLayerDefaultId();
		return getLinksAttributeMap(layerId, key);
	}

	/**
	 * Returns the value of an attribute for each demand at the given layer.
	 *
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getLinksAttributeMap(long layerId, String key)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, String> out = new LinkedHashMap<Long, String>();
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		for(long linkId : linkIds_thisLayer)
		{
			out.put(linkId, getLinkAttribute(layerId, linkId, key));
		}
		
		return out;
	}

    /**
     * Returns the identifiers of the links with a given value in a certain attribute.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the links with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getLinksByAttribute(String attribute, String value)
    {
        long layerId = getLayerDefaultId();
        return getLinksByAttribute(layerId, attribute, value);
    }

    /**
     * Returns the identifiers of the links with a given value in a certain attribute.
     * 
     * @param layerId Layer identifier
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the links with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getLinksByAttribute(long layerId, String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> out = new LinkedHashSet<Long>();

        Set<Long> linkIds_thisNetPlan = getLinkIds(layerId);
        for(long linkId : linkIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisLink = getLinkAttributeMap(layerId, linkId);
            if (value.equals(attributeMap_thisLink.get(attribute)))
                out.add(linkId);
        }

        return out;
    }

	/**
	 * <p>Returns the lower layer demand coupled to each link (among the coupled ones). 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 * 
	 * @return Lower layer demand coupled to each link
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getLinksCoupledDemandMap()
	{
		long layerId = getLayerDefaultId();
		return getLinksCoupledDemandMap(layerId);
	}
	
	/**
	 * <p>Returns the lower layer demand coupled to each link (among the coupled ones). 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Lower layer demand coupled to each link
	 * @since 0.3.0
	 */
	public Map<Long, Pair<Long, Long>> getLinksCoupledDemandMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(upperLayerToLowerLayerMapping.row(layerId));
	}
	
	/**
	 * Returns the set of links down. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @return Links down
	 */
	public Set<Long> getLinksDown()
	{
		long layerId = getLayerDefaultId();
		return getLinksDown(layerId);
	}
	
	/**
	 * Returns the set of links down. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @return Links down
	 * @since 0.3.0
	 */
	public Set<Long> getLinksDown(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableSet(linksDown.get(layerId));
	}
	
	/**
	 * Returns the set of links down per layer. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @return Links down per layer
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getLinksDownAllLayers()
	{
		return Collections.unmodifiableMap(linksDown);
	}
	
	/**
	 * Returns the set of oversubscribed links. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @return Oversubscribed links
	 * @since 0.3.0
	 */
	public Set<Long> getLinksOversubscribed()
	{
		long layerId = getLayerDefaultId();
		return getLinksOversubscribed(layerId);
	}
	
	/**
	 * Returns the set of oversubscribed links. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @return Oversubscribed links
	 * @since 0.3.0
	 */
	public Set<Long> getLinksOversubscribed(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableSet(cache_linksOverSubscribed.get(layerId));
	}
	
	/**
	 * Returns the set of links with a capacity equal to {@code Double.MAX_VALUE}. 
	 * The output set is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.
	 * 
	 * @return Links with unlimited capacity
	 * @since 0.3.0
	 */
	public Set<Long> getLinksWithUnlimitedCapacity()
	{
		long layerId = getLayerDefaultId();
		return getLinksWithUnlimitedCapacity(layerId);
	}
	
	/**
	 * Returns the set of links with a capacity equal to {@code Double.MAX_VALUE}. 
	 * The output set is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @return Links with unlimited capacity
	 * @since 0.3.0
	 */
	public Set<Long> getLinksWithUnlimitedCapacity(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableSet(cache_linksWithUnlimitedCapacity.get(layerId));
	}

	/**
	 * Returns the set of links with a capacity equal to zero. 
	 * The output set is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.
	 * 
	 * @return Links with zero capacity
	 * @since 0.3.0
	 */
	public Set<Long> getLinksWithZeroCapacity()
	{
		long layerId = getLayerDefaultId();
		return getLinksWithZeroCapacity(layerId);
	}
	
	/**
	 * Returns the set of links with a capacity equal to zero. 
	 * The output set is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @return Links with zero capacity
	 * @since 0.3.0
	 */
	public Set<Long> getLinksWithZeroCapacity(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableSet(cache_linksWithZeroCapacity.get(layerId));
	}
	
	/**
	 * Finds a new path after merging a {@code partialPath} into an {@code originalPath}, 
	 * both of them composed of links (and maybe protection segments). Note that the 
	 * {@code partialPath} must be contained into the {@code originalPath} in 
	 * downstream direction.
	 * 
	 * @param originalPath Original sequence of links and protection segments
	 * @param partialPath Original sequence of links and protection segments
	 * @return New path
	 * @since 0.3.0
	 */
	public List<Long> getMergedPath(List<Long> originalPath, List<Long> partialPath)
	{
		long layerId = getLayerDefaultId();
		return getMergedPath(layerId, originalPath, partialPath);
	}
	
	/**
	 * Finds a new path after merging a {@code partialPath} into an {@code originalPath}, 
	 * both of them composed of links (and maybe protection segments). Note that the 
	 * {@code partialPath} must be contained into the {@code originalPath} in 
	 * downstream direction.
	 * 
	 * @param layerId Layer identifier
	 * @param originalPath Original sequence of links and protection segments
	 * @param partialPath Original sequence of links and protection segments
	 * @return New path
	 * @since 0.3.0
	 */
	public List<Long> getMergedPath(long layerId, List<Long> originalPath, List<Long> partialPath)
	{
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = getLinkMap(layerId);
		List<Long> actualSeqLinks_originalRoute = convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, originalPath);
		List<Long> actualSeqLinks_partialRoute = convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, partialPath);
		GraphUtils.checkRouteContinuity(linkMap_thisLayer, actualSeqLinks_originalRoute, CheckRoutingCycleType.NO_CHECK);
		GraphUtils.checkRouteContinuity(linkMap_thisLayer, actualSeqLinks_partialRoute, CheckRoutingCycleType.NO_CHECK);

		long firstLinkId_partialRoute = partialPath.get(0);
		long lastLinkId_partialRoute = partialPath.get(partialPath.size() - 1);
		long originNodeId_partialRoute = firstLinkId_partialRoute >= 0 ? getLinkOriginNode(layerId, firstLinkId_partialRoute) : getProtectionSegmentOriginNode(layerId, -1 - firstLinkId_partialRoute);
		long destinationNodeId_partialRoute = lastLinkId_partialRoute >= 0 ? getLinkDestinationNode(layerId, lastLinkId_partialRoute) : getProtectionSegmentDestinationNode(layerId, -1 - lastLinkId_partialRoute);
		
		List<Long> mergedRoute = new LinkedList<Long>();
		ListIterator<Long> originalRoute_it = originalPath.listIterator();
		int originalRoute_headIndex = -1;
		while(originalRoute_it.hasNext())
		{
			int index = originalRoute_it.nextIndex();
			long linkId = originalRoute_it.next();
			if (linkId >= 0)
			{
				if (getLinkOriginNode(layerId, linkId) == originNodeId_partialRoute)
				{
					originalRoute_headIndex = index;
					break;
				}
			}
			else
			{
				if (getProtectionSegmentOriginNode(layerId, -1 - linkId) == originNodeId_partialRoute)
				{
					originalRoute_headIndex = index;
					break;
				}
			}
			
			mergedRoute.add(linkId);
		}
		
		if (originalRoute_headIndex == -1) throw new Net2PlanException("Origin node of the partial route is not in the original route");

		originalRoute_it = originalPath.listIterator(originalPath.size());
		int originalRoute_tailIndex = -1;
		while(originalRoute_it.hasPrevious())
		{
			int index = originalRoute_it.previousIndex();
			long linkId = originalRoute_it.previous();
			if (linkId >= 0)
			{
				if (getLinkDestinationNode(layerId, linkId) == destinationNodeId_partialRoute)
				{
					originalRoute_tailIndex = index;
					break;
				}
			}
			else
			{
				if (getProtectionSegmentDestinationNode(layerId, -1 - linkId) == destinationNodeId_partialRoute)
				{
					originalRoute_tailIndex = index;
					break;
				}
			}
		}
		
		if (originalRoute_tailIndex == -1) throw new Net2PlanException("Destination node of the partial route is not in the original route");
		if (originalRoute_tailIndex < originalRoute_headIndex) throw new Net2PlanException("Partial route goes upstream");
		
		mergedRoute.addAll(partialPath);

		originalRoute_it = originalPath.listIterator(originalRoute_tailIndex + 1);
		while(originalRoute_it.hasNext()) mergedRoute.add(originalRoute_it.next());
		return mergedRoute;
	}
	
	/**
	 * Returns the value of a network attribute.
	 *
	 * @param key Attribute name
	 * @return Attribute value
	 * @since 0.3.0
	 */
	public String getNetworkAttribute(String key)
	{
		return networkAttributes.get(key);
	}

	/**
	 * Returns the attributes of the network. The output map is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @return Attributes
	 * @since 0.2.0
	 */
	public Map<String, String> getNetworkAttributeMap()
	{
		return Collections.unmodifiableMap(networkAttributes);
	}

	/**
	 * Returns the description of the network.
	 *
	 * @return Network description
	 * @since 0.3.0
	 */
	public String getNetworkDescription()
	{
		return networkDescription;
	}
	
	/**
	 * Returns the name of the network.
	 *
	 * @return Network name
	 * @since 0.3.0
	 */
	public String getNetworkName()
	{
		return networkName;
	}
	
	/**
	 * Gets all the links which has a given node as origin or destination.
	 *
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedDemands(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeAssociatedDemands(layerId, nodeId);
	}

	/**
	 * Gets all the links which has a given node as origin or destination.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedDemands(long layerId, long nodeId)
	{
		Set<Long> demandSet = new TreeSet<Long>();
		demandSet.addAll(getNodeIncomingDemands(layerId, nodeId));
		demandSet.addAll(getNodeOutgoingDemands(layerId, nodeId));

		return demandSet;
	}
	
	/**
	 * Gets all the links which has a given node as origin or destination.
	 *
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedLinks(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeAssociatedLinks(layerId, nodeId);
	}

	/**
	 * Gets all the links which has a given node as origin or destination.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedLinks(long layerId, long nodeId)
	{
		Set<Long> linkSet = new TreeSet<Long>();
		linkSet.addAll(getNodeIncomingLinks(layerId, nodeId));
		linkSet.addAll(getNodeOutgoingLinks(layerId, nodeId));

		return linkSet;
	}
	
	/**
	 * <p>Returns the protection segments traversing a given node.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedProtectionSegments(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeAssociatedProtectionSegments(layerId, nodeId);
	}

	/**
	 * <p>Returns the protection segments traversing a given node.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedProtectionSegments(long layerId, long nodeId)
	{
		Set<Long> segmentIds_thisNode = new LinkedHashSet<Long>();
		Set<Long> traversingLinkIds = getNodeAssociatedLinks(layerId, nodeId);
		for(long linkId : traversingLinkIds)
			segmentIds_thisNode.addAll(getLinkAssociatedProtectionSegments(layerId, linkId));
		
		return new TreeSet<Long>(segmentIds_thisNode);
	}
	
	/**
	 * <p>Returns the routes traversing a given node (but not starting/ending there).</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedRoutes(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeAssociatedRoutes(layerId, nodeId);
	}

	/**
	 * <p>Returns the routes traversing a given node (but not starting/ending there).</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeAssociatedRoutes(long layerId, long nodeId)
	{
		Set<Long> routeIds_thisNode = new LinkedHashSet<Long>();
		Set<Long> traversingLinkIds = getNodeAssociatedLinks(layerId, nodeId);
		for(long linkId : traversingLinkIds)
			routeIds_thisNode.addAll(getLinkAssociatedRoutes(layerId, linkId));
		
		return new TreeSet<Long>(routeIds_thisNode);
	}
	
	/**
	 * Returns the value of a given attribute for a node. If not defined,
	 * it is search for the network.
	 *
	 * @param nodeId Node identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getNodeAttribute(long nodeId, String key)
	{
		checkNodeIsActive(nodeId);
		String value = nodeAttributes.get(nodeId).get(key);
		return value == null ? getNetworkAttribute(key) : value;
	}

	/**
	 * Returns the attributes defined for a given node. The output map is backed
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param nodeId Node identifier
	 * @return Node attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getNodeAttributeMap(long nodeId)
	{
		checkNodeIsActive(nodeId);
		return Collections.unmodifiableMap(nodeAttributes.get(nodeId));
	}

    /**
     * Returns the identifier of the first node with a given value in a 
     * certain attribute, or -1 if not found.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first node with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getNodeByAttribute(String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> nodeIds_thisNetPlan = getNodeIds();
        for(long nodeId : nodeIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisNode = getNodeAttributeMap(nodeId);
            if (value.equals(attributeMap_thisNode.get(attribute)))
                return nodeId;
        }

        return -1;
    }

    /**
     * Returns the identifier of the first node matching the given name, or -1 if 
     * not found.
     * 
     * @param name Node name
     * @return Identifier of the first node matching the given name, or -1 if not found
     * @since 0.3.0
     */
    public long getNodeByName(String name)
    {
        Map<Long, String> nodeNameMap = getNodeNameMap();
        List<Long> nodeIds_thisName = CollectionUtils.find(nodeNameMap, name, Constants.SearchType.FIRST);
        return nodeIds_thisName.isEmpty() ? -1 : nodeIds_thisName.get(0);
    }

	/**
	 * <p>Given a map of demands, it computes the adjacency matrix. This is a matrix 
	 * with as many rows and columns as nodes. Position (<i>i</i>,<i>j</i>) accounts 
	 * for the number of demands from node <i>i</i> (<i>i = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>i = 1</i> refers to the second one, 
	 * and so on) to node <i>j</i> (<i>j = 0</i> refers to the first node in 
	 * {@code NetPlan} object, <i>j = 1</i> refers to the second one, and so on).</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could 
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to 
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandAdjacencyMatrix().toArray();}</p>
	 * 
	 * @return Adjacency matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom/'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandAdjacencyMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeDemandAdjacencyMatrix(layerId);
	}
	
	/**
	 * <p>Given a map of demands, it computes the adjacency matrix. This is a matrix 
	 * with as many rows and columns as nodes. Position (<i>i</i>,<i>j</i>) accounts 
	 * for the number of demands from node <i>i</i> (<i>i = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>i = 1</i> refers to the second one, 
	 * and so on) to node <i>j</i> (<i>j = 0</i> refers to the first node in 
	 * {@code NetPlan} object, <i>j = 1</i> refers to the second one, and so on).</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could 
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to 
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandAdjacencyMatrix(layerId).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Adjacency matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom/'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandAdjacencyMatrix(long layerId)
	{
		return GraphUtils.getAdjacencyMatrix(getNodeIds(), getDemandMap(layerId));
	}
	
	/**
	 * <p>Returns the set of forwarding rules installed in a given node for a given demand.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @param demandId Demand identifier
	 * @return Set of demand-outgoing link pairs
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getNodeDemandForwardingRules(long nodeId, long demandId)
	{
		long layerId = getLayerDefaultId();
		return getNodeDemandForwardingRules(layerId, nodeId, demandId);
	}

	/**
	 * <p>Returns the set of forwarding rules installed in a given node for a given demand.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @param demandId Demand identifier
	 * @return Set of demand-outgoing link pairs
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getNodeDemandForwardingRules(long layerId, long nodeId, long demandId)
	{
		Set<Pair<Long, Long>> forwardingRules_thisNodeAndDemand = new LinkedHashSet<Pair<Long, Long>>();
		forwardingRules_thisNodeAndDemand.addAll(getNodeForwardingRules(layerId, nodeId));
		forwardingRules_thisNodeAndDemand.retainAll(getDemandForwardingRules(layerId, demandId));
		
		return forwardingRules_thisNodeAndDemand;
	}
	
	/**
	 * <p>Given a map of demands, it computes the node-demand incidence matrix. This 
	 * is a matrix with as many rows as nodes, and as many columns as demands. 
	 * Position (<i>n</i>, <i>d</i>) has a 1 if link <i>d</i> (<i>d = 0</i> refers
	 * to the first demand in {@code NetPlan} object, <i>d = 1</i> refers to the
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second
	 * one, and so on), -1 if it ends in node n, and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandIncidenceMatrix().toArray();}</p>
	 * 
	 * @return Incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandIncidenceMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeDemandIncidenceMatrix(layerId);
	}

	/**
	 * <p>Given a map of demands, it computes the node-demand incidence matrix. This 
	 * is a matrix with as many rows as nodes, and as many columns as demands. 
	 * Position (<i>n</i>, <i>d</i>) has a 1 if link <i>d</i> (<i>d = 0</i> refers
	 * to the first demand in {@code NetPlan} object, <i>d = 1</i> refers to the
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second
	 * one, and so on), -1 if it ends in node n, and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandIncidenceMatrix(layerId).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandIncidenceMatrix(long layerId)
	{
		return GraphUtils.getIncidenceMatrix(getNodeIds(), getDemandMap(layerId));
	}

	/**
	 * <p>Given a map of demands, it computes the node-demand incoming incidence 
	 * matrix. This is a matrix with as many rows as nodes, and as many columns
	 * as demands. Position (<i>n</i>, <i>d</i>) has a 1 if demand <i>d</i>
	 * (<i>d = 0</i> refers to the first demand in {@code NetPlan} object,
	 * <i>d = 1</i> refers to the second one, and so on) is terminated in node
	 * <i>n</i> (<i>n = 0</i> refers to the first node in {@code NetPlan} object,
	 * <i>n = 1</i> refers to the second one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandIncomingIncidenceMatrix().toArray();}</p>
	 * 
	 * @return Node-demand incoming incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandIncomingIncidenceMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeDemandIncomingIncidenceMatrix(layerId);
	}

	/**
	 * <p>Given a map of demands, it computes the node-demand incoming incidence 
	 * matrix. This is a matrix with as many rows as nodes, and as many columns
	 * as demands. Position (<i>n</i>, <i>d</i>) has a 1 if demand <i>d</i>
	 * (<i>d = 0</i> refers to the first demand in {@code NetPlan} object,
	 * <i>d = 1</i> refers to the second one, and so on) is terminated in node
	 * <i>n</i> (<i>n = 0</i> refers to the first node in {@code NetPlan} object,
	 * <i>n = 1</i> refers to the second one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandIncomingIncidenceMatrix(layerId).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Node-demand incoming incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandIncomingIncidenceMatrix(long layerId)
	{
		return GraphUtils.getIncomingIncidenceMatrix(getNodeIds(), getDemandMap(layerId));
	}

	/**
	 * <p>Given a map of demands, it computes the node-demand outgoing incidence matrix. 
	 * This is a matrix with as many rows as nodes, and as many columns as demands. 
	 * Position (<i>n</i>, <i>d</i>) has a 1 if demand <i>d</i> (<i>d = 0</i> refers 
	 * to the first demand in {@code demandMap.keySet()}, <i>d = 1</i> refers to the 
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second 
	 * one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandOutgoingIncidenceMatrix().toArray();}</p>
	 * 
	 * @return Node-demand outgoing incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandOutgoingIncidenceMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeDemandOutgoingIncidenceMatrix(layerId);
	}
	
	/**
	 * <p>Given a map of demands, it computes the node-demand outgoing incidence matrix. 
	 * This is a matrix with as many rows as nodes, and as many columns as demands. 
	 * Position (<i>n</i>, <i>d</i>) has a 1 if demand <i>d</i> (<i>d = 0</i> refers 
	 * to the first demand in {@code demandMap.keySet()}, <i>d = 1</i> refers to the 
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second 
	 * one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeDemandOutgoingIncidenceMatrix(layerId).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Node-demand outgoing incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeDemandOutgoingIncidenceMatrix(long layerId)
	{
		return GraphUtils.getOutgoingIncidenceMatrix(getNodeIds(), getDemandMap(layerId));
	}
	
	/**
	 * Returns the egress traffic of a node (in demand traffic units), that is, 
	 * the traffic leaving the network through that node.
	 * 
	 * @param nodeId Node identifier
	 * @return Egress traffic
	 * @since 0.3.0
	 */
	public double getNodeEgressTraffic(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeEgressTraffic(layerId, nodeId);
	}
	
	/**
	 * Returns the egress traffic of a node (in demand traffic units), that is, 
	 * the traffic leaving the network through that node.
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Egress traffic
	 * @since 0.3.0
	 */
	public double getNodeEgressTraffic(long layerId, long nodeId)
	{
		Set<Long> incomingDemandIds_thisNode = getNodeIncomingDemands(layerId, nodeId);
		double egressTraffic_thisNode = 0;
		for(long demandId : incomingDemandIds_thisNode)
			egressTraffic_thisNode += getDemandCarriedTraffic(layerId, demandId);
		
		return egressTraffic_thisNode;
	}
	
	/**
	 * Returns the egress traffic per node (in demand traffic units), that is, 
	 * the traffic leaving the network through each node.
	 * 
	 * @return Egress traffic per node
	 * @since 0.3.0
	 */
	public Map<Long, Double> getNodeEgressTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getNodeEgressTrafficMap(layerId);
	}

	/**
	 * Returns the egress traffic per node (in demand traffic units), that is, 
	 * the traffic leaving the network through each node.
	 * 
	 * @param layerId Layer identifier
	 * @return Egress traffic per node
	 * @since 0.3.0
	 */
	public Map<Long, Double> getNodeEgressTrafficMap(long layerId)
	{
		Map<Long, Double> egressTraffic_n = new LinkedHashMap<Long, Double>();
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		for(long nodeId : nodeIds_thisNetPlan) egressTraffic_n.put(nodeId, getNodeEgressTraffic(layerId, nodeId));
		
		return Collections.unmodifiableMap(egressTraffic_n);
	}
	
	/**
	 * Returns the egress traffic per node (in demand traffic units), that is, 
	 * the traffic leaving the network through each node, in increasing order of 
	 * identifier.
	 * 
	 * @return Egress traffic per node
	 * @since 0.3.0
	 */
	public double[] getNodeEgressTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getNodeEgressTrafficVector(layerId);
	}

	/**
	 * Returns the egress traffic per node (in demand traffic units), that is, 
	 * the traffic leaving the network through each node, in increasing order of 
	 * identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Egress traffic per node
	 * @since 0.3.0
	 */
	public double[] getNodeEgressTrafficVector(long layerId)
	{
		return DoubleUtils.toArray(getNodeEgressTrafficMap(layerId).values());
	}

	/**
	 * Returns the <i>N</i>x<i>N</i> Euclidean distance matrix (derived 
	 * from node coordinates), where <i>N</i> is the number of nodes within the network.
	 *
	 * @return Eucledian distance matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getNodeEuclideanDistanceMatrix()
	{
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds_thisNetPlan);

		int N = nodeIds_thisNetPlan.size();
		DoubleMatrix2D out = DoubleFactory2D.dense.make(N, N);
		for(long nodeId_1 : nodeIds_thisNetPlan)
		{
			int nodeIndex_1 = nodeId2LinearIndex.get(nodeId_1);
			for(long nodeId_2 : nodeIds_thisNetPlan)
			{
				if (nodeId_1 >= nodeId_2) continue;
				
				int nodeIndex_2 = nodeId2LinearIndex.get(nodeId_2);
				double physicalDistance = getNodePairEuclideanDistance(nodeId_1, nodeId_2);
				out.setQuick(nodeIndex_1, nodeIndex_2, physicalDistance);
				out.setQuick(nodeIndex_2, nodeIndex_1, physicalDistance);
			}
		}
		
		return out;
	}

	/**
	 * <p>Returns the set of forwarding rules installed in a given node.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @return Set of demand-outgoing link pairs
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getNodeForwardingRules(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeForwardingRules(layerId, nodeId);
	}
	
	/**
	 * <p>Returns the set of forwarding rules installed in a given node.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Set of demand-outgoing link pairs
	 * @since 0.3.0
	 */
	public Set<Pair<Long, Long>> getNodeForwardingRules(long layerId, long nodeId)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		checkNodeIsActive(nodeId);
		return Collections.unmodifiableSet(cache_nodeForwardingRuleMap.get(nodeId).get(layerId));
	}
	
	/**
	 * Returns the <i>N</i>x<i>N</i> Haversine distance matrix (derived 
	 * from node coordinates, where 'xCoord' is equal to longitude and 'yCoord' 
	 * is equal to latitude), where <i>N</i> is the number of nodes within the network.
	 *
	 * @return Haversine distance matrix
	 * @since 0.3.1
	 * @see <a href="http://www.movable-type.co.uk/scripts/latlong.html">Calculate distance, bearing and more between Latitude/Longitude points</a>
	 */
	public DoubleMatrix2D getNodeHaversineDistanceInKmMatrix()
	{
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds_thisNetPlan);

		int N = nodeIds_thisNetPlan.size();
		DoubleMatrix2D out = DoubleFactory2D.dense.make(N, N);
		for(long nodeId_1 : nodeIds_thisNetPlan)
		{
			int nodeIndex_1 = nodeId2LinearIndex.get(nodeId_1);
			for(long nodeId_2 : nodeIds_thisNetPlan)
			{
				if (nodeId_1 >= nodeId_2) continue;
				
				int nodeIndex_2 = nodeId2LinearIndex.get(nodeId_2);
				double physicalDistance = getNodePairHaversineDistanceInKm(nodeId_1, nodeId_2);
				out.setQuick(nodeIndex_1, nodeIndex_2, physicalDistance);
				out.setQuick(nodeIndex_2, nodeIndex_1, physicalDistance);
			}
		}
		
		return out;
	}
	
	/**
	 * Returns the set of identifiers of the active nodes. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @return Node identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeIds()
	{
		return Collections.unmodifiableSet(nodeIds);
	}

	/**
	 * Returns the set of identifiers of the active nodes, in increasing order of 
	 * identifier.
	 *
	 * @return Node identifiers
	 * @since 0.3.0
	 */
	public long[] getNodeIdsVector()
	{
		return LongUtils.toArray(getNodeIds());
	}
	
	/**
	 * Returns the set of nodes whose links are incoming to a given node.
	 *
	 * @param nodeId Node identifier
	 * @return Set of nodes reachable whose links are incoming to the node
	 * @since 0.3.0
	 */
	public Set<Long> getNodeInNeighbors(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeInNeighbors(layerId, nodeId);
	}
	
	/**
	 * Returns the set of nodes whose links are incoming to a given node.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Set of nodes reachable whose links are incoming to the node
	 * @since 0.3.0
	 */
	public Set<Long> getNodeInNeighbors(long layerId, long nodeId)
	{
		Set<Long> incomingLinks_thisNode = getNodeIncomingLinks(layerId, nodeId);
		Set<Long> inNeighbors = new TreeSet<Long>();
		for(long linkId : incomingLinks_thisNode)
			inNeighbors.add(getLinkOriginNode(layerId, linkId));
		
		return inNeighbors;
	}
	
	/**
	 * Returns the demands ending in a given node. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param nodeId Node identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeIncomingDemands(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeIncomingDemands(layerId, nodeId);
	}
	
	/**
	 * Returns the demands ending in a given node. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeIncomingDemands(long layerId, long nodeId)
	{
		checkLayerIsActive(layerId);
		checkNodeIsActive(nodeId);
		
		return Collections.unmodifiableSet(cache_nodeIncomingDemands.get(nodeId).get(layerId));
	}
	
	/**
	 * Returns the links ending in a given node. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeIncomingLinks(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeIncomingLinks(layerId, nodeId);
	}

	/**
	 * Returns the links ending in a given node. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeIncomingLinks(long layerId, long nodeId)
	{
		checkLayerIsActive(layerId);
		checkNodeIsActive(nodeId);
		
		return Collections.unmodifiableSet(cache_nodeIncomingLinks.get(nodeId).get(layerId));
	}

	/**
	 * <p>Returns the routes ending in a given node. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeIncomingRoutes(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeIncomingRoutes(layerId, nodeId);
	}
		
	/**
	 * <p>Returns the routes ending in a given node. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeIncomingRoutes(long layerId, long nodeId)
	{
		checkLayerIsActive(layerId);
		checkNodeIsActive(nodeId);
		
		Set<Long> out = new TreeSet<Long>();
		Set<Long> incomingDemandIds_thisNode = getNodeIncomingDemands(layerId, nodeId);
		for(long demandId : incomingDemandIds_thisNode) out.addAll(getDemandRoutes(layerId, demandId));
		
		return out;
	}
	
	/**
	 * Returns the ingress traffic of a node node (in demand traffic units), that is, 
	 * the traffic entering the network through that node.
	 * 
	 * @param nodeId Node identifier
	 * @return Ingress traffic
	 * @since 0.3.0
	 */
	public double getNodeIngressTraffic(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeIngressTraffic(layerId, nodeId);
	}
	
	/**
	 * Returns the ingress traffic of a node node (in demand traffic units), that is, 
	 * the traffic entering the network through that node.
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Ingress traffic
	 * @since 0.3.0
	 */
	public double getNodeIngressTraffic(long layerId, long nodeId)
	{
		Set<Long> outgoingDemandIds_thisNode = getNodeOutgoingDemands(layerId, nodeId);
		double ingressTraffic_thisNode = 0;
		for(long demandId : outgoingDemandIds_thisNode)
			ingressTraffic_thisNode += getDemandCarriedTraffic(layerId, demandId);
		
		return ingressTraffic_thisNode;
	}
	
	/**
	 * Returns the ingress traffic per node (in demand traffic units), that is, 
	 * the traffic entering the network through each node.
	 * 
	 * @return Ingress traffic per node
	 * @since 0.3.0
	 */
	public Map<Long, Double> getNodeIngressTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getNodeIngressTrafficMap(layerId);
	}

	/**
	 * Returns the ingress traffic per node (in demand traffic units), that is, 
	 * the traffic entering the network through each node.
	 * 
	 * @param layerId Layer identifier
	 * @return Ingress traffic per node
	 * @since 0.3.0
	 */
	public Map<Long, Double> getNodeIngressTrafficMap(long layerId)
	{
		Map<Long, Double> ingressTraffic_n = new LinkedHashMap<Long, Double>();
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		for(long nodeId : nodeIds_thisNetPlan) ingressTraffic_n.put(nodeId, getNodeIngressTraffic(layerId, nodeId));
		
		return Collections.unmodifiableMap(ingressTraffic_n);
	}

	/**
	 * Returns the ingress traffic per node (in demand traffic units), that is, 
	 * the traffic entering the network through each node, in increasing order of 
	 * identifier.
	 * 
	 * @return Ingress traffic per node
	 * @since 0.3.0
	 */
	public double[] getNodeIngressTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getNodeIngressTrafficVector(layerId);
	}

	/**
	 * Returns the ingress traffic per node (in demand traffic units), that is, 
	 * the traffic entering the network through each node, in increasing order of 
	 * identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Ingress traffic per node
	 * @since 0.3.0
	 */
	public double[] getNodeIngressTrafficVector(long layerId)
	{
		return DoubleUtils.toArray(getNodeIngressTrafficMap(layerId).values());
	}

	/**
	 * <p>Given a map of links, it computes the adjacency matrix. This is a matrix 
	 * with as many rows and columns as nodes. Position (<i>i</i>,<i>j</i>) accounts 
	 * for the number of links from node <i>i</i> (<i>i = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>i = 1</i> refers to the second one, 
	 * and so on) to node <i>j</i> (<i>j = 0</i> refers to the first node in 
	 * {@code NetPlan} object, <i>j = 1</i> refers to the second one, and so on).</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could 
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to 
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeLinkAdjacencyMatrix().toArray();}</p>
	 * 
	 * @return Adjacency matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom/'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkAdjacencyMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeLinkAdjacencyMatrix(layerId);
	}
	
	/**
	 * <p>Given a map of links, it computes the adjacency matrix. This is a matrix 
	 * with as many rows and columns as nodes. Position (<i>i</i>,<i>j</i>) accounts 
	 * for the number of links from node <i>i</i> (<i>i = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>i = 1</i> refers to the second one, 
	 * and so on) to node <i>j</i> (<i>j = 0</i> refers to the first node in 
	 * {@code NetPlan} object, <i>j = 1</i> refers to the second one, and so on).</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could 
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to 
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeLinkAdjacencyMatrix(layerId).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Adjacency matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom/'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkAdjacencyMatrix(long layerId)
	{
		return GraphUtils.getAdjacencyMatrix(getNodeIds(), getLinkMap(layerId));
	}
	
	/**
	 * <p>Given a map of links, it computes the node-link incidence matrix. This 
	 * is a matrix with as many rows as nodes, and as many columns as links. 
	 * Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers
	 * to the first link in {@code NetPlan} object, <i>e = 1</i> refers to the
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second
	 * one, and so on), -1 if it ends in node n, and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code getNodeLinkIncidenceMatrix} format, so that could
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getIncidenceMatrix(nodeIds, linkMap).toArray();}</p>
	 * 
	 * @return Incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkIncidenceMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeLinkIncidenceMatrix(layerId);
	}

	/**
	 * <p>Given a map of links, it computes the node-link incidence matrix. This 
	 * is a matrix with as many rows as nodes, and as many columns as links. 
	 * Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers
	 * to the first link in {@code NetPlan} object, <i>e = 1</i> refers to the
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second
	 * one, and so on), -1 if it ends in node n, and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could
	 * be directly used along with the <a href='#jom'>JOM</a> library in order to
	 * solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeLinkIncidenceMatrix(nodeIds, linkMap).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkIncidenceMatrix(long layerId)
	{
		return GraphUtils.getIncidenceMatrix(getNodeIds(), getLinkMap(layerId));
	}

	/**
	 * <p>Given a map of links, it computes the node-link incoming incidence 
	 * matrix. This is a matrix with as many rows as nodes, and as many columns
	 * as links. Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i>
	 * (<i>e = 0</i> refers to the first link in {@code NetPlan} object,
	 * <i>e = 1</i> refers to the second one, and so on) is terminated in node
	 * <i>n</i> (<i>n = 0</i> refers to the first node in {@code NetPlan} object,
	 * <i>n = 1</i> refers to the second one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeLinkIncomingIncidenceMatrix().toArray();}</p>
	 * 
	 * @return Node-link incoming incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkIncomingIncidenceMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeLinkIncomingIncidenceMatrix(layerId);
	}

	/**
	 * <p>Given a map of links, it computes the node-link incoming incidence 
	 * matrix. This is a matrix with as many rows as nodes, and as many columns
	 * as links. Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i>
	 * (<i>e = 0</i> refers to the first link in {@code NetPlan} object,
	 * <i>e = 1</i> refers to the second one, and so on) is terminated in node
	 * <i>n</i> (<i>n = 0</i> refers to the first node in {@code NetPlan} object,
	 * <i>n = 1</i> refers to the second one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeLinkIncomingIncidenceMatrix(layerId).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Node-link incoming incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkIncomingIncidenceMatrix(long layerId)
	{
		return GraphUtils.getIncomingIncidenceMatrix(getNodeIds(), getLinkMap(layerId));
	}

	/**
	 * <p>Given a map of links, it computes the node-link outgoing incidence matrix. 
	 * This is a matrix with as many rows as nodes, and as many columns as links. 
	 * Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers 
	 * to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the 
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second 
	 * one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeLinkOutgoingIncidenceMatrix().toArray();}</p>
	 * 
	 * @return Node-link outgoing incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkOutgoingIncidenceMatrix()
	{
		long layerId = getLayerDefaultId();
		return getNodeLinkOutgoingIncidenceMatrix(layerId);
	}
	
	/**
	 * <p>Given a map of links, it computes the node-link outgoing incidence matrix. 
	 * This is a matrix with as many rows as nodes, and as many columns as links. 
	 * Position (<i>n</i>, <i>e</i>) has a 1 if link <i>e</i> (<i>e = 0</i> refers 
	 * to the first link in {@code linkMap.keySet()}, <i>e = 1</i> refers to the 
	 * second one, and so on) is initiated in node <i>n</i> (<i>n = 0</i> refers 
	 * to the first node in {@code NetPlan} object, <i>n = 1</i> refers to the second 
	 * one, and so on), and 0 otherwise.</p>
	 * 
	 * <p>The output is in the sparse {@code DoubleMatrix2D} format, so that could be directly used along with the <a href='#jom'>JOM</a> library in order to solve optimization problems.</p>
	 *
	 * <p>For users not interested in this format, a classical dense {@code double[][]} matrix could be obtained via the command:</p>
	 *
	 * <p>{@code double[][] matrix = getNodeLinkOutgoingIncidenceMatrix(layerId).toArray();}</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Node-link outgoing incidence matrix
	 * @since 0.3.1
	 * @see <a href='http://www.net2plan.com/jom'>Java Optimization Modeler (JOM) website</a>
	 */
	public DoubleMatrix2D getNodeLinkOutgoingIncidenceMatrix(long layerId)
	{
		return GraphUtils.getOutgoingIncidenceMatrix(getNodeIds(), getLinkMap(layerId));
	}
	
	/**
	 * Returns the node name.
	 *
	 * @param nodeId Node identifier
	 * @return Node name
	 * @since 0.3.0
	 */
	public String getNodeName(long nodeId)
	{
		checkNodeIsActive(nodeId);
		return nodeNames.get(nodeId);
	}

	/**
	 * Returns the name for each node. The output map is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 *
	 * @return Node names
	 * @since 0.3.0
	 */
	public Map<Long, String> getNodeNameMap()
	{
		return Collections.unmodifiableMap(nodeNames);
	}
	
	/**
	 * Returns the name for each node, in increasing order of identifier.
	 *
	 * @return Node names
	 * @since 0.2.0
	 */
	public String[] getNodeNameVector()
	{
		return StringUtils.toArray(nodeNames.values());
	}
	
	/**
	 * Returns the set of nodes reachable from the outgoing links of a given node.
	 *
	 * @param nodeId Node identifier
	 * @return Set of nodes reachable from the outgoing links of the node
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutNeighbors(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeOutNeighbors(layerId, nodeId);
	}
	
	/**
	 * Returns the set of nodes reachable from the outgoing links of a given node.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Set of nodes reachable from the outgoing links of the node
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutNeighbors(long layerId, long nodeId)
	{
		Set<Long> outgoingLinks_thisNode = getNodeOutgoingLinks(layerId, nodeId);
		Set<Long> outNeighbors = new TreeSet<Long>();
		for(long linkId : outgoingLinks_thisNode)
			outNeighbors.add(getLinkDestinationNode(layerId, linkId));
		
		return outNeighbors;
	}
	
	/**
	 * Returns the next identifier for a new node.
	 *
	 * @return Next node identifier
	 * @since 0.3.0
	 */
	public long getNodeNextId()
	{
		long nodeId = nextNodeId.longValue();
		if (nodeId < 0) throw new Net2PlanException(TEMPLATE_NO_MORE_NODES_ALLOWED);
		
		return nodeId;
	}
	
	/**
	 * Returns the demands starting in a given node. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param nodeId Node identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutgoingDemands(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeOutgoingDemands(layerId, nodeId);
	}
	
	/**
	 * Returns the demands starting in a given node. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutgoingDemands(long layerId, long nodeId)
	{
		checkLayerIsActive(layerId);
		checkNodeIsActive(nodeId);
		
		return Collections.unmodifiableSet(cache_nodeOutgoingDemands.get(nodeId).get(layerId));
	}
	
	/**
	 * Returns the links starting in a given node. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutgoingLinks(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeOutgoingLinks(layerId, nodeId);
	}

	/**
	 * Returns the links starting in a given node. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutgoingLinks(long layerId, long nodeId)
	{
		checkLayerIsActive(layerId);
		checkNodeIsActive(nodeId);
		
		return Collections.unmodifiableSet(cache_nodeOutgoingLinks.get(nodeId).get(layerId));
	}

	/**
	 * <p>Returns the routes starting in a given node.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutgoingRoutes(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeOutgoingRoutes(layerId, nodeId);
	}
		
	/**
	 * <p>Returns the routes starting in a given node at the given layer.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeOutgoingRoutes(long layerId, long nodeId)
	{
		checkLayerIsActive(layerId);
		checkNodeIsActive(nodeId);
		
		Set<Long> out = new TreeSet<Long>();
		Set<Long> outgoingDemandIds_thisNode = getNodeOutgoingDemands(layerId, nodeId);
		for(long demandId : outgoingDemandIds_thisNode) out.addAll(getDemandRoutes(layerId, demandId));
		
		return out;
	}

	/**
	 * Returns the average offered traffic per node pair. It is computed as the 
	 * quotient between {@link #getDemandTotalOfferedTraffic() getDemandTotalOfferedTraffic()} 
	 * and {@link #getNumberOfNodePairs() getNumberOfNodePairs()}.
	 * 
	 * @return Aveverage offered traffic per node pair
	 * @since 0.3.0
	 */
	public double getNodePairAverageOfferedTraffic()
	{
		long layerId = getLayerDefaultId();
		return getNodePairAverageOfferedTraffic(layerId);
	}

	/**
	 * Returns the average offered traffic per node pair. It is computed as the 
	 * quotient between {@link #getDemandTotalOfferedTraffic(long) getDemandTotalOfferedTraffic(layerId)} 
	 * and {@link #getNumberOfNodePairs() getNumberOfNodePairs()}.
	 * 
	 * @param layerId Layer identifier
	 * @return Aveverage offered traffic per node pair
	 * @since 0.3.0
	 */
	public double getNodePairAverageOfferedTraffic(long layerId)
	{
		int nodePairs = getNumberOfNodePairs();
		double H_d = getDemandTotalOfferedTraffic(layerId);
		return nodePairs == 0 ? 0 : H_d / nodePairs;
	}
	
	/**
	 * Returns links between two nodes (in both directions).
	 *
	 * @param node1 Node 1
	 * @param node2 Node 2
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairBidirectionalLinks(long node1, long node2)
	{
		long layerId = getLayerDefaultId();
		return getNodePairBidirectionalLinks(layerId, node1, node2);
	}

	/**
	 * Returns links between two nodes (in both directions).
	 *
	 * @param layerId Layer identifier
	 * @param node1 Node 1
	 * @param node2 Node 2
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairBidirectionalLinks(long layerId, long node1, long node2)
	{
		Set<Long> downstreamLinks = getNodePairLinks(layerId, node1, node2);
		Set<Long> upstreamLinks = getNodePairLinks(layerId, node2, node1);
		Set<Long> out = new TreeSet<Long>(downstreamLinks);
		out.addAll(upstreamLinks);

		return out;
	}

	/**
	 * Returns demands between two nodes (in both directions).
	 *
	 * @param node1 Node 1
	 * @param node2 Node 2
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairBidirectionalDemands(long node1, long node2)
	{
		long layerId = getLayerDefaultId();
		return getNodePairBidirectionalDemands(layerId, node1, node2);
	}

	/**
	 * Returns demands between two nodes (in both directions).
	 *
	 * @param layerId Layer identifier
	 * @param node1 Node 1
	 * @param node2 Node 2
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairBidirectionalDemands(long layerId, long node1, long node2)
	{
		Set<Long> downstreamDemands = getNodePairDemands(layerId, node1, node2);
		Set<Long> upstreamDemands = getNodePairDemands(layerId, node2, node1);
		Set<Long> out = new TreeSet<Long>(downstreamDemands);
		out.addAll(upstreamDemands);

		return out;
	}
	
	/**
	 * Returns the demands from a node to other one.
	 *
	 * @param ingressNodeId Ingress node
	 * @param egressNodeId Egress node
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairDemands(long ingressNodeId, long egressNodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodePairDemands(layerId, ingressNodeId, egressNodeId);
	}
	
	/**
	 * Returns the demands from a node to other one.
	 *
	 * @param layerId Layer identifier
	 * @param ingressNodeId Ingress node
	 * @param egressNodeId Egress node
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairDemands(long layerId, long ingressNodeId, long egressNodeId)
	{
		Set<Long> demandSet = new TreeSet<Long>();
		demandSet.addAll(getNodeOutgoingDemands(layerId, ingressNodeId));
		demandSet.retainAll(getNodeIncomingDemands(layerId, egressNodeId));

		return demandSet;
	}
	
	/**
	 * Returns the links from a node to other one.
	 *
	 * @param originNodeId Origin node
	 * @param destinationNodeId Destination node
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairLinks(long originNodeId, long destinationNodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodePairLinks(layerId, originNodeId, destinationNodeId);
	}
	
	/**
	 * Returns the links from a node to other one.
	 *
	 * @param layerId Layer identifier
	 * @param originNodeId Origin node
	 * @param destinationNodeId Destination node
	 * @return Link identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairLinks(long layerId, long originNodeId, long destinationNodeId)
	{
		Set<Long> linkSet = new TreeSet<Long>();
		linkSet.addAll(getNodeOutgoingLinks(layerId, originNodeId));
		linkSet.retainAll(getNodeIncomingLinks(layerId, destinationNodeId));

		return linkSet;
	}
	
	/**
	 * Gets the Euclidean distance (derived from node coordinates, 
	 * where 'xCoord' is equal to longitude and 'yCoord' is equal to latitude) for a node pair.
	 *
	 * @param nodeId_1 First node
	 * @param nodeId_2 Second node
	 * @return Euclidean distance between the node pair
	 * @since 0.3.0
	 */
	public double getNodePairEuclideanDistance(long nodeId_1, long nodeId_2)
	{
		Point2D src = getNodeXYPosition(nodeId_1);
		Point2D dst = getNodeXYPosition(nodeId_2);
		return GraphUtils.computeEuclideanDistance(src, dst);
	}

	/**
	 * Gets the Haversine distance matrix (derived from node coordinates, 
	 * where 'xCoord' is equal to longitude and 'yCoord' is equal to latitude) for a node pair.
	 *
	 * @param nodeId_1 First node
	 * @param nodeId_2 Second node
	 * @return Haversine distance between the node pair
	 * @since 0.3.0
	 * @see <a href="http://www.movable-type.co.uk/scripts/latlong.html">Calculate distance, bearing and more between Latitude/Longitude points</a>
	 */
	public double getNodePairHaversineDistanceInKm(long nodeId_1, long nodeId_2)
	{
		Point2D src = getNodeXYPosition(nodeId_1);
		Point2D dst = getNodeXYPosition(nodeId_2);
		
		return GraphUtils.computeHaversineDistanceInKm(src, dst);
	}
	
	/**
	 * <p>Returns the routes from a node to other one.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param ingressNodeId Ingress node
	 * @param egressNodeId Egress node
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairRoutes(long ingressNodeId, long egressNodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodePairRoutes(layerId, ingressNodeId, egressNodeId);
	}
	
	/**
	 * <p>Returns the routes from a node to other one.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param ingressNodeId Ingress node
	 * @param egressNodeId Egress node
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodePairRoutes(long layerId, long ingressNodeId, long egressNodeId)
	{
		Set<Long> routeSet = new TreeSet<Long>();
		routeSet.addAll(getNodeOutgoingRoutes(layerId, ingressNodeId));
		routeSet.retainAll(getNodeIncomingRoutes(layerId, egressNodeId));

		return routeSet;
	}
	
	/**
	 * Returns all SRGs associated to a given node. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 * 
	 * @param nodeId Node identifier
	 * @return SRGs associated to the node
	 * @since 0.3.0
	 */
	public Set<Long> getNodeSRGs(long nodeId)
	{
		checkNodeIsActive(nodeId);
		return Collections.unmodifiableSet(cache_nodeSRGIds.get(nodeId));
	}
	
	/**
	 * Returns all SRGs associated to each node. The output map is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 * 
	 * @return SRGs associated to each node
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getNodeSRGsMap()
	{
		return Collections.unmodifiableMap(cache_nodeSRGIds);
	}
	
	/**
	 * <p>Returns the routes traversing a given node (but not starting/ending there).</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeTraversingRoutes(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeTraversingRoutes(layerId, nodeId);
	}

	/**
	 * <p>Returns the routes traversing a given node (but not starting/ending there).</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getNodeTraversingRoutes(long layerId, long nodeId)
	{
		Set<Long> routeIds_thisNode = new LinkedHashSet<Long>(getNodeAssociatedRoutes(layerId, nodeId));
		routeIds_thisNode.removeAll(getNodeIncomingRoutes(layerId, nodeId));
		routeIds_thisNode.removeAll(getNodeOutgoingRoutes(layerId, nodeId));
		
		return new TreeSet<Long>(routeIds_thisNode);
	}

	/**
	 * Returns the traversing traffic of a node (in demand traffic units), that is, 
	 * the processed traffic neither entering or leaving the network through that node.
	 * 
	 * @param nodeId Node identifier
	 * @return Traversing traffic
	 * @since 0.3.0
	 */
	public double getNodeTraversingTraffic(long nodeId)
	{
		long layerId = getLayerDefaultId();
		return getNodeTraversingTraffic(layerId, nodeId);
	}
	
	/**
	 * Returns the traversing traffic of a node (in demand traffic units), that is, 
	 * the processed traffic neither entering or leaving the network through that node.
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @return Traversing traffic
	 * @since 0.3.0
	 */
	public double getNodeTraversingTraffic(long layerId, long nodeId)
	{
		double traversingTraffic_thisNode = 0;

		RoutingType routingType = getRoutingType(layerId);
		switch(routingType)
		{
			case SOURCE_ROUTING:
				Set<Long> traversingRoutes_thisNode = getNodeTraversingRoutes(layerId, nodeId);
				for(long routeId : traversingRoutes_thisNode)
					if (!isRouteDown(layerId, routeId))
						traversingTraffic_thisNode += getRouteCarriedTraffic(layerId, routeId);
				break;
				
			case HOP_BY_HOP_ROUTING:
				Set<Pair<Long, Long>> forwardingRules_thisNode = getNodeForwardingRules(layerId, nodeId);
				for(Pair<Long, Long> demandLinkPair : forwardingRules_thisNode)
					traversingTraffic_thisNode += getForwardingRuleCarriedTraffic(layerId, demandLinkPair);
				traversingTraffic_thisNode -= getNodeIngressTraffic(layerId, nodeId);
				break;
		}

		return traversingTraffic_thisNode;
	}
	
	/**
	 * Returns the traversing traffic per node (in demand traffic units), that is, 
	 * the processed traffic neither entering or leaving the network through each node.
	 * 
	 * @return Traversing traffic per node
	 * @since 0.3.0
	 */
	public Map<Long, Double> getNodeTraversingTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getNodeTraversingTrafficMap(layerId);
	}

	/**
	 * Returns the traversing traffic per node (in demand traffic units), that is, 
	 * the processed traffic neither entering or leaving the network through each node.
	 * 
	 * @param layerId Layer identifier
	 * @return Traversing traffic per node
	 * @since 0.3.0
	 */
	public Map<Long, Double> getNodeTraversingTrafficMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, Double> traversingTraffic_n = new LinkedHashMap<Long, Double>();
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		for(long nodeId : nodeIds_thisNetPlan) traversingTraffic_n.put(nodeId, getNodeTraversingTraffic(layerId, nodeId));
		
		return Collections.unmodifiableMap(traversingTraffic_n);
	}

	/**
	 * Returns the traversing traffic per node (in demand traffic units), that is, 
	 * the processed traffic neither entering or leaving the network through each node, 
	 * in increasing order of identifier.
	 * 
	 * @return Traversing traffic per node
	 * @since 0.3.0
	 */
	public double[] getNodeTraversingTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getNodeTraversingTrafficVector(layerId);
	}
	
	/**
	 * Returns the traversing traffic per node (in demand traffic units), that is, 
	 * the processed traffic neither entering or leaving the network through each node, 
	 * in increasing order of identifier.
	 * 
	 * @param layerId Layer identifier
	 * @return Traversing traffic per node
	 * @since 0.3.0
	 */
	public double[] getNodeTraversingTrafficVector(long layerId)
	{
		return DoubleUtils.toArray(getNodeTraversingTrafficMap(layerId).values());
	}
	
	/**
	 * Returns the node position in a 2D Cartesian plane.
	 *
	 * @param nodeId Node identifier
	 * @return 2D coordinates (x, y)
	 * @since 0.3.0
	 */
	public Point2D getNodeXYPosition(long nodeId)
	{
		checkNodeIsActive(nodeId);
		return nodeXYPositionMap.get(nodeId);
	}
	
	/**
	 * Returns the position for each node. The output map is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @return Node XY position map
	 * @since 0.3.0
	 */
	public Map<Long, Point2D> getNodeXYPositionMap()
	{
		return Collections.unmodifiableMap(nodeXYPositionMap);
	}
	
	/**
	 * Returns the value of an attribute for each node.
	 *
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getNodesAttributeMap(String key)
	{
		Map<Long, String> out = new LinkedHashMap<Long, String>();
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		for(long nodeId : nodeIds_thisNetPlan)
		{
			out.put(nodeId, getNodeAttribute(nodeId, key));
		}
		
		return out;
	}

    /**
     * Returns the identifiers of the nodes with a given value in a certain attribute.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the nodes with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getNodesByAttribute(String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> out = new LinkedHashSet<Long>();

        Set<Long> nodeIds_thisNetPlan = getNodeIds();
        for(long nodeId : nodeIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisNode = getNodeAttributeMap(nodeId);
            if (value.equals(attributeMap_thisNode.get(attribute)))
                out.add(nodeId);
        }

        return out;
    }

	/**
	 * Returns the set of nodes down. The output set is backed in the {@code NetPlan} 
	 * object, hence users should copy it in case they want to modify it.
	 * 
	 * @return Nodes down
	 * @since 0.3.0
	 */
	public Set<Long> getNodesDown()
	{
		return Collections.unmodifiableSet(nodesDown);
	}
	
	/**
	 * Returns the number of traffic demands defined.
	 *
	 * @return Number of traffic demands
	 * @since 0.2.0
	 */
	public int getNumberOfDemands()
	{
		long layerId = getLayerDefaultId();
		return getNumberOfDemands(layerId);
	}

	/**
	 * Returns the number of traffic demands defined.
	 *
	 * @param layerId Layer identifier
	 * @return Number of traffic demands
	 * @since 0.3.0
	 */
	public int getNumberOfDemands(long layerId)
	{
		checkLayerIsActive(layerId);
		return demandIds.get(layerId).size();
	}
	
	/**
	 * <p>Returns the number of forwarding rules.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @return Number of forwarding rules
	 * @since 0.3.0
	 */
	public int getNumberOfForwardingRules()
	{
		long layerId = getLayerDefaultId();
		return getNumberOfForwardingRules(layerId);
	}
	
	/**
	 * <p>Returns the number of forwarding rules.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Number of forwarding rules
	 * @since 0.3.0
	 */
	public int getNumberOfForwardingRules(long layerId)
	{
		return getForwardingRules(layerId).size();
	}
	
	/**
	 * Returns the number of layers defined in the network.
	 *
	 * @return Number of layers
	 * @since 0.3.0
	 */
	public int getNumberOfLayers()
	{
		return layerIds.size();
	}

	/**
	 * Returns the number of unidirectional links.
	 *
	 * @return Number of unidirectional links
	 * @since 0.2.0
	 */
	public int getNumberOfLinks()
	{
		long layerId = getLayerDefaultId();
		return getNumberOfLinks(layerId);
	}

	/**
	 * Returns the number of unidirectional links.
	 *
	 * @param layerId Layer identifier
	 * @return Number of unidirectional links
	 * @since 0.3.0
	 */
	public int getNumberOfLinks(long layerId)
	{
		checkLayerIsActive(layerId);
		return linkIds.get(layerId).size();
	}

	/**
	 * Returns the number of node pairs. This is equal to <i>N x (N-1)</i>, where 
	 * <i>N</i> is equal to the number of nodes.
	 *
	 * @return Number of node pairs
	 * @since 0.3.0
	 */
	public int getNumberOfNodePairs()
	{
		int N = getNumberOfNodes();
		return N * (N - 1);
	}
	
	/**
	 * Returns the number of nodes defined within the network.
	 *
	 * @return Number of nodes
	 * @since 0.2.0
	 */
	public int getNumberOfNodes()
	{
		return nodeIds.size();
	}

	/**
	 * <p>Returns the number of protection segments.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 *
	 * @return Number of protection segments
	 * @since 0.2.0
	 */
	public int getNumberOfProtectionSegments()
	{
		long layerId = getLayerDefaultId();
		return getNumberOfProtectionSegments(layerId);
	}

	/**
	 * <p>Returns the number of protection segments.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 *
	 * @param layerId Layer identifier
	 * @return Number of protection segments
	 * @since 0.3.0
	 */
	public int getNumberOfProtectionSegments(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return segmentIds.get(layerId).size();
	}

	/**
	 * <p>Returns the number of routes for traffic demands defined.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Number of routes for traffic demands
	 * @since 0.2.0
	 */
	public int getNumberOfRoutes()
	{
		long layerId = getLayerDefaultId();
		return getNumberOfRoutes(layerId);
	}

	/**
	 * <p>Returns the number of routes for traffic demands defined.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Number of routes for traffic demands
	 * @since 0.3.0
	 */
	public int getNumberOfRoutes(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return routeIds.get(layerId).size();
	}

	/**
	 * Returns the number of SRGs defined.
	 * 
	 * @return Number of SRGs
	 * @since 0.2.3
	 */
	public int getNumberOfSRGs()
	{
		return srgIds.size();
	}
	
	/**
	 * <p>Returns the list with the sequence of links for each protection segment. 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Sequence of links per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, List<Long>> getProtectionSegmentAllSequenceOfLinks()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentAllSequenceOfLinks(layerId);
	}
	
	/**
	 * <p>Returns the list with the sequence of links for each protection segment. 
	 * The output map is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Sequence of links per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, List<Long>> getProtectionSegmentAllSequenceOfLinks(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return Collections.unmodifiableMap(segmentSequenceOfLinks.get(layerId));
	}
	
	/**
	 * <p>Returns the value of a given attribute for a protection segment. If not defined,
	 * it is search for the default layer in the network.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getProtectionSegmentAttribute(long segmentId, String key)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentAttribute(layerId, segmentId, key);
	}

	/**
	 * <p>Returns the value of a given attribute for a protection segment. If not defined,
	 * it is search for the given layer in the network.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getProtectionSegmentAttribute(long layerId, long segmentId, String key)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		String value = segmentAttributes.get(layerId).get(segmentId).get(key);
		return value == null ? getLayerAttribute(layerId, key) : value;
	}

	/**
	 * <p>Returns the attributes defined for a given protection segment. The output 
	 * map is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Protection segment attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getProtectionSegmentAttributeMap(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentAttributeMap(layerId, segmentId);
	}
	
	/**
	 * <p>Returns the attributes defined for a given protection segment. The output 
	 * map is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Protection segment attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getProtectionSegmentAttributeMap(long layerId, long segmentId)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		return Collections.unmodifiableMap(segmentAttributes.get(layerId).get(segmentId));
	}

    /**
     * <p>Returns the identifier of the first protection segment with a given value in a 
     * certain attribute, or -1 if not found.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first protection segment with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getProtectionSegmentByAttribute(String attribute, String value)
    {
        long layerId = getLayerDefaultId();
        return getProtectionSegmentByAttribute(layerId, attribute, value);
    }

    /**
     * <p>Returns the identifier of the first protection segment with a given value in a 
     * certain attribute, or -1 if not found.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param layerId Layer identifier
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first protection segment with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getProtectionSegmentByAttribute(long layerId, String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> segmentIds_thisNetPlan = getProtectionSegmentIds(layerId);
        for(long segmentId : segmentIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisProtectionSegment = getProtectionSegmentAttributeMap(layerId, segmentId);
            if (value.equals(attributeMap_thisProtectionSegment.get(attribute)))
                return segmentId;
        }

        return -1;
    }

	/**
	 * <p>Returns the carried traffic for a given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getProtectionSegmentCarriedTraffic(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentCarriedTraffic(layerId, segmentId);
	}
	
	/**
	 * <p>Returns the carried traffic for a given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getProtectionSegmentCarriedTraffic(long layerId, long segmentId)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		return cache_segmentCarriedTrafficMap.get(layerId).get(segmentId);
	}
	
	/**
	 * <p>Returns the carried traffic for each protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Carried traffic per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentCarriedTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentReservedCapacityMap(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic for each protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentCarriedTrafficMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(cache_segmentCarriedTrafficMap.get(layerId));
	}

	/**
	 * <p>Returns the carried traffic for each protection segment, in increasing 
	 * order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Carried traffic per protection segment
	 * @since 0.3.0
	 */
	public double[] getProtectionSegmentCarriedTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentCarriedTrafficVector(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic for each protection segment, in increasing 
	 * order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per protection segment
	 * @since 0.3.0
	 */
	public double[] getProtectionSegmentCarriedTrafficVector(long layerId)
	{
		return DoubleUtils.toArray(getProtectionSegmentCarriedTrafficMap(layerId).values());
	}
	
	/**
	 * <p>Returns the identifier of the destination node of the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Destination node identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentDestinationNode(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentDestinationNode(layerId, segmentId);
	}

	/**
	 * <p>Returns the identifier of the destination node of the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Destination node identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentDestinationNode(long layerId, long segmentId)
	{
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		return getLinkDestinationNode(layerId, seqLinks.get(seqLinks.size() - 1));
	}
	
	/**
	 * <p>Returns the set of identifiers of the protection segments. The output 
	 * set is backed in the {@code NetPlan} object, hence users should copy in 
	 * case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentIds()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentIds(layerId);
	}

	/**
	 * <p>Returns the set of identifiers of the protection segments. The output 
	 * set is backed in the {@code NetPlan} object, hence users should copy in 
	 * case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentIds(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return Collections.unmodifiableSet(segmentIds.get(layerId));
	}

	/**
	 * <p>Returns the set of identifiers of the protection segments, in increasing 
	 * order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public long[] getProtectionSegmentIdsVector()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentIdsVector(layerId);
	}

	/**
	 * <p>Returns the set of identifiers of the protection segments, in increasing 
	 * order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public long[] getProtectionSegmentIdsVector(long layerId)
	{
		return LongUtils.toArray(getProtectionSegmentIds(layerId));
	}
	
	/**
	 * <p>Returns the length in kilometers for the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Length in km
	 * @since 0.3.0
	 */
	public double getProtectionSegmentLengthInKm(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentLengthInKm(layerId, segmentId);
	}

	/**
	 * <p>Returns the length in kilometers for the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Length in km
	 * @since 0.3.0
	 */
	public double getProtectionSegmentLengthInKm(long layerId, long segmentId)
	{
		double routeLength = 0;
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		for (long linkId : seqLinks)
		{
			routeLength += getLinkLengthInKm(layerId, linkId);
		}

		return routeLength;
	}
	
	/**
	 * <p>Returns the length in kilometers for each protection segment. The output 
	 * map is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Length in km per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentLengthInKmMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentLengthInKmMap(layerId);
	}

	/**
	 * <p>Returns the length in kilometers for each protection segment. The output 
	 * map is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Length in km per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentLengthInKmMap(long layerId)
	{
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		for(long segmentId : segmentIds_thisLayer)
			out.put(segmentId, getProtectionSegmentLengthInKm(layerId, segmentId));
		
		return out;
	}
	
	/**
	 * <p>Returns the length in kilometers for each protection segment, in increasing 
	 * order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Length in km per protection segment
	 * @since 0.2.0
	 */
	public double[] getProtectionSegmentLengthInKmVector()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentLengthInKmVector(layerId);
	}

	/**
	 * <p>Returns the length in kilometers for each protection segment, in increasing 
	 * order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Length in km per protection segment
	 * @since 0.3.0
	 */
	public double[] getProtectionSegmentLengthInKmVector(long layerId)
	{
		return DoubleUtils.toArray(getProtectionSegmentLengthInKmMap(layerId).values());
	}
	
	/**
	 * <p>Returns the next identifier for a new protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Next protection segment identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentNextId()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentNextId(layerId);
	}

	/**
	 * <p>Returns the next identifier for a new protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Next protection segment identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentNextId(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		long segmentId = nextSegmentId.get(layerId);
		if (segmentId < 0) throw new Net2PlanException(String.format(TEMPLATE_NO_MORE_SEGMENTS_ALLOWED, layerId));
		
		return segmentId;
	}
	
	/**
	 * <p>Returns the number of hops for a given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Number of hops
	 * @since 0.3.0
	 */
	public int getProtectionSegmentNumberOfHops(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentNumberOfHops(layerId, segmentId);
	}

	/**
	 * <p>Returns the number of hops for a given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Number of hops
	 * @since 0.3.0
	 */
	public int getProtectionSegmentNumberOfHops(long layerId, long segmentId)
	{
		return getProtectionSegmentSequenceOfLinks(layerId, segmentId).size();
	}
	
	/**
	 * <p>Returns the number of hops per protection segment.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Number of hops per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Integer> getProtectionSegmentNumberOfHopsMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentNumberOfHopsMap(layerId);
	}
	
	/**
	 * <p>Returns the number of hops per protection segment.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Number of hops per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Integer> getProtectionSegmentNumberOfHopsMap(long layerId)
	{
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		Map<Long, Integer> out = new LinkedHashMap<Long, Integer>();
		for(long segmentId : segmentIds_thisLayer)
		{
			out.put(segmentId, getProtectionSegmentNumberOfHops(layerId, segmentId));
		}
		
		return out;
	}
	
	/**
	 * <p>Returns the number of hops per protection segment, in increasing order of identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Number of hops per protection segment
	 * @since 0.2.0
	 */
	public int[] getProtectionSegmentNumberOfHopsVector()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentNumberOfHopsVector(layerId);
	}

	/**
	 * <p>Returns the number of hops per protection segment, in increasing order of identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Number of hops per protection segment
	 * @since 0.3.0
	 */
	public int[] getProtectionSegmentNumberOfHopsVector(long layerId)
	{
		return IntUtils.toArray(getProtectionSegmentNumberOfHopsMap(layerId).values());
	}
	
	/**
	 * <p>Returns the origin-destination node pair for the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Origin-destination node pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getProtectionSegmentOriginDestinationNodePair(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentOriginDestinationNodePair(layerId, segmentId);
	}

	/**
	 * <p>Returns the origin-destination node pair for the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Origin-destination node pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getProtectionSegmentOriginDestinationNodePair(long layerId, long segmentId)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		long firstLinkId = seqLinks.get(0);
		long lastLinkId = seqLinks.get(seqLinks.size() - 1);
		return Pair.of(getLinkOriginNode(layerId, firstLinkId), getLinkDestinationNode(layerId, lastLinkId));
	}

	/**
	 * <p>Returns the identifier of the origin node of the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Origin node identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentOriginNode(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentOriginNode(layerId, segmentId);
	}

	/**
	 * <p>Returns the identifier of the origin node of the given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Origin node identifier
	 * @since 0.3.0
	 */
	public long getProtectionSegmentOriginNode(long layerId, long segmentId)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		return getLinkOriginNode(layerId, seqLinks.iterator().next());
	}
	
	/**
	 * <p>Returns the propagation delay in seconds of a given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Propagation delay
	 * @since 0.3.0
	 */
	public double getProtectionSegmentPropagationDelayInSeconds(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentPropagationDelayInSeconds(layerId, segmentId);
	}

	/**
	 * <p>Returns the propagation delay in seconds of a given protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Propagation delay
	 * @since 0.3.0
	 */
	public double getProtectionSegmentPropagationDelayInSeconds(long layerId, long segmentId)
	{
		double t_s_prop = 0;
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		for (long linkId : seqLinks)
		{
			t_s_prop += getLinkPropagationDelayInSeconds(layerId, linkId);
		}

		return t_s_prop;
	}
	
	/**
	 * <p>Returns the propagation delay in seconds per protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Propagation delay per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentPropagationDelayInSecondsMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentPropagationDelayInSecondsMap(layerId);
	}

	/**
	 * <p>Returns the propagation delay in seconds per protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Propagation delay per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentPropagationDelayInSecondsMap(long layerId)
	{
		Map<Long, Double> t_s_prop = new LinkedHashMap<Long, Double>();
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		for(long segmentId : segmentIds_thisLayer)
		{
			t_s_prop.put(segmentId, getProtectionSegmentPropagationDelayInSeconds(layerId, segmentId));
		}
		
		return t_s_prop;
	}

	/**
	 * <p>Returns the propagation delay in seconds per protection segment, in 
	 * increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Propagation delay per protection segment
	 * @since 0.2.0
	 */
	public double[] getProtectionSegmentPropagationDelayInSecondsVector()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentPropagationDelayInSecondsVector(layerId);
	}

	/**
	 * <p>Returns the propagation delay in seconds per protection segment, in 
	 * increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Propagation delay per protection segment
	 * @since 0.3.0
	 */
	public double[] getProtectionSegmentPropagationDelayInSecondsVector(long layerId)
	{
		return DoubleUtils.toArray(getProtectionSegmentPropagationDelayInSecondsMap(layerId).values());
	}
	
	/**
	 * <p>Returns the reserved capacity of a protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Reserved capacity
	 * @since 0.3.0
	 */
	public double getProtectionSegmentReservedCapacity(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentReservedCapacity(layerId, segmentId);
	}
	
	/**
	 * <p>Returns the reserved capacity of a protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Reserved capacity
	 * @since 0.3.0
	 */
	public double getProtectionSegmentReservedCapacity(long layerId, long segmentId)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		return segmentReservedCapacity.get(layerId).get(segmentId);
	}
	
	/**
	 * <p>Returns the reserved capacity per protection segment. The output map 
	 * is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Reserved capacity per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentReservedCapacityMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentReservedCapacityMap(layerId);
	}
	
	/**
	 * <p>Returns the reserved capacity per protection segment. The output map 
	 * is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Reserved capacity per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentReservedCapacityMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(segmentReservedCapacity.get(layerId));
	}

	/**
	 * <p>Returns the reserved capacity per protection segment, in increasing order 
	 * of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Reserved capacity per protection segment
	 * @since 0.2.0
	 */
	public double[] getProtectionSegmentReservedCapacityVector()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentReservedCapacityVector(layerId);
	}
	
	/**
	 * <p>Returns the reserved capacity per protection segment, in increasing order 
	 * of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Reserved capacity per protection segment
	 * @since 0.3.0
	 */
	public double[] getProtectionSegmentReservedCapacityVector(long layerId)
	{
		return DoubleUtils.toArray(getProtectionSegmentReservedCapacityMap(layerId).values());
	}

	/**
	 * <p>Returns the routes which share a given protection segment. The output 
	 * set is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentRoutes(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentRoutes(layerId, segmentId);
	}

	/**
	 * <p>Returns the routes which share a given protection segment. The output 
	 * set is backed in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentRoutes(long layerId, long segmentId)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		return Collections.unmodifiableSet(segmentRoutes.get(layerId).get(segmentId));
	}

	/**
	 * <p>Returns the routes which share each protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getProtectionSegmentRoutesMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentRoutesMap(layerId);
	}

	/**
	 * <p>Returns the routes which share each protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getProtectionSegmentRoutesMap(long layerId)
	{
		checkLayerIsActive(layerId);

		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		for(long segmentId : segmentIds_thisLayer) out.put(segmentId, getProtectionSegmentRoutes(layerId, segmentId));
		
		return out;
	}
	
	/**
	 * <p>Returns all SRGs associated to a given protection segment. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Segment identifier
	 * @return SRGs associated to the protection segment
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentSRGs(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentSRGs(layerId, segmentId);
	}

	/**
	 * <p>Returns all SRGs associated to a given protection segment. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return SRGs associated to the protection segment
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentSRGs(long layerId, long segmentId)
	{
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		List<Long> seqNodes = getProtectionSegmentSequenceOfNodes(layerId, segmentId);

		Set<Long> out = new TreeSet<Long>();
		for(long nodeId : seqNodes) out.addAll(getNodeSRGs(nodeId));
		for(long linkId : seqLinks) out.addAll(getLinkSRGs(layerId, linkId));
		
		return out;
	}
	
	/**
	 * <p>Returns all SRGs associated to each protection segment. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return SRGs associated to each protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getProtectionSegmentSRGsMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentSRGsMap(layerId);
	}
	
	/**
	 * <p>Returns all SRGs associated to each protection segment. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return SRGs associated to each protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getProtectionSegmentSRGsMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		for(long segmentId : segmentIds_thisLayer)
			out.put(segmentId, getProtectionSegmentSRGs(layerId, segmentId));
		
		return out;
	}
	
	/**
	 * <p>Returns the sequence of links traversed by a protection segment. The
	 * output list is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> getProtectionSegmentSequenceOfLinks(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentSequenceOfLinks(layerId, segmentId);
	}

	/**
	 * <p>Returns the sequence of links traversed by a protection segment. The
	 * output list is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> getProtectionSegmentSequenceOfLinks(long layerId, long segmentId)
	{
		checkProtectionSegmentIsActive(layerId, segmentId);
		return Collections.unmodifiableList(segmentSequenceOfLinks.get(layerId).get(segmentId));
	}

	/**
	 * <p>Returns the sequence of nodes traversed by a protection segment. The
	 * output list is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Sequence of nodes
	 * @since 0.3.0
	 */
	public List<Long> getProtectionSegmentSequenceOfNodes(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentSequenceOfNodes(layerId, segmentId);
	}

	/**
	 * <p>Returns the sequence of nodes traversed by a protection segment. The
	 * output list is backed in the {@code NetPlan} object, hence users should 
	 * copy in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Sequence of nodes
	 * @since 0.3.0
	 */
	public List<Long> getProtectionSegmentSequenceOfNodes(long layerId, long segmentId)
	{
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = getLinkMap(layerId);
		return GraphUtils.convertSequenceOfLinksToSequenceOfNodes(linkMap_thisLayer, seqLinks);
	}
	
	/**
	 * <p>Returns the available capacity for a protection segment.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return Available capacity
	 * @since 0.3.0
	 */
	public double getProtectionSegmentSpareCapacity(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentSpareCapacity(layerId, segmentId);
	}
	
	/**
	 * <p>Returns the available capacity for a protection segment.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return Available capacity
	 * @since 0.3.0
	 */
	public double getProtectionSegmentSpareCapacity(long layerId, long segmentId)
	{
		if (isProtectionSegmentDown(layerId, segmentId)) return 0;
		
		double u_s = getProtectionSegmentReservedCapacity(layerId, segmentId);
		double y_s = getProtectionSegmentCarriedTraffic(layerId, segmentId);
		return Math.max(u_s - y_s, 0);
	}
	
	/**
	 * <p>Returns the available capacity per protection segment.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Available capacity per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentSpareCapacityMap()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentSpareCapacityMap(layerId);
	}
	
	/**
	 * <p>Returns the available capacity per protection segment.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Available capacity per protection segment
	 * @since 0.3.0
	 */
	public Map<Long, Double> getProtectionSegmentSpareCapacityMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		for(long segmentId : segmentIds_thisLayer) out.put(segmentId, getProtectionSegmentSpareCapacity(layerId, segmentId));
		
		return out;
	}
	
	/**
	 * <p>Returns the available capacity per protection segment, in increasing 
	 * order of identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Available capacity per protection segment
	 * @since 0.3.0
	 */
	public double[] getProtectionSegmentSpareCapacityVector()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentSpareCapacityVector(layerId);
	}
	
	/**
	 * <p>Returns the available capacity per protection segment, in increasing 
	 * order of identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Available capacity per protection segment
	 * @since 0.3.0
	 */
	public double[] getProtectionSegmentSpareCapacityVector(long layerId)
	{
		return DoubleUtils.toArray(getProtectionSegmentSpareCapacityMap(layerId).values());
	}
	
	/**
	 * <p>Returns the value of an attribute for each protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.2.0
	 */
	public Map<Long, String> getProtectionSegmentsAttributeMap(String key)
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentsAttributeMap(layerId, key);
	}

	/**
	 * <p>Returns the value of an attribute for each protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getProtectionSegmentsAttributeMap(long layerId, String key)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, String> out = new LinkedHashMap<Long, String>();
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		for(long segmentId : segmentIds_thisLayer)
		{
			out.put(segmentId, getProtectionSegmentAttribute(layerId, segmentId, key));
		}
		
		return out;
	}
	
    /**
     * <p>Returns the identifiers of the protection segments with a given value in a certain attribute.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the protection segments with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getProtectionSegmentsByAttribute(String attribute, String value)
    {
        long layerId = getLayerDefaultId();
        return getProtectionSegmentsByAttribute(layerId, attribute, value);
    }

    /**
     * <p>Returns the identifiers of the protection segments with a given value in a certain attribute.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param layerId Layer identifier
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the protection segments with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getProtectionSegmentsByAttribute(long layerId, String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> out = new LinkedHashSet<Long>();

        Set<Long> segmentIds_thisNetPlan = getProtectionSegmentIds(layerId);
        for(long segmentId : segmentIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisProtectionSegment = getProtectionSegmentAttributeMap(layerId, segmentId);
            if (value.equals(attributeMap_thisProtectionSegment.get(attribute)))
                out.add(segmentId);
        }

        return out;
    }

	/**
	 * <p>Returns the set of protection segments assigned only to one route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentsDedicated()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentsDedicated(layerId);
	}

	/**
	 * <p>Returns the set of protection segments assigned only to one route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Protection segment identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentsDedicated(long layerId)
	{
		checkLayerIsActive(layerId);

		Set<Long> dedicatedSegmentIds = new LinkedHashSet<Long>();
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		for(long segmentId : segmentIds_thisLayer)
			if (isProtectionSegmentDedicated(layerId, segmentId))
				dedicatedSegmentIds.add(segmentId);
		
		return dedicatedSegmentIds;
	}
	
	/**
	 * <p>Returns the set of protection segments down. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Set of protection segments down
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentsDown()
	{
		long layerId = getLayerDefaultId();
		return getProtectionSegmentsDown(layerId);
	}

	/**
	 * <p>Returns the set of protection segments down. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Set of protection segments down
	 * @since 0.3.0
	 */
	public Set<Long> getProtectionSegmentsDown(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return Collections.unmodifiableSet(segmentsDown.get(layerId));
	}

	/**
	 * <p>Returns the set of protection segments down per layer.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Set of protection segments down per layer
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getProtectionSegmentsDownMap()
	{
		Map<Long, Set<Long>> segmentsDown_thisNetPlan = new LinkedHashMap<Long, Set<Long>>();
		Set<Long> layerIds_thisNetPlan = getLayerIds();
		for(long layerId : layerIds_thisNetPlan)
			segmentsDown_thisNetPlan.put(layerId, getProtectionSegmentsDown(layerId));
		
		return Collections.unmodifiableMap(segmentsDown_thisNetPlan);
	}
	
	/**
	 * <p>Returns the sequence of links (and maybe protection segments) for each 
	 * route. The output map is backed in the {@code NetPlan} object, hence users 
	 * should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Sequence of links (and maybe protection segments) per route
	 * @since 0.3.0
	 */
	public Map<Long, List<Long>> getRouteAllSequenceOfLinks()
	{
		long layerId = getLayerDefaultId();
		return getRouteAllSequenceOfLinks(layerId);
	}
	
	/**
	 * <p>Returns the sequence of links (and maybe protection segments) for each 
	 * route. The output map is backed in the {@code NetPlan} object, hence users 
	 * should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Sequence of links (and maybe protection segments) per route
	 * @since 0.3.0
	 */
	public Map<Long, List<Long>> getRouteAllSequenceOfLinks(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return Collections.unmodifiableMap(routeSequenceOfLinks.get(layerId));
	}

	/**
	 * <p>Returns the value of a given attribute for a route. If not defined,
	 * it is search for the default layer in the network.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getRouteAttribute(long routeId, String key)
	{
		long layerId = getLayerDefaultId();
		return getRouteAttribute(layerId, routeId, key);
	}

	/**
	 <p>* Returns the value of a given attribute for a route. If not defined,
	 * it is search for the given layer.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getRouteAttribute(long layerId, long routeId, String key)
	{
		checkRouteIsActive(layerId, routeId);
		String value = routeAttributes.get(layerId).get(routeId).get(key);
		return value == null ? getLayerAttribute(layerId, key) : value;
	}
	
	/**
	 * <p>Returns the attributes defined for a given route. The output map is 
	 * backed in the {@code NetPlan} object, hence users should copy it in case 
	 * they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Route attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getRouteAttributeMap(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteAttributeMap(layerId, routeId);
	}

	/**
	 * <p>Returns the attributes defined for a given route. The output map is 
	 * backed in the {@code NetPlan} object, hence users should copy it in case 
	 * they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Route attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getRouteAttributeMap(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return Collections.unmodifiableMap(routeAttributes.get(layerId).get(routeId));
	}
	
	/**
	 * <p>Returns the protection segments defined for a given route. The output 
	 * list is backed in the {@code NetPlan} object, hence users should copy it 
	 * in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Backup segment list
	 * @since 0.3.0
	 */
	public List<Long> getRouteBackupSegmentList(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteBackupSegmentList(layerId, routeId);
	}

	/**
	 * <p>Returns the protection segments defined for a given route. The output 
	 * list is backed in the {@code NetPlan} object, hence users should copy it 
	 * in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Backup segment list
	 * @since 0.3.0
	 */
	public List<Long> getRouteBackupSegmentList(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return Collections.unmodifiableList(routeBackupSegmentList.get(layerId).get(routeId));
	}
	
	/**
	 * <p>Returns the backup segment list per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Backup segment list per route
	 * @since 0.3.0
	 */
	public Map<Long, List<Long>> getRouteBackupSegmentListMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteBackupSegmentListMap(layerId);
	}

	/**
	 * <p>Returns the backup segment list per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Backup segment list per route
	 * @since 0.3.0
	 */
	public Map<Long, List<Long>> getRouteBackupSegmentListMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		Map<Long, List<Long>> out = new LinkedHashMap<Long, List<Long>>();
		for (long routeId : routeIds_thisLayer)
			out.put(routeId, getRouteBackupSegmentList(layerId, routeId));
		
		return out;
	}
	
    /**
     * <p>Returns the identifier of the first route with a given value in a 
     * certain attribute, or -1 if not found.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first route with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getRouteByAttribute(String attribute, String value)
    {
        long layerId = getLayerDefaultId();
        return getRouteByAttribute(layerId, attribute, value);
    }

    /**
     * <p>Returns the identifier of the first route with a given value in a 
     * certain attribute, or -1 if not found.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param layerId Layer identifier
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first route with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getRouteByAttribute(long layerId, String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> routeIds_thisNetPlan = getRouteIds(layerId);
        for(long routeId : routeIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisRoute = getRouteAttributeMap(layerId, routeId);
            if (value.equals(attributeMap_thisRoute.get(attribute)))
                return routeId;
        }

        return -1;
    }

	/**
	 * <p>Returns the carried traffic (in demand traffic units) for a given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getRouteCarriedTraffic(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteCarriedTraffic(layerId, routeId);
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units) for a given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getRouteCarriedTraffic(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return routeCarriedTrafficMap.get(layerId).get(routeId);
	}
	
	/**
	 * <p>Returns the ratio between the carried traffic per route with respect to 
	 * the total offered traffic by the corresponding demand.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Carried traffic fraction
	 * @since 0.3.0
	 */
	public double getRouteCarriedTrafficFraction(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteCarriedTrafficFraction(layerId, routeId);
	}
	
	/**
	 * <p>Returns the ratio between the carried traffic per route with respect to 
	 * the total offered traffic by the corresponding demand.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Carried traffic fraction
	 * @since 0.3.0
	 */
	public double getRouteCarriedTrafficFraction(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		
		long demandId = getRouteDemand(layerId, routeId);
		double r_d = getDemandCarriedTraffic(demandId);
		if (r_d == 0) return 0;
		
		double x_p = getRouteCarriedTraffic(layerId, routeId);
		return x_p / r_d;
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units) per route. The
	 * output map is backed in the {@code NetPlan} object, hence users should 
	 * copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Carried traffic per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteCarriedTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteCarriedTrafficMap(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units) per route. The
	 * output map is backed in the {@code NetPlan} object, hence users should 
	 * copy it in case they want to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteCarriedTrafficMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(routeCarriedTrafficMap.get(layerId));
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units), with respect to 
	 * the total offered traffic by the corresponding demand, per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Carried traffic fraction per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteCarriedTrafficFractionMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteCarriedTrafficFractionMap(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units), with respect to 
	 * the total offered traffic by the corresponding demand, per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic fraction per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteCarriedTrafficFractionMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, Double> routeCarriedTrafficFractionMap = new LinkedHashMap<Long, Double>();
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		for(long routeId : routeIds_thisLayer) routeCarriedTrafficFractionMap.put(routeId, getRouteCarriedTrafficFraction(layerId, routeId));
		
		return routeCarriedTrafficFractionMap;
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units) per route, in 
	 * increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Carried traffic per route
	 * @since 0.2.0
	 */
	public double[] getRouteCarriedTrafficVector()
	{
		long layerId = getLayerDefaultId();
		return getRouteCarriedTrafficVector(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units) per route, in 
	 * increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic per route
	 * @since 0.3.0
	 */
	public double[] getRouteCarriedTrafficVector(long layerId)
	{
		return DoubleUtils.toArray(getRouteCarriedTrafficMap(layerId).values());
	}

	/**
	 * <p>Returns the carried traffic (in demand traffic units), with respect to 
	 * the total offered traffic by the corresponding demand, per route, in increasing 
	 * order of identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Carried traffic fraction per route
	 * @since 0.2.0
	 */
	public double[] getRouteCarriedTrafficFractionVector()
	{
		long layerId = getLayerDefaultId();
		return getRouteCarriedTrafficFractionVector(layerId);
	}
	
	/**
	 * <p>Returns the carried traffic (in demand traffic units), with respect to 
	 * the total offered traffic by the corresponding demand, per route, in increasing 
	 * order of identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Carried traffic fraction per route
	 * @since 0.3.0
	 */
	public double[] getRouteCarriedTrafficFractionVector(long layerId)
	{
		return DoubleUtils.toArray(getRouteCarriedTrafficFractionMap(layerId).values());
	}
	
	/**
	 * <p>Returns the cycle type of a route (i.e., loopless, or with loops).</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Type of cycle (loopless, or with loops)
	 * @since 0.3.0
	 */
	public RoutingCycleType getRouteCycleType(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteCycleType(layerId, routeId);
	}
	
	/**
	 * <p>Returns the cycle type of a route (i.e., loopless, or with loops).</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Type of cycle (loopless, or with loops)
	 * @since 0.3.0
	 */
	public RoutingCycleType getRouteCycleType(long layerId, long routeId)
	{
		List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
		if (LongUtils.unique(seqLinks).size() != seqLinks.size()) return RoutingCycleType.OPEN_CYCLES;
		
		List<Long> seqNodes = getRouteSequenceOfNodes(layerId, routeId);
		if (LongUtils.unique(seqNodes).size() != seqNodes.size()) return RoutingCycleType.OPEN_CYCLES;
		
		return RoutingCycleType.LOOPLESS;
	}
	
	/**
	 * <p>Returns the associated demand for a given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long getRouteDemand(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteDemand(layerId, routeId);
	}

	/**
	 * <p>Returns the associated demand for a given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Demand identifier
	 * @since 0.3.0
	 */
	public long getRouteDemand(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return routeDemandIds.get(layerId).get(routeId);
	}
	
	/**
	 * <p>Returns the associated demand per route. The output map is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Map<Long, Long> getRouteDemandMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteDemandMap(layerId);
	}

	/**
	 * <p>Returns the associated demand per route. The output map is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public Map<Long, Long> getRouteDemandMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(routeDemandIds.get(layerId));
	}

	/**
	 * <p>Returns the associated demand per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public long[] getRouteDemandVector()
	{
		long layerId = getLayerDefaultId();
		return getRouteDemandVector(layerId);
	}

	/**
	 * <p>Returns the associated demand per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Demand identifiers
	 * @since 0.3.0
	 */
	public long[] getRouteDemandVector(long layerId)
	{
		checkLayerIsActive(layerId);
		return LongUtils.toArray(routeDemandIds.get(layerId).values());
	}
	
	/**
	 * <p>Returns the identifier of the egress node of the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Egress node identifier
	 * @since 0.3.0
	 */
	public long getRouteEgressNode(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteEgressNode(layerId, routeId);
	}

	/**
	 * <p>Returns the identifier of the egress node of the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Egress node identifier
	 * @since 0.3.0
	 */
	public long getRouteEgressNode(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		long demandId = routeDemandIds.get(layerId).get(routeId);
		return getDemandEgressNode(layerId, demandId);
	}
	
	/**
	 * <p>Returns the set of identifiers of the routes. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getRouteIds()
	{
		long layerId = getLayerDefaultId();
		return getRouteIds(layerId);
	}
	
	/**
	 * <p>Returns the set of identifiers of the routes. The output set is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getRouteIds(long layerId)
	{
		checkLayerIsActive(layerId);
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return Collections.unmodifiableSet(routeIds.get(layerId));
	}

	/**
	 * <p>Returns the set of identifiers of the routes, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public long[] getRouteIdsVector()
	{
		long layerId = getLayerDefaultId();
		return getRouteIdsVector(layerId);
	}
	
	/**
	 * <p>Returns the set of identifiers of the routes, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Route identifiers
	 * @since 0.3.0
	 */
	public long[] getRouteIdsVector(long layerId)
	{
		return LongUtils.toArray(getRouteIds(layerId));
	}
	
	/**
	 * <p>Returns the ingress-egress node pair of a route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Ingress-egress node pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getRouteIngressEgressNodePair(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteIngressEgressNodePair(layerId, routeId);
	}

	/**
	 * <p>Returns the ingress-egress node pair of a route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Ingress-egress node pair
	 * @since 0.3.0
	 */
	public Pair<Long, Long> getRouteIngressEgressNodePair(long layerId, long routeId)
	{
		long demandId = getRouteDemand(layerId, routeId);
		return getDemandIngressEgressNodePair(layerId, demandId);
	}

	/**
	 * <p>Returns the identifier of the ingress node of the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Ingress node identifier
	 * @since 0.3.0
	 */
	public long getRouteIngressNode(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteIngressNode(layerId, routeId);
	}

	/**
	 * <p>Returns the identifier of the ingress node of the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Ingress node identifier
	 * @since 0.3.0
	 */
	public long getRouteIngressNode(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		long demandId = routeDemandIds.get(layerId).get(routeId);
		return getDemandIngressNode(layerId, demandId);
	}
	
	/**
	 * <p>Returns the length in kilometers for the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Length in km
	 * @since 0.3.0
	 */
	public double getRouteLengthInKm(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteLengthInKm(layerId, routeId);
	}

	/**
	 * <p>Returns the length in kilometers for the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Length in km
	 * @since 0.3.0
	 */
	public double getRouteLengthInKm(long layerId, long routeId)
	{
		double routeLength = 0;
		List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
		for (long linkId : seqLinks)
		{
			if (linkId >= 0) routeLength += getLinkLengthInKm(layerId, linkId);
			else routeLength += getProtectionSegmentLengthInKm(layerId, -1 - linkId);
		}

		return routeLength;
	}
	
	/**
	 * <p>Returns the length in kilometers per route. The output map is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Length in km per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteLengthInKmMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteLengthInKmMap(layerId);
	}

	/**
	 * <p>Returns the length in kilometers per route. The output map is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Length in km per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteLengthInKmMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		for(long routeId : routeIds_thisLayer)
		{
			out.put(routeId, getRouteLengthInKm(layerId, routeId));
		}
		
		return out;
	}
	
	/**
	 * <p>Returns the length in kilometers per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Length in km per route
	 * @since 0.2.0
	 */
	public double[] getRouteLengthInKmVector()
	{
		long layerId = getLayerDefaultId();
		return getRouteLengthInKmVector(layerId);
	}

	/**
	 * <p>Returns the length in kilometers per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Length in km per route
	 * @since 0.3.0
	 */
	public double[] getRouteLengthInKmVector(long layerId)
	{
		return DoubleUtils.toArray(getRouteLengthInKmMap(layerId).values());
	}
	
	/**
	 * <p>Returns the next identifier for a new route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Next route identifier
	 * @since 0.3.0
	 */
	public long getRouteNextId()
	{
		long layerId = getLayerDefaultId();
		return getRouteNextId(layerId);
	}

	/**
	 * <p>Returns the next identifier for a new route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Next route identifier
	 * @since 0.3.0
	 */
	public long getRouteNextId(long layerId)
	{
		checkLayerIsActive(layerId);
		long routeId = nextRouteId.get(layerId);
		if (routeId < 0) throw new Net2PlanException(String.format(TEMPLATE_NO_MORE_ROUTES_ALLOWED, layerId));
		
		return routeId;
	}
	
	/**
	 * <p>Returns the number of hops for a given route. In case some protection 
	 * segment is traversed, all its links are also taken into account.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Number of hops
	 * @since 0.3.0
	 */
	public int getRouteNumberOfHops(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteNumberOfHops(layerId, routeId);
	}

	/**
	 * <p>Returns the number of hops for a given route. In case some protection 
	 * segment is traversed, all its links are also taken into account.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Number of hops
	 * @since 0.3.0
	 */
	public int getRouteNumberOfHops(long layerId, long routeId)
	{
		return convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, getRouteSequenceOfLinks(layerId, routeId)).size();
	}
	
	/**
	 * <p>Returns the number of hops per route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Number of hops per route
	 * @since 0.3.0
	 */
	public Map<Long, Integer> getRouteNumberOfHopsMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteNumberOfHopsMap(layerId);
	}
	
	/**
	 * <p>Returns the number of hops per route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Number of hops per route
	 * @since 0.3.0
	 */
	public Map<Long, Integer> getRouteNumberOfHopsMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		Map<Long, Integer> out = new LinkedHashMap<Long, Integer>();
		for(long routeId : routeIds_thisLayer)
		{
			out.put(routeId, getRouteNumberOfHops(layerId, routeId));
		}
		
		return out;
	}
	
	/**
	 * <p>Returns the number of hops per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Number of hops per route
	 * @since 0.2.0
	 */
	public int[] getRouteNumberOfHopsVector()
	{
		long layerId = getLayerDefaultId();
		return getRouteNumberOfHopsVector(layerId);
	}

	/**
	 * <p>Returns the number of hops per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Number of hops per route
	 * @since 0.3.0
	 */
	public int[] getRouteNumberOfHopsVector(long layerId)
	{
		return IntUtils.toArray(getRouteNumberOfHopsMap(layerId).values());
	}

	/**
	 * <p>Returns the capacity occupied (in link capacity units) by the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Occupied capacity
	 * @since 0.3.0
	 */
	public double getRouteOccupiedCapacity(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteOccupiedCapacity(layerId, routeId);
	}
	
	/**
	 * <p>Returns the capacity occupied (in link capacity units) by the given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Occupied capacity
	 * @since 0.3.0
	 */
	public double getRouteOccupiedCapacity(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return routeOccupiedCapacityMap.get(layerId).get(routeId);
	}

	/**
	 * <p>Returns the capacity occupied (in link capacity units) per route. The 
	 * output map is backed in the {@code NetPlan} object, hence users should 
	 * copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Occupied capacity per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteOccupiedCapacityMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteOccupiedCapacityMap(layerId);
	}
	
	/**
	 * <p>Returns the capacity occupied (in link capacity units) per route. The 
	 * output map is backed in the {@code NetPlan} object, hence users should 
	 * copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Occupied capacity per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRouteOccupiedCapacityMap(long layerId)
	{
		checkLayerIsActive(layerId);
		return Collections.unmodifiableMap(routeOccupiedCapacityMap.get(layerId));
	}

	/**
	 * <p>Returns the capacity occupied (in link capacity units) per route, in 
	 * increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Occupied capacity per route
	 * @since 0.3.0
	 */
	public double[] getRouteOccupiedCapacityVector()
	{
		long layerId = getLayerDefaultId();
		return getRouteOccupiedCapacityVector(layerId);
	}
	
	/**
	 * <p>Returns the capacity occupied (in link capacity units) per route, in 
	 * increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Occupied capacity per route
	 * @since 0.3.0
	 */
	public double[] getRouteOccupiedCapacityVector(long layerId)
	{
		return DoubleUtils.toArray(getRouteOccupiedCapacityMap(layerId).values());
	}
	
	/**
	 * <p>Returns the original carried traffic (in demand traffic units) by the route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getRouteOriginalCarriedTraffic(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteOriginalCarriedTraffic(layerId, routeId);
	}

	/**
	 * <p>Returns the original carried traffic (in demand traffic units) by the route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Carried traffic
	 * @since 0.3.0
	 */
	public double getRouteOriginalCarriedTraffic(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return routeOriginalCarriedTrafficMap.get(layerId).get(routeId);
	}

	/**
	 * <p>Returns the original occupied capacity (in link capacity units) by the route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Occupied capacity
	 * @since 0.3.0
	 */
	public double getRouteOriginalOccupiedCapacity(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteOriginalOccupiedCapacity(layerId, routeId);
	}

	/**
	 * <p>Returns the original occupied capacity (in link capacity units) by the route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Occupied capacity
	 * @since 0.3.0
	 */
	public double getRouteOriginalOccupiedCapacity(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return routeOriginalOccupiedCapacityMap.get(layerId).get(routeId);
	}

	/**
	 * <p>Returns the original sequence of links of the route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> getRouteOriginalSequenceOfLinks(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteOriginalSequenceOfLinks(layerId, routeId);
	}

	/**
	 * <p>Returns the original sequence of links of the route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Sequence of links
	 * @since 0.3.0
	 */
	public List<Long> getRouteOriginalSequenceOfLinks(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return routeOriginalSequenceOfLinks.get(layerId).get(routeId);
	}
	
	
	/**
	 * <p>Returns the propagation delay in seconds in a given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Propagation delay
	 * @since 0.3.0
	 */
	public double getRoutePropagationDelayInSeconds(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRoutePropagationDelayInSeconds(layerId, routeId);
	}

	/**
	 * <p>Returns the propagation delay in seconds in a given route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Propagation delay
	 * @since 0.3.0
	 */
	public double getRoutePropagationDelayInSeconds(long layerId, long routeId)
	{
		double t_r_prop = 0;
		List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
		for (long linkId : seqLinks)
		{
			if (linkId >= 0) t_r_prop += getLinkPropagationDelayInSeconds(layerId, linkId);
			else t_r_prop += getProtectionSegmentPropagationDelayInSeconds(layerId, -1 - linkId);
		}

		return t_r_prop;
	}
	
	/**
	 * <p>Returns the propagation delay per route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Propagation delay per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRoutePropagationDelayInSecondsMap()
	{
		long layerId = getLayerDefaultId();
		return getRoutePropagationDelayInSecondsMap(layerId);
	}

	/**
	 * <p>Returns the propagation delay per route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Propagation delay per route
	 * @since 0.3.0
	 */
	public Map<Long, Double> getRoutePropagationDelayInSecondsMap(long layerId)
	{
		Map<Long, Double> t_r_prop = new LinkedHashMap<Long, Double>();
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		for(long routeId : routeIds_thisLayer)
		{
			t_r_prop.put(routeId, getRoutePropagationDelayInSeconds(layerId, routeId));
		}
		
		return t_r_prop;
	}

	/**
	 * <p>Returns the propagation delay per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Propagation delay per route
	 * @since 0.2.0
	 */
	public double[] getRoutePropagationDelayInSecondsVector()
	{
		long layerId = getLayerDefaultId();
		return getRoutePropagationDelayInSecondsVector(layerId);
	}

	/**
	 * <p>Returns the propagation delay per route, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Propagation delay per route
	 * @since 0.3.0
	 */
	public double[] getRoutePropagationDelayInSecondsVector(long layerId)
	{
		return DoubleUtils.toArray(getRoutePropagationDelayInSecondsMap(layerId).values());
	}

	/**
	 * <p>Returns all SRGs associated to a given route. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return SRGs associated to the route
	 * @since 0.3.0
	 */
	public Set<Long> getRouteSRGs(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteSRGs(layerId, routeId);
	}

	/**
	 * <p>Returns all SRGs associated to a given route. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return SRGs associated to the route
	 * @since 0.3.0
	 */
	public Set<Long> getRouteSRGs(long layerId, long routeId)
	{
		List<Long> seqLinks = convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(layerId, getRouteSequenceOfLinks(layerId, routeId));
		List<Long> seqNodes = getRouteSequenceOfNodes(layerId, routeId);
		
		Set<Long> out = new TreeSet<Long>();
		for(long nodeId : seqNodes) out.addAll(getNodeSRGs(nodeId));
		for(long linkId : seqLinks) out.addAll(getLinkSRGs(layerId, linkId));
		
		return out;
	}
	
	/**
	 * <p>Returns all SRGs associated per route. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return SRGs per route
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getRouteSRGsMap()
	{
		long layerId = getLayerDefaultId();
		return getRouteSRGsMap(layerId);
	}
	
	/**
	 * <p>Returns all SRGs associated per route. It is equal to
	 * the union of the SRGs associated to each traversed node/link.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return SRGs per route
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getRouteSRGsMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		for(long routeId : routeIds_thisLayer)
			out.put(routeId, getRouteSRGs(layerId, routeId));
		
		return out;
	}
	
	/**
	 * <p>Returns the sequence of links (and maybe protection segments) traversed by 
	 * a route. The output list is backed in the {@code NetPlan} object, hence 
	 * users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Protection segments are indicated here as (-1 - segmentId), 
	 * that is, protection segment 1 would be represented as -2.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Sequence of links and protection segments
	 * @since 0.3.0
	 */
	public List<Long> getRouteSequenceOfLinks(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteSequenceOfLinks(layerId, routeId);
	}

	/**
	 * <p>Returns the sequence of links (and maybe protection segments) traversed by 
	 * a route. The output list is backed in the {@code NetPlan} object, hence 
	 * users should copy it in case they want to modify it.</p>
	 *
	 * <p><b>Important</b>: Protection segments are indicated here as (-1 - segmentId), 
	 * that is, protection segment 1 would be represented as -2.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Sequence of links and protection segments
	 * @since 0.3.0
	 */
	public List<Long> getRouteSequenceOfLinks(long layerId, long routeId)
	{
		checkRouteIsActive(layerId, routeId);
		return Collections.unmodifiableList(routeSequenceOfLinks.get(layerId).get(routeId));
	}

	/**
	 * <p>Returns the sequence of nodes traversed by a route (including those in 
	 * protection segments).</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return Sequence of nodes
	 * @since 0.3.0
	 */
	public List<Long> getRouteSequenceOfNodes(long routeId)
	{
		long layerId = getLayerDefaultId();
		return getRouteSequenceOfNodes(layerId, routeId);
	}

	/**
	 * <p>Returns the sequence of nodes traversed by a route (including those in 
	 * protection segments).</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return Sequence of nodes
	 * @since 0.3.0
	 */
	public List<Long> getRouteSequenceOfNodes(long layerId, long routeId)
	{
		List<Long> seqLinks = convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(getRouteSequenceOfLinks(layerId, routeId));
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = getLinkMap(layerId);
		return GraphUtils.convertSequenceOfLinksToSequenceOfNodes(linkMap_thisLayer, seqLinks);
	}
	
	/**
	 * <p>Returns the value of an attribute for each route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getRoutesAttributeMap(String key)
	{
		long layerId = getLayerDefaultId();
		return getRoutesAttributeMap(layerId, key);
	}

	/**
	 * <p>Returns the value of an attribute for each route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getRoutesAttributeMap(long layerId, String key)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, String> out = new LinkedHashMap<Long, String>();
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		for(long routeId : routeIds_thisLayer)
		{
			out.put(routeId, getRouteAttribute(layerId, routeId, key));
		}
		
		return out;
	}
	
    /**
     * <p>Returns the identifiers of the routes with a given value in a certain attribute.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the routes with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getRoutesByAttribute(String attribute, String value)
    {
        long layerId = getLayerDefaultId();
        return getRoutesByAttribute(layerId, attribute, value);
    }

    /**
     * <p>Returns the identifiers of the routes with a given value in a certain attribute.</p>
     * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
     * @param layerId Layer identifier
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the routes with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getRoutesByAttribute(long layerId, String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> out = new LinkedHashSet<Long>();

        Set<Long> routeIds_thisNetPlan = getRouteIds(layerId);
        for(long routeId : routeIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisRoute = getRouteAttributeMap(layerId, routeId);
            if (value.equals(attributeMap_thisRoute.get(attribute)))
                out.add(routeId);
        }

        return out;
    }

	/**
	 * <p>Returns the set of routes in down state. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want 
	 * to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Routes down
	 * @since 0.3.0
	 */
	public Set<Long> getRoutesDown()
	{
		long layerId = getLayerDefaultId();
		return getRoutesDown(layerId);
	}
	
	/**
	 * <p>Returns the set of routes in down state. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want 
	 * to modify it.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Routes down
	 * @since 0.3.0
	 */
	public Set<Long> getRoutesDown(long layerId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return Collections.unmodifiableSet(routesDown.get(layerId));
	}

	/**
	 * <p>Returns the set of routes in down state per layer.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Routes down per layer
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getRoutesDownMap()
	{
		Map<Long, Set<Long>> routesDown_thisNetPlan = new LinkedHashMap<Long, Set<Long>>();
		Set<Long> layerIds_thisNetPlan = getLayerIds();
		for(long layerId : layerIds_thisNetPlan)
			routesDown_thisNetPlan.put(layerId, getRoutesDown(layerId));
		
		return Collections.unmodifiableMap(routesDown_thisNetPlan);
	}

	/**
	 * Returns the current {@link com.net2plan.utils.Constants.RoutingType RoutingType}.
	 * 
	 * @return Current {@link com.net2plan.utils.Constants.RoutingType RoutingType}
	 * @since 0.3.0
	 */
	public RoutingType getRoutingType()
	{
		long layerId = getLayerDefaultId();
		return getRoutingType(layerId);
	}
	
	/**
	 * Returns the current {@link com.net2plan.utils.Constants.RoutingType RoutingType}.
	 * 
	 * @param layerId Layer identifier
	 * @return Current {@link com.net2plan.utils.Constants.RoutingType RoutingType}
	 * @since 0.3.0
	 */
	public RoutingType getRoutingType(long layerId)
	{
		checkLayerIsActive(layerId);
		return layerRoutingType.get(layerId);
	}
	
	/**
	 * Returns the current {@link com.net2plan.utils.Constants.RoutingType RoutingType} per layer.
	 * 
	 * @return Current {@link com.net2plan.utils.Constants.RoutingType RoutingType} per layer
	 * @since 0.3.0
	 */
	public Map<Long, RoutingType> getRoutingTypeMap()
	{
		return Collections.unmodifiableMap(layerRoutingType);
	}

	/**
	 * Returns the value of a given attribute for an SRG at the first layer. If not defined,
	 * it is search for the first layer in the network.
	 *
	 * @param srgId SRG identifier
	 * @param key Attribute name
	 * @return Attribute value (or null, if not defined)
	 * @since 0.3.0
	 */
	public String getSRGAttribute(long srgId, String key)
	{
		checkSRGIsActive(srgId);
		String value = srgAttributes.get(srgId).get(key);
		return value == null ? getNetworkAttribute(key) : value;
	}
	
	/**
	 * Returns the attributes defined for a given SRG. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @param srgId SRG identifier
	 * @return SRG attributes
	 * @since 0.3.0
	 */
	public Map<String, String> getSRGAttributeMap(long srgId)
	{
		checkSRGIsActive(srgId);
		return Collections.unmodifiableMap(srgAttributes.get(srgId));
	}
	
	/**
	 * Returns the availability of the given SRG. The availability is given by 
	 * the quotient of MTTF and the mean time between failures (MTBF), which is 
	 * equal to the sum of MTTF and MTTR.
	 * 
	 * @param srgId SRG identifier
	 * @return Availability
	 * @since 0.3.0
	 */
	public double getSRGAvailability(long srgId)
	{
		double mttfInHours = getSRGMeanTimeToFailInHours(srgId);
		if (mttfInHours == Double.MAX_VALUE) return 1;

		double mttrInHours = getSRGMeanTimeToRepairInHours(srgId);
		return mttfInHours / (mttfInHours + mttrInHours);
	}

	/**
	 * Returns the availability per SRG.
	 * 
	 * @return Availability per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Double> getSRGAvailabilityMap()
	{
		Set<Long> srgIds_thisNetwork = getSRGIds();
		Map<Long, Double> out = new LinkedHashMap<Long, Double>();
		for(long srgId : srgIds_thisNetwork)
		{
			out.put(srgId, getSRGAvailability(srgId));
		}
		
		return out;
	}

	/**
	 * Returns the availability per SRG, in increasing order of identifier.
	 * 
	 * @return Availability per SRG
	 * @since 0.3.0
	 */
	public double[] getSRGAvailabilityVector()
	{
		return DoubleUtils.toArray(getSRGAvailabilityMap().values());
	}
	
    /**
     * Returns the identifier of the first SRG with a given value in a 
     * certain attribute, or -1 if not found.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifier of the first SRG with a given value in a certain attribute, or -1 if not found
     * @since 0.3.0
     */
    public long getSRGByAttribute(String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> srgIds_thisNetPlan = getSRGIds();
        for(long srgId : srgIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisSRG = getSRGAttributeMap(srgId);
            if (value.equals(attributeMap_thisSRG.get(attribute)))
                return srgId;
        }

        return -1;
    }

	/**
	 * Returns the set of identifiers of the SRGs. The output set is backed in 
	 * the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 *
	 * @return SRG identifiers
	 * @since 0.3.0
	 */
	public Set<Long> getSRGIds()
	{
		return Collections.unmodifiableSet(srgIds);
	}

	/**
	 * Returns the set of identifiers of the SRGs, in increasing order of identifier.
	 *
	 * @return SRG identifiers
	 * @since 0.3.0
	 */
	public long[] getSRGIdsVector()
	{
		return LongUtils.toArray(getSRGIds());
	}
	
	/**
	 * Returns the set of links in a given SRG. The output set is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 * 
	 * @param srgId SRG identifier
	 * @return Set of links in the SRG
	 * @since 0.3.0
	 */
	public Set<Long> getSRGLinks(long srgId)
	{
		long layerId = getLayerDefaultId();
		return getSRGLinks(layerId, srgId);
	}

	/**
	 * Returns the set of links in a given SRG. The output set is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 * 
	 * @param layerId Layer identifier
	 * @param srgId SRG identifier
	 * @return Set of links in the SRG
	 * @since 0.3.0
	 */
	public Set<Long> getSRGLinks(long layerId, long srgId)
	{
		checkLayerIsActive(layerId);
		checkSRGIsActive(srgId);
		return Collections.unmodifiableSet(srgLinks.get(srgId).get(layerId));
	}

	/**
	 * Returns the set of links in a given SRG per layer.
	 * 
	 * @param srgId SRG identifier
	 * @return Set of links in the SRG per layer
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGLinksAllLayers(long srgId)
	{
		checkSRGIsActive(srgId);
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		for(Entry<Long, Set<Long>> entry : srgLinks.get(srgId).entrySet())
			out.put(entry.getKey(), Collections.unmodifiableSet(entry.getValue()));
		
		return out;
	}
	
	/**
	 * Returns the set of links per SRG.
	 * 
	 * @return Set of links per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGLinksMap()
	{
		long layerId = getLayerDefaultId();
		return getSRGLinksMap(layerId);
	}

	/**
	 * Returns the set of links per SRG.
	 * 
	 * @param layerId Layer identifier
	 * @return Set of links per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGLinksMap(long layerId)
	{
		checkLayerIsActive(layerId);
		
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		Set<Long> srgIds_thisNetPlan = getSRGIds();
		for(long srgId : srgIds_thisNetPlan)
		{
			out.put(srgId, getSRGLinks(layerId, srgId));
		}
		
		return out;
	}
	
	/**
	 * Returns the Mean Time To Fail (in hours) of the given SRG.
	 * 
	 * @param srgId SRG identifier
	 * @return Mean Time To Fail (in hours)
	 * @since 0.3.0
	 */
	public double getSRGMeanTimeToFailInHours(long srgId)
	{
		checkSRGIsActive(srgId);
		return srgMeanTimeToFailInHours.get(srgId);
	}
	
	/**
	 * Returns the Mean Time To Fail (in hours) per SRG. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 * 
	 * @return Mean Time To Fail (in hours) per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Double> getSRGMeanTimeToFailInHoursMap()
	{
		return Collections.unmodifiableMap(srgMeanTimeToFailInHours);
	}

	/**
	 * Returns the Mean Time To Fail (in hours) per SRG, in increasing order of identifier.
	 * 
	 * @return Mean Time To Fail (in hours) per SRG
	 * @since 0.3.0
	 */
	public double[] getSRGMeanTimeToFailInHoursVector()
	{
		return DoubleUtils.toArray(getSRGMeanTimeToFailInHoursMap().values());
	}
	
	/**
	 * Returns the Mean Time To Repair (in hours) of the given SRG.
	 * 
	 * @param srgId SRG identifier
	 * @return Mean Time To Repair (in hours)
	 * @since 0.3.0
	 */
	public double getSRGMeanTimeToRepairInHours(long srgId)
	{
		checkSRGIsActive(srgId);
		return srgMeanTimeToRepairInHours.get(srgId);
	}
	
	/**
	 * Returns the Mean Time To Repair (in hours) per SRG. The output map is backed 
	 * in the {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 * 
	 * @return Mean Time To Repair (in hours) per SRG
	 * @since 0.2.3
	 */
	public Map<Long, Double> getSRGMeanTimeToRepairInHoursMap()
	{
		return Collections.unmodifiableMap(srgMeanTimeToRepairInHours);
	}

	/**
	 * Returns the Mean Time To Repair (in hours) per SRG, in increasing order of identifier.
	 * 
	 * @return Mean Time To Repair (in hours) per SRG
	 * @since 0.3.0
	 */
	public double[] getSRGMeanTimeToRepairInHoursVector()
	{
		return DoubleUtils.toArray(getSRGMeanTimeToRepairInHoursMap().values());
	}

	/**
	 * Returns the next identifier for a new SRG.
	 *
	 * @return Next SRG identifier
	 * @since 0.3.0
	 */
	public long getSRGNextId()
	{
		long srgId = nextSRGId.longValue();
		if (srgId < 0) throw new Net2PlanException(String.format(TEMPLATE_NO_MORE_SRGS_ALLOWED));
		
		return srgId;
	}
	
	/**
	 * Returns the set of nodes in a given SRG. The output set is backed in the 
	 * {@code NetPlan} object, hence users should copy it in case they want to modify it.
	 * 
	 * @param srgId SRG identifier
	 * @return Set of nodes in the SRG
	 * @since 0.3.0
	 */
	public Set<Long> getSRGNodes(long srgId)
	{
		checkSRGIsActive(srgId);
		return Collections.unmodifiableSet(srgNodes.get(srgId));
	}

	/**
	 * Returns the set of nodes per SRG.
	 * 
	 * @return Set of nodes per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGNodesMap()
	{
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		Set<Long> srgIds_thisNetPlan = getSRGIds();
		for(long srgId : srgIds_thisNetPlan)
			out.put(srgId, getSRGNodes(srgId));
		
		return out;
	}
	
	/**
	 * <p>Returns the set of protection segments affected by a given SRG.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param srgId SRG identifier
	 * @return Set of protection segments
	 * @since 0.3.0
	 */
	public Set<Long> getSRGProtectionSegments(long srgId)
	{
		long layerId = getLayerDefaultId();
		return getSRGProtectionSegments(layerId, srgId);
	}

	/**
	 * <p>Returns the set of protection segments affected by a given SRG.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param srgId SRG identifier
	 * @return Set of protection segments
	 * @since 0.3.0
	 */
	public Set<Long> getSRGProtectionSegments(long layerId, long srgId)
	{
		Set<Long> out = new LinkedHashSet<Long>();
		Set<Long> nodeIds_thisSRG = getSRGNodes(srgId);
		Set<Long> linkIds_thisSRG = getSRGLinks(layerId, srgId);
		for(long nodeId : nodeIds_thisSRG) out.addAll(getNodeAssociatedProtectionSegments(layerId, nodeId));
		for(long linkId : linkIds_thisSRG) out.addAll(getLinkAssociatedProtectionSegments(layerId, linkId));
		
		return out;
	}
	
	/**
	 * <p>Returns the set of protection segments per SRG.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Set of protection segments per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGProtectionSegmentsMap()
	{
		long layerId = getLayerDefaultId();
		return getSRGProtectionSegmentsMap(layerId);
	}
	
	/**
	 * <p>Returns the set of protection segments per SRG.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Set of protection segments per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGProtectionSegmentsMap(long layerId)
	{
		Set<Long> srgIds_thisNetPlan = getSRGIds();
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		for(long srgId : srgIds_thisNetPlan) out.put(srgId, getSRGProtectionSegments(layerId, srgId));
		
		return out;
	}

	/**
	 * <p>Returns the set of routes affected by a given SRG.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param srgId SRG identifier
	 * @return Set of routes
	 * @since 0.3.0
	 */
	public Set<Long> getSRGRoutes(long srgId)
	{
		long layerId = getLayerDefaultId();
		return getSRGRoutes(layerId, srgId);
	}

	/**
	 * <p>Returns the set of routes affected by a given SRG.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param srgId SRG identifier
	 * @return Set of routes
	 * @since 0.3.0
	 */
	public Set<Long> getSRGRoutes(long layerId, long srgId)
	{
		Set<Long> out = new LinkedHashSet<Long>();
		Set<Long> nodeIds_thisSRG = getSRGNodes(srgId);
		Set<Long> linkIds_thisSRG = getSRGLinks(layerId, srgId);
		for(long nodeId : nodeIds_thisSRG) out.addAll(getNodeAssociatedRoutes(layerId, nodeId));
		for(long linkId : linkIds_thisSRG) out.addAll(getLinkAssociatedRoutes(layerId, linkId));
		
		return out;
	}

	/**
	 * <p>Returns the set of routes per SRG.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return Set of routes per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGRoutesMap()
	{
		long layerId = getLayerDefaultId();
		return getSRGRoutesMap(layerId);
	}
	
	/**
	 * <p>Returns the set of routes per SRG.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return Set of routes per SRG
	 * @since 0.3.0
	 */
	public Map<Long, Set<Long>> getSRGRoutesMap(long layerId)
	{
		Set<Long> srgIds_thisNetPlan = getSRGIds();
		Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
		for(long srgId : srgIds_thisNetPlan) out.put(srgId, getSRGRoutes(layerId, srgId));
		
		return out;
	}

	/**
	 * Returns the value of an attribute for each SRG.
	 *
	 * @param key Attribute name
	 * @return Attribute values
	 * @since 0.3.0
	 */
	public Map<Long, String> getSRGsAttributeMap(String key)
	{
		Map<Long, String> out = new LinkedHashMap<Long, String>();
		Set<Long> srgIds_thisNetPlan = getSRGIds();
		for(long srgId : srgIds_thisNetPlan)
			out.put(srgId, getSRGAttribute(srgId, key));
		
		return out;
	}
	
    /**
     * Returns the identifiers of the SRGs with a given value in a certain attribute.
     * 
     * @param attribute Attribute (not null)
     * @param value Value (not null)
     * @return Identifiers of the SRGs with a given value in a certain attribute
     * @since 0.3.0
     */
    public Set<Long> getSRGsByAttribute(String attribute, String value)
    {
        if (attribute == null || value == null) throw new Net2PlanException("Attribute and value cannot be null");

        Set<Long> out = new LinkedHashSet<Long>();

        Set<Long> srgIds_thisNetPlan = getSRGIds();
        for(long srgId : srgIds_thisNetPlan)
        {
            Map<String, String> attributeMap_thisSRG = getSRGAttributeMap(srgId);
            if (value.equals(attributeMap_thisSRG.get(attribute)))
                out.add(srgId);
        }

        return out;
    }

	/**
	 * Returns the aggregate traffic between each node pair.
	 * 
	 * @return Aggregate traffic per node pair
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getTrafficMap()
	{
		long layerId = getLayerDefaultId();
		return getTrafficMap(layerId);
	}
	
	/**
	 * Returns the aggregate traffic between each node pair.
	 * 
	 * @param layerId Layer identifier
	 * @return Aggregate traffic per node pair
	 * @since 0.3.0
	 */
	public Map<Pair<Long, Long>, Double> getTrafficMap(long layerId)
	{
		Map<Long, Pair<Long, Long>> demandMap_thisLayer = getDemandMap(layerId);
		Map<Long, Double> h_d = getDemandOfferedTrafficMap(layerId);
		
		Map<Pair<Long, Long>, Double> trafficMap = new LinkedHashMap<Pair<Long, Long>, Double>();
		for(Entry<Long, Pair<Long, Long>> entry : demandMap_thisLayer.entrySet())
		{
			long demandId = entry.getKey();
			Pair<Long, Long> nodePair = entry.getValue();
			double h_d_thisNodePair = h_d.get(demandId);
			if (trafficMap.containsKey(nodePair)) trafficMap.put(nodePair, trafficMap.get(nodePair) + h_d_thisNodePair);
			else trafficMap.put(nodePair, h_d_thisNodePair);
		}
		
		return trafficMap;
	}
	
	/**
	 * Returns the traffic matrix, where rows and columns represent the ingress 
	 * node and the egress node, respectively, in increasing order of identifier.
	 *
	 * @return Traffic matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getTrafficMatrix()
	{
		long layerId = getLayerDefaultId();
		return getTrafficMatrix(layerId);
	}

	/**
	 * Returns the traffic matrix, where rows and columns represent the ingress 
	 * node and the egress node, respectively, in increasing order of identifier.
	 *
	 * @param layerId Layer identifier
	 * @return Traffic matrix
	 * @since 0.3.1
	 */
	public DoubleMatrix2D getTrafficMatrix(long layerId)
	{
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		Set<Long> demandIds_thisLayer = getDemandIds(layerId);
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds);
		Map<Long, Double> h_d = getDemandOfferedTrafficMap(layerId);
		
		int N = nodeIds_thisNetPlan.size();
		DoubleMatrix2D trafficMatrix = DoubleFactory2D.dense.make(N, N);
		for (long demandId : demandIds_thisLayer)
		{
			long ingressNodeId = getDemandIngressNode(layerId, demandId);
			int a_d = nodeId2LinearIndex.get(ingressNodeId);
			long egressNodeId = getDemandEgressNode(layerId, demandId);
			int b_d = nodeId2LinearIndex.get(egressNodeId);

			trafficMatrix.setQuick(a_d, b_d, trafficMatrix.getQuick(a_d, b_d) + h_d.get(demandId));
		}

		return trafficMatrix;
	}
	
	/**
	 * Returns {@code true} if there is at least one demand. It is equivalent to {@code {@link #getNumberOfDemands() getNumberOfDemands()} > 0}.
	 * 
	 * @return {@code true} if there are demands, and {@code false} otherwise
	 * @since 0.2.0
	 */
	public boolean hasDemands()
	{
		long layerId = getLayerDefaultId();
		return hasDemands(layerId);
	}

	/**
	 * Returns {@code true} if there is at least one demand. It is equivalent to {@code {@link #getNumberOfDemands(long) getNumberOfDemands(layerId)} > 0}.
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if there are demands, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasDemands(long layerId)
	{
		return getNumberOfDemands(layerId) > 0;
	}

	/**
	 * <p>Returns {@code true} if there is at least one forwarding rule. It is equivalent to {@code {@link #getNumberOfForwardingRules() getNumberOfForwardingRules()} > 0}.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @return {@code true} if there are forwarding rules, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasForwardingRules()
	{
		long layerId = getLayerDefaultId();
		return hasForwardingRules(layerId);
	}
	
	/**
	 * <p>Returns {@code true} if there is at least one forwarding rule. It is equivalent to {@code {@link #getNumberOfForwardingRules(long) getNumberOfForwardingRules(layerId)} > 0}.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if there are forwarding rules, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasForwardingRules(long layerId)
	{
		return getNumberOfForwardingRules(layerId) > 0;
	}
	
	/**
	 * Returns {@code true} if there is at least one link. It is equivalent to {@code {@link #getNumberOfLinks() getNumberOfLinks()} > 0}.
	 * 
	 * @return {@code true} if there are links, and {@code false} otherwise
	 * @since 0.2.0
	 */
	public boolean hasLinks()
	{
		long layerId = getLayerDefaultId();
		return hasLinks(layerId);
	}

	/**
	 * Returns {@code true} if there is at least one link. It is equivalent to {@code {@link #getNumberOfLinks(long) getNumberOfLinks(layerId)} > 0}.
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if there are links, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasLinks(long layerId)
	{
		return getNumberOfLinks(layerId) > 0;
	}
	
	/**
	 * Returns {@code true} if there is at least one node. It is equivalent to {@code {@link #getNumberOfNodes() getNumberOfNodes()} > 0}.
	 * 
	 * @return {@code true} if there are nodes, and {@code false} otherwise
	 * @since 0.2.0
	 */
	public boolean hasNodes()
	{
		return getNumberOfNodes() > 0;
	}

	/**
	 * <p>Returns {@code true} if there is at least one protection segment. It is equivalent to {@code {@link #getNumberOfProtectionSegments() getNumberOfProtectionSegments()} > 0}.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return {@code true} if there are protection segments, and {@code false} otherwise
	 * @since 0.2.0
	 */
	public boolean hasProtectionSegments()
	{
		long layerId = getLayerDefaultId();
		return hasProtectionSegments(layerId);
	}

	/**
	 * <p>Returns {@code true} if there is at least one protection segment. It is equivalent to {@code {@link #getNumberOfProtectionSegments(long) getNumberOfProtectionSegments(layerId)} > 0}.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if there are protection segments, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasProtectionSegments(long layerId)
	{
		return getNumberOfProtectionSegments(layerId) > 0;
	}
	
	/**
	 * <p>Returns {@code true} if there is at least one route. It is equivalent to {@code {@link #getNumberOfRoutes() getNumberOfRoutes()} > 0}.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return {@code true} if there are routes, and {@code false} otherwise
	 * @since 0.2.0
	 */
	public boolean hasRoutes()
	{
		long layerId = getLayerDefaultId();
		return hasRoutes(layerId);
	}

	/**
	 * <p>Returns {@code true} if there is at least one route. It is equivalent to {@code {@link #getNumberOfRoutes(long) getNumberOfRoutes(layerId)} > 0}.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if there are routes, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasRoutes(long layerId)
	{
		return getNumberOfRoutes(layerId) > 0;
	}
	
	/**
	 * Indicates whether or not at least one demand has routing loops.
	 * 
	 * @return {@code true} if at least one demand has routing loops, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasRoutingLoops()
	{
		long layerId = getLayerDefaultId();
		return hasRoutingLoops(layerId);
	}

	/**
	 * Indicates whether or not at least one demand has routing loops.
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if at least one demand has routing loops, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean hasRoutingLoops(long layerId)
	{
		switch(getRoutingType(layerId))
		{
			case SOURCE_ROUTING:
				Set<Long> routeIds_thisLayer = getRouteIds(layerId);
				for(long routeId : routeIds_thisLayer)
				{
					List<Long> seqNodes = getRouteSequenceOfNodes(layerId, routeId);
					if (LongUtils.unique(seqNodes).size() != seqNodes.size()) return true;
				}

				return false;
				
			case HOP_BY_HOP_ROUTING:
				Set<Long> demandIds_thisLayer = getDemandIds(layerId);
				for(long demandId : demandIds_thisLayer)
					if (getDemandRoutingCycleType(layerId, demandId) != RoutingCycleType.LOOPLESS)
						return true;
				return false;
				
			default:
				throw new RuntimeException("Bad");
		}
	}
	
	/**
	 * <p>Returns {@code true} if there is at least one SRG. It is equivalent to {@code {@link #getNumberOfSRGs() getNumberOfSRGs()} > 0}.</p>
	 * 
	 * @return {@code true} if there are SRGs, and {@code false} otherwise
	 * @since 0.2.2
	 */
	public boolean hasSRGs()
	{
		return getNumberOfSRGs() > 0;
	}
	
	/**
	 * Indicates whether or not action monitoring is activated.
	 * 
	 * @return {@code true} if action monitoring is activated, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isActionMonitoringActivated()
	{
		return monitoringActivated;
	}
	
	/**
	 * Indicates whether or not a demand is active.
	 * 
	 * @param demandId Demand identifier
	 * @return {@code true} if demand is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isDemandActive(long demandId)
	{
		long layerId = getLayerDefaultId();
		return isDemandActive(layerId, demandId);
	}

	/**
	 * Indicates whether or not a demand is active.
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return {@code true} if demand is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isDemandActive(long layerId, long demandId)
	{
		checkLayerIsActive(layerId);
		return demandIds.get(layerId).contains(demandId);
	}
	
	/**
	 * <p>Indicates whether or a demand is bifurcated, that is, traffic is being sent 
	 * through more than one traffic route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param demandId Demand identifier
	 * @return {@code true} if traffic for the demand is bifurcated, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isDemandBifurcated(long demandId)
	{
		long layerId = getLayerDefaultId();
		return isDemandBifurcated(layerId, demandId);
	}
	
	/**
	 * <p>Indicates whether or a demand is bifurcated, that is, traffic is being sent 
	 * through more than one traffic route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return {@code true} if traffic for the demand is bifurcated, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isDemandBifurcated(long layerId, long demandId)
	{
		return getDemandBifurcationDegree(layerId, demandId) > 1;
	}
	
	/**
	 * Indicates whether or not a demand is coupled to an upper layer link.
	 * 
	 * @param demandId Demand identifier
	 * @return {@code true} if the demand is coupled, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isDemandCoupled(long demandId)
	{
		long layerId = getLayerDefaultId();
		return isDemandCoupled(layerId, demandId);
	}
	
	/**
	 * Indicates whether or not a demand is coupled to an upper layer link.
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @return {@code true} if the demand is coupled, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isDemandCoupled(long layerId, long demandId)
	{
		checkDemandIsActive(layerId, demandId);
		return lowerLayerToUpperLayerMapping.contains(layerId, demandId);
	}
	
	/**
	 * Returns {@code true} if the network is empty (i.e. without nodes...). It is equivalent to {@code !{@link #hasNodes hasNodes()}}.
	 *
	 * @return {@code true} if the network is empty, and {@code false} otherwise
	 * @since 0.2.0
	 */
	public boolean isEmpty()
	{
		return !hasNodes();
	}

	/**
	 * <p>Indicates whether or not a forwarding rule is active.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return {@code true} if forwarding rule is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isForwardingRuleActive(Pair<Long, Long> demandLinkPair)
	{
		long layerId = getLayerDefaultId();
		return isForwardingRuleActive(layerId, demandLinkPair);
	}
	
	/**
	 * <p>Indicates whether or not a forwarding rule is active.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return {@code true} if forwarding rule is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isForwardingRuleActive(long layerId, Pair<Long, Long> demandLinkPair)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		long demandId = demandLinkPair.getFirst();
		long linkId = demandLinkPair.getSecond();
		checkDemandIsActive(layerId, demandId);
		checkLinkIsActive(layerId, linkId);
		
		return forwardingRuleDemandLinkPairMap.get(layerId).contains(demandLinkPair);
	}
	
	/**
	 * <p>Indicates whether or not a forwarding rule is down. A forwarding rule is 
	 * down if: (i) the node where the rule is installed is down, (ii) the associated 
	 * outgoing link is down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return {@code true} if the forwarding rule is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isForwardingRuleDown(Pair<Long, Long> demandLinkPair)
	{
		long layerId = getLayerDefaultId();
		return isForwardingRuleActive(layerId, demandLinkPair);
	}
	
	/**
	 * <p>Indicates whether or not a forwarding rule is down. A forwarding rule is 
	 * down if: (i) the node where the rule is installed is down, (ii) the associated 
	 * outgoing link is down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @return {@code true} if the forwarding rule is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isForwardingRuleDown(long layerId, Pair<Long, Long> demandLinkPair)
	{
		checkRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		long demandId = demandLinkPair.getFirst();
		long linkId = demandLinkPair.getSecond();
		checkDemandIsActive(layerId, demandId);
		checkLinkIsActive(layerId, linkId);
		
		return forwardingRuleDemandLinkPairDownMap.get(layerId).contains(demandLinkPair);
	}
	
	/**
	 * Indicates whether or not a layer is active.
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if layer is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLayerActive(long layerId)
	{
		return layerIds.contains(layerId);
	}

	/**
	 * Indicates whether or not a further coupling between two layers would be valid.
	 * This can be used to check if a new coupling would create cycles in the topology 
	 * of layers, that is, if layer hierarchy is not broken.
	 * 
	 * @param lowerLayerId Lower layer identifier
	 * @param upperLayerId Upper layer identifier
	 * @return {@code true} if coupling would be valid, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLayerCouplingValid(long lowerLayerId, long upperLayerId)
	{
		checkLayerIsActive(lowerLayerId);
		checkLayerIsActive(upperLayerId);
		
		DemandLinkMapping coupling_thisLayerPair = interLayerCoupling.getEdge(lowerLayerId, upperLayerId);
		if (coupling_thisLayerPair == null)
		{
			coupling_thisLayerPair = new DemandLinkMapping();
			boolean valid;
			try { valid = interLayerCoupling.addDagEdge(lowerLayerId, upperLayerId, coupling_thisLayerPair); }
			catch (DirectedAcyclicGraph.CycleFoundException ex) { valid = false; }
			if (valid) interLayerCoupling.removeEdge(lowerLayerId, upperLayerId);
			else return false;
		}
		
		return true;
	}
	
	/**
	 * Indicates whether or not a link is active.
	 * 
	 * @param linkId Link identifier
	 * @return {@code true} if link is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkActive(long linkId)
	{
		long layerId = getLayerDefaultId();
		return isLinkActive(layerId, linkId);
	}

	/**
	 * Indicates whether or not a link is active.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return {@code true} if link is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkActive(long layerId, long linkId)
	{
		checkLayerIsActive(layerId);
		return linkIds.get(layerId).contains(linkId);
	}

	/**
	 * Indicates whether or not a link is a bottleneck, that is, if has the largest 
	 * utilization among the link set.
	 * 
	 * @param linkId Link identifier
	 * @return {@code true} if link is a bottleneck, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkBottleneck(long linkId)
	{
		long layerId = getLayerDefaultId();
		return isLinkBottleneck(layerId, linkId);
	}
	
	/**
	 * Indicates whether or not a link is a bottleneck, that is, if has the largest 
	 * utilization among the link set.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return {@code true} if link is a bottleneck, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkBottleneck(long layerId, long linkId)
	{
		double rho_e = getLinkUtilization(layerId, linkId);
		double max_rho_e = getLinkMaximumUtilization(layerId);
		
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		if (max_rho_e < PRECISION_FACTOR) return false;
		
		return DoubleUtils.isEqualWithinRelativeTolerance(max_rho_e, rho_e, PRECISION_FACTOR);
	}
	
	/**
	 * Indicates whether or not a link is coupled to a lower layer demand.
	 * 
	 * @param linkId Link identifier
	 * @return {@code true} if the link is coupled, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkCoupled(long linkId)
	{
		long layerId = getLayerDefaultId();
		return isLinkCoupled(layerId, linkId);
	}
	
	/**
	 * Indicates whether or not a link is coupled to a lower layer demand.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return {@code true} if the link is coupled, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkCoupled(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return upperLayerToLowerLayerMapping.contains(layerId, linkId);
	}
	
	/**
	 * Indicates whether or not a link is down.
	 * 
	 * @param linkId Link identifier
	 * @return {@code true} if the link is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkDown(long linkId)
	{
		long layerId = getLayerDefaultId();
		return isLinkDown(layerId, linkId);
	}
	
	/**
	 * Indicates whether or not a link is down.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @return {@code true} if the link is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isLinkDown(long layerId, long linkId)
	{
		checkLinkIsActive(layerId, linkId);
		return linksDown.get(layerId).contains(linkId);
	}

	/**
	 * Indicates whether the network design is multilayer. It is equivalent to {@link #isSingleLayer() !isSingleLayer()}.
	 * 
	 * @return {@code true} if the network design is multilayer, and {@code false} otherwise.
	 * @since 0.3.0
	 */
	public boolean isMultiLayer()
	{
		return !isSingleLayer();
	}

	/**
	 * Indicates whether or not a node is active.
	 * 
	 * @param nodeId Node identifier
	 * @return {@code true} if node is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isNodeActive(long nodeId)
	{
		return nodeIds.contains(nodeId);
	}

	/**
	 * Indicates whether or not a node is down.
	 * 
	 * @param nodeId Node identifier
	 * @return {@code true} if the node is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isNodeDown(long nodeId)
	{
		checkNodeIsActive(nodeId);
		return nodesDown.contains(nodeId);
	}
	
	/**
	 * Indicates whether or not a path (sequence of links, and maybe protection segments) 
	 * is restorable, that is, whether the whole set of traversed resources is up.
	 * 
	 * @param path Sequence of links and protection segments
	 * @return {@code true} if the path is restorable, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isPathRestorable(List<Long> path)
	{
		long layerId = getLayerDefaultId();
		return isPathRestorable(layerId, path);
	}
	
	/**
	 * Indicates whether or not a path (sequence of links, and maybe protection segments) 
	 * is restorable, that is, whether the whole set of traversed resources is up.
	 * 
	 * @param layerId Layer identifier
	 * @param path Sequence of links and protection segments
	 * @return {@code true} if the path is restorable, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isPathRestorable(long layerId, List<Long> path)
	{
		Set<Long> linksDown_thisLayer = getLinksDown(layerId);
		Set<Long> segmentsDown_thisLayer = getProtectionSegmentsDown(layerId);
		for(long linkId : path)
		{
			if (linkId >= 0)
			{
				if (linksDown_thisLayer.contains(linkId)) return false;
			}
			else
			{
				long segmentId = -1 - linkId;
				if (segmentsDown_thisLayer.contains(segmentId)) return false;
			}
		}

		Set<Long> nodesDown_thisNetPlan = getNodesDown();
		List<Long> actualSeqLinks = convertSequenceOfLinksAndProtectionSegmentsToSequenceOfLinks(path);
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = getLinkMap(layerId);
		List<Long> seqNodes = GraphUtils.convertSequenceOfLinksToSequenceOfNodes(linkMap_thisLayer, actualSeqLinks);
		return !CollectionUtils.containsAny(seqNodes, nodesDown_thisNetPlan);
	}
	
	/**
	 * <p>Indicates whether or not a protection segment is active.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if protection segment is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentActive(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return isProtectionSegmentActive(layerId, segmentId);
	}

	/**
	 * <p>Indicates whether or not a protection segment is active.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if protection segment is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentActive(long layerId, long segmentId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return segmentIds.get(layerId).contains(segmentId);
	}

	/**
	 * <p>Indicates whether or not a protection segment is dedicated, that is, if it 
	 * is associated to only one route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if the protection segment is dedicated. Otherwise, {@code false}
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentDedicated(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return isProtectionSegmentDedicated(layerId, segmentId);
	}
	
	/**
	 * <p>Indicates whether or not a protection segment is dedicated, that is, if it 
	 * is associated to only one route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if the protection segment is dedicated. Otherwise, {@code false}
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentDedicated(long layerId, long segmentId)
	{
		return getProtectionSegmentRoutes(layerId, segmentId).size() == 1;
	}
	
	/**
	 * <p>Indicates whether or not a protection segment is down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if the protection segment is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentDown(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return isProtectionSegmentDown(layerId, segmentId);
	}

	/**
	 * <p>Indicates whether or not a protection segment is down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if the protection segment is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentDown(long layerId, long segmentId)
	{
		return getProtectionSegmentsDown(layerId).contains(segmentId);
	}
	
	/**
	 * <p>Indicates whether or not a protection segment is restorable, that is, whether the 
	 * whole set of resources is up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if the protection segment is restorable, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentRestorable(long segmentId)
	{
		long layerId = getLayerDefaultId();
		return isProtectionSegmentRestorable(layerId, segmentId);
	}
	
	/**
	 * <p>Indicates whether or not a protection segment is restorable, that is, whether the 
	 * whole set of resources is up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @return {@code true} if the protection segment is restorable, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isProtectionSegmentRestorable(long layerId, long segmentId)
	{
		if (!isProtectionSegmentDown(layerId, segmentId)) throw new Net2PlanException("Protection segment " + segmentId + " at layer " + layerId + " is not down");

		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		return isPathRestorable(layerId, seqLinks);
	}
	
	/**
	 * <p>Indicates whether or not a route is active.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return {@code true} if route is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isRouteActive(long routeId)
	{
		long layerId = getLayerDefaultId();
		return isRouteActive(layerId, routeId);
	}

	/**
	 * <p>Indicates whether or not a route is active.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return {@code true} if route is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isRouteActive(long layerId, long routeId)
	{
		checkRoutingType(layerId, RoutingType.SOURCE_ROUTING);
		return routeIds.get(layerId).contains(routeId);
	}
	
	/**
	 * <p>Indicates whether or not a route is down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return {@code true} if the route is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isRouteDown(long routeId)
	{
		long layerId = getLayerDefaultId();
		return isRouteDown(layerId, routeId);
	}

	/**
	 * <p>Indicates whether or not a route is down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return {@code true} if the route is down, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isRouteDown(long layerId, long routeId)
	{
		return getRoutesDown(layerId).contains(routeId);
	}
	
	/**
	 * <p>Indicates whether or not a route is restorable, that is, whether the 
	 * whole set of resources is up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param routeId Route identifier
	 * @return {@code true} if the route is restorable, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isRouteRestorable(long routeId)
	{
		long layerId = getLayerDefaultId();
		return isRouteRestorable(layerId, routeId);
	}
	
	/**
	 * <p>Indicates whether or not a route is restorable, that is, whether the 
	 * whole set of resources is up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @return {@code true} if the route is restorable, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isRouteRestorable(long layerId, long routeId)
	{
		if (!isRouteDown(layerId, routeId)) throw new Net2PlanException("Route " + routeId + " at layer " + layerId + " is not down");
		
		List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
		return isPathRestorable(layerId, seqLinks);
	}
	
	/**
	 * <p>Indicates whether or not routing is bifurcated, that is, whether any demand 
	 * is being actually carried through more than one route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @return {@code true} if routing is bifurcated, and {@code false} otherwise
	 * @since 0.3..0
	 */
	public boolean isRoutingBifurcated()
	{
		long layerId = getLayerDefaultId();
		return isRoutingBifurcated(layerId);
	}
	
	/**
	 * <p>Indicates whether or not routing is bifurcated, that is, whether any demand 
	 * is being actually carried through more than one route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @return {@code true} if routing is bifurcated, and {@code false} otherwise
	 * @since 0.3..0
	 */
	public boolean isRoutingBifurcated(long layerId)
	{
		Map<Long, Integer> aux = getDemandBifurcationDegreeMap();
		if (aux.isEmpty()) return false;
		
		return IntUtils.maxValue(aux) > 1;
	}
	
	/**
	 * Indicates whether the network design is multilayer. It is equivalent to {@link #getNumberOfLayers() getNumberOfLayers() == 1}.
	 * 
	 * @return {@code true} if the network design is multilayer, and {@code false} otherwise.
	 * @since 0.3.0
	 */
	public boolean isSingleLayer()
	{
		return getNumberOfLayers() == 1;
	}

	/**
	 * Indicates whether or not an SRG is active.
	 * 
	 * @param srgId SRG identifier
	 * @return {@code true} if SRG is active, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isSRGActive(long srgId)
	{
		return srgIds.contains(srgId);
	}
	
	/**
	 * Static factory method to get a {@link NetPlan NetPlan} from a {@code .n2p} file.
	 * 
	 * @param file {@code .n2p} file
	 * @return Network design
	 * @since 0.3.0
	 */
	public static NetPlan loadFromFile(File file)
	{
		try
		{
			InputStream inputStream = new FileInputStream(file);
			return new NetPlan(inputStream);
		}
		catch(FileNotFoundException e)
		{
			throw new Net2PlanException(e.getMessage());
		}
	}

	/**
	 * Static factory method to get a {@link NetPlan NetPlan} from an input stream.
	 * 
	 * @param inputStream Input stream
	 * @return Network design
	 * @since 0.3.1
	 */
	public static NetPlan loadFromInputStream(InputStream inputStream)
	{
		return new NetPlan(inputStream);
	}

	/**
	 * <p>Removes all forwarding rules associated to a demand.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void removeAllDemandForwardingRules(long demandId)
	{
		long layerId = getLayerDefaultId();
		removeAllDemandForwardingRules(layerId, demandId);
	}
	
	/**
	 * <p>Removes all forwarding rules associated to a demand.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void removeAllDemandForwardingRules(long layerId, long demandId)
	{
		removeForwardingRules(layerId, getDemandForwardingRules(layerId, demandId));
	}
	
	/**
	 * <p>Removes all the demands defined.</p>
	 *
	 * <p><b>Important</b>: All the associated routes and upper layer links will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} actions will be issued.</p>
	 * 
	 * @since 0.2.0
	 */
	public void removeAllDemands()
	{
		long layerId = getLayerDefaultId();
		removeAllDemands(layerId);
	}

	/**
	 * <p>Removes all the demands defined.</p>
	 *
	 * <p><b>Important</b>: All the associated routes and upper layer links will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void removeAllDemands(long layerId)
	{
		RoutingType routingType = getRoutingType(layerId);
		switch(routingType)
		{
			case SOURCE_ROUTING:
				removeAllRoutes(layerId);
				break;

			case HOP_BY_HOP_ROUTING:
				removeAllForwardingRules(layerId);
				break;

			default:
				throw new RuntimeException("Bad");
		}
		
		removeDemands(layerId, getDemandIds(layerId));
	}
	
	/**
	 * <p>Removes all forwarding rules.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @since 0.3.0
	 */
	public void removeAllForwardingRules()
	{
		long layerId = getLayerDefaultId();
		removeAllForwardingRules(layerId);
	}
	
	/**
	 * <p>Removes all forwarding rules.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void removeAllForwardingRules(long layerId)
	{
		checkIsModifiable();
		
		disableCacheUpdating = true;
		for(Pair<Long, Long> demandLinkPair : new LinkedHashSet<Pair<Long, Long>>(getForwardingRules(layerId)))
			removeForwardingRule(layerId, demandLinkPair);
		
		disableCacheUpdating = false;

		for(long demandId : getDemandIds(layerId))
		{
			cache_demandCarriedTrafficMap.get(layerId).put(demandId, 0.0);
			cache_demandRoutingCycleType.get(layerId).put(demandId, RoutingCycleType.LOOPLESS);
			updateCacheDemandBlockedTraffic(layerId, demandId);
		}

		for(long linkId : getLinkIds(layerId))
		{
			cache_linkCarriedTrafficMap.get(layerId).put(linkId, 0.0);
		}
	}
	
	/**
	 * <p>Removes all forwarding rules associated to a link.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void removeAllLinkForwardingRules(long linkId)
	{
		long layerId = getLayerDefaultId();
		removeAllLinkForwardingRules(layerId, linkId);
	}
	
	/**
	 * <p>Removes all forwarding rules associated to a link.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void removeAllLinkForwardingRules(long layerId, long linkId)
	{
		removeForwardingRules(layerId, getLinkForwardingRules(layerId, linkId));
	}
	
	/**
	 * <p>Removes all the links defined.</p>
	 *
	 * <p><b>Important</b>: All the associated routes, protection segments and lower layer demands will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} actions will be issued.</p>
	 * 
	 * @since 0.2.0
	 */
	public void removeAllLinks()
	{
		long layerId = getLayerDefaultId();
		removeAllLinks(layerId);
	}

	/**
	 * <p>Removes all the links defined.</p>
	 *
	 * <p><b>Important</b>: All the associated routes, protection segments and lower layer demands will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void removeAllLinks(long layerId)
	{
		RoutingType routingType = getRoutingType(layerId);
		switch(routingType)
		{
			case SOURCE_ROUTING:
				removeAllRoutes(layerId);
				removeAllProtectionSegments(layerId);
				break;

			case HOP_BY_HOP_ROUTING:
				removeAllForwardingRules(layerId);
				break;

			default:
				throw new RuntimeException("Bad");
		}
		
		removeLinks(layerId, getLinkIds(layerId));
	}
	
	/**
	 * <p>Removes all forwarding rules associated to a node.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void removeAllNodeForwardingRules(long nodeId)
	{
		long layerId = getLayerDefaultId();
		removeAllNodeForwardingRules(layerId, nodeId);
	}
	
	/**
	 * <p>Removes all forwarding rules associated to a node.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void removeAllNodeForwardingRules(long layerId, long nodeId)
	{
		removeForwardingRules(layerId, getNodeForwardingRules(layerId, nodeId));
	}
	
	/**
	 * <p>Removes all the nodes defined within the network.</p>
	 *
	 * <p><b>Important</b>: All the links, traffic demands, traffic routes, protection segments and SRGs will be also removed. Layers are maintained.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_REMOVED NODE_REMOVED} actions will be issued.</p>
	 * 
	 * @since 0.2.0
	 */
	public void removeAllNodes()
	{
		checkIsModifiable();
		
		Set<Long> layerIds_thisNetPlan = getLayerIds();
		for(long layerId : layerIds_thisNetPlan)
		{
			RoutingType routingType = getRoutingType(layerId);
			switch(routingType)
			{
				case SOURCE_ROUTING:
					removeAllRoutes(layerId);
					removeAllProtectionSegments(layerId);
					break;
					
				case HOP_BY_HOP_ROUTING:
					removeAllForwardingRules(layerId);
					break;
					
				default:
					throw new RuntimeException("Bad");
			}
			
			removeAllLinks(layerId);
			removeAllDemands(layerId);
		}
		
		removeNodes(getNodeIds());
	}

	/**
	 * <p>Removes all the protection segments defined.</p>
	 *
	 * <p><b>Important</b>: All the list of protection segments from traffic routes will be cleared.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @since 0.2.0
	 */
	public void removeAllProtectionSegments()
	{
		long layerId = getLayerDefaultId();
		removeAllProtectionSegments(layerId);
	}

	/**
	 * <p>Removes all the protection segments defined.</p>
	 *
	 * <p><b>Important</b>: All the list of protection segments from traffic routes will be cleared.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void removeAllProtectionSegments(long layerId)
	{
		removeProtectionSegments(layerId, getProtectionSegmentIds(layerId));
	}

	/**
	 * <p>Removes all the routes defined.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @since 0.2.0
	 */
	public void removeAllRoutes()
	{
		long layerId = getLayerDefaultId();
		removeAllRoutes(layerId);
	}

	/**
	 * <p>Removes all the routes defined.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void removeAllRoutes(long layerId)
	{
		removeRoutes(layerId, getRouteIds(layerId));
	}
	
	/**
	 * <p>Removes all the routing information, irrespective of the routing type 
	 * setting. For source routing, all routes and protection segments are removed. 
	 * For hop-by-hop routing, all forwarding rules are removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED}, {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} or {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * @since 0.3.1
	 */
	public void removeAllRoutingInformation()
	{
		long layerId = getLayerDefaultId();
		removeAllRoutingInformation(layerId);
	}
	
	/**
	 * <p>Removes all the routing information, irrespective of the routing type 
	 * setting. For source routing, all routes and protection segments are removed. 
	 * For hop-by-hop routing, all forwarding rules are removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED}, {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} or {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED}actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.1
	 */
	public void removeAllRoutingInformation(long layerId)
	{
		checkIsModifiable();
		RoutingType routingType = getRoutingType(layerId);
		switch(routingType)
		{
			case SOURCE_ROUTING:
				removeAllRoutes(layerId);
				removeAllProtectionSegments(layerId);
				break;
				
			case HOP_BY_HOP_ROUTING:
				removeAllForwardingRules(layerId);
				break;
				
			default:
				throw new RuntimeException("Bad");
		}
	}
	
	/**
	 * Removes all the SRGs defined within the network.
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_REMOVED SRG_REMOVED} actions will be issued.</p>
	 * 
	 * @since 0.2.3
	 */
	public void removeAllSRGs()
	{
		removeSRGs(getSRGIds());
	}

	/**
	 * <p>Removes a demand, and any associated route or forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: if demand is coupled, its coupled link will be also removed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} action will be issued.</p>
	 *
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void removeDemand(long demandId)
	{
		long layerId = getLayerDefaultId();
		removeDemand(layerId, demandId);
	}

	/**
	 * <p>Removes a demand, and any associated route or forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: if demand is coupled, its coupled link will be also removed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @since 0.3.0
	 */
	public void removeDemand(long layerId, long demandId)
	{
		checkIsModifiable();
		
		RoutingType routingType = getRoutingType(layerId);
		if (routingType == RoutingType.SOURCE_ROUTING)
		{
			Set<Long> routeIds_thisDemand = getDemandRoutes(layerId, demandId);
			removeRoutes(layerId, routeIds_thisDemand);
		}
		else
		{
			Set<Pair<Long, Long>> forwardingRules_thisDemand = getDemandForwardingRules(layerId, demandId);
			removeForwardingRules(layerId, forwardingRules_thisDemand);
		}
		
		if (isDemandCoupled(layerId, demandId))
		{
			Pair<Long, Long> upperLayerLink = getDemandCoupledUpperLayerLink(layerId, demandId);
			decoupleDemand(layerId, demandId);
			
			long upperLayerId = upperLayerLink.getFirst();
			long linkId = upperLayerLink.getSecond();
			removeLink(upperLayerId, linkId);
		}

		Pair<Long, Long> endNodes = demandMap.get(layerId).get(demandId);
		long ingressNodeId = endNodes.getFirst();
		long egressNodeId = endNodes.getSecond();
		
		cache_nodeIncomingDemands.get(egressNodeId).get(layerId).remove(demandId);
		cache_nodeOutgoingDemands.get(ingressNodeId).get(layerId).remove(demandId);
		Map<String, String> attributes = demandAttributesMap.get(layerId).remove(demandId);
		demandIds.get(layerId).remove(demandId);
		double h_d = demandOfferedTrafficMap.get(layerId).remove(demandId);
		cache_demandCarriedTrafficMap.get(layerId).remove(demandId);
		cache_demandBlockedTrafficMap.get(layerId).remove(demandId);
		cache_demandBlockedIds.get(layerId).remove(demandId);
		demandMap.get(layerId).remove(demandId);
		demandRouteIdsMap.get(layerId).remove(demandId);
		cache_demandForwardingRuleMap.get(layerId).remove(demandId);
		cache_demandRoutingCycleType.get(layerId).remove(demandId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.demandRemoved(layerId, demandId, ingressNodeId, egressNodeId, h_d, attributes));
	}
	
	/**
	 * <p>Removes a given attribute from a demand. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 *
	 * @param demandId Demand identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeDemandAttribute(long demandId, String key)
	{
		long layerId = getLayerDefaultId();
		removeDemandAttribute(layerId, demandId, key);
	}

	/**
	 * <p>Removes a given attribute from a demand. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeDemandAttribute(long layerId, long demandId, String key)
	{
		checkIsModifiable();
		checkDemandIsActive(layerId, demandId);
		
		demandAttributesMap.get(layerId).get(demandId).remove(key);
		Pair<Long, Long> aux = getDemandCoupledUpperLayerLink(layerId, demandId);
		if (aux == null) return;
		
		long upperLayerId = aux.getFirst();
		long linkId = aux.getSecond();
		linkAttributesMap.get(upperLayerId).get(linkId).remove(key);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.demandModified(layerId, demandId));
	}

	/**
	 * <p>Removes a set of demands.</p>
	 *
	 * <p><b>Important</b>: All the associated routes and upper layer links will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} actions will be issued.</p>
	 * 
	 * @param demandIds Demand identifiers
	 * @since 0.3.0
	 */
	public void removeDemands(Set<Long> demandIds)
	{
		long layerId = getLayerDefaultId();
		removeDemands(layerId, demandIds);
	}
	
	/**
	 * <p>Removes a set of demands.</p>
	 *
	 * <p><b>Important</b>: All the associated routes and upper layer links will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandIds Demand identifiers
	 * @since 0.3.0
	 */
	public void removeDemands(long layerId, Set<Long> demandIds)
	{
		checkIsModifiable();
		for(long demandId : demandIds) checkDemandIsActive(layerId, demandId);
		
		for(long demandId : new LinkedHashSet<Long>(demandIds)) removeDemand(layerId, demandId);
	}

	/**
	 * <p>Removes a forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action will be issued.</p>
	 *
	 * @param demandLinkPair Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public void removeForwardingRule(Pair<Long, Long> demandLinkPair)
	{
		long layerId = getLayerDefaultId();
		removeForwardingRule(layerId, demandLinkPair);
	}
	
	/**
	 * <p>Removes a forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public void removeForwardingRule(long layerId, Pair<Long, Long> demandLinkPair)
	{
		checkIsModifiable();
		double splittingRatio = getForwardingRuleSplittingRatio(layerId, demandLinkPair);
		Map<String, String> attributes = getForwardingRuleAttributeMap(layerId, demandLinkPair);
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.forwardingRuleRemoved(layerId, demandLinkPair, splittingRatio, attributes));
		
		long demandId = demandLinkPair.getFirst();

		if (!disableCacheUpdating)
		{
			forwardingRuleSplittingRatioMap.get(layerId).put(demandLinkPair, 0.0);
			updateCacheDemandForwarding(layerId, demandId);
		}

		forwardingRuleAttributeMap.get(layerId).remove(demandLinkPair);
		forwardingRuleDemandLinkPairMap.get(layerId).remove(demandLinkPair);
		forwardingRuleSplittingRatioMap.get(layerId).remove(demandLinkPair);
		forwardingRuleCarriedTrafficMap.get(layerId).remove(demandLinkPair);
		forwardingRuleDemandLinkPairDownMap.get(layerId).remove(demandLinkPair);
		
		cache_demandForwardingRuleMap.get(layerId).get(demandId).remove(demandLinkPair);

		long linkId = demandLinkPair.getSecond();
		long originNodeId = getLinkOriginNode(layerId, linkId);
		cache_linkForwardingRuleMap.get(layerId).get(linkId).remove(demandLinkPair);
		cache_nodeForwardingRuleMap.get(originNodeId).get(layerId).remove(demandLinkPair);
	}
	
	/**
	 * <p>Removes a given attribute from a forwarding rule. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeForwardingRuleAttribute(Pair<Long, Long> demandLinkPair, String key)
	{
		long layerId = getLayerDefaultId();
		removeForwardingRuleAttribute(layerId, demandLinkPair, key);
	}

	/**
	 * <p>Removes a given attribute from a forwarding rule. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeForwardingRuleAttribute(long layerId, Pair<Long, Long> demandLinkPair, String key)
	{
		checkIsModifiable();
		checkForwardingRuleIsActive(layerId, demandLinkPair);
		
		forwardingRuleAttributeMap.get(layerId).get(demandLinkPair).remove(key);
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.forwardingRuleModified(layerId, demandLinkPair));
	}

	/**
	 * <p>Removes a set of forwarding rules.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPairs Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public void removeForwardingRules(Set<Pair<Long, Long>> demandLinkPairs)
	{
		long layerId = getLayerDefaultId();
		removeForwardingRules(layerId, demandLinkPairs);
	}
	
	/**
	 * <p>Removes a set of forwarding rules.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPairs Demand-outgoing link pair
	 * @since 0.3.0
	 */
	public void removeForwardingRules(long layerId, Set<Pair<Long, Long>> demandLinkPairs)
	{
		checkIsModifiable();
		
		disableCacheUpdating = true;
		Set<Long> demandIdsToUpdate = new LinkedHashSet<Long>();
		Set<Long> linkIdsToUpdate = new LinkedHashSet<Long>();
		for(Pair<Long, Long> demandLinkPair : new LinkedHashSet<Pair<Long, Long>>(demandLinkPairs))
		{
			long demandId = demandLinkPair.getFirst();
			if (!demandIdsToUpdate.contains(demandId))
			{
				Set<Pair<Long, Long>> forwardingRules_thisDemand = getDemandForwardingRules(layerId, demandId);
				for(Pair<Long, Long> auxDemandLinkPair : forwardingRules_thisDemand)
					linkIdsToUpdate.add(auxDemandLinkPair.getSecond());
				
				demandIdsToUpdate.add(demandId);
			}

			removeForwardingRule(layerId, demandLinkPair);
		}
		
		if (demandIdsToUpdate.isEmpty())
		{
			disableCacheUpdating = false;
			return;
		}
		
		for(long demandId : demandIdsToUpdate) updateCacheDemandForwarding(layerId, demandId);
		disableCacheUpdating = false;
		for(long linkId : linkIdsToUpdate) updateCacheLinkForwarding(layerId, linkId);
	}
	
	/**
	 * <p>Removes a layer, and any associated link, demand, route, protection segment or forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_REMOVED LAYER_REMOVED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void removeLayer(long layerId)
	{
		if (getNumberOfLayers() == 1) throw new Net2PlanException("At least one layer must be active");
		
		Set<Long> linkIds_thisLayer = new LinkedHashSet<Long>(getLinkIds(layerId));
		for(long linkId : linkIds_thisLayer)
		{
			if (isLinkCoupled(layerId, linkId))
			{
				Pair<Long, Long> aux = getLinkCoupledLowerLayerDemand(layerId, linkId);
				if (aux == null) continue;

				long lowerLayerId = aux.getFirst();
				long demandId = aux.getSecond();
				removeDemand(lowerLayerId, demandId);
			}
		}
		
		Set<Long> demandIds_thisLayer = new LinkedHashSet<Long>(getDemandIds(layerId));
		for(long demandId : demandIds_thisLayer)
		{
			if (isDemandCoupled(layerId, demandId))
			{
				Pair<Long, Long> aux = getDemandCoupledUpperLayerLink(layerId, demandId);
				if (aux == null) continue;

				long upperLayerId = aux.getFirst();
				long linkId = aux.getSecond();
				removeLink(upperLayerId, linkId);
			}
		}
		
		interLayerCoupling.removeVertex(layerId);
		
		layerIds.remove(layerId);
		layerAttributeMap.remove(layerId);
		layerDemandTrafficUnitsNameMap.remove(layerId);
		layerDescriptionMap.remove(layerId);
		layerNameMap.remove(layerId);
		layerLinkCapacityUnitsNameMap.remove(layerId);
		layerRoutingType.remove(layerId);
		forwardingRuleSplittingRatioMap.remove(layerId);
		forwardingRuleCarriedTrafficMap.remove(layerId);
		forwardingRuleAttributeMap.remove(layerId);
		forwardingRuleDemandLinkPairMap.remove(layerId);

		nextLinkId.remove(layerId);
		nextDemandId.remove(layerId);
		nextRouteId.remove(layerId);
		nextSegmentId.remove(layerId);

		linkIds.remove(layerId);
		linkMap.remove(layerId);
		linkSRGIds.remove(layerId);
		linkRouteIds.remove(layerId);
		linkSegmentIds.remove(layerId);
		linkCapacityMap.remove(layerId);
		cache_linkCarriedTrafficMap.remove(layerId);
		cache_linksOverSubscribed.remove(layerId);
		cache_linksWithUnlimitedCapacity.remove(layerId);
		cache_linksWithZeroCapacity.remove(layerId);
		linkReservedCapacityMap.remove(layerId);
		linkLengthInKmMap.remove(layerId);
		linkAttributesMap.remove(layerId);
		cache_linkForwardingRuleMap.remove(layerId);
		
		linksDown.remove(layerId);
		routesDown.remove(layerId);
		segmentsDown.remove(layerId);
		forwardingRuleDemandLinkPairDownMap.remove(layerId);

		demandIds.remove(layerId);
		demandMap.remove(layerId);
		demandOfferedTrafficMap.remove(layerId);
		cache_demandCarriedTrafficMap.remove(layerId);
		cache_demandBlockedTrafficMap.remove(layerId);
		cache_demandBlockedIds.remove(layerId);
		demandAttributesMap.remove(layerId);
		demandRouteIdsMap.remove(layerId);
		cache_demandForwardingRuleMap.remove(layerId);
		cache_demandRoutingCycleType.remove(layerId);

		routeIds.remove(layerId);
		routeDemandIds.remove(layerId);
		routeSequenceOfLinks.remove(layerId);
		routeCarriedTrafficMap.remove(layerId);
		routeOccupiedCapacityMap.remove(layerId);
		routeOriginalSequenceOfLinks.remove(layerId);
		routeOriginalCarriedTrafficMap.remove(layerId);
		routeOriginalOccupiedCapacityMap.remove(layerId);
		routeSameOccupiedCapacityAsCarriedTraffic.remove(layerId);
		routeBackupSegmentList.remove(layerId);
		routeAttributes.remove(layerId);

		segmentIds.remove(layerId);
		segmentSequenceOfLinks.remove(layerId);
		segmentReservedCapacity.remove(layerId);
		cache_segmentCarriedTrafficMap.remove(layerId);
		segmentRoutes.remove(layerId);
		segmentAttributes.remove(layerId);

		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		for(long nodeId : nodeIds_thisNetPlan)
		{
			cache_nodeIncomingLinks.get(nodeId).remove(layerId);
			cache_nodeOutgoingLinks.get(nodeId).remove(layerId);
			cache_nodeIncomingDemands.get(nodeId).remove(layerId);
			cache_nodeOutgoingDemands.get(nodeId).remove(layerId);
			cache_nodeForwardingRuleMap.get(nodeId).remove(layerId);
		}

		Set<Long> srgIds_thisNetPlan = getSRGIds();
		for(long srgId : srgIds_thisNetPlan)
			srgLinks.get(srgId).remove(layerId);
		
		if (layerDefaultId == layerId) layerDefaultId = layerIds.iterator().next();
	}
	
	/**
	 * <p>Removes a set of layers, and any associated link, demand, route, protection segment or forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_REMOVED LAYER_REMOVED} actions will be issued.</p>
	 *
	 * @param layerIds Layer identifiers
	 * @since 0.3.0
	 */
	public void removeLayers(Set<Long> layerIds)
	{
		checkIsModifiable();
		for(long layerId : layerIds) checkLayerIsActive(layerId);
		
		for(long layerId : new LinkedHashSet<Long>(layerIds)) removeLayer(layerId);
	}
	
	/**
	 * Removes a given attribute from the given layer. If the attribute is not defined, no action is made.
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeLayerAttribute(long layerId, String key)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		layerAttributeMap.get(layerId).remove(key);
	}
	
	/**
	 * <p>Removes a link, and any associated route, protection segment or forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} action will be issued.</p>
	 *
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void removeLink(long linkId)
	{
		long layerId = getLayerDefaultId();
		removeLink(layerId, linkId);
	}
	
	/**
	 * <p>Removes a link, and any associated route, protection segment or forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: if link is coupled, its coupled demand will be also removed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void removeLink(long layerId, long linkId)
	{
		checkIsModifiable();
		checkLinkIsActive(layerId, linkId);

		RoutingType routingType = getRoutingType(layerId);
		switch(routingType)
		{
			case SOURCE_ROUTING:
				Set<Long> routeIds_thisLink = getLinkAssociatedRoutes(layerId, linkId);
				removeRoutes(layerId, routeIds_thisLink);

				Set<Long> segmentIds_thisLink = getLinkAssociatedProtectionSegments(layerId, linkId);
				removeProtectionSegments(layerId, segmentIds_thisLink);

				break;
				
			case HOP_BY_HOP_ROUTING:
				Set<Pair<Long, Long>> forwardingRules_thisLink = getLinkForwardingRules(layerId, linkId);
				removeForwardingRules(layerId, forwardingRules_thisLink);
				break;
				
			default:
				throw new RuntimeException("Bad");
		}
		
		if (isLinkCoupled(layerId, linkId))
		{
			Pair<Long, Long> lowerLayerDemand = getLinkCoupledLowerLayerDemand(layerId, linkId);
			decoupleLink(layerId, linkId);
			
			long lowerLayerId = lowerLayerDemand.getFirst();
			long demandId = lowerLayerDemand.getSecond();
			removeDemand(lowerLayerId, demandId);
		}
		
		Pair<Long, Long> endNodes = linkMap.get(layerId).get(linkId);
		long originNodeId = endNodes.getFirst();
		long destinationNodeId = endNodes.getSecond();
		
		for(long srgId : new LinkedHashSet<Long>(getLinkSRGs(layerId, linkId))) removeLinkFromSRG(layerId, linkId, srgId);

		cache_nodeIncomingLinks.get(destinationNodeId).get(layerId).remove(linkId);
		cache_nodeOutgoingLinks.get(originNodeId).get(layerId).remove(linkId);
		Map<String, String> attributes = linkAttributesMap.get(layerId).remove(linkId);
		double u_e = linkCapacityMap.get(layerId).remove(linkId);
		cache_linkCarriedTrafficMap.get(layerId).remove(linkId);
		cache_linksOverSubscribed.get(layerId).remove(linkId);
		cache_linksWithUnlimitedCapacity.get(layerId).remove(linkId);
		cache_linksWithZeroCapacity.get(layerId).remove(linkId);
		linkReservedCapacityMap.get(layerId).remove(linkId);
		linkIds.get(layerId).remove(linkId);
		double lengthInKm = linkLengthInKmMap.get(layerId).remove(linkId);
		double propagationSpeedInKmPerSecond = linkPropagationSpeedInKmPerSecondMap.get(layerId).remove(linkId);
		linkMap.get(layerId).remove(linkId);
		linksDown.get(layerId).remove(linkId);
		linkSRGIds.get(layerId).remove(linkId);
		linkRouteIds.get(layerId).remove(linkId);
		linkSegmentIds.get(layerId).remove(linkId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkRemoved(layerId, linkId, originNodeId, destinationNodeId, u_e, lengthInKm, propagationSpeedInKmPerSecond, attributes));
	}
	
	/**
	 * <p>Removes a given attribute from a link. If the attribute is not defined, no action is made.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} action will be issued.</p>
	 *
	 * @param linkId Link identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeLinkAttribute(long linkId, String key)
	{
		long layerId = getLayerDefaultId();
		removeLinkAttribute(layerId, linkId, key);
	}

	/**
	 * <p>Removes a given attribute from a link. If the attribute is not defined, no action is made.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeLinkAttribute(long layerId, long linkId, String key)
	{
		checkIsModifiable();
		checkLinkIsActive(layerId, linkId);
		
		linkAttributesMap.get(layerId).get(linkId).remove(key);
		Pair<Long, Long> aux = upperLayerToLowerLayerMapping.get(layerId, linkId);
		if (aux == null) return;
		
		long lowerLayerId = aux.getFirst();
		long demandId = aux.getSecond();
		demandAttributesMap.get(lowerLayerId).get(demandId).remove(key);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(layerId, linkId));
	}
	
	/**
	 * Removes a link from the set of links of a given SRG at the first layer.
	 * 
	 * @param linkId Link identifier
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void removeLinkFromSRG(long linkId, long srgId)
	{
		long layerId = getLayerDefaultId();
		removeLinkFromSRG(layerId, linkId, srgId);
	}
	
	/**
	 * Removes a link from the set of links of a given SRG at the given layer.
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void removeLinkFromSRG(long layerId, long linkId, long srgId)
	{
		checkIsModifiable();
		
		checkSRGIsActive(srgId);
		checkLinkIsActive(layerId, linkId);
		
		srgLinks.get(srgId).get(layerId).remove(linkId);
		linkSRGIds.get(layerId).get(linkId).remove(srgId);
	}
	
	/**
	 * <p>Removes a set of links.</p>
	 *
	 * <p><b>Important</b>: All the associated routes, protection segments and lower layer demands will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} actions will be issued.</p>
	 *
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void removeLinks(Set<Long> linkIds)
	{
		long layerId = getLayerDefaultId();
		removeLinks(layerId, linkIds);
	}

	/**
	 * <p>Removes a set of links.</p>
	 *
	 * <p><b>Important</b>: All the associated routes, protection segments and lower layer demands will be removed as well.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_REMOVED LINK_REMOVED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void removeLinks(long layerId, Set<Long> linkIds)
	{
		checkIsModifiable();
		for(long linkId : linkIds) checkLinkIsActive(layerId, linkId);

		for(long linkId : new LinkedHashSet<Long>(linkIds)) removeLink(layerId, linkId);
	}
	
	/**
	 * Removes a given attribute from the network.
	 *
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeNetworkAttribute(String key)
	{
		checkIsModifiable();
		networkAttributes.remove(key);
	}

	/**
	 * <p>Removes a node, and any associated link, demand, route, protection segment or forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_REMOVED NODE_REMOVED} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void removeNode(long nodeId)
	{
		checkIsModifiable();
		checkNodeIsActive(nodeId);
		
		Set<Long> layerIds_thisNetPlan = getLayerIds();
		for(long layerId : layerIds_thisNetPlan)
		{
			/* Removing links */
			Set<Long> linksToRemove = getNodeAssociatedLinks(layerId, nodeId);
			removeLinks(layerId, linksToRemove);

			/* Removing demands */
			Set<Long> demandsToRemove = getNodeAssociatedDemands(layerId, nodeId);
			removeDemands(layerId, demandsToRemove);
		}
		
		/* Removing from SRGs */
		Set<Long> aux_srgs = getNodeSRGs(nodeId);
		for(long srgId : aux_srgs)
			removeNodeFromSRG(nodeId, srgId);
		
		Map<String, String> attributes = nodeAttributes.remove(nodeId);
		nodeIds.remove(nodeId);
		String nodeName = nodeNames.remove(nodeId);
		cache_nodeSRGIds.remove(nodeId);
		Point2D xyPosition = nodeXYPositionMap.remove(nodeId);
		nodesDown.remove(nodeId);
		cache_nodeIncomingLinks.remove(nodeId);
		cache_nodeOutgoingLinks.remove(nodeId);
		cache_nodeIncomingDemands.remove(nodeId);
		cache_nodeOutgoingDemands.remove(nodeId);
		cache_nodeForwardingRuleMap.remove(nodeId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeRemoved(nodeId, xyPosition, nodeName, attributes));
	}
	
	/**
	 * <p>Removes a given attribute from a node. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_MODIFIED NODE_MODIFIED} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeNodeAttribute(long nodeId, String key)
	{
		checkIsModifiable();
		
		checkNodeIsActive(nodeId);
		nodeAttributes.get(nodeId).remove(key);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(nodeId));
	}
	
	/**
	 * Removes a node from the set of nodes of a given SRG.
	 * 
	 * @param nodeId Node identifier
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void removeNodeFromSRG(long nodeId, long srgId)
	{
		checkIsModifiable();
		
		checkSRGIsActive(srgId);
		checkNodeIsActive(nodeId);
		
		srgNodes.get(srgId).remove(nodeId);
		cache_nodeSRGIds.get(nodeId).remove(srgId);
	}
	
	/**
	 * <p>Removes a set of nodes.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_REMOVED NODE_REMOVED} actions will be issued.</p>
	 *
	 * @param nodeIds Nodes identifiers
	 * @since 0.3.0
	 */
	public void removeNodes(Set<Long> nodeIds)
	{
		checkIsModifiable();
		for(long nodeId : nodeIds) checkNodeIsActive(nodeId);

		for(long nodeId : new LinkedHashSet<Long>(nodeIds)) removeNode(nodeId);
	}
	
	/**
	 * <p>Removes a protection segment.</p>
	 * 
	 * <p><b>Important</b>: The protection segment will be removed from the list of protection segments of any associated route, and removing any route using it in the current path.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action will be issued.</p>
	 *
	 * @param segmentId Protection segment identifier
	 * @since 0.3.0
	 */
	public void removeProtectionSegment(long segmentId)
	{
		long layerId = getLayerDefaultId();
		removeProtectionSegment(layerId, segmentId);
	}
	
	/**
	 * <p>Removes a protection segment.</p>
	 * 
	 * <p><b>Important</b>: The protection segment will be removed from the list of protection segments of any associated route, and removing any route using it in the current path.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @since 0.3.0
	 */
	public void removeProtectionSegment(long layerId, long segmentId)
	{
		checkIsModifiable();
		checkProtectionSegmentIsActive(layerId, segmentId);
		
		long auxSegmentId = -1 - segmentId;
		Set<Long> routeIds_thisSegment = getProtectionSegmentRoutes(layerId, segmentId);
		for(long routeId : routeIds_thisSegment)
		{
			List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
			if (seqLinks.contains(auxSegmentId)) throw new Net2PlanException("Bad - Protection segment is being used by route " + routeId);
		}
		
		if (!isProtectionSegmentDown(layerId, segmentId))
		{
			double u_s = getProtectionSegmentReservedCapacity(layerId, segmentId);
			List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
			for(long linkId : seqLinks)
			{
				double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
				linkReservedCapacityMap.get(layerId).put(linkId, r_e - u_s);
			}
		}
		
		for(long routeId : routeIds_thisSegment)
		{
			removeProtectionSegmentFromRouteBackupSegmentList(layerId, segmentId, routeId);
		}
		
		for(long linkId : getProtectionSegmentSequenceOfLinks(layerId, segmentId)) linkSegmentIds.get(layerId).get(linkId).remove(segmentId);
		Map<String, String> attributes = segmentAttributes.get(layerId).remove(segmentId);
		double reservedCapacity = segmentReservedCapacity.get(layerId).remove(segmentId);
		cache_segmentCarriedTrafficMap.get(layerId).remove(segmentId);
		segmentRoutes.get(layerId).remove(segmentId);
		segmentIds.get(layerId).remove(segmentId);
		List<Long> seqLinks = segmentSequenceOfLinks.get(layerId).remove(segmentId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.protectionSegmentRemoved(layerId, segmentId, reservedCapacity, seqLinks, attributes));
	}

	/**
	 * <p>Removes a given attribute from a protection segment. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 *
	 * @param segmentId Protection segment identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeProtectionSegmentAttribute(long segmentId, String key)
	{
		long layerId = getLayerDefaultId();
		removeProtectionSegmentAttribute(layerId, segmentId, key);
	}

	/**
	 * <p>Removes a given attribute from a protection segment. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeProtectionSegmentAttribute(long layerId, long segmentId, String key)
	{
		checkIsModifiable();
		checkProtectionSegmentIsActive(layerId, segmentId);
		segmentAttributes.get(layerId).get(segmentId).remove(key);
	}
	
	/**
	 * <p>Removes a protection segment from the list of backup protection segments of a route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param segmentId Segment identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void removeProtectionSegmentFromRouteBackupSegmentList(long segmentId, long routeId)
	{
		long layerId = getLayerDefaultId();
		removeProtectionSegmentFromRouteBackupSegmentList(layerId, segmentId, routeId);
	}
	
	/**
	 * <p>Removes a protection segment from the list of backup protection segments of a route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Segment identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void removeProtectionSegmentFromRouteBackupSegmentList(long layerId, long segmentId, long routeId)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		long auxId = -1 - segmentId;
		List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
		if (seqLinks.contains(auxId))
			throw new Net2PlanException("Protection segment " + segmentId + " is being used in route " + routeId);
		
		routeBackupSegmentList.get(layerId).get(routeId).remove(segmentId);
		segmentRoutes.get(layerId).get(segmentId).remove(routeId);
	}
	
	/**
	 * <p>Removes a set of protection segments.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} actions will be issued.</p>
	 * 
	 * @param segmentIds Protection segment identifiers
	 * @since 0.3.0
	 */
	public void removeProtectionSegments(Set<Long> segmentIds)
	{
		long layerId = getLayerDefaultId();
		removeProtectionSegments(layerId, segmentIds);
	}

	/**
	 * <p>Removes a set of protection segments.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_REMOVED PROTECTION_SEGMENT_REMOVED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentIds Protection segment identifiers
	 * @since 0.3.0
	 */
	public void removeProtectionSegments(long layerId, Set<Long> segmentIds)
	{
		checkIsModifiable();
		for(long segmentId : segmentIds) checkProtectionSegmentIsActive(layerId, segmentId);

		for(long segmentId : new LinkedHashSet<Long>(segmentIds)) removeProtectionSegment(layerId, segmentId);
	}
	
	/**
	 * <p>Removes a route.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED} action will be issued.</p>
	 *
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void removeRoute(long routeId)
	{
		long layerId = getLayerDefaultId();
		removeRoute(layerId, routeId);
	}
	
	/**
	 * <p>Removes a route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void removeRoute(long layerId, long routeId)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		
		double x_p = getRouteCarriedTraffic(layerId, routeId);
		double y_p = getRouteOccupiedCapacity(layerId, routeId);

		try
		{
			addingItem = true;
			setRouteDown(layerId, routeId);
		}
		catch(Throwable e)
		{
			addingItem = false;
		}
		
		List<Long> segmentIds_thisRoute = getRouteBackupSegmentList(layerId, routeId);
		for(long segmentId : segmentIds_thisRoute)
			removeProtectionSegmentFromRouteBackupSegmentList(layerId, segmentId, routeId);
		
		for(long linkId : getRouteSequenceOfLinks(layerId, routeId))
			if (linkId >= 0) linkRouteIds.get(layerId).get(linkId).remove(routeId);
		
		long demandId = routeDemandIds.get(layerId).remove(routeId);
		routesDown.get(layerId).remove(routeId);
		Map<String, String> attributes = routeAttributes.get(layerId).remove(routeId);
		demandRouteIdsMap.get(layerId).get(demandId).remove(routeId);
		List<Long> backupSegmentList = routeBackupSegmentList.get(layerId).remove(routeId);
		routeCarriedTrafficMap.get(layerId).remove(routeId);
		routeIds.get(layerId).remove(routeId);
		routeOccupiedCapacityMap.get(layerId).remove(routeId);
		routeOriginalSequenceOfLinks.get(layerId).remove(routeId);
		routeOriginalCarriedTrafficMap.get(layerId).remove(routeId);
		routeOriginalOccupiedCapacityMap.get(layerId).remove(routeId);
		routeSameOccupiedCapacityAsCarriedTraffic.get(layerId).remove(routeId);
		List<Long> seqLinks = routeSequenceOfLinks.get(layerId).remove(routeId);
		
		if (isActionMonitoringActivated())
		{
			Pair<Long, Long> ingressEgressNodePair = getDemandIngressEgressNodePair(layerId, demandId);
			long ingressNodeId = ingressEgressNodePair.getFirst();
			long egressNodeId = ingressEgressNodePair.getSecond();
			actionsPerformed.add(Factory.routeRemoved(layerId, routeId, demandId, ingressNodeId, egressNodeId, x_p, y_p, seqLinks, backupSegmentList, attributes));
		}
	}
	
	/**
	 * <p>Removes a given attribute from a route. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param routeId Route identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeRouteAttribute(long routeId, String key)
	{
		long layerId = getLayerDefaultId();
		removeRouteAttribute(layerId, routeId, key);
	}

	/**
	 * <p>Removes a given attribute from a route. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeRouteAttribute(long layerId, long routeId, String key)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		routeAttributes.get(layerId).get(routeId).remove(key);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
	}
	
	/**
	 * <p>Removes a set of routes.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED} actions will be issued.</p>
	 * 
	 * @param routeIds Route identifiers
	 * @since 0.3.0
	 */
	public void removeRoutes(Set<Long> routeIds)
	{
		long layerId = getLayerDefaultId();
		removeRoutes(layerId, routeIds);
	}

	/**
	 * <p>Removes a set of routes.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_REMOVED ROUTE_REMOVED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeIds Route identifiers
	 * @since 0.3.0
	 */
	public void removeRoutes(long layerId, Set<Long> routeIds)
	{
		checkIsModifiable();
		for(long routeId : routeIds) checkRouteIsActive(layerId, routeId);
		
		for(long routeId : new LinkedHashSet<Long>(routeIds)) removeRoute(layerId, routeId);
	}
	
	/**
	 * <p>Removes an SRG.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_REMOVED SRG_REMOVED} action will be issued.</p>
	 *
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void removeSRG(long srgId)
	{
		checkIsModifiable();
		checkSRGIsActive(srgId);
		
		for(long nodeId : getSRGNodes(srgId)) cache_nodeSRGIds.get(nodeId).remove(srgId);
		
		for(long layerId : getLayerIds())
			for(long linkId : getSRGLinks(layerId, srgId))
				linkSRGIds.get(layerId).get(linkId).remove(srgId);

		srgAttributes.remove(srgId);
		srgIds.remove(srgId);
		srgLinks.remove(srgId);
		srgMeanTimeToFailInHours.remove(srgId);
		srgMeanTimeToRepairInHours.remove(srgId);
		srgNodes.remove(srgId);
	}
	
	/**
	 * <p>Removes a given attribute from an SRG. If the attribute is not defined, no action is made.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param srgId SRG identifier
	 * @param key Attribute name
	 * @since 0.3.0
	 */
	public void removeSRGAttribute(long srgId, String key)
	{
		checkIsModifiable();
		checkSRGIsActive(srgId);
		srgAttributes.get(srgId).remove(key);
	}
	
	/**
	 * <p>Removes a set of SRGs.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_REMOVED SRG_REMOVED} actions will be issued.</p>
	 *
	 * @param srgIds SRG identifiers
	 * @since 0.3.0
	 */
	public void removeSRGs(Set<Long> srgIds)
	{
		checkIsModifiable();
		for(long srgId : srgIds) checkSRGIsActive(srgId);
		
		for (long srgId : new LinkedHashSet<Long>(srgIds)) removeSRG(srgId);
	}
	
	/**
	 * Resets the state of the network to an empty {@code NetPlan}.
	 * 
	 * @since 0.2.3
	 */
	public void reset()
	{
		assignFrom(new NetPlan());
	}
	
	/**
	 * Saves the current network plan to a given file. If extension {@code .n2p} 
	 * is not in the file name, it will be added automatically.
	 *
	 * @param file .n2p file
	 * @since 0.2.0
	 */
	public void saveToFile(File file)
	{
		String filePath = file.getPath();
		if (!filePath.toLowerCase(Locale.getDefault()).endsWith(".n2p")) file = new File(filePath + ".n2p");
		
		FileOutputStream fos = null;
		try
		{
			fos = new FileOutputStream(file);
			saveToOutputStream(fos);
		}
		catch (FileNotFoundException e)
		{
			if (fos != null)
			{
				try { fos.close(); }
				catch (IOException ex) { }
			}
			
			throw new Net2PlanException(e.getMessage());
		}
	}
	
	/**
	 * Saves the current network plan to a given output stream.
	 *
	 * @param outputStream Output stream
	 * @since 0.3.1
	 */
	public void saveToOutputStream(OutputStream outputStream)
	{
		try
		{
			XMLOutputFactory2 output = (XMLOutputFactory2) XMLOutputFactory2.newFactory();
			XMLStreamWriter2 writer = (XMLStreamWriter2) output.createXMLStreamWriter(outputStream);
			
			writer.writeStartDocument("UTF-8", "1.0");
			
			XMLUtils.indent(writer, 0);
			writer.writeStartElement("network");
			writer.writeAttribute("description", getNetworkDescription());
			writer.writeAttribute("name", getNetworkName());
			writer.writeAttribute("version", Version.getFileFormatVersion());
			
			Set<Long> nodeIds_thisNetPlan = getNodeIds();
			for (long nodeId : nodeIds_thisNetPlan)
			{
				boolean emptyNode = getNodeAttributeMap(nodeId).isEmpty();
				
				XMLUtils.indent(writer, 1);
				if (emptyNode) writer.writeEmptyElement("node");
				else writer.writeStartElement("node");
				
				Point2D position = getNodeXYPosition(nodeId);
				writer.writeAttribute("id", Long.toString(nodeId));
				writer.writeAttribute("xCoord", Double.toString(position.getX()));
				writer.writeAttribute("yCoord", Double.toString(position.getY()));
				writer.writeAttribute("name", getNodeName(nodeId));

				for (Entry<String, String> entry : getNodeAttributeMap(nodeId).entrySet())
				{
					XMLUtils.indent(writer, 2);
					writer.writeEmptyElement("attribute");
					writer.writeAttribute("key", entry.getKey());
					writer.writeAttribute("value", entry.getValue());
				}

				if (!emptyNode)
				{
					XMLUtils.indent(writer, 1);
					writer.writeEndElement();
				}
			}
			
			Set<Long> layerIds_thisNetPlan = getLayerIds();
			for(long layerId : layerIds_thisNetPlan)
			{
				XMLUtils.indent(writer, 1);
				writer.writeStartElement("layer");

				writer.writeAttribute("id", Long.toString(layerId));
				writer.writeAttribute("name", getLayerName(layerId));
				writer.writeAttribute("description", getLayerDescription(layerId));
				writer.writeAttribute("linkCapacityUnitsName", getLinkCapacityUnitsName(layerId));
				writer.writeAttribute("demandTrafficUnitsName", getDemandTrafficUnitsName(layerId));

				Set<Long> linkIds_thisLayer = getLinkIds(layerId);
				for(long linkId : linkIds_thisLayer)
				{
					boolean emptyLink = getLinkAttributeMap(layerId, linkId).isEmpty();
					
					XMLUtils.indent(writer, 2);
					if (emptyLink) writer.writeEmptyElement("link");
					else writer.writeStartElement("link");
					
					writer.writeAttribute("id", Long.toString(linkId));
					writer.writeAttribute("originNodeId", Long.toString(getLinkOriginNode(layerId, linkId)));
					writer.writeAttribute("destinationNodeId", Long.toString(getLinkDestinationNode(layerId, linkId)));
					writer.writeAttribute("capacity", Double.toString(getLinkCapacity(layerId, linkId)));
					writer.writeAttribute("lengthInKm", Double.toString(getLinkLengthInKm(layerId, linkId)));
					writer.writeAttribute("propagationSpeedInKmPerSecond", Double.toString(getLinkPropagationSpeedInKmPerSecond(layerId, linkId)));

					for (Entry<String, String> entry : getLinkAttributeMap(layerId, linkId).entrySet())
					{
						XMLUtils.indent(writer, 3);
						writer.writeEmptyElement("attribute");
						writer.writeAttribute("key", entry.getKey());
						writer.writeAttribute("value", entry.getValue());
					}

					if (!emptyLink)
					{
						XMLUtils.indent(writer, 2);
						writer.writeEndElement();
					}
				}

				Set<Long> demandIds_thisLayer = getDemandIds(layerId);
				for(long demandId : demandIds_thisLayer)
				{
					boolean emptyDemand = getDemandAttributeMap(layerId, demandId).isEmpty();
					
					XMLUtils.indent(writer, 2);
					if (emptyDemand) writer.writeEmptyElement("demand");
					else writer.writeStartElement("demand");
					
					writer.writeAttribute("id", Long.toString(demandId));
					writer.writeAttribute("ingressNodeId", Long.toString(getDemandIngressNode(layerId, demandId)));
					writer.writeAttribute("egressNodeId", Long.toString(getDemandEgressNode(layerId, demandId)));
					writer.writeAttribute("offeredTraffic", Double.toString(getDemandOfferedTraffic(layerId, demandId)));

					for (Entry<String, String> entry : getDemandAttributeMap(layerId, demandId).entrySet())
					{
						XMLUtils.indent(writer, 3);
						writer.writeEmptyElement("attribute");
						writer.writeAttribute("key", entry.getKey());
						writer.writeAttribute("value", entry.getValue());
					}

					if (!emptyDemand)
					{
						XMLUtils.indent(writer, 2);
						writer.writeEndElement();
					}
				}

				if (getRoutingType(layerId) == RoutingType.SOURCE_ROUTING)
				{
					boolean emptySourceRouting = !hasProtectionSegments(layerId) && !hasRoutes(layerId);
					XMLUtils.indent(writer, 2);
					if (emptySourceRouting) writer.writeEmptyElement("sourceRouting");
					else writer.writeStartElement("sourceRouting");
					
					Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
					for(long segmentId : segmentIds_thisLayer)
					{
						boolean emptySegment = getProtectionSegmentAttributeMap(layerId, segmentId).isEmpty();

						XMLUtils.indent(writer, 3);
						if (emptySegment) writer.writeEmptyElement("protectionSegment");
						else writer.writeStartElement("protectionSegment");

						writer.writeAttribute("id", Long.toString(segmentId));
						writer.writeAttribute("reservedCapacity", Double.toString(getProtectionSegmentReservedCapacity(layerId, segmentId)));

						List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
						writer.writeAttribute("seqLinks", CollectionUtils.join(seqLinks, " "));

						for (Entry<String, String> entry : getProtectionSegmentAttributeMap(layerId, segmentId).entrySet())
						{
							XMLUtils.indent(writer, 4);
							writer.writeEmptyElement("attribute");
							writer.writeAttribute("key", entry.getKey());
							writer.writeAttribute("value", entry.getValue());
						}

						if (!emptySegment)
						{
							XMLUtils.indent(writer, 3);
							writer.writeEndElement();
						}
					}

					Set<Long> routeIds_thisLayer = getRouteIds(layerId);
					for(long routeId : routeIds_thisLayer)
					{
						boolean emptyRoute = getRouteAttributeMap(layerId, routeId).isEmpty();

						XMLUtils.indent(writer, 3);
						if (emptyRoute) writer.writeEmptyElement("route");
						else writer.writeStartElement("route");

						writer.writeAttribute("id", Long.toString(routeId));
						writer.writeAttribute("demandId", Long.toString(getRouteDemand(layerId, routeId)));
						writer.writeAttribute("carriedTraffic", Double.toString(getRouteCarriedTraffic(layerId, routeId)));
						writer.writeAttribute("occupiedCapacity", Double.toString(routeSameOccupiedCapacityAsCarriedTraffic.get(layerId).contains(routeId) ? -1 : getRouteOccupiedCapacity(layerId, routeId)));

						List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
						List<Long> backupSegmentList = getRouteBackupSegmentList(layerId, routeId);
						writer.writeAttribute("seqLinks", CollectionUtils.join(seqLinks, " "));
						writer.writeAttribute("backupSegmentList", CollectionUtils.join(backupSegmentList, " "));

						for (Entry<String, String> entry : getRouteAttributeMap(layerId, routeId).entrySet())
						{
							XMLUtils.indent(writer, 4);
							writer.writeEmptyElement("attribute");
							writer.writeAttribute("key", entry.getKey());
							writer.writeAttribute("value", entry.getValue());
						}

						if (!emptyRoute)
						{
							XMLUtils.indent(writer, 3);
							writer.writeEndElement();
						}
					}

					if (!emptySourceRouting)
					{
						XMLUtils.indent(writer, 2);
						writer.writeEndElement();
					}
				}
				else
				{
					boolean emptyHopByHopRouting = !hasForwardingRules(layerId);
					XMLUtils.indent(writer, 2);
					if (emptyHopByHopRouting) writer.writeEmptyElement("hopByHopRouting");
					else writer.writeStartElement("hopByHopRouting");
					
					Set<Pair<Long, Long>> forwardingRules_thisLayer = getForwardingRules(layerId);
					for(Pair<Long, Long> forwardingRule : forwardingRules_thisLayer)
					{
						boolean emptyForwardingRule = getForwardingRuleAttributeMap(layerId, forwardingRule).isEmpty();

						XMLUtils.indent(writer, 3);
						if (emptyForwardingRule) writer.writeEmptyElement("forwardingRule");
						else writer.writeStartElement("forwardingRule");

						writer.writeAttribute("demandId", Long.toString(forwardingRule.getFirst()));
						writer.writeAttribute("linkId", Long.toString(forwardingRule.getSecond()));
						writer.writeAttribute("splittingRatio", Double.toString(getForwardingRuleSplittingRatio(layerId, forwardingRule)));

						for (Entry<String, String> entry : getForwardingRuleAttributeMap(layerId, forwardingRule).entrySet())
						{
							XMLUtils.indent(writer, 4);
							writer.writeEmptyElement("attribute");
							writer.writeAttribute("key", entry.getKey());
							writer.writeAttribute("value", entry.getValue());
						}

						if (!emptyForwardingRule)
						{
							XMLUtils.indent(writer, 3);
							writer.writeEndElement();
						}
					}

					if (!emptyHopByHopRouting)
					{
						XMLUtils.indent(writer, 2);
						writer.writeEndElement();
					}
				}

				for (Entry<String, String> entry : getLayerAttributeMap(layerId).entrySet())
				{
					XMLUtils.indent(writer, 2);
					writer.writeEmptyElement("attribute");
					writer.writeAttribute("key", entry.getKey());
					writer.writeAttribute("value", entry.getValue());
				}

				XMLUtils.indent(writer, 1);
				writer.writeEndElement();
			}
			
			Set<Long> srgIds_thisNetPlan = getSRGIds();
			for (long srgId : srgIds_thisNetPlan)
			{
				Set<Long> nodeIds_thisSRG = getSRGNodes(srgId);
				Map<Long, Set<Long>> linkIds_thisSRG = getSRGLinksAllLayers(srgId);

				boolean emptySRG = getSRGAttributeMap(srgId).isEmpty() && getSRGNodes(srgId).isEmpty();
				
				if (emptySRG)
				{
					for (Entry<Long, Set<Long>> entry : linkIds_thisSRG.entrySet())
					{
						if (!entry.getValue().isEmpty())
						{
							emptySRG = false;
							break;
						}
					}
				}				
				
				XMLUtils.indent(writer, 1);
				if (emptySRG) writer.writeEmptyElement("srg");
				else writer.writeStartElement("srg");
				
				writer.writeAttribute("id", Long.toString(srgId));
				writer.writeAttribute("meanTimeToFailInHours", Double.toString(getSRGMeanTimeToFailInHours(srgId)));
				writer.writeAttribute("meanTimeToRepairInHours", Double.toString(getSRGMeanTimeToRepairInHours(srgId)));

				for (long nodeId : nodeIds_thisSRG)
				{
					XMLUtils.indent(writer, 2);
					writer.writeEmptyElement("node");
					writer.writeAttribute("id", Long.toString(nodeId));
				}

				for (Entry<Long, Set<Long>> entry : linkIds_thisSRG.entrySet())
				{
					long layerId = entry.getKey();
					for(long linkId : entry.getValue())
					{
						XMLUtils.indent(writer, 2);
						writer.writeEmptyElement("link");
						writer.writeAttribute("layerId", Long.toString(layerId));
						writer.writeAttribute("linkId", Long.toString(linkId));
					}
				}

				for (Entry<String, String> entry : getSRGAttributeMap(srgId).entrySet())
				{
					XMLUtils.indent(writer, 2);
					writer.writeEmptyElement("attribute");
					writer.writeAttribute("key", entry.getKey());
					writer.writeAttribute("value", entry.getValue());
				}

				if (!emptySRG)
				{
					XMLUtils.indent(writer, 1);
					writer.writeEndElement();
				}
			}

			for(Cell<Long, Long, Pair<Long, Long>> entry : upperLayerToLowerLayerMapping.cellSet())
			{
				XMLUtils.indent(writer, 1);
				writer.writeEmptyElement("layerCoupling");
				
				Pair<Long, Long> lowerLayerLink = entry.getValue();
				writer.writeAttribute("upperLayerLinkId", entry.getColumnKey().toString());
				writer.writeAttribute("upperLayerId", entry.getRowKey().toString());
				writer.writeAttribute("lowerLayerDemandId", lowerLayerLink.getSecond().toString());
				writer.writeAttribute("lowerLayerId", lowerLayerLink.getFirst().toString());
			}
			
			for (Entry<String, String> entry : getNetworkAttributeMap().entrySet())
			{
				XMLUtils.indent(writer, 1);
				writer.writeEmptyElement("attribute");
				writer.writeAttribute("key", entry.getKey());
				writer.writeAttribute("value", entry.getValue());
			}

			XMLUtils.indent(writer, 0);
			writer.writeEndElement();
			writer.writeEndDocument();
			writer.flush();
			writer.close();
		}
		catch (XMLStreamException e)
		{
			throw new RuntimeException(e);
		}
	}

	/**
	 * <p>Sets all links currently down as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>
	 *
	 * @since 0.3.0
	 */
	public void setAllLinksUp()
	{
		long layerId = getLayerDefaultId();
		setAllLinksUp(layerId);
	}

	/**
	 * <p>Sets all links currently down as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void setAllLinksUp(long layerId)
	{
		checkIsModifiable();
		
		Set<Long> linkDown_thisLayer = getLinksDown(layerId);
		setLinksUp(layerId, linkDown_thisLayer);
	}

	/**
	 * <p>Sets all nodes currently down as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_UP NODE_UP} actions will be issued.</p>
	 *
	 * @since 0.3.0
	 */
	public void setAllNodesUp()
	{
		Set<Long> nodesDown_thisNetPlan = getNodesDown();
		setNodesUp(nodesDown_thisNetPlan);
	}

	/**
	 * <p>Sets any protection segment currently down, whose resources are up, as up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @since 0.3.0
	 */
	public void setAllProtectionSegmentsUp()
	{
		long layerId = getLayerDefaultId();
		setAllProtectionSegmentsUp(layerId);
	}
	
	/**
	 * <p>Sets any protection segment currently down, whose resources are up, as up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void setAllProtectionSegmentsUp(long layerId)
	{
		Set<Long> segmentIds_thisLayer = new LinkedHashSet<Long>(getProtectionSegmentsDown(layerId));
		Iterator<Long> it = segmentIds_thisLayer.iterator();
		while(it.hasNext())
		{
			long segmentId = it.next();
			if (!isProtectionSegmentRestorable(layerId, segmentId))
				it.remove();
		}
		
		setProtectionSegmentsUp(layerId, segmentIds_thisLayer);
	}
	
	/**
	 * <p>Sets any route currently down, whose resources are up, as up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @since 0.3.0
	 */
	public void setAllRoutesUp()
	{
		long layerId = getLayerDefaultId();
		setAllRoutesUp(layerId);
	}
	
	/**
	 * <p>Sets any route currently down, whose resources are up, as up.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void setAllRoutesUp(long layerId)
	{
		Set<Long> routeIds_thisLayer = new LinkedHashSet<Long>(getRoutesDown(layerId));
		Iterator<Long> it = routeIds_thisLayer.iterator();
		while(it.hasNext())
		{
			long routeId = it.next();
			if (!isRouteRestorable(layerId, routeId))
				it.remove();
		}
		
		setRoutesUp(layerId, routeIds_thisLayer);
	}
	
	/**
	 * <p>Sets an attribute for a given demand. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 * 
	 * @param demandId Demand identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setDemandAttribute(long demandId, String key, String value)
	{
		long layerId = getLayerDefaultId();
		setDemandAttribute(layerId, demandId, key, value);
	}

	/**
	 * <p>Sets an attribute for a given demand. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setDemandAttribute(long layerId, long demandId, String key, String value)
	{
		checkIsModifiable();
		checkDemandIsActive(layerId, demandId);
		
		demandAttributesMap.get(layerId).get(demandId).put(key, value);
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.demandModified(layerId, demandId));
		Pair<Long, Long> aux = lowerLayerToUpperLayerMapping.get(layerId, demandId);
		if (aux == null) return;
		
		long upperLayerId = aux.getFirst();
		long linkId = aux.getSecond();
		linkAttributesMap.get(upperLayerId).get(linkId).put(key, value);
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(upperLayerId, linkId));
	}
	
	/**
	 * <p>Sets the attributes for a given demand. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 * 
	 * @param demandId Demand identifier
	 * @param attributes Map of attributes
	 * @since 0.3.0
	 */
	public void setDemandAttributeMap(long demandId, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		setDemandAttributeMap(layerId, demandId, attributes);
	}

	/**
	 * <p>Sets the attributes for a given demand. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setDemandAttributeMap(long layerId, long demandId, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkDemandIsActive(layerId, demandId);
		
		Map<String, String> auxDemandAttributes = demandAttributesMap.get(layerId).get(demandId);
		auxDemandAttributes.clear();
		auxDemandAttributes.putAll(attributes);
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.demandModified(layerId, demandId));
		Pair<Long, Long> aux = lowerLayerToUpperLayerMapping.get(layerId, demandId);
		if (aux == null) return;
		
		long upperLayerId = aux.getFirst();
		long linkId = aux.getSecond();
		Map<String, String> auxLinkAttributes = linkAttributesMap.get(upperLayerId).get(linkId);
		auxLinkAttributes.clear();
		auxLinkAttributes.putAll(attributes);
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(upperLayerId, linkId));
	}
	
	/**
	 * <p>Sets the offered traffic by a demand.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 * 
	 * @param demandId Demand identifier
	 * @param h_d Offered traffic
	 * @since 0.3.0
	 */
	public void setDemandOfferedTraffic(long demandId, double h_d)
	{
		long layerId = getLayerDefaultId();
		setDemandOfferedTraffic(layerId, demandId, h_d);
	}

	/**
	 * <p>Sets the offered traffic by a demand.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandId Demand identifier
	 * @param h_d Offered traffic
	 * @since 0.3.0
	 */
	public void setDemandOfferedTraffic(long layerId, long demandId, double h_d)
	{
		checkIsModifiable();
		checkDemandIsActive(layerId, demandId);
		if (h_d < 0) throw new Net2PlanException("Offered traffic must be greater or equal than zero");
		
		demandOfferedTrafficMap.get(layerId).put(demandId, h_d);
		RoutingType routingType = getRoutingType(layerId);
		switch(routingType)
		{
			case HOP_BY_HOP_ROUTING:
				updateCacheDemandForwarding(layerId, demandId);
				break;
				
			case SOURCE_ROUTING:
				updateCacheDemandBlockedTraffic(layerId, demandId);
				break;
				
			default:
				throw new RuntimeException("Bad");
		}
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.demandModified(layerId, demandId));
	}
	
	/**
	 * <p>Sets the offered traffic per demand.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} actions will be issued.</p>
	 * 
	 * @param h_d Offered traffic per demand
	 * @since 0.3.0
	 */
	public void setDemandOfferedTrafficMap(Map<Long, Double> h_d)
	{
		long layerId = getLayerDefaultId();
		setDemandOfferedTrafficMap(layerId, h_d);
	}
	
	/**
	 * <p>Sets the offered traffic per demand.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param h_d Offered traffic per demand
	 * @since 0.3.0
	 */
	public void setDemandOfferedTrafficMap(long layerId, Map<Long, Double> h_d)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		
		for(Entry<Long, Double> entry : h_d.entrySet())
		{
			long demandId = entry.getKey();
			double offeredTraffic = entry.getValue();
			setDemandOfferedTraffic(layerId, demandId, offeredTraffic);
		}
	}
	
	/**
	 * <p>Sets the offered traffic per demand, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} actions will be issued.</p>
	 * 
	 * @param h_d Offered traffic per demand
	 * @since 0.2.0
	 */
	public void setDemandOfferedTrafficVector(double[] h_d)
	{
		long layerId = getLayerDefaultId();
		setDemandOfferedTrafficVector(layerId, h_d);
	}

	/**
	 * <p>Sets the offered traffic per demand, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_MODIFIED DEMAND_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param h_d Offered traffic per demand
	 * @since 0.3.0
	 */
	public void setDemandOfferedTrafficVector(long layerId, double[] h_d)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		
		Set<Long> demandIds_thisLayer = getDemandIds(layerId);
		Map<Long, Integer> demandId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(demandIds_thisLayer);
		for(long demandId : demandIds_thisLayer)
		{
			int d = demandId2LinearIndex.get(demandId);
			setDemandOfferedTraffic(layerId, demandId, h_d[d]);
		}
	}

	/**
	 * <p>Sets the demand traffic units name.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param trafficUnitsName Demand traffic units name
	 * @since 0.3.0
	 */
	public void setDemandTrafficUnitsName(String trafficUnitsName)
	{
		long layerId = getLayerDefaultId();
		setDemandTrafficUnitsName(layerId, trafficUnitsName);
	}
	
	/**
	 * <p>Sets the demand traffic units name.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param trafficUnitsName Demand traffic units name
	 * @since 0.3.0
	 */
	public void setDemandTrafficUnitsName(long layerId, String trafficUnitsName)
	{
		checkIsModifiable();
		
		checkLayerIsActive(layerId);
		if (getDemandTrafficUnitsName(layerId).equals(trafficUnitsName)) return;
		
		Set<Long> demandIds_thisLayer = getDemandIds(layerId);
		for(long demandId : demandIds_thisLayer)
			if (isDemandCoupled(layerId, demandId))
				throw new Net2PlanException("Demand traffic units name cannot be modified since there is some coupling with other layers");
		
		layerDemandTrafficUnitsNameMap.put(layerId, trafficUnitsName == null ? "" : trafficUnitsName);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.layerModified(layerId));
	}
	
	/**
	 * <p>Sets an attribute for a given forwarding rule. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setForwardingRuleAttribute(Pair<Long, Long> demandLinkPair, String key, String value)
	{
		long layerId = getLayerDefaultId();
		setForwardingRuleAttribute(layerId, demandLinkPair, key, value);
	}
	
	/**
	 * <p>Sets an attribute for a given forwarding rule. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setForwardingRuleAttribute(long layerId, Pair<Long, Long> demandLinkPair, String key, String value)
	{
		checkIsModifiable();
		checkForwardingRuleIsActive(layerId, demandLinkPair);
		forwardingRuleAttributeMap.get(layerId).get(demandLinkPair).put(key, value);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.forwardingRuleModified(layerId, demandLinkPair));
	}
	
	/**
	 * <p>Sets the attributes for a given forwarding rule. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setForwardingRuleAttributeMap(Pair<Long, Long> demandLinkPair, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		setForwardingRuleAttributeMap(layerId, demandLinkPair, attributes);
	}

	/**
	 * <p>Sets the attributes for a given forwarding rule. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setForwardingRuleAttributeMap(long layerId, Pair<Long, Long> demandLinkPair, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkForwardingRuleIsActive(layerId, demandLinkPair);

		Map<String, String> auxForwardingRuleAttributes = forwardingRuleAttributeMap.get(layerId).get(demandLinkPair);
		auxForwardingRuleAttributes.clear();
		auxForwardingRuleAttributes.putAll(attributes);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.forwardingRuleModified(layerId, demandLinkPair));
	}
	
	/**
	 * <p>Sets the splitting ratio (fraction of traffic from demand 'd' entering 
	 * to the origin node of link 'e', going through link 'e') of a forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param splittingRatio Splitting ratio (fraction of traffic from demand 'd' entering to the origin node of link 'e', going through link 'e')
	 * @since 0.3.0
	 */
	public void setForwardingRuleSplittingRatio(Pair<Long, Long> demandLinkPair, double splittingRatio)
	{
		long layerId = getLayerDefaultId();
		setForwardingRuleSplittingRatio(layerId, demandLinkPair, splittingRatio);
	}
	
	/**
	 * <p>Sets the splitting ratio (fraction of traffic from demand 'd' entering 
	 * to the origin node of link 'e', going through link 'e') of a forwarding rule.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#HOP_BY_HOP_ROUTING HOP_BY_HOP_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_MODIFIED FORWARDING_RULE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param demandLinkPair Demand-outgoing link pair
	 * @param splittingRatio Splitting ratio (fraction of traffic from demand 'd' entering to the origin node of link 'e', going through link 'e')
	 * @since 0.3.0
	 */
	public void setForwardingRuleSplittingRatio(long layerId, Pair<Long, Long> demandLinkPair, double splittingRatio)
	{
		checkIsModifiable();
		checkForwardingRuleIsActive(layerId, demandLinkPair);
		
		if (splittingRatio < 0) throw new Net2PlanException("Splitting ratio must be greater or equal than zero");

		long demandId = demandLinkPair.getFirst();
		long linkId = demandLinkPair.getSecond();
		long originNodeId = getLinkOriginNode(layerId, linkId);
		Set<Pair<Long, Long>> forwardingRules_thisNodeAndDemand = getNodeDemandForwardingRules(layerId, originNodeId, demandId);
		double totalSplittingRatio = 0;
		for(Pair<Long, Long> forwardingRule : forwardingRules_thisNodeAndDemand)
			if (!forwardingRule.equals(demandLinkPair))
				totalSplittingRatio += getForwardingRuleSplittingRatio(layerId, forwardingRule);

		if (totalSplittingRatio + splittingRatio > 1)
		{
			double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
			if (totalSplittingRatio + splittingRatio > 1 + PRECISION_FACTOR) throw new Net2PlanException("Splitting ratio for demand-link pair (d" + demandId + ", e" + linkId + ") exceeds the maximum value");
			else splittingRatio = 1 - totalSplittingRatio;
		}
		
		forwardingRuleSplittingRatioMap.get(layerId).put(demandLinkPair, splittingRatio);
		if (!disableCacheUpdating) updateCacheDemandForwarding(layerId, demandId);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.forwardingRuleModified(layerId, demandLinkPair));
	}

	/**
	 * <p>Sets the forwarding rules for the given design. Any previous routing 
	 * information (either source routing or hop-by-hop routing) will be removed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} actions will be issued.</p>
	 * 
	 * @param splittingRatioMap Splitting ratio mapping, where key is the demand-outgoing link pair and the value is the fraction of traffic sent through the given outgoing link, with respect to the total traffic entering to the node for the given demand
	 * @since 0.3.0
	 */
	public void setForwardingRules(Map<Pair<Long, Long>, Double> splittingRatioMap)
	{
		long layerId = getLayerDefaultId();
		setForwardingRules(layerId, splittingRatioMap);
	}
	
	/**
	 * <p>Sets the forwarding rules for the given design. Any previous routing 
	 * information (either source routing or hop-by-hop routing) will be removed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param splittingRatioMap Splitting ratio mapping, where key is the demand-outgoing link pair and the value is the fraction of traffic sent through the given outgoing link, with respect to the total traffic entering to the node for the given demand
	 * @since 0.3.0
	 */
	public void setForwardingRules(long layerId, Map<Pair<Long, Long>, Double> splittingRatioMap)
	{
		removeAllRoutingInformation(layerId);
		setRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);
		addForwardingRules(layerId, splittingRatioMap, false);
	}
	
	/**
	 * <p>Sets the forwarding rules for the given design. Any previous routing 
	 * information (either source routing or hop-by-hop routing) will be removed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} actions will be issued.</p>
	 * 
	 * @param f_de For each demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandIds}, <i>d = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code NetPlan} object, <i>e = 1</i> refers to the second one, and so on), {@code f_de[d][e]} sets the fraction of the traffic from demand <i>d</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>b(d)</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1
	 * @since 0.3.1
	 */
	public void setForwardingRules(DoubleMatrix2D f_de)
	{
		long layerId = getLayerDefaultId();
		setForwardingRules(layerId, f_de);
	}
	
	/**
	 * <p>Sets the forwarding rules for the given design. Any previous routing 
	 * information (either source routing or hop-by-hop routing) will be removed.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_REMOVED FORWARDING_RULE_REMOVED} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#FORWARDING_RULE_ADDED FORWARDING_RULE_ADDED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param f_de For each demand <i>d</i> (<i>d = 0</i> refers to the first demand in {@code demandIds}, <i>d = 1</i> refers to the second one, and so on), and each link <i>e</i> (<i>e = 0</i> refers to the first link in {@code NetPlan} object, <i>e = 1</i> refers to the second one, and so on), {@code f_de[d][e]} sets the fraction of the traffic from demand <i>d</i> that arrives (or is generated in) node <i>a(e)</i> (the origin node of link <i>e</i>), that is forwarded through link <i>e</i>. It must hold that for every node <i>n</i> different of <i>b(d)</i>, the sum of the fractions <i>f<sub>te</sub></i> along its outgoing links must be lower or equal than 1
	 * @since 0.3.0
	 */
	public void setForwardingRules(long layerId, DoubleMatrix2D f_de)
	{
		int D = getNumberOfDemands(layerId);
		int E = getNumberOfLinks(layerId);
		if (f_de.rows() != D || f_de.columns() != E) throw new Net2PlanException("'f_de' should be a " + D + " x" + E + " matrix (demands x links)");
		
		Map<Pair<Long, Long>, Double> splittingRatioMap = DoubleUtils.convertMatrix2Map(getDemandIds(layerId), getLinkIds(layerId), f_de);
		setForwardingRules(layerId, splittingRatioMap);
	}
			
	/**
	 * <p>Sets an attribute for the given layer. If already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setLayerAttribute(long layerId, String key, String value)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		layerAttributeMap.get(layerId).put(key, value);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.layerModified(layerId));
	}
	
	/**
	 * <p>Sets the layer attributes. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setLayerAttributeMap(long layerId, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		
		Map<String, String> aux = layerAttributeMap.get(layerId);
		aux.clear();
		aux.putAll(attributes);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.layerModified(layerId));
	}
	
	/**
	 * Sets the default layer identifier.
	 * 
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public void setLayerDefaultId(long layerId)
	{
		checkLayerIsActive(layerId);
		layerDefaultId = layerId;
	}

	/**
	 * <p>Sets the layer description.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param description Description (when {@code null}, it will be set to empty)
	 * @since 0.3.0
	 */
	public void setLayerDescription(long layerId, String description)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		if (getLayerDescription(layerId).equals(description)) return;
		
		layerDescriptionMap.put(layerId, description == null ? "" : description);
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.layerModified(layerId));
	}

	/**
	 * <p>Sets the layer name.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param name Name (when {@code null}, it will be set to 'Layer ' + layerId)
	 * @since 0.3.0
	 */
	public void setLayerName(long layerId, String name)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		if (getLayerName(layerId).equals(name)) return;

		layerNameMap.put(layerId, name == null ? "" : name);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.layerModified(layerId));
	}	

	/**
	 * <p>Sets an attribute for a given link. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param linkId Link identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setLinkAttribute(long linkId, String key, String value)
	{
		long layerId = getLayerDefaultId();
		setLinkAttribute(layerId, linkId, key, value);
	}

	/**
	 * <p>Sets an attribute for a given link. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setLinkAttribute(long layerId, long linkId, String key, String value)
	{
		checkIsModifiable();
		checkLinkIsActive(layerId, linkId);
		
		linkAttributesMap.get(layerId).get(linkId).put(key, value);		
		if (isLinkCoupled(layerId, linkId))
		{
			Pair<Long, Long> lowerLayerDemand = getLinkCoupledLowerLayerDemand(layerId, linkId);

			long lowerLayerId = lowerLayerDemand.getFirst();
			long demandId = lowerLayerDemand.getSecond();
			demandAttributesMap.get(lowerLayerId).get(demandId).put(key, value);
			if (isActionMonitoringActivated()) actionsPerformed.add(Factory.demandModified(lowerLayerId, demandId));
		}
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(layerId, linkId));
		
	}
	
	/**
	 * <p>Sets the attributes for a given link. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param linkId Link identifier
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setLinkAttributeMap(long linkId, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		setLinkAttributeMap(layerId, linkId, attributes);
	}

	/**
	 * <p>Sets the attributes for a given link. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setLinkAttributeMap(long layerId, long linkId, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkLinkIsActive(layerId, linkId);
		
		Map<String, String> aux = linkAttributesMap.get(layerId).get(linkId);
		aux.clear();
		aux.putAll(attributes);
		if (isLinkCoupled(layerId, linkId))
		{
			Pair<Long, Long> lowerLayerDemand = getLinkCoupledLowerLayerDemand(layerId, linkId);

			long lowerLayerId = lowerLayerDemand.getFirst();
			long demandId = lowerLayerDemand.getSecond();
			demandAttributesMap.get(lowerLayerId).get(demandId).putAll(attributes);
			if (isActionMonitoringActivated()) actionsPerformed.add(Factory.demandModified(lowerLayerId, demandId));
		}
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(layerId, linkId));
	}

	/**
	 * <p>Sets the capacity of a link.</p>
	 *
	 * <p><b>Important</b>: If link is coupled, link capacity is given by the carried traffic of the lower layer demand.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param linkId Link identifier
	 * @param u_e Link capacity. It must be greater or equal than zero
	 * @since 0.3.0
	 */
	public void setLinkCapacity(long linkId, double u_e)
	{
		long layerId = getLayerDefaultId();
		setLinkCapacity(layerId, linkId, u_e);
	}

	/**
	 * <p>Sets the capacity of a link.</p>
	 *
	 * <p><b>Important</b>: If link is coupled, link capacity is given by the carried traffic of the lower layer demand.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param u_e Link capacity. It must be greater or equal than zero
	 * @since 0.3.0
	 */
	public void setLinkCapacity(long layerId, long linkId, double u_e)
	{
		checkIsModifiable();
		if (isLinkCoupled(layerId, linkId)) throw new Net2PlanException("Link is coupled, its capacity is equal to the carried traffic of its associated lower layer demand");
		if (Math.abs(u_e) < 1E-10) u_e = 0;
		if (u_e < 0) throw new Net2PlanException("Link capacity must be greater or equal than zero");
		
		linkCapacityMap.get(layerId).put(linkId, u_e);
		updateCacheLinkCapacity(layerId, linkId);
		updateCacheLinkOversubscribed(layerId, linkId);
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(layerId, linkId));
	}
	
	/**
	 * <p>Sets the capacity per link.</p>
	 * 
	 * <p><b>Important</b>: If links are coupled, link capacities are given by the carried traffic of the lower layer demands.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} actions will be issued.</p>
	 * 
	 * @param u_e Capacity per link
	 * @since 0.3.0
	 */
	public void setLinkCapacityMap(Map<Long, Double> u_e)
	{
		long layerId = getLayerDefaultId();
		setLinkCapacityMap(layerId, u_e);
	}
	
	/**
	 * <p>Sets the capacity per link.</p>
	 * 
	 * <p><b>Important</b>: If links are coupled, link capacities are given by the carried traffic of the lower layer demands.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param u_e Capacity per link
	 * @since 0.3.0
	 */
	public void setLinkCapacityMap(long layerId, Map<Long, Double> u_e)
	{
		checkIsModifiable();
		
		for(Entry<Long, Double> entry : u_e.entrySet())
		{
			long linkId = entry.getKey();
			double capacity = entry.getValue();
			setLinkCapacity(layerId, linkId, capacity);
		}
	}
	
	/**
	 * <p>Sets the link capacity units name.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param capacityUnitsName Capacity units name
	 * @since 0.3.0
	 */
	public void setLinkCapacityUnitsName(String capacityUnitsName)
	{
		long layerId = getLayerDefaultId();
		setLinkCapacityUnitsName(layerId, capacityUnitsName);
	}
	
	/**
	 * <p>Sets the link capacity units name.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LAYER_MODIFIED LAYER_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param capacityUnitsName Capacity units name
	 * @since 0.3.0
	 */
	public void setLinkCapacityUnitsName(long layerId, String capacityUnitsName)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		if (getLinkCapacityUnitsName(layerId).equals(capacityUnitsName)) return;

		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		for(long linkId : linkIds_thisLayer)
			if (isLinkCoupled(layerId, linkId))
				throw new Net2PlanException("Link capacity units name cannot be modified since there is some coupling with other layers");
		
		layerLinkCapacityUnitsNameMap.put(layerId, capacityUnitsName == null ? "" : capacityUnitsName);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.layerModified(layerId));
	}
	
	/**
	 * <p>Sets the capacity per link, in increasing order of identifier.</p>
	 * 
	 * <p><b>Important</b>: If links are coupled, link capacities are given by the carried traffic of the lower layer demands.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} actions will be issued.</p>
	 * 
	 * @param u_e Capacity per link
	 * @since 0.2.0
	 */
	public void setLinkCapacityVector(double[] u_e)
	{
		long layerId = getLayerDefaultId();
		setLinkCapacityVector(layerId, u_e);
	}

	/**
	 * <p>Sets the capacity per link, in increasing order of identifier.</p>
	 * 
	 * <p><b>Important</b>: If links are coupled, link capacities are given by the carried traffic of the lower layer demands.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param u_e Capacity per link
	 * @since 0.3.0
	 */
	public void setLinkCapacityVector(long layerId, double[] u_e)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		
		Set<Long> linkIds_thisLayer = getLinkIds(layerId);
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds_thisLayer);
		for(long linkId : linkIds_thisLayer)
		{
			int e = linkId2LinearIndex.get(linkId);
			setLinkCapacity(layerId, linkId, u_e[e]);
		}
	}

	/**
	 * <p>Sets a link as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_DOWN LINK_DOWN} action will be issued.</p>
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void setLinkDown(long linkId)
	{
		long layerId = getLayerDefaultId();
		setLinkDown(layerId, linkId);
	}
	
	/**
	 * <p>Sets a link as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_DOWN LINK_DOWN} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void setLinkDown(long layerId, long linkId)
	{
		Set<Long> aux = new LinkedHashSet<Long>();
		aux.add(linkId);
		
		setLinksDown(layerId, aux);
	}
	
	/**
	 * <p>Sets the length of a link.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param linkId Link identifier
	 * @param linkLengthInKm Link length measured in km. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairPhysicalDistance()} method
	 * @since 0.3.0
	 */
	public void setLinkLengthInKm(long linkId, double linkLengthInKm)
	{
		long layerId = getLayerDefaultId();
		setLinkLengthInKm(layerId, linkId, linkLengthInKm);
	}

	/**
	 * <p>Sets the length of a link.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param lengthInKm Link length measured in km. It must be greater or equal than zero. Physical distance between node pairs can be obtained through the {@link #getNodePairEuclideanDistance(long, long) getNodePairPhysicalDistance()} method
	 * @since 0.3.0
	 */
	public void setLinkLengthInKm(long layerId, long linkId, double lengthInKm)
	{
		checkIsModifiable();
		if (lengthInKm < 0) throw new Net2PlanException("Link length must be greater or equal than zero");
		
		checkLinkIsActive(layerId, linkId);
		this.linkLengthInKmMap.get(layerId).put(linkId, lengthInKm);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(layerId, linkId));
	}

	/**
	 * <p>Sets the propagation speed in km/s of a link.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param linkId Link identifier
	 * @param propagationSpeedInKmPerSecond Propagation speed in km/s
	 * @since 0.3.0
	 */
	public void setLinkPropagationSpeedInKmPerSecond(long linkId, double propagationSpeedInKmPerSecond)
	{
		long layerId = getLayerDefaultId();
		setLinkPropagationSpeedInKmPerSecond(layerId, linkId, propagationSpeedInKmPerSecond);
	}

	/**
	 * <p>Sets the propagation speed in km/s of a link.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @param propagationSpeedInKmPerSecond Propagation speed in km/s
	 * @since 0.3.0
	 */
	public void setLinkPropagationSpeedInKmPerSecond(long layerId, long linkId, double propagationSpeedInKmPerSecond)
	{
		checkLinkIsActive(layerId, linkId);
		if (propagationSpeedInKmPerSecond <= 0) throw new Net2PlanException("Propagation speed must be greater than zero");
		
		linkPropagationSpeedInKmPerSecondMap.get(layerId).put(linkId, propagationSpeedInKmPerSecond);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(layerId, linkId));
	}

	/**
	 * <p>Sets the propagation speed in km/s per link.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} actions will be issued.</p>
	 * 
	 * @param propagationSpeedInKmPerSecond Propagation speed in km/s per link
	 * @since 0.3.0
	 */
	public void setLinkPropagationSpeedInKmPerSecondMap(Map<Long, Double> propagationSpeedInKmPerSecond)
	{
		long layerId = getLayerDefaultId();
		setLinkPropagationSpeedInKmPerSecondMap(layerId, propagationSpeedInKmPerSecond);
	}

	/**
	 * <p>Sets the propagation speed in km/s per link.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_MODIFIED LINK_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param propagationSpeedInKmPerSecond Propagation speed in km/s per link
	 * @since 0.3.0
	 */
	public void setLinkPropagationSpeedInKmPerSecondMap(long layerId, Map<Long, Double> propagationSpeedInKmPerSecond)
	{
		for(Entry<Long, Double> entry : propagationSpeedInKmPerSecond.entrySet())
		{
			long linkId = entry.getKey();
			double propagationSpeedInKmPerSecond_thisLink = entry.getValue();
			setLinkPropagationSpeedInKmPerSecond(layerId, linkId, propagationSpeedInKmPerSecond_thisLink);
		}
	}
	
	/**
	 * <p>Sets a link as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} action will be issued.</p>
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void setLinkUp(long linkId)
	{
		long layerId = getLayerDefaultId();
		setLinkUp(layerId, linkId);
	}
	
	/**
	 * <p>Sets a link as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void setLinkUp(long layerId, long linkId)
	{
		Set<Long> aux = new LinkedHashSet<Long>();
		aux.add(linkId);
		
		setLinksUp(layerId, aux);
	}
	
	/**
	 * <p>Sets some links as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_DOWN LINK_DOWN} actions will be issued.</p>

	 * @param linkIds Link identifiers per layer
	 * @since 0.3.0
	 */
	public void setLinksDown(Set<Long> linkIds)
	{
		long layerId = getLayerDefaultId();
		setLinksDown(layerId, linkIds);
	}
	
	/**
	 * <p>Sets some links as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_DOWN LINK_DOWN} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void setLinksDown(long layerId, Set<Long> linkIds)
	{
		setNodesAndLinksDown(layerId, null, linkIds);
	}
	
	/**
	 * <p>Sets some links as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>
	 * 
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void setLinksUp(Set<Long> linkIds)
	{
		long layerId = getLayerDefaultId();
		setLinksUp(layerId, linkIds);
	}

	/**
	 * <p>Sets some links as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>
	 * 
	 * @param linkIds Link identifiers per layer
	 * @since 0.3.0
	 */
	public void setLinksUp(Map<Long, Set<Long>> linkIds)
	{
		setNodesAndLinksUp(null, linkIds);
	}

	/**
	 * <p>Sets some links as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>

	 * @param layerId Layer identifier
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void setLinksUp(long layerId, Set<Long> linkIds)
	{
		setNodesAndLinksUp(layerId, null, linkIds);
	}
	
	/**
	 * Sets a network attribute. If already exists, it will be overriden.
	 *
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.2.0
	 */
	public void setNetworkAttribute(String key, String value)
	{
		checkIsModifiable();
		networkAttributes.put(key, value);
	}

	/**
	 * Sets the network attributes attributes. Any previous attribute will be removed.
	 *
	 * @param attributes Map of attributes
	 * @since 0.2.0
	 */
	public void setNetworkAttributeMap(Map<String, String> attributes)
	{
		checkIsModifiable();
		networkAttributes.clear();
		networkAttributes.putAll(attributes);
	}

	/**
	 * Sets the network description.
	 * 
	 * @param description Description (when {@code null}, it will be set to empty)
	 * @since 0.2.3
	 */
	public void setNetworkDescription(String description)
	{
		checkIsModifiable();
		if (getNetworkDescription().equals(description)) return;
		networkDescription = description == null ? "" : description;
	}

	/**
	 * Sets the network name.
	 *
	 * @param name Name (when {@code null}, it will be set to 'Layer ' + layerId)
	 * @since 0.2.3
	 */
	public void setNetworkName(String name)
	{
		checkIsModifiable();
		if (getNetworkName().equals(name)) return;
		networkName = name == null ? "" : name;
	}	
	
	/**
	 * <p>Sets an attribute for a given node. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_MODIFIED NODE_MODIFIED} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setNodeAttribute(long nodeId, String key, String value)
	{
		checkIsModifiable();
		checkNodeIsActive(nodeId);
		nodeAttributes.get(nodeId).put(key, value);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeModified(nodeId));
	}
	
	/**
	 * <p>Sets the attributes for a given node. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_MODIFIED NODE_MODIFIED} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setNodeAttributeMap(long nodeId, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkNodeIsActive(nodeId);

		Map<String, String> auxNodeAttributes = nodeAttributes.get(nodeId);
		auxNodeAttributes.clear();
		auxNodeAttributes.putAll(attributes);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeModified(nodeId));
	}
	
	/**
	 * <p>Sets a node as down.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_DOWN NODE_DOWN} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void setNodeDown(long nodeId)
	{
		Set<Long> aux = new LinkedHashSet<Long>();
		aux.add(nodeId);
		
		setNodesDown(aux);
	}
	
	/**
	 * <p>Sets the node name.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_MODIFIED NODE_MODIFIED} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @param name Name (when {@code null}, it will be set to 'Node ' + nodeId)
	 * @since 0.3.0
	 */
	public void setNodeName(long nodeId, String name)
	{
		checkIsModifiable();
		checkNodeIsActive(nodeId);
		nodeNames.put(nodeId, name == null ? "Node " + nodeId : name);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeModified(nodeId));
	}

	/**
	 * <p>Sets a node as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_UP NODE_UP} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void setNodeUp(long nodeId)
	{
		Set<Long> aux = new LinkedHashSet<Long>();
		aux.add(nodeId);
		
		setNodesUp(aux);
	}
	
	/**
	 * <p>Sets the node position.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_MODIFIED NODE_MODIFIED} action will be issued.</p>
	 *
	 * @param nodeId Node identifier
	 * @param x Node position in x-axis
	 * @param y Node position in y-axis
	 * @since 0.3.0
	 */
	public void setNodeXYPosition(long nodeId, double x, double y)
	{
		checkIsModifiable();
		checkNodeIsActive(nodeId);
		nodeXYPositionMap.put(nodeId, new UnmodifiablePoint2D(x, y));
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeModified(nodeId));
	}
	
	/**
	 * <p>Sets some nodes and links as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_DOWN NODE_DOWN} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_DOWN LINK_DOWN} actions will be issued.</p>
	 * 
	 * @param nodeIds Node identifiers
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void setNodesAndLinksDown(Set<Long> nodeIds, Set<Long> linkIds)
	{
		long layerId = getLayerDefaultId();
		setNodesAndLinksDown(layerId, nodeIds, linkIds);
	}

	/**
	 * <p>Sets some nodes and links as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_DOWN NODE_DOWN} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_DOWN LINK_DOWN} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeIds Node identifiers
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void setNodesAndLinksDown(long layerId, Set<Long> nodeIds, Set<Long> linkIds)
	{
		Map<Long, Set<Long>> aux = new LinkedHashMap<Long, Set<Long>>();
		aux.put(layerId, linkIds);
		
		setNodesAndLinksDown(nodeIds, aux);
	}
	
	/**
	 * <p>Sets some nodes and links as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_DOWN NODE_DOWN} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_DOWN LINK_DOWN} actions will be issued.</p>
	 * 
	 * @param nodeIds Node identifiers
	 * @param linkIds Link identifiers per layer
	 * @since 0.3.0
	 */
	public void setNodesAndLinksDown(Set<Long> nodeIds, Map<Long, Set<Long>> linkIds)
	{
		checkIsModifiable();

		Map<Long, Set<Long>> demandsToUpdate = new LinkedHashMap<Long, Set<Long>>();
		
		Set<Long> layerIds_thisNetPlan = getLayerTopologicalSortIds();
		for(long layerId : layerIds_thisNetPlan)
		{
			RoutingType routingType_thisLayer = getRoutingType(layerId);
			if (routingType_thisLayer == RoutingType.SOURCE_ROUTING)
			{
				Set<Long> routesDown_thisLayer = new LinkedHashSet<Long>();
				Set<Long> segmentsDown_thisLayer = new LinkedHashSet<Long>();

				if (nodeIds != null)
				{
					for(long nodeId : nodeIds)
					{
						if (isNodeDown(nodeId)) continue;

						routesDown_thisLayer.addAll(getNodeAssociatedRoutes(layerId, nodeId));
						segmentsDown_thisLayer.addAll(getNodeAssociatedProtectionSegments(layerId, nodeId));
					}
				}

				if (linkIds != null && linkIds.containsKey(layerId))
				{
					for(long linkId : linkIds.get(layerId))
					{
						if (isLinkDown(layerId, linkId)) continue;

						routesDown_thisLayer.addAll(getLinkAssociatedRoutes(layerId, linkId));
						segmentsDown_thisLayer.addAll(getLinkAssociatedProtectionSegments(layerId, linkId));

						linksDown.get(layerId).add(linkId);
						if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkDown(layerId, linkId));
					}
				}

				for(long segmentId : segmentsDown_thisLayer) setProtectionSegmentDown(layerId, segmentId);
				for(long routeId : routesDown_thisLayer) setRouteDown(layerId, routeId);
			}
			else
			{
				Set<Pair<Long, Long>> forwardingRulesToDeactivate = new LinkedHashSet<Pair<Long, Long>>();
				
				if (linkIds != null && linkIds.containsKey(layerId))
				{
					for(long linkId : linkIds.get(layerId))
					{
						if (isLinkDown(layerId, linkId)) continue;

						linksDown.get(layerId).add(linkId);
						if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkDown(layerId, linkId));
						
						Set<Pair<Long, Long>> forwardingRules_thisLink = getLinkForwardingRules(layerId, linkId);
						forwardingRulesToDeactivate.addAll(forwardingRules_thisLink);
					}
				}

				Set<Long> demandsToUpdate_thisLayer = new LinkedHashSet<Long>();

				if (nodeIds != null)
				{
					for(long nodeId : nodeIds)
					{
						if (isNodeDown(nodeId)) continue;
						
						forwardingRulesToDeactivate.addAll(getNodeForwardingRules(layerId, nodeId));
						
						Set<Long> incomingLinkIds = getNodeIncomingLinks(layerId, nodeId);
						for(long linkId : incomingLinkIds)
							forwardingRulesToDeactivate.addAll(getLinkForwardingRules(layerId, linkId));
					}
				}
				
				for(Pair<Long, Long> demandLinkPair : forwardingRulesToDeactivate)
				{
					forwardingRuleDemandLinkPairDownMap.get(layerId).add(demandLinkPair);
					demandsToUpdate_thisLayer.add(demandLinkPair.getFirst());
				}
				
				demandsToUpdate.put(layerId, demandsToUpdate_thisLayer);
			}
		}

		if (nodeIds != null)
		{
			for(long nodeId : nodeIds)
			{
				if (isNodeDown(nodeId)) continue;

				nodesDown.add(nodeId);
				if (isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeDown(nodeId));
			}
		}
		
		for(Entry<Long, Set<Long>> entry : demandsToUpdate.entrySet())
		{
			long layerId = entry.getKey();
			Set<Long> demandsToUpdate_thisLayer = entry.getValue();
			for(long demandId : demandsToUpdate_thisLayer) updateCacheDemandForwarding(layerId, demandId);
		}
	}
	
	/**
	 * <p>Sets some nodes and links as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_UP NODE_UP} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>
	 * 
	 * @param nodeIds Node identifiers
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void setNodesAndLinksUp(Set<Long> nodeIds, Set<Long> linkIds)
	{
		long layerId = getLayerDefaultId();
		setNodesAndLinksUp(layerId, nodeIds, linkIds);
	}

	/**
	 * <p>Sets some nodes and links as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_UP NODE_UP} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param nodeIds Node identifiers
	 * @param linkIds Link identifiers
	 * @since 0.3.0
	 */
	public void setNodesAndLinksUp(long layerId, Set<Long> nodeIds, Set<Long> linkIds)
	{
		Map<Long, Set<Long>> aux = new LinkedHashMap<Long, Set<Long>>();
		aux.put(layerId, linkIds);
		
		setNodesAndLinksUp(nodeIds, aux);
	}
	
	/**
	 * <p>Sets some nodes and links as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_UP NODE_UP} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#LINK_UP LINK_UP} actions will be issued.</p>
	 * 
	 * @param nodeIds Node identifiers
	 * @param linkIds Link identifiers per layer
	 * @since 0.3.0
	 */
	public void setNodesAndLinksUp(Set<Long> nodeIds, Map<Long, Set<Long>> linkIds)
	{
		checkIsModifiable();
		
		if (nodeIds != null)
		{
			for(long nodeId : new LinkedHashSet<Long>(nodeIds))
			{
				if (!isNodeDown(nodeId)) throw new Net2PlanException("Node " + nodeId + " was not down");
				
				nodesDown.remove(nodeId);
				if (isActionMonitoringActivated()) actionsPerformed.add(Factory.nodeUp(nodeId));
			}		
		}
		
		if (linkIds != null)
		{
			for(Entry<Long, Set<Long>> entry : linkIds.entrySet())
			{
				long layerId = entry.getKey();
				Set<Long> linkIds_thisLayer = entry.getValue();
				if (linkIds_thisLayer != null)
				{
					for(long linkId : new LinkedHashSet<Long>(linkIds_thisLayer))
					{
						if (!isLinkDown(layerId, linkId)) throw new Net2PlanException("Link " + linkId + " at layer " + layerId + " was not down");

						linksDown.get(layerId).remove(linkId);
						if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkUp(layerId, linkId));
					}
				}
			}		
		}
		
		Set<Long> layerIds_thisNetPlan = getLayerIds();
		for(long layerId : layerIds_thisNetPlan)
		{
			RoutingType routingType_thisLayer = getRoutingType(layerId);
			if (routingType_thisLayer != RoutingType.HOP_BY_HOP_ROUTING) continue;
			
			Set<Long> demandsToUpdate = new LinkedHashSet<Long>();
			Set<Pair<Long, Long>> forwardingRulesDown_thisLayer = new LinkedHashSet<Pair<Long, Long>>(getForwardingRulesDown(layerId));
			for(Pair<Long, Long> demandLinkPair : forwardingRulesDown_thisLayer)
			{
				long linkId = demandLinkPair.getSecond();
				long originNodeId = getLinkOriginNode(layerId, linkId);
				long destinationNodeId = getLinkDestinationNode(layerId, linkId);
				if (!isNodeDown(originNodeId) && !isNodeDown(destinationNodeId) && !isLinkDown(layerId, linkId))
				{
					demandsToUpdate.add(demandLinkPair.getFirst());
					forwardingRuleDemandLinkPairDownMap.get(layerId).remove(demandLinkPair);
				}
			}
			
			for(long demandId : demandsToUpdate) updateCacheDemandForwarding(layerId, demandId);
		}
	}
	
	/**
	 * <p>Sets a set of nodes as down.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_DOWN NODE_DOWN} actions will be issued.</p>
	 * 
	 * @param nodeIds Node identifiers
	 * @since 0.3.0
	 */
	public void setNodesDown(Set<Long> nodeIds)
	{
		setNodesAndLinksDown(nodeIds, new LinkedHashSet<Long>());
	}
	
	/**
	 * <p>Sets a set of nodes as up.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#NODE_UP NODE_UP} actions will be issued.</p>
	 * 
	 * @param nodeIds Node identifiers
	 * @since 0.3.0
	 */
	public void setNodesUp(Set<Long> nodeIds)
	{
		setNodesAndLinksUp(nodeIds, new LinkedHashSet<Long>());
	}
	
	/**
	 * <p>Sets an attribute for a given protection segment. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setProtectionSegmentAttribute(long segmentId, String key, String value)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentAttribute(layerId, segmentId, key, value);
	}

	/**
	 * <p>Sets an attribute for a given protection segment. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setProtectionSegmentAttribute(long layerId, long segmentId, String key, String value)
	{
		checkIsModifiable();
		checkProtectionSegmentIsActive(layerId, segmentId);
		segmentAttributes.get(layerId).get(segmentId).put(key, value);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.protectionSegmentModified(layerId, segmentId));
	}
	
	/**
	 * <p>Sets the attributes for a given protection segment. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @param attributes Map of attributes
	 * @since 0.3.0
	 */
	public void setProtectionSegmentAttributeMap(long segmentId, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentAttributeMap(layerId, segmentId, attributes);
	}

	/**
	 * <p>Sets the attributes for a given protection segment. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setProtectionSegmentAttributeMap(long layerId, long segmentId, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkProtectionSegmentIsActive(layerId, segmentId);

		Map<String, String> auxSegmentAttributes = segmentAttributes.get(layerId).get(segmentId);
		auxSegmentAttributes.clear();
		auxSegmentAttributes.putAll(attributes);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.protectionSegmentModified(layerId, segmentId));
	}
	
	/**
	 * <p>Sets a protection segment as down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param segmentId Protection segment identifier
	 * @since 0.3.0
	 */
	public void setProtectionSegmentDown(long segmentId)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentDown(layerId, segmentId);
	}
	
	/**
	 * <p>Sets a protection segment as down.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Protection segment identifier
	 * @since 0.3.0
	 */
	public void setProtectionSegmentDown(long layerId, long segmentId)
	{
		checkIsModifiable();
		checkProtectionSegmentIsActive(layerId, segmentId);
		
		if (isProtectionSegmentDown(layerId, segmentId)) return;

		double u_s = getProtectionSegmentReservedCapacity(layerId, segmentId);
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		for(long linkId : seqLinks)
		{
			double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
			linkReservedCapacityMap.get(layerId).put(linkId, r_e - u_s);
		}

		long itemId = -1 - segmentId;
		Set<Long> routeIds_thisSegment = getProtectionSegmentRoutes(layerId, segmentId);
		for(long routeId : routeIds_thisSegment)
		{
			List<Long> seqLinks_thisRoute = getRouteSequenceOfLinks(layerId, routeId);
			if (seqLinks_thisRoute.contains(itemId)) setRouteDown(layerId, routeId);
		}
		
		segmentsDown.get(layerId).add(segmentId);
		
		if (isActionMonitoringActivated()) actionsPerformed.add(Factory.protectionSegmentModified(layerId, segmentId));
	}
	
	/**
	 * <p>Sets the reserved capacity of a protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param segmentId Segment identifier
	 * @param u_s Reserved capacity
	 * @since 0.3.0
	 */
	public void setProtectionSegmentReservedCapacity(long segmentId, double u_s)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentReservedCapacity(layerId, segmentId, u_s);
	}

	/**
	 * <p>Sets the reserved capacity of a protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Segment identifier
	 * @param u_s Reserved capacity
	 * @since 0.3.0
	 */
	public void setProtectionSegmentReservedCapacity(long layerId, long segmentId, double u_s)
	{
		checkIsModifiable();
		checkProtectionSegmentIsActive(layerId, segmentId);
		if (u_s <= 0) throw new Net2PlanException("Reserved capacity must be greater than zero");
		
		
		boolean segmentDown = segmentsDown.get(layerId).contains(segmentId);
		
		List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
		if (!segmentDown)
		{
			double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
			double y_s = getProtectionSegmentCarriedTraffic(layerId, segmentId);
			if (y_s > u_s + PRECISION_FACTOR) throw new Net2PlanException("Unable to change reserved capacity for protection segment " + segmentId + ". It will become over-subscribed");
			
			if (segmentReservedCapacity.get(layerId).containsKey(segmentId))
			{
				for(long linkId : seqLinks)
				{
					double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
					linkReservedCapacityMap.get(layerId).put(linkId, r_e - getProtectionSegmentReservedCapacity(layerId, segmentId));
				}
			}

			for(long linkId : seqLinks)
			{
				double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
				linkReservedCapacityMap.get(layerId).put(linkId, r_e + u_s);
			}
		}
		
		segmentReservedCapacity.get(layerId).put(segmentId, u_s);
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.protectionSegmentModified(layerId, segmentId));
	}
	
	/**
	 * <p>Sets the reserved capacity per protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @param u_s Reserved capacity per protection segment
	 * @since 0.3.0
	 */
	public void setProtectionSegmentReservedCapacityMap(Map<Long, Double> u_s)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentReservedCapacityMap(layerId, u_s);
	}
	
	/**
	 * <p>Sets the reserved capacity per protection segment.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param u_s Reserved capacity per protection segment
	 * @since 0.3.0
	 */
	public void setProtectionSegmentReservedCapacityMap(long layerId, Map<Long, Double> u_s)
	{
		checkIsModifiable();
		
		for(Entry<Long, Double> entry : u_s.entrySet())
		{
			long segmentId = entry.getKey();
			double reservedCapacity = entry.getValue();
			setProtectionSegmentReservedCapacity(layerId, segmentId, reservedCapacity);
		}
	}
	
	/**
	 * <p>Sets the reserved capacity per protection segment, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @param u_s Reserved capacity per protection segment
	 * @since 0.2.0
	 */
	public void setProtectionSegmentReservedCapacityVector(double[] u_s)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentReservedCapacityVector(layerId, u_s);
	}

	/**
	 * <p>Sets the reserved capacity per protection segment, in increasing order of identifier.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param u_s Reserved capacity per protection segment
	 * @since 0.3.0
	 */
	public void setProtectionSegmentReservedCapacityVector(long layerId, double[] u_s)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		
		Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
		Map<Long, Integer> segmentId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(segmentIds_thisLayer);
		for(long segmentId : segmentIds_thisLayer)
		{
			int s = segmentId2LinearIndex.get(segmentId);
			setProtectionSegmentReservedCapacity(layerId, segmentId, u_s[s]);
		}
	}

	/**
	 * <p>Sets the given protection segment as up. If it is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param segmentId Segment identifier
	 * @since 0.3.0
	 */
	public void setProtectionSegmentUp(long segmentId)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentUp(layerId, segmentId);
	}
	
	/**
	 * <p>Sets the given protection segment as up. If it is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentId Segment identifier
	 * @since 0.3.0
	 */
	public void setProtectionSegmentUp(long layerId, long segmentId)
	{
		Set<Long> aux = new LinkedHashSet<Long>();
		aux.add(segmentId);
		
		setProtectionSegmentsUp(layerId, aux);
	}
	
	/**
	 * <p>Sets the given protection segments as up. If some of them is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @param segmentIds Segment identifiers
	 * @since 0.3.0
	 */
	public void setProtectionSegmentsUp(Set<Long> segmentIds)
	{
		long layerId = getLayerDefaultId();
		setProtectionSegmentsUp(layerId, segmentIds);
	}
	
	/**
	 * <p>Sets the given protection segments as up. If some of them is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#PROTECTION_SEGMENT_MODIFIED PROTECTION_SEGMENT_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param segmentIds Segment identifiers
	 * @since 0.3.0
	 */
	public void setProtectionSegmentsUp(long layerId, Set<Long> segmentIds)
	{
		checkIsModifiable();
		
		Set<Long> nodesDown_thisNetPlan = getNodesDown();
		Set<Long> linksDown_thisLayer = getLinksDown(layerId);
		for(long segmentId : new LinkedHashSet<Long>(segmentIds))
		{
			checkProtectionSegmentIsActive(layerId, segmentId);

			if (!isProtectionSegmentRestorable(layerId, segmentId))
			{
				List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
				if (CollectionUtils.containsAny(seqLinks, linksDown_thisLayer)) throw new Net2PlanException("Bad - Some link on protection segment " + segmentId + " at layer " + layerId + " is down");

				List<Long> seqNodes = getProtectionSegmentSequenceOfNodes(layerId, segmentId);
				if (CollectionUtils.containsAny(seqNodes, nodesDown_thisNetPlan)) throw new Net2PlanException("Bad - Some node on protection segment " + segmentId + " at layer " + layerId + " is down");
			}
			
			double u_s = getProtectionSegmentReservedCapacity(layerId, segmentId);
			List<Long> seqLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
			for(long linkId : seqLinks)
			{
				double r_e = getLinkCapacityReservedForProtection(layerId, linkId);
				linkReservedCapacityMap.get(layerId).put(linkId, r_e + u_s);
			}
			
			segmentsDown.get(layerId).remove(segmentId);

			if (isActionMonitoringActivated()) actionsPerformed.add(Factory.protectionSegmentModified(layerId, segmentId));
		}
	}

	/**
	 * <p>Sets an attribute for a given route at the first layer. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param routeId Route identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setRouteAttribute(long routeId, String key, String value)
	{
		long layerId = getLayerDefaultId();
		setRouteAttribute(layerId, routeId, key, value);
	}

	/**
	 * <p>Sets an attribute for a given route. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setRouteAttribute(long layerId, long routeId, String key, String value)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		routeAttributes.get(layerId).get(routeId).put(key, value);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
	}
	
	/**
	 * <p>Sets the attributes for a given route. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param routeId Route identifier
	 * @param attributes Map of attributes
	 * @since 0.3.0
	 */
	public void setRouteAttributeMap(long routeId, Map<String, String> attributes)
	{
		long layerId = getLayerDefaultId();
		setRouteAttributeMap(layerId, routeId, attributes);
	}

	/**
	 * <p>Sets the attributes for a given route at the given layer. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @param attributes Map of attributes (in case of null, an empty map will be assumed)
	 * @since 0.3.0
	 */
	public void setRouteAttributeMap(long layerId, long routeId, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);

		Map<String, String> auxRouteAttributes = routeAttributes.get(layerId).get(routeId);
		auxRouteAttributes.clear();
		auxRouteAttributes.putAll(attributes);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
	}

	/**
	 * <p>Sets the route carried traffic.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param routeId Route identifier
	 * @param x_p Carried traffic (greater or equal than zero)
	 * @since 0.3.0
	 */
	public void setRouteCarriedTraffic(long routeId, double x_p)
	{
		long layerId = getLayerDefaultId();
		setRouteCarriedTraffic(layerId, routeId, x_p);
	}
	
	/**
	 * <p>Sets the route carried traffic.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @param x_p Carried traffic (greater or equal than zero)
	 * @since 0.3.0
	 */
	public void setRouteCarriedTraffic(long layerId, long routeId, double x_p)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		if (x_p < 0) throw new Net2PlanException("Carried traffic must be greater or equal than zero");
		
		boolean routeUp = !isRouteDown(layerId, routeId);
		if (routeUp) updateCacheRouteRemoveCarriedTraffic(layerId, routeId);
		routeCarriedTrafficMap.get(layerId).put(routeId, x_p);
		if (routeUp) updateCacheRouteAddCarriedTraffic(layerId, routeId);
		
		if (routeSameOccupiedCapacityAsCarriedTraffic.get(layerId).contains(routeId))
			setRouteOccupiedCapacity(layerId, routeId, -1);
		else if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
	}
	
	/**
	 * <p>Sets the carried traffic per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param x_p Carried traffic per route
	 * @since 0.3.0
	 */
	public void setRouteCarriedTrafficMap(Map<Long, Double> x_p)
	{
		long layerId = getLayerDefaultId();
		setRouteCarriedTrafficMap(layerId, x_p);
	}
	
	/**
	 * <p>Sets the carried traffic per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param x_p Carried traffic per route
	 * @since 0.3.0
	 */
	public void setRouteCarriedTrafficMap(long layerId, Map<Long, Double> x_p)
	{
		checkIsModifiable();
		
		for(Entry<Long, Double> entry : x_p.entrySet())
		{
			long routeId = entry.getKey();
			double carriedTraffic = entry.getValue();
			setRouteCarriedTraffic(layerId, routeId, carriedTraffic);
		}
	}
	
	/**
	 * <p>Sets the carried traffic per route, in increasing order of route identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param x_p Carried traffic per route
	 * @since 0.2.0
	 */
	public void setRouteCarriedTrafficVector(double[] x_p)
	{
		long layerId = getLayerDefaultId();
		setRouteCarriedTrafficVector(layerId, x_p);
	}

	/**
	 * <p>Sets the carried traffic per route, in increasing order of route identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param x_p Carried traffic per route
	 * @since 0.3.0
	 */
	public void setRouteCarriedTrafficVector(long layerId, double[] x_p)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		Map<Long, Integer> routeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(routeIds_thisLayer);
		for(long routeId : routeIds_thisLayer)
		{
			int p = routeId2LinearIndex.get(routeId);
			setRouteCarriedTraffic(layerId, routeId, x_p[p]);
		}
	}

	/**
	 * <p>Sets a route as down.</p>
	 * 
	 * <p><b>Important</b>: Route keeps its carried traffic assignment, but it is not taking into account for metrics.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void setRouteDown(long routeId)
	{
		long layerId = getLayerDefaultId();
		setRouteDown(layerId, routeId);
	}
	
	/**
	 * <p>Sets a route as down.</p>
	 * 
	 * <p><b>Important</b>: Route keeps its carried traffic assignment, but it is not taking into account for metrics.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void setRouteDown(long layerId, long routeId)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		
		if (isRouteDown(layerId, routeId)) return;
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
		
		updateCacheRouteRemoveCarriedTraffic(layerId, routeId);
		updateCacheRouteRemoveOccupiedCapacity(layerId, routeId);
		
		routesDown.get(layerId).add(routeId);
	}
	
	/**
	 * <p>Sets the capacity occupied by the route in links.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param routeId Route identifier
	 * @param y_p Occupied capacity
	 * @since 0.3.0
	 */
	public void setRouteOccupiedCapacity(long routeId, double y_p)
	{
		long layerId = getLayerDefaultId();
		setRouteOccupiedCapacity(layerId, routeId, y_p);
	}

	/**
	 * <p>Sets the capacity occupied by the route in links.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @param y_p Occupied capacity
	 * @since 0.3.0
	 */
	public void setRouteOccupiedCapacity(long layerId, long routeId, double y_p)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		if (y_p < 0 && y_p != -1) throw new Net2PlanException("Occupied capacity must be greater or equal than zero, or -1");
		
		boolean routeUp = !isRouteDown(layerId, routeId);
		if (routeUp)
		{
			double previous_y_p = getRouteOccupiedCapacity(layerId, routeId);
			double new_y_p = y_p == -1 ? getRouteCarriedTraffic(layerId, routeId) : y_p;
			if (new_y_p > previous_y_p)
			{
				double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
				List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
				for(long linkId : seqLinks)
				{
					if (linkId < 0)
					{
						long segmentId = -1 - linkId;
						double u_s = getProtectionSegmentReservedCapacity(layerId, segmentId);
						double y_s = getProtectionSegmentCarriedTraffic(layerId, segmentId);
						if (y_s + new_y_p - previous_y_p > u_s + PRECISION_FACTOR)
							throw new Net2PlanException("New occupied capacity could not be applied. Protection segment " + segmentId + " would become over-subscribed");
					}
				}
			}
			
			updateCacheRouteRemoveOccupiedCapacity(layerId, routeId);
		}
		
		if (y_p == -1)
		{
			routeSameOccupiedCapacityAsCarriedTraffic.get(layerId).add(routeId);
			y_p = getRouteCarriedTraffic(layerId, routeId);
		}
		else
		{
			routeSameOccupiedCapacityAsCarriedTraffic.get(layerId).remove(routeId);
		}

		routeOccupiedCapacityMap.get(layerId).put(routeId, y_p);
		if (routeUp) updateCacheRouteAddOccupiedCapacity(layerId, routeId);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
	}

	/**
	 * <p>Sets the occupied capacity per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param y_p Occupied capacity per route
	 * @since 0.3.0
	 */
	public void setRouteOccupiedCapacityMap(Map<Long, Double> y_p)
	{
		long layerId = getLayerDefaultId();
		setRouteOccupiedCapacityMap(layerId, y_p);
	}
	
	/**
	 * <p>Sets the occupied capacity per route.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param y_p Occupied capacity per route
	 * @since 0.3.0
	 */
	public void setRouteOccupiedCapacityMap(long layerId, Map<Long, Double> y_p)
	{
		checkIsModifiable();
		
		for(Entry<Long, Double> entry : y_p.entrySet())
		{
			long routeId = entry.getKey();
			double occupiedCapacity = entry.getValue();
			setRouteOccupiedCapacity(layerId, routeId, occupiedCapacity);
		}
	}
	
	/**
	 * <p>Sets the occupied capacity per route, in increasing order of route identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param y_p Occupied capacity per route
	 * @since 0.3.0
	 */
	public void setRouteOccupiedCapacityVector(double[] y_p)
	{
		long layerId = getLayerDefaultId();
		setRouteOccupiedCapacityVector(layerId, y_p);
	}

	/**
	 * <p>Sets the occupied capacity per route, in increasing order of route identifier.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 * 
	 * @param layerId Layer identifier
	 * @param y_p Occupied capacity per route
	 * @since 0.3.0
	 */
	public void setRouteOccupiedCapacityVector(long layerId, double[] y_p)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);
		
		Set<Long> routeIds_thisLayer = getRouteIds(layerId);
		Map<Long, Integer> routeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(routeIds_thisLayer);
		for(long routeId : routeIds_thisLayer)
		{
			int p = routeId2LinearIndex.get(routeId);
			setRouteOccupiedCapacity(layerId, routeId, y_p[p]);
		}
	}

	/**
	 * <p>Sets the current path of a route. Original route can be always accessed 
	 * from the {@link #getRouteOriginalSequenceOfLinks(long) getRouteOriginalSequenceOfLinks()} method.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 *
	 * @param routeId Route identifier
	 * @param path Sequence of links (and maybe protection segments)
	 * @since 0.3.0
	 */
	public void setRouteSequenceOfLinks(long routeId, List<Long> path)
	{
		long layerId = getLayerDefaultId();
		setRouteSequenceOfLinks(layerId, routeId, path);
	}
	
	/**
	 * <p>Sets the current path of a route. Original route can be always accessed 
	 * from the {@link #getRouteOriginalSequenceOfLinks(long, long) getRouteOriginalSequenceOfLinks()} method.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @param path Sequence of links (and maybe protection segments)
	 * @since 0.3.0
	 */
	public void setRouteSequenceOfLinks(long layerId, long routeId, List<Long> path)
	{
		checkIsModifiable();
		checkRouteIsActive(layerId, routeId);
		
		long demandId = getRouteDemand(layerId, routeId);
		checkPathValidityForDemand(layerId, path, demandId);
		
		/* Check whether a protection segment will become over-subscribed */
		boolean willBeRouteUp = isPathRestorable(layerId, path);
		if (willBeRouteUp)
		{
			Set<Long> linksToCheck = new LinkedHashSet<Long>(path);
			if (routeSequenceOfLinks.get(layerId).containsKey(routeId) && !isRouteDown(layerId, routeId))
			{
				List<Long> currentSeqLinks = getRouteSequenceOfLinks(layerId, routeId);
				for(long linkId : currentSeqLinks) linksToCheck.remove(linkId);
			}
			
			double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
			for(long linkId : linksToCheck)
			{
				if (linkId >= 0) continue;
				
				long segmentId = -1 - linkId;
				if (!getRouteBackupSegmentList(layerId, routeId).contains(segmentId))
					throw new Net2PlanException("Protection segment " + segmentId + " is not associated to route " + routeId);
				
				double u_s = getProtectionSegmentReservedCapacity(layerId, segmentId);
				double y_s = getProtectionSegmentCarriedTraffic(layerId, segmentId);
				double y_p = getRouteOccupiedCapacity(layerId, routeId);
				if (y_s + y_p > u_s + PRECISION_FACTOR)
					throw new Net2PlanException("Route cannot be rerouted. Protection segment " + segmentId + " would become over-subscribed");
			}
		}
		
		boolean isRouteUp = !isRouteDown(layerId, routeId);
		
		if (routeSequenceOfLinks.get(layerId).containsKey(routeId))
		{
			List<Long> currentSeqLinks = getRouteSequenceOfLinks(layerId, routeId);
			for(long linkId : currentSeqLinks)
				if (linkId >= 0)
					linkRouteIds.get(layerId).get(linkId).remove(routeId);

			if (isRouteUp) updateCacheRouteRemoveOccupiedCapacity(layerId, routeId);
		}
		
		routeSequenceOfLinks.get(layerId).put(routeId, Collections.unmodifiableList(new LinkedList<Long>(path)));
		for(long linkId : path)
			if (linkId >= 0)
				linkRouteIds.get(layerId).get(linkId).add(routeId);
		
		if (willBeRouteUp && isRouteUp) updateCacheRouteAddOccupiedCapacity(layerId, routeId);
		else setRouteDown(layerId, routeId);
		
		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
	}
	
	/**
	 * <p>Sets the given route as up. If it is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 *
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void setRouteUp(long routeId)
	{
		long layerId = getLayerDefaultId();
		setRouteUp(layerId, routeId);
	}
	
	/**
	 * <p>Sets the given route as up. If it is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} action will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param routeId Route identifier
	 * @since 0.3.0
	 */
	public void setRouteUp(long layerId, long routeId)
	{
		Set<Long> aux = new LinkedHashSet<Long>();
		aux.add(routeId);
		
		setRoutesUp(layerId, aux);
	}
	
	/**
	 * <p>Sets the given routes as up. If some of them is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 *
	 * @param routeIds Route identifiers
	 * @since 0.3.0
	 */
	public void setRoutesUp(Set<Long> routeIds)
	{
		long layerId = getLayerDefaultId();
		setRoutesUp(layerId, routeIds);
	}
	
	/**
	 * <p>Sets the given routes as up. If some of them is not restorable, an exception 
	 * will be thrown.</p>
	 * 
	 * <p><b>Important</b>: Routing type must be {@link com.net2plan.utils.Constants.RoutingType#SOURCE_ROUTING SOURCE_ROUTING}.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#ROUTE_MODIFIED ROUTE_MODIFIED} actions will be issued.</p>
	 *
	 * @param layerId Layer identifier
	 * @param routeIds Route identifiers
	 * @since 0.3.0
	 */
	public void setRoutesUp(long layerId, Set<Long> routeIds)
	{
		checkIsModifiable();
		
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		Set<Long> nodesDown_thisNetPlan = getNodesDown();
		Set<Long> linksDown_thisLayer = getLinksDown(layerId);
		Set<Long> segmentsDown_thisLayer = getProtectionSegmentsDown(layerId);
		for(long routeId : new LinkedHashSet<Long>(routeIds))
		{
			List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
			if (!isRouteRestorable(layerId, routeId))
			{
				double y_p = getRouteOccupiedCapacity(layerId, routeId);
				for(long linkId : seqLinks)
				{
					if (linkId >= 0)
					{
						if (linksDown_thisLayer.contains(linkId)) throw new Net2PlanException("Bad - Some link on route " + routeId + " at layer " + layerId + " is down");
					}
					else
					{
						long segmentId = -1 - linkId;
						if (segmentsDown_thisLayer.contains(segmentId)) throw new Net2PlanException("Bad - Some protection segment on route " + routeId + " at layer " + layerId + " is down");
						
						double u_s = getProtectionSegmentReservedCapacity(layerId, segmentId);
						double y_s = getProtectionSegmentCarriedTraffic(layerId, segmentId);
						if (y_s + y_p > u_s + PRECISION_FACTOR)
							throw new Net2PlanException("Route " + routeId + " cannot be put up. Protection segment " + segmentId + " would become over-subscribed");
					}
				}
				
				List<Long> seqNodes = getRouteSequenceOfNodes(layerId, routeId);
				if (CollectionUtils.containsAny(seqNodes, nodesDown_thisNetPlan)) throw new Net2PlanException("Bad - Some node on route " + routeId + " at layer " + layerId + " is down");
			}

			routesDown.get(layerId).remove(routeId);
			
			updateCacheRouteAddCarriedTraffic(layerId, routeId);
			updateCacheRouteAddOccupiedCapacity(layerId, routeId);

			if (isActionMonitoringActivated()) actionsPerformed.add(Factory.routeModified(layerId, routeId));
		}
	}
	
	/**
	 * Sets the routing type. If there is some previous routing information, it 
	 * will be converted to the new type.
	 * 
	 * @param routingType {@link com.net2plan.utils.Constants.RoutingType RoutingType}
	 * @since 0.3.0
	 */
	public void setRoutingType(RoutingType routingType)
	{
		long layerId = getLayerDefaultId();
		setRoutingType(layerId, routingType);
	}
	
	/**
	 * Sets the routing type. If there is some previous routing information, it 
	 * will be converted to the new type.
	 * 
	 * @param layerId Layer identifier
	 * @param routingType {@link com.net2plan.utils.Constants.RoutingType RoutingType}
	 * @since 0.3.0
	 */
	public void setRoutingType(long layerId, RoutingType routingType)
	{
		if (getRoutingType(layerId) == routingType) return;
		
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = getLinkMap(layerId);
		Set<Long> demandIds_thisLayer = getDemandIds(layerId);
		Map<Long, Pair<Long, Long>> demandMap_thisLayer = getDemandMap(layerId);
		switch(routingType)
		{
			case HOP_BY_HOP_ROUTING:
			{
				removeAllProtectionSegments(layerId);
				Map<Long, Long> d_p = getRouteDemandMap(layerId);
				Map<Long, Double> x_p = getRouteCarriedTrafficMap(layerId);
				Map<Long, List<Long>> seqLinks_p = getRouteAllSequenceOfLinks(layerId);
				Map<Pair<Long, Long>, Double> x_de = GraphUtils.convert_xp2xde(linkMap_thisLayer, d_p, x_p, seqLinks_p);
				removeAllRoutes(layerId);
				layerRoutingType.put(layerId, routingType);
				
				Map<Pair<Long, Long>, Double> f_de = GraphUtils.convert_xde2fde(nodeIds_thisNetPlan, linkMap_thisLayer, demandIds_thisLayer, x_de);
				addForwardingRules(layerId, f_de, false);

				break;
			}	
			
			case SOURCE_ROUTING:
			{	
				/* Get current 'f_de' */
				Map<Pair<Long, Long>, Double> f_de_thisLayer = getForwardingRuleSplittingRatioMap(layerId);

				/* Convert 'f_de' to 'x_de' */
				Map<Pair<Long, Long>, Double> x_de = GraphUtils.convert_fde2xde(nodeIds_thisNetPlan, linkMap_thisLayer, demandMap_thisLayer, getDemandOfferedTrafficMap(layerId), f_de_thisLayer).getFirst();
				
				/* Convert 'x_de' to 'x_p' */
				List<Long> d_p = new LinkedList<Long>();
				List<Double> x_p = new LinkedList<Double>();
				List<List<Long>> seqLinks_p = new LinkedList<List<Long>>();
				GraphUtils.convert_xde2xp(linkMap_thisLayer, demandMap_thisLayer, x_de, d_p, x_p, seqLinks_p);
				
				/* Remove 'hop-by-hop' routing information */
				removeAllForwardingRules(layerId);
				layerRoutingType.put(layerId, routingType);
				
				/* Update routing according to 'x_p' */
				Iterator<Long> demands_it = d_p.iterator();
				Iterator<Double> x_p_it = x_p.iterator();
				Iterator<List<Long>> seqLinks_it = seqLinks_p.iterator();
				while(x_p_it.hasNext())
				{
					long demandId_thisPath = demands_it.next();
					List<Long> seqLinks_thisPath = seqLinks_it.next();
					double x_p_thisPath = x_p_it.next();
					addRoute(layerId, demandId_thisPath, x_p_thisPath, seqLinks_thisPath, null);
				}

				break;
			}
			
			default:
				throw new RuntimeException("Bad - Unknown routing type " + routingType);
		}
	}
	
	/**
	 * <p>Sets an attribute for a given SRG. If it already exists, it will be overriden.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param srgId SRG identifier
	 * @param key Attribute name
	 * @param value Attribute value
	 * @since 0.3.0
	 */
	public void setSRGAttribute(long srgId, String key, String value)
	{
		checkIsModifiable();
		checkSRGIsActive(srgId);
		srgAttributes.get(srgId).put(key, value);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.srgModified(srgId));
	}
	
	/**
	 * <p>Sets the attributes for a given SRG. Any previous attribute will be removed.</p>
	 *
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param srgId SRG identifier
	 * @param attributes Map of attributes
	 * @since 0.3.0
	 */
	public void setSRGAttributeMap(long srgId, Map<String, String> attributes)
	{
		checkIsModifiable();
		checkSRGIsActive(srgId);

		Map<String, String> auxSRGAttributes = srgAttributes.get(srgId);
		auxSRGAttributes.clear();
		auxSRGAttributes.putAll(attributes);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.srgModified(srgId));
	}
	
	/**
	 * Sets nodes and links associated to the given SRG as down (in case they are not yet).
	 * 
	 * @param srgId SRG identifier
	 * @since 0.3.0
	 */
	public void setSRGDown(long srgId)
	{
		checkSRGIsActive(srgId);
		setNodesAndLinksDown(srgNodes.get(srgId), srgLinks.get(srgId));
	}
	
	/**
	 * <p>Sets the Mean Time To Fail (in hours) of a given SRG.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param srgId SRG identifier
	 * @param mttfInHours Mean Time To Fail (in hours). Zero or negative value means {@code Double.MAX_VALUE}
	 * @since 0.3.0
	 */
	public void setSRGMeanTimeToFailInHours(long srgId, double mttfInHours)
	{
		checkIsModifiable();
		checkSRGIsActive(srgId);
		
		if (mttfInHours <= 0) mttfInHours = Double.MAX_VALUE;
		
		srgMeanTimeToFailInHours.put(srgId, mttfInHours);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.srgModified(srgId));
	}
	
	/**
	 * <p>Sets the Mean Time To Repair (in hours) of a given SRG.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, a {@link com.net2plan.interfaces.simulation.SimAction.ActionType#SRG_MODIFIED SRG_MODIFIED} action will be issued.</p>
	 *
	 * @param srgId SRG identifier
	 * @param mttrInHours Mean Time To Repair (in hours). Zero or negative value are not allowed
	 * @since 0.3.0
	 */
	public void setSRGMeanTimeToRepairInHours(long srgId, double mttrInHours)
	{
		checkIsModifiable();
		checkSRGIsActive(srgId);

		if (mttrInHours <= 0) throw new Net2PlanException("Mean time to repair must be greater than zero");
		
		srgMeanTimeToRepairInHours.put(srgId, mttrInHours);

		if (!addingItem && isActionMonitoringActivated()) actionsPerformed.add(Factory.srgModified(srgId));
	}
	
	/**
	 * <p>Sets the traffic demands at the given layer from a given traffic matrix, removing any previous
	 * demand.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_ADDED DEMAND_ADDED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: If there are less nodes than those in the traffic matrix
	 * as nodes as needed will be added in position (0,0).</p>
	 * 
	 * @param trafficMatrix Traffic matrix
	 * @since 0.3.1
	 */
	public void setTrafficMatrix(DoubleMatrix2D trafficMatrix)
	{
		long layerId = getLayerDefaultId();
		setTrafficMatrix(layerId, trafficMatrix);
	}

	/**
	 * <p>Sets the traffic demands at the given layer from a given traffic matrix, removing any previous
	 * demand.</p>
	 * 
	 * <p><b>Important</b>: In case that action monitoring is activated, several {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_REMOVED DEMAND_REMOVED} and {@link com.net2plan.interfaces.simulation.SimAction.ActionType#DEMAND_ADDED DEMAND_ADDED} actions will be issued.</p>
	 * 
	 * <p><b>Important</b>: If there are less nodes than those in the traffic matrix
	 * as nodes as needed will be added in position (0,0).</p>
	 * 
	 * @param layerId Layer identifier
	 * @param trafficMatrix Traffic matrix
	 * @since 0.3.1
	 */
	public void setTrafficMatrix(long layerId, DoubleMatrix2D trafficMatrix)
	{
		checkIsModifiable();
		checkLayerIsActive(layerId);

		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		removeAllDemands(layerId);

		int N = getNumberOfNodes();
		int N_aux = trafficMatrix.rows();

		if (N < N_aux)
			for (int nodeIndex = N; nodeIndex < N_aux; nodeIndex++)
				addNode(0, 0, null, null);

		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		Map<Long, Integer> nodeId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(nodeIds_thisNetPlan);
		for (long ingressNodeId : nodeIds_thisNetPlan)
		{
			int a_d = nodeId2LinearIndex.get(ingressNodeId);
			
			for (long egressNodeId : nodeIds_thisNetPlan)
			{
				int b_d = nodeId2LinearIndex.get(egressNodeId);
				if (a_d == b_d || a_d >= N_aux || b_d >= N_aux) continue;
				if (trafficMatrix.getQuick(a_d, b_d) < PRECISION_FACTOR) continue;
				addDemand(layerId, ingressNodeId, egressNodeId, trafficMatrix.getQuick(a_d, b_d), null);
			}
		}
	}
	
	/**
	 * Returns a {@code String} representation of the network design.
	 *
	 * @return String representation of the network design
	 * @since 0.2.0
	 */
	@Override
	public String toString()
	{
		return toString(getLayerIds());
	}
	
	/**
	 * Returns a {@code String} representation of the network design only for the 
	 * given layers.
	 * 
	 * @param layerIds Layer identifiers
	 * @return String representation of the network design
	 * @since 0.3.0
	 */
	public String toString(Set<Long> layerIds)
	{
		StringBuilder netPlanInformation = new StringBuilder();
		String NEWLINE = StringUtils.getLineSeparator();
		
		int N = getNumberOfNodes();
		if (N == 0)
		{
			netPlanInformation.append("Empty network");
			return netPlanInformation.toString();
		}
		
		int L = getNumberOfLayers();
		int numSRGs = getNumberOfSRGs();
		netPlanInformation.append("Network information");
		netPlanInformation.append(NEWLINE);
		netPlanInformation.append("--------------------------------");
		netPlanInformation.append(NEWLINE).append(NEWLINE);
		netPlanInformation.append("Name: ").append(getNetworkName()).append(NEWLINE);
		netPlanInformation.append("Description: ").append(getNetworkDescription()).append(NEWLINE);
		Map<String, String> networkAttributes_thisNetPlan = getNetworkAttributeMap();
		netPlanInformation.append("Attributes: ");
		netPlanInformation.append(networkAttributes_thisNetPlan.isEmpty() ? "none" : StringUtils.mapToString(networkAttributes_thisNetPlan, "=", ", "));
		netPlanInformation.append(NEWLINE);
		netPlanInformation.append("Number of layers: ").append(L).append(" (next id = ").append(nextLayerId).append(")").append(NEWLINE);
		netPlanInformation.append("Default layer: ").append(layerDefaultId).append(NEWLINE);
		netPlanInformation.append("Number of nodes: ").append(N).append(" (next id = ").append(nextNodeId).append(")").append(NEWLINE);
		netPlanInformation.append("Number of SRGs: ").append(numSRGs).append(" (next id = ").append(nextSRGId).append(")").append(NEWLINE);
		
		netPlanInformation.append(NEWLINE);
		netPlanInformation.append("Node information");
		netPlanInformation.append(NEWLINE);
		netPlanInformation.append("--------------------------------");
		netPlanInformation.append(NEWLINE).append(NEWLINE);
		
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		for (long nodeId : nodeIds_thisNetPlan)
		{
			Point2D position = getNodeXYPosition(nodeId);
			String attributes = StringUtils.mapToString(getNodeAttributeMap(nodeId), "=", ", ");

			String nodeInformation = String.format("n%d, state: %s, position: (%.3g, %.3g), name: %s, attributes: %s", nodeId, isNodeDown(nodeId) ? "down" : "up", position.getX(), position.getY(), getNodeName(nodeId), attributes.isEmpty() ? "none" : attributes);
			netPlanInformation.append(nodeInformation);
			netPlanInformation.append(NEWLINE);
		}

		for (long layerId : new TreeSet<Long>(layerIds))
		{
			RoutingType routingType_thisLayer = getRoutingType(layerId);
			Map<String, String> layerAttributes_thisLayer = getLayerAttributeMap(layerId);

			netPlanInformation.append(NEWLINE);
			netPlanInformation.append("Layer ").append(layerId).append(" information");
			netPlanInformation.append(NEWLINE);
			netPlanInformation.append("--------------------------------");
			netPlanInformation.append(NEWLINE).append(NEWLINE);
			netPlanInformation.append("Name: ").append(getLayerName(layerId)).append(NEWLINE);
			netPlanInformation.append("Description: ").append(getLayerDescription(layerId)).append(NEWLINE);
			netPlanInformation.append("Link capacity units: ").append(getLinkCapacityUnitsName(layerId)).append(NEWLINE);
			netPlanInformation.append("Demand traffic units: ").append(getDemandTrafficUnitsName(layerId)).append(NEWLINE);
			netPlanInformation.append("Routing type: ").append(routingType_thisLayer).append(NEWLINE);

			if (!layerAttributeMap.get(layerId).isEmpty())
			{
				netPlanInformation.append("Attributes: ");
				netPlanInformation.append(StringUtils.mapToString(layerAttributes_thisLayer, "=", ", "));
				netPlanInformation.append(NEWLINE);
			}

			netPlanInformation.append("Number of links: ").append(getNumberOfLinks(layerId)).append(" (next id = ").append(nextLinkId.get(layerId)).append(")").append(NEWLINE);
			netPlanInformation.append("Number of demands: ").append(getNumberOfDemands(layerId)).append(" (next id = ").append(nextDemandId.get(layerId)).append(")").append(NEWLINE);
			if (routingType_thisLayer == RoutingType.SOURCE_ROUTING)
			{
				netPlanInformation.append("Number of routes: ").append(getNumberOfRoutes(layerId)).append(" (next id = ").append(nextRouteId.get(layerId)).append(")").append(NEWLINE);
				netPlanInformation.append("Number of protection segments: ").append(getNumberOfProtectionSegments(layerId)).append(" (next id = ").append(nextSegmentId.get(layerId)).append(")").append(NEWLINE);
			}
			else
			{
				netPlanInformation.append("Number of forwarding rules: ").append(getNumberOfForwardingRules(layerId)).append(NEWLINE);
			}

			if (hasLinks(layerId))
			{
				netPlanInformation.append(NEWLINE);
				netPlanInformation.append("Link information");
				netPlanInformation.append(NEWLINE);
				netPlanInformation.append("--------------------------------");
				netPlanInformation.append(NEWLINE).append(NEWLINE);

				Set<Long> linkIds_thisLayer = getLinkIds(layerId);
				for (long linkId : linkIds_thisLayer)
				{
					long originNodeId = getLinkOriginNode(layerId, linkId);
					long destinationNodeId = getLinkDestinationNode(layerId, linkId);

					String attributes = StringUtils.mapToString(getLinkAttributeMap(layerId, linkId), "=", ", ");

					String linkInformation = String.format("e%d, n%d (%s) -> n%d (%s), state: %s, capacity: %.3g, length: %.3g km, propagation speed: %.3g km/s, attributes: %s", linkId, originNodeId, getNodeName(originNodeId), destinationNodeId, getNodeName(destinationNodeId), isLinkDown(layerId, linkId) ? "down" : "up", getLinkCapacity(layerId, linkId), getLinkLengthInKm(layerId, linkId), getLinkPropagationSpeedInKmPerSecond(layerId, linkId), attributes.isEmpty() ? "none" : attributes);
					netPlanInformation.append(linkInformation);
					
					if (isLinkCoupled(layerId, linkId))
					{
						Pair<Long, Long> lowerLayerDemand = getLinkCoupledLowerLayerDemand(layerId, linkId);
						long lowerLayerId = lowerLayerDemand.getFirst();
						long demandId = lowerLayerDemand.getSecond();
						netPlanInformation.append(String.format(", associated to demand %d at layer %d", demandId, lowerLayerId));
					}
					
					netPlanInformation.append(NEWLINE);
				}
			}

			if (hasDemands(layerId))
			{
				netPlanInformation.append(NEWLINE);
				netPlanInformation.append("Demand information");
				netPlanInformation.append(NEWLINE);
				netPlanInformation.append("--------------------------------");
				netPlanInformation.append(NEWLINE).append(NEWLINE);

				Set<Long> demandIds_thisLayer = getDemandIds(layerId);
				for (long demandId : demandIds_thisLayer)
				{
					long ingressNodeId = getDemandIngressNode(layerId, demandId);
					long egressNodeId = getDemandEgressNode(layerId, demandId);

					String attributes = StringUtils.mapToString(getDemandAttributeMap(layerId, demandId), "=", ", ");

					String demandInformation = String.format("d%d, n%d (%s) -> n%d (%s), offered traffic: %.3g, attributes: %s", demandId, ingressNodeId, getNodeName(ingressNodeId), egressNodeId, getNodeName(egressNodeId), getDemandOfferedTraffic(layerId, demandId), attributes.isEmpty() ? "none" : attributes);
					netPlanInformation.append(demandInformation);

					if (isDemandCoupled(layerId, demandId))
					{
						Pair<Long, Long> upperLayerLink = lowerLayerToUpperLayerMapping.get(layerId, demandId);
						long upperLayerId = upperLayerLink.getFirst();
						long linkId = upperLayerLink.getSecond();
						netPlanInformation.append(String.format(", associated to link %d at layer %d", linkId, upperLayerId));
					}
					
					netPlanInformation.append(NEWLINE);
				}
			}

			if (routingType_thisLayer == RoutingType.SOURCE_ROUTING)
			{
				if (hasRoutes(layerId))
				{
					netPlanInformation.append(NEWLINE);
					netPlanInformation.append("Routing information");
					netPlanInformation.append(NEWLINE);
					netPlanInformation.append("--------------------------------");
					netPlanInformation.append(NEWLINE).append(NEWLINE);

					Set<Long> routeIds_thisLayer = getRouteIds(layerId);
					for (long routeId : routeIds_thisLayer)
					{
						List<Long> sequenceOfLinks = getRouteSequenceOfLinks(layerId, routeId);
						List<Long> sequenceOfNodes = getRouteSequenceOfNodes(layerId, routeId);

						String str_seqNodes = CollectionUtils.join(sequenceOfNodes, " => ");
						StringBuilder path = new StringBuilder();

						Iterator<Long> it = sequenceOfLinks.iterator();
						while(it.hasNext())
						{
							long linkId = it.next();
							if (linkId >= 0) path.append(String.format("e%d", linkId));
							else path.append(String.format("s%d", (-1 - linkId)));

							if (it.hasNext()) path.append(" => ");
						}

						String str_seqLinks = path.toString();

						List<Long> backupSegmentList_thisRoute = getRouteBackupSegmentList(layerId, routeId);
						String backupSegments = backupSegmentList_thisRoute.isEmpty() ? "none" : CollectionUtils.join(backupSegmentList_thisRoute, ", ");
						String attributes = StringUtils.mapToString(getRouteAttributeMap(layerId, routeId), "=", ", ");

						String routeInformation = String.format("r%d, demand: d%d, carried traffic: %.3g, occupied capacity: %.3g, seq. links: %s, seq. nodes: %s, backup segments: %s, attributes: %s", routeId, getRouteDemand(layerId, routeId), getRouteCarriedTraffic(layerId, routeId), getRouteOccupiedCapacity(layerId, routeId), str_seqLinks, str_seqNodes, backupSegments, attributes.isEmpty() ? "none" : attributes);
						netPlanInformation.append(routeInformation);
						netPlanInformation.append(NEWLINE);
					}
				}
				
				if (hasProtectionSegments(layerId))
				{
					netPlanInformation.append(NEWLINE);
					netPlanInformation.append("Protection segment information");
					netPlanInformation.append(NEWLINE);
					netPlanInformation.append("--------------------------------");
					netPlanInformation.append(NEWLINE).append(NEWLINE);

					Set<Long> segmentIds_thisLayer = getProtectionSegmentIds(layerId);
					for (long segmentId : segmentIds_thisLayer)
					{
						List<Long> sequenceOfLinks = getProtectionSegmentSequenceOfLinks(layerId, segmentId);
						List<Long> sequenceOfNodes = getProtectionSegmentSequenceOfNodes(layerId, segmentId);
						String str_seqLinks = CollectionUtils.join(sequenceOfLinks, " => ");
						String str_seqNodes = CollectionUtils.join(sequenceOfNodes, " => ");
						String attributes = StringUtils.mapToString(getProtectionSegmentAttributeMap(layerId, segmentId), "=", ", ");

						String segmentInformation = String.format("s%d, reserved bandwidth: %.3g, seq. links: %s, seq. nodes: %s, attributes: %s", segmentId, getProtectionSegmentReservedCapacity(layerId, segmentId), str_seqLinks, str_seqNodes, attributes.isEmpty() ? "none" : attributes);
						netPlanInformation.append(segmentInformation);
						netPlanInformation.append(NEWLINE);
					}
				}
				
			}
			else
			{
				if (hasForwardingRules(layerId))
				{
					netPlanInformation.append(NEWLINE);
					netPlanInformation.append("Forwarding information");
					netPlanInformation.append(NEWLINE);
					netPlanInformation.append("--------------------------------");
					netPlanInformation.append(NEWLINE).append(NEWLINE);

					Set<Long> demandIds_thisLayer = getDemandIds(layerId);
					for(long nodeId : nodeIds_thisNetPlan)
					{
						Set<Long> outgoingLinks_thisNode = getNodeOutgoingLinks(layerId, nodeId);
						netPlanInformation.append(String.format("Forwarding table for n%d%n", nodeId));

						boolean thereAreRoutes = false;
						if (!outgoingLinks_thisNode.isEmpty())
						{
							for (long demandId : demandIds_thisLayer)
							{
								boolean thereAreRoutes_thisDemand = false;

								StringBuilder out_thisDemand = new StringBuilder();
								out_thisDemand.append(String.format("dst: d%d, gw:", demandId));
								for (long linkId : outgoingLinks_thisNode)
								{
									Pair<Long, Long> demandLinkPair = Pair.of(demandId, linkId);
									if (!isForwardingRuleActive(layerId, demandLinkPair)) continue;
									
									double splittingRatio = getForwardingRuleSplittingRatio(layerId, demandLinkPair);
									if (splittingRatio <= 0) continue;

									out_thisDemand.append(String.format(" [e%d (n%d): %f]", linkId, getLinkDestinationNode(layerId, linkId), splittingRatio));
									thereAreRoutes_thisDemand = true;
									thereAreRoutes = true;
								}

								out_thisDemand.append(String.format("%n"));
								if (thereAreRoutes_thisDemand) netPlanInformation.append(out_thisDemand);
							}
						}

						if (!thereAreRoutes) netPlanInformation.append("Empty");

						netPlanInformation.append(NEWLINE);
					}
				}
			}
		}
		
		if (numSRGs > 0)
		{
			netPlanInformation.append(NEWLINE);
			netPlanInformation.append("Shared-risk group information");
			netPlanInformation.append(NEWLINE);
			netPlanInformation.append("--------------------------------");
			netPlanInformation.append(NEWLINE).append(NEWLINE);

			Set<Long> srgIds_thisNetPlan = getSRGIds();
			for (long srgId : srgIds_thisNetPlan)
			{
				Set<Long> nodeIds_thisSRG = getSRGNodes(srgId);
				Map<Long, Set<Long>> linkIds_thisSRG = getSRGLinksAllLayers(srgId);

				String aux_nodes = nodeIds_thisSRG.isEmpty() ? "none" : CollectionUtils.join(nodeIds_thisSRG, " ");
				StringBuilder aux_links = new StringBuilder();
				for(Entry<Long, Set<Long>> entry : linkIds_thisSRG.entrySet())
				{
					Set<Long> linkIds_thisSRG_thisLayer = entry.getValue();
					if (linkIds_thisSRG_thisLayer.isEmpty()) continue;
					
					aux_links.append(String.format("[layer %d: %s]", entry.getKey(), CollectionUtils.join(linkIds_thisSRG_thisLayer, ", ")));
				}
				
				if (aux_links.length() == 0) aux_links.append("none");
				
				String aux_mttf = StringUtils.secondsToYearsDaysHoursMinutesSeconds(getSRGMeanTimeToFailInHours(srgId) * 3600);
				String aux_mttr = StringUtils.secondsToYearsDaysHoursMinutesSeconds(getSRGMeanTimeToRepairInHours(srgId) * 3600);

				String attributes = StringUtils.mapToString(getSRGAttributeMap(srgId), "=", ", ");

				String srgInformation = String.format("srg%d, nodes: %s, links: %s, mttf: %s, mttr: %s, attributes: %s", srgId, aux_nodes, aux_links.toString(), aux_mttf, aux_mttr, attributes.isEmpty() ? "none" : attributes);
				netPlanInformation.append(srgInformation);
				netPlanInformation.append(NEWLINE);
			}
		}

		return netPlanInformation.toString();
	}

	/**
	 * Returns an unmodifiable view of the network.
	 *
	 * @return An unmodifiable view of the network
	 * @since 0.2.3
	 */
	public NetPlan unmodifiableView()
	{
		NetPlan netPlan = new NetPlan();
		netPlan.assignFrom(this);
		netPlan.isModifiable = false;

		return netPlan;
	}
	
	private void updateCacheDemandBlockedTraffic(long layerId, long demandId)
	{
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		double h_d = getDemandOfferedTraffic(layerId, demandId);
		double r_d = getDemandCarriedTraffic(layerId, demandId);
		double blockedTraffic_d = Math.max(0, h_d - r_d);
		if (blockedTraffic_d > PRECISION_FACTOR) 		
		{
			cache_demandBlockedTrafficMap.get(layerId).put(demandId, blockedTraffic_d);
			cache_demandBlockedIds.get(layerId).add(demandId);
		}
		else
		{
			cache_demandBlockedTrafficMap.get(layerId).put(demandId, 0.0);
			cache_demandBlockedIds.get(layerId).remove(demandId);
		}
	}

	private void updateCacheDemandForwarding(long layerId, long demandId)
	{
		Set<Long> nodeIds_thisNetPlan = getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap_thisLayer = getLinkMap(layerId);
		Map<Long, Pair<Long, Long>> demandMap_thisLayer = getDemandMap(layerId);
		Map<Long, Double> h_d_thisLayer = getDemandOfferedTrafficMap(layerId);
		Map<Pair<Long, Long>, Double> f_de = getDemandForwardingRuleSplittingRatioMap(layerId, demandId);
		
		long egressNodeId = getDemandEgressNode(layerId, demandId);
		if (isNodeDown(egressNodeId))
		{
			f_de = new LinkedHashMap<Pair<Long, Long>, Double>(f_de);

			for(Pair<Long, Long> forwardingRule : f_de.keySet())
				f_de.put(forwardingRule, 0.0);
		}
		else
		{
			Set<Pair<Long, Long>> forwardingRulesToRemove = new LinkedHashSet<Pair<Long, Long>>();
			for(Pair<Long, Long> forwardingRule : f_de.keySet())
				if (isForwardingRuleDown(layerId, forwardingRule))
					forwardingRulesToRemove.add(forwardingRule);

			if (!forwardingRulesToRemove.isEmpty())
			{
				f_de = new LinkedHashMap<Pair<Long, Long>, Double>(f_de);
				for(Pair<Long, Long> forwardingRule : forwardingRulesToRemove)
					f_de.put(forwardingRule, 0.0);
			}
		}
		
		try
		{
			Quadruple<Map<Pair<Long, Long>, Double>, Double, Map<Long, Double>, RoutingCycleType> out = GraphUtils.convert_fde2xde(demandId, nodeIds_thisNetPlan, linkMap_thisLayer, demandMap_thisLayer, h_d_thisLayer, f_de);
			Map<Pair<Long, Long>, Double> x_de = out.getFirst();
			cache_demandCarriedTrafficMap.get(layerId).put(demandId, out.getSecond());
			cache_demandRoutingCycleType.get(layerId).put(demandId, out.getFourth());
			
			Map<Pair<Long, Long>, Double> x_de_thisLayer = forwardingRuleCarriedTrafficMap.get(layerId);
			for(Pair<Long, Long> forwardingRule : f_de.keySet())
			{
				Double newValue = x_de.get(forwardingRule);
				if (newValue == null) x_de_thisLayer.put(forwardingRule, 0.0);
				else x_de_thisLayer.put(forwardingRule, newValue);
			}
		}
		catch(GraphUtils.ClosedCycleRoutingException e)
		{
			cache_demandRoutingCycleType.get(layerId).put(demandId, RoutingCycleType.CLOSED_CYCLES);
			cache_demandCarriedTrafficMap.get(layerId).put(demandId, 0.0);
			for(Entry<Pair<Long, Long>, Double> entry : f_de.entrySet()) forwardingRuleCarriedTrafficMap.get(layerId).put(entry.getKey(), 0.0);
		}
		
		if (isDemandCoupled(layerId, demandId))
		{
			Pair<Long, Long> upperLayerLink = lowerLayerToUpperLayerMapping.get(layerId, demandId);

			long upperLayerId = upperLayerLink.getFirst();
			long linkId = upperLayerLink.getSecond();
			linkCapacityMap.get(upperLayerId).put(linkId, cache_demandCarriedTrafficMap.get(layerId).get(demandId));
			updateCacheLinkCapacity(upperLayerId, linkId);
			updateCacheLinkOversubscribed(upperLayerId, linkId);
			if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(upperLayerId, linkId));
		}

		if (!disableCacheUpdating)
		{
			for(Entry<Pair<Long, Long>, Double> entry : f_de.entrySet())
			{
				long linkId = entry.getKey().getSecond();
				updateCacheLinkForwarding(layerId, linkId);
			}
		}
		
		updateCacheDemandBlockedTraffic(layerId, demandId);
	}
	
	private void updateCacheLinkCapacity(long layerId, long linkId)
	{
		double u_e = getLinkCapacity(layerId, linkId);
		cache_linksWithZeroCapacity.get(layerId).remove(linkId);
		cache_linksWithUnlimitedCapacity.get(layerId).remove(linkId);
		
		if (u_e <= 0) cache_linksWithZeroCapacity.get(layerId).add(linkId);
		else if (u_e == Double.MAX_VALUE) cache_linksWithUnlimitedCapacity.get(layerId).add(linkId);
	}

	private void updateCacheLinkForwarding(long layerId, long linkId)
	{
		if (isLinkDown(layerId, linkId))
		{
			cache_linkCarriedTrafficMap.get(layerId).put(linkId, 0.0);
		}
		else
		{
			Set<Pair<Long, Long>> forwardingRules_thisLink = getLinkForwardingRules(layerId, linkId);
			double y_e = 0;
			for(Pair<Long, Long> forwardingRule : forwardingRules_thisLink)
				y_e += getForwardingRuleCarriedTraffic(layerId, forwardingRule);

			cache_linkCarriedTrafficMap.get(layerId).put(linkId, y_e);
		}
		
		updateCacheLinkOversubscribed(layerId, linkId);
	}

	private void updateCacheLinkOversubscribed(long layerId, long linkId)
	{
		double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		if (getLinkUtilization(layerId, linkId) > 1 + PRECISION_FACTOR) cache_linksOverSubscribed.get(layerId).add(linkId);
		else cache_linksOverSubscribed.get(layerId).remove(linkId);
	}

	private void updateCacheRouteAddCarriedTraffic(long layerId, long routeId)
	{
		long demandId = getRouteDemand(layerId, routeId);

		double x_p = getRouteCarriedTraffic(layerId, routeId);
		double r_d = getDemandCarriedTraffic(layerId, demandId) + x_p;
		cache_demandCarriedTrafficMap.get(layerId).put(demandId, r_d);
		
		if (isDemandCoupled(layerId, demandId))
		{
			Pair<Long, Long> upperLayerLink = lowerLayerToUpperLayerMapping.get(layerId, demandId);

			long upperLayerId = upperLayerLink.getFirst();
			long linkId = upperLayerLink.getSecond();
			linkCapacityMap.get(upperLayerId).put(linkId, r_d);
			updateCacheLinkCapacity(upperLayerId, linkId);
			updateCacheLinkOversubscribed(upperLayerId, linkId);
			if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(upperLayerId, linkId));
		}
		
		updateCacheDemandBlockedTraffic(layerId, demandId);
	}
	
	private void updateCacheRouteAddOccupiedCapacity(long layerId, long routeId)
	{
		List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
		double y_p = getRouteOccupiedCapacity(layerId, routeId);
		
		for(long linkId : seqLinks)
		{
			if (linkId >= 0)
			{
				double y_e = getLinkCarriedTraffic(layerId, linkId);
				cache_linkCarriedTrafficMap.get(layerId).put(linkId, y_e + y_p);
				updateCacheLinkOversubscribed(layerId, linkId);
			}
			else
			{
				long segmentId = -1 - linkId;
				double y_s = getProtectionSegmentCarriedTraffic(layerId, segmentId);
				cache_segmentCarriedTrafficMap.get(layerId).put(segmentId, y_s + y_p);
			}
		}
	}
	
	private void updateCacheRouteRemoveCarriedTraffic(long layerId, long routeId)
	{
		if (routeCarriedTrafficMap.get(layerId).containsKey(routeId))
		{
			long demandId = getRouteDemand(layerId, routeId);
			double r_d = Math.max(0, getDemandCarriedTraffic(layerId, demandId) - getRouteCarriedTraffic(layerId, routeId));
			cache_demandCarriedTrafficMap.get(layerId).put(demandId, r_d);
			
			if (isDemandCoupled(layerId, demandId))
			{
				Pair<Long, Long> upperLayerLink = lowerLayerToUpperLayerMapping.get(layerId, demandId);

				long upperLayerId = upperLayerLink.getFirst();
				long linkId = upperLayerLink.getSecond();
				linkCapacityMap.get(upperLayerId).put(linkId, r_d);
				updateCacheLinkCapacity(upperLayerId, linkId);
				updateCacheLinkOversubscribed(upperLayerId, linkId);
				if (isActionMonitoringActivated()) actionsPerformed.add(Factory.linkModified(upperLayerId, linkId));
			}
			
			updateCacheDemandBlockedTraffic(layerId, demandId);
		}
	}

	private void updateCacheRouteRemoveOccupiedCapacity(long layerId, long routeId)
	{
		double current_y_p = getRouteOccupiedCapacity(layerId, routeId);
		List<Long> seqLinks = getRouteSequenceOfLinks(layerId, routeId);
		for(long linkId : seqLinks)
		{
			if (linkId >= 0)
			{
				double y_e = getLinkCarriedTraffic(layerId, linkId);
				cache_linkCarriedTrafficMap.get(layerId).put(linkId, Math.max(0, y_e - current_y_p));
				updateCacheLinkOversubscribed(layerId, linkId);
			}
			else
			{
				long segmentId = -1 - linkId;
				double y_s = getProtectionSegmentCarriedTraffic(layerId, segmentId);
				cache_segmentCarriedTrafficMap.get(layerId).put(segmentId, Math.max(0, y_s - current_y_p));
			}
		}
	}

	private static class DemandLinkMapping
	{
		public final Map<Long, Long> demandLinkMapping;
		
		public DemandLinkMapping()
		{
			demandLinkMapping = new LinkedHashMap<Long, Long>();
		}
		
		public boolean isEmpty()
		{
			return demandLinkMapping.isEmpty();
		}
		
		public void put(long demandId, long linkId)
		{
			demandLinkMapping.put(demandId, linkId);
		}
		
		public void remove(long demandId)
		{
			demandLinkMapping.remove(demandId);
		}
		
		@Override
		public String toString()
		{
			return demandLinkMapping.toString();
		}
	}

	private static interface INetPlanFormat
	{
		public void create(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException;
	}
	
	private static class NetPlanFormat_v1 implements INetPlanFormat
	{
		Map<Long, List<Long>> backupSegmentMap = new LinkedHashMap<Long, List<Long>>();
		
		@Override
		public void create(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			parseNetwork(netPlan, xmlStreamReader);
			
			for(Entry<Long, List<Long>> entry : backupSegmentMap.entrySet())
			{
				long routeId = entry.getKey();
				List<Long> backupSegmentList = entry.getValue();
				for(long segmentId : backupSegmentList)
					netPlan.addProtectionSegmentToRouteBackupSegmentList(segmentId, routeId);
			}
		}

		protected void parseDemand(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long ingressNodeId = -1;
			long egressNodeId = -1;
			double offeredTraffic = 0;
			Map<String, String> attributeMap = new LinkedHashMap<String, String>();

			int numAttributes = xmlStreamReader.getAttributeCount();
			for(int i = 0; i < numAttributes; i++)
			{
				String localName = xmlStreamReader.getAttributeLocalName(i);
				switch(localName)
				{
					case "ingressNodeId":
						ingressNodeId = xmlStreamReader.getAttributeAsLong(i);
						break;
						
					case "egressNodeId":
						egressNodeId = xmlStreamReader.getAttributeAsLong(i);
						break;
						
					case "offeredTrafficInErlangs":
						offeredTraffic = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					default:
						attributeMap.put(localName, xmlStreamReader.getAttributeValue(i));
						break;
				}
			}

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						throw new RuntimeException("No nested elements allowed for 'Demand' element");

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("demandEntry"))
						{
							netPlan.addDemand(ingressNodeId, egressNodeId, offeredTraffic, attributeMap);
							return;
						}
						break;
				}
			}

			throw new RuntimeException("'Demand' element not parsed correctly (end tag not found)");
		}

		protected void parseDemandSet(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "demandEntry":
								parseDemand(netPlan, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("demandSet")) return;
						break;
				}
			}

			throw new RuntimeException("'Demand set' element not parsed correctly (end tag not found)");
		}
		
		protected void parseLink(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			Map<String, String> attributeMap = new LinkedHashMap<String, String>();
			double linkCapacity = 0;
			double linkLengthInKm = 0;
			long originNodeId = -1;
			long destinationNodeId = -1;

			int numAttributes = xmlStreamReader.getAttributeCount();
			for(int i = 0; i < numAttributes; i++)
			{
				String localName = xmlStreamReader.getAttributeLocalName(i);
				switch(localName)
				{
					case "originNodeId":
						originNodeId = xmlStreamReader.getAttributeAsLong(i);
						break;
						
					case "destinationNodeId":
						destinationNodeId = xmlStreamReader.getAttributeAsLong(i);
						break;
						
					case "linkCapacityInErlangs":
						linkCapacity = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					case "linkLengthInKm":
						linkLengthInKm = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					default:
						attributeMap.put(localName, xmlStreamReader.getAttributeValue(i));
						break;
				}
			}

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						throw new RuntimeException("No nested elements allowed for 'Link' element");

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("link"))
						{
							netPlan.addLink(originNodeId, destinationNodeId, linkCapacity, linkLengthInKm, attributeMap);
							return;
						}
						break;
				}
			}

			throw new RuntimeException("'Link' element not parsed correctly (end tag not found)");
		}

		protected void parseNetwork(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			int numAttributes = xmlStreamReader.getAttributeCount();
			for(int i = 0; i < numAttributes; i++)
			{
				String localName = xmlStreamReader.getAttributeLocalName(i);
				switch(localName)
				{
					case "description":
						netPlan.setNetworkDescription(xmlStreamReader.getAttributeValue(i));
						break;
						
					case "name":
						netPlan.setNetworkName(xmlStreamReader.getAttributeValue(i));
						break;
						
					default:
						netPlan.setNetworkAttribute(localName, xmlStreamReader.getAttributeValue(i));
						break;
				}
			}
			
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();
				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "physicalTopology":
								parsePhysicalTopology(netPlan, xmlStreamReader);
								break;

							case "demandSet":
								parseDemandSet(netPlan, xmlStreamReader);
								break;
								
							case "routingInfo":
								parseRoutingInfo(netPlan, xmlStreamReader);
								break;

							case "protectionInfo":
								parseProtectionInfo(netPlan, xmlStreamReader);
								break;
								
							case "srgInfo":
								parseSRGInfo(netPlan, xmlStreamReader);
								
							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("network")) return;
						break;
				}
			}

			throw new RuntimeException("'Network' element not parsed correctly (end tag not found)");
		}

		protected void parseNode(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			double xCoord = 0;
			double yCoord = 0;
			String nodeName = null;
			Map<String, String> attributeMap = new LinkedHashMap<String, String>();
			
			int numAttributes = xmlStreamReader.getAttributeCount();
			for(int i = 0; i < numAttributes; i++)
			{
				String localName = xmlStreamReader.getAttributeLocalName(i);
				switch(localName)
				{
					case "xCoord":
						xCoord = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					case "yCoord":
						yCoord = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					case "name":
						nodeName = xmlStreamReader.getAttributeValue(i);
						break;
						
					default:
						attributeMap.put(localName, xmlStreamReader.getAttributeValue(i));
				}
			}
			
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						throw new RuntimeException("No nested elements allowed for 'Node' element");

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("node"))
						{
							netPlan.addNode(xCoord, yCoord, nodeName, attributeMap);
							return;
						}
						break;
				}
			}

			throw new RuntimeException("'Node' element not parsed correctly (end tag not found)");
		}

		protected void parsePhysicalTopology(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "node":
								parseNode(netPlan, xmlStreamReader);
								break;
								
							case "link":
								parseLink(netPlan, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("physicalTopology")) return;
						break;
				}
			}

			throw new RuntimeException("'Physical topology' element not parsed correctly (end tag not found)");
		}

		protected void parseProtectionInfo(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "protectionSegment":
								parseProtectionSegment(netPlan, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("protectionInfo")) return;
						break;
				}
			}

			throw new RuntimeException("'Protection info' element not parsed correctly (end tag not found)");
		}

		protected void parseProtectionSegment(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			double reservedBandwidthInErlangs = 0;
			Map<String, String> attributeMap = new LinkedHashMap<String, String>();
			
			int numAttributes = xmlStreamReader.getAttributeCount();
			for(int i = 0; i < numAttributes; i++)
			{
				String localName = xmlStreamReader.getAttributeLocalName(i);
				switch(localName)
				{
					case "reservedBandwidthInErlangs":
						reservedBandwidthInErlangs = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					default:
						attributeMap.put(localName, xmlStreamReader.getAttributeValue(i));
				}
			}
			
			List<Long> seqLinks = new LinkedList<Long>();
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "linkEntry":
								seqLinks.add(xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id")));
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("protectionSegment"))
						{
							netPlan.addProtectionSegment(seqLinks, reservedBandwidthInErlangs, attributeMap);
							return;
						}
						break;
				}
			}

			throw new RuntimeException("'Protection segment' element not parsed correctly (end tag not found)");
		}

		protected void parseRoute(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long demandId = -1;
			double carriedTrafficInErlangs = 0;
			Map<String, String> attributeMap = new LinkedHashMap<String, String>();
			
			int numAttributes = xmlStreamReader.getAttributeCount();
			for(int i = 0; i < numAttributes; i++)
			{
				String localName = xmlStreamReader.getAttributeLocalName(i);
				switch(localName)
				{
					case "carriedTrafficInErlangs":
						carriedTrafficInErlangs = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					case "demandId":
						demandId = xmlStreamReader.getAttributeAsLong(i);
						break;
						
					default:
						attributeMap.put(localName, xmlStreamReader.getAttributeValue(i));
				}
			}
			
			List<Long> seqLinks = new LinkedList<Long>();
			List<Long> backupSegmentList = new LinkedList<Long>();
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "linkEntry":
								seqLinks.add(xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id")));
								break;
								
							case "protectionSegmentEntry":
								backupSegmentList.add(xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id")));
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("route"))
						{
							long routeId = netPlan.addRoute(demandId, carriedTrafficInErlangs, seqLinks, attributeMap);
							backupSegmentMap.put(routeId, backupSegmentList);
							
							return;
						}
						break;
				}
			}

			throw new RuntimeException("'Route' element not parsed correctly (end tag not found)");
		}

		protected void parseRoutingInfo(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "route":
								parseRoute(netPlan, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("routingInfo")) return;
						break;
				}
			}

			throw new RuntimeException("'Routing info' element not parsed correctly (end tag not found)");
		}

		protected void parseSRG(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			double mttf = 0;
			double mttr = 0;
			Map<String, String> attributeMap = new LinkedHashMap<String, String>();
			
			int numAttributes = xmlStreamReader.getAttributeCount();
			for(int i = 0; i < numAttributes; i++)
			{
				String localName = xmlStreamReader.getAttributeLocalName(i);
				switch(localName)
				{
					case "mttf":
						mttf = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					case "mttr":
						mttr = xmlStreamReader.getAttributeAsDouble(i);
						break;
						
					default:
						attributeMap.put(localName, xmlStreamReader.getAttributeValue(i));
				}
			}
			
			Set<Long> nodeIds_thisSRG = new LinkedHashSet<Long>();
			Set<Long> linkIds_thisSRG = new LinkedHashSet<Long>();

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "linkEntry":
								linkIds_thisSRG.add(xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id")));
								break;

							case "nodeEntry":
								nodeIds_thisSRG.add(xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id")));
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("srg"))
						{
							long srgId = netPlan.addSRG(mttf, mttr, attributeMap);
							for(long nodeId : nodeIds_thisSRG) netPlan.addNodeToSRG(nodeId, srgId);
							for(long linkId : linkIds_thisSRG) netPlan.addLinkToSRG(linkId, srgId);
							return;
						}
						break;
				}
			}

			throw new RuntimeException("'SRG' element not parsed correctly (end tag not found)");
		}

		protected void parseSRGInfo(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "srg":
								parseSRG(netPlan, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("srgInfo")) return;
						break;
				}
			}

			throw new RuntimeException("'SRG info' element not parsed correctly (end tag not found)");
		}
	}
	
	private static class NetPlanFormat_v2 implements INetPlanFormat
	{
		@Override
		public void create(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			parseNetwork(netPlan, xmlStreamReader);

			Set<Long> layerIds_thisNetPlan = netPlan.getLayerIds();
			Set<Long> nodeIds_thisNetPlan = netPlan.getNodeIds();
			Set<Long> srgIds_thisNetPlan = netPlan.getSRGIds();

			netPlan.layerDefaultId = LongUtils.minValue(layerIds_thisNetPlan);
			netPlan.nextLayerId = new MutableLong(LongUtils.maxValue(layerIds_thisNetPlan) + 1);
			if (!nodeIds_thisNetPlan.isEmpty()) netPlan.nextNodeId = new MutableLong(LongUtils.maxValue(nodeIds_thisNetPlan) + 1);
			if (!srgIds_thisNetPlan.isEmpty()) netPlan.nextSRGId = new MutableLong(LongUtils.maxValue(srgIds_thisNetPlan) + 1);

			for(long layerId : layerIds_thisNetPlan)
			{
				Set<Long> linkIds_thisLayer = netPlan.getLinkIds(layerId);
				if (!linkIds_thisLayer.isEmpty()) netPlan.nextLinkId.put(layerId, LongUtils.maxValue(linkIds_thisLayer) + 1);

				Set<Long> demandIds_thisLayer = netPlan.getDemandIds(layerId);
				if (!demandIds_thisLayer.isEmpty()) netPlan.nextDemandId.put(layerId, LongUtils.maxValue(demandIds_thisLayer) + 1);

				switch(netPlan.getRoutingType(layerId))
				{
					case SOURCE_ROUTING:
						Set<Long> routeIds_thisLayer = netPlan.getRouteIds(layerId);
						if (!routeIds_thisLayer.isEmpty()) netPlan.nextRouteId.put(layerId, LongUtils.maxValue(routeIds_thisLayer) + 1);

						Set<Long> segmentIds_thisLayer = netPlan.getProtectionSegmentIds(layerId);
						if (!segmentIds_thisLayer.isEmpty()) netPlan.nextSegmentId.put(layerId, LongUtils.maxValue(segmentIds_thisLayer) + 1);

						break;

					case HOP_BY_HOP_ROUTING:
						netPlan.nextRouteId.put(layerId, 0L);
						netPlan.nextSegmentId.put(layerId, 0L);
						break;

					default:
						throw new RuntimeException("Bad");
				}
			}
		}
		
		protected void parseDemand(NetPlan netPlan, long layerId, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long demandId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			long ingressNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "ingressNodeId"));
			long egressNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "egressNodeId"));
			double offeredTraffic = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "offeredTraffic"));
			netPlan.nextDemandId.put(layerId, demandId);
			netPlan.addDemand(layerId, ingressNodeId, egressNodeId, offeredTraffic, null);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setDemandAttribute(layerId, demandId, key, name);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("demand")) return;
						break;
				}
			}

			throw new RuntimeException("'Demand' element not parsed correctly (end tag not found)");
		}

		protected void parseLayer(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long layerId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			if (layerId == 0) netPlan.hasReadLayerZero = true;
			String demandTrafficUnitsName = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "demandTrafficUnitsName"));
			String layerDescription = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "description"));
			String layerName = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "name"));
			String linkCapacityUnitsName = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "linkCapacityUnitsName"));
			netPlan.nextLayerId = new MutableLong(layerId);
			netPlan.addLayer(layerName, layerDescription, linkCapacityUnitsName, demandTrafficUnitsName, null);
			
			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setLayerAttribute(layerId, key, name);
								break;

							case "demand":
								parseDemand(netPlan, layerId, xmlStreamReader);
								break;

							case "link":
								parseLink(netPlan, layerId, xmlStreamReader);
								break;

							case "protectionSegment":
								parseProtectionSegment(netPlan, layerId, xmlStreamReader);
								break;

							case "route":
								parseRoute(netPlan, layerId, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("layer")) return;
						break;
				}
			}

			throw new RuntimeException("'Layer' element not parsed correctly (end tag not found)");
		}

		protected void parseLink(NetPlan netPlan, long layerId, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long linkId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			long originNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "originNodeId"));
			long destinationNodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "destinationNodeId"));
			double capacity = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "capacity"));
			double lengthInKm = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "lengthInKm"));
			double propagationSpeedInKmPerSecond = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "propagationSpeedInKmPerSecond"));
			netPlan.nextLinkId.put(layerId, linkId);
			netPlan.addLink(layerId, originNodeId, destinationNodeId, capacity, lengthInKm, propagationSpeedInKmPerSecond, null);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();
				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setLinkAttribute(layerId, linkId, key, name);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("link")) return;
						break;
				}
			}

			throw new RuntimeException("'Link' element not parsed correctly (end tag not found)");
		}

		protected void parseNetwork(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			String networkDescription_thisNetPlan = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "description"));
			String networkName_thisNetPlan = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "name"));
			netPlan.setNetworkDescription(networkDescription_thisNetPlan);
			netPlan.setNetworkName(networkName_thisNetPlan);

			netPlan.hasReadLayerZero = false;

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();
				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setNetworkAttribute(key, name);
								break;

							case "layer":
								parseLayer(netPlan, xmlStreamReader);
								break;

							case "node":
								parseNode(netPlan, xmlStreamReader);
								break;

							case "srg":
								parseSRG(netPlan, xmlStreamReader);
								break;

							case "layerCoupling":
								long upperLayerLinkId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "upperLayerLinkId"));
								long upperLayerId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "upperLayerId"));
								long lowerLayerDemandId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "lowerLayerDemandId"));
								long lowerLayerId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "lowerLayerId"));
								netPlan.coupleDemandWithUpperLayerLink(lowerLayerId, lowerLayerDemandId, upperLayerId, upperLayerLinkId);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("network"))
						{
							if (!netPlan.hasReadLayerZero && netPlan.getNumberOfLayers() > 1) netPlan.removeLayer(0);
							return;
						}

						break;
				}
			}

			throw new RuntimeException("'Network' element not parsed correctly (end tag not found)");
		}

		protected void parseNode(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long nodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			double xCoord = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "xCoord"));
			double yCoord = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "yCoord"));
			String nodeName = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "name"));
			netPlan.nextNodeId = new MutableLong(nodeId);
			netPlan.addNode(xCoord, yCoord, nodeName, null);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();
				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setNodeAttribute(nodeId, key, name);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("node")) return;
						break;
				}
			}

			throw new RuntimeException("'Node' element not parsed correctly (end tag not found)");
		}

		protected void parseProtectionSegment(NetPlan netPlan, long layerId, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long segmentId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			double reservedCapacity = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "reservedCapacity"));
			List<Long> seqLinks = LongUtils.toList(xmlStreamReader.getAttributeAsLongArray(xmlStreamReader.getAttributeIndex(null, "seqLinks")));
			netPlan.nextSegmentId.put(layerId, segmentId);
			netPlan.addProtectionSegment(layerId, seqLinks, reservedCapacity, null);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setProtectionSegmentAttribute(layerId, segmentId, key, name);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("protectionSegment")) return;
						break;
				}
			}

			throw new RuntimeException("'Protection segment' element not parsed correctly (end tag not found)");
		}

		protected void parseRoute(NetPlan netPlan, long layerId, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long routeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			long demandId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "demandId"));
			double carriedTraffic = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "carriedTraffic"));
			double occupiedCapacity = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "occupiedCapacity"));
			List<Long> seqLinks = LongUtils.toList(xmlStreamReader.getAttributeAsLongArray(xmlStreamReader.getAttributeIndex(null, "seqLinks")));
			List<Long> backupSegmentList = LongUtils.toList(xmlStreamReader.getAttributeAsLongArray(xmlStreamReader.getAttributeIndex(null, "backupSegmentList")));
			netPlan.nextRouteId.put(layerId, routeId);
			netPlan.addRoute(layerId, demandId, carriedTraffic, occupiedCapacity, seqLinks, null);
			for(long segmentId : backupSegmentList) netPlan.addProtectionSegmentToRouteBackupSegmentList(layerId, segmentId, routeId);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setRouteAttribute(layerId, routeId, key, name);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("route")) return;
						break;
				}
			}

			throw new RuntimeException("'Route' element not parsed correctly (end tag not found)");
		}

		protected void parseSRG(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long srgId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			double meanTimeToFailInHours = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "meanTimeToFailInHours"));
			double meanTimeToRepairInHours = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "meanTimeToRepairInHours"));
			netPlan.nextSRGId = new MutableLong(srgId);
			netPlan.addSRG(meanTimeToFailInHours, meanTimeToRepairInHours, null);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setSRGAttribute(srgId, key, name);
								break;

							case "node":
								long nodeId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
								netPlan.addNodeToSRG(nodeId, srgId);
								break;

							case "link":
								long layerId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "layerId"));
								long linkId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "linkId"));
								netPlan.addLinkToSRG(layerId, linkId, srgId);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("srg")) return;
						break;
				}
			}

			throw new RuntimeException("'SRG' element not parsed correctly (end tag not found)");
		}
	}
	
	private static class NetPlanFormat_v3 extends NetPlanFormat_v2
	{
		@Override
		protected void parseLayer(NetPlan netPlan, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long layerId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "id"));
			if (layerId == 0) netPlan.hasReadLayerZero = true;
			String demandTrafficUnitsName = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "demandTrafficUnitsName"));
			String layerDescription = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "description"));
			String layerName = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "name"));
			String linkCapacityUnitsName = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "linkCapacityUnitsName"));
			netPlan.nextLayerId = new MutableLong(layerId);
			netPlan.addLayer(layerName, layerDescription, linkCapacityUnitsName, demandTrafficUnitsName, null);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setLayerAttribute(layerId, key, name);
								break;

							case "demand":
								parseDemand(netPlan, layerId, xmlStreamReader);
								break;

							case "link":
								parseLink(netPlan, layerId, xmlStreamReader);
								break;

							case "hopByHopRouting":
								parseHopByHopRouting(netPlan, layerId, xmlStreamReader);
								break;

							case "sourceRouting":
								parseSourceRouting(netPlan, layerId, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad child (" + startElementName + ") for layer element");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("layer")) return;
						break;
				}
			}

			throw new RuntimeException("'Layer' element not parsed correctly (end tag not found)");
		}

		protected void parseForwardingRule(NetPlan netPlan, long layerId, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			long linkId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "linkId"));
			long demandId = xmlStreamReader.getAttributeAsLong(xmlStreamReader.getAttributeIndex(null, "demandId"));
			double splittingRatio = xmlStreamReader.getAttributeAsDouble(xmlStreamReader.getAttributeIndex(null, "splittingRatio"));
			Pair<Long, Long> demandLinkPair = netPlan.addForwardingRule(layerId, demandId, linkId, splittingRatio, null);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "attribute":
								String key = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "key"));
								String name = xmlStreamReader.getAttributeValue(xmlStreamReader.getAttributeIndex(null, "value"));
								netPlan.setForwardingRuleAttribute(layerId, demandLinkPair, key, name);
								break;

							default:
								throw new RuntimeException("Bad");
						}

						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("forwardingRule")) return;
						break;
				}
			}

			throw new RuntimeException("'Forwarding rule' element not parsed correctly (end tag not found)");
		}

		protected void parseHopByHopRouting(NetPlan netPlan, long layerId, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			netPlan.setRoutingType(layerId, RoutingType.HOP_BY_HOP_ROUTING);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "forwardingRule":
								parseForwardingRule(netPlan, layerId, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("hopByHopRouting")) return;
						break;
				}
			}

			throw new RuntimeException("'Hop-by-hop routing' element not parsed correctly (end tag not found)");
		}

		protected void parseSourceRouting(NetPlan netPlan, long layerId, XMLStreamReader2 xmlStreamReader) throws XMLStreamException
		{
			netPlan.setRoutingType(layerId, RoutingType.SOURCE_ROUTING);

			while(xmlStreamReader.hasNext())
			{
				xmlStreamReader.next();

				switch(xmlStreamReader.getEventType())
				{
					case XMLEvent.START_ELEMENT:
						String startElementName = xmlStreamReader.getName().toString();
						switch(startElementName)
						{
							case "protectionSegment":
								parseProtectionSegment(netPlan, layerId, xmlStreamReader);
								break;

							case "route":
								parseRoute(netPlan, layerId, xmlStreamReader);
								break;

							default:
								throw new RuntimeException("Bad");
						}
						break;

					case XMLEvent.END_ELEMENT:
						String endElementName = xmlStreamReader.getName().toString();
						if (endElementName.equals("sourceRouting")) return;
						break;
				}
			}

			throw new RuntimeException("'Source routing' element not parsed correctly (end tag not found)");
		}
	}
}